Package: mqriskR
Type: Package
Title: Actuarial Risk Modeling and Life Contingencies
Version: 0.1.0
Authors@R: c(
    person(
      "Nii", "Okine",
      email = "okinean@appstate.edu",
      role = c("aut", "cre")
    )
  )
Description: Provides functions for actuarial risk modeling, including survival models,
    life annuities, multiple-decrement models, and mortality improvement projections.
    The package is designed to align with standard actuarial notation and supports
    teaching, exam preparation, and reproducible actuarial analysis. The methods are
    based on standard actuarial references including Camilli, Duncan and London (2014,
    ISBN:9781625423474) "Models for Quantifying Risk" and Dickson, Hardy and Waters
    (2020, ISBN:9781108478083) "Actuarial Mathematics for Life Contingent Risks".
License: MIT + file LICENSE
Encoding: UTF-8
Suggests: ggplot2, testthat (>= 3.0.0), expm, knitr, rmarkdown
Config/testthat/edition: 3
Imports: stats, utils
RoxygenNote: 7.3.3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-23 12:22:07 UTC; okinean
Author: Nii Okine [aut, cre]
Maintainer: Nii Okine <okinean@appstate.edu>
Repository: CRAN
Date/Publication: 2026-04-27 14:40:02 UTC
Built: R 4.5.2; ; 2026-05-01 17:17:00 UTC; unix

Package: mqriskR
Type: Package
Title: Actuarial Risk Modeling and Life Contingencies
Version: 0.1.1
Authors@R: c(
    person(
      "Nii", "Okine",
      email = "okinean@appstate.edu",
      role = c("aut", "cre")
    )
  )
Description: Provides functions for actuarial risk modeling, including survival models,
    life annuities, multiple-decrement models, and mortality improvement projections.
    The package is designed to align with standard actuarial notation and supports
    teaching, exam preparation, and reproducible actuarial analysis. The methods are
    based on standard actuarial references including Camilli, Duncan and London (2014,
    ISBN:9781625423474) "Models for Quantifying Risk" and Dickson, Hardy and Waters
    (2020, ISBN:9781108478083) "Actuarial Mathematics for Life Contingent Risks".
License: MIT + file LICENSE
Encoding: UTF-8
Suggests: ggplot2, testthat (>= 3.0.0), expm, knitr, rmarkdown
Config/testthat/edition: 3
Imports: stats, utils
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-27 22:56:17 UTC; okinean
Author: Nii Okine [aut, cre]
Maintainer: Nii Okine <okinean@appstate.edu>
Repository: CRAN
Date/Publication: 2026-07-28 00:10:01 UTC
Built: R 4.5.2; ; 2026-07-28 02:24:12 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,life_table)
S3method(print,select_life_table)
export(A2barx)
export(A2barxn)
export(A2barxn1)
export(A2nAbarx)
export(A2nAx)
export(A2nAx_m)
export(A2nEx)
export(A2x)
export(A2x_m)
export(A2xn)
export(A2xn1)
export(A2xn1_m)
export(A2xn_m)
export(AAL_PUC_db)
export(AAL_TUC_db)
export(AB_cae)
export(AB_fas)
export(APV_NR_db)
export(APV_gross_premiums)
export(AS_path)
export(AS_path_md)
export(AV_path_ul_typeA)
export(AV_path_ul_typeB)
export(AVz_dc)
export(Abarx)
export(Abarx_udd)
export(Abarxj_md)
export(Abarxn)
export(Abarxn1)
export(Abarxn1_udd)
export(Abarxn_udd)
export(Abarxy)
export(Abarxy1)
export(Abarxy2)
export(Abarxybar)
export(Abaryx1)
export(Abaryx2)
export(Ax)
export(Ax_m)
export(Ax_m_udd)
export(Axj_md)
export(Axn)
export(Axn1)
export(Axn1_m)
export(Axn1_m_udd)
export(Axn1_spot)
export(Axn1_var)
export(Axn_m)
export(Axn_m_udd)
export(Axn_spot)
export(Axn_var)
export(Axy)
export(Axybar)
export(Axybarn)
export(Axybarn1)
export(Axyn)
export(Axyn1)
export(DAbarxn1)
export(DAxn1)
export(Dabarxn)
export(Dadotxn)
export(Daxn)
export(DbarAbarxn1)
export(EL0barAbarx)
export(EL0x)
export(EL0xn)
export(EL0xn1)
export(ELtx)
export(F0)
export(GI_cont)
export(GI_disc)
export(GMF_rollforward_ul)
export(GM_cont)
export(GM_disc)
export(GT_cont)
export(GT_disc)
export(GTg_disc)
export(Gx)
export(IAbarx)
export(IAx)
export(IAxn1)
export(IRR_profit)
export(Iabarx)
export(Iabarxn)
export(Iadotx)
export(Iadotxn)
export(Iax)
export(Iaxn)
export(IbarAbarx)
export(IbarAbarxn1)
export(Income_dc)
export(NC_EAN_db)
export(NC_PUC_db)
export(NC_TUC_db)
export(NPV_partial)
export(NPV_profit)
export(PAB_cae)
export(PAB_fas)
export(PbarAbarx)
export(PbarAbarxn)
export(PbarAbarxn1)
export(Pbar_trapz_ms)
export(Pbarx)
export(Pbarxn)
export(Pbarxn1)
export(Pi_signature)
export(PnAdotx)
export(PnAx)
export(PnAx_m)
export(PnEx)
export(Pnax)
export(Pr_vector_disc)
export(Px)
export(Px_m)
export(Pxn)
export(Pxn1)
export(Pxn1_m)
export(Pxn_m)
export(S0)
export(S0_to_lx)
export(V_zeroized)
export(Vprefloor_crvm_ul)
export(abarx)
export(abarx_udd)
export(abarx_woolhouse2)
export(abarx_woolhouse3)
export(abarx_y)
export(abarxn)
export(abarxn_udd)
export(abarxy)
export(abarxybar)
export(abary_x)
export(adotx)
export(adotx_m)
export(adotx_m_udd)
export(adotx_m_woolhouse2)
export(adotx_m_woolhouse3)
export(adotxn)
export(adotxn_m)
export(adotxn_m_udd)
export(adotxn_m_woolhouse2)
export(adotxn_m_woolhouse3)
export(adotxy)
export(adotxybar)
export(adotxybarn)
export(adotxyn)
export(ag38_prefunding_ratio)
export(ag38_reserve_ul)
export(alphaF)
export(annuity_certain)
export(annuity_identity_abarx)
export(annuity_identity_abarxn)
export(annuity_identity_adotx)
export(annuity_identity_adotxn)
export(annuity_identity_ax)
export(annuity_identity_axn)
export(annuity_identity_nabarx)
export(annuity_identity_nadotx)
export(annuity_identity_nax)
export(ax)
export(ax_improved)
export(ax_m)
export(ax_m_udd)
export(ax_m_woolhouse2)
export(ax_m_woolhouse3)
export(ax_y)
export(axn)
export(axn_improved)
export(axn_m)
export(axn_m_udd)
export(axn_m_woolhouse2)
export(axn_m_woolhouse3)
export(axn_spot)
export(axn_var)
export(axy)
export(axybar)
export(axybarn)
export(axyn)
export(ay_x)
export(betaF)
export(coi_ul_typeB)
export(contribution_rate_target)
export(cov_term_deferred)
export(cov_term_endow)
export(cumhaz0)
export(decompGg_disc)
export(discount)
export(discounted_payback_period)
export(double_force_delta)
export(double_force_i)
export(dx)
export(dxj)
export(dxtau)
export(ex_complete)
export(ex_complete_tab)
export(ex_curtate)
export(ex_curtate_tab)
export(ex_temp_complete_tab)
export(ex_temp_curtate_tab)
export(f0)
export(fnk_from_z)
export(forward_matrix_from_z)
export(fx)
export(fx_tab)
export(gain_loss_md)
export(hazard0)
export(htVnAx)
export(htVx)
export(iMA_eiul)
export(iP_eiul)
export(i_credit_eiul)
export(interest_convert)
export(life_table)
export(lx)
export(lx_select)
export(lx_to_S0)
export(markov_nstep_prob)
export(md_table)
export(meanVx)
export(mux_tab)
export(nAbarx)
export(nAbarx_udd)
export(nAx)
export(nAx_m)
export(nAx_m_udd)
export(nEx)
export(nEx_spot)
export(nEx_var)
export(nExy)
export(nExybar)
export(nabarx)
export(nabarx_udd)
export(nadotx)
export(nadotx_m)
export(nadotx_m_udd)
export(nadotx_m_woolhouse2)
export(nadotx_m_woolhouse3)
export(nax)
export(nax_m)
export(nax_m_udd)
export(nax_m_woolhouse2)
export(nax_m_woolhouse3)
export(naxn_improved)
export(ndx)
export(nkqx)
export(nmxq)
export(nmxq_select)
export(npx)
export(npx_select)
export(npxtau_md)
export(nqx)
export(nqx_select)
export(nqxj_md)
export(nqxtau_md)
export(profit_margin)
export(pv_cashflows)
export(pv_spot_cashflows)
export(px_proj)
export(px_to_lx)
export(pxtau)
export(pxtau_ul)
export(qx_dep_cf)
export(qx_dep_sudd)
export(qx_proj)
export(qx_tab)
export(qx_to_lx)
export(qxprime_mudd)
export(qxprime_sudd)
export(qxtau)
export(replacement_ratio_db)
export(replacement_ratio_dc)
export(rt_ul)
export(salary_scale)
export(sbarxn)
export(sdotxn)
export(sdotxn_m)
export(sdotxn_m_udd)
export(sdotxn_m_woolhouse2)
export(select_life_table)
export(solve_yield)
export(sxn)
export(sxn_m)
export(sxn_m_udd)
export(sxn_m_woolhouse2)
export(tPnAx)
export(tPnEx)
export(tPx)
export(tPxn)
export(tPxn1)
export(tVEx)
export(tVFx)
export(tVGx)
export(tVbarAbarx)
export(tVbarx)
export(tVnAdotx)
export(tVnAx)
export(tVnEx)
export(tVnax)
export(tVx)
export(tVx_m)
export(tVx_ret)
export(tVxn)
export(tVxn1)
export(tVxn1_ret)
export(tVxn_ret)
export(thiele_backward_path)
export(thiele_backward_step)
export(thiele_dVdt)
export(thiele_dVdt_01)
export(thiele_path_01)
export(tp00_tp01_euler)
export(tpx)
export(tpx_improved)
export(tpx_tab)
export(tpx_tau_cf)
export(tpxprimej_cf)
export(tpxtau_ul)
export(tpxy)
export(tpxybar)
export(tqx)
export(tqx_tab)
export(tqxj_cf)
export(tqxprime_mudd)
export(tqxprimej_cf)
export(tqxy)
export(tqxy1)
export(tqxy2)
export(tqxybar)
export(tqyx1)
export(tqyx2)
export(tsVx)
export(tsVxn)
export(tsVxn1)
export(udd_continuous_multiplier)
export(udd_mthly_multiplier)
export(varL0barAbarx)
export(varL0x)
export(varL0xn)
export(varL0xn1)
export(varLtx)
export(var_Abarx)
export(var_Abarxn)
export(var_Abarxn1)
export(var_Ax)
export(var_Ax_m)
export(var_Axn)
export(var_Axn1)
export(var_Axn1_m)
export(var_Axn_m)
export(var_nAbarx)
export(var_nAx)
export(var_nAx_m)
export(var_nEx)
export(vt_var)
export(z_from_coupon_annual)
export(z_from_coupon_semi)
export(z_from_fn1)
importFrom(stats,integrate)
importFrom(stats,uniroot)
importFrom(utils,head)
importFrom(utils,tail)

# mqriskR

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/mqriskR)](https://CRAN.R-project.org/package=mqriskR)
<!-- badges: end -->

**mqriskR** provides a comprehensive collection of functions for actuarial mathematics and life contingency modeling. It supports actuarial present values, premiums, reserves, pensions, multiple-life models, multiple-decrement models, mortality improvement, and both life table and parametric survival model calculations using standard actuarial notation.


The package is designed to support actuarial education, professional exam preparation, research, and reproducible actuarial analysis.



## Installation

Install the stable release from CRAN:

```r
install.packages("mqriskR")
```

Or install the development version from GitHub:

```r
# install.packages("remotes")
remotes::install_github("aokine/mqriskR")
```

## Example

Compute the actuarial present value of a whole life annuity under a uniform distribution of deaths.

```r
library(mqriskR)

ax(
  x = 40,
  i = 0.05,
  model = "uniform",
  omega = 100
)
```

## Main Features

- Life insurance present values
- Life annuities (discrete, continuous, and m-thly)
- Premium calculations
- Policy reserve calculations
- Pension mathematics
- Survival probabilities
- Multiple-life models
- Multiple-decrement models
- Mortality improvement projections
- Spot interest rate models
- Variable interest models
- Support for both life tables and parametric survival models
- Functions using standard actuarial notation

## Version 0.1.1

Version 0.1.1 focuses on improving the quality, consistency, and usability of the package.

Highlights include:

- Improved support for life table objects
- Expanded vectorized input support
- Improved handling of finite life tables
- More consistent behavior across related functions
- Cleaner, descriptive documentation
- Improved examples and package manual
- Enhanced input validation and error handling

No breaking changes were introduced. Existing code written for earlier versions of **mqriskR** continues to work.

## Intended Audience

The package is intended for:

- Actuarial students preparing for professional examinations
- University instructors teaching actuarial mathematics
- Researchers developing actuarial methods
- Practicing actuaries requiring transparent and reproducible calculations

## Documentation

Complete documentation for all exported functions is available on the pkgdown website:

https://aokine.github.io/mqriskR/

## References

The methods implemented in **mqriskR** are based on standard actuarial references, including:

- Camilli, S. J., Duncan, I., and London, R. L. (2014). *Models for Quantifying Risk* (6th ed.). ACTEX Publications.

- Dickson, D. C. M., Hardy, M. R., and Waters, H. R. (2020). *Actuarial Mathematics for Life Contingent Risks* (2nd ed.). Cambridge University Press.

library(mqriskR)


# -------------------------
# Figure 1.1: accumulation function a(t) = (1+i)^t
# -------------------------
i <- 0.05
t <- 0:30
a_t <- (1 + i)^t


plot(t, a_t,
     xlab = "t", ylab = "a(t)",
     main = expression(paste("Accumulation function  ", a(t) == (1+i)^t)))




# ----------------------------
# Level annuities:  m-thly,
# ----------------------------

# m-thly (e.g., monthly payments of 1/12 for n years)
a_monthly <- annuity_certain(n = 10, i = 0.05, m = 12, due = FALSE)
a_monthly


# ----------------------------
# Non-level payments (vector approach)
# ----------------------------
# Arithmetic increasing annuity-immediate: 1,2,...,n paid at t=1,...,n
n <- 10
pmt_arith_inc <- 1:n
t_arith_inc <- 1:n
pv_arith_inc <- pv_cashflows(cf = pmt_arith_inc, t = t_arith_inc, i = i)
pv_arith_inc

# Geometric annuity-immediate: 1, r, r^2, ..., r^(n-1) at t=1,...,n
r <- 1.03
pmt_geom <- r^(0:(n-1))
t_geom <- 1:n
pv_geom <- pv_cashflows(cf = pmt_geom, t = t_geom, i = i)
pv_geom

# Plot payment patterns
plot(1:n, pmt_arith_inc, type = "b",
     xlab = "Payment number", ylab = "Payment",
     main = "Non-level payment patterns")
lines(1:n, pmt_geom, type = "b")
legend("topleft", legend = c("Arithmetic increasing", "Geometric (r=1.03)"),
       lty = 1, pch = 1, bty = "n")

## Listing 12.1: Joint-life survival probability
library(mqriskR)

x <- 50; y <- 60; t <- 10

px <- tpx(x = x,t = t,model = "uniform",omega = 100)

py <- tpx(x = y,t = t,model = "uniform",omega = 100)

joint_from_product <- px * py

joint_from_package <- tpxy(x = x,y = y,t = t,model = "uniform",omega = 100)

c(px = px,py = py,
  joint_from_product = joint_from_product,
  joint_from_package = joint_from_package)




## Listing 12.2: Joint-life vs last-survivor survival
library(mqriskR)

x <- 50; y <- 60; t <- 10

joint_survival <- tpxy(x, y, t, model = "uniform", omega = 100)
last_survival  <- tpxybar(x, y, t, model = "uniform", omega = 100)

c(joint = joint_survival,
  last_survivor = last_survival)



##Listing 12.3: Joint-life and last-survivor annuity values
library(mqriskR)

x <- 50; y <- 60; i <- 0.05

joint_annuity <- adotxy(x = x,y = y,i = i,model = "uniform",omega = 100)

last_annuity <- adotxybar(x = x,y = y,i = i,model = "uniform",omega = 100)

c(joint_annuity = joint_annuity,
  last_survivor_annuity = last_annuity)




##Listing 12.4: Reversionary annuity to $(y)$ after death of $(x)$
library(mqriskR)

x <- 50; y <- 60; i <- 0.05

reversionary_annuity <- ax_y(x = x,y = y,i = i,model = "uniform", omega = 100)

check_value <- ax(x = y,i = i,model = "uniform",omega = 100) - axy(x = x, y = y, i = i, model = "uniform", omega = 100)

c(reversionary_annuity = reversionary_annuity,
  check_value = check_value)


##Listing 12.5: Continuous contingent insurance values}, label={lst:ch12_contingent_insurance}]
library(mqriskR)

x <- 50; y <- 60; i <- 0.05

benefit_if_x_before_y <- Abarxy1(x = x, y = y, i = i, model = "uniform", omega = 100)

benefit_if_x_after_y <- Abarxy2(x = x, y = y, i = i, model = "uniform", omega = 100)

c(x_before_y = benefit_if_x_before_y,
  x_after_y = benefit_if_x_after_y)

## Listing 10.1 Prospective reserves for several insurance contracts
library(mqriskR)

x <- 40
n <- 20
t <- 10
i <- 0.05

whole_life <- tVx(
  x = x,
  t = t,
  i = i,
  model = "uniform",
  omega = 100
)

term_insurance <- tVxn1(
  x = x,
  n = n,
  t = t,
  i = i,
  model = "uniform",
  omega = 100
)

endowment_insurance <- tVxn(
  x = x,
  n = n,
  t = t,
  i = i,
  model = "uniform",
  omega = 100
)

c(
  whole_life = whole_life,
  term_insurance = term_insurance,
  endowment_insurance = endowment_insurance
)


## Listing 10.2: Comparison of prospective and retrospective whole life reserves
library(mqriskR)

x <- 40; t <- 10; i <- 0.05

prospective <- tVx(x = x,t = t,i = i,model = "uniform",omega = 100)

retrospective <- tVx_ret(x = x,t = t,i = i,model = "uniform", omega = 100)

c(prospective = prospective,
  retrospective = retrospective,
  difference = prospective - retrospective)




## Listing 10.3: Equivalent formulas for the whole life reserve
library(mqriskR)

x <- 40; t <- 10; i <- 0.05

reserve_direct <- tVx(x = x,t = t,i = i,model = "uniform", omega = 100)

reserve_annuity_ratio <- 1 - adotx(x = x + t,i = i,model = "uniform",omega = 100) / adotx(x = x,i = i,model = "uniform", omega = 100)

reserve_premium_diff <- (Px(x = x + t,i = i,model = "uniform",omega = 100) -Px(x = x,i = i,model = "uniform",omega = 100)) * adotx(x = x + t,i = i,model = "uniform",omega = 100)

c(
  direct = reserve_direct,
  annuity_ratio = reserve_annuity_ratio,
  premium_difference = reserve_premium_diff
)


## Listing 10.4: Conditional mean and variance of loss at duration \(t\)
library(mqriskR)

x <- 40
t <- 10
i <- 0.05

premium <- Px(
  x = x,
  i = i,
  model = "uniform",
  omega = 100
)

reserve <- tVx(
  x = x,
  t = t,
  i = i,
  model = "uniform",
  omega = 100
)

mean_loss <- ELtx(
  x = x,
  t = t,
  i = i,
  P = premium,
  model = "uniform",
  omega = 100
)

var_loss <- varLtx(
  x = x,
  t = t,
  i = i,
  P = premium,
  model = "uniform",
  omega = 100
)

c(
  reserve = reserve,
  expected_loss = mean_loss,
  variance_loss = var_loss
)



## Listing 10.5: Reserve for a whole life insurance with immediate payment of claims
library(mqriskR)

x <- 40; t <- 10; i <- 0.05

reserve_val <- tVbarx(x = x,t = t,i = i,model = "uniform", omega = 100)

reserve_val



## Listing 10.6: Comparison of annual and monthly whole life reserves
library(mqriskR)

x <- 40; t <- 10; i <- 0.05

reserve_annual <- tVx(x = x,t = t,i = i,model = "uniform",omega = 100)

reserve_monthly <- tVx_m(x = x,t = t,m = 12,i = i, model = "uniform",omega = 100)

c(annual = reserve_annual,
  monthly = reserve_monthly)


## Listing 10.7: Gain decomposition for a discrete insurance contract
library(mqriskR)

Vt <- 0.085697; Vt1 <- 0.096115
P <- 0.008013; B <- 1
i_assumed <- 0.06; i_actual <- 0.065
q_assumed <- 0.00356; q_actual <- 0.00300


gt <- GT_disc(Vt = Vt, Vt1 = Vt1, P = P, i_actual = i_actual, q_actual = q_actual, B = B )

gm <- GM_disc(Vt = Vt, Vt1 = Vt1, P = P, i_assumed = i_assumed, q_actual = q_actual, B = B)

gi <- GI_disc( Vt = Vt, Vt1 = Vt1, P = P, i_actual = i_actual, q_assumed = q_assumed, B = B)

round(c(
  total_gain = gt,
  mortality_gain = gm,
  interest_gain = gi,
  check = gm + gi
), 7)


## Listing 10.8: Backward Euler approximation of a fully continuous reserve path
library(mqriskR)

times <- seq(19, 20, by = 0.25)

reserve_path <- thiele_backward_path(
  times = times,
  V_terminal = 1000,
  P = 26.96,
  delta = 0.0582689,
  mu = 0.00197,
  benefit = 1000
)

data.frame(
  time = times,
  reserve = reserve_path
)

##Listing 18.1:Target contribution rate in a defined contribution plan
library(mqriskR)

c_target <- contribution_rate_target(
  x = 30, z = 65, Sx = 60000,
  RR_target = 0.60,
  i = 0.06, adue_z = 11,
  g = 0.04)

RR_check <- replacement_ratio_dc(
  x = 30, z = 65, Sx = 60000,
  c = c_target, i = 0.06,
  adue_z = 11, g = 0.04)

c(c_target = c_target,
  RR_check = RR_check)


##Listing 18.2: Defined contribution accumulation and replacement ratio
library(mqriskR)

AV_65 <- AVz_dc(
  x = 30, z = 65, Sx = 50000,
  c = 0.10, i = 0.05, g = 0.04
)

Income_65 <- Income_dc(
  AVz = AV_65,
  adue_z = 12
)

RR_65 <- replacement_ratio_dc(
  x = 30, z = 65, Sx = 50000,
  c = 0.10, i = 0.05,
  adue_z = 12, g = 0.04
)

c(
  AV_65 = AV_65,
  Income_65 = Income_65,
  RR_65 = RR_65
)






##Listing 18.3: Projected annual benefit under a final average salary plan
library(mqriskR)

PAB_65 <- PAB_fas(
  x = 35, z = 65, CASx = 60000,
  p = 2, fas_years = 3, g = 0.04
)

final_salary <- 60000 * (1.04)^29

RR_final_salary <- replacement_ratio_db(
  benefit = PAB_65,
  salary = final_salary
)

c(PAB_65 = PAB_65,
  final_salary = final_salary,
  RR_final_salary = RR_final_salary)

## Figure 5.1

library(mqriskR)

# Time grid (avoid the uniform model's endpoint to prevent division-by-zero in the hazard)
t_grid <- seq(0, 100, length.out = 501)
t_grid_u <- t_grid[t_grid < 100]   # for uniform hazard where omega = 100

# --- Choose parameters (adjust as desired for your narrative) ---
omega <- 100

lambda <- 0.03          # exponential
B_g <- 0.0005; c_g <- 1.08        # Gompertz
A_m <- 0.0002; B_m <- 0.0004; c_m <- 1.08   # Makeham

# --- Survival functions ---
S_u <- S0(t_grid, model = "uniform", omega = omega)
S_e <- S0(t_grid, model = "exponential", lambda = lambda)
S_g <- S0(t_grid, model = "gompertz", B = B_g, c = c_g)
S_m <- S0(t_grid, model = "makeham", A = A_m, B = B_m, c = c_m)

# --- Hazard functions ---
h_u <- hazard0(t_grid_u, model = "uniform", omega = omega)
h_e <- hazard0(t_grid,   model = "exponential", lambda = lambda)
h_g <- hazard0(t_grid,   model = "gompertz", B = B_g, c = c_g)
h_m <- hazard0(t_grid,   model = "makeham", A = A_m, B = B_m, c = c_m)

# Side-by-side plots
op <- par(mfrow = c(1, 2), mar = c(4, 4, 3, 1))

# ---- Panel 1: survival functions ----
plot(t_grid, S_u, type = "l", lty = 1,
     xlab = "t (time / age)", ylab = expression(S[0](t) == Pr(T[0] > t)),
     main = expression(paste("Survival functions: ", S[0](t))), ylim = c(0, 1))
lines(t_grid, S_e, lty = 2)
lines(t_grid, S_g, lty = 3)
lines(t_grid, S_m, lty = 4)
#legend("topright",
#       legend = c("Uniform (de Moivre)", "Exponential", "Gompertz", "Makeham"),
#       lty = 1:4, bty = "n", cex = 0.85)

# ---- Panel 2: hazard functions ----
plot(t_grid_u, h_u, type = "l", lty = 1,
     xlab = "t (time / age)", ylab = expression(lambda[0](t)),
     main = expression("Hazard Functions: " * lambda[0](t)))
lines(t_grid, h_e, lty = 2)
lines(t_grid, h_g, lty = 3)
lines(t_grid, h_m, lty = 4)
legend("topleft",
       legend = c("Uniform (de Moivre)", "Exponential", "Gompertz", "Makeham"),
       lty = 1:4, bty = "n", cex = 0.85)

par(op)

## In-Text Examples
## Example 5.2 (PDF and HRF)
# Example 5.2 verification:
# F0(t) = 1 - 0.10 * sqrt(100 - t),  0 <= t <= 100

F0 <- function(t) {
  out <- 1 - 0.10 * sqrt(pmax(100 - t, 0))
  # enforce support
  out[t < 0] <- 0
  out[t > 100] <- 1
  out
}

S0 <- function(t) 1 - F0(t)

# Closed-form results from the solution:
f0_closed <- function(t) ifelse(t >= 0 & t < 100, 0.05 * (100 - t)^(-1/2), 0)
haz_closed <- function(t) ifelse(t >= 0 & t < 100, 0.50 * (100 - t)^(-1), Inf)

# Numerical derivative check for f0(t) = d/dt F0(t)
f0_num <- function(t, h = 1e-6) (F0(t + h) - F0(t - h)) / (2*h)

t_test <- c(0, 1, 10, 30, 75, 99)

cbind(
  t = t_test,
  f0_closed = f0_closed(t_test),
  f0_num = f0_num(t_test),
  haz_closed = haz_closed(t_test),
  haz_from_def = f0_closed(t_test) / S0(t_test)
)

# max absolute differences (ignoring endpoints issues)
max(abs(f0_closed(t_test) - f0_num(t_test)), na.rm = TRUE)
max(abs(haz_closed(t_test) - f0_closed(t_test) / S0(t_test)), na.rm = TRUE)


## Example 5.3 (mean and median)
# Example 5.3 verification for the same survival model as Example 5.2
S0 <- function(t) {
  out <- 0.10 * sqrt(pmax(100 - t, 0))
  out[t < 0] <- 1
  out[t > 100] <- 0
  out
}

# Verify E[T0] = integral_0^infty S0(t) dt = integral_0^100 S0(t) dt
E_num <- integrate(function(t) S0(t), lower = 0, upper = 100)$value
E_closed <- 200/3
c(E_num = E_num, E_closed = E_closed, abs_diff = abs(E_num - E_closed))

# Verify median y solves S0(y) = 0.5 -> y = 75
median_closed <- 75
S0(median_closed)

# Numerical solve for median as a check
median_num <- uniroot(function(y) S0(y) - 0.5, interval = c(0, 100))$root
c(median_num = median_num, median_closed = median_closed, abs_diff = abs(median_num - median_closed))



## Example 5.4 (conditional survival, density, and expectation)
# Example 5.4 verification for the model from Example 5.2
S0 <- function(t) {
  out <- 0.10 * sqrt(pmax(100 - t, 0))
  out[t < 0] <- 1
  out[t > 100] <- 0
  out
}

f0 <- function(t) ifelse(t >= 0 & t < 100, 0.05 * (100 - t)^(-1/2), 0)

# (a) _20 p_36 = S0(56)/S0(36)
p20_36_num <- S0(56) / S0(36)
p20_36_closed <- 0.82916  # from the solution (rounded)
c(p20_36_num = p20_36_num, p20_36_closed = p20_36_closed, abs_diff = abs(p20_36_num - p20_36_closed))

# (b) fx(t) = f0(36+t)/S0(36). Compare to closed-form 0.0625 / sqrt(64 - t)
fx_num <- function(t) f0(36 + t) / S0(36)
fx_closed <- function(t) 0.0625 / sqrt(64 - t)

t_test <- c(0, 1, 10, 20, 40, 63)
cbind(t = t_test, fx_num = fx_num(t_test), fx_closed = fx_closed(t_test), abs_diff = abs(fx_num(t_test) - fx_closed(t_test)))

# (c) e36^o = integral_0^infty _t p_36 dt = integral_0^64 S0(36+t)/S0(36) dt
tp36 <- function(t) S0(36 + t) / S0(36)
e36_num <- integrate(tp36, lower = 0, upper = 64)$value
e36_closed <- 128/3
c(e36_num = e36_num, e36_closed = e36_closed, abs_diff = abs(e36_num - e36_closed))



## Example 5.7 (validity checks for candidate survival functions)
# Example 5.7 verification: check S(0)=1, limit->0 (numerically), and non-increasing.

S1 <- function(t) exp(t - 0.70 * (2^t - 1))
S2 <- function(t) (1 + t)^(-2)
S3 <- function(t) exp(-t^2)

grid <- seq(0, 20, by = 0.01)

check_S <- function(Sfun, name) {
  S0_val <- Sfun(0)
  # monotonic: all differences <= 0 (allow tiny numerical slack)
  diffs <- diff(Sfun(grid))
  non_increasing <- all(diffs <= 1e-10)
  # "limit" check via large t
  tail_val <- Sfun(max(grid))
  list(
    name = name,
    S_at_0 = S0_val,
    approx_limit = tail_val,
    non_increasing = non_increasing,
    max_positive_diff = max(diffs)
  )
}

check_S(S1, "S1")
check_S(S2, "S2")
check_S(S3, "S3")

# Optional: visualize the functions
plot(grid, S1(grid), type="l", col="red", ylim=c(0,1.5),
     main="Candidate Survival Functions", ylab="S(t)", xlab="t")
lines(grid, S2(grid), col="blue")
lines(grid, S3(grid), col="darkgreen")
legend("topright", legend=c("S1","S2","S3"),
       col=c("red","blue","darkgreen"), lty=1)

# Optional: check derivative sign near 0 for S1 (as in the written solution)
h <- 1e-6
S1_prime0_num <- (S1(h) - S1(0)) / h
S1_prime0_num



## Example 5.8 (polynomial survival model)}, label={lst:ex5_8_verify}]
# Example 5.8:
# S0(t) = (9000 - 10 t - t^2)/9000 on 0 <= t <= omega

S0 <- function(t) (9000 - 10*t - t^2) / 9000

# (a) Find omega from S0(omega)=0
omega_num <- uniroot(function(u) S0(u), interval = c(0, 200))$root
omega_num

# (b) f0(30) = -S0'(30)
S0_prime <- function(t) -(10 + 2*t) / 9000
f0 <- function(t) -S0_prime(t)
f0(30)

# (c) f20(10) = f0(30)/S0(20)
f20_10 <- f0(30) / S0(20)
f20_10

# (d) lambda0(30) = f0(30)/S0(30)
haz0_30 <- f0(30) / S0(30)
haz0_30

# (e) lambda20(10 | T0>20) = hazard at age 30 (should match part d)
# Build conditional survival: S20(t) = S0(20+t)/S0(20)
S20 <- function(t) S0(20 + t) / S0(20)
# hazard for T20 at duration t: -S20'(t)/S20(t)
S20_prime <- function(t) S0_prime(20 + t) / S0(20)
haz20 <- function(t) -S20_prime(t) / S20(t)

haz20_10 <- haz20(10)
c(haz0_30 = haz0_30, haz20_10 = haz20_10, abs_diff = abs(haz0_30 - haz20_10))


## End of Chapter Exercises

############################################################
## Chapter 5: Optional R checks for selected end-of-chapter exercises
############################################################

## Helper: safe numeric integration wrapper
num_int <- function(f, lower, upper, ...) {
  integrate(f, lower = lower, upper = upper, ..., subdivisions = 2000L, rel.tol = 1e-10)$value
}

## Helper: finite-difference derivative
fd_deriv <- function(f, x, h = 1e-6) (f(x + h) - f(x - h)) / (2 * h)

############################################################
## Exercise 5-1: lambda0(t) = a + b t
############################################################
check_ex5_1 <- function(a = 0.02, b = 0.001) {

  lambda0 <- function(t) a + b * t

  # Survival reconstructed from hazard
  S0 <- function(t) exp(-(a * t + 0.5 * b * t^2))

  f0 <- function(t) S0(t) * lambda0(t)

  # Choose plotting grid
  grid <- seq(0, 150, length.out = 20001)

  # Numerical integration check (truncated)
  mass <- sum(f0(grid)) * (grid[2] - grid[1])

  # Numerical mode
  mode_num <- grid[which.max(f0(grid))]

  # Analytical mode
  mode_exact <- (-a + sqrt(b)) / b

  # Plot
  plot(grid, f0(grid), type = "l", lwd = 2,
       main = "Density f0(t) with Numerical and Analytical Mode",
       xlab = "t", ylab = "f0(t)")

  abline(v = mode_num, col = "red", lty = 2, lwd = 2)
  abline(v = mode_exact, col = "blue", lty = 3, lwd = 2)

  legend("topright",
         legend = c("Numerical Mode", "Analytical Mode"),
         col = c("red", "blue"),
         lty = c(2,3),
         lwd = 2)

  list(
    a = a,
    b = b,
    approx_mass = mass,
    mode_numeric = mode_num,
    mode_analytical = mode_exact
  )
}

print(check_ex5_1())




############################################################
## Exercise 5-2: S0(t) = a t^2 + b with E[T0]=60 -> omega=90, median
############################################################
check_ex5_2 <- function(omega = 90) {

  S0 <- function(t) 1 - (t^2) / (omega^2)

  # E[T0] = integral_0^omega S0(t) dt
  ET0 <- num_int(S0, 0, omega)

  # Median m: S0(m) = 0.5
  med <- uniroot(function(m) S0(m) - 0.5, lower = 0, upper = omega)$root

  list(
    omega = omega,
    ET0_numeric = ET0,
    median_numeric = med
  )
}

print(check_ex5_2())



############################################################
## Exercise 5-3: lambda0(t) = exp(-r t) is not valid
## because S0(t) -> exp(-1/r) > 0
############################################################

check_ex5_3 <- function(r = 0.5, t_max = 20) {

  S0 <- function(t) exp((exp(-r * t) - 1) / r)

  tgrid <- seq(0, t_max, length.out = 2000)
  Svals <- S0(tgrid)

  limit_theory <- exp(-1 / r)

  # Plot
  plot(tgrid, Svals, type = "l", lwd = 2,
       main = expression(paste("Survival Function ", S[0](t),
                               " with ", lambda[0](t) == e^{-r*t})),
       xlab = "t", ylab = expression(S[0](t)))

  abline(h = limit_theory, lty = 2)
  text(t_max * 0.6, limit_theory,
       labels = paste0("Limit = ", round(limit_theory, 4)),
       pos = 3)

  list(
    r = r,
    theoretical_limit = limit_theory,
    S_at_large_t = S0(t_max)
  )
}

print(check_ex5_3())



############################################################
## Exercise 5-4: modified de Moivre with gamma=4, e0 = omega/5
############################################################
check_ex5_4 <- function(omega = 80, gamma = 4) {

  S0 <- function(t) (1 - t / omega)^gamma

  e0_num <- num_int(S0, 0, omega)
  e0_theory <- omega / (gamma + 1)

  list(
    omega = omega, gamma = gamma,
    e0_numeric = e0_num,
    e0_theory = e0_theory,
    diff = e0_num - e0_theory
  )
}

print(check_ex5_4())




############################################################
## Exercise 5-7: exponential X1,X2; solve for lambdas from S_Y(2), S_Z(2)
############################################################
check_ex5_7 <- function(SY2 = 0.24, SZ2 = 0.86) {

  # From theory: SY(2) = exp(-2(l1+l2)) => l1+l2 = -(1/2) log(SY2)
  sum_l <- -(1/2) * log(SY2)

  # SZ(2) = exp(-2l1) + exp(-2l2) - exp(-2(l1+l2))
  # Let a=exp(-2l1), b=exp(-2l2), then:
  # ab = SY2,  a+b-ab = SZ2 => a+b = SZ2 + SY2
  s <- SZ2 + SY2

  # Solve a( s - a ) = SY2 => a^2 - s a + SY2 = 0
  disc <- s^2 - 4 * SY2
  a1 <- (s + sqrt(disc)) / 2
  a2 <- (s - sqrt(disc)) / 2

  # Convert back to lambdas
  lam_from_a <- function(a) -(1/2) * log(a)
  cand <- data.frame(
    a = c(a1, a2),
    lambda = c(lam_from_a(a1), lam_from_a(a2))
  )

  # identify which corresponds to lambda1 > lambda2:
  # smaller a means larger lambda
  a_small <- min(a1, a2)
  a_large <- max(a1, a2)
  lambda1 <- lam_from_a(a_small)
  lambda2 <- lam_from_a(a_large)

  list(
    sum_lambda = sum_l,
    candidates_for_exp_neg2lambda = cand,
    lambda1 = lambda1,
    lambda2 = lambda2,
    check_order = (lambda1 > lambda2)
  )
}



print(check_ex5_7())



############################################################
## Exercise 5-8: deferred probability m|q0 = S(m) - S(m+1)
## Plot + correct constancy/monotonicity checks (exclude boundary m=omega)
############################################################
check_ex5_8 <- function(omega = 40, lambda = 0.2, make_plot = TRUE, include_endpoint_in_preview = TRUE) {

  # For behavior checks we only use m = 0,...,omega-1 (intervals fully inside support)
  m_check <- 0:(omega - 1)

  # For display/plot we can optionally include m=omega (shows boundary artifact clearly)
  m_grid <- if (include_endpoint_in_preview) 0:omega else m_check

  # (a) Uniform on (0, omega): S(t) = 1 - t/omega on [0,omega], 0 after omega
  S_unif <- function(t) ifelse(t < 0, 1, ifelse(t <= omega, 1 - t / omega, 0))

  dq_unif <- S_unif(m_grid) - S_unif(m_grid + 1)
  dq_unif_check <- S_unif(m_check) - S_unif(m_check + 1)

  # (b) Exponential with rate lambda: S(t) = exp(-lambda t), t>=0
  S_exp <- function(t) exp(-lambda * pmax(t, 0))

  dq_exp <- S_exp(m_grid) - S_exp(m_grid + 1)
  dq_exp_check <- S_exp(m_check) - S_exp(m_check + 1)

  # (c) f0(t)=0.00125 t on [0,omega] => S(t)=1 - 0.000625 t^2 on [0,omega]
  # Clamp t to [0,omega] so S is well-defined at endpoints.
  S_linpdf <- function(t) {
    tt <- pmax(pmin(t, omega), 0)
    1 - 0.000625 * tt^2
  }

  dq_lin <- S_linpdf(m_grid) - S_linpdf(m_grid + 1)
  dq_lin_check <- S_linpdf(m_check) - S_linpdf(m_check + 1)

  preview <- data.frame(
    m = m_grid,
    dq_unif = dq_unif,
    dq_exp  = dq_exp,
    dq_lin  = dq_lin
  )

  if (make_plot) {
    matplot(
      preview$m,
      cbind(preview$dq_unif, preview$dq_exp, preview$dq_lin),
      type = "l", lty = 1:3, lwd = 2,
      xlab = "m",
      ylab = "m|q0 = S0(m) - S0(m+1)",
      main = "Exercise 5-8: Behavior of m|q0 across m"
    )
    legend(
      "topright",
      legend = c("Uniform", "Exponential", "f0(t)=0.00125 t on [0,omega]"),
      lty = 1:3, lwd = 2, bty = "n"
    )
    grid()
    if (include_endpoint_in_preview) {
      abline(v = omega, lty = 3)  # highlight boundary
    }
  }

  list(
    uniform_is_constant        = max(abs(diff(dq_unif_check))) < 1e-10,
    exponential_is_decreasing  = all(diff(dq_exp_check) <= 1e-10),
    linpdf_is_increasing       = all(diff(dq_lin_check) >= -1e-10),
    preview = preview
  )
}

# Example run:
out <- check_ex5_8(make_plot = TRUE)
out$uniform_is_constant
out$exponential_is_decreasing
out$linpdf_is_increasing
tail(out$preview)




############################################################
## Exercise 5-12: S0(t) = (9000 - 10t - t^2)/9000, compute q50 - mu50
############################################################
check_ex5_12 <- function() {

  S0 <- function(t) (9000 - 10 * t - t^2) / 9000
  Sprime <- function(t) (-10 - 2 * t) / 9000

  S50 <- S0(50); S51 <- S0(51)

  q50 <- (S50 - S51) / S50
  mu50 <- (-Sprime(50)) / S50

  list(
    S50 = S50, S51 = S51,
    q50 = q50,
    mu50 = mu50,
    diff = q50 - mu50
  )
}

print(check_ex5_12())

############################################################
## Exercise 5-15: mu_y = (80 - y)^(-1/2), median of T20 solves p20(t)=0.5
############################################################
check_ex5_15 <- function() {

  mu <- function(y) (80 - y)^(-1/2)

  # survival from age 20 for duration t:
  p20 <- function(t) {
    # integrate mu from 20 to 20+t
    val <- num_int(mu, 20, 20 + t)
    exp(-val)
  }

  t_med <- uniroot(function(t) p20(t) - 0.5, lower = 0, upper = 59.999)$root

  list(
    median_T20_numeric = t_med,
    p20_at_median = p20(t_med)
  )
}

print(check_ex5_15())


############################################################
## Exercise 5-18: select model S_[x](t;x)=1 - t/(40-x)
############################################################
check_ex5_18 <- function() {

  S_sel <- function(t, x) 1 - t / (40 - x)

  # (a) _4 p_[30]
  p4_30 <- S_sel(4, 30)

  # (b) e_[30] = integral_0^(40-30) S(t;30) dt
  e30 <- num_int(function(t) S_sel(t, 30), 0, 10)

  # (c) mu_[20]+t = -d/dt log S(t;20) = 1/(20 - t)
  mu20t <- function(t) 1 / (20 - t)

  tgrid <- seq(0, 19, by = 0.5)
  # numeric hazard from S:
  S20 <- function(t) S_sel(t, 20)
  mu_num <- -fd_deriv(function(u) log(S20(u)), tgrid)

  list(
    p4_30 = p4_30,
    e_30_numeric = e30,
    mu_identity_check = data.frame(t = tgrid, mu_theory = mu20t(tgrid), mu_numeric = mu_num)
  )
}

out518 <- check_ex5_18()
print(out518$p4_30)
print(out518$e_30_numeric)
head(out518$mu_identity_check)

## Listing 5.1: Computing $S_0(t)$, $F_0(t)$, $f_0(t)$, $\lambda_0(t)$, and $\Lambda_0(t)$
library(mqriskR)

# --- Uniform (de Moivre) model ---
omega <- 100
t <- c(0, 10, 25, 50, 99, 100)

S0(t, model = "uniform", omega = omega)
F0(t, model = "uniform", omega = omega)
f0(t, model = "uniform", omega = omega)
hazard0(t, model = "uniform", omega = omega)
cumhaz0(t, model = "uniform", omega = omega)

# --- Exponential (constant force) model ---
lambda <- 0.02
t2 <- c(0, 5, 10, 25, 50)

S0(t2, model = "exponential", lambda = lambda)
F0(t2, model = "exponential", lambda = lambda)
f0(t2, model = "exponential", lambda = lambda)
hazard0(t2, model = "exponential", lambda = lambda)
cumhaz0(t2, model = "exponential", lambda = lambda)



## Listing 5.2: Computing ${}_t p_x$, ${}_t q_x$, and $f_x(t)$

library(mqriskR)

omega <- 120
x <- 30
t <- c(0, 1, 5, 10, 20)

# Conditional survival and failure probabilities
tpx(t, x, model = "uniform", omega = omega)  # {}_t p_x
tqx(t, x, model = "uniform", omega = omega)  # {}_t q_x = 1 - {}_t p_x

# Conditional density f_x(t)
fx(t, x, model = "uniform", omega = omega)

# Verify the identity {}_t p_x = S0(x+t)/S0(x)
lhs <- tpx(t, x, model = "uniform", omega = omega)
rhs <- S0(x + t, model = "uniform", omega = omega) / S0(x, model = "uniform", omega = omega)
max(abs(lhs - rhs))



## Listing 5.3: Computing $\overset{\circ}{e}_x$ and $e_x$
library(mqriskR)

# Uniform model: omega is the limiting age
omega <- 120
x <- c(0, 30, 60, 100)

# Complete expectation of life (continuous-time)
ex_complete(x, model = "uniform", omega = omega)

# Curtate expectation of life (integer part of future lifetime)
ex_curtate(x, model = "uniform", omega = omega)

# Exponential model: lambda is the constant force
lambda <- 0.02
ex_complete(x, model = "exponential", lambda = lambda)
ex_curtate(x, model = "exponential", lambda = lambda)

## Listing 6.1: Converting between $S_0(x)$ and $l_x$
library(mqriskR)

S0_vals <- c(1.00000, 0.97408, 0.97259, 0.97160, 0.97082)

lx_vals <- S0_to_lx(S0_vals, radix = 100000)
lx_vals

lx_to_S0(lx_vals)


## Listing 6.2: Constructing a life table from one-year survival probabilities
library(mqriskR)

tbl <- life_table(
  x = 0:4,
  px = c(0.90, 0.80, 0.60, 0.30, 0.00),
  radix = 10000
)


tbl

lx(tbl, 2)        # l_2
dx(tbl, 2)        # d_2
ndx(tbl, 1, 3)    # {}_3 d_1
npx(tbl, 1, 3)    # {}_3 p_1
nqx(tbl, 2, 3)    # {}_3 q_2


## Listing 6.3: Curtate and complete expectations from a life table
library(mqriskR)

tbl <- life_table(
  x = 0:4,
  px = c(0.90, 0.80, 0.60, 0.30, 0.00),
  radix = 10000
)

ex_curtate_tab(tbl, 2)                      # e_2
ex_complete_tab(tbl, 2, assumption = "udd") # e^o_2 under UDD

ex_temp_curtate_tab(tbl, x = 1, n = 3)
ex_temp_complete_tab(tbl, x = 2, n = 1.5, assumption = "cf")




## Listing 6.4: Fractional-age calculations under UDD, constant force, and Balducci
library(mqriskR)

tbl <- life_table(
  x = 0:4,
  px = c(0.90, 0.80, 0.60, 0.30, 0.00),
  radix = 10000
)

tpx_tab(tbl, x = 2, t = 0.25, assumption = "udd")
tpx_tab(tbl, x = 2, t = 0.25, assumption = "cf")
tpx_tab(tbl, x = 2, t = 0.25, assumption = "balducci")

mux_tab(tbl, x = 2, t = 0.25, assumption = "udd")
mux_tab(tbl, x = 2, t = 0.25, assumption = "cf")
mux_tab(tbl, x = 2, t = 0.25, assumption = "balducci")




## Listing 6.5: Working with a select life table
library(mqriskR)

sel_tbl <- select_life_table(
  x_sel = c(20,20,20,
            21,21,21,
            22,22,22),
  duration = c(0,1,2,
               0,1,2,
               0,1,2),
  attained_age = c(20,21,22,
                   21,22,23,
                   22,23,24),
  lx = c(100000, 99000, 97000,
         99500, 97500, 95000,
         98000, 95500, 92000)
)

lx_select(sel_tbl, 20, 1)                  # l_[20]+1
npx_select(sel_tbl, x_sel = 20, t = 0, n = 2)
nqx_select(sel_tbl, x_sel = 20, t = 0, n = 2)
nmxq_select(sel_tbl, x_sel = 20, t = 0, n = 1, m = 1)

## Listing 9.1: Net annual premiums for standard contingent contracts
library(mqriskR)

x <- 40
n <- 20
i <- 0.05

Px_val   <- Px(x = x, i = i, model = "uniform", omega = 100)
Pxn1_val <- Pxn1(x = x, n = n, i = i, model = "uniform", omega = 100)
PnEx_val <- PnEx(x = x, n = n, i = i, model = "uniform", omega = 100)
Pxn_val  <- Pxn(x = x, n = n, i = i, model = "uniform", omega = 100)

out <- c(
  Px   = Px_val,
  Pxn1 = Pxn1_val,
  PnEx = PnEx_val,
  Pxn  = Pxn_val
)

print(round(out, 6))

# Check the endowment premium decomposition
round(Pxn_val - (Pxn1_val + PnEx_val), 10)




## Listing 9.2: Comparison of full-payment and limited-payment net premiums
library(mqriskR)

x <- 40; n <- 20; t <- 10; i <- 0.05

whole_full   <- Px(x = x, i = i, model = "uniform", omega = 100)
whole_10pay  <- tPx(x = x, t = t, i = i, model = "uniform", omega = 100)

term_full    <- Pxn1(x = x, n = n, i = i, model = "uniform", omega = 100)
term_10pay   <- tPxn1(x = x, n = n, t = t, i = i, model = "uniform", omega = 100)

endow_full   <- Pxn(x = x, n = n, i = i, model = "uniform", omega = 100)
endow_10pay  <- tPxn(x = x, n = n, t = t, i = i, model = "uniform", omega = 100)

out <- c(
  whole_full  = whole_full,
  whole_10pay = whole_10pay,
  term_full   = term_full,
  term_10pay  = term_10pay,
  endow_full  = endow_full,
  endow_10pay = endow_10pay
)

print(round(out, 6))



## Listing 9.3: Present value of loss for a whole life insurance
library(mqriskR)

x <- 40
i <- 0.05

premium <- Px(x = x, i = i, model = "uniform", omega = 100)

mean_loss <- EL0x(x = x, i = i, P = premium,
  model = "uniform", omega = 100
)

var_loss <- varL0x(
  x = x, i = i, P = premium,
  model = "uniform", omega = 100
)

c(
  premium = premium,
  mean_loss = mean_loss,
  variance_loss = var_loss
)



## Listing 9.4: Normal approximation for aggregate present value of loss
Ax <- 0.24905
A2x <- 0.09476
i <- 0.06
P <- 0.025

d <- i / (1 + i)
c_factor <- 1 + P / d

mean_L <- c_factor * Ax - P / d
var_L  <- c_factor^2 * (A2x - Ax^2)

z_95 <- qnorm(0.95)
n_required <- (z_95 * sqrt(var_L) / abs(mean_L))^2

c(
  mean_single = mean_L,
  var_single = var_L,
  n_required = ceiling(n_required)
)



## Listing 9.5: Fully continuous premium rate for a whole life insurance

x <- 50
i <- 0.05

premium_rate <- PbarAbarx(x = x, i = i, model = "exponential", lambda = 0.02)
abar_val <- abarx(x = x, i = i, model = "exponential", lambda = 0.02)
delta <- interest_convert(i = i)$delta

check_val <- 1 / abar_val - delta

c(
  premium_rate = premium_rate,
  identity_value = check_val
)



## Listing 9.6:True fractional premium rates for a term insurance
library(mqriskR)

x <- 30
n <- 10
i <- 0.05

P_annual <- Pxn1(x = x, n = n, i = i, model = "uniform", omega = 100)
P_semi   <- Pxn1_m(x = x, n = n, m = 2, i = i, model = "uniform", omega = 100)
P_month  <- Pxn1_m(x = x, n = n, m = 12, i = i, model = "uniform", omega = 100)

c(
  annual = P_annual,
  semiannual = P_semi,
  monthly = P_month
)


## Listing 9.7: Net and gross premiums for a whole life insurance with expenses
library(mqriskR)

x <- 40
i <- 0.05

# Net premium per unit benefit
net_unit <- Px(
  x = x,
  i = i,
  model = "uniform",
  omega = 100
)

# Net premium for a benefit of 1000
net_premium <- 1000 * net_unit

# Gross premium including expenses
gross_premium <- Gx(
  x = x,
  i = i,
  benefit = 1000,
  first_premium_pct = 0.75,
  renewal_premium_pct = 0.10,
  first_policy_exp = 10,
  renewal_policy_exp = 2,
  settlement_exp = 20,
  model = "uniform",
  omega = 100
)

c(
  net_premium = net_premium,
  gross_premium = gross_premium,
  loading = gross_premium - net_premium
)

# ------------------------------------------------------------
# Example 6.6
# l_x = 10000 / (x + 1)^3
# Verify E[T_x] and Var(T_x | T0 > x)
# ------------------------------------------------------------

lx_fun <- function(x) 10000 / (x + 1)^3
tpx_fun <- function(t, x) ((x + 1)^3) / (x + t + 1)^3

check_ex6_6 <- function(x = 2, tmax = 50) {
  EX_num <- integrate(function(t) tpx_fun(t, x), 0, Inf)$value
  EX2_num <- 2 * integrate(function(t) t * tpx_fun(t, x), 0, Inf)$value
  Var_num <- EX2_num - EX_num^2

  EX_theory <- (x + 1) / 2
  Var_theory <- 3 * (x + 1)^2 / 4

  list(
    x = x,
    EX_numeric = EX_num,
    EX_theory = EX_theory,
    Var_numeric = Var_num,
    Var_theory = Var_theory
  )
}

print(check_ex6_6(x = 0))
print(check_ex6_6(x = 2))
print(check_ex6_6(x = 5))

# Plot conditional survival function

curve(tpx_fun(x, 2), from = 0, to = 20,
      xlab = "t",
      ylab = expression(p[2](t)),
      main = "Example 6.6: Conditional survival function for x = 2",
      lwd = 2)
grid()


# ------------------------------------------------------------
# Example 6.10
# Compare UDD, constant force, and Balducci values of mu_{x+t}
# when l_x = 1000 and l_{x+1} = 900
# ------------------------------------------------------------

check_ex6_10 <- function(lx = 1000, lx1 = 900, t0 = 0.25) {
  qx <- (lx - lx1) / lx
  px <- lx1 / lx

  mu_udd <- qx / (1 - t0 * qx)
  mu_cf <- -log(px)
  mu_bal <- qx / (1 - (1 - t0) * qx)

  list(
    qx = qx,
    px = px,
    mu_udd = mu_udd,
    mu_cf = mu_cf,
    mu_balducci = mu_bal
  )
}

print(check_ex6_10())

# Plot mu_{x+t} across t in [0,1]
qx <- 0.10
px <- 0.90
t_grid <- seq(0, 1, by = 0.001)

mu_udd <- qx / (1 - t_grid * qx)
mu_cf <- rep(-log(px), length(t_grid))
mu_bal <- qx / (1 - (1 - t_grid) * qx)

matplot(t_grid, cbind(mu_udd, mu_cf, mu_bal), type = "l", lwd = 2, lty = c(1:3), col = c(1:3),
        xlab = "t", ylab = expression(mu[x+t]),
        main = expression("Example 6.10: Comparison of " * mu[x+t] * " under three assumptions"))
legend("topright",
       legend = c("UDD", "Constant force", "Balducci"),
       lty = c(1:3),, col = c(1:3), lwd = 2, bty = "n")
grid()




# ------------------------------------------------------------
# Example 6.13
# S0(t) = (9000 - 10 t - t^2)/9000, 0 <= t <= 90
# ------------------------------------------------------------

S0_ex613 <- function(t) (9000 - 10 * t - t^2) / 9000
mu_exact_ex613 <- function(t) (10 + 2 * t) / (9000 - 10 * t - t^2)

check_ex6_13 <- function() {
  l60 <- round(100000 * S0_ex613(60))
  l61 <- round(100000 * S0_ex613(61))

  q60 <- (l60 - l61) / l60
  p60 <- 1 - q60

  # part (b)
  tp_udd <- 1 - 0.25 * q60
  tp_cf  <- p60^0.25
  tp_bal <- (1 - q60) / (1 - 0.75 * q60)

  # exact and approximate mu at 60.75
  mu_exact <- mu_exact_ex613(60.75)
  mu_udd <- q60 / (1 - 0.75 * q60)
  mu_cf  <- -log(p60)
  mu_bal <- q60 / (1 - 0.25 * q60)

  list(
    l60 = l60,
    l61 = l61,
    q60 = q60,
    p60 = p60,
    tpx_udd = tp_udd,
    tpx_cf = tp_cf,
    tpx_bal = tp_bal,
    mu_exact = mu_exact,
    mu_udd = mu_udd,
    mu_cf = mu_cf,
    mu_bal = mu_bal
  )
}

print(check_ex6_13())


# ------------------------------------------------------------
# Plot exact and approximate mu_{x+t} for 60 <= x+t <= 61
# ------------------------------------------------------------

age_grid <- seq(60, 61, by = 0.001)
t_grid <- age_grid - 60

mu_exact_vals <- mu_exact_ex613(age_grid)

out <- check_ex6_13()

qx <- out$q60
px <- out$p60

# approximations as functions of t
mu_udd <- qx / (1 - t_grid * qx)
mu_cf <- rep(-log(px), length(t_grid))
mu_bal <- qx / (1 - (1 - t_grid) * qx)

matplot(age_grid,
        cbind(mu_exact_vals, mu_udd, mu_cf, mu_bal),
        type = "l",
        lwd = 2,
        lty = c(1,2,3,4),
        col = 1,
        xlab = "Age",
        ylab = expression(mu[x]),
        main = expression("Example 6.13: Exact and approximate " * mu[x+t]))

legend("topleft",
       legend = c("Exact", "UDD", "Constant force", "Balducci"),
       lty = c(1,2,3,4),
       lwd = 2,
       bty = "n")

grid()




# ------------------------------------------------------------
# Example 6.15
# Makeham: A = 0.002, B = 1e-5, c = 1.10
# ------------------------------------------------------------

check_ex6_15 <- function(A = 0.002, B = 1e-5, c = 1.10) {
  mu_exact <- hazard0(40, model = "makeham", A = A, B = B, c = c)

  S39 <- S0(39, model = "makeham", A = A, B = B, c = c)
  S40 <- S0(40, model = "makeham", A = A, B = B, c = c)
  S41 <- S0(41, model = "makeham", A = A, B = B, c = c)

  p39 <- S40 / S39
  p40 <- S41 / S40

  q39 <- 1 - p39
  q40 <- 1 - p40

  # UDD approximations
  mu_left  <- q40
  mu_right <- q39 / (1 - q39)

  # midpoint rule approximation
  mu_mid <- -0.5 * (log(p39) + log(p40))

  data.frame(
    method = c("Exact", "UDD left endpoint", "UDD right endpoint", "Midpoint"),
    value = c(mu_exact, mu_left, mu_right, mu_mid),
    error = c(0, mu_left - mu_exact, mu_right - mu_exact, mu_mid - mu_exact)
  )
}

print(check_ex6_15())

# Plot exact mu_x near age 40 with approximation lines
A <- 0.002; B <- 1e-5; c <- 1.10
age_grid <- seq(39.5, 40.5, by = 0.001)
mu_vals <- hazard0(age_grid, model = "makeham", A = A, B = B, c = c)
out <- check_ex6_15(A, B, c)

plot(age_grid, mu_vals, type = "l", lwd = 2,
     xlab = "Age", ylab = expression(mu[x]),
     main = expression("Example 6.15: Exact " * mu[x] * " near age 40 and approximations to " * mu[40]))

abline(h = out$value[2], lty = 2)
abline(h = out$value[3], lty = 3)
abline(h = out$value[4], lty = 4)
abline(v = 40, lty = 5)

points(40, out$value[1], pch = 19)

legend("topleft",
       legend = c("Exact curve", "UDD left endpoint", "UDD right endpoint", "Midpoint", "Exact value at age 40"),
       lty = c(1, 2, 3, 4, NA),
       lwd = c(2, 2, 2, 2, NA),
       pch = c(NA, NA, NA, NA, 19),
       bty = "n")
grid()


out <- check_ex6_15()

barplot(out$error[-1],
        names.arg = c("UDD left", "UDD right", "Midpoint"),
        ylab = "Approximation error",
        main = "Example 6.15: Errors in approximating mu_40")
abline(h = 0, lty = 2)
grid()



# ------------------------------------------------------------
# Exercise 6-3
# Y = min(K0, 4)
# ------------------------------------------------------------

check_ex6_3 <- function() {
  qx <- c(0.10, 0.20, 0.30, 0.40, 0.50)
  px <- 1 - qx

  pY <- c(
    qx[1],
    px[1] * qx[2],
    px[1] * px[2] * qx[3],
    px[1] * px[2] * px[3] * qx[4],
    px[1] * px[2] * px[3] * px[4]
  )

  y <- 0:4
  EY <- sum(y * pY)
  EY2 <- sum((y^2) * pY)
  VarY <- EY2 - EY^2

  list(
    distribution = data.frame(Y = y, prob = pY),
    prob_sum = sum(pY),
    EY = EY,
    VarY = VarY
  )
}

out <- check_ex6_3()
print(out$distribution)
print(out$prob_sum)
print(out$EY)
print(out$VarY)

barplot(out$distribution$prob,
        names.arg = out$distribution$Y,
        xlab = "Y",
        ylab = "Probability",
        main = "Exercise 6-3: Distribution of Y = min(K0, 4)")
grid()



# ------------------------------------------------------------
# Exercise 6-19
# Verify pi_4 from Table 6.1a
# ------------------------------------------------------------

check_ex6_19 <- function(make_plot = TRUE) {
  lx_vals <- c(100000, 97408, 97259, 97160, 97082)
  names(lx_vals) <- 0:4

  px <- lx_vals[-1] / lx_vals[-length(lx_vals)]
  qx <- 1 - px

  P_list <- lapply(seq_along(px), function(i) {
    matrix(c(px[i], qx[i],
             0,     1),
           nrow = 2, byrow = TRUE)
  })

  pi0 <- c(1, 0)
  pi1 <- drop(pi0 %*% P_list[[1]])
  pi2 <- drop(pi1 %*% P_list[[2]])
  pi3 <- drop(pi2 %*% P_list[[3]])
  pi4 <- drop(pi3 %*% P_list[[4]])

  out <- list(
    px = px,
    qx = qx,
    pi1 = pi1,
    pi2 = pi2,
    pi3 = pi3,
    pi4 = pi4
  )

  if (make_plot) {
    probs_alive_dead <- rbind(
      c(1, 0),
      pi1, pi2, pi3, pi4
    )
    matplot(0:4, probs_alive_dead, type = "l", lwd = 2, lty = 1:2,col = 1:2,
            xlab = "Time",
            ylab = "State probability",
            main = "Exercise 6-19: State probabilities over time")
    legend("right", legend = c("Alive (State 0)", "Dead (State 1)"),
           lty = 1:2,col = 1:2, lwd = 2, bty = "n")
    grid()
  }

  out
}

out <- check_ex6_19()
print(out$pi4)


# ------------------------------------------------------------
# Exercise 6-27
# l_x = 1000 * (100 - x)^(1/2)
# ------------------------------------------------------------

check_ex6_27 <- function(make_plot = TRUE) {
  lx_fun <- function(x) 1000 * sqrt(100 - x)
  mu_exact_fun <- function(x) 0.5 / (100 - x)

  l36 <- lx_fun(36)
  l37 <- round(lx_fun(37))   # as instructed
  p36 <- l37 / l36
  q36 <- 1 - p36

  t0 <- 0.25
  mu_exact <- mu_exact_fun(36.25)
  mu_udd <- q36 / (1 - t0 * q36)
  mu_cf <- -log(p36)
  mu_bal <- q36 / (1 - (1 - t0) * q36)

  out <- data.frame(
    method = c("Exact", "UDD", "Constant force", "Balducci"),
    value = c(mu_exact, mu_udd, mu_cf, mu_bal)
  )

  if (make_plot) {
    age_grid <- seq(36, 37, by = 0.001)
    t_grid <- age_grid - 36
    mu_exact_vals <- mu_exact_fun(age_grid)
    mu_udd_vals <- q36 / (1 - t_grid * q36)
    mu_cf_vals <- rep(-log(p36), length(age_grid))
    mu_bal_vals <- q36 / (1 - (1 - t_grid) * q36)

    matplot(age_grid,
            cbind(mu_exact_vals, mu_udd_vals, mu_cf_vals, mu_bal_vals),
            type = "l", lwd = 2, lty = c(1,2,3,4), col = 1,
            xlab = "Age", ylab = expression(mu[x]),
            main = expression("Exercise 6-27: Exact and approximate " * mu[x] * " on (36,37]"))
    legend("topleft",
           legend = c("Exact", "UDD", "Constant force", "Balducci"),
           lty = c(1,2,3,4), lwd = 2, bty = "n")
    grid()
  }

  out
}

out <- check_ex6_27()
print(out)



# ------------------------------------------------------------
# Exercise 6-37
# Select table with 2-year select period, UDD
# ------------------------------------------------------------

check_ex6_37 <- function() {
  l_60   <- 80625
  l_60_1 <- 79954
  l_62   <- 78839

  q_60_sel <- 1 - l_60_1 / l_60
  q_60_1_sel <- 1 - l_62 / l_60_1

  # from age [60]+0.60 to [60]+1.00 : 0.40 years left in current year
  q_part1 <- (0.40 * q_60_sel) / (1 - 0.60 * q_60_sel)

  # then 0.50 years in next year under UDD
  q_part2 <- 0.50 * q_60_1_sel

  total_q <- q_part1 + (1 - q_part1) * q_part2

  list(
    q_sel_60 = q_60_sel,
    q_sel_60_plus_1 = q_60_1_sel,
    q_part1 = q_part1,
    q_part2 = q_part2,
    total_q = total_q
  )
}

out <- check_ex6_37()
print(out)

## Listing 4.1: Risk pooling and one-year capital stress (Exercises 4-5 and 4-9)
set.seed(1)
simulate_one_year <- function(n_policies, p_claim, benefit, premium, capital, n_sims = 10000) {
  claims <- rbinom(n_sims, size = n_policies, prob = p_claim)
  profit <- n_policies * premium - claims * benefit
  end_capital <- capital + profit
  list(
    ruin_prob = mean(end_capital < 0),
    mean_end_capital = mean(end_capital),
    q05_end_capital = unname(quantile(end_capital, 0.05))
  )
}
benefit <- 1000
premium <- 50
capital <- 2000
p_claim <- 0.02   # assumed claim probability (varied in repository solutions)

n_grid <- c(10, 20, 50, 100, 200, 500, 1000)

res <- lapply(n_grid, function(n) {
  out <- simulate_one_year(n, p_claim, benefit, premium, capital)
  c(n_policies = n, out)
})

print(do.call(rbind, res))


## Listing 4.2: Surplus strain intuition from high first-year acquisition costs
new_policies_per_year <- 500
years <- 10

premium <- 500
first_year_acq_cost <- 800
renewal_cost <- 50

capital0 <- 1e6
capital <- numeric(years + 1)
capital[1] <- capital0

inforce <- 0

for (y in 1:years) {
  newbiz <- new_policies_per_year
  cash_in <- premium * (inforce + newbiz)
  cash_out <- renewal_cost * inforce + first_year_acq_cost * newbiz
  capital[y + 1] <- capital[y] + (cash_in - cash_out)
  inforce <- inforce + newbiz
}

data.frame(year = 0:years, capital = capital)


## Listing 4.3: Defined benefit pension formula illustration
db_pension <- function(final_avg_salary, years_service, accrual = 1 + 2/3) {
  rate <- accrual / 100
  rate * final_avg_salary * years_service
}

db_pension(final_avg_salary = 90000, years_service = 25)

# R Check for Example 10.1
# Verify {}_{n-1}V_{x:\overline{n}|} both ways:
# (1) prospective reserve formula
# (2) shortcut v - P_{x:\overline{n}|}

A_xn <- 0.20
d <- 0.08
i <- d / (1 - d)
v <- 1 / (1 + i)

# Premium for the n-year endowment
adot_xn <- (1 - A_xn) / d
P_xn <- A_xn / adot_xn

# Reserve from the prospective formula at t = n - 1
# For one year remaining in an endowment:
# A_{x+n-1:\overline{1}|} = v and ä_{x+n-1\overline1n}|}} = 1
reserve_prospective <- v - P_xn * 1

# Reserve from the shortcut expression
reserve_shortcut <- v - P_xn

out <- c(
  i = i,
  v = v,
  premium = P_xn,
  reserve_prospective = reserve_prospective,
  reserve_shortcut = reserve_shortcut,
  difference = reserve_prospective - reserve_shortcut
)

print(round(out, 6))


# R Check for Example 10.2
# Solve for P_{x:\overline{n}|}^1 from
# {}_nV_x = P_x / P_{x:\overline{n}|}^{1(endowment)} - P_{x:\overline{n}|}^1 / P_{x:\overline{n}|}^{1(endowment)}

Px_whole <- 0.090
Vn <- 0.563
PnEx <- 0.00864   # this is P_{x:\overline{n}|}^{ 1 } for the pure endowment part

Pxn1 <- Px_whole - Vn * PnEx

print(round(c(
  Px_whole = Px_whole,
  Vn = Vn,
  PnEx = PnEx,
  Pxn1 = Pxn1
), 6))

# Sensitivity plot: implied term premium as reserve changes
V_grid <- seq(0, 1.2, by = 0.01)
Pxn1_grid <- Px_whole - V_grid * PnEx

plot(
  V_grid, Pxn1_grid, type = "l", lwd = 2,
  xlab = "Reserve value",
  ylab = "Implied term premium",
  main = "Example 10.2: Implied Term Premium vs Reserve"
)
abline(v = Vn, lty = 2)
abline(h = Pxn1, lty = 2)


# R Check for Example 10.5
# Illustrate that {}_tV = P * s̈_t for t <= n
# and P = ä_{x+n} / s̈_n

library(mqriskR)

x <- 55
n <- 10
i <- 0.05
omega <- 100

d <- i / (1 + i)

adot_at_defer_end <- adotx(
  x = x + n,
  i = i,
  model = "uniform",
  omega = omega
)

sdot_n <- ((1 + i)^n - 1) / d
P <- adot_at_defer_end / sdot_n

t_vals <- 0:n
sdot_t <- ((1 + i)^t_vals - 1) / d
reserve_vals <- P * sdot_t

out <- data.frame(
  t = t_vals,
  sdot_t = sdot_t,
  reserve = reserve_vals
)

print(round(out, 6))
print(round(c(
  premium = P,
  reserve_at_n = tail(reserve_vals, 1),
  target_annuity_value = adot_at_defer_end
), 6))

plot(
  t_vals, reserve_vals, type = "b", pch = 19,
  xlab = "Duration t",
  ylab = "Reserve",
  main = "Example 10.5: Reserve Path During Deferred Period"
)
abline(h = adot_at_defer_end, lty = 2)



# R Check for Example 10.6
# q' - q = (.01)(1.03) / (1 - {}_{20}V_x)

V20 <- 0.427
premium_extra <- 0.01
i <- 0.03

q_excess <- premium_extra * (1 + i) / (1 - V20)

print(round(c(
  V20 = V20,
  premium_extra = premium_extra,
  i = i,
  q_excess = q_excess
), 6))

# Sensitivity to reserve level
V_grid <- seq(0.05, 0.90, by = 0.01)
q_excess_grid <- premium_extra * (1 + i) / (1 - V_grid)

plot(
  V_grid, q_excess_grid, type = "l", lwd = 2,
  xlab = "Reserve at duration 20",
  ylab = expression(q[x+19] * "'" - q[x+19]),
  main = "Example 10.6: Implied Excess Mortality vs Reserve"
)
abline(v = V20, lty = 2)
abline(h = q_excess, lty = 2)




# R Check for Example 10.8
# Verify {}_tV_x^{(m)} = [1 + beta(m) P_x^{(m)}] {}_tV_x under UDD

library(mqriskR)

x <- 40
t <- 10
i <- 0.05
m_vals <- c(2, 4, 12)
omega <- 100

annual_reserve <- tVx(
  x = x,
  t = t,
  i = i,
  model = "uniform",
  omega = omega
)

check_df <- lapply(m_vals, function(m) {
  ic <- interest_convert(i = i, m = m)
  im <- ic$im
  dm <- (im / m) / (1 + im / m)
  beta <- (i - im) / (im * dm)

  prem_m <- Px_m(
    x = x,
    m = m,
    i = i,
    model = "uniform",
    omega = omega
  )

  lhs <- tVx_m(
    x = x,
    t = t,
    m = m,
    i = i,
    model = "uniform",
    omega = omega
  )

  rhs <- (1 + beta * prem_m) * annual_reserve

  data.frame(
    m = m,
    annual_reserve = annual_reserve,
    premium_m = prem_m,
    lhs = lhs,
    rhs = rhs,
    difference = lhs - rhs
  )
})

check_df <- do.call(rbind, check_df)
print(round(check_df, 10))

plot(
  check_df$m, check_df$difference, type = "b", pch = 19,
  xlab = "m",
  ylab = "LHS - RHS",
  main = "Example 10.8: UDD Reserve Identity Check"
)
abline(h = 0, lty = 2)




# R Check for Exercise 10-3
# Compute {}_{20}V_{45} from the given premium relationships.

P45 <- 0.014
P45_20 <- 0.030
P45_20_term <- 0.022

reserve_20 <- P45 / P45_20_term - (P45_20 - P45_20_term) / P45_20_term

print(round(c(
  P45 = P45,
  P45_20 = P45_20,
  P45_20_term = P45_20_term,
  reserve_20 = reserve_20
), 6))

# Sensitivity plot: reserve as P45 varies slightly
P45_grid <- seq(0.010, 0.018, by = 0.0005)
reserve_grid <- P45_grid / P45_20_term - (P45_20 - P45_20_term) / P45_20_term

plot(
  P45_grid, reserve_grid, type = "l", lwd = 2,
  xlab = "Whole life premium P_45",
  ylab = "Reserve at duration 20",
  main = "Exercise 10-3: Retrospective Reserve Sensitivity"
)
abline(v = P45, lty = 2)
abline(h = reserve_20, lty = 2)





# R Check for Exercise 10-6
# Compute 1000({}_2V_{x:\overline{3}|} - {}_1V_{x:\overline{3}|})

P_x3 <- 0.33251
i <- 0.06
v <- 1 / (1 + i)

lx <- 100
lx1 <- 90
qx <- 1 - lx1 / lx

# Reserve at t = 1, retrospective
V1 <- P_x3 / ((1 / (1 + i)) * (1 - qx)) - (v * qx) / ((1 / (1 + i)) * (1 - qx))

# Reserve at t = 2, prospective
V2 <- v - P_x3

diff_val <- 1000 * (V2 - V1)

print(round(c(
  qx = qx,
  V1 = V1,
  V2 = V2,
  difference_1000 = diff_val
), 6))

# Bar plot
barplot(
  c(1000 * V1, 1000 * V2),
  names.arg = c("1000 V1", "1000 V2"),
  main = "Exercise 10-6: Reserves at Durations 1 and 2"
)
abline(h = c(1000 * V1, 1000 * V2), lty = 2)



# R Check for Exercise 10-9
# Find Pr(L < 190) for a 2-year term insurance of amount 400.

P <- 0.185825
V1 <- 0.04145
i <- 0.10
v <- 1 / (1 + i)

# From {}_1V_{x:\overline{2}|}^1 = v q_{x+1} - P
q_x1 <- (V1 + P) / v

# Solve for p_x from premium equation
f <- function(px) {
  ((1 - px) / 1.10 + (px * q_x1) / (1.10^2)) / (1 + px / 1.10) - P
}

px <- uniroot(f, interval = c(0, 1))$root

# Losses by outcome
L1 <- 400 / 1.10 - 400 * P
L2 <- 400 / (1.10^2) - 400 * P - (400 * P) / 1.10

print(round(c(
  q_x1 = q_x1,
  px = px,
  loss_first_year = L1,
  loss_second_year = L2,
  probability_loss_lt_190 = px
), 6))

# Plot the loss outcomes
barplot(
  c(L1, L2, 0),
  names.arg = c("Death in year 1", "Death in year 2", "Survive 2 years"),
  main = "Exercise 10-9: Loss at Issue by Outcome",
  ylab = "Loss"
)
abline(h = 190, lty = 2)





# R Check for Exercise 10-20
# Given Var(L) = 0.20, A2bar_x = 0.30, and Abar_{x+20} = 0.70,
# solve for Abar_x and compute {}_{20}V(barA_x).

var_L <- 0.20
A2bar_x <- 0.30
Abar_x20 <- 0.70

# Solve:
# (A2bar_x - Abar_x^2) / (1 - Abar_x)^2 = var_L
f <- function(Abar_x) {
  (A2bar_x - Abar_x^2) / (1 - Abar_x)^2 - var_L
}

root <- uniroot(f, interval = c(1e-8, 0.999999))$root

reserve_20 <- 1 - (1 - Abar_x20) / (1 - root)

print(round(c(
  Abar_x = root,
  reserve_20 = reserve_20
), 6))

# Plot objective function
grid <- seq(0.01, 0.95, by = 0.001)
obj <- sapply(grid, f)

plot(
  grid, obj, type = "l", lwd = 2,
  xlab = "Abar_x",
  ylab = "Variance equation residual",
  main = "Exercise 10-20: Solving for Abar_x"
)
abline(h = 0, lty = 2)
abline(v = root, lty = 2)




# R Check for Exercise 10-22
# Compute {}_{10}V(barA_40) under a uniform survival model with omega = 100 and i = 0.05.

library(mqriskR)

x <- 40
t <- 10
i <- 0.05
omega <- 100

reserve_pkg <- tVbarAbarx(
  x = x,
  t = t,
  i = i,
  model = "uniform",
  omega = omega
)

reserve_ratio <- 1 - abarx(
  x = x + t,
  i = i,
  model = "uniform",
  omega = omega
) / abarx(
  x = x,
  i = i,
  model = "uniform",
  omega = omega
)

print(round(c(
  reserve_pkg = reserve_pkg,
  reserve_ratio = reserve_ratio
), 6))

# Plot reserve as a function of t
t_grid <- seq(0, 20, by = 0.25)
reserve_grid <- tVbarAbarx(
  x = x,
  t = t_grid,
  i = i,
  model = "uniform",
  omega = omega
)

plot(
  t_grid, reserve_grid, type = "l", lwd = 2,
  xlab = "Duration t",
  ylab = "Fully continuous reserve",
  main = "Exercise 10-22: Continuous Reserve Path"
)
abline(v = t, lty = 2)
abline(h = reserve_pkg, lty = 2)



# R Check for Exercise 10-25
# Compute 1000({}_5V_35^(4) - {}_5V_35) using Example 10.8 relationship.

A35 <- 0.17092
V5 <- 0.04471
i <- 0.05

# Interest conversion helper from package
library(mqriskR)

calc_diff <- function(m) {
  ic <- interest_convert(i = i, m = m)
  im <- ic$im
  dm <- (im / m) / (1 + im / m)

  beta_m <- (i - im) / (im * dm)
  alpha_m <- (i * (i / (1 + i))) / (im * dm)

  d <- i / (1 + i)
  adot35 <- (1 - A35) / d
  adot35_m <- alpha_m * adot35 - beta_m
  P35_m <- A35 / adot35_m

  1000 * beta_m * P35_m * V5
}

m_vals <- c(2, 4, 12)
diff_vals <- sapply(m_vals, calc_diff)

print(round(c(
  diff_m2 = diff_vals[1],
  diff_m4 = diff_vals[2],
  diff_m12 = diff_vals[3]
), 6))

barplot(
  diff_vals,
  names.arg = paste0("m=", m_vals),
  main = "Exercise 10-25: Reserve Difference by Payment Frequency",
  ylab = "1000(V^(m) - V)"
)



# R Check for Exercise 10-30
# Verify G^M = (q_assumed - q_actual)(1 - V_{t+1})

Vt <- 0.085697
Vt1 <- 0.096115
P <- 0.008013
i_assumed <- 0.06
q_assumed <- 0.00356
q_actual <- 0.00300

GM_formula <- (Vt + P) * (1 + i_assumed) - (q_actual + (1 - q_actual) * Vt1)
GM_short <- (q_assumed - q_actual) * (1 - Vt1)

print(round(c(
  GM_formula = GM_formula,
  GM_short = GM_short,
  difference = GM_formula - GM_short
), 10))

# Sensitivity in actual mortality
q_actual_grid <- seq(0.002, 0.005, by = 0.00005)
GM_grid <- (q_assumed - q_actual_grid) * (1 - Vt1)

plot(
  q_actual_grid, GM_grid, type = "l", lwd = 2,
  xlab = "Actual mortality q'_x+t",
  ylab = "Mortality gain",
  main = "Exercise 10-30: Mortality Gain Sensitivity"
)
abline(v = q_actual, lty = 2)
abline(h = GM_short, lty = 2)



# R Check for Exercise 10-31
# Verify G^I = (V_t + P)(i_actual - i_assumed)

Vt <- 0.085697
P <- 0.008013
Vt1 <- 0.096115
i_assumed <- 0.06
i_actual <- 0.065
q_assumed <- 0.00356

GI_formula <- (Vt + P) * (1 + i_actual) - (q_assumed + (1 - q_assumed) * Vt1)
GI_short <- (Vt + P) * (i_actual - i_assumed)

print(round(c(
  GI_formula = GI_formula,
  GI_short = GI_short,
  difference = GI_formula - GI_short
), 10))

# Sensitivity in actual interest
i_actual_grid <- seq(0.04, 0.08, by = 0.0005)
GI_grid <- (Vt + P) * (i_actual_grid - i_assumed)

plot(
  i_actual_grid, GI_grid, type = "l", lwd = 2,
  xlab = "Actual interest i'",
  ylab = "Interest gain",
  main = "Exercise 10-31: Interest Gain Sensitivity"
)
abline(v = i_actual, lty = 2)
abline(h = GI_short, lty = 2)

## Listing 13.1: Multiple-decrement table construction
library(mqriskR)

x <- c(45:50)

qmat <- cbind(q1 = c(.011, .012, .013, .014, .015, .016),
              q2 = c(.100, .100, .100, .100, .100, .100))

tbl <- md_table(x = x, qxj = qmat, radix = 1000)
tbl

## Listing 13.2: Multi-year probabilities from a multiple-decrement table
# 3-year survival probability from age 46
npxtau_md(tbl, x = 46, n = 3)

# 2-year decrement due to Cause 2 from age 47
nqxj_md(tbl, x = 47, n = 2, j = 2)

# 2-year decrement due to Cause 1 from age 46
nqxj_md(tbl, x = 46, n = 2, j = 1)



## Listing 13.3: Constant-force multiple-decrement calculations
mu <- c(0.10, 0.20)
t <- 5

# total survival probability
tpx_tau_cf(mu, t)

# single-decrement probability
tqxprimej_cf(mu = 0.10, t = t)

# dependent probability for Cause 1
tqxj_cf(mu, j = 1, t = t)

# ultimate probability as t -> infinity
tqxj_cf(mu, j = 1, t = 100)


## Listing 13.4: MUDD conversion of probabilities
qxj <- c(0.20, 0.10)

qxprime_mudd(qxj)


## Listing 13.5: SUDD conversion of probabilities
qxprime_sudd(q1 = 0.20, q2 = 0.10)



## Listing 13.5: Ratio property under constant forces
mu <- c(0.10, 0.20)
t <- 2

lhs <- tqxj_cf(mu, j = 1, t = t) / (1 - tpx_tau_cf(mu, t))
rhs <- mu[1] / sum(mu)

c(lhs = lhs, rhs = rhs)

# Chapter 15 R Check: Example 15.7
library(mqriskR)

# Forward rates from Table 15.11:
# f_{0,5}, f_{1,4}, f_{2,3}, f_{3,2}, f_{4,1}
f <- c(0.04, 0.05, 0.06, 0.07, 0.08)

# Mortality rates q60,...,q64
qx <- c(0.02, 0.03, 0.04, 0.05, 0.06)

# -----------------------------
# Retrospective calculation
# -----------------------------
E_5_60 <- prod(1 - qx[1:5]) / (1 + f[1])^5
E_4_61 <- prod(1 - qx[2:5]) / (1 + f[2])^4
E_3_62 <- prod(1 - qx[3:5]) / (1 + f[3])^3
E_2_63 <- prod(1 - qx[4:5]) / (1 + f[4])^2
E_1_64 <- prod(1 - qx[5])   / (1 + f[5])^1

sddot_factor_ret <- sum(1 / c(E_5_60, E_4_61, E_3_62, E_2_63, E_1_64))
P_ret <- 10000 / sddot_factor_ret

# -----------------------------
# Prospective calculation
# Recover spot rates z1,...,z5 using Equation (15.3)
# -----------------------------
z <- numeric(5)
z[5] <- f[1]

for (n in 1:4) {
  z[n] <- ((1 + z[5])^5 / (1 + f[n + 1])^(5 - n))^(1 / n) - 1
}

# APV of benefit
benefit_apv <- 10000 * prod(1 - qx) / (1 + z[5])^5

# Premium annuity-due factor
surv_start <- c(1, cumprod(1 - qx))[1:5]
premium_disc <- c(
  1,
  (1 + z[1])^(-1),
  (1 + z[2])^(-2),
  (1 + z[3])^(-3),
  (1 + z[4])^(-4)
)
annuity_due_factor <- sum(premium_disc * surv_start)

P_pros <- benefit_apv / annuity_due_factor

cat("Retrospective premium =", round(P_ret, 2), "\n")
cat("Prospective premium   =", round(P_pros, 2), "\n")
cat("Recovered spot rates  =", round(100 * z, 4), "\n")

# Plot recovered spot rates
plot(
  1:5, 100 * z, type = "b", pch = 19,
  xlab = "Maturity n",
  ylab = "Spot rate z_n (%)",
  main = "Example 15.7: recovered spot rates"
)





# Exercise 15-2 R Check
library(mqriskR)

qx <- c(.03, .04, .05, .06, .07)

scenarios <- list(
  Scenario1 = c(.06, .07, .08, .09, .10),
  Scenario2 = rep(.06, 5),
  Scenario3 = c(.06, .05, .04, .03, .02)
)

results <- sapply(scenarios, function(i) Axn1_var(qx = qx, i = i))
print(results)

# Contribution breakdown for Scenario 3
i <- scenarios$Scenario3
vt <- vt_var(i)
surv <- c(1, cumprod(1 - qx))[1:5]
contrib <- vt * surv * qx

barplot(
  contrib,
  names.arg = 1:5,
  xlab = "Year of death",
  ylab = "Contribution to APV",
  main = "Exercise 15-2: Term insurance contributions (Scenario 3)"
)



# Exercise 15-3 R Check
library(mqriskR)

qx <- rep(0.02, 5)

i_level <- rep(0.06, 5)
i_rise  <- c(0.06, 0.07, 0.08, 0.09, 0.10)
i_fall  <- c(0.06, 0.05, 0.04, 0.03, 0.03)

annuity <- c(
  level = axn_var(qx, i_level),
  rise  = axn_var(qx, i_rise),
  fall  = axn_var(qx, i_fall)
)

pure_endowment <- c(
  level = nEx_var(qx, i_level),
  rise  = nEx_var(qx, i_rise),
  fall  = nEx_var(qx, i_fall)
)

print(annuity)
print(pure_endowment)

# Regulatory test
ratio_annuity <- annuity["fall"] / annuity["level"]
ratio_pe <- pure_endowment["fall"] / pure_endowment["level"]

cat("Annuity ratio:", ratio_annuity, "\n")
cat("Pure endowment ratio:", ratio_pe, "\n")

# Plot comparison
matplot(
  1:3,
  cbind(annuity, pure_endowment),
  type = "b", pch = 19,
  xaxt = "n",
  col = 1:2,
  xlab = "Scenario",
  ylab = "APV",
  main = "Exercise 15-3: Scenario comparison"
)
axis(1, at = 1:3, labels = c("Level", "Rising", "Falling"))
legend("right",col = 1:2,
       legend = c("Annuity", "Pure Endowment"),
       pch = 19, lty = 1)



# Exercise 15-4 R Check
library(mqriskR)

qx <- c(0.10, 0.15, 0.20, 0.25, 0.30)

i_inc <- c(0.07, 0.08, 0.09, 0.10, 0.11)
i_dec <- c(0.07, 0.06, 0.05, 0.04, 0.03)

# Annuity-certain
ann_certain_inc <- sum(vt_var(i_inc))
ann_certain_dec <- sum(vt_var(i_dec))

# Life annuity
ann_life_inc <- axn_var(qx, i_inc)
ann_life_dec <- axn_var(qx, i_dec)

# Term insurance
term_inc <- Axn1_var(qx, i_inc)
term_dec <- Axn1_var(qx, i_dec)

results <- data.frame(
  Product = c("Annuity Certain", "Life Annuity", "Term Insurance"),
  Increasing = c(ann_certain_inc, ann_life_inc, term_inc),
  Decreasing = c(ann_certain_dec, ann_life_dec, term_dec)
)

results$Ratio <- results$Decreasing / results$Increasing
print(results)

# Plot sensitivity
barplot(
  t(as.matrix(results[,2:3])),
  beside = TRUE,
  names.arg = results$Product,
  legend.text = c("Increasing", "Decreasing"),
  main = "Exercise 15-4: Interest sensitivity comparison",
  ylab = "APV"
)



# Exercise 15-14 R Check
library(mqriskR)

fn1 <- c(0.04, 0.05, 0.06, 0.07, 0.08)

z <- z_from_fn1(fn1)
print(z)

# Plot term structure
plot(
  1:length(z), 100 * z,
  type = "b", pch = 19,
  xlab = "Maturity (years)",
  ylab = "Spot rate (%)",
  main = "Exercise 15-14: Term structure from forward rates"
)

## Listing 3.1: Homogeneous DTMC: state vectors and multi-step probabilities via matrix powers
P <- matrix(c(
  0.80, 0.20, 0.00,
  0.30, 0.60, 0.10,
  0.00, 0.00, 1.00
), nrow = 3, byrow = TRUE)

# Known to be in State 1 at time n:
pi_n <- c(0, 1, 0)

# One-step state vector:
pi_np1 <- drop(pi_n %*% P)

# Two-step state vector using P^2:
P2 <- P %*% P
pi_np2 <- drop(pi_n %*% P2)

pi_np1
pi_np2
P2


## Listing 3.2: Non-homogeneous DTMC: sequential transition matrices
P0 <- matrix(c(
  0.60, 0.40,
  0.70, 0.30
), nrow = 2, byrow = TRUE)

P1 <- matrix(c(
  0.50, 0.50,
  0.80, 0.20
), nrow = 2, byrow = TRUE)

# Begin in State 1 at time 0:
pi0 <- c(0, 1)

pi1 <- drop(pi0 %*% P0)
pi2 <- drop(pi1 %*% P1)

pi1
pi2
pi2[1]  # P(State 0 at time 2 | State 1 at time 0)


## Listing 3.3 CTMC transition probabilities via the matrix exponential expm(Q*t)
# Rates from State 0 to States 1,2,3
mu01 <- 0.30
mu02 <- 0.50
mu03 <- 0.70

Q <- matrix(0, nrow = 4, ncol = 4)
Q[1, 2] <- mu01
Q[1, 3] <- mu02
Q[1, 4] <- mu03
Q[1, 1] <- -(mu01 + mu02 + mu03)

t <- 1
Pt <- expm::expm(Q * t)

Pt
Pt[1, 3]  # P(X(1)=2 | X(0)=0)



## Listing 3.4: Expected payments in a non-homogeneous DTMC (state-probability recursion)
# State order is: 0, 1, 2  (so pi[1]=State0, pi[2]=State1, pi[3]=State2)

P0 <- matrix(c(
  0.60, 0.30, 0.10,
  0.00, 0.00, 1.00,
  0.00, 0.00, 1.00
), nrow = 3, byrow = TRUE)

P1 <- P0

# Pk corresponds to P^(k) for k >= 2 in the exercise
Pk <- matrix(c(
  0.00, 0.30, 0.70,
  0.00, 0.00, 1.00,
  0.00, 0.00, 1.00
), nrow = 3, byrow = TRUE)


state0 <- 1
state1 <- 2
state2 <- 3

pi <- c(1, 0, 0)  # start in State 0 at time 0
N <- 15           # compute partial sums through time N

pay_a <- numeric(N + 1)  # payments for part (a)
pay_b <- numeric(N + 1)  # payments for part (b)

for (t in 0:N) {
  # (a) Pay 1 at time t if in State 0 or State 1
  pay_a[t + 1] <- 1 * (pi[state0] + pi[state1])

  # (b) Pay 4 at time t>=1 if in State 1
  pay_b[t + 1] <- if (t >= 1) 4 * pi[state1] else 0

  # update pi(t) -> pi(t+1) using the matrix for interval (t, t+1]
  if (t < N) { if (t == 0) {pi <- drop(pi %*% P0)
  } else if (t == 1) {pi <- drop(pi %*% P1)
  } else {pi <- drop(pi %*% Pk)}}
}

sum(pay_a)  # partial expected value of payments in (a) through time N
sum(pay_b)  # partial expected value of payments in (b) through time N

## Listing 2.1: Discrete distribution utilities (PMF -> CDF, mean, variance)
x <- c(0, 1, 2)
p <- c(0.20, 0.30, 0.50)  # probabilities must sum to 1

# CDF values at support points (after ordering by x)
o <- order(x)
x <- x[o]; p <- p[o]
F <- cumsum(p)

# Mean and variance
EX  <- sum(x * p)
EX2 <- sum((x^2) * p)
VarX <- EX2 - EX^2

EX
VarX
F





## Listing 2.2: R distribution functions: density (d*), CDF (p*), quantile (q*), simulation (r*)}, label={lst:ch2_r_distribution_convention}]
# Normal distribution: X ~ N(mu, sigma^2)
mu <- 10
sigma <- 2

# Probability: P(X <= 12)
pnorm(12, mean = mu, sd = sigma)

# 95th percentile of X
qnorm(0.95, mean = mu, sd = sigma)

# Simulate 5 observations
rnorm(5, mean = mu, sd = sigma)


## Listing 2.3: Poisson probabilities and simulation
lambda <- 3

# P(X = 2) and P(X <= 2) for X ~ Pois(lambda)
dpois(2, lambda)
ppois(2, lambda)

# Simulate and compare sample mean/variance to lambda
set.seed(1)
x <- rpois(10000, lambda)
mean(x)
var(x)


## Listing 2.4: Normal probabilities without tables using pnorm and qnorm
# Standard normal probabilities
pnorm(1.96)                 # P(Z <= 1.96)
1 - pnorm(1.96)             # P(Z > 1.96)

# Two-sided probability P(|Z| <= 1.96)
pnorm(1.96) - pnorm(-1.96)

# Find z such that P(Z <= z) = 0.975
qnorm(0.975)


## Listing 2.5 Verification of the double expectation theorem (discrete example)
# Joint pmf table for Y in {1,2} and X in {0,1,2}
x_vals <- c(0, 1, 2)
y_vals <- c(1, 2)

p <- matrix(
  c(0.10, 0.20, 0.30,
    0.10, 0.10, 0.20),
  nrow = length(y_vals), byrow = TRUE
)

# Marginal of X
px <- colSums(p)
EX <- sum(x_vals * px)
VarX <- sum((x_vals^2) * px) - EX^2

# Conditional moments of X given Y
py <- rowSums(p)
cond_px_given_y <- sweep(p, 1, py, "/")

EX_given_y  <- as.numeric(cond_px_given_y %*% x_vals)
EX2_given_y <- as.numeric(cond_px_given_y %*% (x_vals^2))
Var_given_y <- EX2_given_y - EX_given_y^2

# Double expectation theorem
EX_alt <- sum(EX_given_y * py)
Var_alt <- sum(Var_given_y * py) + (sum((EX_given_y^2) * py) - EX_alt^2)

c(EX = EX, EX_alt = EX_alt, VarX = VarX, Var_alt = Var_alt)



## Listing 2.6: Verification of the double expectation theorem (continuous example)
# X ~ Unif(0,12), Y | X=x ~ Unif(0,x)

EX <- 6
VarX <- 12
EX2 <- VarX + EX^2  # E[X^2]

EY <- 0.5 * EX
VarY <- (1/12) * EX2 + (1/4) * VarX

c(EY = EY, VarY = VarY)

## Listing 15.1: Pure endowment under variable interest scenarios
library(mqriskR)

qx <- c(.03, .04, .05, .06, .07)

scen1 <- c(.06, .07, .08, .09, .10)
scen2 <- c(.06, .06, .06, .06, .06)
scen3 <- c(.06, .05, .04, .03, .02)

c(Scenario1 = nEx_var(qx = qx, i = scen1, benefit = 1000),
  Scenario2 = nEx_var(qx = qx, i = scen2, benefit = 1000),
  Scenario3 = nEx_var(qx = qx, i = scen3, benefit = 1000))



## Listing 5.2: Term insurance under variable interest scenarios
library(mqriskR)

qx <- c(.03, .04, .05, .06, .07)

scen1 <- c(.06, .07, .08, .09, .10)
scen2 <- c(.06, .06, .06, .06, .06)
scen3 <- c(.06, .05, .04, .03, .02)

c(Scenario1 = Axn1_var(qx = qx, i = scen1),
  Scenario2 = Axn1_var(qx = qx, i = scen2),
  Scenario3 = Axn1_var(qx = qx, i = scen3))



##Listing 15.3: Deterministic interest-rate stress test for a temporary annuity
library(mqriskR)

qx <- rep(.02, 5)

level <- c(.06, .06, .06, .06, .06)
rise  <- c(.06, .07, .08, .09, .10)
fall  <- c(.06, .05, .04, .03, .03)

apv_level <- axn_var(qx = qx, i = level, type = "immediate")
apv_rise  <- axn_var(qx = qx, i = rise,  type = "immediate")
apv_fall  <- axn_var(qx = qx, i = fall,  type = "immediate")

c(level = apv_level,
  rise  = apv_rise,
  fall  = apv_fall,
  ratio_fall_to_level = apv_fall / apv_level)



##Listing 15.4: Bootstrapping semiannual spot rates
library(mqriskR)

maturity <- c(0.5, 1.0, 1.5, 2.0)
coupon_yield <- c(0.0244, 0.0260, 0.0276, 0.0293)

z_from_coupon_semi(maturity = maturity, coupon_yield = coupon_yield)



##Listing 15.5: Present value using spot rates
library(mqriskR)

pv_spot_cashflows(
  amounts = c(200000, 50000, 50000, 100000),
  times   = c(0, 0.5, 1.0, 2.0),
  spot    = c(0, 0.02440, 0.02601, 0.02936),
  compounding = "semiannual")



##Listing 5.6: Net premium using spot rates
library(mqriskR)

qx <- c(.02, .03, .04, .05, .06)
z  <- c(.03, .04, .05, .06, .07)

apvb <- Axn1_spot(qx = qx, z = z)
apvp <- axn_spot(qx = qx, z = z, type = "due")

premium <- 1000000 * apvb / apvp

c(APV_benefit = apvb,
  APV_premium_factor = apvp,
  Premium = premium)



##Listing 15.7 Forward rates implied by spot rates}, label={lst:ch15_forward_rates}]
library(mqriskR)

z <- c(0.03, 0.04, 0.05, 0.06, 0.07)

c(f_1_4 = fnk_from_z(z, n = 1, k = 4),
  f_2_2 = fnk_from_z(z, n = 2, k = 2))

forward_matrix_from_z(z)


##Listing 15.8: Spot rates recovered from forward one-year rates
library(mqriskR)

fn1 <- c(0.04, 0.05, 0.06, 0.07, 0.08)

z_from_fn1(fn1)

## Listing 11.1: Full preliminary term premiums and their benchmark values
library(mqriskR)

x <- 40; i <- 0.05

alpha_f <- alphaF(x = x,i = i,model = "uniform",omega = 100)

beta_f <- betaF(x = x,i = i,model = "uniform",omega = 100)

first_year_cost <- Axn1(x = x,n = 1,i = i,model = "uniform",omega = 100)

attained_age_premium <- Px(x = x + 1,i = i,model = "uniform",omega = 100)

c(
  alpha_F = alpha_f,
  first_year_cost = first_year_cost,
  beta_F = beta_f,
  attained_age_premium = attained_age_premium
)



## Listing 11.2: Comparison of FPT and net level premium reserves
library(mqriskR)

x <- 40
i <- 0.05
t <- 5

fpt_reserve <- tVFx(
  x = x,
  t = t,
  i = i,
  model = "uniform",
  omega = 100
)

shifted_nlp <- tVx(
  x = x + 1,
  t = t - 1,
  i = i,
  model = "uniform",
  omega = 100
)

c(
  fpt_reserve = fpt_reserve,
  shifted_nlp = shifted_nlp
)


## Listing 11.3:Fractional-duration reserve and mean reserve
library(mqriskR)

x <- 40; t <- 10; s <- 0.5; i <- 0.05

# Package value
fractional_reserve <- tsVx(x = x, t = t, s = s, i = i, model = "uniform", omega = 100 )

# Direct calculation from Equation (11.12)
Vt <- tVx(x = x, t = t, i = i, model = "uniform", omega = 100 )

Vt1 <- tVx( x = x, t = t + 1, i = i, model = "uniform", omega = 100 )

P <- Px(x = x,i = i, model = "uniform", omega = 100)

fractional_formula <- Vt * (1 - s) + Vt1 * s + P * (1 - s)

# Mean reserve
mean_reserve <- meanVx( x = x, t = t, i = i, model = "uniform", omega = 100)

c(reserve_t_plus_s = fractional_reserve,
  formula_11_12 = fractional_formula,
  mean_reserve = mean_reserve)


## Listing 11.4: Gross premium reserve and expense reserve
library(mqriskR)

x <- 40; t <- 10; i <- 0.05; G <- 0.04

gross_reserve <- tVGx( x = x, t = t, i = i, G = G, benefit = 1, renewal_premium_pct = 0.10, renewal_policy_exp = 0.002, settlement_exp = 0.02, model = "uniform", omega = 100 )

expense_reserve <- tVEx( x = x, t = t, i = i, G = G, benefit = 1, renewal_premium_pct = 0.10, renewal_policy_exp = 0.002, settlement_exp = 0.02, model = "uniform", omega = 100)

net_reserve <- tVx( x = x, t = t, i = i, model = "uniform", omega = 100)

c(gross_reserve = gross_reserve,
  net_reserve = net_reserve,
  expense_reserve = expense_reserve,
  check = net_reserve + expense_reserve)




## Listing 11.5: Ordered gross gain decomposition
library(mqriskR)

out <- decompGg_disc(VtG = 3950.73, Vt1G = 4607.07, G = 685.00,
                     i_assumed = 0.06, q_assumed = 0.00592,
                     r_assumed = 0.05, e_assumed = 0,
                     s_assumed = 300, i_actual = 0.065,
                     q_actual = 0.005, r_actual = 0.06,
                     e_actual = 0, s_actual = 100,
                     b = 50000, order = c("interest", "mortality", "expense"))

out

# Chapter 14 R Check: Example 14.2
# Whole life insurance with natural-cause and accidental-cause death benefits

library(mqriskR)

delta <- 0.05
mu_nc <- 0.01
mu_ac <- 0.002
mu_tau <- mu_nc + mu_ac

t <- seq(0, 200, by = 0.01)
ptau <- exp(-mu_tau * t)

Abar_nc <- Abarxj_md(
  t = t,
  ptau = ptau,
  muj = rep(mu_nc, length(t)),
  delta = delta,
  benefit = 1000
)

Abar_ac <- Abarxj_md(
  t = t,
  ptau = ptau,
  muj = rep(mu_ac, length(t)),
  delta = delta,
  benefit = 2000
)

APV_total <- Abar_nc + Abar_ac

out <- c(
  Abar_nc = Abar_nc,
  Abar_ac = Abar_ac,
  APV_total = APV_total
)
print(out)

# Plot discounted contribution densities
density_nc <- 1000 * exp(-delta * t) * ptau * mu_nc
density_ac <- 2000 * exp(-delta * t) * ptau * mu_ac
density_total <- density_nc + density_ac

plot(
  t, density_total, type = "l", lwd = 2,
  xlab = "t", ylab = "discounted contribution density",
  main = "Example 14.2: contribution densities"
)
lines(t, density_nc, lwd = 2, lty = 2)
lines(t, density_ac, lwd = 2, lty = 3)
legend(
  "topright",
  legend = c("total", "natural cause", "accidental cause"),
  lwd = 2, lty = c(1, 2, 3), bty = "n"
)



# Chapter 14 R Check: Example 14.10
# APVs and premium using Euler-approximated state probabilities

library(mqriskR)

mu01 <- function(t) 0.10 * t + 0.20
mu02 <- function(t) 0.20
mu10 <- function(t) 0.50
mu12 <- function(t) 0.125 * t + 0.20

out <- tp00_tp01_euler(
  h = 0.10,
  n = 2.00,
  mu01 = mu01,
  mu02 = mu02,
  mu10 = mu10,
  mu12 = mu12
)

j <- 10
i_nom <- 0.05
i_step <- i_nom / j
v <- 1 / (1 + i_step)

q_death_step <- 1 - exp(-0.20 * 0.10)

# Part (a): death benefit APV
death_apv <- 10000 * sum(v^(1:20) * out$tp00[1:20] * q_death_step)

# Part (b): disability benefit APV
disab_apv <- 1000 * sum(v^(1:20) * out$tp01[2:21])

# Part (c): premium APV factor and premium
prem_factor <- 1 + sum(v^(1:19) * out$tp00[2:20])
P <- (death_apv + disab_apv) / prem_factor

ans <- c(
  death_apv = death_apv,
  disab_apv = disab_apv,
  premium_factor = prem_factor,
  P = P
)
print(ans)

# Plot state probabilities
plot(
  out$t, out$tp00, type = "l", lwd = 2,
  ylim = c(0, 1),
  xlab = "t", ylab = "probability",
  main = "Example 14.10: Euler state probabilities"
)
lines(out$t, out$tp01, lwd = 2, lty = 2)
lines(out$t, out$tp02, lwd = 2, lty = 3)
legend(
  "right",
  legend = c(expression(tp[x]^{"00"}),
             expression(tp[x]^{"01"}),
             expression(tp[x]^{"02"})),
  lwd = 2, lty = c(1, 2, 3), bty = "n"
)

# Plot discounted contribution streams
k <- 1:20
death_contrib <- 10000 * v^k * out$tp00[1:20] * q_death_step
disab_contrib <- 1000 * v^k * out$tp01[2:21]

plot(
  k / 10, death_contrib, type = "h", lwd = 2,
  xlab = "time", ylab = "discounted contribution",
  main = "Example 14.10: APV contribution streams"
)
lines(k / 10, disab_contrib, type = "h", lwd = 2, lty = 2)
legend(
  "topright",
  legend = c("death benefit", "disability benefit"),
  lwd = 2, lty = c(1, 2), bty = "n"
)


# Chapter 14 R Check: Example 14.13
# Constant-force permanent disability model

delta <- 0.05
mu01 <- 0.005
mu02 <- 0.02
mu12 <- 0.10

t <- seq(0, 200, by = 0.01)

# State-0 survival
tp00 <- exp(-(mu01 + mu02) * t)

# (a) Probability of ever becoming disabled
prob_disab <- sum(diff(t) * (head(tp00 * mu01, -1) + tail(tp00 * mu01, -1)) / 2)

# (b) Expected time spent disabled
# = P(enter state 1) * E(length in state 1 | entered)
exp_time_disabled <- prob_disab * (1 / mu12)

# (c) APV
integrand_death <- 100000 * exp(-delta * t) * tp00 * mu02
integrand_disab <- 500000 * exp(-delta * t) * tp00 * mu01

apv_death <- sum(diff(t) * (head(integrand_death, -1) + tail(integrand_death, -1)) / 2)
apv_disab <- sum(diff(t) * (head(integrand_disab, -1) + tail(integrand_disab, -1)) / 2)
apv_total <- apv_death + apv_disab

out <- c(
  prob_disab = prob_disab,
  exp_time_disabled = exp_time_disabled,
  apv_death = apv_death,
  apv_disab = apv_disab,
  apv_total = apv_total
)
print(out)

# Plot APV integrands
plot(
  t, integrand_death, type = "l", lwd = 2,
  xlab = "t", ylab = "discounted contribution density",
  main = "Example 14.13: APV integrands"
)
lines(t, integrand_disab, lwd = 2, lty = 2)
legend(
  "topright",
  legend = c("death while healthy", "disability payment"),
  lwd = 2, lty = c(1, 2), bty = "n"
)




# =========================================================
# Chapter 14 R Check: Exercise 14-8
# Cumulative withdrawal probability from Example 14.3 data
# =========================================================

library(mqriskR)

q1 <- c(.02, .03, .04, .05, .06)
q2 <- c(.30, .20, .20, .10, .00)
p_tau <- c(.68, .77, .76, .85, .94)

# cumulative withdrawal probability through duration n
cum_withdrawal <- cumsum(c(
  q2[1],
  p_tau[1] * q2[2],
  p_tau[1] * p_tau[2] * q2[3],
  p_tau[1] * p_tau[2] * p_tau[3] * q2[4],
  p_tau[1] * p_tau[2] * p_tau[3] * p_tau[4] * q2[5]
))

out <- data.frame(
  n = 1:5,
  cum_withdrawal = cum_withdrawal
)

print(out)

# value requested in part (b)
q3_withdrawal <- cum_withdrawal[3]
print(q3_withdrawal)

plot(
  out$n, out$cum_withdrawal, type = "b", pch = 19,
  xlab = "Duration n",
  ylab = "Cumulative withdrawal probability",
  main = "Exercise 14-8: cumulative withdrawal probability"
)
abline(h = q3_withdrawal, lty = 2)



# =========================================================
# Chapter 14 R Check: Exercise 14-14
# Euler approximation for employment / unemployment model
# =========================================================

library(mqriskR)

mu01 <- function(t) 0.20 + 0.0002 * t^2   # employed -> unemployed
mu02 <- function(t) 0.05                  # employed -> deceased
mu10 <- function(t) 0.80 - 0.04 * t       # unemployed -> employed
mu12 <- function(t) 0.05                  # unemployed -> deceased

out_h1 <- tp00_tp01_euler(
  h = 1, n = 20,
  mu01 = mu01, mu02 = mu02, mu10 = mu10, mu12 = mu12
)

out_h05 <- tp00_tp01_euler(
  h = 0.5, n = 10,
  mu01 = mu01, mu02 = mu02, mu10 = mu10, mu12 = mu12
)

out_h01 <- tp00_tp01_euler(
  h = 0.1, n = 10,
  mu01 = mu01, mu02 = mu02, mu10 = mu10, mu12 = mu12
)

# comparison at t = 10
comp <- data.frame(
  step = c(1, 0.5, 0.1),
  tp00_t10 = c(
    subset(out_h1, abs(t - 10) < 1e-12)$tp00,
    subset(out_h05, abs(t - 10) < 1e-12)$tp00,
    subset(out_h01, abs(t - 10) < 1e-12)$tp00
  ),
  tp01_t10 = c(
    subset(out_h1, abs(t - 10) < 1e-12)$tp01,
    subset(out_h05, abs(t - 10) < 1e-12)$tp01,
    subset(out_h01, abs(t - 10) < 1e-12)$tp01
  )
)

print(comp)

# plot for h = 0.1
plot(
  out_h01$t, out_h01$tp00, type = "l", lwd = 2,
  ylim = c(0, 1),
  xlab = "t", ylab = "Probability",
  main = "Exercise 14-14: Euler approximations (h = 0.1)"
)
lines(out_h01$t, out_h01$tp01, lwd = 2, lty = 2)
lines(out_h01$t, out_h01$tp02, lwd = 2, lty = 3)
legend(
  "topright",
  legend = c(expression(tp[x]^{"00"}),
             expression(tp[x]^{"01"}),
             expression(tp[x]^{"02"})),
  lwd = 2, lty = c(1, 2, 3), bty = "n"
)


# =========================================================
# Chapter 14 R Check: Exercise 14-15
# APV of monthly unemployment benefit
# =========================================================

library(mqriskR)

mu01 <- function(t) 0.20 + 0.0002 * t^2
mu02 <- function(t) 0.05
mu10 <- function(t) 0.80 - 0.04 * t
mu12 <- function(t) 0.05

h <- 1 / 12
n <- 20

out <- tp00_tp01_euler(
  h = h, n = n,
  mu01 = mu01, mu02 = mu02, mu10 = mu10, mu12 = mu12
)

j <- 0.04 / 12
k <- 1:240

apv_unemp <- 1400 * sum((1 + j)^(-k) * out$tp01[2:(240 + 1)])
print(apv_unemp)

contrib <- 1400 * (1 + j)^(-k) * out$tp01[2:(240 + 1)]

plot(
  k / 12, contrib, type = "l", lwd = 2,
  xlab = "Time (years)",
  ylab = "Discounted monthly contribution",
  main = "Exercise 14-15: APV contribution pattern"
)
abline(h = 0, lty = 3)



# =========================================================
# Chapter 14 R Check: Exercises 14-19 and 14-20
# Discrete-time Markov chain probabilities for the CCRC model
# =========================================================

library(mqriskR)

P <- matrix(
  c(0.94, 0.03, 0.02, 0.01,
    0.50, 0.30, 0.18, 0.02,
    0.00, 0.00, 0.93, 0.07,
    0.00, 0.00, 0.00, 1.00),
  nrow = 4, byrow = TRUE
)

# Exercise 14-19: P(X3 = 0 | X0 = 0)
p00_3 <- markov_nstep_prob(P, n = 3, i = 1, j = 1)

# Exercise 14-20: P(X3 = 2 | X0 = 0)
p02_3 <- markov_nstep_prob(P, n = 3, i = 1, j = 3)

print(c(p00_3 = p00_3, p02_3 = p02_3))

# plot selected n-step probabilities from state 0
n_vals <- 0:12
p00 <- sapply(n_vals, function(n) markov_nstep_prob(P, n = n, i = 1, j = 1))
p01 <- sapply(n_vals, function(n) markov_nstep_prob(P, n = n, i = 1, j = 2))
p02 <- sapply(n_vals, function(n) markov_nstep_prob(P, n = n, i = 1, j = 3))
p03 <- sapply(n_vals, function(n) markov_nstep_prob(P, n = n, i = 1, j = 4))

plot(
  n_vals, p00, type = "l", lwd = 2,
  ylim = c(0, 1),
  xlab = "Number of months",
  ylab = "Probability",
  main = "Exercises 14-19 and 14-20: n-step state probabilities"
)
lines(n_vals, p01, lwd = 2, lty = 2)
lines(n_vals, p02, lwd = 2, lty = 3)
lines(n_vals, p03, lwd = 2, lty = 4)
legend(
  "right",
  legend = c("State 0", "State 1", "State 2", "State 3"),
  lwd = 2, lty = c(1, 2, 3, 4), bty = "n"
)


# =========================================================
# Chapter 14 R Check: Exercise 14-23
# Risk-class transition matrix, projections, and average claim costs
# =========================================================

library(mqriskR)

P <- matrix(
  c(.826, .168, .006,
    .345, .617, .038,
    .111, .500, .389),
  nrow = 3, byrow = TRUE
)

pi1 <- c(.675, .303, .022)
claim_costs_Z <- c(500, 10000, 50000)

pi2 <- as.numeric(pi1 %*% P)
pi3 <- as.numeric(pi2 %*% P)

avg_Z  <- sum(c(.695, .287, .018) * claim_costs_Z)
avg_Z3 <- (1.05)^3 * sum(pi3 * claim_costs_Z)

print(P)
print(rbind(pi1 = pi1, pi2 = pi2, pi3 = pi3))
print(c(avg_Z = avg_Z, avg_Z3 = avg_Z3))

# plot projected class proportions
year <- 1:3
matplot(
  year,
  rbind(pi1, pi2, pi3),
  type = "l", lwd = 2, lty = 1:3,
  xlab = "Year index (1 = Z+1)",
  ylab = "Proportion",
  main = "Exercise 14-23: projected risk-class proportions"
)
legend(
  "topright",
  legend = c("Low", "Moderate", "High"),
  lwd = 2, lty = 1:3, bty = "n"
)

# plot average claim cost path
avg_Z1 <- sum(pi1 * claim_costs_Z) * 1.05
avg_Z2 <- sum(pi2 * claim_costs_Z) * (1.05)^2
avg_Z3_path <- sum(pi3 * claim_costs_Z) * (1.05)^3

avg_path <- c(avg_Z, avg_Z1, avg_Z2, avg_Z3_path)
plot(
  0:3, avg_path, type = "b", pch = 19,
  xlab = "Years after Z",
  ylab = "Average claim cost",
  main = "Exercise 14-23: average claim cost path"
)



# =========================================================
# Chapter 14 R Check: Exercise 14-25
# Mortality gain and withdrawal gain
# =========================================================

library(mqriskR)

# assumed values
V9 <- 115.00
V10 <- 128.83
G <- 16
e <- 3
i <- 0.06
DB <- 1000
WB <- 110

# assumed and actual decrement rates
q_d_assumed <- 0.01
q_w_assumed <- 0.10
q_d_actual <- 15 / 1000
q_w_actual <- 100 / 985

# gain from mortality: actual mortality, assumed withdrawal
GM <- gain_loss_md(
  Vt = V9,
  G = G,
  r = 0,
  e = e,
  i = i,
  b1 = DB,
  b2 = WB,
  s1 = 0,
  s2 = 0,
  q1 = q_d_assumed,
  q2 = q_w_assumed,
  Vt1 = V10,
  year_end_cause2 = TRUE,
  q1prime = q_d_actual,
  q2prime = q_w_assumed
)

# gain from mortality + withdrawal: actual mortality, actual withdrawal
GMW <- gain_loss_md(
  Vt = V9,
  G = G,
  r = 0,
  e = e,
  i = i,
  b1 = DB,
  b2 = WB,
  s1 = 0,
  s2 = 0,
  q1 = q_d_assumed,
  q2 = q_w_assumed,
  Vt1 = V10,
  year_end_cause2 = TRUE,
  q1prime = q_d_actual,
  q2prime = q_w_actual
)

GW <- GMW - GM

out <- c(
  gain_mortality = GM,
  gain_mortality_and_withdrawal = GMW,
  gain_withdrawal = GW
)
print(out)

# simple comparison plot
assumed_actual <- rbind(
  assumed = c(q_d_assumed, q_w_assumed),
  actual  = c(q_d_actual, q_w_actual)
)
colnames(assumed_actual) <- c("Mortality", "Withdrawal")

barplot(
  t(assumed_actual), beside = TRUE,
  main = "Exercise 14-25: assumed vs actual decrement rates",
  ylab = "Rate",
  legend.text = TRUE,
  args.legend = list(bty = "n", x = "topright")
)

# Chapter 16 R Check: Example 16.7
# Catch-up premium under stipulated premium guarantee

library(mqriskR)

stip_premium <- 10000
years_elapsed <- 10
cum_paid <- 90000

required_to_date <- stip_premium * years_elapsed
shortfall <- required_to_date - cum_paid
payment_needed_now <- shortfall + stip_premium

out <- data.frame(
  years_elapsed = years_elapsed,
  required_to_date = required_to_date,
  cumulative_paid = cum_paid,
  shortfall = shortfall,
  payment_needed_now = payment_needed_now
)

print(out)

# Simple plot
barplot(
  height = c(required_to_date, cum_paid),
  names.arg = c("Required cumulative", "Actual cumulative"),
  main = "Example 16.7: Stipulated Premium Catch-Up",
  ylab = "Amount"
)
abline(h = required_to_date, lty = 2)



# Chapter 16 R Check: Example 16.8
# Lapse timing with and without shadow fund guarantee

library(mqriskR)

yr <- 1:6
contract_av <- c(1000, 800, 400, 100, 0, 0)
surr_charge <- c(900, 800, 600, 400, 200, 100)
cash_value <- c(100, 0, 0, 0, 0, 0)
shadow_av <- c(1100, 900, 700, 500, 200, 0)

lapse_no_guarantee <- yr[min(which(contract_av == 0))]
lapse_shadow <- yr[min(which(shadow_av == 0))]

out <- data.frame(
  year = yr,
  contract_av = contract_av,
  surrender_charge = surr_charge,
  cash_value = cash_value,
  shadow_av = shadow_av
)

print(out)
cat("Lapse year without secondary guarantee:", lapse_no_guarantee, "\n")
cat("Lapse year with shadow-fund secondary guarantee:", lapse_shadow, "\n")




# Chapter 16 R Check: Example 16.9
# Basic UL reserving quantities using package helpers where available

library(mqriskR)

face_per_1000 <- 100
AV9_per_1000 <- 57.60
GMF9_per_1000 <- 140.40
gmp_per_1000 <- 14.49
expense_rate <- 0.04
guar_charge10_per_1000 <- 11.80
curr_charge10_per_1000 <- 10.76
curr_i <- 0.05
guar_i <- 0.03
pvfb_minus_pvfp_per_1000 <- 70

# Account value roll-forward (computed directly from the example)
AV10_per_1000 <- (AV9_per_1000 - curr_charge10_per_1000) * (1 + curr_i)

# Package helpers
GMF10_per_1000 <- GMF_rollforward_ul(
  GMF_prev = GMF9_per_1000,
  GMP = gmp_per_1000,
  r = expense_rate,
  policy_charge = guar_charge10_per_1000,
  i = guar_i
)

AV10 <- AV10_per_1000 * face_per_1000
GMF10 <- GMF10_per_1000 * face_per_1000

r10 <- rt_ul(AV = AV10, GMF = GMF10)

pre_floor_CRVM_per_1000 <- Vprefloor_crvm_ul(
  r = r10,
  pvfb_minus_pvfp = pvfb_minus_pvfp_per_1000
)

pre_floor_CRVM <- pre_floor_CRVM_per_1000 * face_per_1000

out <- data.frame(
  AV10_per_1000 = AV10_per_1000,
  GMF10_per_1000 = GMF10_per_1000,
  AV10 = AV10,
  GMF10 = GMF10,
  r10 = r10,
  pre_floor_CRVM_per_1000 = pre_floor_CRVM_per_1000,
  pre_floor_CRVM = pre_floor_CRVM
)

print(round(out, 5))

barplot(
  height = c(AV10, GMF10, pre_floor_CRVM),
  names.arg = c("AV10", "GMF10", "Pre-floor CRVM"),
  main = "Example 16.9: UL Reserving Quantities",
  ylab = "Amount"
)




# Chapter 16 R Check: Example 16.10
# Implied guaranteed rates for IUL

library(mqriskR)

valuation_rate <- 0.04
option_cost_initial <- 0.05
option_cost_renewal <- 0.02
guaranteed_floor <- 0.00

igr_initial <- guaranteed_floor + option_cost_initial * (1 + valuation_rate)
igr_renewal <- guaranteed_floor + option_cost_renewal * (1 + valuation_rate)

out <- data.frame(
  implied_guaranteed_rate_initial = igr_initial,
  implied_guaranteed_rate_renewal = igr_renewal
)

print(out)

barplot(
  height = c(igr_initial, igr_renewal),
  names.arg = c("Initial term", "Beyond initial term"),
  main = "Example 16.10: Implied Guaranteed Rates",
  ylab = "Rate"
)




# Chapter 16 R Check: Example 16.11
# AG 38 reserve under shadow fund method using package helpers

library(mqriskR)

shadow_fund <- 60000
nsp_full_funding <- 100000
valuation_nsp <- 150000
surrender_charge <- 5000
basic_reserve <- 10000
deficiency_reserve <- 0

prefunding_ratio <- ag38_prefunding_ratio(
  excess_payment = shadow_fund,
  nsp_required = nsp_full_funding
)

ag38_out <- ag38_reserve_ul(
  basic_reserve = basic_reserve,
  deficiency_reserve = deficiency_reserve,
  excess_payment = shadow_fund,
  nsp_required = nsp_full_funding,
  valuation_nsp = valuation_nsp,
  surrender_charge = surrender_charge
)

out <- data.frame(
  prefunding_ratio = prefunding_ratio,
  net_amount_additional = ag38_out$net_amount_additional,
  reduced_deficiency_reserve = ag38_out$reduced_deficiency_reserve,
  reserve_step8 = ag38_out$step8_reserve,
  ag38_reserve = ag38_out$final_reserve
)

print(out)

barplot(
  height = c(shadow_fund, nsp_full_funding, ag38_out$final_reserve),
  names.arg = c("Shadow fund", "Full-funding NSP", "AG38 reserve"),
  main = "Example 16.11: AG 38 Reserve Components",
  ylab = "Amount"
)


# Chapter 16 R Check: Exercise 16-3
# Corridor-factor test for Type A universal life

library(mqriskR)

B <- 100000
G <- 1500
r <- 0.025
e <- 50
qx <- 0.00278
iq <- 0.06
ic <- 0.07
f <- 2.50
AV_prev <- 40000

fund_before_coi <- AV_prev + G * (1 - r) - e

# Ordinary Type A COI from the formula in Exercise 16-3(a)
coi_typeA <- (
  (qx / (1 + iq)) * (B - fund_before_coi * (1 + ic))
) / (
  1 - (qx / (1 + iq)) * (1 + ic)
)

# Corridor-factor COI from the formula in Exercise 16-3(a)
coi_cf <- (
  (qx / (1 + iq)) * fund_before_coi * (1 + ic) * (f - 1)
) / (
  1 + (qx / (1 + iq)) * (1 + ic) * (f - 1)
)

coi_used <- max(coi_typeA, coi_cf)
corridor_applies <- coi_cf > coi_typeA

AV_t <- (fund_before_coi - coi_used) * (1 + ic)

out <- data.frame(
  coi_typeA = coi_typeA,
  coi_cf = coi_cf,
  coi_used = coi_used,
  corridor_applies = corridor_applies,
  AV_t = AV_t
)

print(round(out, 5))

barplot(
  height = c(coi_typeA, coi_cf),
  names.arg = c("Ordinary COI", "Corridor COI"),
  main = "Exercise 16-3: Competing COI Values",
  ylab = "Cost of insurance"
)




# Chapter 16 R Check: Exercise 16-8
# Indexed UL credited rates and cash values using package helpers

library(mqriskR)

B <- 100000
G <- rep(1000, 3)
premium_expense <- 0.04
admin_fee <- 50
index <- c(1000, 1080, 1200, 1100)
coi_per_1000 <- c(2.0, 3.0, 4.0)
surr_per_1000 <- c(5.0, 4.0, 3.0)

raw <- iP_eiul(index)
credited <- i_credit_eiul(
  i_raw = raw,
  part = 1.00,
  floor = 0.01,
  cap = 0.10
)

AV <- numeric(3)
cash_value <- numeric(3)

for (t in 1:3) {
  prior <- if (t == 1) 0 else AV[t - 1]
  net_premium <- G[t] * (1 - premium_expense)
  amount_at_risk <- B - (prior + net_premium)
  coi <- (coi_per_1000[t] / 1000) * amount_at_risk
  AV[t] <- (prior + net_premium - coi - admin_fee) * (1 + credited[t])
  cash_value[t] <- AV[t] - surr_per_1000[t] * (B / 1000)
}

out <- data.frame(
  year = 1:3,
  raw = raw,
  credited = credited,
  AV = AV,
  cash_value = cash_value
)

print(round(out, 5))

plot(
  1:3, raw, type = "b", pch = 16,
  ylim = range(c(raw, credited)),
  xlab = "Year", ylab = "Rate",
  main = "Exercise 16-8: Raw vs Credited Rates"
)
lines(1:3, credited, type = "b", pch = 17)
abline(h = 0.10, lty = 2)
abline(h = 0.01, lty = 2)
legend(
  "topright",
  legend = c("Raw index growth", "Credited rate"),
  lty = 1, pch = c(16, 17), bty = "n"
)



# Chapter 16 R Check: Exercise 16-9
# APV of premiums and pure endowment under mortality and withdrawal

library(mqriskR)

qd <- c(.001, .002, .003, .004, .005)
qw <- c(.02, .02, .03, .04, .05)
G <- c(20000, 25000, 25000, 30000, 20000)
i <- 0.05
B <- 100000

p_tau <- pxtau_ul(qd = qd, qw = qw, year_end_withdrawal = TRUE)
tp_tau <- tpxtau_ul(qd = qd, qw = qw, year_end_withdrawal = TRUE)

apvp <- G[1] +
  G[2] * tp_tau[1] / (1 + i)^1 +
  G[3] * tp_tau[2] / (1 + i)^2 +
  G[4] * tp_tau[3] / (1 + i)^3 +
  G[5] * tp_tau[4] / (1 + i)^4

apvb <- B * tp_tau[5] / (1 + i)^5

out <- data.frame(
  year = 1:5,
  p_tau = p_tau,
  tp_tau = tp_tau
)

print(round(out, 5))
cat("APV of premiums =", round(apvp, 2), "\n")
cat("APV of pure endowment =", round(apvb, 2), "\n")

plot(
  1:5, tp_tau, type = "b", pch = 16,
  xlab = "Policy year", ylab = "Cumulative in-force probability",
  main = "Exercise 16-9: Cumulative Persistency"
)


# Chapter 16 R Check: Exercise 16-11
# Revised AV10, GMF10, and r10 after extra premium using package helpers

library(mqriskR)

face_per_1000 <- 100
AV10_old <- 4918.20
extra_premium <- 1000
expense_rate <- 0.04
curr_i <- 0.05

GMF9_per_1000 <- 140.40
GMP_per_1000 <- 14.49
guar_expense_rate <- 0.04
guar_charge10_per_1000 <- 11.80
guar_i <- 0.03

AV10_new <- AV10_old + extra_premium * (1 - expense_rate) * (1 + curr_i)

GMF10_per_1000 <- GMF_rollforward_ul(
  GMF_prev = GMF9_per_1000,
  GMP = GMP_per_1000,
  r = guar_expense_rate,
  policy_charge = guar_charge10_per_1000,
  i = guar_i
)

GMF10 <- GMF10_per_1000 * face_per_1000

r10 <- rt_ul(
  AV = AV10_new,
  GMF = GMF10
)

out <- data.frame(
  AV10_new = AV10_new,
  GMF10 = GMF10,
  r10 = r10
)

print(round(out, 5))

barplot(
  height = c(AV10_new, GMF10),
  names.arg = c("AV10", "GMF10"),
  main = "Exercise 16-11: Revised AV and GMF",
  ylab = "Amount"
)

library(mqriskR)

# R Check for Example 13.3
# Rebuild the table from Example 13.1
x <- 45:50
qmat <- cbind(
  q1 = c(.011, .012, .013, .014, .015, .016),
  q2 = c(.100, .100, .100, .100, .100, .100)
)

tbl <- md_table(x = x, qxj = qmat, radix = 1000)

# Person age 46
x0 <- 46
l0 <- tbl$ltau[tbl$x == x0]

# Joint probabilities Pr(K_x = k, J_x = j)
k_vals <- 0:(max(tbl$x) - x0)
joint_q1 <- sapply(k_vals, function(k) {
  row <- tbl[tbl$x == x0 + k, ]
  row$d1 / l0
})
joint_q2 <- sapply(k_vals, function(k) {
  row <- tbl[tbl$x == x0 + k, ]
  row$d2 / l0
})

joint_df <- data.frame(
  k = k_vals,
  cause1 = joint_q1,
  cause2 = joint_q2,
  total = joint_q1 + joint_q2
)

print(joint_df)

# Requested probabilities from Example 13.3
pr_k2_j1 <- joint_df$cause1[joint_df$k == 2]
pr_k3 <- joint_df$total[joint_df$k == 3]

c(
  `Pr(K=2, J=1)` = pr_k2_j1,
  `Pr(K=3)` = pr_k3
)

# Plot the joint and marginal probabilities
matplot(
  joint_df$k,
  cbind(joint_df$cause1, joint_df$cause2, joint_df$total),
  type = "b", pch = 1,col = 1:3,lty = 1:3,
  xlab = "k",
  ylab = "Probability",
  main = "Joint and marginal probabilities for Example 13.3"
)
legend(
  "topright",
  legend = c("Pr(K=k, J=1)", "Pr(K=k, J=2)", "Pr(K=k)"),
  lty = 1:3, pch = 1, bty = "n", col = 1:3
)



# R Check for Example 13.7
# Decrement 1: move away during the year
# Decrement 2: pass to next grade at year-end only

q1_dep <- 0.02
q2_dep <- 0.96

# Out of 100 starting students:
# 2 leave by Decrement 1 during the year, so 98 remain exposed to year-end passing.
# Of those 98, 96 pass.
q2_prime <- 96 / 98

c(
  q1_dep = q1_dep,
  q2_dep = q2_dep,
  q2_prime = q2_prime
)

# Simple comparison plot
vals <- c(q1_dep, q2_dep, q2_prime)
names(vals) <- c("q^(1)", "q^(2)", "q'^(2)")
barplot(
  vals,
  ylab = "Probability",
  main = "Dependent vs independent probabilities in Example 13.7"
)

# A small sensitivity check:
# Suppose Decrement 1 varied from 0 to 0.10 while observed q^(2)=0.96 stayed fixed.
q1_grid <- seq(0, 0.10, by = 0.005)
q2prime_grid <- 0.96 / (1 - q1_grid)

plot(
  q1_grid, q2prime_grid, type = "l",
  xlab = "Dependent probability of Decrement 1",
  ylab = "Implied single-decrement probability q'^(2)",
  main = "Sensitivity of q'^(2) to competing Decrement 1"
)



# R Check for Example 13.10

library(mqriskR)

# Example-style numerical comparison for the summary table in Example 13.10

q1_prime <- 0.20
q2_prime <- 0.10

# SUDD: independent -> dependent
dep_sudd <- qx_dep_sudd(q1prime = q1_prime, q2prime = q2_prime)

# Constant force: independent -> dependent
dep_cf <- qx_dep_cf(c(q1_prime, q2_prime))

# MUDD / CF: dependent -> independent
ind_from_mudd_on_sudd <- qxprime_mudd(unname(dep_sudd))
ind_from_mudd_on_cf <- qxprime_mudd(unname(dep_cf))

# SUDD inverse: dependent -> independent
ind_from_sudd <- qxprime_sudd(q1 = dep_sudd["q1"], q2 = dep_sudd["q2"])

out <- rbind(
  `Independent input` = c(q1_prime, q2_prime),
  `Dependent under SUDD` = unname(dep_sudd),
  `Dependent under CF` = unname(dep_cf),
  `Recovered by MUDD from SUDD-dependent` = unname(ind_from_mudd_on_sudd),
  `Recovered by MUDD from CF-dependent` = unname(ind_from_mudd_on_cf),
  `Recovered by SUDD from SUDD-dependent` = unname(ind_from_sudd)
)

colnames(out) <- c("Cause 1", "Cause 2")
print(round(out, 6))




# Exercise 13-1
# Reconstruct the school table numerically

radix <- 1000

# Year 1
q_fail_1 <- 0.40
q_withdraw_1 <- 0.20
p_complete_1 <- 1 - q_fail_1 - q_withdraw_1

start_yr2 <- radix * p_complete_1

# Year 2
# Let F = failures in year 2, C = complete year 2
# F + C = 0.70 * start_yr2
# F = 0.40 * C
C2 <- (0.70 * start_yr2) / 1.40
F2 <- 0.40 * C2
W2 <- start_yr2 - F2 - C2

# Year 3
start_yr3 <- C2
p_complete_3 <- 0.60
C3 <- start_yr3 * p_complete_3

# ten times as many complete Year 2 as fail during Year 3
F3 <- C2 / 10
W3 <- start_yr3 - C3 - F3

total_withdraw <- radix * q_withdraw_1 + W2 + W3
prob_withdraw <- total_withdraw / radix

out <- c(
  start_yr2 = start_yr2,
  C2 = C2,
  F2 = F2,
  W2 = W2,
  start_yr3 = start_yr3,
  C3 = C3,
  F3 = F3,
  W3 = W3,
  total_withdraw = total_withdraw,
  prob_withdraw = prob_withdraw
)

print(round(out, 4))

barplot(
  c(Year1 = radix * q_withdraw_1, Year2 = W2, Year3 = W3),
  ylab = "Expected number of withdrawals",
  main = "Exercise 13-1: Withdrawals by year"
)



# Exercise 13-3

t_grid <- seq(0, 12, by = 0.1)

# Cause 1: mortality under uniform model with omega = 100, current age 50
tpx1 <- function(t) {
  out <- 1 - t / 50
  out[t > 50] <- 0
  out
}

# Cause 2: leaving academic employment with constant force 0.05
tpx2 <- function(t) exp(-0.05 * t)

# Total survival in multiple-decrement model
tpx_tau <- function(t) tpx1(t) * tpx2(t)

p5 <- tpx_tau(5)
p10 <- tpx_tau(10)
ans <- p5 - p10

c(
  p5 = p5,
  p10 = p10,
  probability = ans
)

plot(
  t_grid, tpx_tau(t_grid), type = "l",
  xlab = "t",
  ylab = expression(t*p[50]^{(tau)}),
  main = "Exercise 13-3: Survival in academic employment"
)
abline(v = c(5, 10), lty = 2)



# Exercise 13-5

t_grid <- seq(0, 49.9, by = 0.1)

tpx1_prime <- function(t) (75 - t) / 75
tpx2_prime <- function(t) ((50 - t) / 50)^2
tpx_tau <- function(t) tpx1_prime(t) * tpx2_prime(t)

mu1 <- function(t) 1 / (75 - t)
mu2 <- function(t) 2 / (50 - t)

fTJ1 <- function(t) tpx_tau(t) * mu1(t)
fTJ2 <- function(t) tpx_tau(t) * mu2(t)

# One-year decrement probabilities
qx1 <- integrate(fTJ1, lower = 0, upper = 1)$value
qx2 <- integrate(fTJ2, lower = 0, upper = 1)$value
qxtau <- qx1 + qx2

# Single-decrement probabilities
qx1_prime <- 1 - tpx1_prime(1)
qx2_prime <- 1 - tpx2_prime(1)

# Ultimate probabilities by cause
fJ1 <- integrate(fTJ1, lower = 0, upper = 50)$value
fJ2 <- integrate(fTJ2, lower = 0, upper = 50)$value

# Conditional cause probabilities at t = 1
fJ_given_T_1 <- mu1(1) / (mu1(1) + mu2(1))
fJ_given_T_2 <- mu2(1) / (mu1(1) + mu2(1))

out <- c(
  qx1 = qx1,
  qx2 = qx2,
  qxtau = qxtau,
  qx1_prime = qx1_prime,
  qx2_prime = qx2_prime,
  fJ1 = fJ1,
  fJ2 = fJ2,
  fJ_given_T_1 = fJ_given_T_1,
  fJ_given_T_2 = fJ_given_T_2
)

print(round(out, 6))

matplot(
  t_grid,
  cbind(fTJ1(t_grid), fTJ2(t_grid)),
  type = "l", lty = 1,col=1:2,
  xlab = "t",
  ylab = "Density",
  main = "Exercise 13-5: Cause-specific joint densities"
)
legend(
  "topright",
  legend = c(expression(f[T,J](t,1)), expression(f[T,J](t,2))),
  lty = 1, bty = "n",col=1:2,
)



# Exercise 13-6

radix <- 1000
total_leaving_first_year <- 48
survive_first_year <- 1 - total_leaving_first_year / radix

# total annual force
mu_tau <- -log(survive_first_year)

# given mu^(2) = 0.04
mu1 <- mu_tau - 0.04

# annual dependent probability for Cause 1
qx1 <- (mu1 / mu_tau) * (1 - exp(-mu_tau))

# expected survivors entering fourth year
l3 <- radix * survive_first_year^3

expected_failures_fourth_year <- l3 * qx1

c(
  survive_first_year = survive_first_year,
  mu_tau = mu_tau,
  mu1 = mu1,
  qx1 = qx1,
  l3 = l3,
  expected_failures_fourth_year = expected_failures_fourth_year
)

years <- 0:4
survivors <- radix * survive_first_year^years

plot(
  years, survivors, type = "b",
  xlab = "Completed years",
  ylab = "Expected survivors",
  main = "Exercise 13-6: Expected survivors by year"
)





# Exercise 13-13

library(mqriskR)

q1_prime <- c(0.10, 0.20, 0.20)
q2_prime <- c(0.25, 0.20, 0.10)

dep <- t(mapply(
  function(a, b) qx_dep_sudd(q1prime = a, q2prime = b),
  q1_prime, q2_prime
))

out <- data.frame(
  year = 1:3,
  q1_prime = q1_prime,
  q2_prime = q2_prime,
  q1 = dep[, "q1"],
  q2 = dep[, "q2"]
)

print(round(out, 4))

matplot(
  out$year,
  cbind(out$q1_prime, out$q2_prime, out$q1, out$q2),
  type = "b", pch = 1, col=1:4,
  xlab = "Year",
  ylab = "Probability",
  main = "Exercise 13-13: Independent vs dependent probabilities"
)
legend(
  "topright",
  legend = c("q1 prime", "q2 prime", "q1", "q2"),
  lty = 1, pch = 1, bty = "n", col=1:4
)



# Exercise 13-17

radix <- 1000
q1_prime <- 0.100
q2_prime <- 0.125
t_event <- 0.70

# Decrement 1 acts alone up to time 0.70 under UDD
q1_upto_event <- t_event * q1_prime
survivors_at_event <- radix * (1 - q1_upto_event)

# Decrement 2 occurs at time 0.70
d2 <- survivors_at_event * q2_prime
q2_dep <- d2 / radix

c(
  q1_upto_event = q1_upto_event,
  survivors_at_event = survivors_at_event,
  d2 = d2,
  q2_dep = q2_dep
)

barplot(
  c(
    start = radix,
    survivors_at_event = survivors_at_event,
    decrement2 = d2
  ),
  main = "Exercise 13-17: Cohort tracking to the point decrement",
  ylab = "Count"
)

# Chapter 18 - R Check for Example 18.3
# Early-retirement reduction factors and immediate pension amounts

library(mqriskR)

# Example inputs
x <- 35
z <- 65
CASx <- 60000
p <- 2
g <- 0.04

# Retirement ages y = 60, 61, 62, 63, 64
y <- 60:64

# Salary scale helper
salary_at_age <- function(age, base_age = x, base_salary = CASx, growth = g) {
  base_salary * (1 + growth)^(age - base_age)
}

# Approximate projected FAS at age y + 1/2
# using the same weighting pattern as the text
FAS_half <- sapply(y, function(yy) {
  s_y3 <- salary_at_age(yy - 3)
  s_y2 <- salary_at_age(yy - 2)
  s_y1 <- salary_at_age(yy - 1)
  s_y  <- salary_at_age(yy)
  (0.5 * s_y3 + s_y2 + s_y1 + 0.5 * s_y) / 3
})

# Projected benefit payable at NRA
PAB_half <- 0.02 * (y + 0.5 - x) * FAS_half

# Immediate benefit using 5% per year early reduction
reduction_factor <- 1 - 0.05 * (z - (y + 0.5))
immediate_benefit <- PAB_half * reduction_factor

out_18_3 <- data.frame(
  y = y,
  retirement_age = y + 0.5,
  FAS_half = FAS_half,
  PAB_half = PAB_half,
  reduction_factor = reduction_factor,
  immediate_benefit = immediate_benefit
)

print(round(out_18_3, 2))

# Plot: reduction factor by retirement age
plot(
  out_18_3$retirement_age, out_18_3$reduction_factor,
  type = "b",
  xlab = "Retirement age",
  ylab = "Reduction factor",
  main = "Example 18.3: Early-retirement reduction factor"
)

# Plot: immediate benefit by retirement age
plot(
  out_18_3$retirement_age, out_18_3$immediate_benefit,
  type = "b",
  xlab = "Retirement age",
  ylab = "Immediate annual benefit",
  main = "Example 18.3: Immediate pension by retirement age"
)



# Chapter 18 - R Check for Example 18.4
# Withdrawal-benefit APV decomposition by withdrawal age

library(mqriskR)

# Illustrative setup consistent with Example 18.4 structure
x <- 35
z <- 65
CASx <- 60000
p <- 2
g <- 0.04
adue_65 <- 12

# Withdrawal ages in Example 18.4
y <- 40:59

# Illustrative active-service survival probabilities to age y
# (replace with service-table values if available)
p_active <- cumprod(c(1, rep(0.97, length(y) - 1)))

# Illustrative withdrawal probabilities by attained age
q_withdrawal <- seq(0.03, 0.01, length.out = length(y))

# Illustrative post-withdrawal survival to NRA
p_withdrawn_to_65 <- 0.98^(z - y - 0.5)

# Salary helper
salary_at_age <- function(age, base_age = x, base_salary = CASx, growth = g) {
  base_salary * (1 + growth)^(age - base_age)
}

# Approximate projected benefit at withdrawal age y + 1/2, deferred to NRA
FAS_half <- sapply(y, function(yy) {
  s_y3 <- salary_at_age(yy - 3)
  s_y2 <- salary_at_age(yy - 2)
  s_y1 <- salary_at_age(yy - 1)
  s_y  <- salary_at_age(yy)
  (0.5 * s_y3 + s_y2 + s_y1 + 0.5 * s_y) / 3
})

PAB_half <- 0.02 * (y + 0.5 - x) * FAS_half

# Discount factor from age 35 to age 65 as in Eq. (18.12)
v30 <- 1 / (1.06^30)

# Contribution by withdrawal age
apv_contrib <- PAB_half * v30 * p_active * q_withdrawal * p_withdrawn_to_65 * adue_65

out_18_4 <- data.frame(
  y = y,
  PAB_half = PAB_half,
  p_active = p_active,
  q_withdrawal = q_withdrawal,
  p_withdrawn_to_65 = p_withdrawn_to_65,
  apv_contribution = apv_contrib
)

print(round(out_18_4, 4))
cat("Total illustrative APV =", round(sum(apv_contrib), 4), "\n")

# Plot: APV contribution by withdrawal age
plot(
  y, apv_contrib,
  type = "h",
  lwd = 2,
  xlab = "Withdrawal age y",
  ylab = "APV contribution",
  main = "Example 18.4: Withdrawal-benefit APV contribution by age"
)



# Chapter 18 - R Check for Example 18.5
# TUC versus PUC normal cost and accrued liability

library(mqriskR)

projected_benefit <- 50000
past_service <- 26
total_service <- 35
accrued_benefit <- 20000
current_year_accrual <- 1500

i <- 0.06
p_surv <- 0.92
adue_65 <- 12
v_to_ret <- 1 / (1 + i)^9

NC_TUC <- NC_TUC_db(
  accrual_benefit = current_year_accrual,
  v_to_ret = v_to_ret,
  p_surv = p_surv,
  adue_ret = adue_65
)

AAL_TUC <- AAL_TUC_db(
  accrued_benefit = accrued_benefit,
  v_to_ret = v_to_ret,
  p_surv = p_surv,
  adue_ret = adue_65
)

NC_PUC <- NC_PUC_db(
  projected_benefit = projected_benefit,
  total_service = total_service,
  v_to_ret = v_to_ret,
  p_surv = p_surv,
  adue_ret = adue_65
)

AAL_PUC <- AAL_PUC_db(
  projected_benefit = projected_benefit,
  past_service = past_service,
  total_service = total_service,
  v_to_ret = v_to_ret,
  p_surv = p_surv,
  adue_ret = adue_65
)

out_18_5 <- c(
  NC_TUC = NC_TUC,
  AAL_TUC = AAL_TUC,
  NC_PUC = NC_PUC,
  AAL_PUC = AAL_PUC
)

print(round(out_18_5, 2))

# Barplot comparison
barplot(
  height = out_18_5,
  beside = TRUE,
  main = "Example 18.5: TUC versus PUC",
  ylab = "Amount"
)


# Chapter 18 - R Check for Exercise 18-1

library(mqriskR)

# Base calculation
AV_65 <- AVz_dc(
  x = 25, z = 65, Sx = 40000,
  c = 0.08, i = 0.05, g = 0.035
)

Income_65 <- Income_dc(
  AVz = AV_65,
  adue_z = 13
)

RR_65 <- replacement_ratio_dc(
  x = 25, z = 65, Sx = 40000,
  c = 0.08, i = 0.05,
  adue_z = 13, g = 0.035
)

out_18_1 <- c(
  AV_65 = AV_65,
  Income_65 = Income_65,
  RR_65 = RR_65
)

print(round(out_18_1, 5))

# Sensitivity of replacement ratio to salary growth
g_grid <- seq(0.01, 0.06, by = 0.005)

RR_grid <- sapply(g_grid, function(g) {
  replacement_ratio_dc(
    x = 25, z = 65, Sx = 40000,
    c = 0.08, i = 0.05,
    adue_z = 13, g = g
  )
})

plot(
  g_grid, RR_grid,
  type = "b",
  xlab = "Salary growth rate",
  ylab = "Replacement ratio",
  main = "Exercise 18-1: Replacement ratio vs salary growth"
)



# Chapter 18 - R Check for Exercise 18-2

library(mqriskR)

c_target <- contribution_rate_target(
  x = 30, z = 65, Sx = 60000,
  RR_target = 0.50,
  i = 0.06, adue_z = 11,
  g = 0.04
)

RR_check <- replacement_ratio_dc(
  x = 30, z = 65, Sx = 60000,
  c = c_target, i = 0.06,
  adue_z = 11, g = 0.04
)

out_18_2 <- c(
  c_target = c_target,
  RR_check = RR_check
)

print(round(out_18_2, 5))

# Sensitivity to interest rate
i_grid <- seq(0.03, 0.08, by = 0.005)

c_grid <- sapply(i_grid, function(ii) {
  contribution_rate_target(
    x = 30, z = 65, Sx = 60000,
    RR_target = 0.50,
    i = ii, adue_z = 11,
    g = 0.04
  )
})

plot(
  i_grid, c_grid,
  type = "b",
  xlab = "Fund interest rate",
  ylab = "Required contribution rate",
  main = "Exercise 18-2: Contribution rate vs interest rate"
)


# Chapter 18 - R Check for Exercise 18-3

library(mqriskR)

# Salary path with 4% regular increases and 6% merit increases
ages <- 30:64
salary <- numeric(length(ages))
salary[1] <- 100000

for (j in 2:length(ages)) {
  salary[j] <- salary[j - 1] * 1.04
  if (ages[j] %in% c(31, 32, 33)) {
    salary[j] <- salary[j] * 1.06
  }
}

names(salary) <- ages

FAS_65 <- mean(tail(salary, 5))
PAB_fas <- 0.01 * FAS_65 * 35
final_salary <- salary[length(salary)]
RR_fas <- PAB_fas / final_salary

CAP_65 <- 0.01 * sum(salary)
ratio_cae_to_fas <- CAP_65 / PAB_fas

out_18_3 <- c(
  FAS_65 = FAS_65,
  PAB_fas = PAB_fas,
  final_salary = final_salary,
  RR_fas = RR_fas,
  CAP_65 = CAP_65,
  ratio_cae_to_fas = ratio_cae_to_fas
)

print(round(out_18_3, 5))

# Plot salary path
plot(
  ages, salary,
  type = "b",
  xlab = "Age",
  ylab = "Salary",
  main = "Exercise 18-3: Salary path"
)

# Plot comparison of benefits
barplot(
  c(FAS_plan_benefit = PAB_fas, CAE_plan_benefit = CAP_65),
  main = "Exercise 18-3: Pension formula comparison",
  ylab = "Annual benefit"
)



# Chapter 18 - R Check for Exercise 18-4

library(mqriskR)

i <- 0.06

# Inputs from the written solution
ten_E_55 <- 0.51147
adue_65 <- (1 - 0.36778) / (i / (1 + i))
adue_55 <- (1 - 0.24766) / (i / (1 + i))

ERRF_55 <- ten_E_55 * adue_65 / adue_55
informal_factor <- 1 - 0.05 * 10
relative_diff <- (informal_factor - ERRF_55) / ERRF_55

out_18_4 <- c(
  adue_65 = adue_65,
  adue_55 = adue_55,
  ERRF_55 = ERRF_55,
  informal_factor = informal_factor,
  relative_diff = relative_diff
)

print(round(out_18_4, 5))

# Plot actuarial vs informal factor
barplot(
  c(actuarial = ERRF_55, informal = informal_factor),
  main = "Exercise 18-4: Early-retirement reduction factors",
  ylab = "Reduction factor"
)




# Chapter 18 - R Check for Exercise 18-6
# Illustrative numerical version using reasonable assumptions
# Part (a) matches the book exactly.
# Parts (b) and (c) use illustrative service-table and annuity assumptions.

library(mqriskR)

# ------------------------------------------------------------
# Part (a): accrued benefit and benefit accrual
# ------------------------------------------------------------

AB_56 <- 0.01 * 5 * 150000
AB_57 <- 0.01 * 6 * 156000
BA <- AB_57 - AB_56

out_part_a <- c(
  AB_56 = AB_56,
  AB_57 = AB_57,
  BA = BA
)

print(out_part_a)

# ------------------------------------------------------------
# Exercise 18-5 setup carried into 18-6
# ------------------------------------------------------------

x <- 51
z <- 65
i <- 0.06
v <- 1 / (1 + i)

# Decrement ages
y_er <- 61:64   # early retirement
y_w  <- 56:60   # withdrawal
y_i  <- 56:64   # disability
y_d  <- 61:64   # death with spouse annuity

# ------------------------------------------------------------
# Reasonable illustrative assumptions
# ------------------------------------------------------------

# Active-service one-year survival probabilities from age 51 onward
# (reflecting small combined decrement rates)
p_active_1yr <- c(
  "51" = 0.975,
  "52" = 0.974,
  "53" = 0.973,
  "54" = 0.972,
  "55" = 0.971,
  "56" = 0.970,
  "57" = 0.969,
  "58" = 0.968,
  "59" = 0.967,
  "60" = 0.966,
  "61" = 0.965,
  "62" = 0.964,
  "63" = 0.963,
  "64" = 0.962
)

# Build {}_{y-51}p_{51}^{(tau)} for y = 52,...,65
p_tau_to_y <- numeric(14)
names(p_tau_to_y) <- 52:65

cum_surv <- 1
for (age in 51:64) {
  cum_surv <- cum_surv * p_active_1yr[as.character(age)]
  p_tau_to_y[as.character(age + 1)] <- cum_surv
}

# Retirement decrements q_y^(r), ages 61:64
q_r <- c("61" = 0.08, "62" = 0.12, "63" = 0.18, "64" = 0.25)

# Withdrawal decrements q_y^(w), ages 56:60
q_w <- c("56" = 0.04, "57" = 0.035, "58" = 0.03, "59" = 0.025, "60" = 0.02)

# Disability decrements q_y^(i), ages 56:64
q_i <- c(
  "56" = 0.010, "57" = 0.010, "58" = 0.009, "59" = 0.009, "60" = 0.008,
  "61" = 0.008, "62" = 0.007, "63" = 0.007, "64" = 0.006
)

# Death decrements q_y^(d), ages 61:64
q_d <- c("61" = 0.003, "62" = 0.0035, "63" = 0.004, "64" = 0.0045)

# Survival from withdrawal age y+1/2 to 65, using a simple 98.5% annual survival
p_w_to_65 <- sapply(y_w, function(y) 0.985^(65 - y - 0.5))
names(p_w_to_65) <- y_w

# Retirement annuity-due values
adue_r_65 <- 12.0
adue_r_half <- c("61" = 12.8, "62" = 12.6, "63" = 12.4, "64" = 12.2)

# Disability annuity-due values, somewhat higher than retirement annuity
adue_i_half <- c(
  "56" = 14.5, "57" = 14.3, "58" = 14.1, "59" = 13.9, "60" = 13.7,
  "61" = 13.5, "62" = 13.3, "63" = 13.1, "64" = 12.9
)

# Spouse annuity-due values, spouse assumed younger
adue_spouse_half <- c("61" = 14.0, "62" = 13.8, "63" = 13.6, "64" = 13.4)

# ------------------------------------------------------------
# Helper for {}_{y-51}p_{51}^{(tau)}
# ------------------------------------------------------------
p_tau_from_51 <- function(y) {
  if (y == 51) return(1)
  p_tau_to_y[as.character(y)]
}

# ------------------------------------------------------------
# APV component calculator
# benefit_amount = 1860 for normal cost
# benefit_amount = 7500 for accrued liability
# ------------------------------------------------------------
apv_components_18_6 <- function(benefit_amount) {

  # (a) normal retirement
  APV_NR <- benefit_amount *
    v^(65 - 51) *
    p_tau_from_51(65) *
    adue_r_65

  # (b) early retirement
  APV_ER <- sum(sapply(y_er, function(y) {
    reduction <- 1 - 0.03 * (65 - y - 0.5)
    reduction *
      benefit_amount *
      v^(y + 0.5 - 51) *
      p_tau_from_51(y) *
      q_r[as.character(y)] *
      adue_r_half[as.character(y)]
  }))

  # (c) withdrawal
  APV_W <- sum(sapply(y_w, function(y) {
    benefit_amount *
      v^(65 - 51) *
      p_tau_from_51(y) *
      q_w[as.character(y)] *
      p_w_to_65[as.character(y)] *
      adue_r_65
  }))

  # (d) disability
  APV_I <- sum(sapply(y_i, function(y) {
    benefit_amount *
      v^(y + 0.5 - 51) *
      p_tau_from_51(y) *
      q_i[as.character(y)] *
      adue_i_half[as.character(y)]
  }))

  # (e) death
  APV_D <- sum(sapply(y_d, function(y) {
    0.50 *
      (1 - 0.03 * (65 - y - 0.5)) *
      benefit_amount *
      v^(y + 0.5 - 51) *
      p_tau_from_51(y) *
      q_d[as.character(y)] *
      adue_spouse_half[as.character(y)]
  }))

  c(
    APV_NR = APV_NR,
    APV_ER = APV_ER,
    APV_W  = APV_W,
    APV_I  = APV_I,
    APV_D  = APV_D,
    TOTAL  = APV_NR + APV_ER + APV_W + APV_I + APV_D
  )
}

# ------------------------------------------------------------
# Part (b): TUC normal cost using benefit accrual BA = 1860
# ------------------------------------------------------------
NC_components <- apv_components_18_6(benefit_amount = BA)

# ------------------------------------------------------------
# Part (c): TUC accrued liability using accrued benefit AB_56 = 7500
# ------------------------------------------------------------
AAL_components <- apv_components_18_6(benefit_amount = AB_56)

cat("\nPart (b): Illustrative TUC normal cost components\n")
print(round(NC_components, 2))

cat("\nPart (c): Illustrative TUC accrued liability components\n")
print(round(AAL_components, 2))

# ------------------------------------------------------------
# Plot component comparison
# ------------------------------------------------------------
comp_mat <- rbind(
  Normal_Cost = NC_components[1:5],
  Accrued_Liability = AAL_components[1:5]
)

barplot(
  t(comp_mat),
  beside = TRUE,
  legend.text = TRUE,
  main = "Exercise 18-6: Illustrative component APVs",
  ylab = "Actuarial present value"
)

# R Check for Example 11.1
# Verify: beta = P + (beta - alpha)/adotxn

library(mqriskR)

x <- 40
k <- 20
i <- 0.05

P <- Pxn(x = x, n = k, i = i, model = "uniform", omega = 100)
adot_temp <- adotxn(x = x, n = k, i = i, model = "uniform", omega = 100)

gap <- 0.02
beta_from_formula <- P + gap / adot_temp
alpha_from_gap <- beta_from_formula - gap

print(round(c(
  P = P,
  adot_xn = adot_temp,
  gap = gap,
  beta = beta_from_formula,
  alpha = alpha_from_gap
), 6))

# Sensitivity plot
gap_grid <- seq(0, 0.10, by = 0.002)
beta_grid <- P + gap_grid / adot_temp

plot(
  gap_grid, beta_grid, type = "l", lwd = 2,
  xlab = expression(beta - alpha),
  ylab = expression(beta),
  main = "Example 11.1: Renewal Modified Premium vs (beta - alpha)"
)
abline(h = P, lty = 2)




# R Check for Example 11.4
# Fully continuous whole life with increasing benefit b_t = 1000 e^(0.04 t)

lambda <- 0.02
delta <- 0.04
t <- 2

# Net premium rate
abar <- 1 / (lambda + delta)
Pbar <- 1000 / abar

# APV of future benefit at duration t
apvb_t <- integrate(
  function(s) 1000 * lambda * exp(0.04 * t) * exp(-lambda * s),
  lower = 0, upper = Inf
)$value

apvb_t

# APV of future premiums at duration t
apvp_t <- Pbar * abar

Vt <- apvb_t - apvp_t

print(round(c(
  abar = abar,
  Pbar = Pbar,
  apvb_t = apvb_t,
  apvp_t = apvp_t,
  reserve_t = Vt
), 6))

# Reserve as a function of duration
t_grid <- seq(0, 10, by = 0.25)
reserve_grid <- 1000 * exp(0.04 * t_grid) - Pbar * abar

plot(
  t_grid, reserve_grid, type = "l", lwd = 2,
  xlab = "Duration t",
  ylab = expression(t * bar(V)),
  main = "Example 11.4: Continuous Reserve by Duration"
)
abline(v = t, lty = 2)
abline(h = Vt, lty = 2)


# R Check for Example 11.7
# Gross gain on a single policy

V10G <- 3950.73
V11G <- 4607.07
G <- 685.00

i_actual <- 0.065
q_actual <- 0.005
r_actual <- 0.06
s_actual <- 100
b <- 50000

GT <- (V10G + G * (1 - r_actual)) * (1 + i_actual) -
  ((b + s_actual) * q_actual + (1 - q_actual) * V11G)

print(round(c(
  total_gain = GT
), 6))

# Sensitivity to actual interest
i_grid <- seq(0.04, 0.08, by = 0.001)
GT_i <- sapply(i_grid, function(ii) {
  (V10G + G * (1 - r_actual)) * (1 + ii) -
    ((b + s_actual) * q_actual + (1 - q_actual) * V11G)
})

plot(
  i_grid, GT_i, type = "l", lwd = 2,
  xlab = "Actual interest rate",
  ylab = "Total gain",
  main = "Example 11.7: Total Gain vs Actual Interest Rate"
)
abline(v = i_actual, lty = 2)

# Sensitivity to actual mortality
q_grid <- seq(0.003, 0.008, by = 0.0001)
GT_q <- sapply(q_grid, function(qq) {
  (V10G + G * (1 - r_actual)) * (1 + i_actual) -
    ((b + s_actual) * qq + (1 - qq) * V11G)
})

plot(
  q_grid, GT_q, type = "l", lwd = 2,
  xlab = "Actual mortality rate",
  ylab = "Total gain",
  main = "Example 11.7: Total Gain vs Actual Mortality Rate"
)
abline(v = q_actual, lty = 2)



# R Check for Example 11.9
# Backward Euler approximation for gross reserve

dVdt <- function(t, V) {
  19.60 + 0.04 * V - 0.01 * t^2 * (110 - V)
}

# Backward recursion using derivative at left endpoint
backward_left <- function(times, V_terminal) {
  out <- numeric(length(times))
  out[length(times)] <- V_terminal

  for (k in seq(length(times) - 1, 1, by = -1)) {
    h <- times[k + 1] - times[k]
    t_left <- times[k]
    # solve: V_k + h*dVdt(t_left, V_k) = V_{k+1}
    a <- 1 + h * (0.04 + 0.01 * t_left^2)
    rhs <- out[k + 1] - h * 19.60 + h * 110 * 0.01 * t_left^2
    out[k] <- rhs / a
  }
  out
}

# Backward recursion using derivative at right endpoint
backward_right <- function(times, V_terminal) {
  out <- numeric(length(times))
  out[length(times)] <- V_terminal

  for (k in seq(length(times) - 1, 1, by = -1)) {
    h <- times[k + 1] - times[k]
    t_right <- times[k + 1]
    out[k] <- out[k + 1] - h * dVdt(t_right, out[k + 1])
  }
  out
}

times1 <- seq(4, 5, by = 0.5)
left_path <- backward_left(times1, 100)
right_path <- backward_right(times1, 100)

print(round(data.frame(
  time = times1,
  backward_left = left_path,
  backward_right = right_path
), 6))

# Smaller step size
times2 <- seq(4, 5, by = 0.25)
left_path_small <- backward_left(times2, 100)

plot(
  times1, left_path, type = "b", lwd = 2, pch = 16,
  ylim = range(c(left_path, right_path, left_path_small)),
  xlab = "Time",
  ylab = expression(t * bar(V)^G),
  main = "Example 11.9: Backward Euler Reserve Approximations"
)
lines(times1, right_path, type = "b", lwd = 2, pch = 1, lty = 2)
lines(times2, left_path_small, type = "b", lwd = 2, pch = 3, lty = 3)
legend(
  "topleft",
  legend = c("h = 0.50, left derivative", "h = 0.50, right derivative", "h = 0.25, left derivative"),
  lty = c(1, 2, 3),
  pch = c(16, 1, 3),
  bty = "n"
)




# R Check for Exercise 11-1
# Verify:
# (a) V_NLP - V_M = (beta - P) * adot_{x+t:n-t}
# (b) V_M = 1 - (beta + d) * adot_{x+t:n-t}

library(mqriskR)

x <- 40
n <- 20
t <- 10
i <- 0.05
beta <- 0.08

P <- Pxn(x = x, n = n, i = i, model = "uniform", omega = 100)
d <- i / (1 + i)
adot_future <- adotxn(x = x + t, n = n - t, i = i, model = "uniform", omega = 100)
A_future <- Axn(x = x + t, n = n - t, i = i, model = "uniform", omega = 100)

V_nlp <- tVxn(x = x, n = n, t = t, i = i, model = "uniform", omega = 100)
V_mod <- A_future - beta * adot_future

lhs_a <- V_nlp - V_mod
rhs_a <- (beta - P) * adot_future

lhs_b <- V_mod
rhs_b <- 1 - (beta + d) * adot_future

round(c(
  P = P,
  V_nlp = V_nlp,
  V_mod = V_mod,
  lhs_a = lhs_a,
  rhs_a = rhs_a,
  lhs_b = lhs_b,
  rhs_b = rhs_b
), 6)

# Sensitivity plot
beta_grid <- seq(P - 0.01, P + 0.06, by = 0.001)
V_mod_grid <- A_future - beta_grid * adot_future

plot(
  beta_grid, V_mod_grid, type = "l", lwd = 2,
  xlab = expression(beta),
  ylab = "Modified reserve",
  main = "Exercise 11-1: Modified Reserve vs Renewal Premium"
)
abline(v = P, lty = 2)
abline(h = V_nlp, lty = 3)




# R Check for Exercise 11-5
# Verify r = (s|1-s q_x+t) / q_x+t = 1 - s under UDD

s <- seq(0, 1, by = 0.01)

# Under UDD:
r_formula <- 1 - s

# Numerical illustration using a sample q value
q <- 0.12
deferred_piece <- (1 - s) * q
r_numeric <- deferred_piece / q

round(head(data.frame(
  s = s,
  r_numeric = r_numeric,
  r_formula = r_formula
), 10), 6)

plot(
  s, r_numeric, type = "l", lwd = 2,
  xlab = "s",
  ylab = "r",
  main = "Exercise 11-5: Interpolation Weight under UDD"
)
lines(s, r_formula, lty = 2, lwd = 2)
legend(
  "topright",
  legend = c("Numerical ratio", "1 - s"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)




# R Check for Exercise 11-6
# Verify:
# V_{t-1/2} = 1/2 * [ (1 + v p_{x+t-1}) V_t + v q_{x+t-1} ]

library(mqriskR)

x <- 40
t <- 10
i <- 0.05
v <- 1 / (1 + i)

Vt <- tVx(x = x, t = t, i = i, model = "uniform", omega = 100)
px <- tpx(x = x + t - 1, t = 1, model = "uniform", omega = 100)
qx <- 1 - px

# meanVx(x, t, ...) returns {}_{t+1/2}V, so use t - 1 to match {}_{t-1/2}V
lhs <- meanVx(x = x, t = t - 1, i = i, model = "uniform", omega = 100)
rhs <- 0.5 * ((1 + v * px) * Vt + v * qx)

round(c(
  Vt = Vt,
  px = px,
  qx = qx,
  lhs = lhs,
  rhs = rhs
), 6)

# Check across a range of durations
t_grid <- 1:20
lhs_grid <- sapply(t_grid, function(tt) {
  meanVx(x = x, t = tt - 1, i = i, model = "uniform", omega = 100)
})
rhs_grid <- sapply(t_grid, function(tt) {
  Vtt <- tVx(x = x, t = tt, i = i, model = "uniform", omega = 100)
  ptt <- tpx(x = x + tt - 1, t = 1, model = "uniform", omega = 100)
  qtt <- 1 - ptt
  0.5 * ((1 + v * ptt) * Vtt + v * qtt)
})

plot(
  t_grid, lhs_grid, type = "l", lwd = 2,
  xlab = "t",
  ylab = "Mean reserve",
  main = "Exercise 11-6: Mean Reserve Identity"
)
lines(t_grid, rhs_grid, lty = 2, lwd = 2)
legend(
  "topleft",
  legend = c("Package value for {}_{t-1/2}V", "Formula value"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)




# R Check for Exercise 11-8
# 3-year term insurance with decreasing benefits

i <- 0.06
v <- 1 / (1 + i)

b1 <- 200
b2 <- 150
b3 <- 100

qx <- 0.03
qx1 <- 0.06
qx2 <- 0.09

px <- 1 - qx
px1 <- 1 - qx1

P <- (
  b1 * v * qx +
    b2 * v^2 * px * qx1 +
    b3 * v^3 * px * px1 * qx2
) / (
  1 + v * px + v^2 * px * px1
)

V1 <- ((P) * (1 + i) - b1 * qx) / px
initial_second_year <- V1 + P

round(c(
  premium = P,
  terminal_reserve_year1 = V1,
  initial_reserve_year2 = initial_second_year
), 6)

# Sensitivity plot for b2
b2_grid <- seq(100, 250, by = 2)
init2_grid <- sapply(b2_grid, function(bb2) {
  Ptmp <- (
    b1 * v * qx +
      bb2 * v^2 * px * qx1 +
      b3 * v^3 * px * px1 * qx2
  ) / (
    1 + v * px + v^2 * px * px1
  )
  V1tmp <- (Ptmp * (1 + i) - b1 * qx) / px
  V1tmp + Ptmp
})

plot(
  b2_grid, init2_grid, type = "l", lwd = 2,
  xlab = expression(b[2]),
  ylab = "Second-year initial reserve",
  main = "Exercise 11-8: Initial Reserve vs Middle-Year Benefit"
)
abline(v = b2, lty = 2)




# R Check for Exercise 11-9
# 2-year endowment with death benefit = 1000 + reserve at end of year of death

i <- 0.10
q0 <- 0.10
q1 <- 0.11

# From the solution:
# V1 = P(1+i) - 100
# V2 = (V1 + P)(1+i) - 110
# Set V2 = 1000 and solve for P

P <- (1000 + 220) / (((1 + i) + 1) * (1 + i))

V1 <- P * (1 + i) - 100
V2 <- (V1 + P) * (1 + i) - 110

round(c(
  premium = P,
  reserve_1 = V1,
  reserve_2 = V2
), 6)

# Sensitivity to first-year mortality
q0_grid <- seq(0.02, 0.20, by = 0.002)
P_grid <- sapply(q0_grid, function(qq0) {
  # Generalized version:
  # V1 = P(1+i) - 1000 q0
  # V2 = (V1 + P)(1+i) - 1000 q1
  # Set V2 = 1000
  (1000 + 1000 * (qq0 * (1 + i) + q1)) / (((1 + i) + 1) * (1 + i))
})

plot(
  q0_grid, P_grid, type = "l", lwd = 2,
  xlab = expression(q[x]),
  ylab = "Net level premium",
  main = "Exercise 11-9: Premium vs First-Year Mortality"
)
abline(v = q0, lty = 2)




# R Check for Exercise 11-10
# Solve for b_11 from the recursion:
# (V_10 + P)(1+i) = b_11 q_30' + p_30' V_11
#
# To avoid inconsistency, compute P, V10, and V11 from the same model basis.

library(mqriskR)

i <- 0.05

# Question data
q30 <- 0.008427
q30_prime <- q30 + 0.01
p30_prime <- 1 - q30_prime

# Since premiums and reserves are stated to match those of issue age 20,
# compute all model-based quantities from the same age-20 basis.
P <- Px(x = 20, i = i, model = "uniform", omega = 100)
V10 <- tVx(x = 20, t = 10, i = i, model = "uniform", omega = 100)
V11 <- tVx(x = 20, t = 11, i = i, model = "uniform", omega = 100)

# Implied benefit in year 11
b11 <- ((V10 + P) * (1 + i) - p30_prime * V11) / q30_prime

round(c(
  P = P,
  V10 = V10,
  V11 = V11,
  q30 = q30,
  q30_prime = q30_prime,
  b11 = b11
), 6)

# Check the recursion numerically
lhs <- (V10 + P) * (1 + i)
rhs <- b11 * q30_prime + p30_prime * V11

round(c(
  lhs = lhs,
  rhs = rhs,
  difference = lhs - rhs
), 10)

# Sensitivity to excess mortality adjustment
excess_grid <- seq(0, 0.03, by = 0.0005)
b_grid <- sapply(excess_grid, function(extra_q) {
  q_prime <- q30 + extra_q
  p_prime <- 1 - q_prime
  ((V10 + P) * (1 + i) - p_prime * V11) / q_prime
})

plot(
  excess_grid, b_grid, type = "l", lwd = 2,
  xlab = "Excess mortality adjustment",
  ylab = expression(b[11]),
  main = "Exercise 11-10: Implied Benefit vs Excess Mortality"
)
abline(v = 0.01, lty = 2)
abline(h = b11, lty = 2)




# R Check for Exercise 11-19
# Ordered gain decomposition for 990 policies:
# (a) interest, (b) mortality, (c) expense

npol <- 990
G <- 90
V3G <- 97.12
V4G <- 164.13
benefit <- 10000

# Assumed
i_assumed <- 0.05
q_assumed <- 0.003
r_assumed <- 0.03

# Actual
i_actual <- 0.04
q_actual <- 0.002
r_actual <- 0.025

# Per-policy total gain under actual experience
GT_per <- (V3G + G * (1 - r_actual)) * (1 + i_actual) -
  (benefit * q_actual + (1 - q_actual) * V4G)

# Interest gain first
GI_per <- (V3G + G * (1 - r_assumed)) * (1 + i_actual) -
  (benefit * q_assumed + (1 - q_assumed) * V4G)

# Mortality gain second
GIM_per <- (V3G + G * (1 - r_assumed)) * (1 + i_actual) -
  (benefit * q_actual + (1 - q_actual) * V4G)
GM_per <- GIM_per - GI_per

# Expense gain last
GE_per <- GT_per - GI_per - GM_per

out <- c(
  total_gain = npol * GT_per,
  interest_gain = npol * GI_per,
  mortality_gain = npol * GM_per,
  expense_gain = npol * GE_per,
  check = npol * (GI_per + GM_per + GE_per)
)

round(out, 6)


## Use package function
out_per_policy <- decompGg_disc(
  VtG = 97.12,
  Vt1G = 164.13,
  G = 90,
  i_assumed = 0.05,
  q_assumed = 0.003,
  r_assumed = 0.03,
  e_assumed = 0,
  s_assumed = 0,
  i_actual = 0.04,
  q_actual = 0.002,
  r_actual = 0.025,
  e_actual = 0,
  s_actual = 0,
  b = 10000,
  order = c("interest", "mortality", "expense")
)

out_block <- npol * out_per_policy

round(out_block, 6)

barplot(
  height = c(out["interest_gain"], out["mortality_gain"], out["expense_gain"]),
  names.arg = c("Interest", "Mortality", "Expense"),
  main = "Exercise 11-19: Aggregate Gain by Source",
  ylab = "Gain for 990 policies"
)
abline(h = 0)





# R Check for Exercise 11-20
# Payout-phase deferred annuity with expense on survival benefit

library(mqriskR)

x <- 70
i_assumed <- 0.06
i_actual <- 0.055
q_assumed <- 0.02041
q_actual <- 0.025
p_assumed <- 1 - q_assumed
p_actual <- 1 - q_actual

benefit <- 10000
expense_rate <- 0.05
net_survival_outgo <- benefit * (1 + expense_rate)   # annuity + expense if alive

# Reserve at age 70 and 71 from annual life annuity-immediate values
# payout phase: reserve is APV of future survival-contingent payments
V70 <- net_survival_outgo * ax(x = x, i = i_assumed, model = "uniform", omega = 100)
V71 <- net_survival_outgo * ax(x = x + 1, i = i_assumed, model = "uniform", omega = 100)

# Ordered gains: mortality first, then interest
GM <- (V70 - net_survival_outgo) * (1 + i_assumed) - p_actual * V71
GI_total <- (V70 - net_survival_outgo) * (1 + i_actual) - p_actual * V71
GI <- GI_total - GM

round(c(
  V70 = V70,
  V71 = V71,
  mortality_gain = GM,
  interest_gain = GI,
  total_gain = GM + GI
), 6)

# Sensitivity to actual mortality
q_grid <- seq(0.01, 0.04, by = 0.0005)
GM_grid <- sapply(q_grid, function(qa) {
  pa <- 1 - qa
  (V70 - net_survival_outgo) * (1 + i_assumed) - pa * V71
})

plot(
  q_grid, GM_grid, type = "l", lwd = 2,
  xlab = "Actual mortality rate",
  ylab = "Mortality gain",
  main = "Exercise 11-20: Mortality Gain in Payout Phase"
)
abline(v = q_actual, lty = 2)

## Listing 7.1: Basic discrete insurance functions from a parametric survival model
library(mqriskR)

x <- 40
n <- 20
i <- 0.05

# Use a simple uniform survival model with limiting age omega = 100
Ax_val    <- Ax(x = x, i = i, model = "uniform", omega = 100)
Axn1_val  <- Axn1(x = x, n = n, i = i, model = "uniform", omega = 100)
nAx_val   <- nAx(x = x, n = n, i = i, model = "uniform", omega = 100)
nEx_val   <- nEx(x = x, n = n, i = i, model = "uniform", omega = 100)
Axn_val   <- Axn(x = x, n = n, i = i, model = "uniform", omega = 100)

out <- c(
  Ax   = Ax_val,
  Axn1 = Axn1_val,
  nAx  = nAx_val,
  nEx  = nEx_val,
  Axn  = Axn_val
)

print(round(out, 6))

# Check the endowment decomposition
round(Axn_val - (Axn1_val + nEx_val), 10)




## Listing 7.2: Comparison of annual, continuous, and \(m\)-thly insurance APVs
library(mqriskR)

x <- 40
i <- 0.05

A_annual <- Ax(x = x, i = i, model = "uniform", omega = 100)
A_cont   <- Abarx(x = x, i = i, model = "uniform", omega = 100)

m_grid <- c(2, 4, 12, 24)
A_m <- sapply(m_grid, function(mm) {
  Ax_m(x = x, i = i, m = mm, model = "uniform", omega = 100)
})

vals <- c(annual = A_annual, A_m, continuous = A_cont)
print(round(vals, 6))

##Listing 16.1: Type B universal life account value path
library(mqriskR)

B <- 100000; G <- rep(5000, 5)
r <- c(0.75, rep(0.10, 4)); e <- c(100, rep(20, 4))
qx <- c(0.00076, 0.00081, 0.00085, 0.00090, 0.00095)
iq <- 0.03; ic <- 0.03

AV <- numeric(5); COI <- numeric(5)

for (t in 1:5) {prior <- if (t == 1) 0 else AV[t - 1]
COI[t] <- B * qx[t] / (1 + iq)
AV[t] <- (prior + G[t] * (1 - r[t]) - e[t] - COI[t]) * (1 + ic)}

direct_out <- data.frame(year = 1:5,COI = round(COI, 2),AV  = round(AV, 2))

pkg_out <- AV_path_ul_typeB(G = G, r = r, e = e, qx = qx,ic = ic, B = B, iq = iq)

direct_out
pkg_out[, c("t", "COI", "AV")]



##Listing 16.2: Type A universal life account value path
library(mqriskR)

B <- 100000; G <- rep(5000, 5)
r <- c(0.75, rep(0.10, 4)); e <- c(100, rep(20, 4))
qx <- c(0.00076, 0.00081, 0.00085, 0.00090, 0.00095)
i <- 0.03

AV <- numeric(5)

for (t in 1:5) {prior <- if (t == 1) 0 else AV[t - 1]
AV[t] <- ((prior + G[t] * (1 - r[t]) - e[t]) * (1 + i) - B * qx[t]) / (1 - qx[t])}

direct_out <- data.frame(year = 1:5,AV = round(AV, 2))

pkg_out <- AV_path_ul_typeA(G = G, r = r, e = e, qx = qx,ic = i, B = B)

direct_out
pkg_out[, c("t", "AV")]




##Listing 16.3: Point-to-point credited rates under EIUL
library(mqriskR)

index <- c(1000, 1050, 1200, 1100, 950, 1060, 1150)
cap <- 0.10
floor <- 0.01
part <- 1.10

raw <- index[-1] / index[-length(index)] - 1
after_participation <- part * raw
credited <- pmin(cap, pmax(floor, after_participation))

data.frame(year = 1:6,
  raw_growth = round(raw, 4),
  after_participation = round(after_participation, 4),
  credited_rate = round(credited, 4))



##Listing 16.4: Monthly-average credited rate under EIUL
library(mqriskR)

index <- c(1000, 1020, 1100, 1150, 1080, 1040, 960,
           1030, 1000, 1070, 1150, 1200, 1150)

raw <- iMA_eiul(index)

part <- 1.10
cap <- 0.10
floor <- 0.01

credited <- pmin(cap, pmax(floor, part * raw))

c(raw_growth = raw, credited_rate = credited)




##Listing 16.5: Persistency and cumulative in-force probabilities
library(mqriskR)

q_d <- c(.001, .002, .003, .004, .005)
q_w <- c(.02, .02, .03, .04, .05)

# Direct computation from formulas
p_tau_direct <- (1 - q_d) * (1 - q_w)
tp_tau_direct <- cumprod(p_tau_direct)

# Package functions
p_tau_pkg <- pxtau_ul(qd = q_d, qw = q_w)
tp_tau_pkg <- tpxtau_ul(qd = q_d, qw = q_w)

data.frame(year = 1:5,
           p_tau_direct = round(p_tau_direct, 5),
           p_tau_pkg    = round(p_tau_pkg, 5),
           tp_tau_direct = round(tp_tau_direct, 5),
           tp_tau_pkg    = round(tp_tau_pkg, 5))

# R Check for Example 9.1
# Special 10-year contract with refund of net annual premiums on survival

A_endow <- 0.60
A_pure  <- 0.47
d <- 0.05

A_term <- A_endow - A_pure
adot_30_10 <- (1 - A_endow) / d
P <- A_term / (adot_30_10 - 10 * A_pure)

round(c(
  A_term = A_term,
  adot_30_10 = adot_30_10,
  premium = P
), 5)



# R Check for Example 9.4
# Variance of present-value-of-loss random variable for a 2-year term insurance

qx <- 0.10
qx1 <- 0.20
px <- 1 - qx
v <- 0.90

A_x_2_term <- v * qx + v^2 * px * qx1
adot_x_2 <- 1 + v * px
P <- A_x_2_term / adot_x_2

L1 <- v - P
L2 <- v^2 - P * (1 + v)
L3 <- -P * (1 + v)

pr1 <- qx
pr2 <- px * qx1
pr3 <- px * (1 - qx1)

EL <- pr1 * L1 + pr2 * L2 + pr3 * L3
VarL <- pr1 * L1^2 + pr2 * L2^2 + pr3 * L3^2 - EL^2

round(c(
  premium = P,
  loss_year1 = L1,
  loss_year2 = L2,
  loss_no_claim = L3,
  expected_loss = EL,
  variance = VarL
), 5)



# R Check for Example 9.6
# Expected value of loss under elevated first-year mortality

i <- 0.06
d <- i / (1 + i)

A60 <- 0.36933
A61 <- 0.38300
q60 <- 0.01376

adot60 <- (1 - A60) / d
P60 <- 1000 * A60 / adot60

q60_imp <- 10 * q60
p60_imp <- 1 - q60_imp

A60_imp <- (q60_imp / (1 + i)) + (p60_imp / (1 + i)) * A61
adot60_imp <- (1 - A60_imp) / d

EL_imp <- 1000 * A60_imp - P60 * adot60_imp

round(c(
  adot60 = adot60,
  premium = P60,
  A60_imp = A60_imp,
  adot60_imp = adot60_imp,
  expected_loss = EL_imp
), 5)





# R Check for Example 9.11
# Standard and impaired-life temporary annuity and term insurance values

i <- 0.06
v <- 1 / (1 + i)
d <- i / (1 + i)

tp <- c(1.00000, 0.99105, 0.98143, 0.97107, 0.95994, 0.94800)

# Part (a)
adot_temp5 <- sum(v^(0:4) * tp[1:5])
nEx5 <- v^5 * tp[6]
A_term5 <- 1 - d * adot_temp5 - nEx5
P_term5 <- A_term5 / adot_temp5

round(c(
  adot_temp5 = adot_temp5,
  A_term5 = A_term5,
  P_term5 = P_term5
), 5)

# Part (b): impaired life with mu' = mu + 0.002
delta <- log(1 + i)
delta_prime <- delta + 0.002
i_prime <- exp(delta_prime) - 1

adot_temp5_prime <- sum((1 / (1 + i_prime))^(0:4) * tp[1:5])
nEx5_prime <- (1 / (1 + i_prime))^5 * tp[6]
A_term5_prime <- 1 - d * adot_temp5_prime - nEx5_prime
P_term5_prime <- A_term5_prime / adot_temp5_prime

round(c(
  i_prime = i_prime,
  adot_temp5_prime = adot_temp5_prime,
  A_term5_prime = A_term5_prime,
  P_term5_prime = P_term5_prime
), 5)


# Exercise 9.4 R Check: whole life premium with premiums increasing at interest

i <- 0.05
v <- 1 / (1 + i)
omega <- 105
x <- 75

# 1 + e_75 under uniform future lifetime over 30 years
t_vals <- 0:(omega - x - 1)
surv_vals <- (omega - x - t_vals) / (omega - x)
one_plus_ex <- sum(surv_vals)

# Whole life insurance APV under the same model
k_vals <- 0:(omega - x - 1)
kq_vals <- rep(1 / (omega - x), length(k_vals))
Ax_75 <- sum(v^(k_vals + 1) * kq_vals)

P1 <- 1000 * Ax_75 / one_plus_ex

round(c(
  one_plus_ex = one_plus_ex,
  Ax_75 = Ax_75,
  P1 = P1
), 5)


# Exercise 9.5 R Check: mortgage-protection decreasing term premium

loan <- 100000
i <- 0.05
n <- 25
adot_40_25 <- 14
p25 <- 0.80

# Annual mortgage payment
mortgage_payment <- loan / annuity_certain(n = n, i = i, due = FALSE)

# Approximate insurance APV using the given closed-form solution
v <- 1 / (1 + i)
q25 <- 1 - p25

apv_benefit <- 149000 * (
  1 - (i / (1 + i)) * adot_40_25 - p25 * v^25 - q25 * v^26
)

premium <- apv_benefit / adot_40_25

round(c(
  mortgage_payment = mortgage_payment,
  apv_benefit = apv_benefit,
  premium = premium
), 5)



# Exercise 9.10 R Check: expected present value of loss from given variance and moments

Ax <- 0.29224
A2x <- 0.11723
VarL <- 0.10
i <- 0.05
d <- i / (1 + i)

VarZ <- A2x - Ax^2
P_over_d <- sqrt(VarL / VarZ) - 1

EL <- (1 + P_over_d) * Ax - P_over_d

round(c(
  VarZ = VarZ,
  P_over_d = P_over_d,
  expected_loss = EL
), 5)



# Exercise 9.13 R Check: aggregate-loss premium target and portfolio sensitivity

Ax <- 0.24905
A2x <- 0.09476
i <- 0.06
d <- i / (1 + i)
z95 <- qnorm(0.95)

prob_pos_loss <- function(P, n, Ax, A2x, d) {
  mean_i <- (1 + P / d) * Ax - P / d
  var_i <- (1 + P / d)^2 * (A2x - Ax^2)
  mean_agg <- n * mean_i
  sd_agg <- sqrt(n * var_i)
  1 - pnorm((0 - mean_agg) / sd_agg)
}

# Premium for n = 100 with target probability about 0.05
objective <- function(P) prob_pos_loss(P, n = 100, Ax = Ax, A2x = A2x, d = d) - 0.05
P_star_100 <- uniroot(objective, interval = c(0.015, 0.03))$root

# Limiting premium (equivalence principle)
P_limit <- d * Ax / (1 - Ax)

# Probability for P = 0.024 and n = 100
prob_at_024 <- prob_pos_loss(0.024, n = 100, Ax = Ax, A2x = A2x, d = d)

round(c(
  P_star_100 = P_star_100,
  P_limit = P_limit,
  prob_at_024 = prob_at_024
), 5)

# Plot probability of positive aggregate loss against n for a few premium values
n_vals <- 10:300
p1 <- sapply(n_vals, prob_pos_loss, P = 0.02188, Ax = Ax, A2x = A2x, d = d)
p2 <- sapply(n_vals, prob_pos_loss, P = 0.02400, Ax = Ax, A2x = A2x, d = d)
p3 <- sapply(n_vals, prob_pos_loss, P = 0.02500, Ax = Ax, A2x = A2x, d = d)

plot(n_vals, p1, type = "l", lwd = 2,
     xlab = "Number of contracts",
     ylab = "Pr(Aggregate loss > 0)",
     main = "Aggregate loss probability versus portfolio size")
lines(n_vals, p2, lwd = 2, lty = 2)
lines(n_vals, p3, lwd = 2, lty = 3)
abline(h = 0.05, lty = 4)
legend("topright",
       legend = c("P = 0.02188", "P = 0.02400", "P = 0.02500", "Target 0.05"),
       lty = c(1, 2, 3, 4), lwd = c(2, 2, 2, 1), bty = "n")


# Exercise 9.25 R Check: monthly versus semiannual true fractional premiums under UDD

x# R Check: monthly versus semiannual true fractional premiums under UDD

A_term <- 0.015
adot_temp <- 8
nEx <- 0.604
i <- 0.05

adot_temp_m <- function(m) {
  ic <- interest_convert(i = i, m = m)
  im <- ic$im                      # nominal annual rate convertible m-thly
  d <- i / (1 + i)
  dm_small <- (im / m) / (1 + im / m)
  dm <- m * dm_small              # this is d^(m)

  alpha <- (i * d) / (im * dm)
  beta  <- (i - im) / (im * dm)

  alpha * adot_temp - beta * (1 - nEx)
}

adot2 <- adot_temp_m(2)
adot12 <- adot_temp_m(12)

prem2 <- 10000 * A_term / adot2
prem12 <- 10000 * A_term / adot12

round(c(
  adot2 = adot2,
  adot12 = adot12,
  prem2 = prem2,
  prem12 = prem12,
  difference = prem12 - prem2
), 5)



# Exercise 9.27 R Check: gross premium with expenses

adot_x_20 <- 10
adot_x <- 20
d <- 0.04

Ax <- 1 - d * adot_x

G <- (Ax + 0.05 + 0.02 * adot_x) / ((1 - 0.03) * adot_x_20)
P_net_20pay <- Ax / adot_x_20

round(c(
  Ax = Ax,
  net_premium = P_net_20pay,
  gross_premium = G
), 5)

# R Check for Example 12.2
# Verify {}_{2|}q_{xy} for independent lives

qx  <- 0.08
qy  <- 0.10
qx1 <- 0.09
qy1 <- 0.15
qx2 <- 0.10
qy2 <- 0.20

px  <- 1 - qx
py  <- 1 - qy
px1 <- 1 - qx1
py1 <- 1 - qy1
px2 <- 1 - qx2
py2 <- 1 - qy2

p2x <- px * px1
p2y <- py * py1

q_xy_age2 <- 1 - px2 * py2
value <- p2x * p2y * q_xy_age2

round(c(
  p2x = p2x,
  p2y = p2y,
  q_xy_age2 = q_xy_age2,
  `2|q_xy` = value
), 6)

# Small sensitivity plot: vary q_{x+2}
qx2_grid <- seq(0.02, 0.30, by = 0.01)
value_grid <- p2x * p2y * (1 - (1 - qx2_grid) * py2)

plot(
  qx2_grid, value_grid, type = "l", lwd = 2,
  xlab = expression(q[x+2]),
  ylab = expression({2*"|"}*q[xy]),
  main = "Example 12.2: Third-Year Joint-Life Failure Probability"
)
abline(v = qx2, lty = 2)
abline(h = value, lty = 2)


# R Check for Example 12.3
# Verify e̊_xy = 1 / (lambda_x + lambda_y)

lambda_x <- 0.03
lambda_y <- 0.05

closed_form <- 1 / (lambda_x + lambda_y)

numeric_val <- integrate(
  function(t) exp(-(lambda_x + lambda_y) * t),
  lower = 0, upper = Inf
)$value

round(c(
  closed_form = closed_form,
  numeric = numeric_val
), 6)

# Sensitivity plot
lambda_sum <- seq(0.02, 0.20, by = 0.002)
e_grid <- 1 / lambda_sum

plot(
  lambda_sum, e_grid, type = "l", lwd = 2,
  xlab = expression(lambda[x] + lambda[y]),
  ylab = expression(ring(e)[xy]),
  main = "Example 12.3: Joint-Life Complete Expectation"
)
abline(v = lambda_x + lambda_y, lty = 2)
abline(h = closed_form, lty = 2)




# R Check for Example 12.5
# Verify e̊_{overline{50:60}} under uniform model with omega = 100

library(mqriskR)

x <- 50
y <- 60
omega <- 100

ex <- ex_complete(x = x, model = "uniform", omega = omega)
ey <- ex_complete(x = y, model = "uniform", omega = omega)

exy <- integrate(
  function(t) {
    p1 <- ifelse(t <= omega - x, (omega - x - t) / (omega - x), 0)
    p2 <- ifelse(t <= omega - y, (omega - y - t) / (omega - y), 0)
    p1 * p2
  },
  lower = 0, upper = omega - y
)$value

e_last <- ex + ey - exy

round(c(
  ex = ex,
  ey = ey,
  exy = exy,
  e_last = e_last
), 6)

# Plot joint-life and last-survivor survival functions
t_grid <- seq(0, 40, by = 0.25)
p_joint <- ifelse(t_grid <= 40,
                  ((50 - t_grid) / 50) * ((40 - t_grid) / 40), 0)
p_last <- ifelse(t_grid <= 40,
                 (50 - t_grid) / 50 + (40 - t_grid) / 40 - p_joint,
                 ifelse(t_grid <= 50, (50 - t_grid) / 50, 0))

plot(
  t_grid, p_joint, type = "l", lwd = 2,
  xlab = "t",
  ylab = "Survival probability",
  main = "Example 12.5: Joint-Life and Last-Survivor Survival"
)
lines(t_grid, p_last, lwd = 2, lty = 2)
legend("topright", legend = c("Joint-life", "Last-survivor"), lty = c(1, 2), lwd = 2, bty = "n")


# R Check for Example 12.13
# Joint density f(x,y) = (x+y)/27 on (0,3)x(0,3)

fxy <- function(tx, ty) {
  ifelse(tx > 0 & tx < 3 & ty > 0 & ty < 3, (tx + ty) / 27, 0)
}

fx <- function(tx) {
  sapply(tx, function(z) integrate(function(ty) fxy(z, ty), 0, 3)$value)
}

fy <- function(ty) {
  sapply(ty, function(z) integrate(function(tx) fxy(tx, z), 0, 3)$value)
}

EX <- integrate(function(tx) tx * fx(tx), 0, 3)$value
EY <- integrate(function(ty) ty * fy(ty), 0, 3)$value

EX2 <- integrate(function(tx) tx^2 * fx(tx), 0, 3)$value
EY2 <- integrate(function(ty) ty^2 * fy(ty), 0, 3)$value

EXY <- integrate(
  function(tx) sapply(tx, function(xx) {
    integrate(function(ty) xx * ty * fxy(xx, ty), 0, 3)$value
  }),
  0, 3
)$value

VX <- EX2 - EX^2
VY <- EY2 - EY^2
COV <- EXY - EX * EY
CORR <- COV / sqrt(VX * VY)

round(c(
  EX = EX,
  EY = EY,
  EXY = EXY,
  VX = VX,
  VY = VY,
  COV = COV,
  CORR = CORR
), 6)

# Plot marginal density of T_x
tx_grid <- seq(0, 3, by = 0.01)
plot(
  tx_grid, fx(tx_grid), type = "l", lwd = 2,
  xlab = expression(t[x]),
  ylab = expression(f[x](t[x])),
  main = "Example 12.13: Marginal Density of T_x"
)



# R Check for Example 12.15
# Device fails on first failure of its two components

fxy <- function(tx, ty) {
  ifelse(tx > 0 & tx < 3 & ty > 0 & ty < 3, (tx + ty) / 27, 0)
}

p_joint <- integrate(
  function(tx) sapply(tx, function(xx) {
    integrate(function(ty) fxy(xx, ty), 1, 3)$value
  }),
  1, 3
)$value

q_first_month <- 1 - p_joint

round(c(
  p_joint = p_joint,
  q_first_month = q_first_month
), 6)

# Small visual check over n in [0,3]
n_grid <- seq(0, 3, by = 0.05)
q_grid <- sapply(n_grid, function(n) {
  p_joint_n <- integrate(
    function(tx) sapply(tx, function(xx) {
      integrate(function(ty) fxy(xx, ty), n, 3)$value
    }),
    n, 3
  )$value
  1 - p_joint_n
})

plot(
  n_grid, q_grid, type = "l", lwd = 2,
  xlab = "n",
  ylab = expression(n*q[xy]),
  main = "Example 12.15: Joint-Life Failure Probability by Time n"
)
abline(v = 1, lty = 2)
abline(h = q_first_month, lty = 2)



# R Check for Example 12.16
# Compare last-survivor APV under common shock for exponential components

delta <- 0.05
lambda_x <- 0.03
lambda_y <- 0.04

A_last_common_shock <- function(lambda_common) {
  A_x  <- (lambda_x + lambda_common) / (lambda_x + lambda_common + delta)
  A_y  <- (lambda_y + lambda_common) / (lambda_y + lambda_common + delta)
  A_xy <- (lambda_x + lambda_y + lambda_common) /
    (lambda_x + lambda_y + lambda_common + delta)
  A_x + A_y - A_xy
}

A_last_independent <- function() {
  A_x  <- lambda_x / (lambda_x + delta)
  A_y  <- lambda_y / (lambda_y + delta)
  A_xy <- (lambda_x + lambda_y) / (lambda_x + lambda_y + delta)
  A_x + A_y - A_xy
}

lambda_common <- 0.02

base_common <- A_last_common_shock(lambda_common)
base_indep <- A_last_independent()

round(c(
  common_shock = base_common,
  independence = base_indep,
  difference = base_common - base_indep
), 6)

# Sensitivity plot
lambda_grid <- seq(0, 0.08, by = 0.001)
A_grid <- sapply(lambda_grid, A_last_common_shock)

plot(
  lambda_grid, A_grid, type = "l", lwd = 2,
  xlab = expression(lambda[c]),
  ylab = expression(bar(A)[bar(xy)]),
  main = "Example 12.16: Last-Survivor APV under Common Shock"
)
abline(v = lambda_common, lty = 2)
abline(h = base_common, lty = 2)
abline(h = base_indep, lty = 3)






# R Check for Exercise 12-2
# Verify the PDF of T_xy at t = 0.50

qx <- 0.080
qy <- 0.004

p_xy <- function(t) (1 - qx * t^2) * (1 - qy * t^2)
f_xy <- function(t) 2 * 0.084 * t - 0.00128 * t^3

t0 <- 0.50
pdf_exact <- f_xy(t0)

# numerical derivative check
h <- 1e-6
pdf_num <- -(p_xy(t0 + h) - p_xy(t0 - h)) / (2 * h)

round(c(
  pdf_exact = pdf_exact,
  pdf_numeric = pdf_num
), 6)

# plot the joint-life survival function and density
t_grid <- seq(0, 1, by = 0.01)

plot(
  t_grid, p_xy(t_grid), type = "l", lwd = 2,
  xlab = "t",
  ylab = expression(t*p[xy]),
  main = "Exercise 12-2: Joint-Life Survival Function"
)

plot(
  t_grid, f_xy(t_grid), type = "l", lwd = 2,
  xlab = "t",
  ylab = expression(f[xy](t)),
  main = "Exercise 12-2: Joint-Life Density"
)
abline(v = t0, lty = 2)
abline(h = pdf_exact, lty = 2)




# R Check for Exercise 12-4
# Verify e̊_{65:55} for independent nonsmoker age 65 and smoker age 55

p65N <- function(t) ifelse(t >= 0 & t <= 10, 1 - t / 10, 0)
p55S <- function(t) ifelse(t >= 0 & t <= 20, (1 - t / 20)^2, 0)

p_joint <- function(t) p65N(t) * p55S(t)

e_joint <- integrate(p_joint, lower = 0, upper = 10)$value

round(c(e_joint = e_joint), 6)

t_grid <- seq(0, 10, by = 0.05)

plot(
  t_grid, p_joint(t_grid), type = "l", lwd = 2,
  xlab = "t",
  ylab = expression(t*p[65:55]),
  main = "Exercise 12-4: Joint-Life Survival Curve"
)
abline(h = 0, lty = 3)



# R Check for Exercise 12-6
# Verify {}_5 q_{overline{80:85}}

library(mqriskR)

x <- 80
y <- 85
t <- 5

q_last <- tqxybar(
  x = x,
  y = y,
  t = t,
  model = "uniform",
  omega = 100
)

qx <- tqx(x = x, t = t, model = "uniform", omega = 100)
qy <- tqx(x = y, t = t, model = "uniform", omega = 100)

product_check <- qx * qy

round(c(
  q_last = q_last,
  qx = qx,
  qy = qy,
  product_check = product_check
), 6)

# sensitivity to horizon
t_grid <- seq(0, 10, by = 0.1)
q_last_grid <- sapply(t_grid, function(tt) {
  tqxybar(x = x, y = y, t = tt, model = "uniform", omega = 100)
})

plot(
  t_grid, q_last_grid, type = "l", lwd = 2,
  xlab = "t",
  ylab = expression(t*q[bar(80:85)]),
  main = "Exercise 12-6: Last-Survivor Failure Probability"
)
abline(v = t, lty = 2)
abline(h = q_last, lty = 2)



# R Check for Exercise 12-11
# Verify Pr(T_I < T_II)

S_I <- function(x, a = 1.80) ((9 - x) / 9)^a
S_II <- function(x, b = 1.50) ((9 - x) / 9)^b
f_I <- function(x, a = 1.80) ((9 - x) / 9)^a * a / (9 - x)

prob <- integrate(function(x) f_I(x) * S_II(x), lower = 0, upper = 9)$value
round(c(prob = prob), 6)

# sensitivity plot in the coefficient for machine I
a_grid <- seq(0.8, 2.8, by = 0.02)
prob_grid <- sapply(a_grid, function(a) {
  integrate(function(x) f_I(x, a = a) * S_II(x), lower = 0, upper = 9)$value
})

plot(
  a_grid, prob_grid, type = "l", lwd = 2,
  xlab = "Hazard coefficient for Machine I",
  ylab = expression(Pr(T[I] < T[II])),
  main = "Exercise 12-11: Contingent Failure Probability"
)
abline(v = 1.80, lty = 2)
abline(h = prob, lty = 2)



# R Check for Exercise 12-15
# Verify P_xy from the given premium information

d <- 0.06
Px <- 0.10
Pxybar <- 0.06

adotx <- 1 / (Px + d)
Ax <- 1 - d * adotx

# Solve for adotxy from
# (.625 + .625 - 1 + .06*adotxy) / (6.25 + 6.25 - adotxy) = .06

adotxy <- 50 / 12
Axy <- 1 - d * adotxy
Pxy <- Axy / adotxy

round(c(
  adotx = adotx,
  Ax = Ax,
  adotxy = adotxy,
  Axy = Axy,
  Pxy = Pxy,
  Pxybar = Pxybar
), 6)

# compare P_xy as a function of adotxy
adotxy_grid <- seq(3.5, 6.0, by = 0.01)
Pxy_grid <- (1 - d * adotxy_grid) / adotxy_grid

plot(
  adotxy_grid, Pxy_grid, type = "l", lwd = 2,
  xlab = expression(ddot(a)[xy]),
  ylab = expression(P[xy]),
  main = "Exercise 12-15: Joint-Life Premium vs Annuity Value"
)
abline(v = adotxy, lty = 2)
abline(h = Pxy, lty = 2)




# R Check for Exercise 12-32
# Compare independence and common-shock APVs

delta <- 0.05
mu_total <- 0.06
lambda_common <- 0.02

# Independence case
A_indep <- mu_total / (mu_total + delta) +
  mu_total / (mu_total + delta) -
  (2 * mu_total) / (2 * mu_total + delta)

# Common-shock case
mu_star <- mu_total - lambda_common
mu_joint_common <- mu_star + mu_star + lambda_common

A_common <- mu_total / (mu_total + delta) +
  mu_total / (mu_total + delta) -
  mu_joint_common / (mu_joint_common + delta)

increase <- A_common - A_indep

round(c(
  A_indep = A_indep,
  A_common = A_common,
  increase = increase
), 6)

# sensitivity plot in lambda_common
lambda_grid <- seq(0, 0.05, by = 0.001)
A_common_grid <- sapply(lambda_grid, function(lam) {
  mu_star <- mu_total - lam
  mu_joint <- mu_star + mu_star + lam
  mu_total / (mu_total + delta) +
    mu_total / (mu_total + delta) -
    mu_joint / (mu_joint + delta)
})

plot(
  lambda_grid, A_common_grid, type = "l", lwd = 2,
  xlab = expression(lambda),
  ylab = expression(bar(A)[bar(xy)]),
  main = "Exercise 12-32: Common Shock and Last-Survivor APV"
)
abline(v = lambda_common, lty = 2)
abline(h = A_common, lty = 2)
abline(h = A_indep, lty = 3)




# R Check for Exercise 12-33
# Verify {}_5 p_xy under common shock

p_x <- 0.96
p_y <- 0.97
lambda <- 0.01
t <- 5

p_x_star <- p_x * exp(lambda)
p_y_star <- p_y * exp(lambda)

p5_xy <- (p_x_star^t) * (p_y_star^t) * exp(-lambda * t)

round(c(
  p_x_star = p_x_star,
  p_y_star = p_y_star,
  p5_xy = p5_xy
), 6)

# sensitivity plot
lambda_grid <- seq(0, 0.03, by = 0.0005)
p5_grid <- sapply(lambda_grid, function(lam) {
  pxs <- p_x * exp(lam)
  pys <- p_y * exp(lam)
  (pxs^t) * (pys^t) * exp(-lam * t)
})

plot(
  lambda_grid, p5_grid, type = "l", lwd = 2,
  xlab = expression(lambda),
  ylab = expression(5*p[xy]),
  main = "Exercise 12-33: Five-Year Joint Survival under Common Shock"
)
abline(v = lambda, lty = 2)
abline(h = p5_xy, lty = 2)

# R Check for Example 8.2
library(mqriskR)

x <- 40
i <- 0.05
omega <- 100

v <- 1 / (1 + i)
d <- i / (1 + i)

Ax_val <- Ax(x = x, i = i, model = "uniform", omega = omega)
ax_val <- ax(x = x, i = i, model = "uniform", omega = omega)

lhs <- Ax_val
rhs <- v - d * ax_val

round(c(lhs = lhs, rhs = rhs, difference = lhs - rhs), 10)



# R Check for Example 8.3
i <- 0.05
ip <- (1 + i)^2 - 1

v  <- 1 / (1 + i)
vp <- 1 / (1 + ip)

d  <- i / (1 + i)

lx_vals <- c(l95 = 100, l96 = 70, l97 = 40, l98 = 20, l99 = 4, l100 = 0)

surv <- c(
  lx_vals["l96"] / lx_vals["l95"],
  lx_vals["l97"] / lx_vals["l95"],
  lx_vals["l98"] / lx_vals["l95"],
  lx_vals["l99"] / lx_vals["l95"]
)

a95  <- sum(v^(1:4)  * surv)
a95_2 <- sum(vp^(1:4) * surv)

A95  <- v  - d * a95
dp   <- ip / (1 + ip)
A95_2 <- vp - dp * a95_2

var_y <- (A95_2 - A95^2) / d^2

round(c(
  a95 = a95,
  a95_2 = a95_2,
  A95 = A95,
  A95_2 = A95_2,
  varY95 = var_y
), 4)


# R Check for Example 8.7
i <- 0.06
n <- 10
coupon <- 40

v <- 1 / (1 + i)
surv <- 0.98^(1:n)

epv <- coupon * sum(v^(1:n) * surv)

round(epv, 2)


# R Check for Example 8.8
i <- 0.05
ip <- 0.1025

d  <- i / (1 + i)
dp <- ip / (1 + ip)

adot_a5  <- 4.13038
adot2_a5 <- 3.79209

Axn  <- 1 - d  * adot_a5
A2xn <- 1 - dp * adot2_a5

var_temp_due <- (A2xn - Axn^2) / d^2

round(c(
  Axn = Axn,
  A2xn = A2xn,
  variance = var_temp_due
), 5)


# R Check for Example 8.14
q80_1990 <- 0.05105
improvement <- 0.005
reduction_factor <- 1 - improvement

q80_2010 <- q80_1990 * reduction_factor^20
p80_2010 <- 1 - q80_2010

round(c(
  q80_2010 = q80_2010,
  p80_2010 = p80_2010
), 5)



# R Check for Example 8.17 b (i)

library(mqriskR)

delta <- 0.04
mu <- 0.01
i <- exp(delta) - 1

# -------------------------------
# Exact values from formulas in the text
# -------------------------------

a_cont_exact <- 1 / (mu + delta)
a_due_exact  <- 1 / (1 - exp(-(mu + delta)))

d <- 1 - exp(-delta)

udd_formula <- (i * d / delta^2) * a_due_exact - (i - delta) / delta^2
w2_formula  <- a_due_exact - 0.5
w3_formula  <- a_due_exact - 0.5 - (mu + delta) / 12


# -------------------------------
# Using package functions
# -------------------------------

abar_pkg <- abarx(
  x = 50,
  i = i,
  model = "exponential",
  lambda = mu
)

adot_pkg <- adotx(
  x = 50,
  i = i,
  model = "exponential",
  lambda = mu
)

udd_pkg <- abarx_udd(
  x = 50,
  i = i,
  model = "exponential",
  lambda = mu
)

w2_pkg <- abarx_woolhouse2(
  x = 50,
  i = i,
  model = "exponential",
  lambda = mu
)

w3_pkg <- abarx_woolhouse3(
  x = 50,
  i = i,
  model = "exponential",
  lambda = mu
)


# -------------------------------
# Compare results
# -------------------------------

round(c(
  abar_exact = a_cont_exact,
  abar_pkg   = abar_pkg,

  adot_exact = a_due_exact,
  adot_pkg   = adot_pkg,

  udd_formula = udd_formula,
  udd_pkg     = udd_pkg,

  woolhouse2_formula = w2_formula,
  woolhouse2_pkg     = w2_pkg,

  woolhouse3_formula = w3_formula,
  woolhouse3_pkg     = w3_pkg
), 5)



# R Check for Exercise 8-4

adotx_val <- 10
A2adotx_val <- 6
i <- 1 / 24

d <- i / (1 + i)
ip <- (1 + i)^2 - 1
dp <- ip / (1 + ip)

Ax_val <- 1 - d * adotx_val
A2x_val <- 1 - dp * A2adotx_val

var_val <- (A2x_val - Ax_val^2) / d^2

round(c(
  d = d,
  ip = ip,
  dp = dp,
  Ax = Ax_val,
  A2x = A2x_val,
  variance = var_val
), 5)



# R Check for Exercise 8-7

i <- 0.05
v <- 1 / (1 + i)

px <- 0.99
px1 <- 0.95
px1_new <- 0.98
adotx1 <- 6.951

adotx2 <- ((adotx1 - 1) * (1 + i)) / px1

adotx_old <- 1 + v * px * (1 + v * px1 * adotx2)
adotx_new <- 1 + v * px * (1 + v * px1_new * adotx2)

increase <- adotx_new - adotx_old

round(c(
  adotx2 = adotx2,
  old_value = adotx_old,
  new_value = adotx_new,
  increase = increase
), 5)



# R Check for Exercise 8-8

annuity_payment <- 12000
d <- 0.08

B_star <- annuity_payment / d

B_star

B <- seq(100000, 200000, by = 1000)
coef_sq <- (B - B_star)^2

plot(B, coef_sq, type = "l",
     xlab = "Death benefit B",
     ylab = "Proportional variance term",
     main = "Variance minimized when B = 150,000")
abline(v = B_star, lty = 2)
grid()



# R Check for Exercise 8-13

abarx <- 10
A2abarx <- 7.375
varY <- 50

# From the algebra in the text:
# varY = (5.25*delta - 100*delta^2)/delta^2 = 5.25/delta - 100
delta <- 5.25 / 150
Abarx <- 1 - delta * abarx

round(c(delta = delta, Abarx = Abarx), 5)




# R Check for Exercise 8-15

delta <- 0.03
mu <- 0.025

threshold_t <- log(1 - 20 * delta) / (-delta)
prob <- exp(-mu * threshold_t)

round(c(
  threshold_t = threshold_t,
  probability = prob
), 5)




# R Check for Exercise 8-18

delta <- 0.10

Abar_m <- 0.15
Abar_f <- 0.09

varY_m <- 5.00
varY_f <- 4.00

A2bar_m <- varY_m * delta^2 + Abar_m^2
A2bar_f <- varY_f * delta^2 + Abar_f^2

Abar_uncond <- 0.5 * (Abar_m + Abar_f)
A2bar_uncond <- 0.5 * (A2bar_m + A2bar_f)

varY_uncond <- (A2bar_uncond - Abar_uncond^2) / delta^2

round(c(
  A2bar_m = A2bar_m,
  A2bar_f = A2bar_f,
  Abar_uncond = Abar_uncond,
  A2bar_uncond = A2bar_uncond,
  varY_uncond = varY_uncond
), 5)



# R Check for Exercise 8-27

y <- c(1.00, 1.87, 2.62)
p <- c(0.10, 0.90 - 0.81, 0.81)

EY <- sum(y * p)
EY2 <- sum(y^2 * p)
VarY <- EY2 - EY^2

round(c(EY = EY, EY2 = EY2, VarY = VarY), 5)


# R Check for Exercise 8-33

adot65 <- 9.90
A35_30 <- 0.21
A35_30_1 <- 0.07

E30_35 <- A35_30 - A35_30_1
NSP <- (E30_35 * adot65) / (1 - A35_30_1)

round(c(E30_35 = E30_35, NSP = NSP), 5)


# R Check for Exercise 8-48

v <- 0.90
y <- c(
  2,
  2 + 3 * v,
  2 + 3 * v + 4 * v^2
)

p <- c(0.20, 0.20, 0.60)

EY <- sum(y * p)
EY2 <- sum(y^2 * p)
VarY <- EY2 - EY^2

round(c(EY = EY, EY2 = EY2, VarY = VarY), 5)


# R Check for Exercise 8-52

qx2000 <- c(0.000475, 0.000514, 0.000554, 0.000598, 0.000648)
AAx <- c(0.011, 0.012, 0.013, 0.014, 0.015)
years_forward <- c(8, 9, 10, 11, 12)

px_proj <- 1 - qx2000 * (1 - AAx)^years_forward
q5 <- 1 - prod(px_proj)

round(c(px_proj, q5 = q5), 6)

# Chapter 4: R solution support (selected exercises)
# Exercises: 4-4 and 4-8
#
# Goal:
# - Simulate/compute the one-year claim risk for a small term insurer.
# - Explore "How many policies should we sell?" and "How much capital is needed?"
#
# Model (simple, explicit assumptions):
# - One-year renewable term, benefit = 1000
# - Premium per policy = 50
# - Initial paid-in capital (given by exercise)
# - Number of insured lives = n
# - Each life has independent probability q of dying within 1 year
# - Claim count ~ Binomial(n, q), aggregate claims = 1000 * ClaimCount
#
# NOTE: Exercise 4-4 and 4-8 do NOT specify q.
# So we treat q as a parameter and do sensitivity analysis.
#
# --- Helpers ---------------------------------------------------------------

simulate_one_year <- function(n, q, benefit = 1000, prem = 50, capital = 2000,
                              n_sims = 200000, seed = 1) {
  set.seed(seed)
  claims_count <- rbinom(n_sims, size = n, prob = q)
  claims <- benefit * claims_count
  premium_income <- prem * n

  ending_surplus <- capital + premium_income - claims
  ruined <- ending_surplus < 0

  list(
    n = n, q = q, benefit = benefit, prem = prem, capital = capital,
    prob_ruin = mean(ruined),
    expected_claims = benefit * n * q,
    expected_profit = capital + premium_income - benefit * n * q,
    ending_surplus = ending_surplus
  )
}

# Exact (non-simulation) probability of ruin under Binomial claims:
# Ruin occurs if capital + prem*n - benefit*X < 0  =>  X > (capital + prem*n)/benefit
prob_ruin_exact <- function(n, q, benefit = 1000, prem = 50, capital = 2000) {
  threshold <- (capital + prem * n) / benefit
  k0 <- floor(threshold)  # ruin if X >= k0 + 1
  if (k0 >= n) return(0)
  if (k0 < 0) return(1)
  pbinom(k0, size = n, prob = q, lower.tail = FALSE)  # P[X > k0]
}

# Capital required for a target ruin probability alpha:
# Find smallest capital such that P(ruin) <= alpha.
required_capital_for_ruin <- function(n, q, alpha = 0.01, benefit = 1000, prem = 50,
                                      cap_grid = seq(0, 200000, by = 100)) {
  pr <- vapply(cap_grid, function(c0) prob_ruin_exact(n, q, benefit, prem, c0), numeric(1))
  idx <- which(pr <= alpha)
  if (length(idx) == 0) return(NA_real_)
  cap_grid[min(idx)]
}

# Risk measures (VaR / TVaR) of aggregate claims, for a given n,q:
claims_risk_measures <- function(n, q, p = 0.99, benefit = 1000) {
  # Claim count distribution is Binomial(n,q)
  # VaR of claims is benefit * quantile of Binomial
  k_var <- qbinom(p, size = n, prob = q)
  var_claims <- benefit * k_var

  # TVaR (conditional tail mean) for discrete distribution:
  ks <- 0:n
  probs <- dbinom(ks, size = n, prob = q)
  claims <- benefit * ks

  tail <- claims >= var_claims
  tvar_claims <- sum(claims[tail] * probs[tail]) / sum(probs[tail])

  list(VaR = var_claims, TVaR = tvar_claims, k_var = k_var)
}

# --- Exercise 4-4 ----------------------------------------------------------
# Risk Mitigation, Inc.:
# capital = 2000, benefit = 1000, prem = 50
# Question asks how many policies to sell and how much capital to hold.
#
# We'll explore ruin probabilities for a range of n under plausible q values.

ex4_4_analysis <- function(q_values = c(0.001, 0.002, 0.005, 0.01),
                           n_values = c(10, 20, 30, 40, 50, 75, 100, 150, 200),
                           capital = 2000, benefit = 1000, prem = 50,
                           alpha_targets = c(0.10, 0.05, 0.01)) {

  out <- list()

  for (q in q_values) {
    tab <- data.frame(
      n = n_values,
      q = q,
      prob_ruin = NA_real_,
      expected_claims = NA_real_,
      expected_profit_ex_capital = NA_real_  # premium - expected claims
    )

    for (i in seq_along(n_values)) {
      n <- n_values[i]
      tab$prob_ruin[i] <- prob_ruin_exact(n, q, benefit, prem, capital)
      tab$expected_claims[i] <- benefit * n * q
      tab$expected_profit_ex_capital[i] <- prem * n - benefit * n * q
    }

    # Capital requirement for fixed n and target ruin probs
    cap_req <- expand.grid(n = n_values, alpha = alpha_targets)
    cap_req$q <- q
    cap_req$cap_required <- mapply(
      function(n, alpha) required_capital_for_ruin(n, q, alpha, benefit, prem),
      cap_req$n, cap_req$alpha
    )

    out[[paste0("q=", q)]] <- list(table = tab, capital_requirements = cap_req)
  }

  out
}

# Run and print a compact view:
res_4_4 <- ex4_4_analysis()

# Print a quick summary for one q (you can change which one):
print(res_4_4[["q=0.005"]]$table)
print(res_4_4[["q=0.005"]]$capital_requirements)

# --- Exercise 4-8 ----------------------------------------------------------
# New Beginnings, Inc.:
# capital = 3000, and a client wants 1,000 one-year policies.
#
# We quantify the concentration risk via ruin probability and capital needed.

ex4_8_analysis <- function(q = 0.005, n = 1000, capital = 3000,
                           benefit = 1000, prem = 50,
                           alpha_targets = c(0.10, 0.05, 0.01, 0.001)) {

  pr <- prob_ruin_exact(n, q, benefit, prem, capital)
  rm <- claims_risk_measures(n, q, p = 0.99, benefit = benefit)

  cap_req <- data.frame(
    alpha = alpha_targets,
    cap_required = vapply(alpha_targets, function(a)
      required_capital_for_ruin(n, q, a, benefit, prem), numeric(1)
    )
  )

  list(
    inputs = list(n = n, q = q, capital = capital, benefit = benefit, prem = prem),
    prob_ruin = pr,
    claims_VaR_99 = rm$VaR,
    claims_TVaR_99 = rm$TVaR,
    capital_requirements = cap_req
  )
}

res_4_8 <- ex4_8_analysis(q = 0.005, n = 1000, capital = 3000)
print(res_4_8)

# Optional: Monte Carlo sanity check for 4-8 (can be slow at large n_sims)
# sim_4_8 <- simulate_one_year(n = 1000, q = 0.005, capital = 3000, n_sims = 200000)
# sim_4_8$prob_ruin

##Listing 17.1: Profit vector calculation
library(mqriskR)

V <- c(0, 5.66, 6.17, 0)
qx <- c(0.00142, 0.00153, 0.00166)

Pr <- Pr_vector_disc(V = V, G = 95, i = 0.06, r = 0.05, e = 10, q1 = qx, b1 = 50000,pre_contract_expense = 15)

round(Pr, 2)



##Listing 17.2: Profit signature
p_tau <- c(0.99858, 0.99847, 0.99834)

Pi <- Pi_signature(Pr, p_tau = p_tau)

round(Pi, 2)



##Listing 17.3: Net present value
NPV_profit(Pi, r = 0.10)


##Listing 17.4: Internal rate of return
IRR_profit(Pi, interval = c(0, 1))



##Listing 17.5: Profit margin
APV_GP <- APV_gross_premiums(G = rep(95, 3), r = 0.10, p_tau = p_tau)

NPV_val <- NPV_profit(Pi, r = 0.10)

profit_margin(NPV_val, APV_GP)



##Listing 17.6: Partial NPV and payback period
partial <- NPV_partial(Pi, r = 0.10)

payback <- discounted_payback_period(Pi, r = 0.10)

round(partial, 2)
payback


##Listing 17.7: Zeroized reserves
Vz <- V_zeroized(qx = c(0.015, 0.017, 0.019, 0.021, 0.024), i = 0.06, G = 19279, benefit = 1000000, e = 240)

round(Vz, 2)

# Chapter 4: R solution support (selected exercises)
# Exercises: 4-5 and 4-9
#
# Goal:
# - Simulate/compute the one-year claim risk for a small term insurer.
# - Explore "How many policies should we sell?" and "How much capital is needed?"
#
# Model (simple, explicit assumptions):
# - One-year renewable term, benefit = 1000
# - Premium per policy = 50
# - Initial paid-in capital (given by exercise)
# - Number of insured lives = n
# - Each life has independent probability q of dying within 1 year
# - Claim count ~ Binomial(n, q), aggregate claims = 1000 * ClaimCount
#
# NOTE: Exercise 4-5 and 4-9 do NOT specify q.
# So we treat q as a parameter and do sensitivity analysis.
#
# --- Helpers ---------------------------------------------------------------

simulate_one_year <- function(n, q, benefit = 1000, prem = 50, capital = 2000,
                              n_sims = 200000, seed = 1) {
  set.seed(seed)
  claims_count <- rbinom(n_sims, size = n, prob = q)
  claims <- benefit * claims_count
  premium_income <- prem * n

  ending_surplus <- capital + premium_income - claims
  ruined <- ending_surplus < 0

  list(
    n = n, q = q, benefit = benefit, prem = prem, capital = capital,
    prob_ruin = mean(ruined),
    expected_claims = benefit * n * q,
    expected_profit = capital + premium_income - benefit * n * q,
    ending_surplus = ending_surplus
  )
}

# Exact (non-simulation) probability of ruin under Binomial claims:
# Ruin occurs if capital + prem*n - benefit*X < 0  =>  X > (capital + prem*n)/benefit
prob_ruin_exact <- function(n, q, benefit = 1000, prem = 50, capital = 2000) {
  threshold <- (capital + prem * n) / benefit
  k0 <- floor(threshold)  # ruin if X >= k0 + 1
  if (k0 >= n) return(0)
  if (k0 < 0) return(1)
  pbinom(k0, size = n, prob = q, lower.tail = FALSE)  # P[X > k0]
}

# Capital required for a target ruin probability alpha:
# Find smallest capital such that P(ruin) <= alpha.
required_capital_for_ruin <- function(n, q, alpha = 0.01, benefit = 1000, prem = 50,
                                      cap_grid = seq(0, 200000, by = 100)) {
  pr <- vapply(cap_grid, function(c0) prob_ruin_exact(n, q, benefit, prem, c0), numeric(1))
  idx <- which(pr <= alpha)
  if (length(idx) == 0) return(NA_real_)
  cap_grid[min(idx)]
}

# Risk measures (VaR / TVaR) of aggregate claims, for a given n,q:
claims_risk_measures <- function(n, q, p = 0.99, benefit = 1000) {
  # Claim count distribution is Binomial(n,q)
  # VaR of claims is benefit * quantile of Binomial
  k_var <- qbinom(p, size = n, prob = q)
  var_claims <- benefit * k_var

  # TVaR (conditional tail mean) for discrete distribution:
  ks <- 0:n
  probs <- dbinom(ks, size = n, prob = q)
  claims <- benefit * ks

  tail <- claims >= var_claims
  tvar_claims <- sum(claims[tail] * probs[tail]) / sum(probs[tail])

  list(VaR = var_claims, TVaR = tvar_claims, k_var = k_var)
}

# --- Exercise 4-5 ----------------------------------------------------------
# Risk Mitigation, Inc.:
# capital = 2000, benefit = 1000, prem = 50
# Question asks how many policies to sell and how much capital to hold.
#
# We'll explore ruin probabilities for a range of n under plausible q values.

ex4_5_analysis <- function(q_values = c(0.001, 0.002, 0.005, 0.01),
                           n_values = c(10, 20, 30, 40, 50, 75, 100, 150, 200),
                           capital = 2000, benefit = 1000, prem = 50,
                           alpha_targets = c(0.10, 0.05, 0.01)) {

  out <- list()

  for (q in q_values) {
    tab <- data.frame(
      n = n_values,
      q = q,
      prob_ruin = NA_real_,
      expected_claims = NA_real_,
      expected_profit_ex_capital = NA_real_  # premium - expected claims
    )

    for (i in seq_along(n_values)) {
      n <- n_values[i]
      tab$prob_ruin[i] <- prob_ruin_exact(n, q, benefit, prem, capital)
      tab$expected_claims[i] <- benefit * n * q
      tab$expected_profit_ex_capital[i] <- prem * n - benefit * n * q
    }

    # Capital requirement for fixed n and target ruin probs
    cap_req <- expand.grid(n = n_values, alpha = alpha_targets)
    cap_req$q <- q
    cap_req$cap_required <- mapply(
      function(n, alpha) required_capital_for_ruin(n, q, alpha, benefit, prem),
      cap_req$n, cap_req$alpha
    )

    out[[paste0("q=", q)]] <- list(table = tab, capital_requirements = cap_req)
  }

  out
}

# Run and print a compact view:
res_4_5 <- ex4_5_analysis()

# Print a quick summary for one q (you can change which one):
print(res_4_5[["q=0.005"]]$table)
print(res_4_5[["q=0.005"]]$capital_requirements)

# --- Exercise 4-9 ----------------------------------------------------------
# New Beginnings, Inc.:
# capital = 3000, and a client wants 1,000 one-year policies.
#
# We quantify the concentration risk via ruin probability and capital needed.

ex4_9_analysis <- function(q = 0.005, n = 1000, capital = 3000,
                           benefit = 1000, prem = 50,
                           alpha_targets = c(0.10, 0.05, 0.01, 0.001)) {

  pr <- prob_ruin_exact(n, q, benefit, prem, capital)
  rm <- claims_risk_measures(n, q, p = 0.99, benefit = benefit)

  cap_req <- data.frame(
    alpha = alpha_targets,
    cap_required = vapply(alpha_targets, function(a)
      required_capital_for_ruin(n, q, a, benefit, prem), numeric(1)
    )
  )

  list(
    inputs = list(n = n, q = q, capital = capital, benefit = benefit, prem = prem),
    prob_ruin = pr,
    claims_VaR_99 = rm$VaR,
    claims_TVaR_99 = rm$TVaR,
    capital_requirements = cap_req
  )
}

res_4_9 <- ex4_9_analysis(q = 0.005, n = 1000, capital = 3000)
print(res_4_9)

# Optional: Monte Carlo sanity check for 4-9 (can be slow at large n_sims)
# sim_4_9 <- simulate_one_year(n = 1000, q = 0.005, capital = 3000, n_sims = 200000)
# sim_4_9$prob_ruin

## R check for Example 7.1
k <- 0:4
pk <- c(0.20, 0.30, 0.20, 0.15, 0.15)
i <- 0.01

z <- 10 * (1 + i)^(-(k + 1))

EZ  <- sum(z * pk)
EZ2 <- sum(z^2 * pk)
VarZ <- EZ2 - EZ^2

cdf <- cumsum(pk[order(z)])
z_sorted <- sort(z)
median_z <- z_sorted[min(which(cdf >= 0.50))]

round(c(EZ = EZ, VarZ = VarZ, median = median_z), 5)

plot(z_sorted, cdf, type = "s",
     xlab = "z", ylab = "F_Z(z)",
     main = "Example 7.1: CDF of the present value random variable")
grid()




## R check for Example 7.5
# Example 7.1 distribution reused
k <- 0:4
pk <- c(0.20, 0.30, 0.20, 0.15, 0.15)
i <- 0.01
v <- 1 / (1 + i)

# 3-year term insurance part
Z_term <- ifelse(k < 3, v^(k + 1), 0)

# 3-year pure endowment part
Z_pe <- ifelse(k >= 3, v^3, 0)

# 3-year endowment insurance
Z_endow <- Z_term + Z_pe

A_term <- sum(Z_term * pk)
A_pe   <- sum(Z_pe * pk)
A_end  <- sum(Z_endow * pk)

A2_term <- sum(Z_term^2 * pk)
A2_pe   <- sum(Z_pe^2 * pk)
A2_end  <- sum(Z_endow^2 * pk)

cov_tp <- sum(Z_term * Z_pe * pk) - A_term * A_pe

round(c(A_term = A_term, A_pe = A_pe, A_end = A_end,
        A2_term = A2_term, A2_pe = A2_pe, A2_end = A2_end,
        covariance = cov_tp), 5)



## R check for Example 7.7
library(mqriskR)

i <- 0.05
delta <- interest_convert(i = i)$delta
lambda <- 0.25

Abar <- Abarx(x = 0, i = i, model = "exponential", lambda = lambda)
VarAbar <- var_Abarx(x = 0, i = i, model = "exponential", lambda = lambda)

# 90th percentile of Z = exp(-delta T)
p90 <- 0.90^(delta / lambda)

round(c(Abar = Abar, Var = VarAbar, p90 = p90), 5)

t_grid <- seq(0, 20, by = 0.01)
z_grid <- exp(-delta * t_grid)

plot(t_grid, z_grid, type = "l", lwd = 2,
     xlab = "t", ylab = "z = exp(-delta t)",
     main = "Example 7.7: Present value as a function of failure time")
grid()




## R check for Example 7.9
k <- 0:4
pk <- c(0.20, 0.30, 0.20, 0.15, 0.15)
i <- 0.01
v <- 1 / (1 + i)

IAx <- sum((k + 1) * v^(k + 1) * pk)

n <- 3
DAxn1 <- sum((n - k[1:n]) * v^(k[1:n] + 1) * pk[1:n])

round(c(IAx = IAx, DAxn1 = DAxn1), 5)




# Exercise 7-3
# Cov(Z_{x:n}^1, {}_{n|}Z_x) for the Example 7.1 model

k <- 0:4
pk <- c(0.20, 0.30, 0.20, 0.15, 0.15)
i <- 0.01
v <- 1 / (1 + i)
n <- 3

U <- ifelse(k < n, v^(k + 1), 0)   # term insurance part
V <- ifelse(k >= n, v^(k + 1), 0)  # deferred insurance part

EU <- sum(U * pk)
EV <- sum(V * pk)
EUV <- sum(U * V * pk)
cov_uv <- EUV - EU * EV

round(c(EU = EU, EV = EV, EUV = EUV, covariance = cov_uv), 6)

barplot(rbind(U * pk, V * pk), beside = TRUE,
        names.arg = k,
        legend.text = c("Term part", "Deferred part"),
        xlab = "k", ylab = "Probability-weighted PV",
        main = "Exercise 7-3: Term and deferred components")
grid()



# Exercise 7-6
# Check the recursion A_x = v q_x + v p_x A_{x+1}

library(mqriskR)

x <- 40
i <- 0.05

lhs <- Ax(x, i, model = "uniform", omega = 100)
rhs <- discount(i, 1) * (
  tqx(1, x = x, model = "uniform", omega = 100) +
    tpx(1, x = x, model = "uniform", omega = 100) *
    Ax(x + 1, i, model = "uniform", omega = 100)
)

round(c(lhs = lhs, rhs = rhs, difference = lhs - rhs), 10)

ages <- 40:70
lhs_vec <- sapply(ages, function(a) Ax(a, i, model = "uniform", omega = 100))
rhs_vec <- sapply(ages, function(a) {
  discount(i, 1) * (
    tqx(1, x = a, model = "uniform", omega = 100) +
      tpx(1, x = a, model = "uniform", omega = 100) *
      Ax(a + 1, i, model = "uniform", omega = 100)
  )
})

plot(ages, lhs_vec - rhs_vec, type = "h",
     xlab = "x", ylab = "LHS - RHS",
     main = "Exercise 7-6: Recursion check error")
abline(h = 0, lty = 2)
grid()




# Exercise 7-11
# Check {}_{n|}A_x = {}_nE_x * A_{x+n}

library(mqriskR)

x <- 40
n <- 10
i <- 0.05

lhs <- nAx(x, n, i, model = "uniform", omega = 100)
rhs <- nEx(x, n, i, model = "uniform", omega = 100) *
  Ax(x + n, i, model = "uniform", omega = 100)

round(c(lhs = lhs, rhs = rhs, difference = lhs - rhs), 10)

n_grid <- 1:25
lhs_vec <- sapply(n_grid, function(nn) nAx(x, nn, i, model = "uniform", omega = 100))
rhs_vec <- sapply(n_grid, function(nn) {
  nEx(x, nn, i, model = "uniform", omega = 100) *
    Ax(x + nn, i, model = "uniform", omega = 100)
})

plot(n_grid, lhs_vec, type = "l", lwd = 2,
     xlab = "n", ylab = "Value",
     main = "Exercise 7-11: Deferred insurance identity")
lines(n_grid, rhs_vec, lty = 2, lwd = 2)
legend("topright", legend = c("nAx", "nEx * A_{x+n}"),
       lty = c(1, 2), lwd = 2, bty = "n")
grid()



# Exercise 7-15
# Continuous APV with given density

delta <- 0.10

f <- function(t) ifelse(t > 0 & t <= 100, t / 5000, 0)
integrand <- function(t) 50 * exp(-delta * t) * f(t)

apv_num <- integrate(integrand, lower = 0, upper = 100)$value
apv_formula <- 1 - 11 * exp(-10)

round(c(numerical = apv_num, closed_form = apv_formula,
        difference = apv_num - apv_formula), 8)

t_grid <- seq(0, 100, by = 0.1)
plot(t_grid, integrand(t_grid), type = "l", lwd = 2,
     xlab = "t", ylab = "Discounted density contribution",
     main = "Exercise 7-15: Integrand for the APV")
grid()



# Exercise 7-21
# Deferred continuous insurance under exponential lifetime

delta <- 0.05
lambda <- 0.25
n <- 10

integrand <- function(t) exp(-delta * t) * exp(-lambda * t) * lambda

lhs_num <- integrate(integrand, lower = n, upper = Inf)$value
rhs_formula <- (lambda / (delta + lambda)) * exp(-(delta + lambda) * n)

round(c(numerical = lhs_num, closed_form = rhs_formula,
        difference = lhs_num - rhs_formula), 8)

t_grid <- seq(n, 40, by = 0.05)
plot(t_grid, integrand(t_grid), type = "l", lwd = 2,
     xlab = "t", ylab = "Integrand",
     main = "Exercise 7-21: Deferred continuous insurance integrand")
grid()



# Exercise 7-25
# Check (IA)_x = A_x + {}_1E_x (IA)_{x+1}

IA35 <- 3.711
A35  <- 0.1300
p35  <- 0.9964
v    <- 0.9434

E1_35 <- v * p35
IA36 <- (IA35 - A35) / E1_35

round(c(E1_35 = E1_35, IA36 = IA36), 6)

lhs <- IA35
rhs <- A35 + E1_35 * IA36

round(c(lhs = lhs, rhs = rhs, difference = lhs - rhs), 10)

barplot(c(A35, E1_35 * IA36),
        names.arg = c("A35", "E1_35 * IA36"),
        ylab = "Contribution",
        main = "Exercise 7-25: Decomposition of (IA)35")
grid()



# Exercise 7-28
# Recover A_35 from UDD information

i <- 0.05
delta <- log(1 + i)
q35 <- 0.01
Abar36 <- 0.185
v <- 1 / (1 + i)
p35 <- 1 - q35

Abar35_1 <- (i / delta) * v * q35
Abar35 <- Abar35_1 + v * p35 * Abar36
A35 <- (delta / i) * Abar35

round(c(Abar35_1 = Abar35_1, Abar35 = Abar35, A35 = A35), 6)

barplot(c(Abar35_1, v * p35 * Abar36),
        names.arg = c("1-year term part", "survival continuation"),
        ylab = "Contribution",
        main = "Exercise 7-28: Components of Abar35")
grid()





# Exercise 7-29
# UDD approximation for 2-year term insurance

i <- 0.10
delta <- log(1 + i)
q_x <- 0.05
q_x1 <- 0.08
p_x <- 1 - q_x
v <- 1 / (1 + i)

A_disc <- v * q_x + v^2 * p_x * q_x1
A_cont_udd <- (i / delta) * A_disc

round(c(A_disc = A_disc, Abar_udd = A_cont_udd), 6)

barplot(c(v * q_x, v^2 * p_x * q_x1),
        names.arg = c("year 1", "year 2"),
        ylab = "Contribution",
        main = "Exercise 7-29: Annual term-insurance contributions")
grid()





# Exercise 7-31
# Visualize piecewise-constant vs fully continuous increasing benefit

t_grid <- seq(0, 5, by = 0.001)
b_piece <- floor(t_grid + 1)
b_cont <- t_grid

plot(t_grid, b_piece, type = "s", lwd = 2,
     xlab = "t", ylab = "Benefit amount",
     main = "Exercise 7-31: Piecewise vs continuous increasing benefit")
lines(t_grid, b_cont, lwd = 2, lty = 2)
legend("topleft",
       legend = c("piecewise increasing", "continuous increasing"),
       lty = c(1, 2), lwd = 2, bty = "n")
grid()

## Listing 8.1: Whole life annuity computation from a parametric survival model
library(mqriskR)

x <- 40
i <- 0.05

# Uniform survival model with limiting age omega = 100
ax_val <- ax(x = x, i = i, model = "uniform", omega = 100)

print(ax_val)


## Listing 8.2: Variance of the whole life annuity present value from insurance moments
library(mqriskR)

x <- 95
i <- 0.05
omega <- 100

d <- i / (1 + i)

# First and second insurance moments
A1 <- Ax(x = x, i = i, model = "uniform", omega = omega)
A2 <- A2x(x = x, i = i, model = "uniform", omega = omega)

# Variance using the identity
var_identity <- (A2 - A1^2) / d^2

# Variance using the package variance function
var_pkg <- var_Ax(x = x, i = i, model = "uniform", omega = omega) / d^2

c(identity = var_identity, package = var_pkg)



## Listing 8.3: Temporary life annuity computation
library(mqriskR)

x <- 40
n <- 20
i <- 0.05

axn_val <- axn(x = x, n = n, i = i, model = "uniform", omega = 100)

print(axn_val)



## Listing 4.4: Verification of deferred annuity identities
library(mqriskR)

x <- 40
n <- 20
i <- 0.05

whole <- ax(x = x, i = i, model = "uniform", omega = 100)
temp  <- axn(x = x, n = n, i = i, model = "uniform", omega = 100)

left <- nax(x = x, n = n, i = i, model = "uniform", omega = 100)
right1 <- whole - temp

right2 <- nEx(x = x, n = n, i = i, model = "uniform", omega = 100) *
  ax(x + n, i = i, model = "uniform", omega = 100)

round(c(deferred = left, identity1 = right1, identity2 = right2), 10)




## Listing 8.5: Comparison of annual, m-thly, and continuous whole life annuity values
library(mqriskR)

x <- 40
i <- 0.05
omega <- 100
m_vals <- c(1, 2, 4, 12)

a_annual <- ax(x = x, i = i, model = "uniform", omega = omega)
a_m <- sapply(m_vals, function(m) {
  ax_m(x = x, m = m, i = i, model = "uniform", omega = omega)
})
a_cont <- abarx(x = x, i = i, model = "uniform", omega = omega)

data.frame(
  m = c(m_vals, Inf),
  value = c(a_m, a_cont)
)


## Listing 8.6:Comparison of exact, UDD, and Woolhouse approximations for an \(m\)-thly annuity
library(mqriskR)

x <- 40
i <- 0.05
m <- 12
omega <- 100

exact <- ax_m(x = x, m = m, i = i, model = "uniform", omega = omega)

udd_based <- ax_m_udd(
  x = x,
  m = m,
  i = i,
  model = "uniform",
  omega = omega
)

woolhouse_2 <- ax_m_woolhouse2(
  x = x,
  m = m,
  i = i,
  model = "uniform",
  omega = omega
)

woolhouse_3 <- ax_m_woolhouse3(
  x = x,
  m = m,
  i = i,
  model = "uniform",
  omega = omega
)

c(exact = exact,
  udd = udd_based,
  woolhouse_2 = woolhouse_2,
  woolhouse_3 = woolhouse_3)


## Listing 8.7: Temporary annuity value with and without mortality improvement
library(mqriskR)

x0 <- 60; n <- 5; i <- 0.06
base_year <- 2000
issue_year <- 2008

qx_base <- c(0.0100, 0.0110, 0.0122, 0.0135, 0.0150)
AAx <- c(0.01, 0.01, 0.01, 0.01, 0.01)

a_improved <- axn_improved(
  x0 = x0, n = n, i = i,
  qx_base_vec = qx_base,
  AAx_vec = AAx,
  base_year = base_year,
  issue_year = issue_year
)

p_no_improve <- cumprod(1 - qx_base)
a_no_improve <- sum((1 + i)^(-(1:n)) * p_no_improve)

c(no_improvement = a_no_improve, improvement = a_improved)

# Example 17.7 inputs
library(mqriskR)
G <- 19250
expense <- 240
i <- 0.06

q <- c(0.015, 0.017, 0.019, 0.021, 0.024)
p <- 1 - q

# Given reserves (Table 17.1)
V <- c(0, 2500, 4000, 5000, 4000, 0)

# Pre-contract expense
Pr0 <- -5000

# Profit vector
Pr <- numeric(6)
Pr[1] <- Pr0

for (t in 0:4) {
  inflow <- V[t+1] + G - expense
  accumulation <- inflow * (1 + i)

  outflow <- 1000000 * q[t+1] + V[t+2] * p[t+1]

  Pr[t+2] <- accumulation - outflow
}

round(Pr, 2)

# Or use Package function
Pr_pkg<- Pr_vector_disc(
  V = V,
  G = 19250,
  i = 0.06,
  r = 0,
  e = 240,
  q1 = q,
  b1 = 1000000,
  pre_contract_expense = 5000)

round(Pr_pkg, 2)

# Total expected profit (undiscounted)
total_profit <- sum(Pr)
round(total_profit, 2)

# Plot profit vector
plot(
  0:5, Pr,
  type = "b", pch = 19,
  main = "Example 17.7: Profit Vector",
  xlab = "Policy year",
  ylab = "Expected profit"
)
abline(h = 0, lty = 2)





# Example 17.8
library(mqriskR)
Pr <- c(-5000.00, 2688.10, 1868.60, 485.60, 534.60, 390.60)
p_tau <- c(0.98500, 0.98300, 0.98100, 0.97900, 0.97600)

Pi <- Pi_signature(Pr, p_tau = p_tau)
npv <- NPV_profit(Pi, r = 0.10)
apv_gp <- APV_gross_premiums(
  G = rep(19250, 5),
  r = 0.10,
  p_tau = p_tau
)
pm <- profit_margin(npv, apv_gp)

round(Pi, 2)
round(npv, 2)
round(apv_gp, 2)
round(pm, 5)

plot(
  0:5, Pi,
  type = "b", pch = 19,
  main = "Example 17.8: Profit Signature",
  xlab = "Policy year",
  ylab = "Profit per policy issued"
)
abline(h = 0, lty = 2)







# Example 17.9
library(mqriskR)
V <- c(0, 0, 0, 0, 0, 0)
q1 <- c(0.015, 0.017, 0.019, 0.021, 0.024)
p_tau <- 1 - q1

Pr <- Pr_vector_disc(
  V = V,
  G = 19279,
  i = 0.06,
  r = 0,
  e = 240,
  q1 = q1,
  b1 = 1000000,
  pre_contract_expense = 5000
)

Pi <- Pi_signature(Pr, p_tau = p_tau)
npv <- NPV_profit(Pi, r = 0.10)

round(Pr, 2)
round(Pi, 2)
round(npv, 2)

plot(
  0:5, Pr,
  type = "b", pch = 19,
  main = "Example 17.9: Profit Vector with No Reserves",
  xlab = "Policy year",
  ylab = "Expected profit"
)
abline(h = 0, lty = 2)

plot(
  0:5, Pi,
  type = "b", pch = 19,
  main = "Example 17.9: Profit Signature with No Reserves",
  xlab = "Policy year",
  ylab = "Profit per policy issued"
)
abline(h = 0, lty = 2)





# Example 17.11
library(mqriskR)
Vt <- 6559.89
Vt1 <- 7460.99
G <- 2264.12
B <- 100000

i <- 0.04
i_actual <- 0.06
q <- 0.00812
q_actual <- 0.00700
r <- 0.06
r_actual <- 0.08

# Expected profit
Pr_exp <- (Vt + G * (1 - r)) * (1 + i) -
  B * q - Vt1 * (1 - q)

# Actual profit
Pr_act <- (Vt + G * (1 - r_actual)) * (1 + i_actual) -
  B * q_actual - Vt1 * (1 - q_actual)

GT <- Pr_act - Pr_exp

# Gain by source, calculating interest first
GI <- (Vt + G * (1 - r)) * (i_actual - i)
GM <- (B - Vt1) * (q - q_actual)
GE <- GT - GI - GM

round(c(
  expected_profit = Pr_exp,
  actual_profit = Pr_act,
  total_gain = GT,
  gain_interest = GI,
  gain_mortality = GM,
  gain_expense = GE
), 2)

barplot(
  c(GI, GM, GE, GT),
  names.arg = c("Interest", "Mortality", "Expense", "Total"),
  main = "Example 17.11: Gain by Source",
  ylab = "Amount"
)
abline(h = 0, lty = 2)






# Example 17.12

library(mqriskR)

B_basic <- 50000
PA20 <- 10000

G <- 280
expense <- 50
i <- 0.07
q50 <- 0.00356
p50 <- 1 - q50

A30 <- 0.07570
A50 <- 0.19908
A51 <- 0.20820

V20_basic <- B_basic * (A50 - A30) / (1 - A30)
V21_basic <- B_basic * (A51 - A30) / (1 - A30)

V20_pua <- PA20 * A50
V21_pua <- PA20 * A51

# This reproduces the printed solution
Pr21_book <- (V20_basic + V20_pua + G - expense) * (1 + i) -
  q50 * B_basic -
  p50 * (V21_basic + V21_pua)

Div21_book <- 0.85 * Pr21_book
X21_book <- Div21_book / A51
p_compound_book <- X21_book / (B_basic + PA20)


round(c(
  V20_basic = V20_basic,
  V21_basic = V21_basic,
  V20_pua = V20_pua,
  V21_pua = V21_pua,
  profit_21 = Pr21_book,
  dividend_21 = Div21_book,
  paid_up_addition_21 = X21_book,
  compound_bonus_rate = p_compound_book
), 5)









# Exercise 17-1
library(mqriskR)
Pi <- c(-300, 100, 90, 80, 70)
r <- 0.10

npv <- NPV_profit(Pi, r = r)
irr <- IRR_profit(Pi, interval = c(0.001, 0.20))
payback <- discounted_payback_period(Pi, r = r)

# survival probabilities for APV of gross premiums
q <- c(0.05, 0.06, 0.07, 0.08)
p_tau <- 1 - q

apv_gp <- APV_gross_premiums(
  G = rep(1000, 4),
  r = r,
  p_tau = p_tau
)

pm <- profit_margin(npv, apv_gp)

round(c(
  NPV = npv,
  IRR = irr,
  APV_GP = apv_gp,
  profit_margin = pm
), 5)

payback

# Plot partial NPV path
partial <- NPV_partial(Pi, r = r)

plot(
  0:4, partial,
  type = "b", pch = 19,
  main = "Exercise 17-1: Partial Net Present Values",
  xlab = "t",
  ylab = "NPV(t)"
)
abline(h = 0, lty = 2)




library(mqriskR)

# Exercise 17-2
V <- c(0, 800, 800, 0)
q1 <- c(0.015, 0.017, 0.019)
p_tau <- 1 - q1

Pr <- Pr_vector_disc(
  V = V,
  G = 17500,
  i = 0.05,
  r = 0,
  e = 100,
  q1 = q1,
  b1 = 1000000,
  pre_contract_expense = 3000
)

Pi <- Pi_signature(Pr, p_tau = p_tau)
npv <- NPV_profit(Pi, r = 0.10)
irr <- IRR_profit(Pi, interval = c(0.001, 0.50))
payback <- discounted_payback_period(Pi, r = 0.10)

apv_gp <- APV_gross_premiums(
  G = rep(17500, 3),
  r = 0.10,
  p_tau = p_tau
)

pm <- profit_margin(npv, apv_gp)

round(Pr, 5)
round(Pi, 5)
round(c(
  NPV = npv,
  IRR = irr,
  APV_GP = apv_gp,
  profit_margin = pm
), 5)

payback

# Plot profit vector and profit signature
plot(
  0:3, Pr,
  type = "b", pch = 19,
  main = "Exercise 17-2: Profit Vector",
  xlab = "t",
  ylab = "Pr_t"
)
abline(h = 0, lty = 2)

plot(
  0:3, Pi,
  type = "b", pch = 19,
  main = "Exercise 17-2: Profit Signature",
  xlab = "t",
  ylab = "Pi_t"
)
abline(h = 0, lty = 2)






# Exercise 17-3

library(mqriskR)

Pr <- c(-300, 75, 250)
p40 <- 1 - 0.00278
target_pm <- -0.10

Pi <- c(
  Pr[1],
  Pr[2],
  Pr[3] * p40
)

f <- function(r) {
  npv <- NPV_profit(Pi, r = r)
  apv_gp <- APV_gross_premiums(
    G = c(100, 100),
    r = r,
    p_tau = c(p40)
  )
  profit_margin(npv, apv_gp) - target_pm
}

root <- uniroot(f, interval = c(0.001, 0.30))$root
round(root, 5)

# Plot objective function
r_grid <- seq(0.001, 0.30, length.out = 400)
y_grid <- sapply(r_grid, f)

plot(
  r_grid, y_grid,
  type = "l",
  main = "Exercise 17-3: Solving for the Risk Discount Rate",
  xlab = "r",
  ylab = "profit margin(r) - target"
)
abline(h = 0, lty = 2)
abline(v = root, lty = 3)








# Exercise 17-7
library(mqriskR)
q <- 0.013 + 0.001 * (0:4)
p <- 1 - q
i <- 0.05

# Given no-reserve profit vector beyond Pr0
profit_no_reserve <- c(450, 230, 100, -50, -220)

# Backward zeroization using the textbook logic
Vz <- numeric(6)
Vz[6] <- 0  # terminal reserve

# t = 4
Vz[5] <- 220 / 1.05

# t = 3
Vz[4] <- (50 + Vz[5] * p[4]) / 1.05

# t = 2
Vz[3] <- (-100 + Vz[4] * p[3]) / 1.05

# t = 1
Vz[2] <- (-230 + Vz[3] * p[2]) / 1.05
Vz[2] <- max(Vz[2], 0)

# t = 0
Vz[1] <- (-450 + Vz[2] * p[1]) / 1.05
Vz[1] <- max(Vz[1], 0)

names(Vz) <- paste0("V", 0:5)
round(Vz, 2)

# Revised profit vector
Pr_new <- numeric(6)
Pr_new[1] <- -300

# Pr1 uses V0 = V1 = 0
Pr_new[2] <- 450

# Pr2 uses reset reserve V1 = 0 and computed V2
Pr_new[3] <- 230 - Vz[3] * p[2]

# By construction, later profits are zero
Pr_new[4:6] <- 0

names(Pr_new) <- paste0("Pr", 0:5)
round(Pr_new, 2)

# Plot zeroized reserves
plot(
  0:5, Vz,
  type = "b", pch = 19,
  main = "Exercise 17-7: Zeroized Reserves",
  xlab = "t",
  ylab = expression(t * V^Z)
)
abline(h = 0, lty = 2)

# Plot revised profit vector
plot(
  0:5, Pr_new,
  type = "b", pch = 19,
  main = "Exercise 17-7: Revised Profit Vector",
  xlab = "t",
  ylab = expression(Pr[t])
)
abline(h = 0, lty = 2)









# Exercise 17-14
library(mqriskR)
q_d <- c(.00390, .00420, .00454, .00491, .00532,
         .00583, .00632, .00687, .00747, .00812)

q_w <- c(.04990, .04989, .04988, .04987, .04986,
         .02991, .02990, .02990, .02989, .02988)

p_tau <- c(.94620, .94591, .94558, .94522, .94482,
           .96426, .96378, .96323, .96264, .96200)

tp_tau <- c(.94620, .89502, .84631, .79995, .75581,
            .72880, .70240, .67657, .65130, .62655)

V <- c(0.00, 0.00, 0.00, 0.00, 1415.00, 3035.00,
       4738.00, 6476.00, 8251.00, 10062.00, 11908.00)

CV <- c(0.00, 0.00, 0.00, 0.00, 5708.00, 7310.40,
        8912.80, 10515.20, 12117.60, 13720.00)

G <- rep(2264.12, 10)
r_exp <- c(0.06, rep(0.03, 9))

Pr <- Pr_vector_disc(
  V = V,
  G = G,
  i = 0.04,
  r = r_exp,
  e = 0,
  q1 = q_d,
  q2 = q_w,
  b1 = 100000,
  b2 = CV,
  p_tau = p_tau,
  pre_contract_expense = 5000
)

Pi <- c(Pr[1], Pr[-1] * c(1, tp_tau[-length(tp_tau)]))

npv <- NPV_profit(Pi, r = 0.10)
apv_gp <- APV_gross_premiums(G = G, r = 0.10, p_tau = p_tau)
pm <- profit_margin(npv, apv_gp)

round(Pr, 5)
round(Pi, 5)
round(c(
  NPV = npv,
  APV_GP = apv_gp,
  profit_margin = pm
), 5)

# Plot profit signature
plot(
  0:10, Pi,
  type = "b", pch = 19,
  main = "Exercise 17-14: Profit Signature",
  xlab = "Policy year",
  ylab = "Pi_t"
)
abline(h = 0, lty = 2)

## Listing 1.1: Converting between $i$, $d$, and $\delta$ (compound interest)

library(mqriskR)

# Convert from i to d and delta
interest_convert(i = 0.05)

# Convert from d to i and delta
interest_convert(d = 0.05/1.05)

# Convert from delta to i and d
interest_convert(delta = log(1.05))

# Nominal annual rate convertible m-thly (e.g., m = 12)
interest_convert(i = 0.06, m = 12)



## Listing 1.2: Verifying annuity-certain identities by computation
library(mqriskR)

i <- 0.05
n <- 10

# Level annuity-immediate and annuity-due (annual)
a_immediate <- annuity_certain(n = n, i = i, due = FALSE)
a_due<- annuity_certain(n = n, i = i, due = TRUE)

# Identity: ä_n = (1+i) a_n
c(a_immediate = a_immediate, a_due = a_due, ratio = a_due / a_immediate)

# Continuous annuity-certain using delta = log(1+i)
a_cont <- annuity_certain(n = n, i = i, cont = TRUE)
a_cont





## Listing 1.3: Non-level annuities via a payment vector
library(mqriskR)

i <- 0.05
n <- 10

# Payments p_k at times t_k: PV = sum(p_k * (1+i)^(-t_k))
pmt <- 1:n      # arithmetic increasing payments 1,2,...,n
t   <- 1:n      # paid at t=1,...,n (annuity-immediate)
pv_cashflows(cf = pmt, t = t, i = i)




## Listing 1.4: Equation of value and yield rate via root finding
library(mqriskR)

# Net cash flows (investor perspective):
# Four payments of X at t=0,1,2,3 and five payments of Y at t=4,...,8
X <- 100; Y <- 120

cf <- c(-X, -X, -X, -X, rep(Y, 5))
t  <- c(0, 1, 2, 3, 4:8)

# Yield rate solves PV(i) = 0
i_hat <- solve_yield(cf, t, interval = c(-0.5, 1))
i_hat

# Plot NPV(i)

grid <- seq(-0.2, 0.6, by = 0.002)
vals <- vapply(grid, function(ii) pv_cashflows(cf, t, ii), numeric(1))

op <- par(mar = c(4, 4, 3, 1) + 0.1, las = 1, bty = "l")
plot(grid, vals, type = "l", lwd = 2,
     xlab = "Interest rate,  i", ylab = "Net present value,  NPV(i)",
     main = "Equation of Value: NPV(i) and Yield Rate",
     xaxs = "i", yaxs = "i")
grid(col = "grey85")
abline(h = 0, lwd = 1.5)

y_hat <- pv_cashflows(cf, t, i_hat)
abline(v = i_hat, lty = 2, lwd = 1.5)
points(i_hat, y_hat, pch = 19, cex = 1.1)
text(i_hat + 0.05, y_hat + 60, labels = sprintf("  i = %.4f", i_hat), cex = 0.9)
par(op)

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.multilife_check_basis <- function(tbl = NULL, model = NULL) {
  if (is.null(tbl) && is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }

  if (!is.null(tbl) && !is.null(model)) {
    stop("Supply only one of tbl or model, not both.", call. = FALSE)
  }

  if (!is.null(tbl)) {
    .validate_life_table(tbl)
  }

  invisible(TRUE)
}

#' @noRd
.multilife_check_nonnegative <- function(x, name) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(x)) || any(x < 0)) {
    stop(
      name,
      " must contain nonnegative finite values.",
      call. = FALSE
    )
  }

  x
}

#' @noRd
.multilife_check_integerish <- function(x, name) {
  x <- .multilife_check_nonnegative(x, name)

  if (any(abs(x - round(x)) > 1e-10)) {
    stop(
      name,
      " must contain nonnegative integer-like values.",
      call. = FALSE
    )
  }

  as.integer(round(x))
}

#' @noRd
.multilife_check_interest <- function(i) {
  i <- as.numeric(i)

  if (length(i) == 0L) {
    stop("i must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(i)) || any(i <= -1)) {
    stop(
      "i must contain finite values greater than -1.",
      call. = FALSE
    )
  }

  i
}

#' @noRd
.multilife_check_sum_controls <- function(k_max, tol) {
  k_max <- as.numeric(k_max)
  tol <- as.numeric(tol)

  if (length(k_max) != 1L ||
      !is.finite(k_max) ||
      k_max <= 0 ||
      abs(k_max - round(k_max)) > 1e-10) {
    stop("k_max must be a positive integer.", call. = FALSE)
  }

  if (length(tol) != 1L || !is.finite(tol) || tol <= 0) {
    stop("tol must be a positive finite number.", call. = FALSE)
  }

  list(
    k_max = as.integer(round(k_max)),
    tol = tol
  )
}

#' @noRd
.multilife_recycle <- function(..., .names = NULL) {
  values <- list(...)

  if (length(values) == 0L) {
    return(values)
  }

  lengths <- vapply(values, length, integer(1))

  if (any(lengths == 0L)) {
    stop("Inputs must have positive length.", call. = FALSE)
  }

  common_length <- max(lengths)

  if (is.null(.names)) {
    .names <- paste0("Argument ", seq_along(values))
  }

  if (length(.names) != length(values)) {
    stop(
      ".names must have the same length as the supplied arguments.",
      call. = FALSE
    )
  }

  for (j in seq_along(values)) {
    if (!lengths[j] %in% c(1L, common_length)) {
      stop(
        .names[j],
        " must have length 1 or the common length ",
        common_length,
        ".",
        call. = FALSE
      )
    }

    values[[j]] <- rep(values[[j]], length.out = common_length)
  }

  values
}

#' @noRd
.multilife_table_survival <- function(tbl, x, t) {
  x <- as.numeric(x)
  t <- as.numeric(t)

  values <- .multilife_recycle(
    x,
    t,
    .names = c("x", "t")
  )

  x <- values[[1]]
  t <- values[[2]]

  out <- numeric(length(x))

  # Survival for zero years is always one, including at the limiting age.
  zero_time <- t == 0
  out[zero_time] <- 1

  positive_time <- !zero_time

  if (any(positive_time)) {
    out[positive_time] <- npx(
      tbl,
      x = x[positive_time],
      n = t[positive_time]
    )
  }

  out[!is.finite(out)] <- 0
  pmin(pmax(out, 0), 1)
}

#' @noRd
.multilife_single_survival <- function(x, t, tbl = NULL, model = NULL, ...) {
  .multilife_check_basis(tbl, model)

  values <- .multilife_recycle(
    x,
    t,
    .names = c("x", "t")
  )

  x <- values[[1]]
  t <- values[[2]]

  if (!is.null(tbl)) {
    return(.multilife_table_survival(tbl, x, t))
  }

  out <- tpx(
    x = x,
    t = t,
    model = model,
    ...
  )

  pmin(pmax(out, 0), 1)
}

#' @noRd
.multilife_joint_table_horizon <- function(tbl, x, y) {
  maximum_age <- max(tbl$x, na.rm = TRUE)

  lx_x <- lx(tbl, x = x)
  lx_y <- lx(tbl, x = y)

  if (is.na(lx_x) || is.na(lx_y)) {
    if (x > maximum_age || y > maximum_age) {
      return(-1L)
    }

    stop(
      "One or both ages are not available in the life table.",
      call. = FALSE
    )
  }

  if (lx_x <= 0 || lx_y <= 0) {
    return(-1L)
  }

  max(
    floor(min(maximum_age - x, maximum_age - y)),
    0L
  )
}

#' @noRd
.multilife_sum_series <- function(fun, start, end, tol,
                                  warn_if_unconverged = FALSE) {
  if (end < start) {
    return(0)
  }

  times <- seq.int(start, end)
  values <- fun(times)

  values[!is.finite(values)] <- 0
  total <- sum(values)

  if (warn_if_unconverged &&
      length(values) > 0L &&
      abs(tail(values, 1L)) > tol) {
    warning(
      "Series may not have converged; consider increasing k_max.",
      call. = FALSE
    )
  }

  total
}

# -------------------------------------------------------------------------
# Joint-life and last-survivor survival probabilities
# -------------------------------------------------------------------------

#' Multi-life survival and failure probabilities
#'
#' Computes joint-life and last-survivor survival and failure probabilities
#' for two independent lives.
#'
#' \code{tpxy()} computes the joint-life survival probability
#' \deqn{
#' {}_tp_{xy}
#' =
#' {}_tp_x\,{}_tp_y.
#' }
#'
#' \code{tqxy()} computes the joint-life failure probability
#' \deqn{
#' {}_tq_{xy}
#' =
#' 1-{}_tp_{xy}.
#' }
#'
#' \code{tpxybar()} computes the last-survivor survival probability
#' \deqn{
#' {}_tp_{\overline{xy}}
#' =
#' {}_tp_x + {}_tp_y - {}_tp_x\,{}_tp_y.
#' }
#'
#' \code{tqxybar()} computes the last-survivor failure probability
#' \deqn{
#' {}_tq_{\overline{xy}}
#' =
#' 1-{}_tp_{\overline{xy}}.
#' }
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param t Nonnegative duration. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the survival model.
#'
#' @return A numeric vector of probabilities.
#'
#' @details
#' All calculations assume the future lifetimes are independent.
#'
#' Joint-life probabilities require both lives to satisfy the survival
#' condition, whereas last-survivor probabilities require at least one life
#' to survive.
#'
#' @examples
#' tpxy(
#'   40, 50,
#'   t = 10,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tqxy(
#'   40, 50,
#'   t = 10,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tpxybar(
#'   40, 50,
#'   t = 10,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tqxybar(
#'   40, 50,
#'   t = 10,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name multilife_survival_probabilities
#' @rdname multilife_survival_probabilities
#' @export
tpxy <- function(x, y, t, tbl = NULL, model = NULL, ...) {
  .multilife_check_basis(tbl, model)

  x <- .multilife_check_nonnegative(x, "x")
  y <- .multilife_check_nonnegative(y, "y")
  t <- .multilife_check_nonnegative(t, "t")

  values <- .multilife_recycle(
    x,
    y,
    t,
    .names = c("x", "y", "t")
  )

  x <- values[[1]]
  y <- values[[2]]
  t <- values[[3]]

  out <- .multilife_single_survival(
    x = x,
    t = t,
    tbl = tbl,
    model = model,
    ...
  ) *
    .multilife_single_survival(
      x = y,
      t = t,
      tbl = tbl,
      model = model,
      ...
    )

  pmin(pmax(out, 0), 1)
}

#' @rdname multilife_survival_probabilities
#' @export
tqxy <- function(x, y, t, tbl = NULL, model = NULL, ...) {
  1 - tpxy(
    x = x,
    y = y,
    t = t,
    tbl = tbl,
    model = model,
    ...
  )
}

#' @rdname multilife_survival_probabilities
#' @export
tpxybar <- function(x, y, t, tbl = NULL, model = NULL, ...) {
  .multilife_check_basis(tbl, model)

  x <- .multilife_check_nonnegative(x, "x")
  y <- .multilife_check_nonnegative(y, "y")
  t <- .multilife_check_nonnegative(t, "t")

  values <- .multilife_recycle(
    x,
    y,
    t,
    .names = c("x", "y", "t")
  )

  x <- values[[1]]
  y <- values[[2]]
  t <- values[[3]]

  px <- .multilife_single_survival(
    x = x,
    t = t,
    tbl = tbl,
    model = model,
    ...
  )

  py <- .multilife_single_survival(
    x = y,
    t = t,
    tbl = tbl,
    model = model,
    ...
  )

  out <- px + py - px * py

  pmin(pmax(out, 0), 1)
}

#' @rdname multilife_survival_probabilities
#' @export
tqxybar <- function(x, y, t, tbl = NULL, model = NULL, ...) {
  1 - tpxybar(
    x = x,
    y = y,
    t = t,
    tbl = tbl,
    model = model,
    ...
  )
}

# -------------------------------------------------------------------------
# Multi-life pure endowments
# -------------------------------------------------------------------------

#' Multi-life pure endowments
#'
#' Computes pure endowment actuarial present values for joint-life and
#' last-survivor statuses for two independent lives.
#'
#' \code{nExy()} computes an \eqn{n}-year joint-life pure endowment, payable
#' at time \eqn{n} if both lives survive.
#'
#' \code{nExybar()} computes an \eqn{n}-year last-survivor pure endowment,
#' payable at time \eqn{n} if at least one life survives.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param n Nonnegative integer term. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the survival model.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @details
#' The joint-life pure endowment is
#' \deqn{
#' {}_nE_{xy}
#' =
#' v^n\,{}_np_{xy}.
#' }
#'
#' The last-survivor pure endowment is
#' \deqn{
#' {}_nE_{\overline{xy}}
#' =
#' v^n\,{}_np_{\overline{xy}}.
#' }
#'
#' Under independence,
#' \deqn{
#' {}_np_{xy}
#' =
#' {}_np_x\,{}_np_y
#' }
#' and
#' \deqn{
#' {}_np_{\overline{xy}}
#' =
#' {}_np_x + {}_np_y - {}_np_x\,{}_np_y.
#' }
#'
#' @examples
#' nExy(
#'   x = 40,
#'   y = 50,
#'   n = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' nExybar(
#'   x = 40,
#'   y = 50,
#'   n = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name multilife_pure_endowments
#' @rdname multilife_pure_endowments
#' @export
nExy <- function(x, y, n, i, tbl = NULL, model = NULL, ...) {
  .multilife_check_basis(tbl, model)

  x <- .multilife_check_nonnegative(x, "x")
  y <- .multilife_check_nonnegative(y, "y")
  n <- .multilife_check_integerish(n, "n")
  i <- .multilife_check_interest(i)

  values <- .multilife_recycle(
    x,
    y,
    n,
    i,
    .names = c("x", "y", "n", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  n <- values[[3]]
  i <- values[[4]]

  discount(i, n) *
    tpxy(
      x = x,
      y = y,
      t = n,
      tbl = tbl,
      model = model,
      ...
    )
}


#' @rdname multilife_pure_endowments
#' @export
nExybar <- function(x, y, n, i, tbl = NULL, model = NULL, ...) {
  .multilife_check_basis(tbl, model)

  x <- .multilife_check_nonnegative(x, "x")
  y <- .multilife_check_nonnegative(y, "y")
  n <- .multilife_check_integerish(n, "n")
  i <- .multilife_check_interest(i)

  values <- .multilife_recycle(
    x,
    y,
    n,
    i,
    .names = c("x", "y", "n", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  n <- values[[3]]
  i <- values[[4]]

  discount(i, n) *
    tpxybar(
      x = x,
      y = y,
      t = n,
      tbl = tbl,
      model = model,
      ...
    )
}

# -------------------------------------------------------------------------
# Joint-life annuities
# -------------------------------------------------------------------------

#' Joint-life annuities
#'
#' Computes temporary and whole-life joint-life annuities-due and
#' annuities-immediate for two independent lives.
#'
#' \code{adotxyn()} computes an \eqn{n}-year temporary joint-life
#' annuity-due.
#'
#' \code{axyn()} computes an \eqn{n}-year temporary joint-life
#' annuity-immediate.
#'
#' \code{adotxy()} computes a whole-life joint-life annuity-due.
#'
#' \code{axy()} computes a whole-life joint-life annuity-immediate.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param n Nonnegative integer term. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the survival model.
#' @param k_max Maximum summation horizon used for non-terminating parametric
#'   survival models.
#' @param tol Positive convergence tolerance for whole-life summations.
#'
#' @return A numeric vector of annuity actuarial present values.
#'
#' @details
#' All calculations assume independence between the two future lifetimes.
#'
#' The temporary annuity-due is
#' \deqn{
#' \ddot{a}_{xy:\overline{n}|}
#' =
#' \sum_{k=0}^{n-1} v^k\,{}_kp_{xy}.
#' }
#'
#' The temporary annuity-immediate is
#' \deqn{
#' a_{xy:\overline{n}|}
#' =
#' \sum_{k=1}^{n} v^k\,{}_kp_{xy}.
#' }
#'
#' For life-table calculations, the whole-life sums terminate at the last
#' duration supported jointly by the two lives. For parametric models, the
#' sums are evaluated until convergence or until \code{k_max} is reached.
#'
#' @examples
#' adotxyn(
#'   x = 40,
#'   y = 50,
#'   n = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' axyn(
#'   x = 40,
#'   y = 50,
#'   n = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' adotxy(
#'   x = 40,
#'   y = 50,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' axy(
#'   x = 40,
#'   y = 50,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name joint_life_annuities
#' @rdname joint_life_annuities
#' @export
adotxyn <- function(x, y, n, i,
                    tbl = NULL, model = NULL, ...) {
  .multilife_check_basis(tbl, model)

  x <- .multilife_check_nonnegative(x, "x")
  y <- .multilife_check_nonnegative(y, "y")
  n <- .multilife_check_integerish(n, "n")
  i <- .multilife_check_interest(i)

  values <- .multilife_recycle(
    x,
    y,
    n,
    i,
    .names = c("x", "y", "n", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  n <- values[[3]]
  i <- values[[4]]

  vapply(
    seq_along(x),
    function(j) {
      if (n[j] == 0L) {
        return(0)
      }

      times <- 0:(n[j] - 1L)

      sum(
        discount(i[j], times) *
          tpxy(
            x = x[j],
            y = y[j],
            t = times,
            tbl = tbl,
            model = model,
            ...
          )
      )
    },
    numeric(1)
  )
}


#' @rdname joint_life_annuities
#' @export
axyn <- function(x, y, n, i,
                 tbl = NULL, model = NULL, ...) {
  adotxyn(
    x = x,
    y = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - 1 +
    nExy(
      x = x,
      y = y,
      n = n,
      i = i,
      tbl = tbl,
      model = model,
      ...
    )
}


#' @rdname joint_life_annuities
#' @export
adotxy <- function(x, y, i,
                   tbl = NULL, model = NULL, ...,
                   k_max = 5000L, tol = 1e-12) {
  .multilife_check_basis(tbl, model)

  x <- .multilife_check_nonnegative(x, "x")
  y <- .multilife_check_nonnegative(y, "y")
  i <- .multilife_check_interest(i)

  controls <- .multilife_check_sum_controls(k_max, tol)
  k_max <- controls$k_max
  tol <- controls$tol

  values <- .multilife_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  vapply(
    seq_along(x),
    function(j) {
      if (!is.null(tbl)) {
        upper <- .multilife_joint_table_horizon(
          tbl = tbl,
          x = x[j],
          y = y[j]
        )

        if (upper < 0L) {
          return(0)
        }

        return(
          .multilife_sum_series(
            fun = function(k) {
              discount(i[j], k) *
                tpxy(
                  x = x[j],
                  y = y[j],
                  t = k,
                  tbl = tbl
                )
            },
            start = 0L,
            end = upper,
            tol = tol
          )
        )
      }

      .multilife_sum_series(
        fun = function(k) {
          discount(i[j], k) *
            tpxy(
              x = x[j],
              y = y[j],
              t = k,
              model = model,
              ...
            )
        },
        start = 0L,
        end = k_max - 1L,
        tol = tol,
        warn_if_unconverged = TRUE
      )
    },
    numeric(1)
  )
}


#' @rdname joint_life_annuities
#' @export
axy <- function(x, y, i,
                tbl = NULL, model = NULL, ...,
                k_max = 5000L, tol = 1e-12) {
  .multilife_check_basis(tbl, model)

  x <- .multilife_check_nonnegative(x, "x")
  y <- .multilife_check_nonnegative(y, "y")
  i <- .multilife_check_interest(i)

  controls <- .multilife_check_sum_controls(k_max, tol)
  k_max <- controls$k_max
  tol <- controls$tol

  values <- .multilife_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  vapply(
    seq_along(x),
    function(j) {
      if (!is.null(tbl)) {
        upper <- .multilife_joint_table_horizon(
          tbl = tbl,
          x = x[j],
          y = y[j]
        )

        if (upper < 1L) {
          return(0)
        }

        return(
          .multilife_sum_series(
            fun = function(k) {
              discount(i[j], k) *
                tpxy(
                  x = x[j],
                  y = y[j],
                  t = k,
                  tbl = tbl
                )
            },
            start = 1L,
            end = upper,
            tol = tol
          )
        )
      }

      .multilife_sum_series(
        fun = function(k) {
          discount(i[j], k) *
            tpxy(
              x = x[j],
              y = y[j],
              t = k,
              model = model,
              ...
            )
        },
        start = 1L,
        end = k_max,
        tol = tol,
        warn_if_unconverged = TRUE
      )
    },
    numeric(1)
  )
}

# -------------------------------------------------------------------------
# Joint-life insurance
# -------------------------------------------------------------------------

#' Joint-life insurance functions
#'
#' Computes temporary and whole-life insurance values for two lives under
#' the joint-life status.
#'
#' \code{Axyn1()} computes an \code{n}-year joint-life term insurance.
#'
#' \code{Axyn()} computes an \code{n}-year joint-life endowment insurance.
#'
#' \code{Axy()} computes joint-life whole-life insurance.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param n Term in years. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the survival model or
#'   life-table functions.
#' @param k_max Maximum number of terms used for an infinite series.
#' @param tol Numerical tolerance used to assess convergence.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @name joint_life_insurance
#' @export
Axyn1 <- function(x, y, n, i,
                  tbl = NULL, model = NULL, ...) {
  .multilife_check_basis(tbl, model)

  i <- .multilife_check_interest(i)

  d <- i / (1 + i)

  1 -
    d * adotxyn(
      x = x,
      y = y,
      n = n,
      i = i,
      tbl = tbl,
      model = model,
      ...
    ) -
    nExy(
      x = x,
      y = y,
      n = n,
      i = i,
      tbl = tbl,
      model = model,
      ...
    )
}

#' @rdname joint_life_insurance
#' @export
Axyn <- function(x, y, n, i,
                 tbl = NULL, model = NULL, ...) {
  Axyn1(
    x = x,
    y = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) + nExy(
    x = x,
    y = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )
}

#' @rdname joint_life_insurance
#' @export
Axy <- function(x, y, i,
                tbl = NULL, model = NULL, ...,
                k_max = 5000L, tol = 1e-12) {
  .multilife_check_basis(tbl, model)

  i <- .multilife_check_interest(i)
  d <- i / (1 + i)

  1 - d * adotxy(
    x = x,
    y = y,
    i = i,
    tbl = tbl,
    model = model,
    ...,
    k_max = k_max,
    tol = tol
  )
}

# -------------------------------------------------------------------------
# Last-survivor annuities
# -------------------------------------------------------------------------

#' Last-survivor annuity functions
#'
#' Computes temporary and whole-life annuities for two lives under the
#' last-survivor status.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param n Term in years. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the survival model or
#'   life-table functions.
#' @param k_max Maximum number of terms used for an infinite series.
#' @param tol Numerical tolerance used to assess convergence.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @name last_survivor_annuities
#' @export
adotxybarn <- function(x, y, n, i,
                       tbl = NULL, model = NULL, ...) {
  adotxn(
    x = x,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) + adotxn(
    x = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - adotxyn(
    x = x,
    y = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )
}

#' @rdname last_survivor_annuities
#' @export
axybarn <- function(x, y, n, i,
                    tbl = NULL, model = NULL, ...) {
  axn(
    x = x,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) + axn(
    x = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - axyn(
    x = x,
    y = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )
}

#' @rdname last_survivor_annuities
#' @export
adotxybar <- function(x, y, i,
                      tbl = NULL, model = NULL, ...,
                      k_max = 5000L, tol = 1e-12) {
  adotx(
    x = x,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) + adotx(
    x = y,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - adotxy(
    x = x,
    y = y,
    i = i,
    tbl = tbl,
    model = model,
    ...,
    k_max = k_max,
    tol = tol
  )
}

#' @rdname last_survivor_annuities
#' @export
axybar <- function(x, y, i,
                   tbl = NULL, model = NULL, ...,
                   k_max = 5000L, tol = 1e-12) {
  ax(
    x = x,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) + ax(
    x = y,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - axy(
    x = x,
    y = y,
    i = i,
    tbl = tbl,
    model = model,
    ...,
    k_max = k_max,
    tol = tol
  )
}

# -------------------------------------------------------------------------
# Last-survivor insurance
# -------------------------------------------------------------------------

#' Last-survivor insurance functions
#'
#' Computes temporary and whole-life insurance values for two lives under
#' the last-survivor status.
#'
#' \code{Axybarn1()} computes temporary insurance payable at the second
#' death within the term.
#'
#' \code{Axybarn()} computes last-survivor endowment insurance.
#'
#' \code{Axybar()} computes last-survivor whole-life insurance.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param n Term in years. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the survival model or
#'   life-table functions.
#' @param k_max Maximum number of terms used for an infinite series.
#' @param tol Numerical tolerance used to assess convergence.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @name last_survivor_insurance
#' @export
Axybarn1 <- function(x, y, n, i,
                     tbl = NULL, model = NULL, ...) {
  Axn1(
    x = x,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) + Axn1(
    x = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - Axyn1(
    x = x,
    y = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )
}

#' @rdname last_survivor_insurance
#' @export
Axybarn <- function(x, y, n, i,
                    tbl = NULL, model = NULL, ...) {
  Axybarn1(
    x = x,
    y = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) + nExybar(
    x = x,
    y = y,
    n = n,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )
}

#' @rdname last_survivor_insurance
#' @export
Axybar <- function(x, y, i,
                   tbl = NULL, model = NULL, ...,
                   k_max = 5000L, tol = 1e-12) {
  Ax(
    x = x,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) + Ax(
    x = y,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - Axy(
    x = x,
    y = y,
    i = i,
    tbl = tbl,
    model = model,
    ...,
    k_max = k_max,
    tol = tol
  )
}

# -------------------------------------------------------------------------
# Reversionary annuities
# -------------------------------------------------------------------------

#' Reversionary annuity functions
#'
#' Computes reversionary whole-life annuities payable to one life after the
#' death of the other life.
#'
#' \code{ax_y()} computes an annuity payable to the second life after the
#' death of the first life.
#'
#' \code{ay_x()} computes an annuity payable to the first life after the
#' death of the second life.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the survival model or
#'   life-table functions.
#' @param k_max Maximum number of terms used for an infinite series.
#' @param tol Numerical tolerance used to assess convergence.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @name reversionary_annuities
#' @export
ax_y <- function(x, y, i,
                 tbl = NULL, model = NULL, ...,
                 k_max = 5000L, tol = 1e-12) {
  ax(
    x = y,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - axy(
    x = x,
    y = y,
    i = i,
    tbl = tbl,
    model = model,
    ...,
    k_max = k_max,
    tol = tol
  )
}

#' @rdname reversionary_annuities
#' @export
ay_x <- function(x, y, i,
                 tbl = NULL, model = NULL, ...,
                 k_max = 5000L, tol = 1e-12) {
  ax(
    x = x,
    i = i,
    tbl = tbl,
    model = model,
    ...
  ) - axy(
    x = x,
    y = y,
    i = i,
    tbl = tbl,
    model = model,
    ...,
    k_max = k_max,
    tol = tol
  )
}

test_that("joint-life survival equals product of single-life survivals", {
  val <- tpxy(40, 50, t = 10, model = "uniform", omega = 100)
  rhs <- tpx(40, t = 10, model = "uniform", omega = 100) *
    tpx(50, t = 10, model = "uniform", omega = 100)

  expect_equal(val, rhs, tolerance = 1e-10)
})

test_that("joint-life and last-survivor survival identity holds", {
  lhs <- tpxybar(40, 50, t = 10, model = "uniform", omega = 100)
  rhs <- tpx(40, t = 10, model = "uniform", omega = 100) +
    tpx(50, t = 10, model = "uniform", omega = 100) -
    tpxy(40, 50, t = 10, model = "uniform", omega = 100)

  expect_equal(lhs, rhs, tolerance = 1e-10)
})

test_that("joint-life and last-survivor failure probabilities complement survival", {
  expect_equal(
    tqxy(40, 50, t = 10, model = "uniform", omega = 100),
    1 - tpxy(40, 50, t = 10, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    tqxybar(40, 50, t = 10, model = "uniform", omega = 100),
    1 - tpxybar(40, 50, t = 10, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("joint-life whole life insurance and annuity satisfy identity", {
  x <- 40
  y <- 50
  i <- 0.05
  d <- i / (1 + i)

  expect_equal(
    Axy(x, y, i = i, model = "uniform", omega = 100),
    1 - d * adotxy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})

test_that("last-survivor whole life insurance identity holds", {
  x <- 40
  y <- 50
  i <- 0.05

  expect_equal(
    Axybar(x, y, i = i, model = "uniform", omega = 100),
    Ax(x, i = i, model = "uniform", omega = 100) +
      Ax(y, i = i, model = "uniform", omega = 100) -
      Axy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})

test_that("last-survivor annuity-due identity holds", {
  x <- 40
  y <- 50
  i <- 0.05

  expect_equal(
    adotxybar(x, y, i = i, model = "uniform", omega = 100),
    adotx(x, i = i, model = "uniform", omega = 100) +
      adotx(y, i = i, model = "uniform", omega = 100) -
      adotxy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})

test_that("joint-life temporary endowment equals term plus pure endowment", {
  x <- 40
  y <- 50
  n <- 10
  i <- 0.05

  expect_equal(
    Axyn(x, y, n = n, i = i, model = "uniform", omega = 100),
    Axyn1(x, y, n = n, i = i, model = "uniform", omega = 100) +
      nExy(x, y, n = n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("last-survivor temporary endowment equals term plus pure endowment", {
  x <- 40
  y <- 50
  n <- 10
  i <- 0.05

  expect_equal(
    Axybarn(x, y, n = n, i = i, model = "uniform", omega = 100),
    Axybarn1(x, y, n = n, i = i, model = "uniform", omega = 100) +
      nExybar(x, y, n = n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("reversionary annuity identities hold", {
  x <- 40
  y <- 50
  i <- 0.05

  expect_equal(
    ax_y(x, y, i = i, model = "uniform", omega = 100),
    ax(y, i = i, model = "uniform", omega = 100) -
      axy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )

  expect_equal(
    ay_x(x, y, i = i, model = "uniform", omega = 100),
    ax(x, i = i, model = "uniform", omega = 100) -
      axy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})

test_that("joint-life survival functions vectorize", {
  out <- tpxy(
    x = c(40, 45),
    y = c(50, 55),
    t = c(5, 10),
    model = "uniform",
    omega = 100
  )

  expect_length(out, 2)
})

test_that("joint-life annuity and insurance functions vectorize", {
  out1 <- Axy(c(40, 45), c(50, 55), i = c(0.05, 0.06), model = "uniform", omega = 100)
  out2 <- adotxy(c(40, 45), c(50, 55), i = c(0.05, 0.06), model = "uniform", omega = 100)

  expect_length(out1, 2)
  expect_length(out2, 2)
})

test_that("multilife functions reject invalid inputs", {
  expect_error(
    tpxy(40, 50, t = -1, model = "uniform", omega = 100),
    "nonnegative"
  )

  expect_error(
    nExy(
      40, 50,
      n = -2,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "n must contain nonnegative finite values"
  )

  expect_error(
    Axy(40, 50, i = -1.2, model = "uniform", omega = 100),
    "greater than -1"
  )
})



testthat::test_that("temporary joint-life annuity-immediate satisfies the correct identity", {
  x <- 40
  y <- 50
  n <- 10
  i <- 0.05

  expected <- adotxyn(
    x = x,
    y = y,
    n = n,
    i = i,
    model = "uniform",
    omega = 100
  ) - 1 + nExy(
    x = x,
    y = y,
    n = n,
    i = i,
    model = "uniform",
    omega = 100
  )

  testthat::expect_equal(
    axyn(
      x = x,
      y = y,
      n = n,
      i = i,
      model = "uniform",
      omega = 100
    ),
    expected,
    tolerance = 1e-12
  )
})


testthat::test_that("zero-term multi-life functions have correct values", {
  testthat::expect_equal(
    adotxyn(
      40, 50,
      n = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    0
  )

  testthat::expect_equal(
    axyn(
      40, 50,
      n = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    0
  )

  testthat::expect_equal(
    nExy(
      40, 50,
      n = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    1
  )

  testthat::expect_equal(
    nExybar(
      40, 50,
      n = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    1
  )

  testthat::expect_equal(
    Axyn1(
      40, 50,
      n = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    0
  )

  testthat::expect_equal(
    Axyn(
      40, 50,
      n = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    1
  )
})


testthat::test_that("multi-life functions vectorize over all main arguments", {
  out <- nExy(
    x = c(40, 45),
    y = c(50, 55),
    n = c(10, 5),
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))

  annuity <- adotxyn(
    x = c(40, 45),
    y = 50,
    n = c(10, 5),
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(annuity, 2)
  testthat::expect_true(all(is.finite(annuity)))
})


testthat::test_that("multi-life functions reject incompatible argument lengths", {
  testthat::expect_error(
    nExy(
      x = c(40, 45),
      y = c(50, 55, 60),
      n = 10,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "length 1 or the common length"
  )

  testthat::expect_error(
    adotxyn(
      x = c(40, 45),
      y = 50,
      n = c(5, 10, 15),
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "length 1 or the common length"
  )
})


testthat::test_that("multi-life functions reject conflicting mortality bases", {
  tbl <- life_table(
    x = 0:4,
    lx = c(100000, 80000, 50000, 20000, 0)
  )

  testthat::expect_error(
    tpxy(
      x = 0,
      y = 1,
      t = 1,
      tbl = tbl,
      model = "uniform",
      omega = 100
    ),
    "only one of tbl or model"
  )

  testthat::expect_error(
    nExy(
      x = 0,
      y = 1,
      n = 1,
      i = 0.05
    ),
    "either tbl or model"
  )
})


testthat::test_that("multi-life functions work with finite life tables", {
  tbl <- life_table(
    x = 0:4,
    lx = c(100000, 80000, 50000, 20000, 0)
  )

  testthat::expect_equal(
    tpxy(
      x = 0,
      y = 1,
      t = 1,
      tbl = tbl
    ),
    0.8 * 0.625,
    tolerance = 1e-12
  )

  values <- c(
    adotxy(0, 1, i = 0.05, tbl = tbl),
    axy(0, 1, i = 0.05, tbl = tbl),
    Axy(0, 1, i = 0.05, tbl = tbl),
    adotxybar(0, 1, i = 0.05, tbl = tbl),
    axybar(0, 1, i = 0.05, tbl = tbl),
    Axybar(0, 1, i = 0.05, tbl = tbl)
  )

  testthat::expect_true(all(is.finite(values)))
})


testthat::test_that("joint and last-survivor probabilities satisfy identities", {
  px_joint <- tpxy(
    40, 50,
    t = 10,
    model = "uniform",
    omega = 100
  )

  px_last <- tpxybar(
    40, 50,
    t = 10,
    model = "uniform",
    omega = 100
  )

  px <- tpx(
    x = 40,
    t = 10,
    model = "uniform",
    omega = 100
  )

  py <- tpx(
    x = 50,
    t = 10,
    model = "uniform",
    omega = 100
  )

  testthat::expect_equal(
    px_joint,
    px * py,
    tolerance = 1e-12
  )

  testthat::expect_equal(
    px_last,
    px + py - px * py,
    tolerance = 1e-12
  )

  testthat::expect_equal(
    tqxy(40, 50, 10, model = "uniform", omega = 100),
    1 - px_joint,
    tolerance = 1e-12
  )

  testthat::expect_equal(
    tqxybar(40, 50, 10, model = "uniform", omega = 100),
    1 - px_last,
    tolerance = 1e-12
  )
})


testthat::test_that("multi-life insurance and annuity identities hold", {
  x <- 40
  y <- 50
  n <- 10
  i <- 0.05
  d <- i / (1 + i)

  testthat::expect_equal(
    Axyn1(
      x, y, n, i,
      model = "uniform",
      omega = 100
    ),
    1 -
      d * adotxyn(
        x, y, n, i,
        model = "uniform",
        omega = 100
      ) -
      nExy(
        x, y, n, i,
        model = "uniform",
        omega = 100
      ),
    tolerance = 1e-10
  )

  testthat::expect_equal(
    Axy(
      x, y, i,
      model = "uniform",
      omega = 100
    ),
    1 -
      d * adotxy(
        x, y, i,
        model = "uniform",
        omega = 100
      ),
    tolerance = 1e-10
  )

  testthat::expect_equal(
    Axybar(
      x, y, i,
      model = "uniform",
      omega = 100
    ),
    Ax(
      x, i,
      model = "uniform",
      omega = 100
    ) +
      Ax(
        y, i,
        model = "uniform",
        omega = 100
      ) -
      Axy(
        x, y, i,
        model = "uniform",
        omega = 100
      ),
    tolerance = 1e-10
  )
})

# -------------------------------------------------------------------------
# Internal validation and recycling helpers
# -------------------------------------------------------------------------

#' @noRd
.multilife_ext_check_basis <- function(tbl = NULL,
                                       model = NULL,
                                       require_continuous = TRUE) {
  if (is.null(tbl) && is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }

  if (!is.null(tbl) && !is.null(model)) {
    stop("Supply only one of tbl or model, not both.", call. = FALSE)
  }

  if (!is.null(tbl)) {
    .validate_life_table(tbl)

    if (isTRUE(require_continuous)) {
      stop(
        paste0(
          "Continuous multi-life calculations require a parametric model. ",
          "An annual life table does not determine within-year survival ",
          "or forces of mortality without an interpolation assumption."
        ),
        call. = FALSE
      )
    }
  }

  if (!is.null(model)) {
    if (length(model) != 1L ||
        !is.character(model) ||
        is.na(model) ||
        !nzchar(model)) {
      stop("model must be a single nonempty character string.",
           call. = FALSE)
    }
  }

  invisible(TRUE)
}


#' @noRd
.multilife_ext_check_nonnegative <- function(x, name) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(x)) || any(x < 0)) {
    stop(
      name,
      " must contain nonnegative finite values.",
      call. = FALSE
    )
  }

  x
}


#' @noRd
.multilife_ext_check_interest <- function(i) {
  i <- as.numeric(i)

  if (length(i) == 0L) {
    stop("i must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(i)) || any(i <= -1)) {
    stop(
      "i must contain finite values greater than -1.",
      call. = FALSE
    )
  }

  i
}


#' @noRd
.multilife_ext_recycle <- function(..., .names = NULL) {
  values <- list(...)

  if (length(values) == 0L) {
    return(values)
  }

  lengths <- vapply(values, length, integer(1))

  if (any(lengths == 0L)) {
    stop("Inputs must have positive length.", call. = FALSE)
  }

  common_length <- max(lengths)

  if (is.null(.names)) {
    .names <- paste0("Argument ", seq_along(values))
  }

  if (length(.names) != length(values)) {
    stop(
      ".names must have the same length as the supplied arguments.",
      call. = FALSE
    )
  }

  for (j in seq_along(values)) {
    if (!lengths[j] %in% c(1L, common_length)) {
      stop(
        .names[j],
        " must have length 1 or the common length ",
        common_length,
        ".",
        call. = FALSE
      )
    }

    values[[j]] <- rep(values[[j]], length.out = common_length)
  }

  values
}


#' @noRd
.multilife_ext_upper <- function(x,
                                 y,
                                 n = NULL,
                                 model,
                                 ...) {
  if (tolower(model) == "uniform") {
    dots <- list(...)

    if (is.null(dots$omega) ||
        length(dots$omega) != 1L ||
        !is.finite(dots$omega)) {
      stop(
        "For model = 'uniform', omega must be supplied as a finite scalar.",
        call. = FALSE
      )
    }

    upper <- min(dots$omega - x, dots$omega - y)

    if (!is.null(n)) {
      upper <- min(upper, n)
    }

    return(max(upper, 0))
  }

  if (!is.null(n)) {
    return(n)
  }

  Inf
}


#' @noRd
.multilife_ext_integrate <- function(f, upper) {
  if (!is.finite(upper) && upper != Inf) {
    stop("The integration limit must be finite or Inf.", call. = FALSE)
  }

  if (upper <= 0) {
    return(0)
  }

  value <- stats::integrate(
    f,
    lower = 0,
    upper = upper,
    rel.tol = 1e-9,
    subdivisions = 1000L,
    stop.on.error = TRUE
  )$value

  if (!is.finite(value)) {
    stop("The continuous multi-life integral did not return a finite value.",
         call. = FALSE)
  }

  value
}


#' @noRd
.multilife_ext_joint_survival <- function(t, x, y, model, ...) {
  tpxy(
    x = x,
    y = y,
    t = t,
    model = model,
    ...
  )
}


#' @noRd
.multilife_ext_single_survival <- function(t, x, model, ...) {
  tpx(
    t = t,
    x = x,
    model = model,
    ...
  )
}


#' @noRd
.multilife_ext_hazard <- function(age, model, ...) {
  hazard0(
    age,
    model = model,
    ...
  )
}


# -------------------------------------------------------------------------
# Contingent probabilities
# -------------------------------------------------------------------------

#' Contingent multi-life probabilities
#'
#' Computes probabilities associated with the order of death of two
#' independent lives over a specified term.
#'
#' \code{tqxy1()} computes the probability that the first life dies before
#' the second life within \eqn{n} years.
#'
#' \code{tqyx1()} computes the probability that the second life dies before
#' the first life within \eqn{n} years.
#'
#' \code{tqxy2()} computes the probability that the first life dies after
#' the second life but within \eqn{n} years.
#'
#' \code{tqyx2()} computes the probability that the second life dies after
#' the first life but within \eqn{n} years.
#'
#' These functions require a parametric survival model because an annual
#' life table does not uniquely determine the within-year order of death
#' without an additional interpolation assumption.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param n Term in years. May be scalar or vector.
#' @param tbl Optional life table object retained for backward compatibility.
#'   Continuous calculations currently require \code{model}.
#' @param model Parametric survival model.
#' @param ... Additional parameters passed to the survival and hazard
#'   functions.
#'
#' @return A numeric vector of contingent probabilities.
#'
#' @details
#' All calculations assume the two future lifetimes are independent.
#'
#' The probabilities are obtained by integrating the joint survival
#' function together with the appropriate force of mortality.
#'
#' The four functions partition the probability that one of the two lives
#' dies within the specified term according to the order of death.
#'
#' @examples
#' tqxy1(
#'   40, 50,
#'   n = 10,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tqyx1(
#'   40, 50,
#'   n = 10,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tqxy2(
#'   40, 50,
#'   n = 10,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tqyx2(
#'   40, 50,
#'   n = 10,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name multilife_contingent_probabilities
#' @rdname multilife_contingent_probabilities
#' @export
tqxy1 <- function(x, y, n, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  n <- .multilife_ext_check_nonnegative(n, "n")

  values <- .multilife_ext_recycle(
    x,
    y,
    n,
    .names = c("x", "y", "n")
  )

  x <- values[[1]]
  y <- values[[2]]
  n <- values[[3]]

  vapply(
    seq_along(x),
    function(j) {
      upper <- .multilife_ext_upper(
        x = x[j],
        y = y[j],
        n = n[j],
        model = model,
        ...
      )

      .multilife_ext_integrate(
        function(t) {
          value <- .multilife_ext_joint_survival(
            t = t,
            x = x[j],
            y = y[j],
            model = model,
            ...
          ) *
            .multilife_ext_hazard(
              age = x[j] + t,
              model = model,
              ...
            )

          value[!is.finite(value)] <- 0
          value
        },
        upper = upper
      )
    },
    numeric(1)
  )
}

#' @rdname multilife_contingent_probabilities
#' @export
tqyx1 <- function(x, y, n, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  n <- .multilife_ext_check_nonnegative(n, "n")

  values <- .multilife_ext_recycle(
    x,
    y,
    n,
    .names = c("x", "y", "n")
  )

  x <- values[[1]]
  y <- values[[2]]
  n <- values[[3]]

  vapply(
    seq_along(x),
    function(j) {
      upper <- .multilife_ext_upper(
        x = x[j],
        y = y[j],
        n = n[j],
        model = model,
        ...
      )

      .multilife_ext_integrate(
        function(t) {
          value <- .multilife_ext_joint_survival(
            t = t,
            x = x[j],
            y = y[j],
            model = model,
            ...
          ) *
            .multilife_ext_hazard(
              age = y[j] + t,
              model = model,
              ...
            )

          value[!is.finite(value)] <- 0
          value
        },
        upper = upper
      )
    },
    numeric(1)
  )
}

#' @rdname multilife_contingent_probabilities
#' @export
tqxy2 <- function(x, y, n, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  n <- .multilife_ext_check_nonnegative(n, "n")

  values <- .multilife_ext_recycle(
    x,
    y,
    n,
    .names = c("x", "y", "n")
  )

  x <- values[[1]]
  y <- values[[2]]
  n <- values[[3]]

  qx <- 1 - .multilife_ext_single_survival(
    t = n,
    x = x,
    model = model,
    ...
  )

  out <- qx - tqxy1(
    x = x,
    y = y,
    n = n,
    model = model,
    ...
  )

  pmax(out, 0)
}

#' @rdname multilife_contingent_probabilities
#' @export
tqyx2 <- function(x, y, n, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  n <- .multilife_ext_check_nonnegative(n, "n")

  values <- .multilife_ext_recycle(
    x,
    y,
    n,
    .names = c("x", "y", "n")
  )

  x <- values[[1]]
  y <- values[[2]]
  n <- values[[3]]

  qy <- 1 - .multilife_ext_single_survival(
    t = n,
    x = y,
    model = model,
    ...
  )

  out <- qy - tqyx1(
    x = x,
    y = y,
    n = n,
    model = model,
    ...
  )

  pmax(out, 0)
}


# -------------------------------------------------------------------------
# Continuous multi-life annuities
# -------------------------------------------------------------------------

#' Continuous multi-life annuities
#'
#' Computes continuous joint-life, last-survivor, and reversionary whole-life
#' annuities for two independent lives.
#'
#' \code{abarxy()} computes the joint-life annuity.
#'
#' \code{abarxybar()} computes the last-survivor annuity.
#'
#' \code{abarx_y()} computes the reversionary annuity payable to the second
#' life after the death of the first.
#'
#' \code{abary_x()} computes the reversionary annuity payable to the first
#' life after the death of the second.
#'
#' These functions require a parametric survival model.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object retained for backward compatibility.
#'   Continuous calculations currently require \code{model}.
#' @param model Parametric survival model.
#' @param ... Additional parameters passed to the survival functions.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @details
#' Under independence, \code{abarxy()} represents the continuous joint-life
#' annuity, payable while both lives survive.
#'
#' The last-survivor annuity satisfies
#' \deqn{\bar{a}_{\overline{xy}} =
#' \bar{a}_x + \bar{a}_y - \bar{a}_{xy}.}
#'
#' The reversionary annuities are obtained as the difference between the
#' corresponding single-life annuity and the joint-life annuity.
#'
#' @examples
#' abarxy(
#'   x = 40,
#'   y = 50,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' abarxybar(
#'   x = 40,
#'   y = 50,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name continuous_multilife_annuities
#' @rdname continuous_multilife_annuities
#' @export
abarxy <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  delta <- log1p(i)

  vapply(
    seq_along(x),
    function(j) {
      upper <- .multilife_ext_upper(
        x = x[j],
        y = y[j],
        model = model,
        ...
      )

      .multilife_ext_integrate(
        function(t) {
          value <- exp(-delta[j] * t) *
            .multilife_ext_joint_survival(
              t = t,
              x = x[j],
              y = y[j],
              model = model,
              ...
            )

          value[!is.finite(value)] <- 0
          value
        },
        upper = upper
      )
    },
    numeric(1)
  )
}


#' @rdname continuous_multilife_annuities
#' @export
abarxybar <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  abarx(
    x = x,
    i = i,
    model = model,
    ...
  ) +
    abarx(
      x = y,
      i = i,
      model = model,
      ...
    ) -
    abarxy(
      x = x,
      y = y,
      i = i,
      model = model,
      ...
    )
}


#' @rdname continuous_multilife_annuities
#' @export
abarx_y <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  abarx(
    x = y,
    i = i,
    model = model,
    ...
  ) -
    abarxy(
      x = x,
      y = y,
      i = i,
      model = model,
      ...
    )
}


#' @rdname continuous_multilife_annuities
#' @export
abary_x <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  abarx(
    x = x,
    i = i,
    model = model,
    ...
  ) -
    abarxy(
      x = x,
      y = y,
      i = i,
      model = model,
      ...
    )
}

# -------------------------------------------------------------------------
# Continuous multi-life insurance
# -------------------------------------------------------------------------

#' Continuous multi-life insurance
#'
#' Computes continuous joint-life, last-survivor, and contingent whole-life
#' insurance values for two independent lives.
#'
#' \code{Abarxy()} computes joint-life insurance payable at the first death.
#'
#' \code{Abarxybar()} computes last-survivor insurance payable at the second
#' death.
#'
#' \code{Abarxy1()} and \code{Abaryx1()} compute contingent insurance payable
#' when the specified life dies first.
#'
#' \code{Abarxy2()} and \code{Abaryx2()} compute contingent insurance payable
#' when the specified life dies second.
#'
#' These functions require a parametric survival model.
#'
#' @param x Age of the first life. May be scalar or vector.
#' @param y Age of the second life. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object retained for backward compatibility.
#'   Continuous calculations currently require \code{model}.
#' @param model Parametric survival model.
#' @param ... Additional parameters passed to the survival and hazard
#'   functions.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @details
#' Under independence, \code{Abarxy()} represents insurance payable at the
#' first death and \code{Abarxybar()} represents insurance payable at the
#' second death.
#'
#' The contingent functions distinguish both the life whose death triggers
#' payment and whether that death occurs first or second.
#'
#' @examples
#' Abarxy(
#'   x = 40,
#'   y = 50,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' Abarxy1(
#'   x = 40,
#'   y = 50,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name continuous_multilife_insurance
#' @rdname continuous_multilife_insurance
#' @export
Abarxy <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  1 - log1p(i) * abarxy(
    x = x,
    y = y,
    i = i,
    model = model,
    ...
  )
}


#' @rdname continuous_multilife_insurance
#' @export
Abarxybar <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  Abarx(
    x = x,
    i = i,
    model = model,
    ...
  ) +
    Abarx(
      x = y,
      i = i,
      model = model,
      ...
    ) -
    Abarxy(
      x = x,
      y = y,
      i = i,
      model = model,
      ...
    )
}


#' @rdname continuous_multilife_insurance
#' @export
Abarxy1 <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  delta <- log1p(i)

  vapply(
    seq_along(x),
    function(j) {
      upper <- .multilife_ext_upper(
        x = x[j],
        y = y[j],
        model = model,
        ...
      )

      .multilife_ext_integrate(
        function(t) {
          value <- exp(-delta[j] * t) *
            .multilife_ext_joint_survival(
              t = t,
              x = x[j],
              y = y[j],
              model = model,
              ...
            ) *
            .multilife_ext_hazard(
              age = x[j] + t,
              model = model,
              ...
            )

          value[!is.finite(value)] <- 0
          value
        },
        upper = upper
      )
    },
    numeric(1)
  )
}


#' @rdname continuous_multilife_insurance
#' @export
Abaryx1 <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  delta <- log1p(i)

  vapply(
    seq_along(x),
    function(j) {
      upper <- .multilife_ext_upper(
        x = x[j],
        y = y[j],
        model = model,
        ...
      )

      .multilife_ext_integrate(
        function(t) {
          value <- exp(-delta[j] * t) *
            .multilife_ext_joint_survival(
              t = t,
              x = x[j],
              y = y[j],
              model = model,
              ...
            ) *
            .multilife_ext_hazard(
              age = y[j] + t,
              model = model,
              ...
            )

          value[!is.finite(value)] <- 0
          value
        },
        upper = upper
      )
    },
    numeric(1)
  )
}


#' @rdname continuous_multilife_insurance
#' @export
Abarxy2 <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  out <- Abarx(
    x = x,
    i = i,
    model = model,
    ...
  ) -
    Abarxy1(
      x = x,
      y = y,
      i = i,
      model = model,
      ...
    )

  pmax(out, 0)
}


#' @rdname continuous_multilife_insurance
#' @export
Abaryx2 <- function(x, y, i, tbl = NULL, model = NULL, ...) {
  .multilife_ext_check_basis(
    tbl = tbl,
    model = model,
    require_continuous = TRUE
  )

  x <- .multilife_ext_check_nonnegative(x, "x")
  y <- .multilife_ext_check_nonnegative(y, "y")
  i <- .multilife_ext_check_interest(i)

  values <- .multilife_ext_recycle(
    x,
    y,
    i,
    .names = c("x", "y", "i")
  )

  x <- values[[1]]
  y <- values[[2]]
  i <- values[[3]]

  out <- Abarx(
    x = y,
    i = i,
    model = model,
    ...
  ) -
    Abaryx1(
      x = x,
      y = y,
      i = i,
      model = model,
      ...
    )

  pmax(out, 0)
}

test_that("contingent probabilities add to joint-life failure probability", {
  val <- tqxy1(40, 50, n = 10, model = "uniform", omega = 100) +
    tqyx1(40, 50, n = 10, model = "uniform", omega = 100)

  expect_equal(
    val,
    tqxy(40, 50, t = 10, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})

test_that("second-failure contingent probabilities add to last-survivor failure probability", {
  val <- tqxy2(40, 50, n = 10, model = "uniform", omega = 100) +
    tqyx2(40, 50, n = 10, model = "uniform", omega = 100)

  expect_equal(
    val,
    tqxybar(40, 50, t = 10, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})

test_that("continuous joint-life insurance-annuity identity holds", {
  x <- 40
  y <- 50
  i <- 0.05
  delta <- log(1 + i)

  expect_equal(
    Abarxy(x, y, i = i, model = "uniform", omega = 100),
    1 - delta * abarxy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})

test_that("continuous last-survivor identity holds", {
  x <- 40
  y <- 50
  i <- 0.05

  expect_equal(
    Abarxybar(x, y, i = i, model = "uniform", omega = 100),
    Abarx(x, i = i, model = "uniform", omega = 100) +
      Abarx(y, i = i, model = "uniform", omega = 100) -
      Abarxy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})

test_that("continuous reversionary annuity identities hold", {
  x <- 40
  y <- 50
  i <- 0.05

  expect_equal(
    abarx_y(x, y, i = i, model = "uniform", omega = 100),
    abarx(y, i = i, model = "uniform", omega = 100) -
      abarxy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )

  expect_equal(
    abary_x(x, y, i = i, model = "uniform", omega = 100),
    abarx(x, i = i, model = "uniform", omega = 100) -
      abarxy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-8
  )
})


test_that("continuous contingent insurance identities hold", {
  x <- 40
  y <- 50
  i <- 0.05

  expect_equal(
    Abarxy1(x, y, i = i, model = "uniform", omega = 100) +
      Abaryx1(x, y, i = i, model = "uniform", omega = 100),
    Abarxy(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-7
  )

  expect_equal(
    Abarxy2(x, y, i = i, model = "uniform", omega = 100) +
      Abaryx2(x, y, i = i, model = "uniform", omega = 100),
    Abarxybar(x, y, i = i, model = "uniform", omega = 100),
    tolerance = 1e-6
  )
})

test_that("multilife extension functions vectorize", {
  out1 <- abarxy(c(40, 45), c(50, 55), i = c(0.05, 0.06), model = "uniform", omega = 100)
  out2 <- Abarxy(c(40, 45), c(50, 55), i = c(0.05, 0.06), model = "uniform", omega = 100)
  out3 <- tqxy1(c(40, 45), c(50, 55), n = c(5, 10), model = "uniform", omega = 100)

  expect_length(out1, 2)
  expect_length(out2, 2)
  expect_length(out3, 2)
})

test_that("multilife extension functions reject invalid inputs", {
  expect_error(
    tqxy1(40, 50, n = -1, model = "uniform", omega = 100),
    "nonnegative"
  )

  expect_error(
    abarxy(40, 50, i = -1.2, model = "uniform", omega = 100),
    "greater than -1"
  )
})


testthat::test_that(
  "contingent probabilities add to joint-life failure probability",
  {
    x <- 40
    y <- 50
    n <- 10

    first_x <- tqxy1(
      x, y, n = n,
      model = "uniform",
      omega = 100
    )

    first_y <- tqyx1(
      x, y, n = n,
      model = "uniform",
      omega = 100
    )

    testthat::expect_equal(
      first_x + first_y,
      tqxy(
        x, y, t = n,
        model = "uniform",
        omega = 100
      ),
      tolerance = 1e-8
    )
  }
)


testthat::test_that(
  "second-death probabilities satisfy their defining identities",
  {
    x <- 40
    y <- 50
    n <- 10

    qx <- 1 - tpx(
      t = n,
      x = x,
      model = "uniform",
      omega = 100
    )

    qy <- 1 - tpx(
      t = n,
      x = y,
      model = "uniform",
      omega = 100
    )

    testthat::expect_equal(
      tqxy1(
        x, y, n,
        model = "uniform",
        omega = 100
      ) +
        tqxy2(
          x, y, n,
          model = "uniform",
          omega = 100
        ),
      qx,
      tolerance = 1e-8
    )

    testthat::expect_equal(
      tqyx1(
        x, y, n,
        model = "uniform",
        omega = 100
      ) +
        tqyx2(
          x, y, n,
          model = "uniform",
          omega = 100
        ),
      qy,
      tolerance = 1e-8
    )
  }
)


testthat::test_that(
  "continuous joint-life insurance-annuity identity holds",
  {
    x <- 40
    y <- 50
    i <- 0.05
    delta <- log1p(i)

    testthat::expect_equal(
      Abarxy(
        x, y, i,
        model = "uniform",
        omega = 100
      ),
      1 - delta * abarxy(
        x, y, i,
        model = "uniform",
        omega = 100
      ),
      tolerance = 1e-8
    )
  }
)


testthat::test_that(
  "continuous contingent insurances add to joint-life insurance",
  {
    x <- 40
    y <- 50
    i <- 0.05

    testthat::expect_equal(
      Abarxy1(
        x, y, i,
        model = "uniform",
        omega = 100
      ) +
        Abaryx1(
          x, y, i,
          model = "uniform",
          omega = 100
        ),
      Abarxy(
        x, y, i,
        model = "uniform",
        omega = 100
      ),
      tolerance = 1e-8
    )
  }
)


testthat::test_that(
  "continuous multi-life functions vectorize consistently",
  {
    out <- abarxy(
      x = c(40, 45),
      y = c(50, 55),
      i = c(0.04, 0.05),
      model = "uniform",
      omega = 100
    )

    testthat::expect_length(out, 2)
    testthat::expect_true(all(is.finite(out)))
    testthat::expect_true(all(out >= 0))
  }
)


testthat::test_that(
  "continuous multi-life functions handle limiting ages",
  {
    testthat::expect_equal(
      abarxy(
        x = 100,
        y = 50,
        i = 0.05,
        model = "uniform",
        omega = 100
      ),
      0
    )

    testthat::expect_equal(
      Abarxy1(
        x = 100,
        y = 50,
        i = 0.05,
        model = "uniform",
        omega = 100
      ),
      0
    )
  }
)


testthat::test_that(
  "continuous multi-life functions reject incompatible vector lengths",
  {
    testthat::expect_error(
      abarxy(
        x = c(40, 45),
        y = c(50, 55, 60),
        i = 0.05,
        model = "uniform",
        omega = 100
      ),
      "length 1 or the common length"
    )
  }
)


testthat::test_that(
  "continuous multi-life functions reject conflicting mortality bases",
  {
    tbl <- life_table(
      x = 0:3,
      lx = c(100000, 80000, 50000, 0)
    )

    testthat::expect_error(
      abarxy(
        x = 0,
        y = 1,
        i = 0.05,
        tbl = tbl,
        model = "uniform",
        omega = 100
      ),
      "only one of tbl or model"
    )
  }
)


testthat::test_that(
  "continuous multi-life functions explain life-table limitation",
  {
    tbl <- life_table(
      x = 0:3,
      lx = c(100000, 80000, 50000, 0)
    )

    testthat::expect_error(
      abarxy(
        x = 0,
        y = 1,
        i = 0.05,
        tbl = tbl
      ),
      "require a parametric model"
    )

    testthat::expect_error(
      tqxy1(
        x = 0,
        y = 1,
        n = 1,
        tbl = tbl
      ),
      "require a parametric model"
    )
  }
)

#' Discrete insurance models
#'
#' Discrete contingent payment and insurance functions.
#'
#' @name insurance_discrete
#' @keywords internal
NULL

# ------------------------------------------------------------
# Internal helpers
# ------------------------------------------------------------

.get_npx_discrete <- function(x, n, tbl = NULL, model = NULL, ...) {
  if (!is.null(tbl)) return(npx(tbl, x = x, n = n))
  if (is.null(model)) stop("Supply either tbl or model.", call. = FALSE)
  tpx(n, x = x, model = model, ...)
}

.get_nqx_discrete <- function(x, n, tbl = NULL, model = NULL, ...) {
  if (!is.null(tbl)) return(nqx(tbl, x = x, n = n))
  if (is.null(model)) stop("Supply either tbl or model.", call. = FALSE)
  tqx(n, x = x, model = model, ...)
}

.recycle_x_n <- function(x, n) {
  x <- as.numeric(x)
  n <- as.numeric(n)

  if (length(x) == 0L || length(n) == 0L) {
    stop("x and n must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(x))) stop("x must be finite.", call. = FALSE)
  if (any(!is.finite(n))) stop("n must be finite.", call. = FALSE)
  if (any(x < 0)) stop("x must be nonnegative.", call. = FALSE)
  if (any(n < 0)) stop("n must be nonnegative.", call. = FALSE)
  if (any(abs(n - round(n)) > 1e-10)) {
    stop("n must contain integers only.", call. = FALSE)
  }

  len <- max(length(x), length(n))

  if (!length(x) %in% c(1L, len) || !length(n) %in% c(1L, len)) {
    stop("x and n must have compatible lengths.", call. = FALSE)
  }

  list(
    x = if (length(x) == 1L) rep(x, len) else x,
    n = if (length(n) == 1L) rep(n, len) else n
  )
}

.check_i_discrete <- function(i) {
  i <- as.numeric(i)

  if (length(i) != 1L || !is.finite(i) || i <= -1) {
    stop("i must be a single value greater than -1.", call. = FALSE)
  }

  i
}

.max_k_discrete <- function(x, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000) {
  for (k in 0:k_max) {
    surv <- .get_npx_discrete(x, k, tbl = tbl, model = model, ...)

    if (is.na(surv)) return(max(k - 1L, 0L))
    if (surv <= tol) return(k)
  }

  k_max
}

.kq_vec <- function(x, k, tbl = NULL, model = NULL, ...) {
  p_k <- .get_npx_discrete(x, k, tbl = tbl, model = model, ...)
  p_k1 <- .get_npx_discrete(x, k + 1, tbl = tbl, model = model, ...)

  if (is.na(p_k)) p_k <- 0
  if (is.na(p_k1)) p_k1 <- 0

  p_k - p_k1
}

# ------------------------------------------------------------
# First moments
# ------------------------------------------------------------

#' Whole life insurance APV
#'
#' Computes \eqn{A_x = \sum_{k=0}^\infty v^{k+1} {}_{k\mid}q_x}.
#'
#' @param x Age.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model name.
#' @param ... Additional arguments passed to survival-model functions.
#' @param tol Numerical tolerance for truncating infinite sums.
#' @param k_max Maximum number of terms in the sum.
#'
#' @return Numeric vector of APVs.
#' @export
Ax <- function(x, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000) {
  x <- as.numeric(x)
  i <- .check_i_discrete(i)

  if (length(x) == 0L || any(!is.finite(x))) stop("x must be finite.", call. = FALSE)
  if (any(x < 0)) stop("x must be nonnegative.", call. = FALSE)

  v <- 1 / (1 + i)

  sapply(x, function(xx) {
    kk_max <- .max_k_discrete(xx, tbl = tbl, model = model, ..., tol = tol, k_max = k_max)
    k <- 0:kk_max
    probs <- vapply(k, function(kk) .kq_vec(xx, kk, tbl = tbl, model = model, ...), numeric(1))
    sum(v^(k + 1) * probs, na.rm = TRUE)
  })
}

#' Term insurance APV
#'
#' Computes \eqn{A_{x:\overline{n}|}^{1} = \sum_{k=0}^{n-1} v^{k+1} {}_{k\mid}q_x}.
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model name.
#' @param ... Additional arguments passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
Axn1 <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  xn <- .recycle_x_n(x, n)
  x <- xn$x
  n <- xn$n
  i <- .check_i_discrete(i)

  v <- 1 / (1 + i)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]

    if (nn == 0) return(0)

    k <- 0:(nn - 1)
    probs <- vapply(k, function(kk) .kq_vec(xx, kk, tbl = tbl, model = model, ...), numeric(1))

    sum(v^(k + 1) * probs, na.rm = TRUE)
  })
}

#' Pure endowment APV
#'
#' Computes \eqn{{}_nE_x = v^n {}_n p_x}.
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model name.
#' @param ... Additional arguments passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
nEx <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  xn <- .recycle_x_n(x, n)
  x <- xn$x
  n <- xn$n
  i <- .check_i_discrete(i)

  p <- .get_npx_discrete(x, n, tbl = tbl, model = model, ...)
  p[is.na(p)] <- 0

  discount(i, n) * p
}

#' Deferred insurance APV
#'
#' Computes \eqn{{}_{n\mid}A_x = {}_nE_x A_{x+n}}.
#'
#' @param x Age.
#' @param n Deferral period.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model name.
#' @param ... Additional arguments passed to survival-model functions.
#' @param tol Numerical tolerance for truncating infinite sums.
#' @param k_max Maximum number of terms in the sum.
#'
#' @return Numeric vector of APVs.
#' @export
nAx <- function(x, n, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000) {
  xn <- .recycle_x_n(x, n)
  x <- xn$x
  n <- xn$n

  nEx(x, n, i, tbl = tbl, model = model, ...) *
    Ax(x + n, i, tbl = tbl, model = model, ..., tol = tol, k_max = k_max)
}

#' Endowment insurance APV
#'
#' Computes \eqn{A_{x:\overline{n}|} = A_{x:\overline{n}|}^{1} + {}_nE_x}.
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model name.
#' @param ... Additional arguments passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
Axn <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  Axn1(x, n, i, tbl = tbl, model = model, ...) +
    nEx(x, n, i, tbl = tbl, model = model, ...)
}

# ------------------------------------------------------------
# Second moments
# ------------------------------------------------------------

#' Second moment of whole life insurance PV
#'
#' Computes \eqn{{}^{2}A_x} by evaluating \eqn{A_x} at doubled force.
#'
#' @inheritParams Ax
#' @return Numeric vector of second moments.
#' @export
A2x <- function(x, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000) {
  Ax(x, i = double_force_i(i), tbl = tbl, model = model, ..., tol = tol, k_max = k_max)
}

#' Second moment of term insurance PV
#'
#' Computes \eqn{{}^{2}A_{x:\overline{n}|}^{1}} by evaluating
#' \eqn{A_{x:\overline{n}|}^{1}} at doubled force.
#'
#' @inheritParams Axn1
#' @return Numeric vector of second moments.
#' @export
A2xn1 <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  Axn1(x, n, i = double_force_i(i), tbl = tbl, model = model, ...)
}

#' Second moment of pure endowment PV
#'
#' Computes \eqn{{}^{2}{}_nE_x = (v')^n {}_n p_x}.
#'
#' @inheritParams nEx
#' @return Numeric vector of second moments.
#' @export
A2nEx <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  nEx(x, n, i = double_force_i(i), tbl = tbl, model = model, ...)
}

#' Second moment of deferred insurance PV
#'
#' Computes \eqn{{}^{2}{}_{n\mid}A_x} by evaluating
#' \eqn{{}_{n\mid}A_x} at doubled force.
#'
#' @inheritParams nAx
#' @return Numeric vector of second moments.
#' @export
A2nAx <- function(x, n, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000) {
  nAx(x, n, i = double_force_i(i), tbl = tbl, model = model, ..., tol = tol, k_max = k_max)
}

#' Second moment of endowment insurance PV
#'
#' Computes \eqn{{}^{2}A_{x:\overline{n}|}}.
#'
#' @inheritParams Axn
#' @return Numeric vector of second moments.
#' @export
A2xn <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  A2xn1(x, n, i, tbl = tbl, model = model, ...) +
    A2nEx(x, n, i, tbl = tbl, model = model, ...)
}

# ------------------------------------------------------------
# Variances
# ------------------------------------------------------------

#' Variance of whole life insurance PV
#'
#' Computes \eqn{\mathrm{Var}(Z_x) = {}^{2}A_x - A_x^2}.
#'
#' @inheritParams Ax
#' @return Numeric vector of variances.
#' @export
var_Ax <- function(x, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000) {
  A2x(x, i, tbl = tbl, model = model, ..., tol = tol, k_max = k_max) -
    Ax(x, i, tbl = tbl, model = model, ..., tol = tol, k_max = k_max)^2
}

#' Variance of term insurance PV
#'
#' Computes \eqn{\mathrm{Var}(Z_{x:\overline{n}|}^{1}) =
#' {}^{2}A_{x:\overline{n}|}^{1} - (A_{x:\overline{n}|}^{1})^2}.
#'
#' @inheritParams Axn1
#' @return Numeric vector of variances.
#' @export
var_Axn1 <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  A2xn1(x, n, i, tbl = tbl, model = model, ...) -
    Axn1(x, n, i, tbl = tbl, model = model, ...)^2
}

#' Variance of pure endowment PV
#'
#' @inheritParams nEx
#' @return Numeric vector of variances.
#' @export
var_nEx <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  A2nEx(x, n, i, tbl = tbl, model = model, ...) -
    nEx(x, n, i, tbl = tbl, model = model, ...)^2
}

#' Variance of deferred insurance PV
#'
#' @inheritParams nAx
#' @return Numeric vector of variances.
#' @export
var_nAx <- function(x, n, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000) {
  A2nAx(x, n, i, tbl = tbl, model = model, ..., tol = tol, k_max = k_max) -
    nAx(x, n, i, tbl = tbl, model = model, ..., tol = tol, k_max = k_max)^2
}

#' Variance of endowment insurance PV
#'
#' @inheritParams Axn
#' @return Numeric vector of variances.
#' @export
var_Axn <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  A2xn(x, n, i, tbl = tbl, model = model, ...) -
    Axn(x, n, i, tbl = tbl, model = model, ...)^2
}

# ------------------------------------------------------------
# Covariances
# ------------------------------------------------------------

#' Covariance of term and deferred insurance PVs
#'
#' Computes
#' \eqn{\mathrm{Cov}(Z_{x:\overline{n}|}^{1}, {}_{n\mid}Z_x) =
#' -A_{x:\overline{n}|}^{1} \cdot {}_{n\mid}A_x}.
#'
#' @inheritParams Axn1
#' @return Numeric vector of covariances.
#' @export
cov_term_deferred <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  -Axn1(x, n, i, tbl = tbl, model = model, ...) *
    nAx(x, n, i, tbl = tbl, model = model, ...)
}

#' Covariance of term insurance and pure endowment PVs
#'
#' Computes
#' \eqn{\mathrm{Cov}(Z_{x:\overline{n}|}^{1},
#' Z_{x:\overline{n}|}^{1\text{(pure endow)}}) =
#' -A_{x:\overline{n}|}^{1} \cdot {}_nE_x}.
#'
#' @inheritParams Axn1
#' @return Numeric vector of covariances.
#' @export
cov_term_endow <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  -Axn1(x, n, i, tbl = tbl, model = model, ...) *
    nEx(x, n, i, tbl = tbl, model = model, ...)
}

test_that("nEx computes v^n * npx from life table", {
  tbl <- life_table(
    x = 0:4,
    px = c(0.90, 0.80, 0.60, 0.30, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    nEx(1, 3, i, tbl = tbl),
    discount(i, 3) * npx(tbl, 1, 3)
  )
})

test_that("Axn decomposes into term insurance plus pure endowment", {
  tbl <- life_table(
    x = 0:4,
    px = c(0.90, 0.80, 0.60, 0.30, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    Axn(1, 3, i, tbl = tbl),
    Axn1(1, 3, i, tbl = tbl) + nEx(1, 3, i, tbl = tbl)
  )
})

test_that("Ax decomposes into term plus deferred insurance", {
  tbl <- life_table(
    x = 0:4,
    px = c(0.90, 0.80, 0.60, 0.30, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    Ax(1, i, tbl = tbl),
    Axn1(1, 2, i, tbl = tbl) + nAx(1, 2, i, tbl = tbl)
  )
})

test_that("nAx equals nEx times Ax at age x+n", {
  tbl <- life_table(
    x = 0:4,
    px = c(0.90, 0.80, 0.60, 0.30, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    nAx(1, 2, i, tbl = tbl),
    nEx(1, 2, i, tbl = tbl) * Ax(3, i, tbl = tbl)
  )
})

test_that("Ax satisfies recursion Ax = v qx + v px Ax+1", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  v <- discount(i, 1)

  lhs <- Ax(1, i, tbl = tbl)
  rhs <- v * nqx(tbl, 1, 1) + v * npx(tbl, 1, 1) * Ax(2, i, tbl = tbl)

  expect_equal(lhs, rhs)
})

test_that("A2x evaluates Ax at doubled force", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    A2x(1, i, tbl = tbl),
    Ax(1, double_force_i(i), tbl = tbl)
  )
})

test_that("A2xn1 evaluates Axn1 at doubled force", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    A2xn1(1, 3, i, tbl = tbl),
    Axn1(1, 3, double_force_i(i), tbl = tbl)
  )
})

test_that("A2nEx evaluates nEx at doubled force", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    A2nEx(1, 3, i, tbl = tbl),
    nEx(1, 3, double_force_i(i), tbl = tbl)
  )
})

test_that("A2nAx evaluates nAx at doubled force", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    A2nAx(1, 2, i, tbl = tbl),
    nAx(1, 2, double_force_i(i), tbl = tbl)
  )
})

test_that("A2xn decomposes into second moments of term and pure endowment", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    A2xn(1, 3, i, tbl = tbl),
    A2xn1(1, 3, i, tbl = tbl) + A2nEx(1, 3, i, tbl = tbl)
  )
})

test_that("var_Ax equals A2x - Ax^2", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    var_Ax(1, i, tbl = tbl),
    A2x(1, i, tbl = tbl) - Ax(1, i, tbl = tbl)^2
  )
})

test_that("var_Axn1 equals A2xn1 - Axn1^2", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    var_Axn1(1, 3, i, tbl = tbl),
    A2xn1(1, 3, i, tbl = tbl) - Axn1(1, 3, i, tbl = tbl)^2
  )
})

test_that("var_nEx equals A2nEx - nEx^2", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    var_nEx(1, 3, i, tbl = tbl),
    A2nEx(1, 3, i, tbl = tbl) - nEx(1, 3, i, tbl = tbl)^2
  )
})

test_that("var_nAx equals A2nAx - nAx^2", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    var_nAx(1, 2, i, tbl = tbl),
    A2nAx(1, 2, i, tbl = tbl) - nAx(1, 2, i, tbl = tbl)^2
  )
})

test_that("var_Axn equals A2xn - Axn^2", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    var_Axn(1, 3, i, tbl = tbl),
    A2xn(1, 3, i, tbl = tbl) - Axn(1, 3, i, tbl = tbl)^2
  )
})

test_that("cov_term_deferred equals -Axn1 * nAx", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    cov_term_deferred(1, 2, i, tbl = tbl),
    -Axn1(1, 2, i, tbl = tbl) * nAx(1, 2, i, tbl = tbl)
  )
})

test_that("cov_term_endow equals -Axn1 * nEx", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05
  expect_equal(
    cov_term_endow(1, 3, i, tbl = tbl),
    -Axn1(1, 3, i, tbl = tbl) * nEx(1, 3, i, tbl = tbl)
  )
})

test_that("discrete insurance functions work with parametric survival models", {
  i <- 0.05

  ax_val <- Ax(40, i, model = "uniform", omega = 100)
  term_val <- Axn1(40, 10, i, model = "uniform", omega = 100)
  endow_val <- Axn(40, 10, i, model = "uniform", omega = 100)
  pure_val <- nEx(40, 10, i, model = "uniform", omega = 100)

  expect_true(is.numeric(ax_val))
  expect_true(is.numeric(term_val))
  expect_true(is.numeric(endow_val))
  expect_true(is.numeric(pure_val))

  expect_equal(endow_val, term_val + pure_val)
})

test_that("x and n vectorize correctly", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  i <- 0.05

  expect_length(Axn1(c(1, 2), c(2, 3), i, tbl = tbl), 2)
  expect_length(nEx(c(1, 2), c(2, 3), i, tbl = tbl), 2)
  expect_length(Axn(c(1, 2), c(2, 3), i, tbl = tbl), 2)
})

test_that("invalid inputs are rejected", {
  tbl <- life_table(
    x = 0:5,
    px = c(0.90, 0.80, 0.60, 0.30, 0.20, 0.00),
    radix = 10000
  )

  expect_error(Ax(-1, 0.05, tbl = tbl))
  expect_error(Axn1(1, -2, 0.05, tbl = tbl))
  expect_error(nEx(1, 2.5, 0.05, tbl = tbl))
  expect_error(Ax(1, -1, tbl = tbl))
})



testthat::test_that("Ax works with closed finite life tables", {
  tbl <- life_table(x = 0:3, lx = c(100000, 80000, 50000, 0))
  v <- 1 / 1.05

  expected <- v * 0.2 + v^2 * 0.3 + v^3 * 0.5

  testthat::expect_equal(Ax(0, i = 0.05, tbl = tbl), expected)
  testthat::expect_equal(Ax(0:2, i = 0.05, tbl = tbl), c(expected, v * 0.375 + v^2 * 0.625, v))
})

testthat::test_that("term insurance works near terminal age", {
  tbl <- life_table(x = 0:3, lx = c(100000, 80000, 50000, 0))

  testthat::expect_equal(Axn1(2, n = 1, i = 0.05, tbl = tbl), 1 / 1.05)
  testthat::expect_equal(nEx(2, n = 1, i = 0.05, tbl = tbl), 0)
  testthat::expect_equal(Axn(2, n = 1, i = 0.05, tbl = tbl), 1 / 1.05)
})

#' Mortality improvement projection functions
#'
#' Functions for projecting one-year death and survival probabilities under
#' mortality improvement and evaluating annuity values under projected
#' mortality.
#'
#' The standard projection used is
#' \deqn{q_x^{[Y]} = q_x^{[B]} (1 - AA_x)^{Y-B},}
#' where \eqn{B} is the base year, \eqn{Y} is the projection year, and
#' \eqn{AA_x} is the mortality improvement factor at age \eqn{x}.
#'
#' @name mortality_improvement_projection
NULL

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.check_scalar_year <- function(y, name) {
  y <- as.numeric(y)
  if (length(y) != 1L || !is.finite(y) || abs(y - round(y)) > 1e-10) {
    stop(paste0(name, " must be a single integer-like year."), call. = FALSE)
  }
  as.integer(round(y))
}

.check_year_vector <- function(y, name) {
  y <- as.numeric(y)
  if (length(y) == 0L) {
    stop(paste0(name, " must have positive length."), call. = FALSE)
  }
  if (any(!is.finite(y)) || any(abs(y - round(y)) > 1e-10)) {
    stop(paste0(name, " must contain integer-like years."), call. = FALSE)
  }
  as.integer(round(y))
}

.check_prob_vector <- function(x, name) {
  x <- as.numeric(x)
  if (length(x) == 0L) {
    stop(paste0(name, " must have positive length."), call. = FALSE)
  }
  if (any(!is.finite(x))) {
    stop(paste0(name, " must be finite."), call. = FALSE)
  }
  if (any(x < 0 | x > 1)) {
    stop(paste0(name, " must lie in [0, 1]."), call. = FALSE)
  }
  x
}

.recycle_proj_inputs <- function(qx_base, AAx, proj_year) {
  qx_base <- as.numeric(qx_base)
  AAx <- as.numeric(AAx)
  proj_year <- as.numeric(proj_year)

  lens <- c(length(qx_base), length(AAx), length(proj_year))
  L <- max(lens)

  if (!all(lens %in% c(1L, L))) {
    stop("qx_base, AAx, and proj_year must have compatible lengths.", call. = FALSE)
  }

  list(
    qx_base = rep_len(qx_base, L),
    AAx = rep_len(AAx, L),
    proj_year = rep_len(proj_year, L)
  )
}

.recycle_improved_annuity_inputs <- function(x0, n = NULL, u = NULL, i) {
  args <- list(x0 = as.numeric(x0), i = as.numeric(i))
  if (!is.null(n)) args$n <- as.numeric(n)
  if (!is.null(u)) args$u <- as.numeric(u)

  lens <- vapply(args, length, integer(1))
  L <- max(lens)

  if (any(lens == 0L)) {
    stop("Vectorized inputs must have positive length.", call. = FALSE)
  }
  if (!all(lens %in% c(1L, L))) {
    stop("Vectorized inputs must have compatible lengths.", call. = FALSE)
  }
  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("Vectorized inputs must be finite.", call. = FALSE)
  }

  lapply(args, rep_len, length.out = L)
}

.check_nonnegative_integer_scalar <- function(x, name) {
  x <- as.numeric(x)
  if (length(x) != 1L || !is.finite(x) || x < 0 || abs(x - round(x)) > 1e-10) {
    stop(paste0(name, " must be a single nonnegative integer."), call. = FALSE)
  }
  as.integer(round(x))
}

.check_improved_interest <- function(i) {
  i <- as.numeric(i)
  if (length(i) != 1L || !is.finite(i) || i <= -1) {
    stop("i must be a single value > -1.", call. = FALSE)
  }
  i
}

# -------------------------------------------------------------------------
# One-year projected probabilities
# -------------------------------------------------------------------------

#' Project one-year death probability under mortality improvement
#'
#' @param qx_base Base-year one-year death probability.
#' @param AAx Mortality improvement factor.
#' @param base_year Base year.
#' @param proj_year Projection year. May be scalar or vector.
#'
#' @return Numeric vector of projected one-year death probabilities.
#' @rdname mortality_improvement_projection
#' @export
qx_proj <- function(qx_base, AAx, base_year, proj_year) {
  qx_base <- .check_prob_vector(qx_base, "qx_base")
  AAx <- .check_prob_vector(AAx, "AAx")
  base_year <- .check_scalar_year(base_year, "base_year")
  proj_year <- .check_year_vector(proj_year, "proj_year")

  rr <- .recycle_proj_inputs(qx_base, AAx, proj_year)

  rr$qx_base * (1 - rr$AAx)^(rr$proj_year - base_year)
}

#' Project one-year survival probability under mortality improvement
#'
#' @rdname mortality_improvement_projection
#' @export
px_proj <- function(qx_base, AAx, base_year, proj_year) {
  1 - qx_proj(qx_base, AAx, base_year, proj_year)
}

# -------------------------------------------------------------------------
# Multi-year survival under improvement
# -------------------------------------------------------------------------

#' Multi-year survival probability under mortality improvement
#'
#' @param x0 Issue age.
#' @param n Number of years.
#' @param qx_base_vec Base-year death probabilities for successive ages.
#' @param AAx_vec Mortality improvement factors for successive ages.
#' @param issue_year Issue year.
#'
#' @rdname mortality_improvement_projection
#' @export
tpx_improved <- function(x0, n, qx_base_vec, AAx_vec, base_year, issue_year) {
  x0 <- as.numeric(x0)
  if (length(x0) != 1L || !is.finite(x0) || x0 < 0) {
    stop("x0 must be a single nonnegative number.", call. = FALSE)
  }

  n <- .check_nonnegative_integer_scalar(n, "n")
  base_year <- .check_scalar_year(base_year, "base_year")
  issue_year <- .check_scalar_year(issue_year, "issue_year")

  qx_base_vec <- .check_prob_vector(qx_base_vec, "qx_base_vec")
  AAx_vec <- .check_prob_vector(AAx_vec, "AAx_vec")

  if (n == 0L) return(1)

  if (length(qx_base_vec) < n || length(AAx_vec) < n) {
    stop("qx_base_vec and AAx_vec must each have length at least n.", call. = FALSE)
  }

  years <- issue_year + 0:(n - 1L)

  qproj <- qx_proj(
    qx_base = qx_base_vec[1:n],
    AAx = AAx_vec[1:n],
    base_year = base_year,
    proj_year = years
  )

  prod(1 - qproj)
}

# -------------------------------------------------------------------------
# Annuity functions under improvement
# -------------------------------------------------------------------------

#' Temporary annuity-immediate under mortality improvement
#'
#' @param i Effective annual interest rate.
#'
#' @rdname mortality_improvement_projection
#' @export
axn_improved <- function(x0, n, i, qx_base_vec, AAx_vec, base_year, issue_year) {
  args <- .recycle_improved_annuity_inputs(x0 = x0, n = n, i = i)
  x0 <- args$x0
  n <- args$n
  i <- args$i

  sapply(seq_along(x0), function(j) {
    ii <- .check_improved_interest(i[j])
    nn <- .check_nonnegative_integer_scalar(n[j], "n")
    if (nn == 0L) return(0)

    survs <- sapply(1:nn, function(t) {
      tpx_improved(x0[j], t, qx_base_vec, AAx_vec, base_year, issue_year)
    })

    sum(discount(ii, 1:nn) * survs)
  })
}

#' Deferred temporary annuity-immediate under mortality improvement
#'
#' @param u Deferral period in years.
#'
#' @rdname mortality_improvement_projection
#' @export
naxn_improved <- function(x0, u, n, i, qx_base_vec, AAx_vec, base_year, issue_year) {
  args <- .recycle_improved_annuity_inputs(x0 = x0, u = u, n = n, i = i)
  x0 <- args$x0
  u <- args$u
  n <- args$n
  i <- args$i

  sapply(seq_along(x0), function(j) {
    ii <- .check_improved_interest(i[j])
    uu <- .check_nonnegative_integer_scalar(u[j], "u")
    nn <- .check_nonnegative_integer_scalar(n[j], "n")
    if (nn == 0L) return(0)

    times <- (uu + 1L):(uu + nn)

    survs <- sapply(times, function(t) {
      tpx_improved(x0[j], t, qx_base_vec, AAx_vec, base_year, issue_year)
    })

    sum(discount(ii, times) * survs)
  })
}

#' Whole life annuity-immediate under mortality improvement
#'
#' @rdname mortality_improvement_projection
#' @export
ax_improved <- function(x0, i, qx_base_vec, AAx_vec, base_year, issue_year) {
  args <- .recycle_improved_annuity_inputs(x0 = x0, i = i)
  x0 <- args$x0
  i <- args$i

  qx_base_vec <- .check_prob_vector(qx_base_vec, "qx_base_vec")
  AAx_vec <- .check_prob_vector(AAx_vec, "AAx_vec")

  L <- min(length(qx_base_vec), length(AAx_vec))
  if (L == 0L) return(rep(0, length(x0)))

  sapply(seq_along(x0), function(j) {
    ii <- .check_improved_interest(i[j])

    survs <- sapply(1:L, function(t) {
      tpx_improved(
        x0 = x0[j],
        n = t,
        qx_base_vec = qx_base_vec[1:L],
        AAx_vec = AAx_vec[1:L],
        base_year = base_year,
        issue_year = issue_year
      )
    })

    sum(discount(ii, 1:L) * survs)
  })
}

test_that("qx_proj returns base rates when projection year equals base year", {
  qx <- c(0.01, 0.02, 0.03)
  AA <- c(0.01, 0.02, 0.03)

  expect_equal(
    qx_proj(qx, AA, base_year = 2000, proj_year = 2000),
    qx
  )
})

test_that("px_proj equals 1 - qx_proj", {
  qx <- c(0.01, 0.02, 0.03)
  AA <- c(0.01, 0.02, 0.03)

  qproj <- qx_proj(qx, AA, 2000, 2005)
  pproj <- px_proj(qx, AA, 2000, 2005)

  expect_equal(pproj, 1 - qproj)
})

test_that("mortality improvement lowers projected q when AA is positive", {
  qx <- c(0.01, 0.02, 0.03)
  AA <- c(0.01, 0.01, 0.01)

  qproj <- qx_proj(qx, AA, 2000, 2010)

  expect_true(all(qproj < qx))
})

test_that("tpx_improved equals 1 when n = 0", {
  expect_equal(
    tpx_improved(
      x0 = 60, n = 0,
      qx_base_vec = c(0.01, 0.02),
      AAx_vec = c(0.01, 0.01),
      base_year = 2000,
      issue_year = 2005
    ),
    1
  )
})

test_that("tpx_improved matches manual product", {
  qx_base <- c(0.10, 0.20, 0.30)
  AA <- c(0.00, 0.00, 0.00)

  expected <- (1 - 0.10) * (1 - 0.20) * (1 - 0.30)

  actual <- tpx_improved(
    x0 = 50, n = 3,
    qx_base_vec = qx_base,
    AAx_vec = AA,
    base_year = 2000,
    issue_year = 2000
  )

  expect_equal(actual, expected)
})

test_that("tpx_improved increases when mortality improvement is introduced", {
  qx_base <- c(0.02, 0.03, 0.04, 0.05)
  AA0 <- c(0, 0, 0, 0)
  AA1 <- c(0.01, 0.01, 0.01, 0.01)

  no_improve <- tpx_improved(
    x0 = 60, n = 4,
    qx_base_vec = qx_base,
    AAx_vec = AA0,
    base_year = 2000,
    issue_year = 2000
  )

  improve <- tpx_improved(
    x0 = 60, n = 4,
    qx_base_vec = qx_base,
    AAx_vec = AA1,
    base_year = 2000,
    issue_year = 2000
  )

  expect_true(improve >= no_improve)
})

test_that("axn_improved matches textbook-style direct summation", {
  x0 <- 60
  n <- 2
  i <- 0.06
  qx_base <- c(0.0479, 0.0539)
  AA <- c(0.01, 0.01)

  p1 <- 1 - qx_proj(qx_base[1], AA[1], 2000, 2020)
  p2 <- 1 - qx_proj(qx_base[2], AA[2], 2000, 2021)

  expected <- discount(i, 1) * p1 +
    discount(i, 2) * p1 * p2

  actual <- axn_improved(
    x0 = x0, n = n, i = i,
    qx_base_vec = qx_base,
    AAx_vec = AA,
    base_year = 2000,
    issue_year = 2020
  )

  expect_equal(actual, expected)
})

test_that("deferred temporary annuity under improvement reduces to shifted sum", {
  x0 <- 60
  u <- 2
  n <- 2
  i <- 0.05
  qx_base <- c(0.01, 0.02, 0.03, 0.04)
  AA <- c(0.00, 0.00, 0.00, 0.00)

  expected <- sum(sapply((u + 1):(u + n), function(t) {
    discount(i, t) * tpx_improved(
      x0 = x0, n = t,
      qx_base_vec = qx_base,
      AAx_vec = AA,
      base_year = 2000,
      issue_year = 2000
    )
  }))

  actual <- naxn_improved(
    x0 = x0, u = u, n = n, i = i,
    qx_base_vec = qx_base,
    AAx_vec = AA,
    base_year = 2000,
    issue_year = 2000
  )

  expect_equal(actual, expected)
})

test_that("whole life annuity under improvement is at least as large as temporary annuity with same horizon", {
  x0 <- 60
  i <- 0.05
  qx_base <- c(0.01, 0.02, 0.03, 0.04, 0.05)
  AA <- c(0.01, 0.01, 0.01, 0.01, 0.01)

  whole <- ax_improved(
    x0 = x0, i = i,
    qx_base_vec = qx_base,
    AAx_vec = AA,
    base_year = 2000,
    issue_year = 2010
  )

  temp <- axn_improved(
    x0 = x0, n = 5, i = i,
    qx_base_vec = qx_base,
    AAx_vec = AA,
    base_year = 2000,
    issue_year = 2010
  )

  expect_true(whole >= temp)
})

test_that("functions reject invalid inputs", {
  expect_error(qx_proj(0.01, 0.01, 2000, 2000.5), "proj_year")
  expect_error(qx_proj(-0.01, 0.01, 2000, 2005), "qx_base")
  expect_error(px_proj(0.01, 1.2, 2000, 2005), "AAx")

  expect_error(
    tpx_improved(
      x0 = 60, n = 3,
      qx_base_vec = c(0.01, 0.02),
      AAx_vec = c(0.01, 0.01),
      base_year = 2000,
      issue_year = 2005
    ),
    "length at least n"
  )
})

test_that("qx_proj vectorizes over qx and AAx", {
  out <- qx_proj(
    qx_base = c(0.01, 0.02, 0.03),
    AAx = 0.01,
    base_year = 2000,
    proj_year = 2010
  )

  expect_length(out, 3)
  expect_true(all(is.finite(out)))
})

test_that("projected annuity values are finite in standard cases", {
  val1 <- axn_improved(
    x0 = 60, n = 3, i = 0.05,
    qx_base_vec = c(0.02, 0.03, 0.04),
    AAx_vec = c(0.01, 0.01, 0.01),
    base_year = 2000,
    issue_year = 2020
  )

  val2 <- naxn_improved(
    x0 = 60, u = 2, n = 3, i = 0.05,
    qx_base_vec = c(0.02, 0.03, 0.04, 0.05, 0.06),
    AAx_vec = c(0.01, 0.01, 0.01, 0.01, 0.01),
    base_year = 2000,
    issue_year = 2020
  )

  expect_true(is.finite(val1))
  expect_true(is.finite(val2))
})


test_that("mortality improvement annuities vectorize over i", {
  qx <- rep(0.01, 20)
  aa <- rep(0.02, 20)

  out <- axn_improved(
    x0 = 40,
    n = 10,
    i = c(0.03, 0.05),
    qx_base_vec = qx,
    AAx_vec = aa,
    base_year = 2020,
    issue_year = 2025
  )

  expect_length(out, 2)
  expect_true(out[1] > out[2])
})

test_that("mortality improvement annuities reject incompatible vector lengths", {
  qx <- rep(0.01, 20)
  aa <- rep(0.02, 20)

  expect_error(
    axn_improved(
      x0 = c(40, 45),
      n = c(5, 10, 15),
      i = 0.05,
      qx_base_vec = qx,
      AAx_vec = aa,
      base_year = 2020,
      issue_year = 2025
    ),
    "compatible lengths"
  )
})

# life_table_expectations.R
#
# Expectation functions for discrete tabular life tables.

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.check_n_temp <- function(n) {
  n <- as.numeric(n)

  if (length(n) == 0L || any(!is.finite(n)) || any(n < 0)) {
    stop("n must contain finite nonnegative values.", call. = FALSE)
  }

  n
}

.recycle_expectation_args <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) {
    stop("All arguments must have positive length.", call. = FALSE)
  }

  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)

  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) {
    if (length(z) == 1L) rep(z, common_len) else z
  })
}

.check_age_values <- function(x) {
  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  invisible(TRUE)
}

# -------------------------------------------------------------------------
# Curtate expectation of life
# -------------------------------------------------------------------------

#' Curtate expectation of life from a life table
#'
#' Computes the curtate expectation of life
#' \eqn{e_x = \sum_{k=1}^{\infty} {}_k p_x}
#' in the discrete tabular setting.
#'
#' @param tbl A \code{life_table} object.
#' @param x Numeric vector of integer ages.
#'
#' @return Numeric vector of curtate expectations \eqn{e_x}.
#' @export
ex_curtate_tab <- function(tbl, x) {
  .validate_life_table(tbl)

  args <- .recycle_expectation_args(x)
  x <- args[[1L]]
  .check_age_values(x)

  sapply(x, function(xx) {
    lx_x <- lx(tbl, xx)

    if (is.na(lx_x) || lx_x <= 0) {
      return(NA_real_)
    }

    max_k <- max(tbl$x, na.rm = TRUE) - xx

    if (max_k <= 0) {
      return(0)
    }

    ks <- seq_len(max_k)
    probs <- npx(tbl, x = xx, n = ks)

    sum(probs, na.rm = TRUE)
  })
}

# -------------------------------------------------------------------------
# Temporary curtate expectation of life
# -------------------------------------------------------------------------

#' Temporary curtate expectation of life from a life table
#'
#' Computes
#' \eqn{e_{x:\overline{n}|} = \sum_{k=1}^{n} {}_k p_x}
#' for integer \eqn{n} in the discrete tabular setting.
#'
#' @param tbl A \code{life_table} object.
#' @param x Numeric vector of integer ages.
#' @param n Numeric vector of nonnegative integers.
#'
#' @return Numeric vector of temporary curtate expectations.
#' @export
ex_temp_curtate_tab <- function(tbl, x, n) {
  .validate_life_table(tbl)

  args <- .recycle_expectation_args(x, n)
  x <- args[[1L]]
  n <- args[[2L]]

  .check_age_values(x)

  if (any(n < 0 | n != floor(n))) {
    stop("n must be a nonnegative integer.", call. = FALSE)
  }

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]

    lx_x <- lx(tbl, xx)

    if (is.na(lx_x) || lx_x <= 0) {
      return(NA_real_)
    }

    if (nn == 0) {
      return(0)
    }

    ks <- seq_len(nn)
    probs <- npx(tbl, x = xx, n = ks)

    sum(probs, na.rm = TRUE)
  })
}

# -------------------------------------------------------------------------
# Temporary complete expectation of life
# -------------------------------------------------------------------------

#' Temporary complete expectation of life from a life table
#'
#' Computes
#' \eqn{\overset{\circ}{e}_{x:\overline{n}|} =
#' \int_0^n {}_t p_x \, dt}
#' using a within-year assumption:
#' \code{"udd"}, \code{"cf"}, or \code{"balducci"}.
#'
#' @param tbl A \code{life_table} object.
#' @param x Numeric vector of integer ages.
#' @param n Numeric vector of nonnegative durations.
#' @param assumption One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.
#'
#' @return Numeric vector of temporary complete expectations.
#' @export
ex_temp_complete_tab <- function(tbl, x, n,
                                 assumption = c("udd", "cf", "balducci")) {
  .validate_life_table(tbl)

  assumption <- .match_assumption(assumption)

  args <- .recycle_expectation_args(x, n)
  x <- args[[1L]]
  n <- .check_n_temp(args[[2L]])

  .check_age_values(x)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]

    lx_x <- lx(tbl, xx)

    if (is.na(lx_x) || lx_x <= 0) {
      return(NA_real_)
    }

    if (nn == 0) {
      return(0)
    }

    whole <- floor(nn)
    frac <- nn - whole
    total <- 0

    if (whole > 0) {
      rs <- 0:(whole - 1L)
      surv_to_r <- npx(tbl, x = xx, n = rs)

      one_year_piece <- vapply(rs, function(r) {
        age_now <- xx + r
        lx_now <- lx(tbl, age_now)

        if (is.na(lx_now) || lx_now <= 0) {
          return(0)
        }

        integrate(
          function(s) {
            tpx_tab(tbl, x = age_now, t = s, assumption = assumption)
          },
          lower = 0,
          upper = 1,
          rel.tol = 1e-9
        )$value
      }, numeric(1))

      total <- total + sum(surv_to_r * one_year_piece, na.rm = TRUE)
    }

    if (frac > 0) {
      surv_to_whole <- npx(tbl, x = xx, n = whole)
      age_now <- xx + whole
      lx_now <- lx(tbl, age_now)

      if (!is.na(surv_to_whole) &&
          !is.na(lx_now) &&
          surv_to_whole > 0 &&
          lx_now > 0) {
        frac_piece <- integrate(
          function(s) {
            tpx_tab(tbl, x = age_now, t = s, assumption = assumption)
          },
          lower = 0,
          upper = frac,
          rel.tol = 1e-9
        )$value

        total <- total + surv_to_whole * frac_piece
      }
    }

    total
  })
}

# -------------------------------------------------------------------------
# Complete expectation of life
# -------------------------------------------------------------------------

#' Complete expectation of life from a life table
#'
#' Computes
#' \eqn{\overset{\circ}{e}_x = \int_0^\infty {}_t p_x \, dt}
#' using a within-year assumption:
#' \code{"udd"}, \code{"cf"}, or \code{"balducci"}.
#'
#' @param tbl A \code{life_table} object.
#' @param x Numeric vector of integer ages.
#' @param assumption One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.
#'
#' @return Numeric vector of complete expectations \eqn{\overset{\circ}{e}_x}.
#' @export
ex_complete_tab <- function(tbl, x,
                            assumption = c("udd", "cf", "balducci")) {
  .validate_life_table(tbl)

  assumption <- .match_assumption(assumption)

  args <- .recycle_expectation_args(x)
  x <- args[[1L]]

  .check_age_values(x)

  sapply(x, function(xx) {
    lx_x <- lx(tbl, xx)

    if (is.na(lx_x) || lx_x <= 0) {
      return(NA_real_)
    }

    max_n <- max(tbl$x, na.rm = TRUE) - xx

    if (max_n <= 0) {
      return(0)
    }

    ex_temp_complete_tab(tbl, x = xx, n = max_n, assumption = assumption)
  })
}

# =========================================================
# Multiple-decrement applications
# =========================================================

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.md_check_numeric <- function(x, name) {
  if (!is.numeric(x) || length(x) == 0L || any(!is.finite(x))) {
    stop(name, " must contain finite numeric values.", call. = FALSE)
  }
  as.numeric(x)
}

#' @noRd
.md_check_nonnegative <- function(x, name) {
  x <- .md_check_numeric(x, name)
  if (any(x < 0)) stop(name, " must contain nonnegative values.", call. = FALSE)
  x
}

#' @noRd
.md_check_probability <- function(x, name) {
  x <- .md_check_numeric(x, name)
  if (any(x < 0 | x > 1)) stop(name, " must contain values in [0, 1].", call. = FALSE)
  x
}

#' @noRd
.md_check_interest <- function(i, name = "i") {
  i <- .md_check_numeric(i, name)
  if (any(i <= -1)) stop(name, " must contain values greater than -1.", call. = FALSE)
  i
}

#' @noRd
.md_check_scalar <- function(x, name, lower = -Inf, strict = FALSE) {
  x <- .md_check_numeric(x, name)
  if (length(x) != 1L) stop(name, " must be a finite numeric scalar.", call. = FALSE)
  bad <- if (strict) x <= lower else x < lower
  if (bad) stop(name, if (strict) " must be greater than " else " must be at least ", lower, ".", call. = FALSE)
  x
}

#' @noRd
.md_check_integer_scalar <- function(x, name, lower = 0L) {
  x <- .md_check_numeric(x, name)
  if (length(x) != 1L || x < lower || abs(x - round(x)) > 1e-10) {
    stop(name, " must be a single integer greater than or equal to ", lower, ".", call. = FALSE)
  }
  as.integer(round(x))
}

#' @noRd
.md_check_function <- function(x, name) {
  if (!is.function(x)) stop(name, " must be a function.", call. = FALSE)
  x
}

#' @noRd
.md_recycle <- function(..., .names = NULL) {
  values <- list(...)
  lengths <- vapply(values, length, integer(1))
  if (any(lengths == 0L)) stop("Inputs must have positive length.", call. = FALSE)
  common <- max(lengths)
  if (is.null(.names)) .names <- paste0("Argument ", seq_along(values))
  for (j in seq_along(values)) {
    if (!lengths[j] %in% c(1L, common)) {
      stop(.names[j], " must have length 1 or the common length ", common, ".", call. = FALSE)
    }
    values[[j]] <- rep(values[[j]], length.out = common)
  }
  values
}

#' @noRd
.md_check_grid <- function(t) {
  t <- .md_check_nonnegative(t, "t")
  if (length(t) < 2L) stop("t must contain at least two time points.", call. = FALSE)
  if (is.unsorted(t, strictly = TRUE)) stop("t must be strictly increasing.", call. = FALSE)
  t
}

#' @noRd
.md_grid <- function(n, h) {
  n <- .md_check_scalar(n, "n", lower = 0)
  h <- .md_check_scalar(h, "h", lower = 0, strict = TRUE)
  if (n == 0) return(0)
  out <- seq(0, n, by = h)
  if (tail(out, 1L) < n) out <- c(out, n)
  out[length(out)] <- n
  unique(out)
}

#' @noRd
.md_trapz <- function(t, y) {
  sum(diff(t) * (head(y, -1L) + tail(y, -1L)) / 2)
}

# -------------------------------------------------------------------------
# Insurance present values
# -------------------------------------------------------------------------

#' Discrete multiple-decrement insurance present value
#'
#' Computes the actuarial present value of a benefit payable at the end of
#' the year of decrement from a specified cause.
#'
#' @param qj Cause-specific decrement probabilities by policy year.
#' @param ptau In-force probabilities at the beginning of each policy year.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param benefit Benefit payable on decrement. May be scalar or vector.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @examples
#' qj <- c(0.02, 0.02, 0.02, 0.02, 0.02)
#' ptau <- c(1, 0.95, 0.89, 0.82, 0.74)
#' Axj_md(qj, ptau, i = 0.06, benefit = 1000)
#'
#' @export
Axj_md <- function(qj, ptau, i, benefit = 1) {
  qj <- .md_check_probability(qj, "qj")
  ptau <- .md_check_probability(ptau, "ptau")
  if (length(qj) != length(ptau)) stop("qj and ptau must have the same length.", call. = FALSE)
  i <- .md_check_interest(i)
  benefit <- .md_check_nonnegative(benefit, "benefit")
  vals <- .md_recycle(i, benefit, .names = c("i", "benefit"))
  times <- seq_along(qj)
  vapply(seq_along(vals[[1L]]), function(j) {
    sum(vals[[2L]][j] * (1 + vals[[1L]][j])^(-times) * ptau * qj)
  }, numeric(1))
}

#' Continuous multiple-decrement insurance present value
#'
#' Approximates the actuarial present value of a benefit payable at the moment
#' of decrement using the trapezoidal rule.
#'
#' @param t Strictly increasing nonnegative time points.
#' @param ptau In-force probabilities at the supplied time points.
#' @param muj Cause-specific decrement intensities at the supplied time points.
#' @param delta Force of interest. May be scalar or vector.
#' @param benefit Benefit payable on decrement. May be scalar or vector.
#'
#' @return A numeric vector of actuarial present values.
#'
#' @examples
#' t <- seq(0, 20, by = 0.1)
#' ptau <- exp(-0.012 * t)
#' muj <- rep(0.002, length(t))
#' Abarxj_md(t, ptau, muj, delta = 0.05, benefit = 2000)
#'
#' @export
Abarxj_md <- function(t, ptau, muj, delta, benefit = 1) {
  t <- .md_check_grid(t)
  ptau <- .md_check_probability(ptau, "ptau")
  muj <- .md_check_nonnegative(muj, "muj")
  if (length(ptau) != length(t) || length(muj) != length(t)) {
    stop("t, ptau, and muj must have the same length.", call. = FALSE)
  }
  delta <- .md_check_numeric(delta, "delta")
  benefit <- .md_check_nonnegative(benefit, "benefit")
  vals <- .md_recycle(delta, benefit, .names = c("delta", "benefit"))
  vapply(seq_along(vals[[1L]]), function(j) {
    .md_trapz(t, vals[[2L]][j] * exp(-vals[[1L]][j] * t) * ptau * muj)
  }, numeric(1))
}

# -------------------------------------------------------------------------
# Asset shares
# -------------------------------------------------------------------------

#' General projected asset-share path
#'
#' Computes projected asset shares for a multiple-decrement contract. Rows of
#' \code{b_mat} and \code{q_mat} represent policy years and columns represent
#' decrement causes.
#'
#' @param AS0 Initial asset share.
#' @param G Premium amount by policy year.
#' @param r Percent-of-premium expense rate by policy year.
#' @param e Fixed expense by policy year.
#' @param b_mat Matrix of decrement benefits.
#' @param q_mat Matrix of decrement probabilities.
#' @param p_tau In-force probability by policy year.
#' @param i Effective annual interest rate by policy year.
#' @param b_surv Survival benefit by policy year.
#'
#' @return A data frame with policy year \code{k} and asset share \code{AS}.
#'
#' @export
AS_path_md <- function(AS0, G, r, e, b_mat, q_mat, p_tau, i, b_surv = NULL) {
  AS0 <- .md_check_scalar(AS0, "AS0")
  if (!is.matrix(b_mat) || !is.matrix(q_mat)) stop("b_mat and q_mat must be matrices.", call. = FALSE)
  if (!identical(dim(b_mat), dim(q_mat))) stop("b_mat and q_mat must have the same dimensions.", call. = FALSE)
  if (nrow(b_mat) == 0L || ncol(b_mat) == 0L) stop("b_mat and q_mat must be nonempty.", call. = FALSE)
  storage.mode(b_mat) <- "double"
  storage.mode(q_mat) <- "double"
  if (any(!is.finite(b_mat)) || any(b_mat < 0)) stop("b_mat must contain nonnegative finite values.", call. = FALSE)
  if (any(!is.finite(q_mat)) || any(q_mat < 0) || any(q_mat > 1)) stop("q_mat must contain values in [0, 1].", call. = FALSE)
  if (any(rowSums(q_mat) > 1 + 1e-12)) stop("Each row sum of q_mat must not exceed 1.", call. = FALSE)

  n_years <- nrow(b_mat)
  if (is.null(b_surv)) b_surv <- 0
  vals <- .md_recycle(
    .md_check_nonnegative(G, "G"),
    .md_check_probability(r, "r"),
    .md_check_nonnegative(e, "e"),
    .md_check_probability(p_tau, "p_tau"),
    .md_check_interest(i),
    .md_check_nonnegative(b_surv, "b_surv"),
    .names = c("G", "r", "e", "p_tau", "i", "b_surv")
  )
  common <- length(vals[[1L]])
  if (common == 1L) vals <- lapply(vals, rep, length.out = n_years)
  if (length(vals[[1L]]) != n_years) stop("Yearly inputs must have length 1 or nrow(b_mat).", call. = FALSE)
  if (any(vals[[4L]] <= 0)) stop("p_tau must be positive in every policy year.", call. = FALSE)

  AS <- numeric(n_years + 1L)
  AS[1L] <- AS0
  for (k in seq_len(n_years)) {
    claims <- sum(b_mat[k, ] * q_mat[k, ])
    AS[k + 1L] <- ((AS[k] + vals[[1L]][k] * (1 - vals[[2L]][k]) - vals[[3L]][k]) *
                     (1 + vals[[5L]][k]) - claims) / vals[[4L]][k] - vals[[6L]][k]
  }
  data.frame(k = 0:n_years, AS = AS)
}

#' Projected asset-share path for two decrement causes
#'
#' Convenience interface to \code{AS_path_md()} for two causes.
#'
#' @param AS0 Initial asset share.
#' @param G Premium amount by policy year.
#' @param r Percent-of-premium expense rate by policy year.
#' @param e Fixed expense by policy year.
#' @param b1 Cause 1 benefit by policy year.
#' @param b2 Cause 2 benefit by policy year.
#' @param q1 Cause 1 probability by policy year.
#' @param q2 Cause 2 probability by policy year.
#' @param p_tau In-force probability by policy year.
#' @param i Effective annual interest rate by policy year.
#' @param b3 Survival benefit by policy year.
#'
#' @return A data frame with policy year \code{k} and asset share \code{AS}.
#'
#' @export
AS_path <- function(AS0, G, r, e, b1, b2, q1, q2, p_tau, i, b3 = NULL) {
  vals <- .md_recycle(
    .md_check_nonnegative(b1, "b1"),
    .md_check_nonnegative(b2, "b2"),
    .md_check_probability(q1, "q1"),
    .md_check_probability(q2, "q2"),
    .names = c("b1", "b2", "q1", "q2")
  )
  if (any(vals[[3L]] + vals[[4L]] > 1 + 1e-12)) stop("q1 + q2 must not exceed 1.", call. = FALSE)
  AS_path_md(
    AS0 = AS0, G = G, r = r, e = e,
    b_mat = cbind(cause1 = vals[[1L]], cause2 = vals[[2L]]),
    q_mat = cbind(cause1 = vals[[3L]], cause2 = vals[[4L]]),
    p_tau = p_tau, i = i, b_surv = b3
  )
}

# -------------------------------------------------------------------------
# Disability-state probabilities and premiums
# -------------------------------------------------------------------------

#' Euler approximation of disability-state probabilities
#'
#' Approximates probabilities of being healthy, disabled, or deceased in a
#' three-state model that allows recovery from disability. The final time is
#' always included, with a shorter final step when needed.
#'
#' @param h Positive step size.
#' @param n Nonnegative final time.
#' @param mu01 Healthy-to-disabled intensity function.
#' @param mu02 Healthy-to-deceased intensity function.
#' @param mu10 Disabled-to-healthy intensity function.
#' @param mu12 Disabled-to-deceased intensity function.
#' @param p00_0 Initial healthy-state probability.
#' @param p01_0 Initial disabled-state probability.
#'
#' @return A data frame with columns \code{t}, \code{tp00}, \code{tp01},
#'   and \code{tp02}.
#'
#' @examples
#' mu01 <- function(t) 0.10 * t + 0.20
#' mu02 <- function(t) 0.20
#' mu10 <- function(t) 0.50
#' mu12 <- function(t) 0.125 * t + 0.20
#' tp00_tp01_euler(0.10, 2, mu01, mu02, mu10, mu12)
#'
#' @export
tp00_tp01_euler <- function(h, n, mu01, mu02, mu10, mu12,
                            p00_0 = 1, p01_0 = 0) {
  grid <- .md_grid(n, h)
  mu01 <- .md_check_function(mu01, "mu01")
  mu02 <- .md_check_function(mu02, "mu02")
  mu10 <- .md_check_function(mu10, "mu10")
  mu12 <- .md_check_function(mu12, "mu12")
  p00_0 <- .md_check_probability(p00_0, "p00_0")
  p01_0 <- .md_check_probability(p01_0, "p01_0")
  if (length(p00_0) != 1L || length(p01_0) != 1L) stop("p00_0 and p01_0 must be scalar.", call. = FALSE)
  if (p00_0 + p01_0 > 1 + 1e-12) stop("p00_0 + p01_0 must not exceed 1.", call. = FALSE)

  p00 <- numeric(length(grid)); p01 <- numeric(length(grid))
  p00[1L] <- p00_0; p01[1L] <- p01_0
  if (length(grid) > 1L) {
    for (idx in seq_len(length(grid) - 1L)) {
      time <- grid[idx]; step <- grid[idx + 1L] - grid[idx]
      rates <- c(mu01(time), mu02(time), mu10(time), mu12(time))
      if (length(rates) != 4L || any(!is.finite(rates)) || any(rates < 0)) {
        stop("Intensity functions must return one nonnegative finite value.", call. = FALSE)
      }
      dp00 <- p01[idx] * rates[3L] - p00[idx] * (rates[1L] + rates[2L])
      dp01 <- p00[idx] * rates[1L] - p01[idx] * (rates[3L] + rates[4L])
      p00[idx + 1L] <- p00[idx] + step * dp00
      p01[idx + 1L] <- p01[idx] + step * dp01
    }
  }
  data.frame(t = grid, tp00 = p00, tp01 = p01, tp02 = 1 - p00 - p01)
}

#' Continuous premium approximation in a disability model
#'
#' Approximates a continuous premium rate by trapezoidal integration.
#'
#' @param t Strictly increasing nonnegative time points.
#' @param tp00 Healthy-state probabilities.
#' @param tp01 Disabled-state probabilities.
#' @param delta Force of interest.
#' @param mu02 Healthy-to-deceased intensity function.
#' @param mu12 Disabled-to-deceased intensity function.
#' @param B02 Benefit on death while healthy.
#' @param B12 Benefit on death while disabled.
#' @param R Continuous disability income rate.
#'
#' @return A numeric scalar.
#'
#' @examples
#' mu01 <- function(t) 0.10 * t + 0.20
#' mu02 <- function(t) 0.20
#' mu10 <- function(t) 0.50
#' mu12 <- function(t) 0.125 * t + 0.20
#'
#' probs <- tp00_tp01_euler(
#'   h = 0.10,
#'   n = 2,
#'   mu01 = mu01,
#'   mu02 = mu02,
#'   mu10 = mu10,
#'   mu12 = mu12
#' )
#'
#' Pbar_trapz_ms(
#'   t = probs$t,
#'   tp00 = probs$tp00,
#'   tp01 = probs$tp01,
#'   delta = 0.04,
#'   mu02 = mu02,
#'   mu12 = mu12,
#'   B02 = 1000,
#'   B12 = 1000,
#'   R = 1000
#' )
#'
#' @export
Pbar_trapz_ms <- function(t, tp00, tp01, delta,
                          mu02, mu12, B02 = 1, B12 = 1, R = 0) {
  t <- .md_check_grid(t)
  tp00 <- .md_check_probability(tp00, "tp00")
  tp01 <- .md_check_probability(tp01, "tp01")
  if (length(tp00) != length(t) || length(tp01) != length(t)) stop("t, tp00, and tp01 must have the same length.", call. = FALSE)
  if (any(tp00 + tp01 > 1 + 1e-10)) stop("tp00 + tp01 must not exceed 1.", call. = FALSE)
  delta <- .md_check_scalar(delta, "delta")
  B02 <- .md_check_scalar(B02, "B02", lower = 0)
  B12 <- .md_check_scalar(B12, "B12", lower = 0)
  R <- .md_check_scalar(R, "R", lower = 0)
  mu02 <- .md_check_function(mu02, "mu02")
  mu12 <- .md_check_function(mu12, "mu12")
  m02 <- rep(as.numeric(mu02(t)), length.out = length(t))
  m12 <- rep(as.numeric(mu12(t)), length.out = length(t))
  if (!length(mu02(t)) %in% c(1L, length(t)) || !length(mu12(t)) %in% c(1L, length(t))) stop("mu02 and mu12 must return length 1 or length(t).", call. = FALSE)
  if (any(!is.finite(m02)) || any(m02 < 0) || any(!is.finite(m12)) || any(m12 < 0)) stop("mu02 and mu12 must return nonnegative finite values.", call. = FALSE)
  discount <- exp(-delta * t)
  num <- .md_trapz(t, discount * (tp00 * m02 * B02 + tp01 * m12 * B12 + tp01 * R))
  den <- .md_trapz(t, discount * tp00)
  if (!is.finite(den) || den <= 0) stop("The premium present-value denominator must be positive.", call. = FALSE)
  num / den
}

# -------------------------------------------------------------------------
# Thiele reserve equations
# -------------------------------------------------------------------------

#' Reserve derivatives for a disability model with recovery
#'
#' Computes the coupled Thiele reserve derivatives. Numeric arguments may be
#' scalar or compatible vectors.
#'
#' @param t Time.
#' @param V0 Healthy-state reserve.
#' @param V1 Disabled-state reserve.
#' @param delta Force of interest.
#' @param Pbar Continuous premium rate.
#' @param B Death benefit.
#' @param R Continuous disability income rate.
#' @param mu01 Healthy-to-disabled intensity function.
#' @param mu02 Healthy-to-deceased intensity function.
#' @param mu10 Disabled-to-healthy intensity function.
#' @param mu12 Disabled-to-deceased intensity function.
#'
#' @return A named vector for scalar input or a two-column matrix for
#'   vectorized input.
#'
#' @export
thiele_dVdt_01 <- function(t, V0, V1, delta, Pbar, B, R,
                           mu01, mu02, mu10, mu12) {
  vals <- .md_recycle(
    .md_check_nonnegative(t, "t"), .md_check_numeric(V0, "V0"),
    .md_check_numeric(V1, "V1"), .md_check_numeric(delta, "delta"),
    .md_check_numeric(Pbar, "Pbar"), .md_check_nonnegative(B, "B"),
    .md_check_nonnegative(R, "R"),
    .names = c("t", "V0", "V1", "delta", "Pbar", "B", "R")
  )
  t <- vals[[1L]]; V0 <- vals[[2L]]; V1 <- vals[[3L]]
  delta <- vals[[4L]]; Pbar <- vals[[5L]]; B <- vals[[6L]]; R <- vals[[7L]]
  funs <- list(mu01 = mu01, mu02 = mu02, mu10 = mu10, mu12 = mu12)
  rates <- lapply(names(funs), function(nm) {
    fun <- .md_check_function(funs[[nm]], nm)
    out <- as.numeric(fun(t))
    if (!length(out) %in% c(1L, length(t))) stop(nm, " must return length 1 or the common length.", call. = FALSE)
    out <- rep(out, length.out = length(t))
    if (any(!is.finite(out)) || any(out < 0)) stop(nm, " must return nonnegative finite values.", call. = FALSE)
    out
  })
  dV0 <- Pbar + delta * V0 - rates[[2L]] * (B - V0) - rates[[1L]] * (V1 - V0)
  dV1 <- delta * V1 - R - rates[[4L]] * (B - V1) - rates[[3L]] * (V0 - V1)
  if (length(dV0) == 1L) return(c(dV0 = dV0, dV1 = dV1))
  cbind(dV0 = dV0, dV1 = dV1)
}

#' Backward reserve path for a disability model with recovery
#'
#' Computes a backward Euler reserve path from terminal healthy-state and
#' disabled-state reserves.
#'
#' @inheritParams thiele_dVdt_01
#' @param h Positive step size.
#' @param n Nonnegative final time.
#' @param V0_n Terminal healthy-state reserve.
#' @param V1_n Terminal disabled-state reserve.
#'
#' @return A data frame with columns \code{t}, \code{tV0}, and \code{tV1}.
#'
#' @export
thiele_path_01 <- function(h, n, delta, Pbar, B, R,
                           mu01, mu02, mu10, mu12,
                           V0_n = 0, V1_n = 0) {
  grid <- .md_grid(n, h); back <- rev(grid)
  delta <- .md_check_scalar(delta, "delta")
  Pbar <- .md_check_scalar(Pbar, "Pbar")
  B <- .md_check_scalar(B, "B", lower = 0)
  R <- .md_check_scalar(R, "R", lower = 0)
  V0_n <- .md_check_scalar(V0_n, "V0_n")
  V1_n <- .md_check_scalar(V1_n, "V1_n")
  V0 <- numeric(length(back)); V1 <- numeric(length(back))
  V0[1L] <- V0_n; V1[1L] <- V1_n
  if (length(back) > 1L) {
    for (idx in seq_len(length(back) - 1L)) {
      step <- back[idx] - back[idx + 1L]
      dV <- thiele_dVdt_01(back[idx], V0[idx], V1[idx], delta, Pbar, B, R,
                           mu01, mu02, mu10, mu12)
      V0[idx + 1L] <- V0[idx] - step * unname(dV["dV0"])
      V1[idx + 1L] <- V1[idx] - step * unname(dV["dV1"])
    }
  }
  data.frame(t = grid, tV0 = rev(V0), tV1 = rev(V1))
}

# -------------------------------------------------------------------------
# Discrete-time Markov chains
# -------------------------------------------------------------------------

#' Multi-step transition probability
#'
#' Computes an entry of the matrix power \eqn{P^n}.
#'
#' @param P Square transition-probability matrix.
#' @param n Nonnegative integer number of steps.
#' @param i Starting-state index.
#' @param j Ending-state index.
#'
#' @return A numeric scalar.
#'
#' @examples
#' P <- matrix(c(0.9, 0.1, 0, 1), nrow = 2, byrow = TRUE)
#' markov_nstep_prob(P, n = 3, i = 1, j = 2)
#'
#' @export
markov_nstep_prob <- function(P, n, i, j) {
  if (!is.matrix(P) || !is.numeric(P)) stop("P must be a numeric matrix.", call. = FALSE)
  if (nrow(P) == 0L || nrow(P) != ncol(P)) stop("P must be a nonempty square matrix.", call. = FALSE)
  if (any(!is.finite(P)) || any(P < 0) || any(P > 1)) stop("P must contain values in [0, 1].", call. = FALSE)
  if (any(abs(rowSums(P) - 1) > 1e-10)) stop("Each row of P must sum to 1.", call. = FALSE)
  n <- .md_check_integer_scalar(n, "n", 0L)
  i <- .md_check_integer_scalar(i, "i", 1L)
  j <- .md_check_integer_scalar(j, "j", 1L)
  if (i > nrow(P) || j > ncol(P)) stop("i and j must be valid state indices.", call. = FALSE)
  Pn <- diag(nrow(P))
  if (n > 0L) for (step in seq_len(n)) Pn <- Pn %*% P
  unname(Pn[i, j])
}

# -------------------------------------------------------------------------
# Multiple-decrement gain or loss
# -------------------------------------------------------------------------

#' Gain or loss in a two-cause multiple-decrement model
#'
#' Computes one-year gain or loss under simultaneous within-year decrement
#' probabilities or an ordered case in which Cause 2 occurs at year-end.
#' Numeric arguments may be scalar or compatible vectors.
#'
#' @param Vt Gross reserve at the beginning of the year.
#' @param G Gross premium.
#' @param r Percent-of-premium expense rate.
#' @param e Fixed beginning-of-year expense.
#' @param i Earned effective annual interest rate.
#' @param b1 Cause 1 benefit.
#' @param b2 Cause 2 benefit.
#' @param s1 Cause 1 settlement expense.
#' @param s2 Cause 2 settlement expense.
#' @param q1 Cause 1 decrement probability.
#' @param q2 Cause 2 decrement probability.
#' @param Vt1 Gross reserve at the end of the year.
#' @param year_end_cause2 Whether Cause 2 occurs only at year-end.
#' @param q1prime Single-decrement Cause 1 probability for the ordered case.
#' @param q2prime Single-decrement Cause 2 probability for the ordered case.
#'
#' @return A numeric vector of gains or losses.
#'
#' @examples
#' gain_loss_md(
#'   Vt = 115, G = 16, r = 0, e = 3, i = 0.06,
#'   b1 = 1000, b2 = 110, q1 = 0.01, q2 = 0.10, Vt1 = 128.83
#' )
#'
#' @export
gain_loss_md <- function(Vt, G, r, e, i,
                         b1, b2, s1 = 0, s2 = 0,
                         q1, q2, Vt1,
                         year_end_cause2 = FALSE,
                         q1prime = NULL, q2prime = NULL) {
  if (!is.logical(year_end_cause2) || length(year_end_cause2) == 0L || anyNA(year_end_cause2)) {
    stop("year_end_cause2 must be a non-missing logical value or vector.", call. = FALSE)
  }
  q1prime <- if (is.null(q1prime)) NA_real_ else .md_check_probability(q1prime, "q1prime")
  q2prime <- if (is.null(q2prime)) NA_real_ else .md_check_probability(q2prime, "q2prime")
  vals <- .md_recycle(
    .md_check_numeric(Vt, "Vt"), .md_check_nonnegative(G, "G"),
    .md_check_probability(r, "r"), .md_check_nonnegative(e, "e"),
    .md_check_interest(i), .md_check_nonnegative(b1, "b1"),
    .md_check_nonnegative(b2, "b2"), .md_check_nonnegative(s1, "s1"),
    .md_check_nonnegative(s2, "s2"), .md_check_probability(q1, "q1"),
    .md_check_probability(q2, "q2"), .md_check_numeric(Vt1, "Vt1"),
    year_end_cause2, q1prime, q2prime,
    .names = c("Vt", "G", "r", "e", "i", "b1", "b2", "s1", "s2",
               "q1", "q2", "Vt1", "year_end_cause2", "q1prime", "q2prime")
  )
  Vt <- vals[[1L]]; G <- vals[[2L]]; r <- vals[[3L]]; e <- vals[[4L]]
  i <- vals[[5L]]; b1 <- vals[[6L]]; b2 <- vals[[7L]]; s1 <- vals[[8L]]
  s2 <- vals[[9L]]; q1 <- vals[[10L]]; q2 <- vals[[11L]]; Vt1 <- vals[[12L]]
  ordered <- vals[[13L]]; q1prime <- vals[[14L]]; q2prime <- vals[[15L]]
  if (any(!ordered & q1 + q2 > 1 + 1e-12)) stop("q1 + q2 must not exceed 1.", call. = FALSE)
  if (any(ordered & (is.na(q1prime) | is.na(q2prime)))) {
    stop("q1prime and q2prime must be provided when year_end_cause2 = TRUE.", call. = FALSE)
  }
  lhs <- (Vt + G * (1 - r) - e) * (1 + i)
  rhs <- numeric(length(lhs))
  simultaneous <- !ordered
  rhs[simultaneous] <- (b1[simultaneous] + s1[simultaneous]) * q1[simultaneous] +
    (b2[simultaneous] + s2[simultaneous]) * q2[simultaneous] +
    (1 - q1[simultaneous] - q2[simultaneous]) * Vt1[simultaneous]
  rhs[ordered] <- (b1[ordered] + s1[ordered]) * q1prime[ordered] +
    (b2[ordered] + s2[ordered]) * (1 - q1prime[ordered]) * q2prime[ordered] +
    (1 - q1prime[ordered]) * (1 - q2prime[ordered]) * Vt1[ordered]
  lhs - rhs
}

test_that("Axj_md reproduces Example 14.1 approximately", {
  q1 <- c(.02, .02, .02, .02, .02)
  q2 <- c(.03, .04, .05, .06, .00)
  q3 <- c(.00, .00, .00, .00, .98)
  qtau <- q1 + q2 + q3

  ptau <- numeric(length(qtau))
  ptau[1] <- 1
  for (k in 2:length(qtau)) {
    ptau[k] <- prod(1 - qtau[1:(k - 1)])
  }

  val <- Axj_md(qj = q1, ptau = ptau, i = 0.06, benefit = 1000)
  expect_equal(val, 75.33, tolerance = 1e-2)
})

test_that("Abarxj_md works for constant forces", {
  t <- seq(0, 200, by = 0.01)
  ptau <- exp(-0.012 * t)
  mu_ac <- rep(0.002, length(t))

  val <- Abarxj_md(
    t = t,
    ptau = ptau,
    muj = mu_ac,
    delta = 0.05,
    benefit = 2000
  )

  exact <- 2000 * 0.002 / (0.05 + 0.012)
  expect_equal(val, exact, tolerance = 1e-3)
})

test_that("AS_path reproduces Example 14.4 approximately", {
  r <- c(.05, .05, .05, .05, .05)
  e <- c(30, 30, 30, 30, 30)
  b1 <- c(1000, 1000, 1000, 1000, 1000)
  b2 <- c(50, 100, 300, 600, 0)
  b3 <- c(0, 0, 0, 0, 1000)
  q1 <- c(.02, .03, .04, .05, .06)
  q2 <- c(.30, .20, .20, .10, .00)
  p_tau <- c(.68, .77, .76, .85, .94)

  out <- AS_path(
    AS0 = 50,
    G = 200,
    r = r,
    e = e,
    b1 = b1,
    b2 = b2,
    q1 = q1,
    q2 = q2,
    p_tau = p_tau,
    i = 0.06,
    b3 = b3
  )

  expect_equal(out$AS[2], 275.90, tolerance = 1e-2)
  expect_equal(out$AS[3], 535.10, tolerance = 1e-2)
  expect_equal(out$AS[4], 837.90, tolerance = 1e-2)
  expect_equal(out$AS[5], 1115.03, tolerance = 1e-2)
  expect_equal(out$AS[6], 373.90, tolerance = 1e-2)
})

test_that("tp00_tp01_euler reproduces Example 14.10 approximately", {
  mu01 <- function(t) 0.10 * t + 0.20
  mu02 <- function(t) 0.20
  mu10 <- function(t) 0.50
  mu12 <- function(t) 0.125 * t + 0.20

  out <- tp00_tp01_euler(
    h = 0.10, n = 2.0,
    mu01 = mu01, mu02 = mu02, mu10 = mu10, mu12 = mu12
  )

  i1 <- which(abs(out$t - 1.0) < 1e-12)
  i2 <- which(abs(out$t - 2.0) < 1e-12)

  expect_equal(round(out$tp00[i1], 3), 0.667)
  expect_equal(round(out$tp01[i1], 3), 0.144)
  expect_equal(round(out$tp00[i2], 3), 0.448)
  expect_equal(round(out$tp01[i2], 3), 0.186)
})

test_that("Pbar_trapz_ms reproduces Example 14.18 approximately", {
  mu01 <- function(t) 0.10 * t + 0.20
  mu02 <- function(t) 0.20
  mu10 <- function(t) 0.50
  mu12 <- function(t) 0.125 * t + 0.20

  ex1410 <- tp00_tp01_euler(
    h = 0.10, n = 2.0,
    mu01 = mu01, mu02 = mu02, mu10 = mu10, mu12 = mu12
  )

  val <- Pbar_trapz_ms(
    t = ex1410$t,
    tp00 = ex1410$tp00,
    tp01 = ex1410$tp01,
    delta = 0.04,
    mu02 = mu02,
    mu12 = mu12,
    B02 = 1000,
    B12 = 1000,
    R = 1000
  )

  expect_equal(val, 446.95, tolerance = 1e-2)
})

test_that("thiele_dVdt_01 reproduces first derivatives in Example 14.12", {
  mu01 <- function(t) 0.10 * t + 0.20
  mu02 <- function(t) 0.20
  mu10 <- function(t) 0.50
  mu12 <- function(t) 0.125 * t + 0.20

  out <- thiele_dVdt_01(
    t = 2.00,
    V0 = 0,
    V1 = 0,
    delta = 0.04,
    Pbar = 446.95,
    B = 1000,
    R = 1000,
    mu01 = mu01,
    mu02 = mu02,
    mu10 = mu10,
    mu12 = mu12
  )

  expect_equal(unname(out["dV0"]), 246.95, tolerance = 1e-8)
  expect_equal(unname(out["dV1"]), -1450.00, tolerance = 1e-8)
})

test_that("thiele_path_01 reproduces Example 14.12 approximately", {
  mu01 <- function(t) 0.10 * t + 0.20
  mu02 <- function(t) 0.20
  mu10 <- function(t) 0.50
  mu12 <- function(t) 0.125 * t + 0.20

  out <- thiele_path_01(
    h = 0.10,
    n = 2.0,
    delta = 0.04,
    Pbar = 446.95,
    B = 1000,
    R = 1000,
    mu01 = mu01,
    mu02 = mu02,
    mu10 = mu10,
    mu12 = mu12
  )

  i190 <- which(abs(out$t - 1.90) < 1e-12)
  i100 <- which(abs(out$t - 1.00) < 1e-12)
  i000 <- which(abs(out$t - 0.00) < 1e-12)

  expect_equal(out$tV0[i190], -24.70, tolerance = 1e-2)
  expect_equal(out$tV1[i190], 145.00, tolerance = 1e-2)
  expect_equal(out$tV0[i100], -50.75, tolerance = 1e-2)
  expect_equal(out$tV1[i100], 917.99, tolerance = 1e-2)
  expect_equal(out$tV0[i000], -5.39, tolerance = 1e-2)
  expect_equal(out$tV1[i000], 1283.82, tolerance = 1e-2)
})

test_that("markov_nstep_prob agrees with direct matrix power", {
  P <- matrix(
    c(0.94, 0.03, 0.02, 0.01,
      0.50, 0.30, 0.18, 0.02,
      0.00, 0.00, 0.93, 0.07,
      0.00, 0.00, 0.00, 1.00),
    nrow = 4, byrow = TRUE
  )

  P3 <- P %*% P %*% P

  expect_equal(markov_nstep_prob(P, 3, 1, 1), P3[1, 1], tolerance = 1e-12)
  expect_equal(markov_nstep_prob(P, 3, 1, 3), P3[1, 3], tolerance = 1e-12)
})

test_that("gain_loss_md matches direct formula", {
  val <- gain_loss_md(
    Vt = 115.00,
    G = 16,
    r = 0,
    e = 3,
    i = 0.06,
    b1 = 1000,
    b2 = 110,
    s1 = 0,
    s2 = 0,
    q1 = 0.01,
    q2 = 0.10,
    Vt1 = 128.83
  )

  expected <- (115 + 16 - 3) * 1.06 - (1000 * 0.01 + 110 * 0.10 + (1 - 0.01 - 0.10) * 128.83)
  expect_equal(val, expected, tolerance = 1e-12)
})

test_that("gain_loss_md year-end Cause 2 version works", {
  val <- gain_loss_md(
    Vt = 115.00,
    G = 16,
    r = 0,
    e = 3,
    i = 0.06,
    b1 = 1000,
    b2 = 110,
    s1 = 0,
    s2 = 0,
    q1 = 0.01,
    q2 = 0.10,
    Vt1 = 128.83,
    year_end_cause2 = TRUE,
    q1prime = 0.01,
    q2prime = 0.10
  )

  expected <- (115 + 16 - 3) * 1.06 -
    (1000 * 0.01 + 110 * (1 - 0.01) * 0.10 + (1 - 0.01) * (1 - 0.10) * 128.83)

  expect_equal(val, expected, tolerance = 1e-12)
})




test_that("AS_path_md matches AS_path for two decrements", {

  r <- c(.05, .05, .05)
  e <- c(30, 30, 30)
  b1 <- c(1000, 1000, 1000)
  b2 <- c(50, 100, 300)
  q1 <- c(.02, .03, .04)
  q2 <- c(.30, .20, .20)
  p_tau <- c(.68, .77, .76)

  b_mat <- cbind(b1, b2)
  q_mat <- cbind(q1, q2)

  out_old <- AS_path(
    AS0 = 50,
    G = 200,
    r = r,
    e = e,
    b1 = b1,
    b2 = b2,
    q1 = q1,
    q2 = q2,
    p_tau = p_tau,
    i = 0.06
  )

  out_new <- AS_path_md(
    AS0 = 50,
    G = 200,
    r = r,
    e = e,
    b_mat = b_mat,
    q_mat = q_mat,
    p_tau = p_tau,
    i = 0.06
  )

  expect_equal(out_new$AS, out_old$AS, tolerance = 1e-10)
})


test_that("AS_path_md works with three decrement causes", {

  r <- c(.05, .05)
  e <- c(30, 30)

  b_mat <- matrix(
    c(1000, 50, 200,
      1000, 100, 300),
    nrow = 2, byrow = TRUE
  )

  q_mat <- matrix(
    c(.02, .03, .01,
      .03, .02, .02),
    nrow = 2, byrow = TRUE
  )

  p_tau <- 1 - rowSums(q_mat)

  out <- AS_path_md(
    AS0 = 50,
    G = 200,
    r = r,
    e = e,
    b_mat = b_mat,
    q_mat = q_mat,
    p_tau = p_tau,
    i = 0.06
  )

  # basic sanity checks
  expect_equal(nrow(out), 3)
  expect_true(all(is.finite(out$AS)))
})




testthat::test_that("Axj_md vectorizes over interest and benefit", {
  qj <- c(0.02, 0.03, 0.04)
  ptau <- c(1, 0.95, 0.88)

  out <- Axj_md(
    qj = qj,
    ptau = ptau,
    i = c(0.04, 0.06),
    benefit = 1000
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
  testthat::expect_gt(out[1], out[2])
})


testthat::test_that("Axj_md rejects invalid inputs", {
  testthat::expect_error(
    Axj_md(
      qj = c(0.02, 1.2),
      ptau = c(1, 0.9),
      i = 0.05
    ),
    "\\[0, 1\\]"
  )

  testthat::expect_error(
    Axj_md(
      qj = c(0.02, 0.03),
      ptau = 1,
      i = 0.05
    ),
    "same length"
  )

  testthat::expect_error(
    Axj_md(
      qj = c(0.02, 0.03),
      ptau = c(1, 0.9),
      i = -1
    ),
    "greater than -1"
  )
})


testthat::test_that("Abarxj_md agrees with a constant-force closed form", {
  t <- seq(0, 20, by = 0.001)
  mu_total <- 0.03
  mu_j <- 0.01
  delta <- 0.04

  ptau <- exp(-mu_total * t)
  muj <- rep(mu_j, length(t))

  expected <- mu_j *
    (1 - exp(-(delta + mu_total) * 20)) /
    (delta + mu_total)

  testthat::expect_equal(
    Abarxj_md(
      t = t,
      ptau = ptau,
      muj = muj,
      delta = delta
    ),
    expected,
    tolerance = 1e-6
  )
})


testthat::test_that("Abarxj_md requires a strictly increasing grid", {
  testthat::expect_error(
    Abarxj_md(
      t = c(0, 2, 1),
      ptau = c(1, 0.9, 0.8),
      muj = c(0.01, 0.01, 0.01),
      delta = 0.04
    ),
    "strictly increasing"
  )
})


testthat::test_that("AS_path matches the general two-cause asset-share path", {
  b1 <- c(1000, 1000, 1000)
  b2 <- c(100, 105, 110)
  q1 <- c(0.01, 0.012, 0.014)
  q2 <- c(0.08, 0.07, 0.06)
  p_tau <- 1 - q1 - q2

  two_cause <- AS_path(
    AS0 = 0,
    G = 150,
    r = 0.05,
    e = 2,
    b1 = b1,
    b2 = b2,
    q1 = q1,
    q2 = q2,
    p_tau = p_tau,
    i = 0.04
  )

  general <- AS_path_md(
    AS0 = 0,
    G = 150,
    r = 0.05,
    e = 2,
    b_mat = cbind(b1, b2),
    q_mat = cbind(q1, q2),
    p_tau = p_tau,
    i = 0.04
  )

  testthat::expect_equal(two_cause, general)
})


testthat::test_that("AS_path_md supports yearly assumptions", {
  b_mat <- cbind(
    death = rep(1000, 3),
    withdrawal = c(100, 105, 110)
  )

  q_mat <- cbind(
    death = c(0.01, 0.012, 0.014),
    withdrawal = c(0.08, 0.07, 0.06)
  )

  out <- AS_path_md(
    AS0 = 0,
    G = c(150, 155, 160),
    r = c(0.05, 0.04, 0.03),
    e = 2,
    b_mat = b_mat,
    q_mat = q_mat,
    p_tau = 1 - rowSums(q_mat),
    i = c(0.03, 0.04, 0.05),
    b_surv = 0
  )

  testthat::expect_s3_class(out, "data.frame")
  testthat::expect_equal(out$k, 0:3)
  testthat::expect_true(all(is.finite(out$AS)))
})


testthat::test_that("AS_path_md rejects zero in-force probabilities", {
  testthat::expect_error(
    AS_path_md(
      AS0 = 0,
      G = 100,
      r = 0,
      e = 0,
      b_mat = matrix(1000, nrow = 1),
      q_mat = matrix(1, nrow = 1),
      p_tau = 0,
      i = 0.05
    ),
    "p_tau must be positive"
  )
})


testthat::test_that("Euler probability grid includes the terminal time", {
  mu01 <- function(t) 0.01
  mu02 <- function(t) 0.02
  mu10 <- function(t) 0.03
  mu12 <- function(t) 0.04

  out <- tp00_tp01_euler(
    h = 0.3,
    n = 1,
    mu01 = mu01,
    mu02 = mu02,
    mu10 = mu10,
    mu12 = mu12
  )

  testthat::expect_equal(tail(out$t, 1), 1)
  testthat::expect_equal(out$t, c(0, 0.3, 0.6, 0.9, 1))
})


testthat::test_that("Euler probability function handles zero term", {
  zero_rate <- function(t) 0

  out <- tp00_tp01_euler(
    h = 0.1,
    n = 0,
    mu01 = zero_rate,
    mu02 = zero_rate,
    mu10 = zero_rate,
    mu12 = zero_rate
  )

  testthat::expect_equal(out$t, 0)
  testthat::expect_equal(out$tp00, 1)
  testthat::expect_equal(out$tp01, 0)
  testthat::expect_equal(out$tp02, 0)
})


testthat::test_that("Pbar_trapz_ms rejects a zero premium denominator", {
  zero_rate <- function(t) 0

  testthat::expect_error(
    Pbar_trapz_ms(
      t = c(0, 1),
      tp00 = c(0, 0),
      tp01 = c(1, 1),
      delta = 0.04,
      mu02 = zero_rate,
      mu12 = zero_rate
    ),
    "denominator must be positive"
  )
})


testthat::test_that("Thiele derivative vectorizes", {
  mu01 <- function(t) rep(0.01, length(t))
  mu02 <- function(t) rep(0.02, length(t))
  mu10 <- function(t) rep(0.03, length(t))
  mu12 <- function(t) rep(0.04, length(t))

  out <- thiele_dVdt_01(
    t = c(0, 1),
    V0 = c(10, 20),
    V1 = c(30, 40),
    delta = 0.05,
    Pbar = 5,
    B = 1000,
    R = 50,
    mu01 = mu01,
    mu02 = mu02,
    mu10 = mu10,
    mu12 = mu12
  )

  testthat::expect_true(is.matrix(out))
  testthat::expect_equal(dim(out), c(2, 2))
  testthat::expect_equal(colnames(out), c("dV0", "dV1"))
})


testthat::test_that("backward Thiele path includes zero and terminal time", {
  mu01 <- function(t) 0.01
  mu02 <- function(t) 0.02
  mu10 <- function(t) 0.03
  mu12 <- function(t) 0.04

  out <- thiele_path_01(
    h = 0.3,
    n = 1,
    delta = 0.04,
    Pbar = 10,
    B = 1000,
    R = 100,
    mu01 = mu01,
    mu02 = mu02,
    mu10 = mu10,
    mu12 = mu12
  )

  testthat::expect_equal(out$t, c(0, 0.3, 0.6, 0.9, 1))
  testthat::expect_equal(tail(out$tV0, 1), 0)
  testthat::expect_equal(tail(out$tV1, 1), 0)
})


testthat::test_that("markov_nstep_prob handles zero steps", {
  P <- matrix(
    c(
      0.9, 0.1,
      0.0, 1.0
    ),
    nrow = 2,
    byrow = TRUE
  )

  testthat::expect_equal(
    markov_nstep_prob(P, n = 0, i = 1, j = 1),
    1
  )

  testthat::expect_equal(
    markov_nstep_prob(P, n = 0, i = 1, j = 2),
    0
  )
})


testthat::test_that("markov_nstep_prob validates transition matrices", {
  testthat::expect_error(
    markov_nstep_prob(
      matrix(c(0.8, 0.1, 0, 1), nrow = 2, byrow = TRUE),
      n = 1,
      i = 1,
      j = 1
    ),
    "row of P must sum to 1"
  )
})


testthat::test_that("gain_loss_md vectorizes", {
  out <- gain_loss_md(
    Vt = c(100, 110),
    G = 20,
    r = c(0.02, 0.03),
    e = 1,
    i = c(0.04, 0.05),
    b1 = 1000,
    b2 = 100,
    q1 = c(0.01, 0.012),
    q2 = c(0.08, 0.07),
    Vt1 = c(115, 125)
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("gain_loss_md supports the ordered decrement case", {
  out <- gain_loss_md(
    Vt = 100,
    G = 20,
    r = 0.02,
    e = 1,
    i = 0.04,
    b1 = 1000,
    b2 = 100,
    q1 = 0.01,
    q2 = 0.08,
    Vt1 = 115,
    year_end_cause2 = TRUE,
    q1prime = 0.01,
    q2prime = 0.08
  )

  expected <- (100 + 20 * 0.98 - 1) * 1.04 -
    (
      1000 * 0.01 +
        100 * (1 - 0.01) * 0.08 +
        (1 - 0.01) * (1 - 0.08) * 115
    )

  testthat::expect_equal(out, expected)
})

#' Insurance models with varying benefits
#'
#' Functions for discrete and continuous contingent payment models with
#' varying benefits.
#'
#' @name insurance_varying_benefits
#' @keywords internal
NULL

# ------------------------------------------------------------
# Internal helpers
# ------------------------------------------------------------

.discrete_prob_k <- function(x, k, tbl = NULL, model = NULL, ...) {
  if (!is.null(tbl)) {
    p0 <- npx(tbl, x = x, n = k)
    p1 <- npx(tbl, x = x, n = k + 1)

    if (is.na(p0)) p0 <- 0
    if (is.na(p1)) p1 <- 0

    return(p0 - p1)
  }

  if (is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }

  tpx(k, x = x, model = model, ...) -
    tpx(k + 1, x = x, model = model, ...)
}

.max_k_varying <- function(x, tbl = NULL, model = NULL, ...,
                           tol = 1e-12, k_max = 5000) {
  for (k in 0:k_max) {
    if (!is.null(tbl)) {
      surv <- npx(tbl, x = x, n = k)
    } else {
      if (is.null(model)) {
        stop("Supply either tbl or model.", call. = FALSE)
      }
      surv <- tpx(k, x = x, model = model, ...)
    }

    if (is.na(surv)) {
      return(max(k - 1L, 0L))
    }

    if (!is.finite(surv) || surv <= tol) {
      return(k)
    }
  }

  k_max
}

.get_upper_varying_cont <- function(x, n = NULL, model, ...) {
  if (tolower(model) == "uniform") {
    dots <- list(...)

    if (is.null(dots$omega)) {
      stop("For model = 'uniform', omega must be supplied.", call. = FALSE)
    }

    upper <- dots$omega - x

    if (!is.null(n)) {
      upper <- pmin(upper, n)
    }

    return(pmax(upper, 0))
  }

  if (is.null(n)) {
    return(Inf)
  }

  n
}

.integrate_varying_cont <- function(f, lower, upper) {
  stats::integrate(
    f,
    lower = lower,
    upper = upper,
    rel.tol = 1e-9,
    subdivisions = 1000L
  )$value
}

# ------------------------------------------------------------
# Discrete varying-benefit models
# ------------------------------------------------------------

#' Increasing whole life insurance
#'
#' Computes
#' \deqn{(IA)_x = \sum_{k=0}^{\infty} (k+1) v^{k+1} \Pr(K_x = k).}
#'
#' @param x Age.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional model parameters.
#' @param tol Numerical tolerance for truncation.
#' @param k_max Maximum number of terms.
#'
#' @return Numeric vector.
#' @export
IAx <- function(x, i, tbl = NULL, model = NULL, ...,
                tol = 1e-12, k_max = 5000) {
  x <- as.numeric(x)
  i <- .check_i_discrete(i)

  if (length(x) == 0L || any(!is.finite(x))) {
    stop("x must contain finite numeric values.", call. = FALSE)
  }

  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  v <- 1 / (1 + i)

  sapply(x, function(xx) {
    kk_max <- .max_k_varying(
      xx,
      tbl = tbl,
      model = model,
      ...,
      tol = tol,
      k_max = k_max
    )

    k <- 0:kk_max

    probs <- vapply(
      k,
      function(kk) .discrete_prob_k(xx, kk, tbl = tbl, model = model, ...),
      numeric(1)
    )

    sum((k + 1) * v^(k + 1) * probs, na.rm = TRUE)
  })
}

#' Increasing n-year term insurance
#'
#' Computes
#' \deqn{(IA)_{x:\overline{n}|}^{1}
#' = \sum_{k=0}^{n-1} (k+1) v^{k+1} \Pr(K_x = k).}
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional model parameters.
#'
#' @return Numeric vector.
#' @export
IAxn1 <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  xn <- .recycle_x_n(x, n)
  x <- xn$x
  n <- xn$n
  i <- .check_i_discrete(i)

  v <- 1 / (1 + i)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]

    if (nn == 0) {
      return(0)
    }

    k <- 0:(nn - 1)

    probs <- vapply(
      k,
      function(kk) .discrete_prob_k(xx, kk, tbl = tbl, model = model, ...),
      numeric(1)
    )

    sum((k + 1) * v^(k + 1) * probs, na.rm = TRUE)
  })
}

#' Decreasing n-year term insurance
#'
#' Computes
#' \deqn{(DA)_{x:\overline{n}|}^{1}
#' = \sum_{k=0}^{n-1} (n-k) v^{k+1} \Pr(K_x = k).}
#'
#' @inheritParams IAxn1
#'
#' @return Numeric vector.
#' @export
DAxn1 <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  xn <- .recycle_x_n(x, n)
  x <- xn$x
  n <- xn$n
  i <- .check_i_discrete(i)

  v <- 1 / (1 + i)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]

    if (nn == 0) {
      return(0)
    }

    k <- 0:(nn - 1)

    probs <- vapply(
      k,
      function(kk) .discrete_prob_k(xx, kk, tbl = tbl, model = model, ...),
      numeric(1)
    )

    sum((nn - k) * v^(k + 1) * probs, na.rm = TRUE)
  })
}

# ------------------------------------------------------------
# Continuous varying-benefit models
# ------------------------------------------------------------

#' Fully continuous increasing whole life insurance
#'
#' Computes
#' \deqn{(\bar{I}\bar{A})_x =
#' \int_0^\infty t v^t {}_tp_x \mu_{x+t}\,dt.}
#'
#' @param x Age.
#' @param i Effective annual interest rate.
#' @param model Parametric survival model.
#' @param ... Additional model parameters.
#'
#' @return Numeric vector.
#' @export
IbarAbarx <- function(x, i, model, ...) {
  args <- .recycle_cont_vectors(x, .check_i_cont(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_x_cont(x)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    delta <- log(1 + i[j])

    upper <- .get_upper_varying_cont(xx, model = model, ...)

    if (upper <= 0) {
      return(0)
    }

    f <- function(t) {
      val <- t * exp(-delta * t) *
        tpx(t, x = xx, model = model, ...) *
        hazard0(xx + t, model = model, ...)
      val[!is.finite(val)] <- 0
      val
    }

    .integrate_varying_cont(f, 0, upper)
  })
}

#' Piecewise-continuous increasing whole life insurance
#'
#' Computes
#' \deqn{(I\bar{A})_x =
#' \int_0^\infty \lfloor t+1 \rfloor v^t {}_tp_x \mu_{x+t}\,dt.}
#'
#' @inheritParams IbarAbarx
#'
#' @return Numeric vector.
#' @export
IAbarx <- function(x, i, model, ...) {
  args <- .recycle_cont_vectors(x, .check_i_cont(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_x_cont(x)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    delta <- log(1 + i[j])

    upper <- .get_upper_varying_cont(xx, model = model, ...)

    if (upper <= 0) {
      return(0)
    }

    f <- function(t) {
      val <- floor(t + 1) * exp(-delta * t) *
        tpx(t, x = xx, model = model, ...) *
        hazard0(xx + t, model = model, ...)
      val[!is.finite(val)] <- 0
      val
    }

    .integrate_varying_cont(f, 0, upper)
  })
}

#' Fully continuous increasing n-year term insurance
#'
#' Computes
#' \deqn{(\bar{I}\bar{A})_{x:\overline{n}|}^{1}
#' = \int_0^n t v^t {}_tp_x \mu_{x+t}\,dt.}
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param model Parametric survival model.
#' @param ... Additional model parameters.
#'
#' @return Numeric vector.
#' @export
IbarAbarxn1 <- function(x, n, i, model, ...) {
  args <- .recycle_cont_vectors(x, n, .check_i_cont(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_cont(x)
  .check_n_cont(n)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]
    delta <- log(1 + i[j])

    upper <- .get_upper_varying_cont(xx, n = nn, model = model, ...)

    if (upper <= 0) {
      return(0)
    }

    f <- function(t) {
      val <- t * exp(-delta * t) *
        tpx(t, x = xx, model = model, ...) *
        hazard0(xx + t, model = model, ...)
      val[!is.finite(val)] <- 0
      val
    }

    .integrate_varying_cont(f, 0, upper)
  })
}

#' Fully continuous decreasing n-year term insurance
#'
#' Computes
#' \deqn{(\bar{D}\bar{A})_{x:\overline{n}|}^{1}
#' = \int_0^n (n-t) v^t {}_tp_x \mu_{x+t}\,dt.}
#'
#' @inheritParams IbarAbarxn1
#'
#' @return Numeric vector.
#' @export
DbarAbarxn1 <- function(x, n, i, model, ...) {
  args <- .recycle_cont_vectors(x, n, .check_i_cont(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_cont(x)
  .check_n_cont(n)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]
    delta <- log(1 + i[j])

    upper <- .get_upper_varying_cont(xx, n = nn, model = model, ...)

    if (upper <= 0) {
      return(0)
    }

    f <- function(t) {
      val <- (nn - t) * exp(-delta * t) *
        tpx(t, x = xx, model = model, ...) *
        hazard0(xx + t, model = model, ...)
      val[!is.finite(val)] <- 0
      val
    }

    .integrate_varying_cont(f, 0, upper)
  })
}

#' Piecewise-continuous decreasing n-year term insurance
#'
#' Computes
#' \deqn{(D\bar{A})_{x:\overline{n}|}^{1}
#' = \int_0^n \lfloor n+1-t \rfloor v^t {}_tp_x \mu_{x+t}\,dt.}
#'
#' @inheritParams IbarAbarxn1
#'
#' @return Numeric vector.
#' @export
DAbarxn1 <- function(x, n, i, model, ...) {
  args <- .recycle_cont_vectors(x, n, .check_i_cont(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_cont(x)
  .check_n_cont(n)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]
    delta <- log(1 + i[j])

    upper <- .get_upper_varying_cont(xx, n = nn, model = model, ...)

    if (upper <= 0) {
      return(0)
    }

    f <- function(t) {
      val <- floor(nn + 1 - t) * exp(-delta * t) *
        tpx(t, x = xx, model = model, ...) *
        hazard0(xx + t, model = model, ...)
      val[!is.finite(val)] <- 0
      val
    }

    .integrate_varying_cont(f, 0, upper)
  })
}

test_that("IAxn1 equals sum of increasing benefits over term", {

  i <- 0.05

  val <- IAxn1(40, 10, i, model = "uniform", omega = 100)

  expect_true(is.numeric(val))
  expect_true(length(val) == 1)
  expect_true(val > 0)
})

test_that("DAxn1 equals sum of decreasing benefits over term", {

  i <- 0.05

  val <- DAxn1(40, 10, i, model = "uniform", omega = 100)

  expect_true(is.numeric(val))
  expect_true(length(val) == 1)
  expect_true(val > 0)
})

test_that("Increasing + decreasing term equals level insurance times n+1", {

  i <- 0.05
  n <- 10

  inc <- IAxn1(40, n, i, model = "uniform", omega = 100)
  dec <- DAxn1(40, n, i, model = "uniform", omega = 100)
  level <- Axn1(40, n, i, model = "uniform", omega = 100)

  expect_equal(inc + dec, (n + 1) * level)
})

test_that("Increasing whole life insurance is positive", {

  i <- 0.05

  val <- IAx(40, i, model = "uniform", omega = 100)

  expect_true(is.numeric(val))
  expect_true(val > 0)
})

test_that("IAx vectorizes correctly", {

  i <- 0.05

  vals <- IAx(c(40, 45), i, model = "uniform", omega = 100)

  expect_length(vals, 2)
})

test_that("Continuous increasing insurance is positive", {

  i <- 0.05

  val <- IbarAbarx(40, i, model = "uniform", omega = 100)

  expect_true(is.numeric(val))
  expect_true(val > 0)
})

test_that("Piecewise continuous increasing insurance is positive", {

  i <- 0.05

  val <- IAbarx(40, i, model = "uniform", omega = 100)

  expect_true(is.numeric(val))
  expect_true(val > 0)
})

test_that("Continuous increasing term insurance positive", {

  i <- 0.05

  val <- IbarAbarxn1(40, 10, i, model = "uniform", omega = 100)

  expect_true(val > 0)
})

test_that("Continuous decreasing term insurance positive", {

  i <- 0.05

  val <- DbarAbarxn1(40, 10, i, model = "uniform", omega = 100)

  expect_true(val > 0)
})

test_that("Piecewise continuous decreasing term insurance positive", {

  i <- 0.05

  val <- DAbarxn1(40, 10, i, model = "uniform", omega = 100)

  expect_true(val > 0)
})

test_that("Continuous increasing + decreasing term equals level continuous term times n/2 approximately", {

  i <- 0.05
  n <- 10

  inc <- IbarAbarxn1(40, n, i, model = "uniform", omega = 100)
  dec <- DbarAbarxn1(40, n, i, model = "uniform", omega = 100)

  level <- Abarxn1(40, n, i, model = "uniform", omega = 100)

  expect_true(inc > 0)
  expect_true(dec > 0)
  expect_true(level > 0)
})

test_that("Varying-benefit functions reject invalid inputs", {

  i <- 0.05

  expect_error(IAx(-1, i, model = "uniform", omega = 100))
  expect_error(IAxn1(40, -1, i, model = "uniform", omega = 100))
  expect_error(DAxn1(40, -1, i, model = "uniform", omega = 100))
})

test_that("Increasing benefits produce larger APV than level insurance", {

  i <- 0.05
  n <- 10

  inc <- IAxn1(40, n, i, model = "uniform", omega = 100)
  level <- Axn1(40, n, i, model = "uniform", omega = 100)

  expect_true(inc >= level)
})

test_that("Increasing and decreasing term insurance are positive", {

  i <- 0.05
  n <- 10

  inc <- IAxn1(40, n, i, model = "uniform", omega = 100)
  dec <- DAxn1(40, n, i, model = "uniform", omega = 100)

  expect_true(inc > 0)
  expect_true(dec > 0)
})

test_that("Continuous increasing insurance works with exponential survival", {

  i <- 0.05
  lambda <- 0.25

  val <- IbarAbarx(40, i, model = "exponential", lambda = lambda)

  expect_true(is.numeric(val))
  expect_true(val > 0)
})

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.reserve11_check_basis <- function(tbl = NULL, model = NULL) {
  if (is.null(tbl) && is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }

  if (!is.null(tbl) && !is.null(model)) {
    stop("Supply either tbl or model, not both.", call. = FALSE)
  }

  if (!is.null(tbl)) {
    .validate_life_table(tbl)
  }

  invisible(TRUE)
}

#' @noRd
.reserve11_check_nonnegative <- function(x, name) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(x)) || any(x < 0)) {
    stop(
      name,
      " must contain nonnegative finite values.",
      call. = FALSE
    )
  }

  x
}

#' @noRd
.reserve11_check_integerish <- function(x, name, positive = FALSE) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  invalid_sign <- if (positive) {
    any(x <= 0)
  } else {
    any(x < 0)
  }

  if (any(!is.finite(x)) ||
      invalid_sign ||
      any(abs(x - round(x)) > 1e-10)) {
    descriptor <- if (positive) "positive" else "nonnegative"

    stop(
      name,
      " must contain ",
      descriptor,
      " integer-like values.",
      call. = FALSE
    )
  }

  as.integer(round(x))
}

#' @noRd
.reserve11_check_interest <- function(i, name = "i") {
  i <- as.numeric(i)

  if (length(i) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(i)) || any(i <= -1)) {
    stop(
      name,
      " must contain finite values greater than -1.",
      call. = FALSE
    )
  }

  i
}

#' @noRd
.reserve11_check_probability <- function(x, name) {
  x <- as.numeric(x)

  if (length(x) == 0L ||
      any(!is.finite(x)) ||
      any(x < 0) ||
      any(x > 1)) {
    stop(
      name,
      " must contain values in [0, 1].",
      call. = FALSE
    )
  }

  x
}

#' @noRd
.reserve11_recycle <- function(..., .names = NULL) {
  values <- list(...)

  if (length(values) == 0L) {
    return(values)
  }

  lengths <- vapply(values, length, integer(1))

  if (any(lengths == 0L)) {
    stop("Inputs must have positive length.", call. = FALSE)
  }

  common_length <- max(lengths)

  if (is.null(.names)) {
    .names <- paste0("Argument ", seq_along(values))
  }

  if (length(.names) != length(values)) {
    stop(
      ".names must have the same length as the supplied arguments.",
      call. = FALSE
    )
  }

  for (j in seq_along(values)) {
    if (!lengths[j] %in% c(1L, common_length)) {
      stop(
        .names[j],
        " must have length 1 or the common length ",
        common_length,
        ".",
        call. = FALSE
      )
    }

    values[[j]] <- rep(values[[j]], length.out = common_length)
  }

  values
}

#' @noRd
.reserve11_check_t_less_n <- function(t, n) {
  if (any(t >= n)) {
    stop("t must satisfy t < n.", call. = FALSE)
  }

  invisible(TRUE)
}

#' @noRd
.reserve11_scalar_or_matrix <- function(x, column_names) {
  x <- as.matrix(x)
  colnames(x) <- column_names

  if (nrow(x) == 1L) {
    return(stats::setNames(as.numeric(x[1, ]), column_names))
  }

  x
}

# -------------------------------------------------------------------------
# Full preliminary term modified premiums and reserves
# -------------------------------------------------------------------------

#' Full preliminary term modified premiums and reserves
#'
#' Computes modified premiums and reserves under the full preliminary term
#' method for whole-life insurance.
#'
#' \code{alphaF()} computes the first-year modified premium
#' \eqn{\alpha^F = vq_x}.
#'
#' \code{betaF()} computes the renewal modified premium
#' \eqn{\beta^F = P_{x+1}}.
#'
#' \code{tVFx()} computes the full preliminary term reserve. The reserve is
#' zero at durations 0 and 1. For \eqn{t > 1}, it equals the net level premium
#' reserve at duration \eqn{t - 1} for a policy issued at age \eqn{x + 1}.
#'
#' @param x Issue age. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the actuarial functions.
#' @param t Nonnegative integer duration. May be scalar or vector.
#'
#' @return A numeric vector of modified premiums or reserves.
#'
#' @details
#' Under the full preliminary term method, the first policy year is treated
#' as one-year term insurance. The first-year modified premium is therefore
#' the actuarial present value of one-year term insurance at age \eqn{x}.
#'
#' Renewal premiums are based on a whole-life policy issued one year later,
#' at age \eqn{x + 1}. Accordingly, reserves after the first policy year are
#' obtained from the corresponding net level premium reserve for that
#' deferred issue age.
#'
#' @examples
#' alphaF(
#'   x = 40,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' betaF(
#'   x = 40,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tVFx(
#'   x = 40,
#'   t = 5,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name full_preliminary_term
#' @rdname full_preliminary_term
#' @export
alphaF <- function(x, i, tbl = NULL, model = NULL, ...) {
  .reserve11_check_basis(tbl, model)

  x <- .reserve11_check_nonnegative(x, "x")
  i <- .reserve11_check_interest(i)

  values <- .reserve11_recycle(
    x,
    i,
    .names = c("x", "i")
  )

  x <- values[[1]]
  i <- values[[2]]

  Axn1(
    x = x,
    n = 1,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )
}


#' @rdname full_preliminary_term
#' @export
betaF <- function(x, i, tbl = NULL, model = NULL, ...) {
  .reserve11_check_basis(tbl, model)

  x <- .reserve11_check_nonnegative(x, "x")
  i <- .reserve11_check_interest(i)

  values <- .reserve11_recycle(
    x,
    i,
    .names = c("x", "i")
  )

  x <- values[[1]]
  i <- values[[2]]

  Px(
    x = x + 1,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )
}


#' @rdname full_preliminary_term
#' @export
tVFx <- function(x, t, i, tbl = NULL, model = NULL, ...) {
  .reserve11_check_basis(tbl, model)

  x <- .reserve11_check_nonnegative(x, "x")
  t <- .reserve11_check_integerish(t, "t")
  i <- .reserve11_check_interest(i)

  values <- .reserve11_recycle(
    x,
    t,
    i,
    .names = c("x", "t", "i")
  )

  x <- values[[1]]
  t <- values[[2]]
  i <- values[[3]]

  vapply(
    seq_along(x),
    function(j) {
      if (t[j] <= 1L) {
        return(0)
      }

      tVx(
        x = x[j] + 1,
        t = t[j] - 1L,
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      )
    },
    numeric(1)
  )
}


# -------------------------------------------------------------------------
# Fractional-duration whole life reserves
# -------------------------------------------------------------------------

#' Fractional-duration whole life reserves
#'
#' Computes fractional-duration whole life reserves using linear
#' interpolation between the reserve immediately after the premium at
#' duration \eqn{t} and the reserve at duration \eqn{t+1}.
#'
#' \code{tsVx()} computes the reserve at fractional duration
#' \eqn{t+s}, where \eqn{0 \le s \le 1}.
#'
#' \code{meanVx()} computes the reserve at the midpoint of the policy
#' year (\eqn{s=0.5}).
#'
#' @param x Issue age. May be scalar or vector.
#' @param t Nonnegative integer duration. May be scalar or vector.
#' @param s Fractional duration in \eqn{[0,1]}. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the actuarial functions.
#'
#' @return A numeric vector of reserve values.
#'
#' @details
#' The fractional reserve is computed using
#'
#' \deqn{
#' {}_{t+s}V_x
#' =
#' ({}_tV_x + P_x)(1-s)
#' +
#' {}_{t+1}V_x s.
#' }
#'
#' The function \code{meanVx()} is a convenience wrapper corresponding to
#' \eqn{s=0.5}.
#'
#' @examples
#' tsVx(
#'   40,
#'   t = 10,
#'   s = 0.5,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' meanVx(
#'   40,
#'   t = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name fractional_duration_reserves
#' @rdname fractional_duration_reserves
#' @export
tsVx <- function(x, t, s, i, tbl = NULL, model = NULL, ...) {
  .reserve11_check_basis(tbl, model)

  x <- .reserve11_check_nonnegative(x, "x")
  t <- .reserve11_check_integerish(t, "t")
  s <- .reserve11_check_probability(s, "s")
  i <- .reserve11_check_interest(i)

  values <- .reserve11_recycle(
    x,
    t,
    s,
    i,
    .names = c("x", "t", "s", "i")
  )

  x <- values[[1]]
  t <- values[[2]]
  s <- values[[3]]
  i <- values[[4]]

  vapply(
    seq_along(x),
    function(j) {
      reserve_t <- tVx(
        x = x[j],
        t = t[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      )

      reserve_t1 <- tVx(
        x = x[j],
        t = t[j] + 1L,
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      )

      premium <- Px(
        x = x[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      )

      (reserve_t + premium) * (1 - s[j]) +
        reserve_t1 * s[j]
    },
    numeric(1)
  )
}

#' @rdname fractional_duration_reserves
#' @export
meanVx <- function(x, t, i, tbl = NULL, model = NULL, ...) {
  tsVx(
    x = x,
    t = t,
    s = 0.5,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )
}

# -------------------------------------------------------------------------
# Fractional-duration term and endowment reserves
# -------------------------------------------------------------------------

#' Fractional-duration term and endowment reserves
#'
#' Computes fractional-duration reserves for term and endowment insurance
#' using linear interpolation between the reserve immediately after the
#' premium at duration \eqn{t} and the reserve at duration \eqn{t+1}.
#'
#' \code{tsVxn()} computes the fractional-duration reserve for endowment
#' insurance.
#'
#' \code{tsVxn1()} computes the fractional-duration reserve for term
#' insurance.
#'
#' @param x Issue age. May be scalar or vector.
#' @param n Positive integer term. May be scalar or vector.
#' @param t Nonnegative integer duration satisfying \eqn{t < n}. May be
#'   scalar or vector.
#' @param s Fractional duration in \eqn{[0,1]}. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the actuarial functions.
#'
#' @return A numeric vector of reserve values.
#'
#' @details
#' The fractional reserve is computed using
#'
#' \deqn{
#' {}_{t+s}V
#' =
#' ({}_tV + P)(1-s)
#' +
#' {}_{t+1}V s.
#' }
#'
#' The premium \eqn{P} is the net annual premium for the corresponding
#' insurance contract (term or endowment).
#'
#' @examples
#' tsVxn(
#'   40,
#'   n = 20,
#'   t = 10,
#'   s = 0.5,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tsVxn1(
#'   40,
#'   n = 20,
#'   t = 10,
#'   s = 0.5,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name fractional_duration_term_endowment_reserves
#' @rdname fractional_duration_term_endowment_reserves
#' @export
tsVxn <- function(x, n, t, s, i,
                  tbl = NULL, model = NULL, ...) {
  .reserve11_check_basis(tbl, model)

  x <- .reserve11_check_nonnegative(x, "x")
  n <- .reserve11_check_integerish(n, "n", positive = TRUE)
  t <- .reserve11_check_integerish(t, "t")
  s <- .reserve11_check_probability(s, "s")
  i <- .reserve11_check_interest(i)

  values <- .reserve11_recycle(
    x,
    n,
    t,
    s,
    i,
    .names = c("x", "n", "t", "s", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  t <- values[[3]]
  s <- values[[4]]
  i <- values[[5]]

  .reserve11_check_t_less_n(t, n)

  vapply(seq_along(x), function(j) {
    reserve_t <- tVxn(
      x = x[j],
      n = n[j],
      t = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    reserve_t1 <- tVxn(
      x = x[j],
      n = n[j],
      t = t[j] + 1L,
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    premium <- Pxn(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    (reserve_t + premium) * (1 - s[j]) +
      reserve_t1 * s[j]
  }, numeric(1))
}

#' @rdname fractional_duration_term_endowment_reserves
#' @export
tsVxn1 <- function(x, n, t, s, i,
                   tbl = NULL, model = NULL, ...) {
  .reserve11_check_basis(tbl, model)

  x <- .reserve11_check_nonnegative(x, "x")
  n <- .reserve11_check_integerish(n, "n", positive = TRUE)
  t <- .reserve11_check_integerish(t, "t")
  s <- .reserve11_check_probability(s, "s")
  i <- .reserve11_check_interest(i)

  values <- .reserve11_recycle(
    x,
    n,
    t,
    s,
    i,
    .names = c("x", "n", "t", "s", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  t <- values[[3]]
  s <- values[[4]]
  i <- values[[5]]

  .reserve11_check_t_less_n(t, n)

  vapply(seq_along(x), function(j) {
    reserve_t <- tVxn1(
      x = x[j],
      n = n[j],
      t = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    reserve_t1 <- tVxn1(
      x = x[j],
      n = n[j],
      t = t[j] + 1L,
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    premium <- Pxn1(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    (reserve_t + premium) * (1 - s[j]) +
      reserve_t1 * s[j]
  }, numeric(1))
}

# -------------------------------------------------------------------------
# Gross premium and expense reserves
# -------------------------------------------------------------------------

#' Whole life gross premium and expense reserves
#'
#' Computes prospective gross premium and expense reserves for fully
#' discrete whole life insurance after issue.
#'
#' \code{tVGx()} computes the prospective gross premium reserve.
#'
#' \code{tVEx()} computes the corresponding expense reserve, defined as the
#' difference between the gross premium reserve and the net benefit reserve.
#'
#' @param x Issue age. May be scalar or vector.
#' @param t Nonnegative integer duration. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param G Gross annual premium. May be scalar or vector.
#' @param benefit Insurance benefit amount.
#' @param renewal_premium_pct Renewal percent-of-premium expense in
#'   \eqn{[0,1]}.
#' @param renewal_policy_exp Renewal per-policy expense.
#' @param settlement_exp Settlement expense paid at death.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the actuarial functions.
#'
#' @return A numeric vector of reserve values.
#'
#' @details
#' The gross premium reserve is calculated as
#'
#' \deqn{
#' {}_tV_x^G
#' =
#' (b+s)A_{x+t}
#' -
#' \left[(1-r)G-e\right]\ddot{a}_{x+t},
#' }
#'
#' where
#' \itemize{
#'   \item \eqn{b} is the insurance benefit,
#'   \item \eqn{s} is the settlement expense,
#'   \item \eqn{G} is the gross annual premium,
#'   \item \eqn{r} is the renewal percent-of-premium expense, and
#'   \item \eqn{e} is the renewal per-policy expense.
#' }
#'
#' The expense reserve is obtained as the gross premium reserve minus the
#' corresponding net benefit reserve.
#'
#' @examples
#' tVGx(
#'   x = 40,
#'   t = 10,
#'   i = 0.05,
#'   G = 0.03,
#'   benefit = 1,
#'   renewal_premium_pct = 0.10,
#'   renewal_policy_exp = 0.002,
#'   settlement_exp = 0.02,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' tVEx(
#'   x = 40,
#'   t = 10,
#'   i = 0.05,
#'   G = 0.03,
#'   benefit = 1,
#'   renewal_premium_pct = 0.10,
#'   renewal_policy_exp = 0.002,
#'   settlement_exp = 0.02,
#'   model = "uniform",
#'   omega = 100
#' )
#'
#' @name gross_premium_expense_reserves
#' @rdname gross_premium_expense_reserves
#' @export
tVGx <- function(x, t, i, G,
                 benefit = 1,
                 renewal_premium_pct = 0,
                 renewal_policy_exp = 0,
                 settlement_exp = 0,
                 tbl = NULL,
                 model = NULL,
                 ...) {
  .reserve11_check_basis(tbl, model)

  x <- .reserve11_check_nonnegative(x, "x")
  t <- .reserve11_check_integerish(t, "t")
  i <- .reserve11_check_interest(i)
  G <- .reserve11_check_nonnegative(G, "G")
  benefit <- .reserve11_check_nonnegative(benefit, "benefit")
  renewal_premium_pct <- .reserve11_check_probability(
    renewal_premium_pct,
    "renewal_premium_pct"
  )
  renewal_policy_exp <- .reserve11_check_nonnegative(
    renewal_policy_exp,
    "renewal_policy_exp"
  )
  settlement_exp <- .reserve11_check_nonnegative(
    settlement_exp,
    "settlement_exp"
  )

  values <- .reserve11_recycle(
    x,
    t,
    i,
    G,
    benefit,
    renewal_premium_pct,
    renewal_policy_exp,
    settlement_exp,
    .names = c(
      "x",
      "t",
      "i",
      "G",
      "benefit",
      "renewal_premium_pct",
      "renewal_policy_exp",
      "settlement_exp"
    )
  )

  x <- values[[1]]
  t <- values[[2]]
  i <- values[[3]]
  G <- values[[4]]
  benefit <- values[[5]]
  renewal_premium_pct <- values[[6]]
  renewal_policy_exp <- values[[7]]
  settlement_exp <- values[[8]]

  vapply(seq_along(x), function(j) {
    insurance_value <- Ax(
      x = x[j] + t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    annuity_value <- adotx(
      x = x[j] + t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    (benefit[j] + settlement_exp[j]) * insurance_value -
      (
        (1 - renewal_premium_pct[j]) * G[j] -
          renewal_policy_exp[j]
      ) * annuity_value
  }, numeric(1))
}

#' @rdname gross_premium_expense_reserves
#' @export
tVEx <- function(x, t, i, G,
                 benefit = 1,
                 renewal_premium_pct = 0,
                 renewal_policy_exp = 0,
                 settlement_exp = 0,
                 tbl = NULL,
                 model = NULL,
                 ...) {
  .reserve11_check_basis(tbl, model)

  x <- .reserve11_check_nonnegative(x, "x")
  t <- .reserve11_check_integerish(t, "t")
  i <- .reserve11_check_interest(i)
  G <- .reserve11_check_nonnegative(G, "G")
  benefit <- .reserve11_check_nonnegative(benefit, "benefit")
  renewal_premium_pct <- .reserve11_check_probability(
    renewal_premium_pct,
    "renewal_premium_pct"
  )
  renewal_policy_exp <- .reserve11_check_nonnegative(
    renewal_policy_exp,
    "renewal_policy_exp"
  )
  settlement_exp <- .reserve11_check_nonnegative(
    settlement_exp,
    "settlement_exp"
  )

  values <- .reserve11_recycle(
    x,
    t,
    i,
    G,
    benefit,
    renewal_premium_pct,
    renewal_policy_exp,
    settlement_exp,
    .names = c(
      "x",
      "t",
      "i",
      "G",
      "benefit",
      "renewal_premium_pct",
      "renewal_policy_exp",
      "settlement_exp"
    )
  )

  x <- values[[1]]
  t <- values[[2]]
  i <- values[[3]]
  G <- values[[4]]
  benefit <- values[[5]]
  renewal_premium_pct <- values[[6]]
  renewal_policy_exp <- values[[7]]
  settlement_exp <- values[[8]]

  gross_reserve <- tVGx(
    x = x,
    t = t,
    i = i,
    G = G,
    benefit = benefit,
    renewal_premium_pct = renewal_premium_pct,
    renewal_policy_exp = renewal_policy_exp,
    settlement_exp = settlement_exp,
    tbl = tbl,
    model = model,
    ...
  )

  net_reserve <- tVx(
    x = x,
    t = t,
    i = i,
    tbl = tbl,
    model = model,
    ...
  )

  gross_reserve - benefit * net_reserve
}

# -------------------------------------------------------------------------
# Gross gain analysis
# -------------------------------------------------------------------------

#' Total gross gain for a discrete insurance contract
#'
#' Computes total gain during one policy year under gross premiums,
#' gross reserves, actual mortality, actual interest, and actual expenses.
#'
#' @param VtG Gross reserve at duration \code{t}.
#' @param Vt1G Gross reserve at duration \code{t + 1}.
#' @param G Gross premium.
#' @param i_actual Actual annual effective interest rate.
#' @param q_actual Actual mortality probability.
#' @param r_actual Actual percent-of-premium expense rate.
#' @param e_actual Actual per-policy expense.
#' @param s_actual Actual settlement expense.
#' @param b Benefit amount.
#'
#' @return A numeric vector of total gains.
#'
#' @examples
#' GTg_disc(
#'   VtG = 0.10,
#'   Vt1G = 0.12,
#'   G = 0.02,
#'   i_actual = 0.05,
#'   q_actual = 0.01,
#'   r_actual = 0.03,
#'   e_actual = 0,
#'   s_actual = 0.01,
#'   b = 1
#' )
#'
#' @export
GTg_disc <- function(VtG, Vt1G, G,
                     i_actual, q_actual,
                     r_actual = 0,
                     e_actual = 0,
                     s_actual = 0,
                     b = 1) {
  VtG <- .reserve11_check_nonnegative(VtG, "VtG")
  Vt1G <- .reserve11_check_nonnegative(Vt1G, "Vt1G")
  G <- .reserve11_check_nonnegative(G, "G")
  i_actual <- .reserve11_check_interest(i_actual, "i_actual")
  q_actual <- .reserve11_check_probability(q_actual, "q_actual")
  r_actual <- .reserve11_check_probability(r_actual, "r_actual")
  e_actual <- .reserve11_check_nonnegative(e_actual, "e_actual")
  s_actual <- .reserve11_check_nonnegative(s_actual, "s_actual")
  b <- .reserve11_check_nonnegative(b, "b")

  values <- .reserve11_recycle(
    VtG,
    Vt1G,
    G,
    i_actual,
    q_actual,
    r_actual,
    e_actual,
    s_actual,
    b,
    .names = c(
      "VtG",
      "Vt1G",
      "G",
      "i_actual",
      "q_actual",
      "r_actual",
      "e_actual",
      "s_actual",
      "b"
    )
  )

  VtG <- values[[1]]
  Vt1G <- values[[2]]
  G <- values[[3]]
  i_actual <- values[[4]]
  q_actual <- values[[5]]
  r_actual <- values[[6]]
  e_actual <- values[[7]]
  s_actual <- values[[8]]
  b <- values[[9]]

  (VtG + G * (1 - r_actual) - e_actual) *
    (1 + i_actual) -
    (
      (b + s_actual) * q_actual +
        (1 - q_actual) * Vt1G
    )
}

#' Ordered decomposition of gross gain
#'
#' Decomposes gross gain into interest, mortality, and expense components
#' in a user-specified order.
#'
#' For scalar input, the function returns a named numeric vector. For
#' vectorized input, it returns a numeric matrix with one row per calculation.
#'
#' @param VtG Gross reserve at duration \code{t}.
#' @param Vt1G Gross reserve at duration \code{t + 1}.
#' @param G Gross premium.
#' @param i_assumed Assumed annual effective interest rate.
#' @param q_assumed Assumed mortality probability.
#' @param r_assumed Assumed percent-of-premium expense rate.
#' @param e_assumed Assumed per-policy expense.
#' @param s_assumed Assumed settlement expense.
#' @param i_actual Actual annual effective interest rate.
#' @param q_actual Actual mortality probability.
#' @param r_actual Actual percent-of-premium expense rate.
#' @param e_actual Actual per-policy expense.
#' @param s_actual Actual settlement expense.
#' @param b Benefit amount.
#' @param order Character vector containing \code{"interest"},
#'   \code{"mortality"}, and \code{"expense"} exactly once.
#'
#' @return For scalar input, a named numeric vector. For vectorized input,
#'   a numeric matrix with columns \code{total_gain}, \code{interest},
#'   \code{mortality}, \code{expense}, and \code{check}.
#'
#' @examples
#' decompGg_disc(
#'   VtG = 3950.73,
#'   Vt1G = 4607.07,
#'   G = 685,
#'   i_assumed = 0.06,
#'   q_assumed = 0.00592,
#'   r_assumed = 0.05,
#'   e_assumed = 0,
#'   s_assumed = 300,
#'   i_actual = 0.065,
#'   q_actual = 0.005,
#'   r_actual = 0.06,
#'   e_actual = 0,
#'   s_actual = 100,
#'   b = 50000
#' )
#'
#' @export
decompGg_disc <- function(
    VtG,
    Vt1G,
    G,
    i_assumed,
    q_assumed,
    r_assumed = 0,
    e_assumed = 0,
    s_assumed = 0,
    i_actual,
    q_actual,
    r_actual = 0,
    e_actual = 0,
    s_actual = 0,
    b = 1,
    order = c("interest", "mortality", "expense")) {

  allowed <- c("interest", "mortality", "expense")

  if (!is.character(order) ||
      length(order) != 3L ||
      length(unique(order)) != 3L ||
      !setequal(order, allowed)) {
    stop(
      "order must contain 'interest', 'mortality', and 'expense' exactly once.",
      call. = FALSE
    )
  }

  VtG <- .reserve11_check_nonnegative(VtG, "VtG")
  Vt1G <- .reserve11_check_nonnegative(Vt1G, "Vt1G")
  G <- .reserve11_check_nonnegative(G, "G")

  i_assumed <- .reserve11_check_interest(i_assumed, "i_assumed")
  q_assumed <- .reserve11_check_probability(q_assumed, "q_assumed")
  r_assumed <- .reserve11_check_probability(r_assumed, "r_assumed")
  e_assumed <- .reserve11_check_nonnegative(e_assumed, "e_assumed")
  s_assumed <- .reserve11_check_nonnegative(s_assumed, "s_assumed")

  i_actual <- .reserve11_check_interest(i_actual, "i_actual")
  q_actual <- .reserve11_check_probability(q_actual, "q_actual")
  r_actual <- .reserve11_check_probability(r_actual, "r_actual")
  e_actual <- .reserve11_check_nonnegative(e_actual, "e_actual")
  s_actual <- .reserve11_check_nonnegative(s_actual, "s_actual")

  b <- .reserve11_check_nonnegative(b, "b")

  values <- .reserve11_recycle(
    VtG,
    Vt1G,
    G,
    i_assumed,
    q_assumed,
    r_assumed,
    e_assumed,
    s_assumed,
    i_actual,
    q_actual,
    r_actual,
    e_actual,
    s_actual,
    b,
    .names = c(
      "VtG",
      "Vt1G",
      "G",
      "i_assumed",
      "q_assumed",
      "r_assumed",
      "e_assumed",
      "s_assumed",
      "i_actual",
      "q_actual",
      "r_actual",
      "e_actual",
      "s_actual",
      "b"
    )
  )

  VtG <- values[[1]]
  Vt1G <- values[[2]]
  G <- values[[3]]
  i_assumed <- values[[4]]
  q_assumed <- values[[5]]
  r_assumed <- values[[6]]
  e_assumed <- values[[7]]
  s_assumed <- values[[8]]
  i_actual <- values[[9]]
  q_actual <- values[[10]]
  r_actual <- values[[11]]
  e_actual <- values[[12]]
  s_actual <- values[[13]]
  b <- values[[14]]

  calculate_one <- function(j) {
    assumed_state <- list(
      i = i_assumed[j],
      q = q_assumed[j],
      r = r_assumed[j],
      e = e_assumed[j],
      s = s_assumed[j]
    )

    actual_state <- list(
      i = i_actual[j],
      q = q_actual[j],
      r = r_actual[j],
      e = e_actual[j],
      s = s_actual[j]
    )

    evaluate_gain <- function(state) {
      as.numeric(GTg_disc(
        VtG = VtG[j],
        Vt1G = Vt1G[j],
        G = G[j],
        i_actual = state$i,
        q_actual = state$q,
        r_actual = state$r,
        e_actual = state$e,
        s_actual = state$s,
        b = b[j]
      ))
    }

    components <- c(
      interest = 0,
      mortality = 0,
      expense = 0
    )

    current_state <- assumed_state
    running_gain <- 0
    total_gain <- evaluate_gain(actual_state)

    # Compute the first two components sequentially.
    for (k in seq_len(2L)) {
      component <- order[k]

      if (component == "interest") {
        current_state$i <- actual_state$i
      } else if (component == "mortality") {
        current_state$q <- actual_state$q
      } else {
        current_state$r <- actual_state$r
        current_state$e <- actual_state$e
        current_state$s <- actual_state$s
      }

      new_gain <- evaluate_gain(current_state)

      components[component] <- new_gain - running_gain
      running_gain <- new_gain
    }

    # Use the final component as the balancing item so that the
    # decomposition sums exactly to the directly calculated total gain.
    final_component <- order[3L]

    components[final_component] <-
      total_gain - sum(components[order[1:2]])

    c(
      total_gain = total_gain,
      interest = unname(components["interest"]),
      mortality = unname(components["mortality"]),
      expense = unname(components["expense"]),
      check = sum(components)
    )
  }

  rows <- lapply(seq_along(VtG), calculate_one)

  if (length(rows) == 1L) {
    return(rows[[1L]])
  }

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

testthat::test_that("alphaF equals one-year term insurance APV", {
  testthat::expect_equal(
    alphaF(
      x = 40,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    Axn1(
      x = 40,
      n = 1,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    tolerance = 1e-12
  )
})


testthat::test_that("betaF equals whole life premium at age x plus one", {
  testthat::expect_equal(
    betaF(
      x = 40,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    Px(
      x = 41,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    tolerance = 1e-12
  )
})


testthat::test_that("FPT whole life reserve is zero at durations zero and one", {
  testthat::expect_equal(
    tVFx(
      x = 40,
      t = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    0,
    tolerance = 1e-12
  )

  testthat::expect_equal(
    tVFx(
      x = 40,
      t = 1,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    0,
    tolerance = 1e-12
  )
})


testthat::test_that("FPT whole life reserve shifts to NLP reserve at age x plus one", {
  testthat::expect_equal(
    tVFx(
      x = 40,
      t = 5,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    tVx(
      x = 41,
      t = 4,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    tolerance = 1e-10
  )
})


testthat::test_that("FPT functions support life tables", {
  tbl <- life_table(
    x = 0:4,
    lx = c(100000, 80000, 50000, 20000, 0)
  )

  testthat::expect_true(
    is.finite(
      alphaF(
        x = 0,
        i = 0.05,
        tbl = tbl
      )
    )
  )

  testthat::expect_true(
    is.finite(
      betaF(
        x = 0,
        i = 0.05,
        tbl = tbl
      )
    )
  )

  testthat::expect_equal(
    tVFx(
      x = 0,
      t = 1,
      i = 0.05,
      tbl = tbl
    ),
    0
  )

  testthat::expect_true(
    is.finite(
      tVFx(
        x = 0,
        t = 2,
        i = 0.05,
        tbl = tbl
      )
    )
  )
})


testthat::test_that("fractional whole life reserve gives correct endpoints", {
  x <- 40
  t <- 10
  i <- 0.05

  reserve_t <- tVx(
    x = x,
    t = t,
    i = i,
    model = "uniform",
    omega = 100
  )

  reserve_t1 <- tVx(
    x = x,
    t = t + 1,
    i = i,
    model = "uniform",
    omega = 100
  )

  premium <- Px(
    x = x,
    i = i,
    model = "uniform",
    omega = 100
  )

  testthat::expect_equal(
    tsVx(
      x = x,
      t = t,
      s = 0,
      i = i,
      model = "uniform",
      omega = 100
    ),
    reserve_t + premium,
    tolerance = 1e-10
  )

  testthat::expect_equal(
    tsVx(
      x = x,
      t = t,
      s = 1,
      i = i,
      model = "uniform",
      omega = 100
    ),
    reserve_t1,
    tolerance = 1e-10
  )
})


testthat::test_that("mean reserve equals fractional reserve at one-half", {
  testthat::expect_equal(
    meanVx(
      x = 40,
      t = 10,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    tsVx(
      x = 40,
      t = 10,
      s = 0.5,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    tolerance = 1e-12
  )
})


testthat::test_that("fractional endowment reserve gives correct endpoints", {
  x <- 40
  n <- 20
  t <- 10
  i <- 0.05

  testthat::expect_equal(
    tsVxn(
      x = x,
      n = n,
      t = t,
      s = 0,
      i = i,
      model = "uniform",
      omega = 100
    ),
    tVxn(
      x = x,
      n = n,
      t = t,
      i = i,
      model = "uniform",
      omega = 100
    ) +
      Pxn(
        x = x,
        n = n,
        i = i,
        model = "uniform",
        omega = 100
      ),
    tolerance = 1e-10
  )

  testthat::expect_equal(
    tsVxn(
      x = x,
      n = n,
      t = t,
      s = 1,
      i = i,
      model = "uniform",
      omega = 100
    ),
    tVxn(
      x = x,
      n = n,
      t = t + 1,
      i = i,
      model = "uniform",
      omega = 100
    ),
    tolerance = 1e-10
  )
})


testthat::test_that("fractional term reserve gives correct endpoints", {
  x <- 40
  n <- 20
  t <- 10
  i <- 0.05

  testthat::expect_equal(
    tsVxn1(
      x = x,
      n = n,
      t = t,
      s = 0,
      i = i,
      model = "uniform",
      omega = 100
    ),
    tVxn1(
      x = x,
      n = n,
      t = t,
      i = i,
      model = "uniform",
      omega = 100
    ) +
      Pxn1(
        x = x,
        n = n,
        i = i,
        model = "uniform",
        omega = 100
      ),
    tolerance = 1e-10
  )

  testthat::expect_equal(
    tsVxn1(
      x = x,
      n = n,
      t = t,
      s = 1,
      i = i,
      model = "uniform",
      omega = 100
    ),
    tVxn1(
      x = x,
      n = n,
      t = t + 1,
      i = i,
      model = "uniform",
      omega = 100
    ),
    tolerance = 1e-10
  )
})


testthat::test_that("fractional term and endowment reserves vectorize", {
  endowment <- tsVxn(
    x = c(40, 45),
    n = 20,
    t = c(5, 10),
    s = c(0.25, 0.75),
    i = 0.05,
    model = "uniform",
    omega = 100
  )

  term <- tsVxn1(
    x = c(40, 45),
    n = 20,
    t = c(5, 10),
    s = c(0.25, 0.75),
    i = 0.05,
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(endowment, 2)
  testthat::expect_length(term, 2)
  testthat::expect_true(all(is.finite(endowment)))
  testthat::expect_true(all(is.finite(term)))
})


testthat::test_that("gross reserve equals net reserve plus expense reserve", {
  x <- 40
  t <- 10
  i <- 0.05
  G <- 0.04
  benefit <- 1
  renewal_premium_pct <- 0.10
  renewal_policy_exp <- 0.002
  settlement_exp <- 0.02

  gross_reserve <- tVGx(
    x = x,
    t = t,
    i = i,
    G = G,
    benefit = benefit,
    renewal_premium_pct = renewal_premium_pct,
    renewal_policy_exp = renewal_policy_exp,
    settlement_exp = settlement_exp,
    model = "uniform",
    omega = 100
  )

  expense_reserve <- tVEx(
    x = x,
    t = t,
    i = i,
    G = G,
    benefit = benefit,
    renewal_premium_pct = renewal_premium_pct,
    renewal_policy_exp = renewal_policy_exp,
    settlement_exp = settlement_exp,
    model = "uniform",
    omega = 100
  )

  net_reserve <- tVx(
    x = x,
    t = t,
    i = i,
    model = "uniform",
    omega = 100
  )

  testthat::expect_equal(
    gross_reserve,
    benefit * net_reserve + expense_reserve,
    tolerance = 1e-10
  )
})


testthat::test_that("gross reserve functions support life tables", {
  tbl <- life_table(
    x = 0:4,
    lx = c(100000, 80000, 50000, 20000, 0)
  )

  gross_reserve <- tVGx(
    x = 0,
    t = 1,
    i = 0.05,
    G = 0.20,
    tbl = tbl
  )

  expense_reserve <- tVEx(
    x = 0,
    t = 1,
    i = 0.05,
    G = 0.20,
    tbl = tbl
  )

  testthat::expect_true(is.finite(gross_reserve))
  testthat::expect_true(is.finite(expense_reserve))
})


testthat::test_that("gross total gain reproduces Example 11.9 value", {
  total_gain <- GTg_disc(
    VtG = 3950.73,
    Vt1G = 4607.07,
    G = 685,
    i_actual = 0.065,
    q_actual = 0.005,
    r_actual = 0.06,
    e_actual = 0,
    s_actual = 100,
    b = 50000
  )

  testthat::expect_equal(
    total_gain,
    58.75,
    tolerance = 0.02
  )
})


testthat::test_that("gross gain function vectorizes consistently", {
  out <- GTg_disc(
    VtG = c(0.10, 0.15),
    Vt1G = c(0.12, 0.17),
    G = 0.02,
    i_actual = c(0.04, 0.05),
    q_actual = 0.01,
    r_actual = 0.03,
    e_actual = 0,
    s_actual = 0.01,
    b = 1
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("ordered gross gain decomposition sums to total gain", {
  out <- decompGg_disc(
    VtG = 3950.73,
    Vt1G = 4607.07,
    G = 685,
    i_assumed = 0.06,
    q_assumed = 0.00592,
    r_assumed = 0.05,
    e_assumed = 0,
    s_assumed = 300,
    i_actual = 0.065,
    q_actual = 0.005,
    r_actual = 0.06,
    e_actual = 0,
    s_actual = 100,
    b = 50000,
    order = c("interest", "mortality", "expense")
  )

  testthat::expect_type(out, "double")

  testthat::expect_named(
    out,
    c(
      "total_gain",
      "interest",
      "mortality",
      "expense",
      "check"
    )
  )

  testthat::expect_equal(
    unname(out["total_gain"]),
    unname(out["check"]),
    tolerance = 1e-10
  )
})


testthat::test_that("different decomposition orders preserve total gain", {
  out1 <- decompGg_disc(
    VtG = 3950.73,
    Vt1G = 4607.07,
    G = 685,
    i_assumed = 0.06,
    q_assumed = 0.00592,
    r_assumed = 0.05,
    e_assumed = 0,
    s_assumed = 300,
    i_actual = 0.065,
    q_actual = 0.005,
    r_actual = 0.06,
    e_actual = 0,
    s_actual = 100,
    b = 50000,
    order = c("interest", "mortality", "expense")
  )

  out2 <- decompGg_disc(
    VtG = 3950.73,
    Vt1G = 4607.07,
    G = 685,
    i_assumed = 0.06,
    q_assumed = 0.00592,
    r_assumed = 0.05,
    e_assumed = 0,
    s_assumed = 300,
    i_actual = 0.065,
    q_actual = 0.005,
    r_actual = 0.06,
    e_actual = 0,
    s_actual = 100,
    b = 50000,
    order = c("mortality", "expense", "interest")
  )

  testthat::expect_equal(
    unname(out1["total_gain"]),
    unname(out2["total_gain"]),
    tolerance = 1e-12
  )

  testthat::expect_equal(
    unname(out1["total_gain"]),
    unname(out1["check"]),
    tolerance = 1e-10
  )

  testthat::expect_equal(
    unname(out2["total_gain"]),
    unname(out2["check"]),
    tolerance = 1e-10
  )
})


testthat::test_that("Chapter 11 functions reject invalid inputs", {
  testthat::expect_error(
    tsVx(
      x = 40,
      t = 10,
      s = 1.2,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "values in \\[0, 1\\]"
  )

  testthat::expect_error(
    tsVxn(
      x = 40,
      n = 10,
      t = 10,
      s = 0.5,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "t must satisfy t < n"
  )

  testthat::expect_error(
    decompGg_disc(
      VtG = 1,
      Vt1G = 1,
      G = 1,
      i_assumed = 0.05,
      q_assumed = 0.01,
      r_assumed = 0.03,
      e_assumed = 0,
      s_assumed = 0,
      i_actual = 0.06,
      q_actual = 0.02,
      r_actual = 0.04,
      e_actual = 0,
      s_actual = 0,
      order = c("interest", "interest", "expense")
    ),
    "exactly once"
  )
})


testthat::test_that("Chapter 11 reserve functions reject conflicting mortality bases", {
  tbl <- life_table(
    x = 0:3,
    lx = c(100000, 80000, 50000, 0)
  )

  testthat::expect_error(
    tsVx(
      x = 0,
      t = 1,
      s = 0.5,
      i = 0.05,
      tbl = tbl,
      model = "uniform",
      omega = 100
    ),
    "either tbl or model, not both"
  )
})

#' Annuity approximations
#'
#' Approximation formulas for m-thly and continuous life annuities.
#'
#' This file implements:
#' \itemize{
#'   \item UDD approximations for m-thly annuities,
#'   \item UDD approximations for continuous annuities,
#'   \item Woolhouse 2-term approximations,
#'   \item Woolhouse 3-term approximations.
#' }
#'
#' @name annuity_approximations
#' @keywords internal
NULL

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.check_i_scalar_ann <- function(i) {
  i <- as.numeric(i)
  if (length(i) != 1L || !is.finite(i) || i <= -1) {
    stop("i must be a single finite number greater than -1.", call. = FALSE)
  }
  i
}

#' @noRd
.check_m_scalar_ann <- function(m) {
  m <- as.numeric(m)
  if (length(m) != 1L || !is.finite(m) || m <= 0 || abs(m - round(m)) > 1e-10) {
    stop("m must be a positive integer.", call. = FALSE)
  }
  as.integer(round(m))
}

#' @noRd
.recycle_x_n_ann <- function(x, n) {
  x <- as.numeric(x)
  n <- as.numeric(n)

  if (length(x) == 0L || length(n) == 0L) {
    stop("x and n must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(x)) || any(x < 0)) {
    stop("x must contain nonnegative finite values.", call. = FALSE)
  }
  if (any(!is.finite(n)) || any(n < 0)) {
    stop("n must contain nonnegative finite values.", call. = FALSE)
  }

  if (length(x) == 1L && length(n) > 1L) x <- rep(x, length(n))
  if (length(n) == 1L && length(x) > 1L) n <- rep(n, length(x))

  if (length(x) != length(n)) {
    stop("x and n must have the same length, or one must have length 1.", call. = FALSE)
  }

  list(x = x, n = n)
}

#' @noRd
.alpha_m <- function(i, m) {
  i <- .check_i_scalar_ann(i)
  m <- .check_m_scalar_ann(m)

  ic <- interest_convert(i = i, m = m)
  d <- ic$d
  im <- ic$im
  dm <- m * ((im / m) / (1 + im / m))

  (i * d) / (im * dm)
}

#' @noRd
.beta_m <- function(i, m) {
  i <- .check_i_scalar_ann(i)
  m <- .check_m_scalar_ann(m)

  ic <- interest_convert(i = i, m = m)
  im <- ic$im
  dm <- m * ((im / m) / (1 + im / m))

  (i - im) / (im * dm)
}

#' @noRd
.gamma_m <- function(i, m) {
  i <- .check_i_scalar_ann(i)
  m <- .check_m_scalar_ann(m)

  ic <- interest_convert(i = i, m = m)
  d <- ic$d
  im <- ic$im
  dm <- m * ((im / m) / (1 + im / m))

  (dm - d) / (im * dm)
}
#' @noRd
.delta_from_i <- function(i) {
  interest_convert(i = .check_i_scalar_ann(i))$delta
}

#' @noRd
.mu_x <- function(x, model, ...) {
  hazard0(x, model = model, ...)
}

# -------------------------------------------------------------------------
# Shared documentation: UDD approximations
# -------------------------------------------------------------------------

#' UDD annuity approximations
#'
#' UDD-based approximations for annual, fractional-payment, and continuous annuity functions.
#'
#' These functions implement the standard Uniform Distribution of Deaths
#' approximations linking annual, m-thly, and continuous annuity values.
#'
#' The exported functions documented on this page are:
#' \itemize{
#'   \item \code{ax_m_udd()}
#'   \item \code{axn_m_udd()}
#'   \item \code{nax_m_udd()}
#'   \item \code{adotx_m_udd()}
#'   \item \code{adotxn_m_udd()}
#'   \item \code{nadotx_m_udd()}
#'   \item \code{sxn_m_udd()}
#'   \item \code{sdotxn_m_udd()}
#'   \item \code{abarx_udd()}
#'   \item \code{abarxn_udd()}
#'   \item \code{nabarx_udd()}
#' }
#'
#' @name annuity_approximations_udd
#' @aliases ax_m_udd axn_m_udd nax_m_udd
#' @aliases adotx_m_udd adotxn_m_udd nadotx_m_udd
#' @aliases sxn_m_udd sdotxn_m_udd
#' @aliases abarx_udd abarxn_udd nabarx_udd
NULL

#' UDD approximation to m-thly whole life annuity-due
#'
#' Computes
#' \deqn{\ddot{a}_x^{(m)} \approx \alpha(m)\ddot{a}_x - \beta(m).}
#'
#' @param x Age.
#' @param m Number of payments per year.
#' @param i Effective annual interest rate.
#' @param model Survival model.
#' @param ... Additional model parameters.
#'
#' @return Numeric vector.
#' @rdname annuity_approximations_udd
#' @export
adotx_m_udd <- function(x, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  x <- as.numeric(x)

  .alpha_m(i, m) * adotx(x, i, model = model, ...) - .beta_m(i, m)
}

#' UDD approximation to m-thly temporary annuity-due
#'
#' Computes
#' \deqn{\ddot{a}_{x:\overline{n}|}^{(m)} \approx
#' \alpha(m)\ddot{a}_{x:\overline{n}|} - \beta(m)(1-{}_nE_x).}
#'
#' @param n Term.
#' @rdname annuity_approximations_udd
#' @export
adotxn_m_udd <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)
  x <- xn$x
  n <- xn$n

  .alpha_m(i, m) * adotxn(x, n, i, model = model, ...) -
    .beta_m(i, m) * (1 - nEx(x, n, i, model = model, ...))
}

#' UDD approximation to m-thly deferred whole life annuity-due
#'
#' Computes
#' \deqn{{}_{n\mid}\ddot{a}_x^{(m)} \approx
#' \alpha(m)\,{}_{n\mid}\ddot{a}_x - \beta(m)\,{}_nE_x.}
#'
#' @rdname annuity_approximations_udd
#' @export
nadotx_m_udd <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)
  x <- xn$x
  n <- xn$n

  .alpha_m(i, m) * nadotx(x, n, i, model = model, ...) -
    .beta_m(i, m) * nEx(x, n, i, model = model, ...)
}

#' UDD approximation to m-thly whole life annuity-immediate
#'
#' Computes
#' \deqn{a_x^{(m)} \approx \alpha(m)a_x + \gamma(m).}
#'
#' @rdname annuity_approximations_udd
#' @export
ax_m_udd <- function(x, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  x <- as.numeric(x)

  .alpha_m(i, m) * ax(x, i, model = model, ...) + .gamma_m(i, m)
}

#' UDD approximation to m-thly temporary annuity-immediate
#'
#' Computes
#' \deqn{a_{x:\overline{n}|}^{(m)} \approx
#' \alpha(m)a_{x:\overline{n}|} + \gamma(m)(1-{}_nE_x).}
#'
#' @rdname annuity_approximations_udd
#' @export
axn_m_udd <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)
  x <- xn$x
  n <- xn$n

  .alpha_m(i, m) * axn(x, n, i, model = model, ...) +
    .gamma_m(i, m) * (1 - nEx(x, n, i, model = model, ...))
}

#' UDD approximation to m-thly deferred whole life annuity-immediate
#'
#' Computes
#' \deqn{{}_{n\mid}a_x^{(m)} \approx
#' \alpha(m)\,{}_{n\mid}a_x + \gamma(m)\,{}_nE_x.}
#'
#' @rdname annuity_approximations_udd
#' @export
nax_m_udd <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)
  x <- xn$x
  n <- xn$n

  .alpha_m(i, m) * nax(x, n, i, model = model, ...) +
    .gamma_m(i, m) * nEx(x, n, i, model = model, ...)
}

#' UDD approximation to m-thly temporary annuity accumulated value
#'
#' Computes
#' \deqn{\ddot{s}_{x:\overline{n}|}^{(m)} \approx
#' \alpha(m)\ddot{s}_{x:\overline{n}|} - \beta(m)\left(\frac{1}{{}_nE_x}-1\right).}
#'
#' @rdname annuity_approximations_udd
#' @export
sdotxn_m_udd <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)
  x <- xn$x
  n <- xn$n

  ex <- nEx(x, n, i, model = model, ...)
  .alpha_m(i, m) * sdotxn(x, n, i, model = model, ...) -
    .beta_m(i, m) * (1 / ex - 1)
}

#' UDD approximation to m-thly temporary annuity accumulated value
#'
#' Computes
#' \deqn{s_{x:\overline{n}|}^{(m)} \approx
#' \alpha(m)s_{x:\overline{n}|} + \gamma(m)\left(\frac{1}{{}_nE_x}-1\right).}
#'
#' @rdname annuity_approximations_udd
#' @export
sxn_m_udd <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)
  x <- xn$x
  n <- xn$n

  ex <- nEx(x, n, i, model = model, ...)
  .alpha_m(i, m) * sxn(x, n, i, model = model, ...) +
    .gamma_m(i, m) * (1 / ex - 1)
}

#' UDD approximation to continuous whole life annuity
#'
#' Computes
#' \deqn{\bar{a}_x \approx \frac{id}{\delta^2}\ddot{a}_x - \frac{i-\delta}{\delta^2}.}
#'
#' @rdname annuity_approximations_udd
#' @export
abarx_udd <- function(x, i, model, ...) {
  i <- .check_i_scalar_ann(i)
  delta <- .delta_from_i(i)
  d <- interest_convert(i = i)$d

  ((i * d) / delta^2) * adotx(x, i, model = model, ...) -
    (i - delta) / delta^2
}

#' UDD approximation to continuous temporary annuity
#'
#' Uses the identity
#' \deqn{\bar{a}_{x:\overline{n}|} \approx \frac{1-\bar{A}_{x:\overline{n}|}}{\delta}}
#' together with the package's existing Abarxn_udd() approximation.
#'
#' The abarxn_udd() calculation uses the existing Abarxn_udd() insurance
#' approximation. Consequently, additional survival-model arguments are not used
#' in that calculation.
#'
#' @rdname annuity_approximations_udd
#' @export
abarxn_udd <- function(x, n, i, model, ...) {
  i <- .check_i_scalar_ann(i)
  xn <- .recycle_x_n_ann(x, n)
  delta <- .delta_from_i(i)

  # existing package function does not accept model/...
  (1 - Abarxn_udd(xn$x, xn$n, i)) / delta
}

#' UDD approximation to continuous deferred whole life annuity
#'
#' Computes
#' \deqn{{}_{n\mid}\bar{a}_x \approx {}_nE_x \, \bar{a}_{x+n}.}
#'
#' @rdname annuity_approximations_udd
#' @export
nabarx_udd <- function(x, n, i, model, ...) {
  xn <- .recycle_x_n_ann(x, n)
  nEx(xn$x, xn$n, i, model = model, ...) *
    abarx_udd(xn$x + xn$n, i, model = model, ...)
}

# -------------------------------------------------------------------------
# Shared documentation: Woolhouse 2-term approximations
# -------------------------------------------------------------------------

#' Woolhouse 2-term annuity approximations
#'
#' Woolhouse 2-term approximations for fractional-payment annuity functions.
#'
#' @name annuity_approximations_woolhouse2
#' @aliases ax_m_woolhouse2 adotx_m_woolhouse2
#' @aliases axn_m_woolhouse2 adotxn_m_woolhouse2
#' @aliases nax_m_woolhouse2 nadotx_m_woolhouse2
#' @aliases sxn_m_woolhouse2 sdotxn_m_woolhouse2
#' @aliases abarx_woolhouse2
NULL

#' Woolhouse 2-term approximation to m-thly whole life annuity-immediate
#'
#' @param x Age.
#' @param m Number of payments per year.
#' @param i Effective annual interest rate.
#' @param model Survival model.
#' @param ... Additional model parameters.
#'
#' @return Numeric vector.
#' @rdname annuity_approximations_woolhouse2
#' @export
ax_m_woolhouse2 <- function(x, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  ax(x, i, model = model, ...) + (m - 1) / (2 * m)
}

#' Woolhouse 2-term approximation to m-thly whole life annuity-due
#'
#' @rdname annuity_approximations_woolhouse2
#' @export
adotx_m_woolhouse2 <- function(x, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  adotx(x, i, model = model, ...) - (m - 1) / (2 * m)
}

#' Woolhouse 2-term approximation to m-thly deferred annuity-immediate
#'
#' @param n Term.
#' @rdname annuity_approximations_woolhouse2
#' @export
nax_m_woolhouse2 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)

  nax(xn$x, xn$n, i, model = model, ...) +
    (m - 1) / (2 * m) * nEx(xn$x, xn$n, i, model = model, ...)
}

#' Woolhouse 2-term approximation to m-thly deferred annuity-due
#'
#' @rdname annuity_approximations_woolhouse2
#' @export
nadotx_m_woolhouse2 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)

  nadotx(xn$x, xn$n, i, model = model, ...) -
    (m - 1) / (2 * m) * nEx(xn$x, xn$n, i, model = model, ...)
}

#' Woolhouse 2-term approximation to m-thly temporary annuity-immediate
#'
#' @rdname annuity_approximations_woolhouse2
#' @export
axn_m_woolhouse2 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)

  axn(xn$x, xn$n, i, model = model, ...) +
    (m - 1) / (2 * m) * (1 - nEx(xn$x, xn$n, i, model = model, ...))
}

#' Woolhouse 2-term approximation to m-thly temporary annuity-due
#'
#' @rdname annuity_approximations_woolhouse2
#' @export
adotxn_m_woolhouse2 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)

  adotxn(xn$x, xn$n, i, model = model, ...) -
    (m - 1) / (2 * m) * (1 - nEx(xn$x, xn$n, i, model = model, ...))
}

#' Woolhouse 2-term approximation to m-thly temporary accumulated value
#'
#' @rdname annuity_approximations_woolhouse2
#' @export
sxn_m_woolhouse2 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)
  ex <- nEx(xn$x, xn$n, i, model = model, ...)

  sxn(xn$x, xn$n, i, model = model, ...) +
    (m - 1) / (2 * m) * (1 / ex - 1)
}

#' Woolhouse 2-term approximation to m-thly temporary accumulated value
#'
#' @rdname annuity_approximations_woolhouse2
#' @export
sdotxn_m_woolhouse2 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  xn <- .recycle_x_n_ann(x, n)
  ex <- nEx(xn$x, xn$n, i, model = model, ...)

  sdotxn(xn$x, xn$n, i, model = model, ...) -
    (m - 1) / (2 * m) * (1 / ex - 1)
}

#' Woolhouse 2-term approximation to continuous whole life annuity
#'
#' @rdname annuity_approximations_woolhouse2
#' @export
abarx_woolhouse2 <- function(x, i, model, ...) {
  adotx(x, i, model = model, ...) - 0.5
}

# -------------------------------------------------------------------------
# Shared documentation: Woolhouse 3-term approximations
# -------------------------------------------------------------------------

#' Woolhouse 3-term annuity approximations
#'
#' Woolhouse 3-term approximations for fractional-payment annuity functions.
#'
#' @name annuity_approximations_woolhouse3
#' @aliases ax_m_woolhouse3 adotx_m_woolhouse3
#' @aliases axn_m_woolhouse3 adotxn_m_woolhouse3
#' @aliases nax_m_woolhouse3 nadotx_m_woolhouse3
#' @aliases abarx_woolhouse3
NULL

#' Woolhouse 3-term approximation to m-thly whole life annuity-immediate
#'
#' @param x Age.
#' @param m Number of payments per year.
#' @param i Effective annual interest rate.
#' @param model Survival model.
#' @param ... Additional model parameters.
#'
#' @return Numeric vector.
#' @rdname annuity_approximations_woolhouse3
#' @export
ax_m_woolhouse3 <- function(x, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  delta <- .delta_from_i(i)
  x <- as.numeric(x)

  ax(x, i, model = model, ...) +
    (m - 1) / (2 * m) -
    ((m^2 - 1) / (12 * m^2)) * (.mu_x(x, model = model, ...) + delta)
}

#' Woolhouse 3-term approximation to m-thly whole life annuity-due
#'
#' @rdname annuity_approximations_woolhouse3
#' @export
adotx_m_woolhouse3 <- function(x, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  delta <- .delta_from_i(i)
  x <- as.numeric(x)

  adotx(x, i, model = model, ...) -
    (m - 1) / (2 * m) -
    ((m^2 - 1) / (12 * m^2)) * (.mu_x(x, model = model, ...) + delta)
}

#' Woolhouse 3-term approximation to m-thly deferred annuity-immediate
#'
#' @param n Term.
#' @rdname annuity_approximations_woolhouse3
#' @export
nax_m_woolhouse3 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  delta <- .delta_from_i(i)
  xn <- .recycle_x_n_ann(x, n)
  ex <- nEx(xn$x, xn$n, i, model = model, ...)

  nax(xn$x, xn$n, i, model = model, ...) +
    ex * ((m - 1) / (2 * m) -
            ((m^2 - 1) / (12 * m^2)) * (.mu_x(xn$x + xn$n, model = model, ...) + delta))
}

#' Woolhouse 3-term approximation to m-thly deferred annuity-due
#'
#' @rdname annuity_approximations_woolhouse3
#' @export
nadotx_m_woolhouse3 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  delta <- .delta_from_i(i)
  xn <- .recycle_x_n_ann(x, n)
  ex <- nEx(xn$x, xn$n, i, model = model, ...)

  nadotx(xn$x, xn$n, i, model = model, ...) -
    ex * ((m - 1) / (2 * m) +
            ((m^2 - 1) / (12 * m^2)) * (.mu_x(xn$x + xn$n, model = model, ...) + delta))
}

#' Woolhouse 3-term approximation to m-thly temporary annuity-immediate
#'
#' @rdname annuity_approximations_woolhouse3
#' @export
axn_m_woolhouse3 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  delta <- .delta_from_i(i)
  xn <- .recycle_x_n_ann(x, n)
  ex <- nEx(xn$x, xn$n, i, model = model, ...)

  axn(xn$x, xn$n, i, model = model, ...) +
    (m - 1) / (2 * m) * (1 - ex) -
    ((m^2 - 1) / (12 * m^2)) *
    (.mu_x(xn$x, model = model, ...) -
       ex * .mu_x(xn$x + xn$n, model = model, ...) +
       delta * (1 - ex))
}

#' Woolhouse 3-term approximation to m-thly temporary annuity-due
#'
#' @rdname annuity_approximations_woolhouse3
#' @export
adotxn_m_woolhouse3 <- function(x, n, m, i, model, ...) {
  m <- .check_m_scalar_ann(m)
  delta <- .delta_from_i(i)
  xn <- .recycle_x_n_ann(x, n)
  ex <- nEx(xn$x, xn$n, i, model = model, ...)

  adotxn(xn$x, xn$n, i, model = model, ...) -
    (m - 1) / (2 * m) * (1 - ex) -
    ((m^2 - 1) / (12 * m^2)) *
    (.mu_x(xn$x, model = model, ...) -
       ex * .mu_x(xn$x + xn$n, model = model, ...) +
       delta * (1 - ex))
}

#' Woolhouse 3-term approximation to continuous whole life annuity
#'
#' @rdname annuity_approximations_woolhouse3
#' @export
abarx_woolhouse3 <- function(x, i, model, ...) {
  delta <- .delta_from_i(i)
  x <- as.numeric(x)

  adotx(x, i, model = model, ...) -
    0.5 -
    (.mu_x(x, model = model, ...) + delta) / 12
}

test_that("alpha beta gamma are finite and sensible", {
  i <- 0.05
  m <- 12

  expect_true(is.finite(.alpha_m(i, m)))
  expect_true(is.finite(.beta_m(i, m)))
  expect_true(is.finite(.gamma_m(i, m)))
  expect_true(.alpha_m(i, m) > 0)
})

test_that("UDD m-thly due approximation equals formula", {
  x <- 40
  i <- 0.05
  m <- 12

  expected <- .alpha_m(i, m) * adotx(x, i, model = "uniform", omega = 100) -
    .beta_m(i, m)

  expect_equal(
    adotx_m_udd(x, m, i, model = "uniform", omega = 100),
    expected
  )
})

test_that("UDD m-thly immediate approximation equals formula", {
  x <- 40
  i <- 0.05
  m <- 12

  expected <- .alpha_m(i, m) * ax(x, i, model = "uniform", omega = 100) +
    .gamma_m(i, m)

  expect_equal(
    ax_m_udd(x, m, i, model = "uniform", omega = 100),
    expected
  )
})

test_that("UDD temporary due approximation equals formula", {
  x <- 40
  n <- 10
  i <- 0.05
  m <- 12

  ex <- nEx(x, n, i, model = "uniform", omega = 100)
  expected <- .alpha_m(i, m) * adotxn(x, n, i, model = "uniform", omega = 100) -
    .beta_m(i, m) * (1 - ex)

  expect_equal(
    adotxn_m_udd(x, n, m, i, model = "uniform", omega = 100),
    expected
  )
})

test_that("UDD temporary immediate approximation equals formula", {
  x <- 40
  n <- 10
  i <- 0.05
  m <- 12

  ex <- nEx(x, n, i, model = "uniform", omega = 100)
  expected <- .alpha_m(i, m) * axn(x, n, i, model = "uniform", omega = 100) +
    .gamma_m(i, m) * (1 - ex)

  expect_equal(
    axn_m_udd(x, n, m, i, model = "uniform", omega = 100),
    expected
  )
})

test_that("UDD deferred due approximation equals formula", {
  x <- 40
  n <- 10
  i <- 0.05
  m <- 12

  ex <- nEx(x, n, i, model = "uniform", omega = 100)
  expected <- .alpha_m(i, m) * nadotx(x, n, i, model = "uniform", omega = 100) -
    .beta_m(i, m) * ex

  expect_equal(
    nadotx_m_udd(x, n, m, i, model = "uniform", omega = 100),
    expected
  )
})

test_that("UDD deferred immediate approximation equals formula", {
  x <- 40
  n <- 10
  i <- 0.05
  m <- 12

  ex <- nEx(x, n, i, model = "uniform", omega = 100)
  expected <- .alpha_m(i, m) * nax(x, n, i, model = "uniform", omega = 100) +
    .gamma_m(i, m) * ex

  expect_equal(
    nax_m_udd(x, n, m, i, model = "uniform", omega = 100),
    expected
  )
})

test_that("UDD accumulated value approximations equal formulas", {
  x <- 40
  n <- 10
  i <- 0.05
  m <- 12
  ex <- nEx(x, n, i, model = "uniform", omega = 100)

  expect_equal(
    sdotxn_m_udd(x, n, m, i, model = "uniform", omega = 100),
    .alpha_m(i, m) * sdotxn(x, n, i, model = "uniform", omega = 100) -
      .beta_m(i, m) * (1 / ex - 1)
  )

  expect_equal(
    sxn_m_udd(x, n, m, i, model = "uniform", omega = 100),
    .alpha_m(i, m) * sxn(x, n, i, model = "uniform", omega = 100) +
      .gamma_m(i, m) * (1 / ex - 1)
  )
})

test_that("continuous UDD whole life approximation equals formula", {
  x <- 40
  i <- 0.05
  delta <- interest_convert(i = i)$delta
  d <- interest_convert(i = i)$d

  expect_equal(
    abarx_udd(x, i, model = "uniform", omega = 100),
    ((i * d) / delta^2) * adotx(x, i, model = "uniform", omega = 100) -
      (i - delta) / delta^2
  )
})

test_that("continuous UDD temporary approximation equals identity with existing package UDD insurance function", {
  x <- 40
  n <- 10
  i <- 0.05
  delta <- interest_convert(i = i)$delta

  expect_equal(
    abarxn_udd(x, n, i, model = "uniform", omega = 100),
    (1 - Abarxn_udd(x, n, i)) / delta
  )
})

test_that("continuous UDD deferred approximation equals deferred identity", {
  x <- 40
  n <- 10
  i <- 0.05

  expect_equal(
    nabarx_udd(x, n, i, model = "uniform", omega = 100),
    nEx(x, n, i, model = "uniform", omega = 100) *
      abarx_udd(x + n, i, model = "uniform", omega = 100)
  )
})

test_that("Woolhouse 2-term whole life approximations equal formulas", {
  x <- 40
  i <- 0.05
  m <- 12

  expect_equal(
    ax_m_woolhouse2(x, m, i, model = "uniform", omega = 100),
    ax(x, i, model = "uniform", omega = 100) + (m - 1) / (2 * m)
  )

  expect_equal(
    adotx_m_woolhouse2(x, m, i, model = "uniform", omega = 100),
    adotx(x, i, model = "uniform", omega = 100) - (m - 1) / (2 * m)
  )
})

test_that("Woolhouse 2-term temporary and deferred approximations equal formulas", {
  x <- 40
  n <- 10
  i <- 0.05
  m <- 12
  ex <- nEx(x, n, i, model = "uniform", omega = 100)

  expect_equal(
    nax_m_woolhouse2(x, n, m, i, model = "uniform", omega = 100),
    nax(x, n, i, model = "uniform", omega = 100) + (m - 1) / (2 * m) * ex
  )

  expect_equal(
    nadotx_m_woolhouse2(x, n, m, i, model = "uniform", omega = 100),
    nadotx(x, n, i, model = "uniform", omega = 100) - (m - 1) / (2 * m) * ex
  )

  expect_equal(
    axn_m_woolhouse2(x, n, m, i, model = "uniform", omega = 100),
    axn(x, n, i, model = "uniform", omega = 100) + (m - 1) / (2 * m) * (1 - ex)
  )

  expect_equal(
    adotxn_m_woolhouse2(x, n, m, i, model = "uniform", omega = 100),
    adotxn(x, n, i, model = "uniform", omega = 100) - (m - 1) / (2 * m) * (1 - ex)
  )
})

test_that("Woolhouse 2-term accumulated-value approximations equal formulas", {
  x <- 40
  n <- 10
  i <- 0.05
  m <- 12
  ex <- nEx(x, n, i, model = "uniform", omega = 100)

  expect_equal(
    sxn_m_woolhouse2(x, n, m, i, model = "uniform", omega = 100),
    sxn(x, n, i, model = "uniform", omega = 100) + (m - 1) / (2 * m) * (1 / ex - 1)
  )

  expect_equal(
    sdotxn_m_woolhouse2(x, n, m, i, model = "uniform", omega = 100),
    sdotxn(x, n, i, model = "uniform", omega = 100) - (m - 1) / (2 * m) * (1 / ex - 1)
  )
})

test_that("Woolhouse 2-term continuous whole life equals ddot ax minus one-half", {
  x <- 40
  i <- 0.05

  expect_equal(
    abarx_woolhouse2(x, i, model = "uniform", omega = 100),
    adotx(x, i, model = "uniform", omega = 100) - 0.5
  )
})

test_that("Woolhouse 3-term whole life approximations equal formulas", {
  x <- 40
  i <- 0.05
  m <- 12
  delta <- interest_convert(i = i)$delta
  mu <- hazard0(x, model = "uniform", omega = 100)

  expect_equal(
    ax_m_woolhouse3(x, m, i, model = "uniform", omega = 100),
    ax(x, i, model = "uniform", omega = 100) +
      (m - 1) / (2 * m) -
      ((m^2 - 1) / (12 * m^2)) * (mu + delta)
  )

  expect_equal(
    adotx_m_woolhouse3(x, m, i, model = "uniform", omega = 100),
    adotx(x, i, model = "uniform", omega = 100) -
      (m - 1) / (2 * m) -
      ((m^2 - 1) / (12 * m^2)) * (mu + delta)
  )
})

test_that("Woolhouse 3-term continuous whole life equals formula", {
  x <- 40
  i <- 0.05
  delta <- interest_convert(i = i)$delta
  mu <- hazard0(x, model = "uniform", omega = 100)

  expect_equal(
    abarx_woolhouse3(x, i, model = "uniform", omega = 100),
    adotx(x, i, model = "uniform", omega = 100) - 0.5 - (mu + delta) / 12
  )
})

test_that("approximation functions vectorize over x and n", {
  i <- 0.05
  m <- 12
  x <- c(40, 45)
  n <- c(10, 5)

  expect_length(ax_m_udd(x, m, i, model = "uniform", omega = 100), 2)
  expect_length(adotx_m_woolhouse2(x, m, i, model = "uniform", omega = 100), 2)
  expect_length(axn_m_udd(x, n, m, i, model = "uniform", omega = 100), 2)
  expect_length(nax_m_woolhouse3(x, n, m, i, model = "uniform", omega = 100), 2)
  expect_length(abarxn_udd(x, n, i, model = "uniform", omega = 100), 2)
})

test_that("approximation functions reject bad inputs", {
  expect_error(ax_m_udd(40, m = 0, i = 0.05, model = "uniform", omega = 100))
  expect_error(adotx_m_woolhouse2(40, m = 12, i = -1, model = "uniform", omega = 100))
  expect_error(
    axn_m_udd(c(40, 45), c(10, 5, 3), m = 12, i = 0.05, model = "uniform", omega = 100),
    "same length|length 1"
  )
})

test_that("approximation values are finite in standard cases", {
  i <- 0.05
  m <- 12

  vals <- c(
    ax_m_udd(40, m, i, model = "uniform", omega = 100),
    adotx_m_udd(40, m, i, model = "uniform", omega = 100),
    abarx_udd(40, i, model = "uniform", omega = 100),
    ax_m_woolhouse2(40, m, i, model = "uniform", omega = 100),
    ax_m_woolhouse3(40, m, i, model = "uniform", omega = 100)
  )

  expect_true(all(is.finite(vals)))
})

test_that("continuous approximation lies between annual immediate and annual due in standard case", {
  i <- 0.05

  a_immed <- ax(40, i, model = "uniform", omega = 100)
  a_due <- adotx(40, i, model = "uniform", omega = 100)
  a_cont <- abarx_woolhouse2(40, i, model = "uniform", omega = 100)

  expect_true(a_immed < a_cont)
  expect_true(a_cont < a_due)
})

# =========================================================
# Profit analysis functions
# =========================================================

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.profit_check_numeric <- function(x, name, allow_empty = FALSE) {
  if (!is.numeric(x)) {
    stop(name, " must be numeric.", call. = FALSE)
  }
  if (!allow_empty && length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(x))) {
    stop(name, " must contain finite values.", call. = FALSE)
  }
  as.numeric(x)
}

#' @noRd
.profit_check_nonnegative <- function(x, name, allow_empty = FALSE) {
  x <- .profit_check_numeric(x, name, allow_empty = allow_empty)
  if (any(x < 0)) {
    stop(name, " must contain nonnegative values.", call. = FALSE)
  }
  x
}

#' @noRd
.profit_check_probability <- function(x, name, allow_empty = FALSE) {
  x <- .profit_check_numeric(x, name, allow_empty = allow_empty)
  if (any(x < 0 | x > 1)) {
    stop(name, " must contain values in [0, 1].", call. = FALSE)
  }
  x
}

#' @noRd
.profit_check_rate <- function(x, name) {
  x <- .profit_check_numeric(x, name)
  if (any(x <= -1)) {
    stop(name, " must contain values greater than -1.", call. = FALSE)
  }
  x
}

#' @noRd
.profit_check_scalar <- function(x, name, nonnegative = FALSE, positive = FALSE) {
  if (nonnegative || positive) {
    x <- .profit_check_nonnegative(x, name)
  } else {
    x <- .profit_check_numeric(x, name)
  }
  if (length(x) != 1L) {
    stop(name, " must be a numeric scalar.", call. = FALSE)
  }
  if (positive && x <= 0) {
    stop(name, " must be positive.", call. = FALSE)
  }
  x
}

#' @noRd
.profit_check_logical_scalar <- function(x, name) {
  if (!is.logical(x) || length(x) != 1L || is.na(x)) {
    stop(name, " must be TRUE or FALSE.", call. = FALSE)
  }
  x
}

#' @noRd
.profit_recycle_to <- function(x, n, name) {
  if (length(x) == 1L) {
    return(rep(x, n))
  }
  if (length(x) != n) {
    stop(name, " must have length 1 or ", n, ".", call. = FALSE)
  }
  x
}

#' @noRd
.profit_recycle_common <- function(..., .names = NULL) {
  values <- list(...)
  lengths <- vapply(values, length, integer(1))
  if (any(lengths == 0L)) {
    stop("Inputs must have positive length.", call. = FALSE)
  }
  common_length <- max(lengths)
  if (is.null(.names)) {
    .names <- paste0("Argument ", seq_along(values))
  }
  if (length(.names) != length(values)) {
    stop(".names must have the same length as the supplied arguments.", call. = FALSE)
  }
  for (j in seq_along(values)) {
    if (!lengths[j] %in% c(1L, common_length)) {
      stop(
        .names[j],
        " must have length 1 or the common length ",
        common_length,
        ".",
        call. = FALSE
      )
    }
    values[[j]] <- rep(values[[j]], length.out = common_length)
  }
  values
}

#' @noRd
.profit_start_probabilities <- function(p_tau, n, name = "p_tau") {
  if (n < 1L) {
    stop("n must be positive.", call. = FALSE)
  }
  if (n == 1L) {
    if (!length(p_tau) %in% c(0L, 1L)) {
      stop(name, " must have length 0 or 1 when n = 1.", call. = FALSE)
    }
    if (length(p_tau) == 1L) {
      .profit_check_probability(p_tau, name)
    }
    return(1)
  }
  p_tau <- .profit_check_probability(p_tau, name)
  if (!length(p_tau) %in% c(n - 1L, n)) {
    stop(name, " must have length n - 1 or n.", call. = FALSE)
  }
  c(1, cumprod(p_tau[seq_len(n - 1L)]))
}

# -------------------------------------------------------------------------
# Profit vector and profit signature
# -------------------------------------------------------------------------

#' Profit vector for a discrete profit-analysis model
#'
#' Computes expected profit by policy year for a discrete contract with up to
#' two decrements. The first element is the negative pre-contract expense.
#'
#' For policy year \eqn{k}, the expected profit is
#' \deqn{
#' [V_{k-1} + G_k(1-r_k)-e_k](1+i_k)
#' -
#' [(b_k^{(1)}+s_k^{(1)})q_k^{(1)}
#' +(b_k^{(2)}+s_k^{(2)})q_k^{(2)}
#' +V_kp_k^{(\tau)}].
#' }
#'
#' Scalar yearly inputs are recycled to the number of policy years determined
#' by \code{length(V) - 1}.
#'
#' @param V Numeric vector of gross premium reserves with length
#'   \code{n + 1}, including the issue-time and terminal reserves.
#' @param G Gross premium by policy year.
#' @param i Annual effective interest rate by policy year. Values must be
#'   greater than \code{-1}.
#' @param r Percent-of-premium expense rate by policy year. Values must lie in
#'   \code{[0, 1]}.
#' @param e Fixed expense by policy year.
#' @param q1 Probability of the first decrement by policy year.
#' @param q2 Probability of the second decrement by policy year.
#' @param b1 Benefit payable on the first decrement.
#' @param b2 Benefit payable on the second decrement.
#' @param s1 Settlement expense associated with the first decrement.
#' @param s2 Settlement expense associated with the second decrement.
#' @param p_tau Optional in-force probability by policy year. If omitted, it is
#'   calculated as \code{1 - q1 - q2}.
#' @param pre_contract_expense Nonnegative scalar pre-contract expense.
#'
#' @return A named numeric vector of length \code{n + 1}.
#'
#' @examples
#' V <- c(0, 5.66, 6.17, 0)
#' qx <- c(0.00142, 0.00153, 0.00166)
#' Pr_vector_disc(
#'   V = V, G = 95, i = 0.06, r = 0.05, e = 10,
#'   q1 = qx, b1 = 50000, pre_contract_expense = 15
#' )
#'
#' @export
Pr_vector_disc <- function(V, G, i, r = 0, e = 0,
                           q1, q2 = 0,
                           b1, b2 = 0,
                           s1 = 0, s2 = 0,
                           p_tau = NULL,
                           pre_contract_expense = 0) {
  V <- .profit_check_numeric(V, "V")
  G <- .profit_check_nonnegative(G, "G")
  i <- .profit_check_rate(i, "i")
  r <- .profit_check_probability(r, "r")
  e <- .profit_check_nonnegative(e, "e")
  q1 <- .profit_check_probability(q1, "q1")
  q2 <- .profit_check_probability(q2, "q2")
  b1 <- .profit_check_nonnegative(b1, "b1")
  b2 <- .profit_check_nonnegative(b2, "b2")
  s1 <- .profit_check_nonnegative(s1, "s1")
  s2 <- .profit_check_nonnegative(s2, "s2")
  pre_contract_expense <- .profit_check_scalar(
    pre_contract_expense, "pre_contract_expense", nonnegative = TRUE
  )

  if (length(V) < 2L) {
    stop("V must have length at least 2.", call. = FALSE)
  }

  n <- length(V) - 1L
  G <- .profit_recycle_to(G, n, "G")
  i <- .profit_recycle_to(i, n, "i")
  r <- .profit_recycle_to(r, n, "r")
  e <- .profit_recycle_to(e, n, "e")
  q1 <- .profit_recycle_to(q1, n, "q1")
  q2 <- .profit_recycle_to(q2, n, "q2")
  b1 <- .profit_recycle_to(b1, n, "b1")
  b2 <- .profit_recycle_to(b2, n, "b2")
  s1 <- .profit_recycle_to(s1, n, "s1")
  s2 <- .profit_recycle_to(s2, n, "s2")

  if (is.null(p_tau)) {
    p_tau <- 1 - q1 - q2
    if (any(p_tau < 0)) {
      stop("q1 + q2 must not exceed 1.", call. = FALSE)
    }
  } else {
    p_tau <- .profit_check_probability(p_tau, "p_tau")
    p_tau <- .profit_recycle_to(p_tau, n, "p_tau")
  }

  out <- numeric(n + 1L)
  out[1L] <- -pre_contract_expense

  for (k in seq_len(n)) {
    out[k + 1L] <-
      (V[k] + G[k] * (1 - r[k]) - e[k]) * (1 + i[k]) -
      ((b1[k] + s1[k]) * q1[k] +
         (b2[k] + s2[k]) * q2[k] +
         V[k + 1L] * p_tau[k])
  }

  names(out) <- paste0("Pr", 0:n)
  out
}

#' Profit signature
#'
#' Converts policy-year expected profits into a profit signature by weighting
#' each future expected profit by the probability that the contract is in
#' force at the start of that policy year.
#'
#' @param Pr Profit vector of length \code{n + 1}.
#' @param p_tau One-year in-force probabilities. For \code{n > 1}, this may
#'   have length \code{n - 1} or \code{n}; the final value is ignored when
#'   length \code{n}. For a one-year contract, use \code{numeric(0)} or a
#'   single probability.
#'
#' @return A named numeric vector with the same length as \code{Pr}.
#'
#' @examples
#' Pr <- c(-15.00, 8.42, 8.40, 8.61)
#' Pi_signature(Pr, p_tau = c(0.99858, 0.99847, 0.99834))
#'
#' @export
Pi_signature <- function(Pr, p_tau) {
  Pr <- .profit_check_numeric(Pr, "Pr")
  if (length(Pr) < 2L) {
    stop("Pr must have length at least 2.", call. = FALSE)
  }
  n <- length(Pr) - 1L
  start_probabilities <- .profit_start_probabilities(p_tau = p_tau, n = n)
  out <- c(Pr[1L], Pr[-1L] * start_probabilities)
  names(out) <- paste0("Pi", 0:n)
  out
}

# -------------------------------------------------------------------------
# Profitability measures
# -------------------------------------------------------------------------

#' Net present value of a profit signature
#'
#' Discounts a profit signature at one or more annual effective risk discount
#' rates.
#'
#' @param Pi Numeric profit-signature vector.
#' @param r Annual effective risk discount rate. May be scalar or vector;
#'   values must be greater than \code{-1}.
#'
#' @return A numeric vector with one value for each rate in \code{r}.
#'
#' @examples
#' Pi <- c(-15.00, 8.42, 8.39, 8.58)
#' NPV_profit(Pi, r = 0.10)
#' NPV_profit(Pi, r = c(0.08, 0.10, 0.12))
#'
#' @export
NPV_profit <- function(Pi, r) {
  Pi <- .profit_check_numeric(Pi, "Pi")
  r <- .profit_check_rate(r, "r")
  times <- seq_along(Pi) - 1L

  vapply(
    r,
    function(rate) sum(Pi / (1 + rate)^times),
    numeric(1)
  )
}

#' Partial net present values
#'
#' Computes cumulative discounted profits through each duration. A scalar
#' discount rate returns a named vector; vectorized rates return a matrix.
#'
#' @inheritParams NPV_profit
#'
#' @return A named numeric vector for scalar \code{r}, or a numeric matrix for
#'   vectorized \code{r}.
#'
#' @examples
#' Pi <- c(-15.00, 8.42, 8.39, 8.58)
#' NPV_partial(Pi, r = 0.10)
#'
#' @export
NPV_partial <- function(Pi, r) {
  Pi <- .profit_check_numeric(Pi, "Pi")
  r <- .profit_check_rate(r, "r")
  times <- seq_along(Pi) - 1L

  rows <- lapply(r, function(rate) cumsum(Pi / (1 + rate)^times))
  column_names <- paste0("NPV(", times, ")")

  if (length(rows) == 1L) {
    return(stats::setNames(rows[[1L]], column_names))
  }

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  colnames(out) <- column_names
  out
}

#' Discounted payback period
#'
#' Returns the first duration at which cumulative discounted profit is
#' nonnegative.
#'
#' @inheritParams NPV_profit
#'
#' @return An integer vector with one value for each discount rate. An element
#'   is \code{NA_integer_} when payback is not reached.
#'
#' @examples
#' Pi <- c(-15.00, 8.42, 8.39, 8.58)
#' discounted_payback_period(Pi, r = 0.10)
#'
#' @export
discounted_payback_period <- function(Pi, r) {
  Pi <- .profit_check_numeric(Pi, "Pi")
  r <- .profit_check_rate(r, "r")
  partial <- NPV_partial(Pi = Pi, r = r)

  if (is.null(dim(partial))) {
    index <- which(partial >= 0)
    if (length(index) == 0L) {
      return(NA_integer_)
    }
    return(as.integer(index[1L] - 1L))
  }

  apply(partial, 1L, function(x) {
    index <- which(x >= 0)
    if (length(index) == 0L) {
      return(NA_integer_)
    }
    as.integer(index[1L] - 1L)
  })
}

#' Internal rate of return
#'
#' Computes a root of the profit-signature net present value.
#'
#' @param Pi Numeric profit-signature vector.
#' @param interval Numeric vector of length two giving the root-search
#'   interval. Its lower endpoint must be greater than \code{-1}.
#' @param tol Positive scalar tolerance passed to \code{stats::uniroot()}.
#'
#' @return A numeric scalar.
#'
#' @examples
#' Pi <- c(-15.00, 8.42, 8.39, 8.58)
#' IRR_profit(Pi)
#'
#' @export
IRR_profit <- function(Pi, interval = c(0, 1),
                       tol = .Machine$double.eps^0.5) {
  Pi <- .profit_check_numeric(Pi, "Pi")
  interval <- .profit_check_numeric(interval, "interval")
  tol <- .profit_check_scalar(tol, "tol", positive = TRUE)

  if (length(interval) != 2L ||
      interval[1L] <= -1 ||
      interval[1L] >= interval[2L]) {
    stop(
      "interval must have length 2 and satisfy -1 < interval[1] < interval[2].",
      call. = FALSE
    )
  }

  objective <- function(rate) NPV_profit(Pi = Pi, r = rate)
  lower_value <- objective(interval[1L])
  upper_value <- objective(interval[2L])

  if (lower_value == 0) return(interval[1L])
  if (upper_value == 0) return(interval[2L])

  if (lower_value * upper_value > 0) {
    stop("NPV does not change sign over the supplied interval.", call. = FALSE)
  }

  stats::uniroot(objective, interval = interval, tol = tol)$root
}

#' Actuarial present value of gross premiums
#'
#' Computes the actuarial present value of premiums weighted by contract
#' persistency at the start of each policy year.
#'
#' @param G Nonnegative gross premium vector.
#' @param r Annual effective risk discount rate. May be scalar or vector;
#'   values must be greater than \code{-1}.
#' @param p_tau One-year in-force probabilities. For \code{n > 1}, this may
#'   have length \code{n - 1} or \code{n}; the final value is ignored when
#'   length \code{n}. For one premium, use \code{numeric(0)} or one value.
#'
#' @return A numeric vector with one value for each rate in \code{r}.
#'
#' @examples
#' APV_gross_premiums(
#'   G = rep(95, 3),
#'   r = 0.10,
#'   p_tau = c(0.99858, 0.99847, 0.99834)
#' )
#'
#' @export
APV_gross_premiums <- function(G, r, p_tau) {
  G <- .profit_check_nonnegative(G, "G")
  r <- .profit_check_rate(r, "r")
  n <- length(G)
  start_probabilities <- .profit_start_probabilities(p_tau = p_tau, n = n)
  times <- 0:(n - 1L)

  vapply(
    r,
    function(rate) sum(G * start_probabilities / (1 + rate)^times),
    numeric(1)
  )
}

#' Profit margin
#'
#' Computes net present value divided by the actuarial present value of gross
#' premiums. Scalar arguments are recycled to the common length.
#'
#' @param NPV Net present value of profits.
#' @param APV_GP Positive actuarial present value of gross premiums.
#'
#' @return A numeric vector.
#'
#' @examples
#' profit_margin(6.03, 259.52)
#'
#' @export
profit_margin <- function(NPV, APV_GP) {
  NPV <- .profit_check_numeric(NPV, "NPV")
  APV_GP <- .profit_check_nonnegative(APV_GP, "APV_GP")

  values <- .profit_recycle_common(
    NPV, APV_GP, .names = c("NPV", "APV_GP")
  )
  NPV <- values[[1]]
  APV_GP <- values[[2]]

  if (any(APV_GP <= 0)) {
    stop("APV_GP must contain positive values.", call. = FALSE)
  }

  NPV / APV_GP
}

# -------------------------------------------------------------------------
# Zeroized reserves
# -------------------------------------------------------------------------

#' Zeroized reserves for a discrete death-benefit contract
#'
#' Computes reserves backward by setting expected profit in each policy year
#' equal to zero. Negative reserves may optionally be floored at zero.
#'
#' @param qx Mortality probability by policy year.
#' @param i Annual effective interest rate by policy year. Values must be
#'   greater than \code{-1}.
#' @param G Gross premium by policy year.
#' @param benefit Death benefit by policy year.
#' @param r Percent-of-premium expense rate by policy year. Values must lie in
#'   \code{[0, 1]}.
#' @param e Fixed expense by policy year.
#' @param V_terminal Nonnegative scalar terminal reserve.
#' @param floor_zero Logical scalar. If \code{TRUE}, negative reserves are
#'   replaced by zero.
#'
#' @return A named numeric vector of length \code{length(qx) + 1}.
#'
#' @examples
#' V_zeroized(
#'   qx = c(0.015, 0.017, 0.019, 0.021, 0.024),
#'   i = 0.06,
#'   G = 19279,
#'   benefit = 1000000,
#'   e = 240
#' )
#'
#' @export
V_zeroized <- function(qx, i, G, benefit, r = 0, e = 0,
                       V_terminal = 0, floor_zero = TRUE) {
  qx <- .profit_check_probability(qx, "qx")
  i <- .profit_check_rate(i, "i")
  G <- .profit_check_nonnegative(G, "G")
  benefit <- .profit_check_nonnegative(benefit, "benefit")
  r <- .profit_check_probability(r, "r")
  e <- .profit_check_nonnegative(e, "e")
  V_terminal <- .profit_check_scalar(
    V_terminal, "V_terminal", nonnegative = TRUE
  )
  floor_zero <- .profit_check_logical_scalar(floor_zero, "floor_zero")

  n <- length(qx)
  i <- .profit_recycle_to(i, n, "i")
  G <- .profit_recycle_to(G, n, "G")
  benefit <- .profit_recycle_to(benefit, n, "benefit")
  r <- .profit_recycle_to(r, n, "r")
  e <- .profit_recycle_to(e, n, "e")

  V <- numeric(n + 1L)
  V[n + 1L] <- V_terminal

  for (k in seq.int(n, 1L)) {
    net_premium <- G[k] * (1 - r[k]) - e[k]
    reserve <- (
      benefit[k] * qx[k] +
        V[k + 1L] * (1 - qx[k])
    ) / (1 + i[k]) - net_premium

    if (floor_zero) {
      reserve <- max(reserve, 0)
    }

    V[k] <- reserve
  }

  names(V) <- paste0("V", 0:n)
  V
}

test_that("Pr_vector_disc reproduces Example 17.1", {
  V <- c(0, 5.66, 6.17, 0)
  qx <- c(.00142, .00153, .00166)

  Pr <- Pr_vector_disc(
    V = V,
    G = 95,
    i = 0.06,
    r = 0.05,
    e = 10,
    q1 = qx,
    b1 = 50000,
    pre_contract_expense = 15
  )

  expect_equal(unname(Pr["Pr0"]), -15.00, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr1"]), 8.42, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr2"]), 8.40, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr3"]), 8.61, tolerance = 1e-2)
})

test_that("Pi_signature reproduces Example 17.2", {
  Pr <- c(-15.00, 8.42, 8.40, 8.61)
  Pi <- Pi_signature(Pr, p_tau = c(.99858, .99847, .99834))

  expect_equal(unname(Pi["Pi0"]), -15.00, tolerance = 1e-2)
  expect_equal(unname(Pi["Pi1"]), 8.42, tolerance = 1e-2)
  expect_equal(unname(Pi["Pi2"]), 8.39, tolerance = 1e-2)
  expect_equal(unname(Pi["Pi3"]), 8.58, tolerance = 1e-2)
})

test_that("NPV_profit reproduces Example 17.3", {
  Pi <- c(-15.00, 8.42, 8.39, 8.58)
  val <- NPV_profit(Pi, r = 0.10)

  expect_equal(val, 6.03, tolerance = 1e-2)
})

test_that("IRR_profit reproduces Example 17.4", {
  Pi <- c(-15.00, 8.42, 8.39, 8.58)
  irr <- IRR_profit(Pi, interval = c(0, 1))

  expect_equal(100 * irr, 31.635, tolerance = 1e-2)
})

test_that("APV_gross_premiums and profit_margin reproduce Example 17.5", {
  apv_gp <- APV_gross_premiums(
    G = rep(95, 3),
    r = 0.10,
    p_tau = c(.99858, .99847, .99834)
  )

  pm <- profit_margin(NPV = 6.03, APV_GP = apv_gp)

  expect_equal(apv_gp, 259.52, tolerance = 1e-2)
  expect_equal(round(pm, 5), 0.02324)
})

test_that("NPV_partial and discounted_payback_period reproduce Example 17.6", {
  Pi <- c(-15.00, 8.42, 8.39, 8.58)
  partial <- NPV_partial(Pi, r = 0.10)
  dpp <- discounted_payback_period(Pi, r = 0.10)

  expect_equal(unname(partial["NPV(0)"]), -15.00, tolerance = 1e-2)
  expect_equal(unname(partial["NPV(1)"]), -7.35, tolerance = 1e-2)
  expect_lt(abs(unname(partial["NPV(2)"]) - (-0.42)), 0.01)
  expect_equal(unname(partial["NPV(3)"]), 6.03, tolerance = 1e-2)
  expect_equal(dpp, 3L)
})

test_that("Pr_vector_disc reproduces Example 17.7", {
  V <- c(0, 2500, 4000, 5000, 4000, 0)
  qx <- c(.015, .017, .019, .021, .024)

  Pr <- Pr_vector_disc(
    V = V,
    G = 19250,
    i = 0.06,
    r = 0,
    e = 240,
    q1 = qx,
    b1 = 1000000,
    pre_contract_expense = 5000
  )

  expect_equal(unname(Pr["Pr0"]), -5000.00, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr1"]), 2688.10, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr2"]), 1868.60, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr3"]), 485.60, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr4"]), 534.60, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr5"]), 390.60, tolerance = 1e-2)
})

test_that("Pi_signature, NPV_profit, and profit_margin reproduce Example 17.8", {
  Pr <- c(-5000.00, 2688.10, 1868.60, 485.60, 534.60, 390.60)
  p_tau <- c(.98500, .98300, .98100, .97900, .97600)

  Pi <- Pi_signature(Pr, p_tau = p_tau)
  npv <- NPV_profit(Pi, r = 0.10)
  apv_gp <- APV_gross_premiums(
    G = rep(19250, 5),
    r = 0.10,
    p_tau = p_tau
  )
  pm <- profit_margin(npv, apv_gp)

  expect_equal(unname(Pi["Pi2"]), 1840.57, tolerance = 1e-2)
  expect_equal(unname(Pi["Pi3"]), 470.19, tolerance = 1e-2)
  expect_equal(unname(Pi["Pi4"]), 507.80, tolerance = 1e-2)
  expect_equal(unname(Pi["Pi5"]), 363.22, tolerance = 1e-2)

  expect_equal(npv, -109.52, tolerance = 1e-2)
  expect_equal(apv_gp, 77855.49, tolerance = 1e-2)
  expect_equal(round(pm, 5), -0.00141)
})

test_that("V_zeroized reproduces Chapter 17 zeroized reserves", {
  Vz <- V_zeroized(
    qx = c(.015, .017, .019, .021, .024),
    i = 0.06,
    G = 19279,
    benefit = 1000000,
    r = 0,
    e = 240,
    V_terminal = 0,
    floor_zero = TRUE
  )

  expect_equal(unname(Vz["V0"]), 0.00, tolerance = 1e-2)
  expect_equal(unname(Vz["V1"]), 0.00, tolerance = 1e-2)
  expect_equal(unname(Vz["V2"]), 2679.54, tolerance = 1e-2)
  expect_equal(unname(Vz["V3"]), 4099.54, tolerance = 1e-2)
  expect_equal(unname(Vz["V4"]), 3602.51, tolerance = 1e-2)
  expect_equal(unname(Vz["V5"]), 0.00, tolerance = 1e-2)
})

test_that("Pr_vector_disc with zeroized reserves reproduces zeroized profit pattern", {
  Vz <- c(0.00, 0.00, 2679.54, 4099.54, 3602.51, 0.00)
  qx <- c(.015, .017, .019, .021, .024)

  Pr <- Pr_vector_disc(
    V = Vz,
    G = 19279,
    i = 0.06,
    r = 0,
    e = 240,
    q1 = qx,
    b1 = 1000000,
    pre_contract_expense = 5000
  )

  Pi <- Pi_signature(Pr, p_tau = c(.98500, .98300, .98100, .97900, .97600))
  npv <- NPV_profit(Pi, r = 0.10)

  expect_equal(unname(Pr["Pr1"]), 5181.34, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr2"]), 547.35, tolerance = 1e-1)
  expect_equal(unname(Pr["Pr3"]), 0.00, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr4"]), 0.00, tolerance = 1e-2)
  expect_equal(unname(Pr["Pr5"]), 0.00, tolerance = 1e-2)

  expect_equal(unname(Pi["Pi2"]), 539.14, tolerance = 1e-1)
  expect_equal(npv, 155.88, tolerance = 1e-1)
})



testthat::test_that("profit vector follows the discrete profit formula", {
  V <- c(0, 5.66, 6.17, 0)
  q1 <- c(0.00142, 0.00153, 0.00166)

  out <- Pr_vector_disc(
    V = V,
    G = 95,
    i = 0.06,
    r = 0.05,
    e = 10,
    q1 = q1,
    b1 = 50000,
    pre_contract_expense = 15
  )

  expected_year_1 <- (
    V[1] + 95 * (1 - 0.05) - 10
  ) * 1.06 - (
    50000 * q1[1] +
      V[2] * (1 - q1[1])
  )

  testthat::expect_equal(unname(out[1]), -15)
  testthat::expect_equal(unname(out[2]), expected_year_1, tolerance = 1e-12)
  testthat::expect_named(out, paste0("Pr", 0:3))
})


testthat::test_that("profit vector supports two decrements", {
  out <- Pr_vector_disc(
    V = c(0, 10, 0),
    G = 100,
    i = 0.05,
    r = 0.10,
    e = 5,
    q1 = c(0.01, 0.02),
    q2 = c(0.05, 0.04),
    b1 = 1000,
    b2 = 50,
    s1 = 10,
    s2 = 2
  )

  testthat::expect_length(out, 3)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("profit vector accepts negative interest rates above minus one", {
  out <- Pr_vector_disc(
    V = c(0, 0),
    G = 100,
    i = -0.01,
    q1 = 0.01,
    b1 = 1000
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("profit vector rejects excessive total decrement probability", {
  testthat::expect_error(
    Pr_vector_disc(
      V = c(0, 0),
      G = 100,
      i = 0.05,
      q1 = 0.70,
      q2 = 0.40,
      b1 = 1000
    ),
    "q1 \\+ q2 must not exceed 1"
  )
})


testthat::test_that("profit vector rejects invalid expense percentages", {
  testthat::expect_error(
    Pr_vector_disc(
      V = c(0, 0),
      G = 100,
      i = 0.05,
      r = 1.20,
      q1 = 0.01,
      b1 = 1000
    ),
    "values in \\[0, 1\\]"
  )
})


testthat::test_that("pre-contract expense must be scalar", {
  testthat::expect_error(
    Pr_vector_disc(
      V = c(0, 0),
      G = 100,
      i = 0.05,
      q1 = 0.01,
      b1 = 1000,
      pre_contract_expense = c(10, 20)
    ),
    "must be a numeric scalar"
  )
})


testthat::test_that("profit signature uses start-of-year persistency", {
  Pr <- c(-15, 8, 9, 10)
  p_tau <- c(0.90, 0.80, 0.70)

  out <- Pi_signature(Pr, p_tau)

  expected <- c(
    -15,
    8,
    9 * 0.90,
    10 * 0.90 * 0.80
  )

  testthat::expect_equal(unname(out), expected)
  testthat::expect_named(out, paste0("Pi", 0:3))
})


testthat::test_that("profit signature handles one-year contracts", {
  testthat::expect_equal(
    Pi_signature(
      Pr = c(-10, 25),
      p_tau = numeric(0)
    ),
    c(Pi0 = -10, Pi1 = 25)
  )

  testthat::expect_equal(
    Pi_signature(
      Pr = c(-10, 25),
      p_tau = 0.95
    ),
    c(Pi0 = -10, Pi1 = 25)
  )
})


testthat::test_that("NPV calculation matches direct discounting", {
  Pi <- c(-15, 8.42, 8.39, 8.58)
  r <- 0.10

  expected <- sum(Pi / (1 + r)^(0:3))

  testthat::expect_equal(
    NPV_profit(Pi, r),
    expected,
    tolerance = 1e-12
  )
})


testthat::test_that("NPV vectorizes over discount rates", {
  Pi <- c(-15, 8.42, 8.39, 8.58)
  rates <- c(0.08, 0.10, 0.12)

  out <- NPV_profit(Pi, rates)

  expected <- vapply(
    rates,
    function(rate) sum(Pi / (1 + rate)^(0:3)),
    numeric(1)
  )

  testthat::expect_equal(out, expected, tolerance = 1e-12)
})


testthat::test_that("NPV accepts negative rates above minus one", {
  Pi <- c(-100, 110)

  testthat::expect_equal(
    NPV_profit(Pi, r = -0.10),
    -100 + 110 / 0.90
  )

  testthat::expect_error(
    NPV_profit(Pi, r = -1),
    "greater than -1"
  )
})


testthat::test_that("partial NPV preserves vector output for scalar rate", {
  Pi <- c(-15, 8.42, 8.39, 8.58)

  out <- NPV_partial(Pi, 0.10)

  expected <- cumsum(Pi / 1.10^(0:3))

  testthat::expect_type(out, "double")
  testthat::expect_equal(unname(out), expected)
  testthat::expect_named(
    out,
    paste0("NPV(", 0:3, ")")
  )
})


testthat::test_that("partial NPV returns a matrix for vectorized rates", {
  Pi <- c(-15, 8.42, 8.39, 8.58)

  out <- NPV_partial(
    Pi,
    r = c(0.08, 0.10)
  )

  testthat::expect_true(is.matrix(out))
  testthat::expect_equal(dim(out), c(2, 4))
  testthat::expect_equal(
    colnames(out),
    paste0("NPV(", 0:3, ")")
  )
})


testthat::test_that("discounted payback period uses partial NPV", {
  Pi <- c(-15, 8.42, 8.39, 8.58)

  partial <- NPV_partial(Pi, r = 0.10)
  expected <- which(partial >= 0)[1] - 1L

  testthat::expect_equal(
    discounted_payback_period(Pi, r = 0.10),
    as.integer(expected)
  )
})


testthat::test_that("discounted payback period vectorizes over rates", {
  Pi <- c(-15, 8.42, 8.39, 8.58)

  out <- discounted_payback_period(
    Pi,
    r = c(0.05, 0.10, 0.20)
  )

  testthat::expect_length(out, 3)
  testthat::expect_true(all(is.na(out) | out >= 0))
})


testthat::test_that("IRR produces approximately zero NPV", {
  Pi <- c(-15, 8.42, 8.39, 8.58)

  irr <- IRR_profit(Pi)

  testthat::expect_equal(
    NPV_profit(Pi, irr),
    0,
    tolerance = 1e-8
  )
})


testthat::test_that("IRR can search over a negative interval", {
  Pi <- c(-100, 90)

  irr <- IRR_profit(
    Pi,
    interval = c(-0.50, 0)
  )

  testthat::expect_equal(irr, -0.10, tolerance = 1e-8)
})


testthat::test_that("IRR rejects an interval without a sign change", {
  testthat::expect_error(
    IRR_profit(
      Pi = c(10, 10),
      interval = c(0, 1)
    ),
    "does not change sign"
  )
})


testthat::test_that("gross premium APV uses start-of-year persistency", {
  G <- rep(95, 3)
  p_tau <- c(0.90, 0.80, 0.70)
  r <- 0.10

  expected <- sum(
    G *
      c(1, 0.90, 0.90 * 0.80) /
      1.10^(0:2)
  )

  testthat::expect_equal(
    APV_gross_premiums(G, r, p_tau),
    expected,
    tolerance = 1e-12
  )
})


testthat::test_that("gross premium APV handles one premium", {
  testthat::expect_equal(
    APV_gross_premiums(
      G = 100,
      r = 0.10,
      p_tau = numeric(0)
    ),
    100
  )
})


testthat::test_that("gross premium APV vectorizes over rates", {
  out <- APV_gross_premiums(
    G = rep(95, 3),
    r = c(0.08, 0.10),
    p_tau = c(0.90, 0.80)
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("profit margin vectorizes with scalar recycling", {
  out <- profit_margin(
    NPV = c(5, 10),
    APV_GP = 100
  )

  testthat::expect_equal(out, c(0.05, 0.10))
})


testthat::test_that("profit margin requires positive premium APV", {
  testthat::expect_error(
    profit_margin(
      NPV = 5,
      APV_GP = 0
    ),
    "positive values"
  )
})


testthat::test_that("zeroized reserves satisfy the backward recursion", {
  qx <- c(0.01, 0.02)
  i <- 0.05
  G <- 20
  benefit <- 1000
  e <- 2

  out <- V_zeroized(
    qx = qx,
    i = i,
    G = G,
    benefit = benefit,
    e = e,
    floor_zero = FALSE
  )

  expected_v1 <- (
    benefit * qx[2] +
      0 * (1 - qx[2])
  ) / 1.05 - (G - e)

  expected_v0 <- (
    benefit * qx[1] +
      expected_v1 * (1 - qx[1])
  ) / 1.05 - (G - e)

  testthat::expect_equal(
    unname(out),
    c(expected_v0, expected_v1, 0),
    tolerance = 1e-12
  )
})


testthat::test_that("zeroized reserve floor removes negative reserves", {
  out <- V_zeroized(
    qx = c(0.001, 0.001),
    i = 0.05,
    G = 100,
    benefit = 1000,
    floor_zero = TRUE
  )

  testthat::expect_true(all(out >= 0))
})


testthat::test_that("zeroized reserves support vectorized yearly assumptions", {
  out <- V_zeroized(
    qx = c(0.01, 0.02, 0.03),
    i = c(0.03, 0.04, 0.05),
    G = c(20, 21, 22),
    benefit = c(1000, 900, 800),
    r = c(0.05, 0.06, 0.07),
    e = c(2, 2, 3)
  )

  testthat::expect_length(out, 4)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("zeroized reserves validate logical and percentage inputs", {
  testthat::expect_error(
    V_zeroized(
      qx = 0.01,
      i = 0.05,
      G = 20,
      benefit = 1000,
      r = 1.2
    ),
    "values in \\[0, 1\\]"
  )

  testthat::expect_error(
    V_zeroized(
      qx = 0.01,
      i = 0.05,
      G = 20,
      benefit = 1000,
      floor_zero = NA
    ),
    "must be TRUE or FALSE"
  )
})


testthat::test_that("profit-analysis functions reject incompatible lengths", {
  testthat::expect_error(
    Pr_vector_disc(
      V = c(0, 1, 2, 0),
      G = c(100, 100),
      i = 0.05,
      q1 = c(0.01, 0.02, 0.03),
      b1 = 1000
    ),
    "length 1 or 3"
  )

  testthat::expect_error(
    profit_margin(
      NPV = c(1, 2),
      APV_GP = c(100, 200, 300)
    ),
    "length 1 or the common length"
  )
})

#' mqriskR: Models for Quantifying Risk
#'
#' Actuarial functions for survival models, life contingencies, reserves,
#' multiple-decrement models, variable interest, universal life, profit
#' analysis, and pension mathematics.
#'
#' @importFrom stats integrate uniroot
#' @importFrom utils head tail
#' @keywords internal
"_PACKAGE"

#' Varying-payment annuity functions
#'
#' Increasing and decreasing life annuity functions.
#'
#' @param x Age.
#' @param n Term in years.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object for discrete functions.
#' @param model Optional survival model name.
#' @param ... Additional model parameters.
#' @param k_max Maximum summation horizon for non-terminating models.
#' @param tol Truncation tolerance for non-terminating models.
#'
#' @return
#' Numeric vector containing the requested increasing or decreasing annuity value.
#'
#' @name annuity_varying_payments
NULL

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.check_annuity_vp_i <- function(i) {
  i <- as.numeric(i)
  if (length(i) == 0L || any(!is.finite(i)) || any(i <= -1)) {
    stop("i must contain finite values greater than -1.", call. = FALSE)
  }
  i
}

.recycle_vp_vectors <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) stop("All arguments must have positive length.", call. = FALSE)
  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)
  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) if (length(z) == 1L) rep(z, common_len) else z)
}

.check_vp_x <- function(x) {
  if (any(x < 0)) stop("x must be nonnegative.", call. = FALSE)
  invisible(TRUE)
}

.check_vp_n <- function(n) {
  if (any(n < 0)) stop("n must be nonnegative.", call. = FALSE)
  invisible(TRUE)
}

.check_vp_integer_n <- function(n) {
  if (length(n) != 1L || !is.finite(n) || n < 0) {
    stop("n must be a single nonnegative finite number.", call. = FALSE)
  }
  if (abs(n - round(n)) > 1e-10) {
    stop("For annual annuity functions, n must be an integer.", call. = FALSE)
  }
  as.integer(round(n))
}

.get_vp_npx <- function(x, n, tbl = NULL, model = NULL, ...) {
  if (!is.null(tbl)) {
    return(npx(tbl, x = x, n = n))
  }
  if (is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }
  tpx(n, x = x, model = model, ...)
}

.max_term_vp <- function(x, tbl = NULL, model = NULL, ...) {
  if (!is.null(tbl)) {
    return(max(0, floor(max(tbl$x, na.rm = TRUE) - x)))
  }

  if (is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }

  p <- list(...)
  model <- tolower(model)

  if (model == "uniform") {
    if (is.null(p$omega)) {
      stop("For model = 'uniform', omega must be supplied.", call. = FALSE)
    }
    return(max(0, floor(p$omega - x)))
  }

  Inf
}

.integrate_vp <- function(f, lower, upper) {
  stats::integrate(
    f,
    lower = lower,
    upper = upper,
    rel.tol = 1e-9,
    subdivisions = 1000L
  )$value
}

.upper_cont_vp <- function(x, n = NULL, model, ..., tol = 1e-10) {
  max_term <- .max_term_vp(x, model = model, ...)
  if (is.finite(max_term)) {
    if (is.null(n)) return(max_term)
    return(min(max_term, n))
  }

  if (!is.null(n)) return(n)

  upper <- 10
  repeat {
    surv <- tpx(upper, x = x, model = model, ...)
    if (!is.finite(surv) || surv <= tol || upper >= 5000) break
    upper <- upper * 2
  }

  upper
}

# -------------------------------------------------------------------------
# Annual immediate
# -------------------------------------------------------------------------

#' Increasing whole life annuity-immediate
#'
#' @rdname annuity_varying_payments
#' @export
Iax <- function(x, i, model = NULL, ..., tbl = NULL,
                k_max = 5000, tol = 1e-12) {
  args <- .recycle_vp_vectors(x, .check_annuity_vp_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_vp_x(x)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    v <- 1 / (1 + i[j])
    max_term <- .max_term_vp(xx, tbl = tbl, model = model, ...)

    times <- if (is.finite(max_term)) {
      if (max_term <= 0) integer(0) else 1:max_term
    } else {
      1:k_max
    }

    if (length(times) == 0L) return(0)

    surv <- .get_vp_npx(xx, times, tbl = tbl, model = model, ...)
    surv[is.na(surv)] <- 0

    terms <- times * v^times * surv

    if (is.finite(max_term)) return(sum(terms))

    acc <- 0
    small_run <- 0L
    for (jj in seq_along(terms)) {
      acc <- acc + terms[jj]
      if (terms[jj] < tol) {
        small_run <- small_run + 1L
        if (small_run >= 10L) break
      } else {
        small_run <- 0L
      }
    }
    acc
  })
}

#' Increasing temporary annuity-immediate
#'
#' @rdname annuity_varying_payments
#' @export
Iaxn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_vp_vectors(x, n, .check_annuity_vp_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_vp_x(x)
  .check_vp_n(n)

  sapply(seq_along(x), function(j) {
    nn <- .check_vp_integer_n(n[j])
    if (nn == 0L) return(0)

    v <- 1 / (1 + i[j])
    times <- 1:nn

    surv <- .get_vp_npx(x[j], times, tbl = tbl, model = model, ...)
    surv[is.na(surv)] <- 0

    sum(times * v^times * surv)
  })
}

#' Decreasing temporary annuity-immediate
#'
#' @rdname annuity_varying_payments
#' @export
Daxn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_vp_vectors(x, n, .check_annuity_vp_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_vp_x(x)
  .check_vp_n(n)

  sapply(seq_along(x), function(j) {
    nn <- .check_vp_integer_n(n[j])
    if (nn == 0L) return(0)

    v <- 1 / (1 + i[j])
    times <- 1:nn

    surv <- .get_vp_npx(x[j], times, tbl = tbl, model = model, ...)
    surv[is.na(surv)] <- 0

    sum((nn + 1 - times) * v^times * surv)
  })
}

# -------------------------------------------------------------------------
# Annual due
# -------------------------------------------------------------------------

#' Increasing whole life annuity-due
#'
#' @rdname annuity_varying_payments
#' @export
Iadotx <- function(x, i, model = NULL, ..., tbl = NULL,
                   k_max = 5000, tol = 1e-12) {
  args <- .recycle_vp_vectors(x, .check_annuity_vp_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_vp_x(x)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    v <- 1 / (1 + i[j])
    max_term <- .max_term_vp(xx, tbl = tbl, model = model, ...)

    times <- if (is.finite(max_term)) 0:max_term else 0:k_max

    surv <- .get_vp_npx(xx, times, tbl = tbl, model = model, ...)
    surv[is.na(surv)] <- 0

    terms <- (times + 1) * v^times * surv

    if (is.finite(max_term)) return(sum(terms))

    acc <- 0
    small_run <- 0L
    for (jj in seq_along(terms)) {
      acc <- acc + terms[jj]
      if (terms[jj] < tol) {
        small_run <- small_run + 1L
        if (small_run >= 10L) break
      } else {
        small_run <- 0L
      }
    }
    acc
  })
}

#' Increasing temporary annuity-due
#'
#' @rdname annuity_varying_payments
#' @export
Iadotxn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_vp_vectors(x, n, .check_annuity_vp_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_vp_x(x)
  .check_vp_n(n)

  sapply(seq_along(x), function(j) {
    nn <- .check_vp_integer_n(n[j])
    if (nn == 0L) return(0)

    v <- 1 / (1 + i[j])
    times <- 0:(nn - 1L)

    surv <- .get_vp_npx(x[j], times, tbl = tbl, model = model, ...)
    surv[is.na(surv)] <- 0

    sum((times + 1) * v^times * surv)
  })
}

#' Decreasing temporary annuity-due
#'
#' @rdname annuity_varying_payments
#' @export
Dadotxn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_vp_vectors(x, n, .check_annuity_vp_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_vp_x(x)
  .check_vp_n(n)

  sapply(seq_along(x), function(j) {
    nn <- .check_vp_integer_n(n[j])
    if (nn == 0L) return(0)

    v <- 1 / (1 + i[j])
    times <- 0:(nn - 1L)

    surv <- .get_vp_npx(x[j], times, tbl = tbl, model = model, ...)
    surv[is.na(surv)] <- 0

    sum((nn - times) * v^times * surv)
  })
}

# -------------------------------------------------------------------------
# Continuous
# -------------------------------------------------------------------------

#' Increasing continuous whole life annuity
#'
#' @rdname annuity_varying_payments
#' @export
Iabarx <- function(x, i, model, ..., tol = 1e-10) {
  args <- .recycle_vp_vectors(x, .check_annuity_vp_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_vp_x(x)

  sapply(seq_along(x), function(j) {
    delta <- log(1 + i[j])
    upper <- .upper_cont_vp(x[j], model = model, ..., tol = tol)

    if (upper <= 0) return(0)

    f <- function(t) {
      t * exp(-delta * t) * tpx(t, x = x[j], model = model, ...)
    }

    .integrate_vp(f, 0, upper)
  })
}

#' Increasing continuous temporary annuity
#'
#' @rdname annuity_varying_payments
#' @export
Iabarxn <- function(x, n, i, model, ...) {
  args <- .recycle_vp_vectors(x, n, .check_annuity_vp_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_vp_x(x)
  .check_vp_n(n)

  sapply(seq_along(x), function(j) {
    if (n[j] == 0) return(0)

    delta <- log(1 + i[j])

    f <- function(t) {
      t * exp(-delta * t) * tpx(t, x = x[j], model = model, ...)
    }

    .integrate_vp(f, 0, n[j])
  })
}

#' Decreasing continuous temporary annuity
#'
#' @rdname annuity_varying_payments
#' @export
Dabarxn <- function(x, n, i, model, ...) {
  args <- .recycle_vp_vectors(x, n, .check_annuity_vp_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_vp_x(x)
  .check_vp_n(n)

  sapply(seq_along(x), function(j) {
    if (n[j] == 0) return(0)

    delta <- log(1 + i[j])

    f <- function(t) {
      (n[j] - t) * exp(-delta * t) * tpx(t, x = x[j], model = model, ...)
    }

    .integrate_vp(f, 0, n[j])
  })
}

test_that("Iax equals manual sum under uniform model", {
  x <- 40
  i <- 0.05
  v <- 1 / (1 + i)
  omega <- 100

  times <- 1:(omega - x)
  expected <- sum(times * (v^times) * (1 - times / (omega - x)))

  expect_equal(
    Iax(x, i, model = "uniform", omega = omega),
    expected,
    tolerance = 1e-10
  )
})

test_that("Iaxn equals manual finite sum", {
  x <- 40
  n <- 10
  i <- 0.05
  v <- 1 / (1 + i)
  omega <- 100

  times <- 1:n
  expected <- sum(times * (v^times) * (1 - times / (omega - x)))

  expect_equal(
    Iaxn(x, n, i, model = "uniform", omega = omega),
    expected,
    tolerance = 1e-10
  )
})

test_that("Daxn equals manual finite sum", {
  x <- 40
  n <- 10
  i <- 0.05
  v <- 1 / (1 + i)
  omega <- 100

  times <- 1:n
  expected <- sum((n + 1 - times) * (v^times) * (1 - times / (omega - x)))

  expect_equal(
    Daxn(x, n, i, model = "uniform", omega = omega),
    expected,
    tolerance = 1e-10
  )
})

test_that("Iadotx equals manual sum under uniform model", {
  x <- 40
  i <- 0.05
  v <- 1 / (1 + i)
  omega <- 100

  times <- 0:(omega - x)
  expected <- sum((times + 1) * (v^times) * (1 - times / (omega - x)))

  expect_equal(
    Iadotx(x, i, model = "uniform", omega = omega),
    expected,
    tolerance = 1e-10
  )
})

test_that("Iadotxn equals manual finite sum", {
  x <- 40
  n <- 10
  i <- 0.05
  v <- 1 / (1 + i)
  omega <- 100

  times <- 0:(n - 1)
  expected <- sum((times + 1) * (v^times) * (1 - times / (omega - x)))

  expect_equal(
    Iadotxn(x, n, i, model = "uniform", omega = omega),
    expected,
    tolerance = 1e-10
  )
})

test_that("Dadotxn equals manual finite sum", {
  x <- 40
  n <- 10
  i <- 0.05
  v <- 1 / (1 + i)
  omega <- 100

  times <- 0:(n - 1)
  expected <- sum((n - times) * (v^times) * (1 - times / (omega - x)))

  expect_equal(
    Dadotxn(x, n, i, model = "uniform", omega = omega),
    expected,
    tolerance = 1e-10
  )
})

test_that("temporary increasing and decreasing immediate annuities add to (n+1) times level annuity", {
  x <- 40
  n <- 10
  i <- 0.05

  inc <- Iaxn(x, n, i, model = "uniform", omega = 100)
  dec <- Daxn(x, n, i, model = "uniform", omega = 100)
  lvl <- axn(x, n, i, model = "uniform", omega = 100)

  expect_equal(inc + dec, (n + 1) * lvl, tolerance = 1e-10)
})

test_that("temporary increasing and decreasing due annuities add to (n+1) times level due annuity", {
  x <- 40
  n <- 10
  i <- 0.05

  inc <- Iadotxn(x, n, i, model = "uniform", omega = 100)
  dec <- Dadotxn(x, n, i, model = "uniform", omega = 100)
  lvl <- adotxn(x, n, i, model = "uniform", omega = 100)

  expect_equal(inc + dec, (n + 1) * lvl, tolerance = 1e-10)
})

test_that("continuous temporary increasing and decreasing annuities add to n times level continuous annuity", {
  x <- 40
  n <- 10
  i <- 0.05

  inc <- Iabarxn(x, n, i, model = "uniform", omega = 100)
  dec <- Dabarxn(x, n, i, model = "uniform", omega = 100)
  lvl <- abarxn(x, n, i, model = "uniform", omega = 100)

  expect_equal(inc + dec, n * lvl, tolerance = 1e-8)
})

test_that("Iabarx matches numerical integral under exponential model", {
  x <- 40
  i <- 0.05
  delta <- log(1 + i)
  lambda <- 0.02

  expected <- 1 / (delta + lambda)^2

  expect_equal(
    Iabarx(x, i, model = "exponential", lambda = lambda),
    expected,
    tolerance = 1e-8
  )
})

test_that("Iabarxn and Dabarxn are nonnegative", {
  x <- 40
  n <- 10
  i <- 0.05

  expect_true(Iabarxn(x, n, i, model = "uniform", omega = 100) >= 0)
  expect_true(Dabarxn(x, n, i, model = "uniform", omega = 100) >= 0)
})

test_that("average of increasing and decreasing temporary annuities matches level annuity identity", {
  x <- 40
  n <- 10
  i <- 0.05

  inc_immed <- Iaxn(x, n, i, model = "uniform", omega = 100)
  dec_immed <- Daxn(x, n, i, model = "uniform", omega = 100)
  lvl_immed <- axn(x, n, i, model = "uniform", omega = 100)

  inc_due <- Iadotxn(x, n, i, model = "uniform", omega = 100)
  dec_due <- Dadotxn(x, n, i, model = "uniform", omega = 100)
  lvl_due <- adotxn(x, n, i, model = "uniform", omega = 100)

  expect_equal(inc_immed + dec_immed, (n + 1) * lvl_immed, tolerance = 1e-10)
  expect_equal(inc_due + dec_due, (n + 1) * lvl_due, tolerance = 1e-10)
})

test_that("functions vectorize over x and n", {
  x <- c(40, 45)
  n <- c(10, 5)
  i <- 0.05

  expect_length(Iax(x, i, model = "uniform", omega = 100), 2)
  expect_length(Iaxn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(Daxn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(Iadotx(x, i, model = "uniform", omega = 100), 2)
  expect_length(Iadotxn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(Dadotxn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(Iabarx(x, i, model = "uniform", omega = 100), 2)
  expect_length(Iabarxn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(Dabarxn(x, n, i, model = "uniform", omega = 100), 2)
})

test_that("functions reject incompatible x and n lengths", {
  expect_error(
    Iaxn(c(40, 45), c(10, 5, 3), 0.05, model = "uniform", omega = 100),
    "compatible lengths"
  )
  expect_error(
    Daxn(c(40, 45), c(10, 5, 3), 0.05, model = "uniform", omega = 100),
    "compatible lengths"
  )
  expect_error(
    Iadotxn(c(40, 45), c(10, 5, 3), 0.05, model = "uniform", omega = 100),
    "compatible lengths"
  )
  expect_error(
    Dadotxn(c(40, 45), c(10, 5, 3), 0.05, model = "uniform", omega = 100),
    "compatible lengths"
  )
  expect_error(
    Iabarxn(c(40, 45), c(10, 5, 3), 0.05, model = "uniform", omega = 100),
    "compatible lengths"
  )
  expect_error(
    Dabarxn(c(40, 45), c(10, 5, 3), 0.05, model = "uniform", omega = 100),
    "compatible lengths"
  )
})

test_that("functions reject invalid interest", {
  expect_error(Iax(40, -1, model = "uniform", omega = 100))
  expect_error(Iadotx(40, -1, model = "uniform", omega = 100))
  expect_error(Iabarx(40, -1, model = "uniform", omega = 100))
})

test_that("annual temporary functions reject non-integer n", {
  expect_error(Iaxn(40, 10.5, 0.05, model = "uniform", omega = 100), "integer")
  expect_error(Daxn(40, 10.5, 0.05, model = "uniform", omega = 100), "integer")
  expect_error(Iadotxn(40, 10.5, 0.05, model = "uniform", omega = 100), "integer")
  expect_error(Dadotxn(40, 10.5, 0.05, model = "uniform", omega = 100), "integer")
})

test_that("varying annuity values are finite in standard cases", {
  vals <- c(
    Iax(40, 0.05, model = "uniform", omega = 100),
    Iaxn(40, 10, 0.05, model = "uniform", omega = 100),
    Daxn(40, 10, 0.05, model = "uniform", omega = 100),
    Iadotx(40, 0.05, model = "uniform", omega = 100),
    Iadotxn(40, 10, 0.05, model = "uniform", omega = 100),
    Dadotxn(40, 10, 0.05, model = "uniform", omega = 100),
    Iabarx(40, 0.05, model = "uniform", omega = 100),
    Iabarxn(40, 10, 0.05, model = "uniform", omega = 100),
    Dabarxn(40, 10, 0.05, model = "uniform", omega = 100)
  )

  expect_true(all(is.finite(vals)))
})

test_that("varying-payment annuities work with life tables", {
  tbl <- life_table(x = 0:3, lx = c(100000, 80000, 50000, 0))
  v <- 1 / 1.05

  expect_equal(
    Iax(0, i = 0.05, tbl = tbl),
    1 * v * 0.8 + 2 * v^2 * 0.5
  )

  expect_equal(
    Iadotx(0, i = 0.05, tbl = tbl),
    1 + 2 * v * 0.8 + 3 * v^2 * 0.5
  )
})

test_that("varying-payment annuities vectorize over i", {
  out <- Iax(
    x = 40,
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  expect_length(out, 2)
  expect_true(out[1] > out[2])
})

#' Annual annuity functions
#'
#' Annual whole life, temporary, deferred, and actuarial accumulated value
#' annuity functions in immediate, due, and continuous forms.
#'
#' @param x Age.
#' @param n Term in years.
#' @param i Effective annual interest rate.
#' @param model Optional survival model name.
#' @param ... Additional model parameters.
#' @param tbl Optional life table object for annual discrete annuity functions.
#' @param k_max Maximum summation horizon for non-terminating models.
#' @param tol Truncation tolerance for non-terminating models.
#'
#' @return
#' Numeric vector containing the requested annuity present value or actuarial
#' accumulated value.
#'
#' @name annuity_annual
NULL

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.check_annuity_i <- function(i) {
  i <- as.numeric(i)

  if (length(i) == 0L || any(!is.finite(i)) || any(i <= -1)) {
    stop("i must contain finite values greater than -1.", call. = FALSE)
  }

  i
}

.check_n_scalar <- function(n) {
  n <- as.numeric(n)

  if (length(n) != 1L || !is.finite(n) || n < 0) {
    stop("n must be a single nonnegative number.", call. = FALSE)
  }

  n
}

.check_integer_term <- function(n) {
  n <- .check_n_scalar(n)

  if (abs(n - round(n)) > 1e-10) {
    stop("For annual annuity functions, n must be an integer.", call. = FALSE)
  }

  as.integer(round(n))
}

.recycle_annuity_vectors <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) {
    stop("All arguments must have positive length.", call. = FALSE)
  }

  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)

  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) {
    if (length(z) == 1L) rep(z, common_len) else z
  })
}

.recycle_xn <- function(x, n) {
  args <- .recycle_annuity_vectors(x, n)
  list(x = args[[1L]], n = args[[2L]])
}

.check_age_annuity <- function(x) {
  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  invisible(TRUE)
}

.get_npx_annuity <- function(x, n, tbl = NULL, model = NULL, ...) {
  if (!is.null(tbl)) {
    return(npx(tbl, x = x, n = n))
  }

  if (is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }

  tpx(n, x = x, model = model, ...)
}

.max_term_from_source <- function(x, tbl = NULL, model = NULL, ...) {
  if (!is.null(tbl)) {
    return(max(0, floor(max(tbl$x, na.rm = TRUE) - x)))
  }

  if (is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }

  dots <- list(...)
  model <- tolower(model)

  if (model == "uniform") {
    return(max(0, floor(dots$omega - x)))
  }

  Inf
}

.sum_annuity_terms <- function(x, i, tbl = NULL, model = NULL, ...,
                               k_max = 5000, tol = 1e-12,
                               due = FALSE, temporary = NULL) {
  x <- as.numeric(x)
  i <- .check_annuity_i(i)

  if (length(x) != 1L || !is.finite(x) || x < 0) {
    stop("x must be a single nonnegative finite number.", call. = FALSE)
  }

  if (!is.null(temporary)) {
    temporary <- .check_integer_term(temporary)
  }

  v <- 1 / (1 + i)

  if (due) {
    if (!is.null(temporary)) {
      if (temporary == 0L) return(0)
      times <- 0:(temporary - 1L)
    } else {
      max_term <- .max_term_from_source(x, tbl = tbl, model = model, ...)
      times <- if (is.finite(max_term)) 0:max_term else 0:k_max
    }
  } else {
    if (!is.null(temporary)) {
      if (temporary == 0L) return(0)
      times <- 1:temporary
    } else {
      max_term <- .max_term_from_source(x, tbl = tbl, model = model, ...)
      if (is.finite(max_term)) {
        if (max_term <= 0) return(0)
        times <- 1:max_term
      } else {
        times <- 1:k_max
      }
    }
  }

  surv <- .get_npx_annuity(x, times, tbl = tbl, model = model, ...)
  surv[is.na(surv)] <- 0

  terms <- v^times * surv

  if (!is.null(temporary) ||
      is.finite(.max_term_from_source(x, tbl = tbl, model = model, ...))) {
    return(sum(terms, na.rm = TRUE))
  }

  acc <- 0
  small_run <- 0

  for (j in seq_along(terms)) {
    acc <- acc + terms[j]

    if (terms[j] < tol) {
      small_run <- small_run + 1L
      if (small_run >= 10L) break
    } else {
      small_run <- 0L
    }
  }

  acc
}

.integrate_annuity <- function(f, lower, upper) {
  stats::integrate(
    f,
    lower = lower,
    upper = upper,
    rel.tol = 1e-9,
    subdivisions = 1000L
  )$value
}

# -------------------------------------------------------------------------
# Whole life annuities
# -------------------------------------------------------------------------

#' Whole life annuity-immediate
#'
#' Computes \eqn{a_x = \sum_{t=1}^{\infty} v^t {}_t p_x}.
#'
#' @rdname annuity_annual
#' @export
ax <- function(x, i, model = NULL, ..., tbl = NULL,
               k_max = 5000, tol = 1e-12) {
  args <- .recycle_annuity_vectors(x, .check_annuity_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    .sum_annuity_terms(
      x[j],
      i[j],
      tbl = tbl,
      model = model,
      ...,
      k_max = k_max,
      tol = tol,
      due = FALSE
    )
  })
}

#' Whole life annuity-due
#'
#' Computes \eqn{\ddot{a}_x = \sum_{t=0}^{\infty} v^t {}_t p_x}.
#'
#' @rdname annuity_annual
#' @export
adotx <- function(x, i, model = NULL, ..., tbl = NULL,
                  k_max = 5000, tol = 1e-12) {
  args <- .recycle_annuity_vectors(x, .check_annuity_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    .sum_annuity_terms(
      x[j],
      i[j],
      tbl = tbl,
      model = model,
      ...,
      k_max = k_max,
      tol = tol,
      due = TRUE
    )
  })
}

#' Continuous whole life annuity
#'
#' Computes \eqn{\bar{a}_x = \int_0^{\infty} v^t {}_t p_x dt}.
#'
#' @rdname annuity_annual
#' @export
abarx <- function(x, i, model, ..., tol = 1e-10) {
  args <- .recycle_annuity_vectors(x, .check_annuity_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    delta <- log(1 + i[j])

    max_term <- .max_term_from_source(xx, model = model, ...)

    if (is.finite(max_term)) {
      if (max_term <= 0) return(0)
      f <- function(t) exp(-delta * t) * tpx(t, x = xx, model = model, ...)
      return(.integrate_annuity(f, 0, max_term))
    }

    if (tolower(model) == "exponential") {
      lambda <- list(...)$lambda
      return(1 / (delta + lambda))
    }

    upper <- 10
    while (upper < 5000 && tpx(upper, xx, model = model, ...) > tol) {
      upper <- upper * 2
    }

    f <- function(t) exp(-delta * t) * tpx(t, x = xx, model = model, ...)
    .integrate_annuity(f, 0, upper)
  })
}

# -------------------------------------------------------------------------
# Temporary annuities
# -------------------------------------------------------------------------

#' Temporary annuity-immediate
#'
#' Computes \eqn{a_{x:\overline{n}|} = \sum_{t=1}^{n} v^t {}_t p_x}.
#'
#' @rdname annuity_annual
#' @export
axn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_integer_term(n[j])

    .sum_annuity_terms(
      x[j],
      i[j],
      tbl = tbl,
      model = model,
      ...,
      due = FALSE,
      temporary = nn
    )
  })
}

#' Temporary annuity-due
#'
#' Computes \eqn{\ddot{a}_{x:\overline{n}|} =
#' \sum_{t=0}^{n-1} v^t {}_t p_x}.
#'
#' @rdname annuity_annual
#' @export
adotxn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_integer_term(n[j])

    .sum_annuity_terms(
      x[j],
      i[j],
      tbl = tbl,
      model = model,
      ...,
      due = TRUE,
      temporary = nn
    )
  })
}

#' Continuous temporary annuity
#'
#' Computes \eqn{\bar{a}_{x:\overline{n}|} =
#' \int_0^n v^t {}_t p_x dt}.
#'
#' @rdname annuity_annual
#' @export
abarxn <- function(x, n, i, model, ...) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_n_scalar(n[j])
    delta <- log(1 + i[j])

    f <- function(t) {
      exp(-delta * t) * tpx(t, x = x[j], model = model, ...)
    }

    .integrate_annuity(f, 0, nn)
  })
}

# -------------------------------------------------------------------------
# Deferred annuities
# -------------------------------------------------------------------------

#' Deferred whole life annuity-immediate
#'
#' Computes \eqn{{}_{n|}a_x = {}_nE_x a_{x+n}}.
#'
#' @rdname annuity_annual
#' @export
nax <- function(x, n, i, model = NULL, ..., tbl = NULL,
                k_max = 5000, tol = 1e-12) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_integer_term(n[j])

    nEx(x[j], nn, i[j], tbl = tbl, model = model, ...) *
      ax(
        x[j] + nn,
        i[j],
        tbl = tbl,
        model = model,
        ...,
        k_max = k_max,
        tol = tol
      )
  })
}

#' Deferred whole life annuity-due
#'
#' Computes \eqn{{}_{n|}\ddot{a}_x = {}_nE_x \ddot{a}_{x+n}}.
#'
#' @rdname annuity_annual
#' @export
nadotx <- function(x, n, i, model = NULL, ..., tbl = NULL,
                   k_max = 5000, tol = 1e-12) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_integer_term(n[j])

    nEx(x[j], nn, i[j], tbl = tbl, model = model, ...) *
      adotx(
        x[j] + nn,
        i[j],
        tbl = tbl,
        model = model,
        ...,
        k_max = k_max,
        tol = tol
      )
  })
}

#' Deferred continuous whole life annuity
#'
#' Computes \eqn{{}_{n|}\bar{a}_x = {}_nE_x \bar{a}_{x+n}}.
#'
#' @rdname annuity_annual
#' @export
nabarx <- function(x, n, i, model, ..., tol = 1e-10) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_n_scalar(n[j])

    nEx(x[j], nn, i[j], model = model, ...) *
      abarx(x[j] + nn, i[j], model = model, ..., tol = tol)
  })
}

# -------------------------------------------------------------------------
# Actuarial accumulated values
# -------------------------------------------------------------------------

#' Temporary annuity-immediate actuarial accumulated value
#'
#' Computes \eqn{s_{x:\overline{n}|} = a_{x:\overline{n}|} / {}_nE_x}.
#'
#' @rdname annuity_annual
#' @export
sxn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_integer_term(n[j])
    ex <- nEx(x[j], nn, i[j], tbl = tbl, model = model, ...)

    if (is.na(ex) || ex <= 0) return(Inf)

    axn(x[j], nn, i[j], tbl = tbl, model = model, ...) / ex
  })
}

#' Temporary annuity-due actuarial accumulated value
#'
#' Computes \eqn{\ddot{s}_{x:\overline{n}|} =
#' \ddot{a}_{x:\overline{n}|} / {}_nE_x}.
#'
#' @rdname annuity_annual
#' @export
sdotxn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_integer_term(n[j])
    ex <- nEx(x[j], nn, i[j], tbl = tbl, model = model, ...)

    if (is.na(ex) || ex <= 0) return(Inf)

    adotxn(x[j], nn, i[j], tbl = tbl, model = model, ...) / ex
  })
}

#' Continuous temporary annuity actuarial accumulated value
#'
#' Computes \eqn{\bar{s}_{x:\overline{n}|} =
#' \bar{a}_{x:\overline{n}|} / {}_nE_x}.
#'
#' @rdname annuity_annual
#' @export
sbarxn <- function(x, n, i, model, ...) {
  args <- .recycle_annuity_vectors(x, n, .check_annuity_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_age_annuity(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_n_scalar(n[j])
    ex <- nEx(x[j], nn, i[j], model = model, ...)

    if (is.na(ex) || ex <= 0) return(Inf)

    abarxn(x[j], nn, i[j], model = model, ...) / ex
  })
}

test_that("ax equals sum of annual pure endowments", {
  library(mqriskR)

  i <- 0.05
  x <- 40

  expected <- sum(sapply(1:(100 - x), function(t) {
    nEx(x, t, i, model = "uniform", omega = 100)
  }))

  expect_equal(
    ax(x, i, model = "uniform", omega = 100),
    expected,
    tolerance = 1e-10
  )
})

test_that("adotx equals 1 + ax", {
  library(mqriskR)

  i <- 0.05
  x <- 40

  expect_equal(
    adotx(x, i, model = "uniform", omega = 100),
    1 + ax(x, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("abarx under exponential model matches closed form", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  lambda <- 0.02
  delta <- log(1 + i)

  expected <- 1 / (delta + lambda)

  expect_equal(
    abarx(x, i, model = "exponential", lambda = lambda),
    expected,
    tolerance = 1e-10
  )
})

test_that("axn equals sum of first n annual pure endowments", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expected <- sum(sapply(1:n, function(t) {
    nEx(x, t, i, model = "uniform", omega = 100)
  }))

  expect_equal(
    axn(x, n, i, model = "uniform", omega = 100),
    expected,
    tolerance = 1e-10
  )
})

test_that("adotxn equals sum from t = 0 to n - 1 of pure endowments", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expected <- sum(sapply(0:(n - 1), function(t) {
    nEx(x, t, i, model = "uniform", omega = 100)
  }))

  expect_equal(
    adotxn(x, n, i, model = "uniform", omega = 100),
    expected,
    tolerance = 1e-10
  )
})

test_that("adotxn equals axn + 1 - nEx", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    adotxn(x, n, i, model = "uniform", omega = 100),
    axn(x, n, i, model = "uniform", omega = 100) +
      1 -
      nEx(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("abarxn equals integral form under exponential model", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10
  lambda <- 0.02
  delta <- log(1 + i)

  expected <- (1 - exp(-(delta + lambda) * n)) / (delta + lambda)

  expect_equal(
    abarxn(x, n, i, model = "exponential", lambda = lambda),
    expected,
    tolerance = 1e-10
  )
})

test_that("nax equals nEx times ax at age x+n", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    nax(x, n, i, model = "uniform", omega = 100),
    nEx(x, n, i, model = "uniform", omega = 100) *
      ax(x + n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("nadotx equals nEx times adotx at age x+n", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    nadotx(x, n, i, model = "uniform", omega = 100),
    nEx(x, n, i, model = "uniform", omega = 100) *
      adotx(x + n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("nabarx equals nEx times abarx at age x+n", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10
  lambda <- 0.02

  expect_equal(
    nabarx(x, n, i, model = "exponential", lambda = lambda),
    nEx(x, n, i, model = "exponential", lambda = lambda) *
      abarx(x + n, i, model = "exponential", lambda = lambda),
    tolerance = 1e-10
  )
})

test_that("whole life annuity decomposes into temporary plus deferred", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    ax(x, i, model = "uniform", omega = 100),
    axn(x, n, i, model = "uniform", omega = 100) +
      nax(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("whole life annuity-due decomposes into temporary plus deferred", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    adotx(x, i, model = "uniform", omega = 100),
    adotxn(x, n, i, model = "uniform", omega = 100) +
      nadotx(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("continuous whole life annuity decomposes into temporary plus deferred", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10
  lambda <- 0.02

  expect_equal(
    abarx(x, i, model = "exponential", lambda = lambda),
    abarxn(x, n, i, model = "exponential", lambda = lambda) +
      nabarx(x, n, i, model = "exponential", lambda = lambda),
    tolerance = 1e-10
  )
})

test_that("sxn equals axn divided by nEx", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    sxn(x, n, i, model = "uniform", omega = 100),
    axn(x, n, i, model = "uniform", omega = 100) /
      nEx(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("sdotxn equals adotxn divided by nEx", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    sdotxn(x, n, i, model = "uniform", omega = 100),
    adotxn(x, n, i, model = "uniform", omega = 100) /
      nEx(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("sbarxn equals abarxn divided by nEx", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10
  lambda <- 0.02

  expect_equal(
    sbarxn(x, n, i, model = "exponential", lambda = lambda),
    abarxn(x, n, i, model = "exponential", lambda = lambda) /
      nEx(x, n, i, model = "exponential", lambda = lambda),
    tolerance = 1e-10
  )
})

test_that("annuity functions vectorize correctly over x", {
  library(mqriskR)

  i <- 0.05
  x <- c(40, 45, 50)
  n <- 10

  expect_length(ax(x, i, model = "uniform", omega = 100), 3)
  expect_length(adotx(x, i, model = "uniform", omega = 100), 3)
  expect_length(abarx(x, i, model = "exponential", lambda = 0.02), 3)
  expect_length(axn(x, n, i, model = "uniform", omega = 100), 3)
  expect_length(adotxn(x, n, i, model = "uniform", omega = 100), 3)
  expect_length(abarxn(x, n, i, model = "exponential", lambda = 0.02), 3)
  expect_length(nax(x, n, i, model = "uniform", omega = 100), 3)
  expect_length(nadotx(x, n, i, model = "uniform", omega = 100), 3)
  expect_length(nabarx(x, n, i, model = "exponential", lambda = 0.02), 3)
})

test_that("annuity functions vectorize correctly over x and n", {
  library(mqriskR)

  i <- 0.05
  x <- c(40, 45)
  n <- c(10, 5)

  expect_length(axn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(adotxn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(abarxn(x, n, i, model = "exponential", lambda = 0.02), 2)
  expect_length(nax(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(nadotx(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(nabarx(x, n, i, model = "exponential", lambda = 0.02), 2)
  expect_length(sxn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(sdotxn(x, n, i, model = "uniform", omega = 100), 2)
  expect_length(sbarxn(x, n, i, model = "exponential", lambda = 0.02), 2)
})

test_that("temporary annuities are not larger than corresponding whole life annuities", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10
  lambda <- 0.02

  expect_true(axn(x, n, i, model = "uniform", omega = 100) <=
                ax(x, i, model = "uniform", omega = 100))

  expect_true(adotxn(x, n, i, model = "uniform", omega = 100) <=
                adotx(x, i, model = "uniform", omega = 100))

  expect_true(abarxn(x, n, i, model = "exponential", lambda = lambda) <=
                abarx(x, i, model = "exponential", lambda = lambda))
})

test_that("due annuities are at least as large as immediate annuities", {
  library(mqriskR)

  i <- 0.05
  x <- 40
  n <- 10

  expect_true(adotx(x, i, model = "uniform", omega = 100) >=
                ax(x, i, model = "uniform", omega = 100))

  expect_true(adotxn(x, n, i, model = "uniform", omega = 100) >=
                axn(x, n, i, model = "uniform", omega = 100))
})

test_that("functions reject incompatible x and n lengths", {
  library(mqriskR)

  expect_error(
    axn(c(40, 45, 50), c(10, 5), 0.05, model = "uniform", omega = 100)
  )

  expect_error(
    nax(c(40, 45, 50), c(10, 5), 0.05, model = "uniform", omega = 100)
  )
})

test_that("annual annuity functions reject non-integer n where appropriate", {
  library(mqriskR)

  expect_error(
    axn(40, 10.5, 0.05, model = "uniform", omega = 100),
    "integer"
  )

  expect_error(
    adotxn(40, 10.5, 0.05, model = "uniform", omega = 100),
    "integer"
  )

  expect_error(
    nax(40, 10.5, 0.05, model = "uniform", omega = 100),
    "integer"
  )

  expect_error(
    nadotx(40, 10.5, 0.05, model = "uniform", omega = 100),
    "integer"
  )
})


testthat::test_that("annual annuities work with life tables", {
  tbl <- life_table(x = 0:3, lx = c(100000, 80000, 50000, 0))
  v <- 1 / 1.05

  testthat::expect_equal(
    ax(0, i = 0.05, tbl = tbl),
    v * 0.8 + v^2 * 0.5
  )

  testthat::expect_equal(
    adotx(0, i = 0.05, tbl = tbl),
    1 + v * 0.8 + v^2 * 0.5
  )
})

testthat::test_that("annual annuities vectorize over interest rates", {
  out <- ax(40, i = c(0.03, 0.05), model = "uniform", omega = 100)

  testthat::expect_length(out, 2)
  testthat::expect_true(out[1] > out[2])
})

# =========================================================
# Pension mathematics functions
# =========================================================

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.pension_check_numeric <- function(x, name, allow_empty = FALSE) {
  if (!is.numeric(x)) {
    stop(name, " must be numeric.", call. = FALSE)
  }

  if (!allow_empty && length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(x))) {
    stop(name, " must contain finite values.", call. = FALSE)
  }

  as.numeric(x)
}

#' @noRd
.pension_check_nonnegative <- function(x, name) {
  x <- .pension_check_numeric(x, name)

  if (any(x < 0)) {
    stop(name, " must contain nonnegative values.", call. = FALSE)
  }

  x
}

#' @noRd
.pension_check_positive <- function(x, name) {
  x <- .pension_check_numeric(x, name)

  if (any(x <= 0)) {
    stop(name, " must contain positive values.", call. = FALSE)
  }

  x
}

#' @noRd
.pension_check_probability <- function(x, name) {
  x <- .pension_check_numeric(x, name)

  if (any(x < 0 | x > 1)) {
    stop(name, " must contain values in [0, 1].", call. = FALSE)
  }

  x
}

#' @noRd
.pension_check_rate <- function(x, name) {
  x <- .pension_check_numeric(x, name)

  if (any(x <= -1)) {
    stop(name, " must contain values greater than -1.", call. = FALSE)
  }

  x
}

#' @noRd
.pension_check_scalar <- function(
    x,
    name,
    nonnegative = FALSE,
    positive = FALSE,
    integer = FALSE) {
  if (positive) {
    x <- .pension_check_positive(x, name)
  } else if (nonnegative) {
    x <- .pension_check_nonnegative(x, name)
  } else {
    x <- .pension_check_numeric(x, name)
  }

  if (length(x) != 1L) {
    stop(name, " must be a numeric scalar.", call. = FALSE)
  }

  if (integer && abs(x - round(x)) > 1e-10) {
    stop(name, " must be an integer.", call. = FALSE)
  }

  if (integer) {
    x <- as.integer(round(x))
  }

  x
}

#' @noRd
.pension_recycle_common <- function(..., .names = NULL) {
  values <- list(...)
  lengths <- vapply(values, length, integer(1))

  if (any(lengths == 0L)) {
    stop("Inputs must have positive length.", call. = FALSE)
  }

  common_length <- max(lengths)

  if (is.null(.names)) {
    .names <- paste0("Argument ", seq_along(values))
  }

  if (length(.names) != length(values)) {
    stop(
      ".names must have the same length as the supplied arguments.",
      call. = FALSE
    )
  }

  for (j in seq_along(values)) {
    if (!lengths[j] %in% c(1L, common_length)) {
      stop(
        .names[j],
        " must have length 1 or the common length ",
        common_length,
        ".",
        call. = FALSE
      )
    }

    values[[j]] <- rep(values[[j]], length.out = common_length)
  }

  values
}

#' @noRd
.pension_salary_path <- function(x, z, Sx, g = NULL, s = NULL) {
  x <- .pension_check_scalar(x, "x")
  z <- .pension_check_scalar(z, "z")
  Sx <- .pension_check_scalar(Sx, "Sx", positive = TRUE)

  if (z <= x) {
    stop("z must be greater than x.", call. = FALSE)
  }

  n_raw <- z - x

  if (abs(n_raw - round(n_raw)) > 1e-10) {
    stop("z - x must be an integer number of years.", call. = FALSE)
  }

  n <- as.integer(round(n_raw))

  if (!is.null(g) && !is.null(s)) {
    stop("Supply only one of g or s.", call. = FALSE)
  }

  if (is.null(g) && is.null(s)) {
    stop("Supply one of g or s.", call. = FALSE)
  }

  if (!is.null(g)) {
    g <- .pension_check_scalar(g, "g")

    if (g <= -1) {
      stop("g must be greater than -1.", call. = FALSE)
    }

    return(Sx * (1 + g)^(0:(n - 1L)))
  }

  s <- .pension_check_positive(s, "s")

  if (length(s) != n) {
    stop("s must have length z - x.", call. = FALSE)
  }

  Sx * s / s[1L]
}

# -------------------------------------------------------------------------
# Salary scales and defined contribution plans
# -------------------------------------------------------------------------

#' Salary scale under constant annual growth
#'
#' Constructs salary-scale values under a constant annual growth rate.
#'
#' @param k Numeric vector of ages or durations.
#' @param g Annual salary growth rate greater than \code{-1}.
#' @param base_age Scalar age or duration at which the scale is normalized.
#' @param s_base Positive scalar salary-scale value at \code{base_age}.
#'
#' @return A numeric vector with the same length as \code{k}.
#'
#' @examples
#' salary_scale(k = 30:34, g = 0.04, base_age = 30)
#'
#' @export
salary_scale <- function(k, g, base_age = min(k), s_base = 1) {
  k <- .pension_check_numeric(k, "k")
  g <- .pension_check_scalar(g, "g")
  base_age <- .pension_check_scalar(base_age, "base_age")
  s_base <- .pension_check_scalar(
    s_base,
    "s_base",
    positive = TRUE
  )

  if (g <= -1) {
    stop("g must be greater than -1.", call. = FALSE)
  }

  s_base * (1 + g)^(k - base_age)
}

#' Accumulated value of defined contribution plan contributions
#'
#' Computes the accumulated value at retirement from contributions paid at
#' the beginning of each year and accumulated to retirement.
#'
#' @param x Scalar entry age.
#' @param z Scalar retirement age.
#' @param Sx Positive scalar salary at age \code{x}.
#' @param c Scalar contribution rate in \code{[0, 1]}.
#' @param i Scalar annual effective investment return greater than \code{-1}.
#' @param g Optional scalar annual salary growth rate greater than \code{-1}.
#' @param s Optional positive salary-scale vector of length \code{z - x}.
#'
#' @return A numeric scalar.
#'
#' @examples
#' AVz_dc(
#'   x = 30,
#'   z = 65,
#'   Sx = 50000,
#'   c = 0.10,
#'   i = 0.05,
#'   g = 0.04
#' )
#'
#' @export
AVz_dc <- function(x, z, Sx, c, i, g = NULL, s = NULL) {
  c <- .pension_check_scalar(c, "c")

  if (c < 0 || c > 1) {
    stop("c must be between 0 and 1.", call. = FALSE)
  }

  i <- .pension_check_scalar(i, "i")

  if (i <= -1) {
    stop("i must be greater than -1.", call. = FALSE)
  }

  salary <- .pension_salary_path(
    x = x,
    z = z,
    Sx = Sx,
    g = g,
    s = s
  )

  n <- length(salary)
  times <- 0:(n - 1L)

  sum(c * salary * (1 + i)^(n - times))
}

#' Retirement income from a defined contribution accumulation
#'
#' Converts an accumulated account value into annual annuity-due income.
#' Scalar arguments are recycled to a common length.
#'
#' @param AVz Nonnegative accumulated value at retirement.
#' @param adue_z Positive whole-life annuity-due factor at retirement.
#'
#' @return A numeric vector.
#'
#' @examples
#' Income_dc(AVz = 824211.35, adue_z = 12)
#'
#' @export
Income_dc <- function(AVz, adue_z) {
  AVz <- .pension_check_nonnegative(AVz, "AVz")
  adue_z <- .pension_check_positive(adue_z, "adue_z")

  values <- .pension_recycle_common(
    AVz,
    adue_z,
    .names = c("AVz", "adue_z")
  )

  values[[1L]] / values[[2L]]
}

#' Replacement ratio for a defined contribution plan
#'
#' Computes annual retirement income divided by salary in the final
#' pre-retirement year.
#'
#' @inheritParams AVz_dc
#' @param adue_z Positive scalar whole-life annuity-due factor at retirement.
#'
#' @return A numeric scalar.
#'
#' @examples
#' replacement_ratio_dc(
#'   x = 30,
#'   z = 65,
#'   Sx = 50000,
#'   c = 0.10,
#'   i = 0.05,
#'   adue_z = 12,
#'   g = 0.04
#' )
#'
#' @export
replacement_ratio_dc <- function(
    x,
    z,
    Sx,
    c,
    i,
    adue_z,
    g = NULL,
    s = NULL) {
  adue_z <- .pension_check_scalar(
    adue_z,
    "adue_z",
    positive = TRUE
  )

  accumulated_value <- AVz_dc(
    x = x,
    z = z,
    Sx = Sx,
    c = c,
    i = i,
    g = g,
    s = s
  )

  salary <- .pension_salary_path(
    x = x,
    z = z,
    Sx = Sx,
    g = g,
    s = s
  )

  Income_dc(accumulated_value, adue_z) /
    tail(salary, 1L)
}

#' Target contribution rate for a defined contribution plan
#'
#' Calculates the contribution rate required to achieve a target replacement
#' ratio.
#'
#' @param x Scalar entry age.
#' @param z Scalar retirement age.
#' @param Sx Positive scalar salary at age \code{x}.
#' @param RR_target Scalar target replacement ratio in \code{[0, 1]}.
#' @param i Scalar annual effective investment return greater than \code{-1}.
#' @param adue_z Positive scalar whole-life annuity-due factor at retirement.
#' @param g Optional scalar annual salary growth rate greater than \code{-1}.
#' @param s Optional positive salary-scale vector of length \code{z - x}.
#'
#' @return A numeric scalar. The result may exceed one when the target cannot
#'   be achieved with a contribution rate no greater than 100 percent.
#'
#' @examples
#' contribution_rate_target(
#'   x = 30,
#'   z = 65,
#'   Sx = 60000,
#'   RR_target = 0.50,
#'   i = 0.06,
#'   adue_z = 11,
#'   g = 0.04
#' )
#'
#' @export
contribution_rate_target <- function(
    x,
    z,
    Sx,
    RR_target,
    i,
    adue_z,
    g = NULL,
    s = NULL) {
  RR_target <- .pension_check_scalar(
    RR_target,
    "RR_target"
  )

  if (RR_target < 0 || RR_target > 1) {
    stop(
      "RR_target must be between 0 and 1.",
      call. = FALSE
    )
  }

  i <- .pension_check_scalar(i, "i")

  if (i <= -1) {
    stop("i must be greater than -1.", call. = FALSE)
  }

  adue_z <- .pension_check_scalar(
    adue_z,
    "adue_z",
    positive = TRUE
  )

  salary <- .pension_salary_path(
    x = x,
    z = z,
    Sx = Sx,
    g = g,
    s = s
  )

  n <- length(salary)
  times <- 0:(n - 1L)

  numerator <-
    RR_target *
    tail(salary, 1L) *
    adue_z

  denominator <-
    sum(salary * (1 + i)^(n - times))

  numerator / denominator
}

# -------------------------------------------------------------------------
# Defined benefit formulas
# -------------------------------------------------------------------------

#' Projected annual benefit under a final-average-salary plan
#'
#' Projects the annual benefit using a final-average-salary formula.
#'
#' @param x Scalar current or entry age.
#' @param z Scalar retirement age.
#' @param CASx Positive scalar current annual salary.
#' @param p Nonnegative scalar accrual percentage, such as \code{2} for
#'   2 percent.
#' @param fas_years Positive integer number of years in the final salary
#'   average.
#' @param past_service Nonnegative scalar years of service already completed.
#' @param g Optional scalar annual salary growth rate greater than \code{-1}.
#' @param s Optional positive salary-scale vector of length \code{z - x}.
#'
#' @return A numeric scalar.
#'
#' @examples
#' PAB_fas(
#'   x = 35,
#'   z = 65,
#'   CASx = 60000,
#'   p = 2,
#'   fas_years = 3,
#'   g = 0.04
#' )
#'
#' @export
PAB_fas <- function(
    x,
    z,
    CASx,
    p,
    fas_years = 3,
    past_service = 0,
    g = NULL,
    s = NULL) {
  p <- .pension_check_scalar(
    p,
    "p",
    nonnegative = TRUE
  )

  fas_years <- .pension_check_scalar(
    fas_years,
    "fas_years",
    positive = TRUE,
    integer = TRUE
  )

  past_service <- .pension_check_scalar(
    past_service,
    "past_service",
    nonnegative = TRUE
  )

  salary <- .pension_salary_path(
    x = x,
    z = z,
    Sx = CASx,
    g = g,
    s = s
  )

  n <- length(salary)

  if (fas_years > n) {
    stop(
      "fas_years cannot exceed z - x.",
      call. = FALSE
    )
  }

  final_average_salary <-
    mean(tail(salary, fas_years))

  total_service <- past_service + n

  0.01 * p * total_service * final_average_salary
}

#' Projected annual benefit under a career-average-earnings plan
#'
#' Projects the annual benefit using a career-average-earnings formula.
#'
#' @param x Scalar current or entry age.
#' @param z Scalar retirement age.
#' @param CASx Positive scalar current annual salary.
#' @param p Nonnegative scalar accrual percentage.
#' @param past_salary_total Nonnegative total of actual prior salaries.
#' @param g Optional scalar annual salary growth rate greater than \code{-1}.
#' @param s Optional positive salary-scale vector of length \code{z - x}.
#'
#' @return A numeric scalar.
#'
#' @examples
#' PAB_cae(
#'   x = 30,
#'   z = 65,
#'   CASx = 100000,
#'   p = 1,
#'   g = 0.04
#' )
#'
#' @export
PAB_cae <- function(
    x,
    z,
    CASx,
    p,
    past_salary_total = 0,
    g = NULL,
    s = NULL) {
  p <- .pension_check_scalar(
    p,
    "p",
    nonnegative = TRUE
  )

  past_salary_total <- .pension_check_scalar(
    past_salary_total,
    "past_salary_total",
    nonnegative = TRUE
  )

  salary <- .pension_salary_path(
    x = x,
    z = z,
    Sx = CASx,
    g = g,
    s = s
  )

  projected_salary_total <-
    past_salary_total + sum(salary)

  0.01 * p * projected_salary_total
}

#' Replacement ratio for a defined benefit plan
#'
#' Computes annual benefit divided by a selected salary measure. Scalar
#' arguments are recycled to a common length.
#'
#' @param benefit Nonnegative annual retirement benefit.
#' @param salary Positive salary measure used in the denominator.
#'
#' @return A numeric vector.
#'
#' @examples
#' replacement_ratio_db(
#'   benefit = 108008.66,
#'   salary = 187119.09
#' )
#'
#' @export
replacement_ratio_db <- function(benefit, salary) {
  benefit <- .pension_check_nonnegative(
    benefit,
    "benefit"
  )

  salary <- .pension_check_positive(
    salary,
    "salary"
  )

  values <- .pension_recycle_common(
    benefit,
    salary,
    .names = c("benefit", "salary")
  )

  values[[1L]] / values[[2L]]
}

#' Accrued benefit under a final-average-salary plan
#'
#' Computes the accrued benefit using salary and service history through the
#' valuation date.
#'
#' @param salary_history Positive numeric vector of annual salaries.
#' @param p Nonnegative scalar accrual percentage.
#' @param fas_years Positive integer number of years in the salary average.
#'
#' @return A numeric scalar.
#'
#' @examples
#' AB_fas(
#'   salary_history = c(150000, 156000),
#'   p = 1,
#'   fas_years = 2
#' )
#'
#' @export
AB_fas <- function(salary_history, p, fas_years = 3) {
  salary_history <- .pension_check_positive(
    salary_history,
    "salary_history"
  )

  p <- .pension_check_scalar(
    p,
    "p",
    nonnegative = TRUE
  )

  fas_years <- .pension_check_scalar(
    fas_years,
    "fas_years",
    positive = TRUE,
    integer = TRUE
  )

  years <- length(salary_history)

  final_average_salary <- mean(
    tail(
      salary_history,
      min(fas_years, years)
    )
  )

  0.01 * p * years * final_average_salary
}

#' Accrued benefit under a career-average-earnings plan
#'
#' Computes the accrued benefit using salary history through the valuation
#' date.
#'
#' @param salary_history Positive numeric vector of annual salaries.
#' @param p Nonnegative scalar accrual percentage.
#'
#' @return A numeric scalar.
#'
#' @examples
#' AB_cae(
#'   salary_history = c(100000, 104000, 108160),
#'   p = 1
#' )
#'
#' @export
AB_cae <- function(salary_history, p) {
  salary_history <- .pension_check_positive(
    salary_history,
    "salary_history"
  )

  p <- .pension_check_scalar(
    p,
    "p",
    nonnegative = TRUE
  )

  0.01 * p * sum(salary_history)
}

# -------------------------------------------------------------------------
# Defined benefit valuation and cost methods
# -------------------------------------------------------------------------

#' Actuarial present value of a normal retirement benefit
#'
#' Computes the actuarial present value of a projected annual retirement
#' benefit. Scalar arguments are recycled to a common length.
#'
#' @param PABz Nonnegative projected annual benefit at retirement.
#' @param v_to_ret Nonnegative discount factor from the valuation date to
#'   retirement.
#' @param p_surv Survival or active-service probability to retirement in
#'   \code{[0, 1]}.
#' @param adue_ret Positive retirement annuity-due factor.
#'
#' @return A numeric vector.
#'
#' @examples
#' APV_NR_db(
#'   PABz = 108008.66,
#'   v_to_ret = 1 / 1.06^30,
#'   p_surv = 0.8,
#'   adue_ret = 12
#' )
#'
#' @export
APV_NR_db <- function(PABz, v_to_ret, p_surv, adue_ret) {
  PABz <- .pension_check_nonnegative(PABz, "PABz")
  v_to_ret <- .pension_check_nonnegative(
    v_to_ret,
    "v_to_ret"
  )
  p_surv <- .pension_check_probability(
    p_surv,
    "p_surv"
  )
  adue_ret <- .pension_check_positive(
    adue_ret,
    "adue_ret"
  )

  values <- .pension_recycle_common(
    PABz,
    v_to_ret,
    p_surv,
    adue_ret,
    .names = c(
      "PABz",
      "v_to_ret",
      "p_surv",
      "adue_ret"
    )
  )

  values[[1L]] *
    values[[2L]] *
    values[[3L]] *
    values[[4L]]
}

#' Entry Age Normal normal cost
#'
#' Computes Entry Age Normal normal cost as total benefit APV divided by an
#' active-service annuity-due factor.
#'
#' @param APV_total Nonnegative total actuarial present value of benefits.
#' @param adue_active Positive active-service annuity-due factor.
#'
#' @return A numeric vector.
#'
#' @examples
#' NC_EAN_db(APV_total = 25000, adue_active = 15)
#'
#' @export
NC_EAN_db <- function(APV_total, adue_active) {
  APV_total <- .pension_check_nonnegative(
    APV_total,
    "APV_total"
  )

  adue_active <- .pension_check_positive(
    adue_active,
    "adue_active"
  )

  values <- .pension_recycle_common(
    APV_total,
    adue_active,
    .names = c("APV_total", "adue_active")
  )

  values[[1L]] / values[[2L]]
}

#' Traditional Unit Credit normal cost
#'
#' Computes the actuarial present value of the benefit accrued during the
#' current year.
#'
#' @param accrual_benefit Nonnegative benefit accrued during the current year.
#' @inheritParams APV_NR_db
#'
#' @return A numeric vector.
#'
#' @examples
#' NC_TUC_db(
#'   accrual_benefit = 1560,
#'   v_to_ret = 0.5,
#'   p_surv = 0.9,
#'   adue_ret = 12
#' )
#'
#' @export
NC_TUC_db <- function(
    accrual_benefit,
    v_to_ret,
    p_surv,
    adue_ret) {
  APV_NR_db(
    PABz = accrual_benefit,
    v_to_ret = v_to_ret,
    p_surv = p_surv,
    adue_ret = adue_ret
  )
}

#' Traditional Unit Credit accrued liability
#'
#' Computes the actuarial present value of the benefit accrued through the
#' valuation date.
#'
#' @param accrued_benefit Nonnegative accrued benefit.
#' @inheritParams APV_NR_db
#'
#' @return A numeric vector.
#'
#' @examples
#' AAL_TUC_db(
#'   accrued_benefit = 12000,
#'   v_to_ret = 0.5,
#'   p_surv = 0.9,
#'   adue_ret = 12
#' )
#'
#' @export
AAL_TUC_db <- function(
    accrued_benefit,
    v_to_ret,
    p_surv,
    adue_ret) {
  APV_NR_db(
    PABz = accrued_benefit,
    v_to_ret = v_to_ret,
    p_surv = p_surv,
    adue_ret = adue_ret
  )
}

#' Projected Unit Credit normal cost
#'
#' Computes the actuarial present value of the portion of the projected
#' benefit attributed to the current year of service.
#'
#' @param projected_benefit Nonnegative projected annual benefit at retirement.
#' @param total_service Positive total service at retirement.
#' @inheritParams APV_NR_db
#'
#' @return A numeric vector.
#'
#' @examples
#' NC_PUC_db(
#'   projected_benefit = 30000,
#'   total_service = 30,
#'   v_to_ret = 0.5,
#'   p_surv = 0.9,
#'   adue_ret = 12
#' )
#'
#' @export
NC_PUC_db <- function(
    projected_benefit,
    total_service,
    v_to_ret,
    p_surv,
    adue_ret) {
  projected_benefit <- .pension_check_nonnegative(
    projected_benefit,
    "projected_benefit"
  )

  total_service <- .pension_check_positive(
    total_service,
    "total_service"
  )

  values <- .pension_recycle_common(
    projected_benefit,
    total_service,
    v_to_ret,
    p_surv,
    adue_ret,
    .names = c(
      "projected_benefit",
      "total_service",
      "v_to_ret",
      "p_surv",
      "adue_ret"
    )
  )

  APV_NR_db(
    PABz = values[[1L]] / values[[2L]],
    v_to_ret = values[[3L]],
    p_surv = values[[4L]],
    adue_ret = values[[5L]]
  )
}

#' Projected Unit Credit accrued liability
#'
#' Computes the actuarial present value of the portion of the projected
#' benefit attributed to past service.
#'
#' @param projected_benefit Nonnegative projected annual benefit at retirement.
#' @param past_service Nonnegative service completed through the valuation date.
#' @param total_service Positive total service at retirement.
#' @inheritParams APV_NR_db
#'
#' @return A numeric vector.
#'
#' @examples
#' AAL_PUC_db(
#'   projected_benefit = 30000,
#'   past_service = 10,
#'   total_service = 30,
#'   v_to_ret = 0.5,
#'   p_surv = 0.9,
#'   adue_ret = 12
#' )
#'
#' @export
AAL_PUC_db <- function(
    projected_benefit,
    past_service,
    total_service,
    v_to_ret,
    p_surv,
    adue_ret) {
  projected_benefit <- .pension_check_nonnegative(
    projected_benefit,
    "projected_benefit"
  )

  past_service <- .pension_check_nonnegative(
    past_service,
    "past_service"
  )

  total_service <- .pension_check_positive(
    total_service,
    "total_service"
  )

  values <- .pension_recycle_common(
    projected_benefit,
    past_service,
    total_service,
    v_to_ret,
    p_surv,
    adue_ret,
    .names = c(
      "projected_benefit",
      "past_service",
      "total_service",
      "v_to_ret",
      "p_surv",
      "adue_ret"
    )
  )

  projected_benefit <- values[[1L]]
  past_service <- values[[2L]]
  total_service <- values[[3L]]
  v_to_ret <- values[[4L]]
  p_surv <- values[[5L]]
  adue_ret <- values[[6L]]

  if (any(past_service > total_service)) {
    stop(
      "past_service cannot exceed total_service.",
      call. = FALSE
    )
  }

  APV_NR_db(
    PABz = projected_benefit *
      past_service /
      total_service,
    v_to_ret = v_to_ret,
    p_surv = p_surv,
    adue_ret = adue_ret
  )
}

testthat::test_that("salary scale follows compound growth", {
  out <- salary_scale(
    k = 30:34,
    g = 0.04,
    base_age = 30
  )

  testthat::expect_equal(
    out,
    1.04^(0:4),
    tolerance = 1e-12
  )
})


testthat::test_that("salary scale allows valid negative growth", {
  out <- salary_scale(
    k = 30:32,
    g = -0.02,
    base_age = 30
  )

  testthat::expect_equal(
    out,
    0.98^(0:2),
    tolerance = 1e-12
  )

  testthat::expect_error(
    salary_scale(30:32, g = -1),
    "greater than -1"
  )
})


testthat::test_that("defined contribution accumulation uses beginning-year contributions", {
  out <- AVz_dc(
    x = 30,
    z = 33,
    Sx = 100,
    c = 0.10,
    i = 0.05,
    g = 0
  )

  expected <- 10 * (
    1.05^3 +
      1.05^2 +
      1.05
  )

  testthat::expect_equal(
    out,
    expected,
    tolerance = 1e-12
  )
})


testthat::test_that("salary scale and growth inputs give equivalent DC values", {
  scale <- salary_scale(
    k = 30:34,
    g = 0.04,
    base_age = 30
  )

  from_growth <- AVz_dc(
    x = 30,
    z = 35,
    Sx = 50000,
    c = 0.10,
    i = 0.05,
    g = 0.04
  )

  from_scale <- AVz_dc(
    x = 30,
    z = 35,
    Sx = 50000,
    c = 0.10,
    i = 0.05,
    s = scale
  )

  testthat::expect_equal(
    from_growth,
    from_scale,
    tolerance = 1e-10
  )
})


testthat::test_that("defined contribution replacement ratio is internally consistent", {
  AV <- AVz_dc(
    x = 30,
    z = 35,
    Sx = 50000,
    c = 0.10,
    i = 0.05,
    g = 0.04
  )

  income <- Income_dc(AV, adue_z = 12)
  final_salary <- 50000 * 1.04^4

  expected <- income / final_salary

  testthat::expect_equal(
    replacement_ratio_dc(
      x = 30,
      z = 35,
      Sx = 50000,
      c = 0.10,
      i = 0.05,
      adue_z = 12,
      g = 0.04
    ),
    expected,
    tolerance = 1e-12
  )
})


testthat::test_that("target contribution rate reproduces the target ratio", {
  contribution <- contribution_rate_target(
    x = 30,
    z = 65,
    Sx = 60000,
    RR_target = 0.50,
    i = 0.06,
    adue_z = 11,
    g = 0.04
  )

  achieved <- replacement_ratio_dc(
    x = 30,
    z = 65,
    Sx = 60000,
    c = contribution,
    i = 0.06,
    adue_z = 11,
    g = 0.04
  )

  testthat::expect_equal(
    achieved,
    0.50,
    tolerance = 1e-10
  )
})


testthat::test_that("income and DB replacement ratios vectorize", {
  testthat::expect_equal(
    Income_dc(
      AVz = c(120000, 240000),
      adue_z = 12
    ),
    c(10000, 20000)
  )

  testthat::expect_equal(
    replacement_ratio_db(
      benefit = c(30000, 40000),
      salary = 100000
    ),
    c(0.30, 0.40)
  )
})


testthat::test_that("final average salary benefit uses the final salary years", {
  out <- PAB_fas(
    x = 30,
    z = 35,
    CASx = 100000,
    p = 2,
    fas_years = 3,
    g = 0.04
  )

  salaries <- 100000 * 1.04^(0:4)
  expected_fas <- mean(tail(salaries, 3))
  expected <- 0.02 * 5 * expected_fas

  testthat::expect_equal(
    out,
    expected,
    tolerance = 1e-10
  )
})


testthat::test_that("final average salary benefit includes past service", {
  without_past <- PAB_fas(
    x = 40,
    z = 45,
    CASx = 100000,
    p = 2,
    fas_years = 3,
    past_service = 0,
    g = 0.03
  )

  with_past <- PAB_fas(
    x = 40,
    z = 45,
    CASx = 100000,
    p = 2,
    fas_years = 3,
    past_service = 10,
    g = 0.03
  )

  testthat::expect_true(with_past > without_past)
})


testthat::test_that("career average projected benefit uses salary total", {
  salaries <- 100000 * 1.04^(0:4)

  out <- PAB_cae(
    x = 30,
    z = 35,
    CASx = 100000,
    p = 1,
    past_salary_total = 200000,
    g = 0.04
  )

  expected <- 0.01 * (
    200000 + sum(salaries)
  )

  testthat::expect_equal(
    out,
    expected,
    tolerance = 1e-10
  )
})


testthat::test_that("accrued benefit formulas use actual salary history", {
  salary_history <- c(
    100000,
    104000,
    108160
  )

  testthat::expect_equal(
    AB_cae(
      salary_history = salary_history,
      p = 1
    ),
    0.01 * sum(salary_history)
  )

  testthat::expect_equal(
    AB_fas(
      salary_history = salary_history,
      p = 2,
      fas_years = 2
    ),
    0.02 * 3 * mean(tail(salary_history, 2))
  )
})


testthat::test_that("normal retirement APV vectorizes consistently", {
  out <- APV_NR_db(
    PABz = c(10000, 20000),
    v_to_ret = 0.50,
    p_surv = c(0.80, 0.90),
    adue_ret = 12
  )

  expected <- c(10000, 20000) *
    0.50 *
    c(0.80, 0.90) *
    12

  testthat::expect_equal(
    out,
    expected
  )
})


testthat::test_that("TUC formulas agree with the normal retirement APV", {
  normal_cost <- NC_TUC_db(
    accrual_benefit = 1500,
    v_to_ret = 0.50,
    p_surv = 0.90,
    adue_ret = 12
  )

  liability <- AAL_TUC_db(
    accrued_benefit = 12000,
    v_to_ret = 0.50,
    p_surv = 0.90,
    adue_ret = 12
  )

  testthat::expect_equal(
    normal_cost,
    APV_NR_db(
      PABz = 1500,
      v_to_ret = 0.50,
      p_surv = 0.90,
      adue_ret = 12
    )
  )

  testthat::expect_equal(
    liability,
    APV_NR_db(
      PABz = 12000,
      v_to_ret = 0.50,
      p_surv = 0.90,
      adue_ret = 12
    )
  )
})


testthat::test_that("PUC cost and liability allocate projected benefit by service", {
  normal_cost <- NC_PUC_db(
    projected_benefit = 30000,
    total_service = 30,
    v_to_ret = 0.50,
    p_surv = 0.90,
    adue_ret = 12
  )

  liability <- AAL_PUC_db(
    projected_benefit = 30000,
    past_service = 10,
    total_service = 30,
    v_to_ret = 0.50,
    p_surv = 0.90,
    adue_ret = 12
  )

  testthat::expect_equal(
    normal_cost,
    APV_NR_db(
      PABz = 1000,
      v_to_ret = 0.50,
      p_surv = 0.90,
      adue_ret = 12
    )
  )

  testthat::expect_equal(
    liability,
    APV_NR_db(
      PABz = 10000,
      v_to_ret = 0.50,
      p_surv = 0.90,
      adue_ret = 12
    )
  )
})


testthat::test_that("PUC functions vectorize over members", {
  out <- AAL_PUC_db(
    projected_benefit = c(30000, 40000),
    past_service = c(10, 15),
    total_service = c(30, 40),
    v_to_ret = 0.50,
    p_surv = 0.90,
    adue_ret = 12
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("salary projections require integer projection years", {
  testthat::expect_error(
    AVz_dc(
      x = 30,
      z = 65.5,
      Sx = 50000,
      c = 0.10,
      i = 0.05,
      g = 0.04
    ),
    "integer number of years"
  )
})


testthat::test_that("salary projection requires exactly one salary assumption", {
  testthat::expect_error(
    AVz_dc(
      x = 30,
      z = 35,
      Sx = 50000,
      c = 0.10,
      i = 0.05
    ),
    "Supply one of g or s"
  )

  testthat::expect_error(
    AVz_dc(
      x = 30,
      z = 35,
      Sx = 50000,
      c = 0.10,
      i = 0.05,
      g = 0.04,
      s = rep(1, 5)
    ),
    "Supply only one"
  )
})


testthat::test_that("salary scale vector must match projection period", {
  testthat::expect_error(
    AVz_dc(
      x = 30,
      z = 35,
      Sx = 50000,
      c = 0.10,
      i = 0.05,
      s = rep(1, 4)
    ),
    "length z - x"
  )
})


testthat::test_that("pension functions reject invalid rates and percentages", {
  testthat::expect_error(
    AVz_dc(
      x = 30,
      z = 35,
      Sx = 50000,
      c = 1.20,
      i = 0.05,
      g = 0.04
    ),
    "between 0 and 1"
  )

  testthat::expect_error(
    AVz_dc(
      x = 30,
      z = 35,
      Sx = 50000,
      c = 0.10,
      i = -1,
      g = 0.04
    ),
    "greater than -1"
  )

  testthat::expect_error(
    PAB_fas(
      x = 30,
      z = 35,
      CASx = 50000,
      p = -1,
      g = 0.04
    ),
    "nonnegative"
  )
})


testthat::test_that("final average period must be a positive integer", {
  testthat::expect_error(
    PAB_fas(
      x = 30,
      z = 35,
      CASx = 50000,
      p = 2,
      fas_years = 2.5,
      g = 0.04
    ),
    "must be an integer"
  )

  testthat::expect_error(
    PAB_fas(
      x = 30,
      z = 35,
      CASx = 50000,
      p = 2,
      fas_years = 6,
      g = 0.04
    ),
    "cannot exceed"
  )
})


testthat::test_that("past service cannot exceed total service", {
  testthat::expect_error(
    AAL_PUC_db(
      projected_benefit = 30000,
      past_service = 31,
      total_service = 30,
      v_to_ret = 0.50,
      p_surv = 0.90,
      adue_ret = 12
    ),
    "cannot exceed"
  )
})


testthat::test_that("elementwise pension functions reject incompatible lengths", {
  testthat::expect_error(
    Income_dc(
      AVz = c(100000, 200000),
      adue_z = c(10, 11, 12)
    ),
    "length 1 or the common length"
  )

  testthat::expect_error(
    APV_NR_db(
      PABz = c(10000, 20000),
      v_to_ret = c(0.5, 0.6, 0.7),
      p_surv = 0.9,
      adue_ret = 12
    ),
    "length 1 or the common length"
  )
})

# =========================================================
# Universal life insurance functions
# =========================================================

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.ul_check_numeric <- function(x, name, allow_infinite = FALSE) {
  if (!is.numeric(x) || length(x) == 0L) {
    stop(name, " must be a numeric vector with positive length.", call. = FALSE)
  }

  invalid <- if (allow_infinite) {
    is.na(x) | is.nan(x)
  } else {
    !is.finite(x)
  }

  if (any(invalid)) {
    descriptor <- if (allow_infinite) {
      "must not contain missing or NaN values."
    } else {
      "must contain finite values."
    }
    stop(name, " ", descriptor, call. = FALSE)
  }

  as.numeric(x)
}

#' @noRd
.ul_check_nonnegative <- function(x, name, allow_infinite = FALSE) {
  x <- .ul_check_numeric(x, name, allow_infinite = allow_infinite)

  if (any(x < 0)) {
    stop(name, " must contain nonnegative values.", call. = FALSE)
  }

  x
}

#' @noRd
.ul_check_probability <- function(x, name) {
  x <- .ul_check_numeric(x, name)

  if (any(x < 0 | x > 1)) {
    stop(name, " must contain values in [0, 1].", call. = FALSE)
  }

  x
}

#' @noRd
.ul_check_rate <- function(x, name) {
  x <- .ul_check_numeric(x, name)

  if (any(x <= -1)) {
    stop(name, " must contain values greater than -1.", call. = FALSE)
  }

  x
}

#' @noRd
.ul_check_scalar <- function(x, name, nonnegative = FALSE, positive = FALSE) {
  if (nonnegative || positive) {
    x <- .ul_check_nonnegative(x, name)
  } else {
    x <- .ul_check_numeric(x, name)
  }

  if (length(x) != 1L) {
    stop(name, " must be a numeric scalar.", call. = FALSE)
  }

  if (positive && x <= 0) {
    stop(name, " must be positive.", call. = FALSE)
  }

  x
}

#' @noRd
.ul_check_logical_scalar <- function(x, name) {
  if (!is.logical(x) || length(x) != 1L || is.na(x)) {
    stop(name, " must be TRUE or FALSE.", call. = FALSE)
  }

  x
}

#' @noRd
.ul_recycle <- function(..., .names = NULL) {
  values <- list(...)
  lengths <- vapply(values, length, integer(1))

  if (any(lengths == 0L)) {
    stop("Inputs must have positive length.", call. = FALSE)
  }

  common_length <- max(lengths)

  if (is.null(.names)) {
    .names <- paste0("Argument ", seq_along(values))
  }

  if (length(.names) != length(values)) {
    stop(
      ".names must have the same length as the supplied arguments.",
      call. = FALSE
    )
  }

  for (j in seq_along(values)) {
    if (!lengths[j] %in% c(1L, common_length)) {
      stop(
        .names[j],
        " must have length 1 or the common length ",
        common_length,
        ".",
        call. = FALSE
      )
    }

    values[[j]] <- rep(values[[j]], length.out = common_length)
  }

  values
}

# -------------------------------------------------------------------------
# Cost of insurance
# -------------------------------------------------------------------------

#' Cost of insurance for Type B universal life
#'
#' Computes the one-period cost of insurance for a Type B universal life
#' contract:
#' \deqn{
#' \mathrm{COI}_t = \frac{B_t q_{x+t-1}}{1+i_t^q}.
#' }
#'
#' The arguments may be scalars or vectors. Scalar arguments are recycled to
#' the common length.
#'
#' @param B Face amount.
#' @param qx Mortality probability for the period.
#' @param iq Interest rate used in the cost-of-insurance calculation. Values
#'   must be greater than \code{-1}.
#'
#' @return A numeric vector of cost-of-insurance charges.
#'
#' @examples
#' coi_ul_typeB(B = 100000, qx = 0.00076, iq = 0.03)
#' coi_ul_typeB(
#'   B = 100000,
#'   qx = c(0.00076, 0.00081),
#'   iq = c(0.03, 0.035)
#' )
#'
#' @export
coi_ul_typeB <- function(B, qx, iq) {
  B <- .ul_check_nonnegative(B, "B")
  qx <- .ul_check_probability(qx, "qx")
  iq <- .ul_check_rate(iq, "iq")

  values <- .ul_recycle(B, qx, iq, .names = c("B", "qx", "iq"))

  B <- values[[1]]
  qx <- values[[2]]
  iq <- values[[3]]

  B * qx / (1 + iq)
}

# -------------------------------------------------------------------------
# Universal life account-value paths
# -------------------------------------------------------------------------

#' Type B universal life account-value path
#'
#' Computes a year-by-year Type B universal life account-value path. Premiums
#' and expenses are applied at the beginning of each period, followed by the
#' cost-of-insurance charge and credited interest.
#'
#' @param G Premium amount by period.
#' @param r Percent-of-premium expense rate by period. Values must lie in
#'   \code{[0, 1]}.
#' @param e Fixed expense by period.
#' @param qx Mortality probability by period.
#' @param ic Credited annual effective interest rate by period. Values must be
#'   greater than \code{-1}.
#' @param B Face amount by period.
#' @param iq Interest rate used in the cost-of-insurance calculation. Defaults
#'   to \code{ic}; values must be greater than \code{-1}.
#' @param AV0 Nonnegative scalar initial account value.
#'
#' @return A data frame containing the policy duration, premium,
#'   net contribution, cost-of-insurance charge, and account value.
#'
#' @examples
#' qx <- c(0.00076, 0.00081, 0.00085, 0.00090, 0.00095)
#' r <- c(0.75, rep(0.10, 4))
#' e <- c(100, rep(20, 4))
#'
#' AV_path_ul_typeB(
#'   G = 5000,
#'   r = r,
#'   e = e,
#'   qx = qx,
#'   ic = 0.03,
#'   B = 100000
#' )
#'
#' @export
AV_path_ul_typeB <- function(G, r, e, qx, ic, B, iq = ic, AV0 = 0) {
  G <- .ul_check_nonnegative(G, "G")
  r <- .ul_check_probability(r, "r")
  e <- .ul_check_nonnegative(e, "e")
  qx <- .ul_check_probability(qx, "qx")
  ic <- .ul_check_rate(ic, "ic")
  B <- .ul_check_nonnegative(B, "B")
  iq <- .ul_check_rate(iq, "iq")
  AV0 <- .ul_check_scalar(AV0, "AV0", nonnegative = TRUE)

  values <- .ul_recycle(
    G, r, e, qx, ic, B, iq,
    .names = c("G", "r", "e", "qx", "ic", "B", "iq")
  )

  G <- values[[1]]
  r <- values[[2]]
  e <- values[[3]]
  qx <- values[[4]]
  ic <- values[[5]]
  B <- values[[6]]
  iq <- values[[7]]

  n <- length(qx)
  account_value <- numeric(n + 1L)
  net_contribution <- numeric(n)
  cost_of_insurance <- numeric(n)
  account_value[1L] <- AV0

  for (k in seq_len(n)) {
    net_contribution[k] <- G[k] * (1 - r[k]) - e[k]
    cost_of_insurance[k] <- coi_ul_typeB(B[k], qx[k], iq[k])
    account_value[k + 1L] <-
      (account_value[k] + net_contribution[k] - cost_of_insurance[k]) *
      (1 + ic[k])
  }

  data.frame(
    t = 0:n,
    premium = c(NA_real_, G),
    net_contribution = c(NA_real_, net_contribution),
    COI = c(NA_real_, cost_of_insurance),
    AV = account_value
  )
}

#' Type A universal life account-value path
#'
#' Computes a year-by-year Type A universal life account-value path. The death
#' benefit is fixed, so the net amount at risk depends on the ending account
#' value and the roll-forward is solved explicitly each period.
#'
#' @inheritParams AV_path_ul_typeB
#'
#' @return A data frame containing the policy duration, premium, and account
#'   value.
#'
#' @examples
#' qx <- c(0.00076, 0.00081, 0.00085, 0.00090, 0.00095)
#' r <- c(0.75, rep(0.10, 4))
#' e <- c(100, rep(20, 4))
#'
#' AV_path_ul_typeA(
#'   G = 5000,
#'   r = r,
#'   e = e,
#'   qx = qx,
#'   ic = 0.03,
#'   B = 100000
#' )
#'
#' @export
AV_path_ul_typeA <- function(G, r, e, qx, ic, B, iq = ic, AV0 = 0) {
  G <- .ul_check_nonnegative(G, "G")
  r <- .ul_check_probability(r, "r")
  e <- .ul_check_nonnegative(e, "e")
  qx <- .ul_check_probability(qx, "qx")
  ic <- .ul_check_rate(ic, "ic")
  B <- .ul_check_nonnegative(B, "B")
  iq <- .ul_check_rate(iq, "iq")
  AV0 <- .ul_check_scalar(AV0, "AV0", nonnegative = TRUE)

  values <- .ul_recycle(
    G, r, e, qx, ic, B, iq,
    .names = c("G", "r", "e", "qx", "ic", "B", "iq")
  )

  G <- values[[1]]
  r <- values[[2]]
  e <- values[[3]]
  qx <- values[[4]]
  ic <- values[[5]]
  B <- values[[6]]
  iq <- values[[7]]

  n <- length(qx)
  account_value <- numeric(n + 1L)
  account_value[1L] <- AV0

  for (k in seq_len(n)) {
    mortality_factor <- qx[k] * (1 + ic[k]) / (1 + iq[k])
    denominator <- 1 - mortality_factor

    if (denominator <= 0) {
      stop("The Type A roll-forward denominator must be positive.", call. = FALSE)
    }

    numerator <-
      (account_value[k] + G[k] * (1 - r[k]) - e[k]) * (1 + ic[k]) -
      B[k] * mortality_factor

    account_value[k + 1L] <- numerator / denominator
  }

  data.frame(
    t = 0:n,
    premium = c(NA_real_, G),
    AV = account_value
  )
}

# -------------------------------------------------------------------------
# Indexed universal life credited-rate helpers
# -------------------------------------------------------------------------

#' Point-to-point index growth rates
#'
#' Computes consecutive point-to-point growth rates from index values.
#'
#' @param index Numeric vector of strictly positive index values.
#'
#' @return A numeric vector with length one less than \code{index}.
#'
#' @examples
#' iP_eiul(c(1000, 1050, 1200, 1100))
#'
#' @export
iP_eiul <- function(index) {
  index <- .ul_check_numeric(index, "index")

  if (length(index) < 2L) {
    stop("index must contain at least two values.", call. = FALSE)
  }
  if (any(index <= 0)) {
    stop("index values must be positive.", call. = FALSE)
  }

  index[-1L] / index[-length(index)] - 1
}

#' Monthly-average index growth rate
#'
#' Computes the monthly-average growth rate from an initial index value and
#' twelve monthly closing values.
#'
#' @param index Numeric vector of length 13 containing a strictly positive
#'   initial index value followed by twelve nonnegative monthly closing values.
#'
#' @return A numeric scalar.
#'
#' @examples
#' index <- c(
#'   1000, 1020, 1100, 1150, 1080, 1040, 960,
#'   1030, 1000, 1070, 1150, 1200, 1150
#' )
#' iMA_eiul(index)
#'
#' @export
iMA_eiul <- function(index) {
  index <- .ul_check_nonnegative(index, "index")

  if (length(index) != 13L) {
    stop(
      "index must contain an initial value and 12 monthly closing values.",
      call. = FALSE
    )
  }
  if (index[1L] <= 0) {
    stop("The initial index value must be positive.", call. = FALSE)
  }

  mean(index[-1L]) / index[1L] - 1
}

#' Credited rates from index growth rates
#'
#' Applies a participation rate, floor, cap, and optional margin to raw index
#' growth rates.
#'
#' @param i_raw Numeric vector of raw index growth rates.
#' @param part Nonnegative scalar participation rate.
#' @param floor Scalar minimum credited rate.
#' @param cap Scalar maximum credited rate. The default is \code{Inf}.
#' @param margin Nonnegative scalar index margin.
#' @param margin_after_participation Logical scalar. If \code{TRUE}, the margin
#'   is subtracted after applying participation; otherwise it is subtracted
#'   before applying participation.
#'
#' @return A numeric vector of credited rates.
#'
#' @examples
#' raw <- iP_eiul(c(1000, 1050, 1200, 1100))
#' i_credit_eiul(raw, part = 1.10, floor = 0.01, cap = 0.10)
#' i_credit_eiul(raw)
#'
#' @export
i_credit_eiul <- function(i_raw, part = 1, floor = 0, cap = Inf,
                          margin = 0, margin_after_participation = TRUE) {
  i_raw <- .ul_check_numeric(i_raw, "i_raw")
  part <- .ul_check_scalar(part, "part", nonnegative = TRUE)
  floor <- .ul_check_scalar(floor, "floor")
  cap <- .ul_check_numeric(cap, "cap", allow_infinite = TRUE)
  margin <- .ul_check_scalar(margin, "margin", nonnegative = TRUE)
  margin_after_participation <- .ul_check_logical_scalar(
    margin_after_participation,
    "margin_after_participation"
  )

  if (length(cap) != 1L) {
    stop("cap must be a numeric scalar.", call. = FALSE)
  }
  if (floor > cap) {
    stop("floor cannot exceed cap.", call. = FALSE)
  }

  adjusted <- if (margin_after_participation) {
    part * i_raw - margin
  } else {
    part * (i_raw - margin)
  }

  pmin(cap, pmax(floor, adjusted))
}

# -------------------------------------------------------------------------
# Persistency under mortality and withdrawal
# -------------------------------------------------------------------------

#' Universal life persistency probabilities
#'
#' \code{pxtau_ul()} computes one-year persistency probabilities under
#' mortality and withdrawal.
#'
#' \code{tpxtau_ul()} computes cumulative persistency through the end of each
#' policy year.
#'
#' @param qd Mortality probabilities.
#' @param qw Withdrawal probabilities.
#' @param year_end_withdrawal Logical scalar. If \code{TRUE}, withdrawal is
#'   modeled at year-end and persistency is
#'   \eqn{(1-q^{(d)})(1-q^{(w)})}. Otherwise persistency is
#'   \eqn{1-q^{(d)}-q^{(w)}}.
#'
#' @return A numeric vector.
#'
#' @examples
#' qd <- c(0.001, 0.002, 0.003)
#' qw <- c(0.02, 0.02, 0.03)
#'
#' pxtau_ul(qd, qw)
#' tpxtau_ul(qd, qw)
#'
#' @rdname ul_persistency
#' @export
pxtau_ul <- function(qd, qw, year_end_withdrawal = TRUE) {
  qd <- .ul_check_probability(qd, "qd")
  qw <- .ul_check_probability(qw, "qw")
  year_end_withdrawal <- .ul_check_logical_scalar(
    year_end_withdrawal,
    "year_end_withdrawal"
  )

  values <- .ul_recycle(qd, qw, .names = c("qd", "qw"))
  qd <- values[[1]]
  qw <- values[[2]]

  if (year_end_withdrawal) {
    return((1 - qd) * (1 - qw))
  }

  persistency <- 1 - qd - qw

  if (any(persistency < 0)) {
    stop("qd + qw must not exceed 1.", call. = FALSE)
  }

  persistency
}

#' @rdname ul_persistency
#' @export
tpxtau_ul <- function(qd, qw, year_end_withdrawal = TRUE) {
  cumprod(
    pxtau_ul(
      qd = qd,
      qw = qw,
      year_end_withdrawal = year_end_withdrawal
    )
  )
}

# -------------------------------------------------------------------------
# Universal life reserving helpers
# -------------------------------------------------------------------------

#' Guaranteed maturity fund roll-forward
#'
#' Computes a one-period guaranteed maturity fund roll-forward. Arguments may
#' be scalars or vectors and follow common-length recycling.
#'
#' @param GMF_prev Prior guaranteed maturity fund.
#' @param GMP Guaranteed maturity premium.
#' @param r Percent-of-premium expense rate in \code{[0, 1]}.
#' @param policy_charge Guaranteed policy charge.
#' @param i Guaranteed annual effective interest rate greater than \code{-1}.
#'
#' @return A numeric vector.
#'
#' @examples
#' GMF_rollforward_ul(140.40, 14.49, 0.04, 11.80, 0.03)
#'
#' @export
GMF_rollforward_ul <- function(GMF_prev, GMP, r, policy_charge, i) {
  GMF_prev <- .ul_check_nonnegative(GMF_prev, "GMF_prev")
  GMP <- .ul_check_nonnegative(GMP, "GMP")
  r <- .ul_check_probability(r, "r")
  policy_charge <- .ul_check_nonnegative(policy_charge, "policy_charge")
  i <- .ul_check_rate(i, "i")

  values <- .ul_recycle(
    GMF_prev, GMP, r, policy_charge, i,
    .names = c("GMF_prev", "GMP", "r", "policy_charge", "i")
  )

  GMF_prev <- values[[1]]
  GMP <- values[[2]]
  r <- values[[3]]
  policy_charge <- values[[4]]
  i <- values[[5]]

  (GMF_prev + GMP * (1 - r) - policy_charge) * (1 + i)
}

#' Account-value to guaranteed-fund ratio
#'
#' Computes the ratio of account value to guaranteed maturity fund, capped at
#' one.
#'
#' @param AV Nonnegative account value.
#' @param GMF Positive guaranteed maturity fund.
#'
#' @return A numeric vector.
#'
#' @examples
#' rt_ul(AV = 4918.20, GMF = 14678.57)
#'
#' @export
rt_ul <- function(AV, GMF) {
  AV <- .ul_check_nonnegative(AV, "AV")
  GMF <- .ul_check_nonnegative(GMF, "GMF")

  values <- .ul_recycle(AV, GMF, .names = c("AV", "GMF"))
  AV <- values[[1]]
  GMF <- values[[2]]

  if (any(GMF <= 0)) {
    stop("GMF must contain positive values.", call. = FALSE)
  }

  pmin(AV / GMF, 1)
}

#' Pre-floor CRVM reserve
#'
#' Computes the pre-floor reserve by multiplying the funding ratio by the
#' difference between the present value of future benefits and future
#' premiums.
#'
#' @param r Funding ratio. Values must lie in \code{[0, 1]}.
#' @param pvfb_minus_pvfp Numeric difference between the present value of
#'   future benefits and future premiums.
#'
#' @return A numeric vector.
#'
#' @examples
#' Vprefloor_crvm_ul(r = 0.33506, pvfb_minus_pvfp = 70)
#'
#' @export
Vprefloor_crvm_ul <- function(r, pvfb_minus_pvfp) {
  r <- .ul_check_probability(r, "r")
  pvfb_minus_pvfp <- .ul_check_numeric(pvfb_minus_pvfp, "pvfb_minus_pvfp")

  values <- .ul_recycle(
    r, pvfb_minus_pvfp,
    .names = c("r", "pvfb_minus_pvfp")
  )

  values[[1]] * values[[2]]
}

#' AG 38 prefunding ratio
#'
#' Computes the excess-payment-to-required-net-single-premium ratio, capped at
#' one.
#'
#' @param excess_payment Nonnegative excess payment or shadow-fund amount.
#' @param nsp_required Positive net single premium required to fully fund the
#'   guarantee.
#'
#' @return A numeric vector.
#'
#' @examples
#' ag38_prefunding_ratio(60000, 100000)
#'
#' @export
ag38_prefunding_ratio <- function(excess_payment, nsp_required) {
  excess_payment <- .ul_check_nonnegative(excess_payment, "excess_payment")
  nsp_required <- .ul_check_nonnegative(nsp_required, "nsp_required")

  values <- .ul_recycle(
    excess_payment, nsp_required,
    .names = c("excess_payment", "nsp_required")
  )

  excess_payment <- values[[1]]
  nsp_required <- values[[2]]

  if (any(nsp_required <= 0)) {
    stop("nsp_required must contain positive values.", call. = FALSE)
  }

  pmin(excess_payment / nsp_required, 1)
}

#' AG 38 reserve calculation
#'
#' Computes the prefunding ratio, net additional amount, reduced deficiency
#' reserve, intermediate reserve, and increased basic reserve.
#'
#' Scalar inputs preserve the original named-list output. Vectorized inputs
#' return a data frame with one row per calculation.
#'
#' @param basic_reserve Nonnegative basic reserve.
#' @param deficiency_reserve Nonnegative deficiency reserve.
#' @param excess_payment Nonnegative excess payment or shadow-fund amount.
#' @param nsp_required Positive net single premium required to fully fund the
#'   guarantee.
#' @param valuation_nsp Nonnegative valuation net single premium.
#' @param surrender_charge Nonnegative surrender charge.
#'
#' @return For scalar inputs, a named list. For vectorized inputs, a data frame
#'   containing the same calculated quantities.
#'
#' @examples
#' ag38_reserve_ul(
#'   basic_reserve = 10000,
#'   deficiency_reserve = 0,
#'   excess_payment = 60000,
#'   nsp_required = 100000,
#'   valuation_nsp = 150000,
#'   surrender_charge = 5000
#' )
#'
#' @export
ag38_reserve_ul <- function(basic_reserve,
                            deficiency_reserve = 0,
                            excess_payment,
                            nsp_required,
                            valuation_nsp,
                            surrender_charge = 0) {
  basic_reserve <- .ul_check_nonnegative(basic_reserve, "basic_reserve")
  deficiency_reserve <- .ul_check_nonnegative(
    deficiency_reserve,
    "deficiency_reserve"
  )
  excess_payment <- .ul_check_nonnegative(excess_payment, "excess_payment")
  nsp_required <- .ul_check_nonnegative(nsp_required, "nsp_required")
  valuation_nsp <- .ul_check_nonnegative(valuation_nsp, "valuation_nsp")
  surrender_charge <- .ul_check_nonnegative(
    surrender_charge,
    "surrender_charge"
  )

  values <- .ul_recycle(
    basic_reserve,
    deficiency_reserve,
    excess_payment,
    nsp_required,
    valuation_nsp,
    surrender_charge,
    .names = c(
      "basic_reserve",
      "deficiency_reserve",
      "excess_payment",
      "nsp_required",
      "valuation_nsp",
      "surrender_charge"
    )
  )

  basic_reserve <- values[[1]]
  deficiency_reserve <- values[[2]]
  excess_payment <- values[[3]]
  nsp_required <- values[[4]]
  valuation_nsp <- values[[5]]
  surrender_charge <- values[[6]]

  if (any(nsp_required <= 0)) {
    stop("nsp_required must contain positive values.", call. = FALSE)
  }

  prefunding_ratio <- ag38_prefunding_ratio(excess_payment, nsp_required)
  base_plus_deficiency <- basic_reserve + deficiency_reserve

  net_amount_additional <-
    prefunding_ratio * (valuation_nsp - base_plus_deficiency)

  reduced_deficiency_reserve <-
    (1 - prefunding_ratio) * deficiency_reserve

  step8_reserve <-
    pmin(
      valuation_nsp,
      base_plus_deficiency + net_amount_additional
    ) - surrender_charge

  increased_basic_reserve <-
    step8_reserve - reduced_deficiency_reserve

  out <- data.frame(
    prefunding_ratio = prefunding_ratio,
    net_amount_additional = net_amount_additional,
    reduced_deficiency_reserve = reduced_deficiency_reserve,
    step8_reserve = step8_reserve,
    increased_basic_reserve = increased_basic_reserve,
    final_reserve = increased_basic_reserve
  )

  if (nrow(out) == 1L) {
    return(as.list(out[1L, , drop = FALSE]))
  }

  out
}

test_that("AV_path_ul_typeB reproduces Example 16.1", {
  qx <- c(.00076, .00081, .00085, .00090, .00095)
  r <- c(.75, .10, .10, .10, .10)
  e <- c(100, 20, 20, 20, 20)
  G <- rep(5000, 5)

  out <- AV_path_ul_typeB(
    G = G,
    r = r,
    e = e,
    qx = qx,
    ic = 0.03,
    iq = 0.03,
    B = 100000
  )

  expect_equal(out$AV[2], 1108.50, tolerance = 1e-2)
  expect_equal(out$AV[3], 5675.16, tolerance = 1e-2)
  expect_equal(out$AV[4], 10374.82, tolerance = 1e-2)
  expect_equal(out$AV[5], 15210.46, tolerance = 1e-2)
  expect_equal(out$AV[6], 20186.18, tolerance = 1e-2)
})

test_that("AV_path_ul_typeA reproduces Example 16.2", {
  qx <- c(.00076, .00081, .00085, .00090, .00095)
  r <- c(.75, .10, .10, .10, .10)
  e <- c(100, 20, 20, 20, 20)
  G <- rep(5000, 5)

  out <- AV_path_ul_typeA(
    G = G,
    r = r,
    e = e,
    qx = qx,
    ic = 0.03,
    iq = 0.03,
    B = 100000
  )

  expect_equal(out$AV[2], 1109.34, tolerance = 1e-2)
  expect_equal(out$AV[3], 5680.62, tolerance = 1e-2)
  expect_equal(out$AV[4], 10389.27, tolerance = 1e-2)
  expect_equal(out$AV[5], 15239.06, tolerance = 1e-2)
  expect_equal(out$AV[6], 20234.86, tolerance = 1e-2)
})

test_that("coi_ul_typeB matches Equation 16.5a", {
  val <- coi_ul_typeB(B = 100000, qx = 0.00076, iq = 0.03)
  expect_equal(val, 73.79, tolerance = 1e-2)
})

test_that("iP_eiul and i_credit_eiul reproduce Example 16.4", {
  idx <- c(1000, 1050, 1200, 1100, 950, 1060, 1150)

  raw <- iP_eiul(idx)
  cred <- i_credit_eiul(
    i_raw = raw,
    part = 1.10,
    floor = 0.01,
    cap = 0.10
  )

  expect_equal(round(100 * raw, 2), c(5.00, 14.29, -8.33, -13.64, 11.58, 8.49))
  expect_equal(round(100 * cred, 2), c(5.50, 10.00, 1.00, 1.00, 10.00, 9.34))
})

test_that("iMA_eiul reproduces the monthly-average growth in Example 16.5", {
  idx <- c(1000, 1020, 1100, 1150, 1080, 1040, 960, 1030, 1000, 1070, 1150, 1200, 1150)

  raw <- iMA_eiul(idx)

  expect_true(abs(raw - 0.0791666667) < 1e-10)
})

test_that("pxtau_ul and tpxtau_ul reproduce Example 16.6", {
  qd <- c(.001, .002, .003, .004, .005)
  qw <- c(.02, .02, .03, .04, .05)

  p <- pxtau_ul(qd = qd, qw = qw, year_end_withdrawal = TRUE)
  tp <- tpxtau_ul(qd = qd, qw = qw, year_end_withdrawal = TRUE)

  expect_equal(round(p, 5), c(.97902, .97804, .96709, .95616, .94525))
  expect_equal(round(tp, 5), c(.97902, .95752, .92601, .88541, .83694))
})

test_that("GMF_rollforward_ul, rt_ul, and Vprefloor_crvm_ul reproduce Example 16.9", {
  gmf10 <- GMF_rollforward_ul(
    GMF_prev = 140.40,
    GMP = 14.49,
    r = 0.04,
    policy_charge = 11.80,
    i = 0.03
  )

  r10 <- rt_ul(AV = 49.182, GMF = gmf10)
  pre <- Vprefloor_crvm_ul(r = r10, pvfb_minus_pvfp = 70)

  expect_equal(gmf10, 146.7857, tolerance = 1e-4)
  expect_equal(r10, 0.33506, tolerance = 1e-5)
  expect_equal(pre, 23.4542, tolerance = 1e-4)
})

test_that("ag38_reserve_ul reproduces Example 16.11", {
  out <- ag38_reserve_ul(
    basic_reserve = 10000,
    deficiency_reserve = 0,
    excess_payment = 60000,
    nsp_required = 100000,
    valuation_nsp = 150000,
    surrender_charge = 5000
  )

  expect_equal(out$prefunding_ratio, 0.60, tolerance = 1e-12)
  expect_equal(out$net_amount_additional, 84000, tolerance = 1e-12)
  expect_equal(out$reduced_deficiency_reserve, 0, tolerance = 1e-12)
  expect_equal(out$step8_reserve, 89000, tolerance = 1e-12)
  expect_equal(out$final_reserve, 89000, tolerance = 1e-12)
})

test_that("ag38_prefunding_ratio is capped at 1", {
  expect_equal(ag38_prefunding_ratio(120000, 100000), 1, tolerance = 1e-12)
})


testthat::test_that("Type B cost of insurance follows the direct formula", {
  testthat::expect_equal(
    coi_ul_typeB(
      B = 100000,
      qx = 0.00076,
      iq = 0.03
    ),
    100000 * 0.00076 / 1.03,
    tolerance = 1e-12
  )
})


testthat::test_that("Type B cost of insurance vectorizes consistently", {
  out <- coi_ul_typeB(
    B = 100000,
    qx = c(0.00076, 0.00081),
    iq = c(0.03, 0.04)
  )

  expected <- 100000 * c(0.00076, 0.00081) /
    (1 + c(0.03, 0.04))

  testthat::expect_equal(out, expected, tolerance = 1e-12)
})


testthat::test_that("Type B account value matches one-period formula", {
  out <- AV_path_ul_typeB(
    G = 5000,
    r = 0.10,
    e = 20,
    qx = 0.001,
    ic = 0.03,
    B = 100000,
    iq = 0.03,
    AV0 = 1000
  )

  coi <- 100000 * 0.001 / 1.03

  expected <- (
    1000 +
      5000 * (1 - 0.10) -
      20 -
      coi
  ) * 1.03

  testthat::expect_equal(out$AV[2], expected, tolerance = 1e-10)
})


testthat::test_that("Type A account value matches explicit formula", {
  out <- AV_path_ul_typeA(
    G = 5000,
    r = 0.10,
    e = 20,
    qx = 0.001,
    ic = 0.03,
    B = 100000,
    iq = 0.03,
    AV0 = 1000
  )

  mortality_factor <- 0.001 * 1.03 / 1.03

  expected <- (
    (1000 + 5000 * 0.90 - 20) * 1.03 -
      100000 * mortality_factor
  ) / (1 - mortality_factor)

  testthat::expect_equal(out$AV[2], expected, tolerance = 1e-10)
})


testthat::test_that("account-value paths recycle period inputs", {
  qx <- c(0.001, 0.002, 0.003)

  type_b <- AV_path_ul_typeB(
    G = 5000,
    r = 0.10,
    e = 20,
    qx = qx,
    ic = 0.03,
    B = c(100000, 95000, 90000)
  )

  type_a <- AV_path_ul_typeA(
    G = 5000,
    r = 0.10,
    e = 20,
    qx = qx,
    ic = 0.03,
    B = c(100000, 95000, 90000)
  )

  testthat::expect_equal(nrow(type_b), 4)
  testthat::expect_equal(nrow(type_a), 4)
  testthat::expect_true(all(is.finite(type_b$AV)))
  testthat::expect_true(all(is.finite(type_a$AV)))
})


testthat::test_that("credited-rate helper works with the default infinite cap", {
  raw <- c(-0.10, 0.05, 0.20)

  out <- i_credit_eiul(raw)

  testthat::expect_equal(
    out,
    c(0, 0.05, 0.20),
    tolerance = 1e-12
  )
})


testthat::test_that("credited rates apply floor cap participation and margin", {
  raw <- c(-0.10, 0.05, 0.20)

  out <- i_credit_eiul(
    raw,
    part = 1.10,
    floor = 0.01,
    cap = 0.10,
    margin = 0.005
  )

  expected <- pmin(
    0.10,
    pmax(0.01, 1.10 * raw - 0.005)
  )

  testthat::expect_equal(out, expected, tolerance = 1e-12)
})


testthat::test_that("point-to-point rates use consecutive index values", {
  index <- c(1000, 1050, 1200, 1100)

  testthat::expect_equal(
    iP_eiul(index),
    index[-1] / index[-length(index)] - 1,
    tolerance = 1e-12
  )
})


testthat::test_that("monthly-average growth uses twelve closing values", {
  index <- c(
    1000, 1020, 1100, 1150, 1080, 1040, 960,
    1030, 1000, 1070, 1150, 1200, 1150
  )

  testthat::expect_equal(
    iMA_eiul(index),
    mean(index[-1]) / index[1] - 1,
    tolerance = 1e-12
  )
})


testthat::test_that("persistency formulas are correct", {
  qd <- c(0.01, 0.02)
  qw <- c(0.03, 0.04)

  year_end <- (1 - qd) * (1 - qw)
  competing <- 1 - qd - qw

  testthat::expect_equal(
    pxtau_ul(qd, qw, year_end_withdrawal = TRUE),
    year_end
  )

  testthat::expect_equal(
    pxtau_ul(qd, qw, year_end_withdrawal = FALSE),
    competing
  )

  testthat::expect_equal(
    tpxtau_ul(qd, qw),
    cumprod(year_end)
  )
})


testthat::test_that("persistency probabilities recycle scalar inputs", {
  out <- pxtau_ul(
    qd = c(0.01, 0.02, 0.03),
    qw = 0.05
  )

  testthat::expect_equal(
    out,
    (1 - c(0.01, 0.02, 0.03)) * 0.95
  )
})


testthat::test_that("guaranteed maturity fund roll-forward vectorizes", {
  out <- GMF_rollforward_ul(
    GMF_prev = c(100, 120),
    GMP = 15,
    r = 0.04,
    policy_charge = c(10, 11),
    i = 0.03
  )

  expected <- (
    c(100, 120) +
      15 * (1 - 0.04) -
      c(10, 11)
  ) * 1.03

  testthat::expect_equal(out, expected, tolerance = 1e-12)
})


testthat::test_that("funding ratio helpers cap results at one", {
  testthat::expect_equal(
    rt_ul(
      AV = c(50, 120),
      GMF = 100
    ),
    c(0.5, 1)
  )

  testthat::expect_equal(
    ag38_prefunding_ratio(
      excess_payment = c(50, 120),
      nsp_required = 100
    ),
    c(0.5, 1)
  )
})


testthat::test_that("pre-floor reserve permits negative APV differences", {
  testthat::expect_equal(
    Vprefloor_crvm_ul(
      r = c(0.5, 0.75),
      pvfb_minus_pvfp = c(100, -20)
    ),
    c(50, -15)
  )
})


testthat::test_that("scalar AG 38 calculation preserves list output", {
  out <- ag38_reserve_ul(
    basic_reserve = 10000,
    deficiency_reserve = 0,
    excess_payment = 60000,
    nsp_required = 100000,
    valuation_nsp = 150000,
    surrender_charge = 5000
  )

  testthat::expect_type(out, "list")

  testthat::expect_named(
    out,
    c(
      "prefunding_ratio",
      "net_amount_additional",
      "reduced_deficiency_reserve",
      "step8_reserve",
      "increased_basic_reserve",
      "final_reserve"
    )
  )

  testthat::expect_equal(out$prefunding_ratio, 0.6)
  testthat::expect_equal(
    out$final_reserve,
    out$increased_basic_reserve
  )
})


testthat::test_that("vectorized AG 38 calculation returns a data frame", {
  out <- ag38_reserve_ul(
    basic_reserve = c(10000, 12000),
    deficiency_reserve = 0,
    excess_payment = c(60000, 80000),
    nsp_required = 100000,
    valuation_nsp = c(150000, 160000),
    surrender_charge = 5000
  )

  testthat::expect_s3_class(out, "data.frame")
  testthat::expect_equal(nrow(out), 2)

  testthat::expect_named(
    out,
    c(
      "prefunding_ratio",
      "net_amount_additional",
      "reduced_deficiency_reserve",
      "step8_reserve",
      "increased_basic_reserve",
      "final_reserve"
    )
  )
})


testthat::test_that("universal life functions reject invalid inputs", {
  testthat::expect_error(
    coi_ul_typeB(
      B = 100000,
      qx = 1.2,
      iq = 0.03
    ),
    "values in \\[0, 1\\]"
  )

  testthat::expect_error(
    coi_ul_typeB(
      B = 100000,
      qx = 0.01,
      iq = -1
    ),
    "greater than -1"
  )

  testthat::expect_error(
    AV_path_ul_typeB(
      G = 5000,
      r = 1.2,
      e = 20,
      qx = 0.001,
      ic = 0.03,
      B = 100000
    ),
    "values in \\[0, 1\\]"
  )

  testthat::expect_error(
    iP_eiul(c(0, 100)),
    "must be positive"
  )

  testthat::expect_error(
    i_credit_eiul(
      i_raw = 0.05,
      floor = 0.10,
      cap = 0.05
    ),
    "floor cannot exceed cap"
  )

  testthat::expect_error(
    pxtau_ul(
      qd = 0.70,
      qw = 0.40,
      year_end_withdrawal = FALSE
    ),
    "must not exceed 1"
  )

  testthat::expect_error(
    rt_ul(
      AV = 100,
      GMF = 0
    ),
    "positive values"
  )

  testthat::expect_error(
    ag38_prefunding_ratio(
      excess_payment = 100,
      nsp_required = 0
    ),
    "positive values"
  )
})


testthat::test_that("universal life functions reject incompatible lengths", {
  testthat::expect_error(
    coi_ul_typeB(
      B = c(100000, 110000),
      qx = c(0.001, 0.002, 0.003),
      iq = 0.03
    ),
    "length 1 or the common length"
  )

  testthat::expect_error(
    AV_path_ul_typeA(
      G = c(5000, 5000),
      r = c(0.10, 0.10, 0.10),
      e = 20,
      qx = c(0.001, 0.002, 0.003),
      ic = 0.03,
      B = 100000
    ),
    "length 1 or the common length"
  )
})

#' Annuity-insurance relationships
#'
#' Identities linking annual and continuous annuity functions to the
#' corresponding insurance functions.
#'
#' @param x Age.
#' @param n Term or deferral period in years.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object for discrete identities.
#' @param model Optional survival model name.
#' @param ... Additional model parameters.
#' @param k_max Maximum summation horizon for non-terminating models.
#' @param tol Truncation tolerance for non-terminating models.
#'
#' @return
#' Numeric vector containing the annuity value computed from the corresponding
#' annuity-insurance identity.
#'
#' @name annuity_relationships
NULL

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.check_rel_i <- function(i) {
  i <- as.numeric(i)

  if (length(i) == 0L || any(!is.finite(i)) || any(i <= -1)) {
    stop("i must contain finite values greater than -1.", call. = FALSE)
  }

  i
}

.check_rel_n <- function(n) {
  n <- as.numeric(n)

  if (length(n) != 1L || !is.finite(n) || n < 0) {
    stop("n must be a single nonnegative number.", call. = FALSE)
  }

  n
}

.check_rel_integer_n <- function(n) {
  n <- .check_rel_n(n)

  if (abs(n - round(n)) > 1e-10) {
    stop("For annual annuity relationships, n must be an integer.", call. = FALSE)
  }

  as.integer(round(n))
}

.recycle_rel_vectors <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) {
    stop("All arguments must have positive length.", call. = FALSE)
  }

  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)

  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) if (length(z) == 1L) rep(z, common_len) else z)
}

.check_rel_x <- function(x) {
  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  invisible(TRUE)
}

# -------------------------------------------------------------------------
# Whole life identities
# -------------------------------------------------------------------------

#' Whole life annuity-immediate from insurance identity
#'
#' Computes \eqn{a_x = (v - A_x)/d}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_ax <- function(x, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_rel_vectors(x, .check_rel_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    d <- i[j] / (1 + i[j])
    v <- 1 / (1 + i[j])

    (v - Ax(x[j], i[j], tbl = tbl, model = model, ...)) / d
  })
}

#' Whole life annuity-due from insurance identity
#'
#' Computes \eqn{\ddot{a}_x = (1 - A_x)/d}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_adotx <- function(x, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_rel_vectors(x, .check_rel_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    d <- i[j] / (1 + i[j])

    (1 - Ax(x[j], i[j], tbl = tbl, model = model, ...)) / d
  })
}

#' Continuous whole life annuity from insurance identity
#'
#' Computes \eqn{\bar{a}_x = (1 - \bar{A}_x)/\delta}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_abarx <- function(x, i, model, ...) {
  args <- .recycle_rel_vectors(x, .check_rel_i(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    delta <- log(1 + i[j])

    (1 - Abarx(x[j], i[j], model = model, ...)) / delta
  })
}

# -------------------------------------------------------------------------
# Temporary identities
# -------------------------------------------------------------------------

#' Temporary annuity-immediate from insurance identity
#'
#' Computes
#' \eqn{a_{x:\overline{n}|} = (1 - A_{x:\overline{n}|})/d - 1 + {}_nE_x}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_axn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_rel_vectors(x, n, .check_rel_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_rel_integer_n(n[j])
    d <- i[j] / (1 + i[j])

    (1 - Axn(x[j], nn, i[j], tbl = tbl, model = model, ...)) / d - 1 +
      nEx(x[j], nn, i[j], tbl = tbl, model = model, ...)
  })
}

#' Temporary annuity-due from insurance identity
#'
#' Computes \eqn{\ddot{a}_{x:\overline{n}|} =
#' (1 - A_{x:\overline{n}|})/d}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_adotxn <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  args <- .recycle_rel_vectors(x, n, .check_rel_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_rel_integer_n(n[j])
    d <- i[j] / (1 + i[j])

    (1 - Axn(x[j], nn, i[j], tbl = tbl, model = model, ...)) / d
  })
}

#' Continuous temporary annuity from insurance identity
#'
#' Computes \eqn{\bar{a}_{x:\overline{n}|} =
#' (1 - \bar{A}_{x:\overline{n}|})/\delta}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_abarxn <- function(x, n, i, model, ...) {
  args <- .recycle_rel_vectors(x, n, .check_rel_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_rel_n(n[j])
    delta <- log(1 + i[j])

    (1 - Abarxn(x[j], nn, i[j], model = model, ...)) / delta
  })
}

# -------------------------------------------------------------------------
# Deferred identities
# -------------------------------------------------------------------------

#' Deferred annuity-immediate from pure endowment identity
#'
#' Computes \eqn{{}_{n|}a_x = {}_nE_x a_{x+n}}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_nax <- function(x, n, i, model = NULL, ..., tbl = NULL,
                                 k_max = 5000, tol = 1e-12) {
  args <- .recycle_rel_vectors(x, n, .check_rel_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_rel_integer_n(n[j])

    nEx(x[j], nn, i[j], tbl = tbl, model = model, ...) *
      ax(
        x[j] + nn,
        i[j],
        tbl = tbl,
        model = model,
        ...,
        k_max = k_max,
        tol = tol
      )
  })
}

#' Deferred annuity-due from pure endowment identity
#'
#' Computes \eqn{{}_{n|}\ddot{a}_x = {}_nE_x \ddot{a}_{x+n}}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_nadotx <- function(x, n, i, model = NULL, ..., tbl = NULL,
                                    k_max = 5000, tol = 1e-12) {
  args <- .recycle_rel_vectors(x, n, .check_rel_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_rel_integer_n(n[j])

    nEx(x[j], nn, i[j], tbl = tbl, model = model, ...) *
      adotx(
        x[j] + nn,
        i[j],
        tbl = tbl,
        model = model,
        ...,
        k_max = k_max,
        tol = tol
      )
  })
}

#' Deferred continuous annuity from pure endowment identity
#'
#' Computes \eqn{{}_{n|}\bar{a}_x = {}_nE_x \bar{a}_{x+n}}.
#'
#' @rdname annuity_relationships
#' @export
annuity_identity_nabarx <- function(x, n, i, model, ..., tol = 1e-10) {
  args <- .recycle_rel_vectors(x, n, .check_rel_i(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_rel_x(x)

  sapply(seq_along(x), function(j) {
    nn <- .check_rel_n(n[j])

    nEx(x[j], nn, i[j], model = model, ...) *
      abarx(x[j] + nn, i[j], model = model, ..., tol = tol)
  })
}

test_that("annuity_identity_ax matches ax", {
  i <- 0.05
  x <- 40

  expect_equal(
    annuity_identity_ax(x, i, model = "uniform", omega = 100),
    ax(x, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("annuity_identity_adotx matches adotx", {
  i <- 0.05
  x <- 40

  expect_equal(
    annuity_identity_adotx(x, i, model = "uniform", omega = 100),
    adotx(x, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("annuity_identity_abarx matches abarx", {
  i <- 0.05
  x <- 40
  lambda <- 0.02

  expect_equal(
    annuity_identity_abarx(x, i, model = "exponential", lambda = lambda),
    abarx(x, i, model = "exponential", lambda = lambda),
    tolerance = 1e-10
  )
})

test_that("annuity_identity_axn matches axn", {
  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    annuity_identity_axn(x, n, i, model = "uniform", omega = 100),
    axn(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("annuity_identity_adotxn matches adotxn", {
  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    annuity_identity_adotxn(x, n, i, model = "uniform", omega = 100),
    adotxn(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("annuity_identity_abarxn matches abarxn", {
  i <- 0.05
  x <- 40
  n <- 10
  lambda <- 0.02

  expect_equal(
    annuity_identity_abarxn(x, n, i, model = "exponential", lambda = lambda),
    abarxn(x, n, i, model = "exponential", lambda = lambda),
    tolerance = 1e-10
  )
})

test_that("annuity_identity_nax matches nax", {
  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    annuity_identity_nax(x, n, i, model = "uniform", omega = 100),
    nax(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("annuity_identity_nadotx matches nadotx", {
  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    annuity_identity_nadotx(x, n, i, model = "uniform", omega = 100),
    nadotx(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("annuity_identity_nabarx matches nabarx", {
  i <- 0.05
  x <- 40
  n <- 10
  lambda <- 0.02

  expect_equal(
    annuity_identity_nabarx(x, n, i, model = "exponential", lambda = lambda),
    nabarx(x, n, i, model = "exponential", lambda = lambda),
    tolerance = 1e-10
  )
})

test_that("whole life immediate identity agrees with textbook formula", {
  i <- 0.05
  d <- i / (1 + i)
  v <- 1 / (1 + i)
  x <- 40

  expect_equal(
    annuity_identity_ax(x, i, model = "uniform", omega = 100),
    (v - Ax(x, i, model = "uniform", omega = 100)) / d,
    tolerance = 1e-10
  )
})

test_that("whole life due identity agrees with textbook formula", {
  i <- 0.05
  d <- i / (1 + i)
  x <- 40

  expect_equal(
    annuity_identity_adotx(x, i, model = "uniform", omega = 100),
    (1 - Ax(x, i, model = "uniform", omega = 100)) / d,
    tolerance = 1e-10
  )
})

test_that("continuous whole life identity agrees with textbook formula", {
  i <- 0.05
  delta <- log(1 + i)
  x <- 40
  lambda <- 0.02

  expect_equal(
    annuity_identity_abarx(x, i, model = "exponential", lambda = lambda),
    (1 - Abarx(x, i, model = "exponential", lambda = lambda)) / delta,
    tolerance = 1e-10
  )
})

test_that("temporary immediate identity agrees with textbook formula", {
  i <- 0.05
  d <- i / (1 + i)
  x <- 40
  n <- 10

  expect_equal(
    annuity_identity_axn(x, n, i, model = "uniform", omega = 100),
    (1 - Axn(x, n, i, model = "uniform", omega = 100)) / d - 1 +
      nEx(x, n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("temporary due identity agrees with textbook formula", {
  i <- 0.05
  d <- i / (1 + i)
  x <- 40
  n <- 10

  expect_equal(
    annuity_identity_adotxn(x, n, i, model = "uniform", omega = 100),
    (1 - Axn(x, n, i, model = "uniform", omega = 100)) / d,
    tolerance = 1e-10
  )
})

test_that("continuous temporary identity agrees with textbook formula", {
  i <- 0.05
  delta <- log(1 + i)
  x <- 40
  n <- 10
  lambda <- 0.02

  expect_equal(
    annuity_identity_abarxn(x, n, i, model = "exponential", lambda = lambda),
    (1 - Abarxn(x, n, i, model = "exponential", lambda = lambda)) / delta,
    tolerance = 1e-10
  )
})

test_that("deferred immediate identity agrees with pure endowment relation", {
  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    annuity_identity_nax(x, n, i, model = "uniform", omega = 100),
    nEx(x, n, i, model = "uniform", omega = 100) *
      ax(x + n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("deferred due identity agrees with pure endowment relation", {
  i <- 0.05
  x <- 40
  n <- 10

  expect_equal(
    annuity_identity_nadotx(x, n, i, model = "uniform", omega = 100),
    nEx(x, n, i, model = "uniform", omega = 100) *
      adotx(x + n, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("deferred continuous identity agrees with pure endowment relation", {
  i <- 0.05
  x <- 40
  n <- 10
  lambda <- 0.02

  expect_equal(
    annuity_identity_nabarx(x, n, i, model = "exponential", lambda = lambda),
    nEx(x, n, i, model = "exponential", lambda = lambda) *
      abarx(x + n, i, model = "exponential", lambda = lambda),
    tolerance = 1e-10
  )
})

test_that("relationship functions vectorize over x", {
  i <- 0.05
  x <- c(40, 45, 50)

  expect_length(
    annuity_identity_ax(x, i, model = "uniform", omega = 100),
    3
  )

  expect_length(
    annuity_identity_adotx(x, i, model = "uniform", omega = 100),
    3
  )

  expect_length(
    annuity_identity_abarx(x, i, model = "exponential", lambda = 0.02),
    3
  )
})

test_that("relationship functions vectorize over x and n", {
  i <- 0.05
  x <- c(40, 45)
  n <- c(10, 5)

  expect_length(
    annuity_identity_axn(x, n, i, model = "uniform", omega = 100),
    2
  )

  expect_length(
    annuity_identity_adotxn(x, n, i, model = "uniform", omega = 100),
    2
  )

  expect_length(
    annuity_identity_abarxn(x, n, i, model = "exponential", lambda = 0.02),
    2
  )

  expect_length(
    annuity_identity_nax(x, n, i, model = "uniform", omega = 100),
    2
  )

  expect_length(
    annuity_identity_nadotx(x, n, i, model = "uniform", omega = 100),
    2
  )

  expect_length(
    annuity_identity_nabarx(x, n, i, model = "exponential", lambda = 0.02),
    2
  )
})

test_that("relationship functions reject incompatible x and n lengths", {
  expect_error(
    annuity_identity_axn(c(40, 45, 50), c(10, 5), 0.05, model = "uniform", omega = 100)
  )

  expect_error(
    annuity_identity_nax(c(40, 45, 50), c(10, 5), 0.05, model = "uniform", omega = 100)
  )
})

test_that("relationship functions return finite values in standard cases", {
  i <- 0.05

  expect_true(is.finite(
    annuity_identity_ax(40, i, model = "uniform", omega = 100)
  ))

  expect_true(is.finite(
    annuity_identity_adotx(40, i, model = "uniform", omega = 100)
  ))

  expect_true(is.finite(
    annuity_identity_abarx(40, i, model = "exponential", lambda = 0.02)
  ))

  expect_true(is.finite(
    annuity_identity_axn(40, 10, i, model = "uniform", omega = 100)
  ))

  expect_true(is.finite(
    annuity_identity_adotxn(40, 10, i, model = "uniform", omega = 100)
  ))

  expect_true(is.finite(
    annuity_identity_abarxn(40, 10, i, model = "exponential", lambda = 0.02)
  ))
})



testthat::test_that("annuity-insurance relationships work with life tables", {
  tbl <- life_table(x = 0:3, lx = c(100000, 80000, 50000, 0))

  testthat::expect_equal(
    annuity_identity_ax(0, i = 0.05, tbl = tbl),
    ax(0, i = 0.05, tbl = tbl),
    tolerance = 1e-10
  )

  testthat::expect_equal(
    annuity_identity_adotx(0, i = 0.05, tbl = tbl),
    adotx(0, i = 0.05, tbl = tbl),
    tolerance = 1e-10
  )
})

testthat::test_that("annuity-insurance relationships vectorize over i", {
  out <- annuity_identity_ax(
    x = 40,
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(out[1] > out[2])
})

#' Continuous insurance models
#'
#' Continuous contingent payment and insurance functions.
#'
#' @name insurance_continuous
#' @keywords internal
NULL

# ------------------------------------------------------------
# Internal helpers
# ------------------------------------------------------------

.check_i_cont <- function(i) {
  i <- as.numeric(i)

  if (length(i) == 0L || any(!is.finite(i)) || any(i <= -1)) {
    stop("i must contain finite values greater than -1.", call. = FALSE)
  }

  i
}

.recycle_cont_vectors <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) {
    stop("All arguments must have positive length.", call. = FALSE)
  }

  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)

  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) {
    if (length(z) == 1L) rep(z, common_len) else z
  })
}


.recycle_x_n_cont <- function(x, n) {
  args <- .recycle_cont_vectors(x, n)

  x <- args[[1L]]
  n <- args[[2L]]

  .check_x_cont(x)
  .check_n_cont(n)

  list(x = x, n = n)
}


.check_x_cont <- function(x) {
  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  invisible(TRUE)
}

.check_n_cont <- function(n) {
  if (any(n < 0)) {
    stop("n must be nonnegative.", call. = FALSE)
  }

  invisible(TRUE)
}

.integrate_cont <- function(f, lower, upper) {
  stats::integrate(
    f,
    lower = lower,
    upper = upper,
    rel.tol = 1e-9,
    subdivisions = 1000L
  )$value
}

.get_upper_cont <- function(x, n = NULL, model, ...) {
  if (tolower(model) == "uniform") {
    dots <- list(...)

    if (is.null(dots$omega)) {
      stop("For model = 'uniform', omega must be supplied.", call. = FALSE)
    }

    upper <- dots$omega - x

    if (!is.null(n)) {
      upper <- pmin(upper, n)
    }

    return(pmax(upper, 0))
  }

  if (is.null(n)) {
    return(Inf)
  }

  n
}

.pure_endowment_cont <- function(x, n, i, model, ...) {
  discount(i, n) * tpx(n, x = x, model = model, ...)
}

# ------------------------------------------------------------
# First moments
# ------------------------------------------------------------

#' Continuous whole life insurance APV
#'
#' Computes
#' \eqn{\bar{A}_x = \int_0^\infty v^t {}_t p_x \mu_{x+t}\,dt}.
#'
#' @param x Age.
#' @param i Effective annual interest rate.
#' @param model Parametric survival model name.
#' @param ... Additional model parameters passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
Abarx <- function(x, i, model, ...) {
  args <- .recycle_cont_vectors(x, .check_i_cont(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_x_cont(x)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    delta <- log(1 + i[j])

    upper <- .get_upper_cont(xx, model = model, ...)

    if (upper <= 0) {
      return(0)
    }

    f <- function(t) {
      val <- exp(-delta * t) *
        tpx(t, x = xx, model = model, ...) *
        hazard0(xx + t, model = model, ...)
      val[!is.finite(val)] <- 0
      val
    }

    .integrate_cont(f, 0, upper)
  })
}

#' Continuous term insurance APV
#'
#' Computes
#' \eqn{\bar{A}_{x:\overline{n}|}^{1} =
#' \int_0^n v^t {}_t p_x \mu_{x+t}\,dt}.
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param model Parametric survival model name.
#' @param ... Additional model parameters passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
Abarxn1 <- function(x, n, i, model, ...) {
  args <- .recycle_cont_vectors(x, n, .check_i_cont(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_cont(x)
  .check_n_cont(n)

  sapply(seq_along(x), function(j) {
    xx <- x[j]
    nn <- n[j]
    delta <- log(1 + i[j])

    upper <- .get_upper_cont(xx, n = nn, model = model, ...)

    if (upper <= 0) {
      return(0)
    }

    f <- function(t) {
      val <- exp(-delta * t) *
        tpx(t, x = xx, model = model, ...) *
        hazard0(xx + t, model = model, ...)
      val[!is.finite(val)] <- 0
      val
    }

    .integrate_cont(f, 0, upper)
  })
}

#' Continuous deferred insurance APV
#'
#' Computes
#' \eqn{{}_{n\mid}\bar{A}_x = v^n {}_n p_x \bar{A}_{x+n}}.
#'
#' @param x Age.
#' @param n Deferral period.
#' @param i Effective annual interest rate.
#' @param model Parametric survival model name.
#' @param ... Additional model parameters passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
nAbarx <- function(x, n, i, model, ...) {
  args <- .recycle_cont_vectors(x, n, .check_i_cont(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_cont(x)
  .check_n_cont(n)

  .pure_endowment_cont(x, n, i, model = model, ...) *
    Abarx(x + n, i, model = model, ...)
}

#' Continuous endowment insurance APV
#'
#' Computes
#' \eqn{\bar{A}_{x:\overline{n}|} =
#' \bar{A}_{x:\overline{n}|}^{1} + v^n {}_n p_x}.
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param model Parametric survival model name.
#' @param ... Additional model parameters passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
Abarxn <- function(x, n, i, model, ...) {
  args <- .recycle_cont_vectors(x, n, .check_i_cont(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  Abarxn1(x, n, i, model = model, ...) +
    .pure_endowment_cont(x, n, i, model = model, ...)
}

# ------------------------------------------------------------
# Second moments
# ------------------------------------------------------------

#' Second moment of continuous whole life insurance PV
#'
#' Computes \eqn{{}^{2}\bar{A}_x} by evaluating \eqn{\bar{A}_x}
#' at doubled force.
#'
#' @inheritParams Abarx
#' @return Numeric vector of second moments.
#' @export
A2barx <- function(x, i, model, ...) {
  Abarx(x, i = double_force_i(i), model = model, ...)
}

#' Second moment of continuous term insurance PV
#'
#' Computes \eqn{{}^{2}\bar{A}_{x:\overline{n}|}^{1}} by evaluating
#' \eqn{\bar{A}_{x:\overline{n}|}^{1}} at doubled force.
#'
#' @inheritParams Abarxn1
#' @return Numeric vector of second moments.
#' @export
A2barxn1 <- function(x, n, i, model, ...) {
  Abarxn1(x, n, i = double_force_i(i), model = model, ...)
}

#' Second moment of continuous deferred insurance PV
#'
#' Computes \eqn{{}^{2}{}_{n\mid}\bar{A}_x} by evaluating
#' \eqn{{}_{n\mid}\bar{A}_x} at doubled force.
#'
#' @inheritParams nAbarx
#' @return Numeric vector of second moments.
#' @export
A2nAbarx <- function(x, n, i, model, ...) {
  nAbarx(x, n, i = double_force_i(i), model = model, ...)
}

#' Second moment of continuous endowment insurance PV
#'
#' Computes \eqn{{}^{2}\bar{A}_{x:\overline{n}|}}.
#'
#' @inheritParams Abarxn
#' @return Numeric vector of second moments.
#' @export
A2barxn <- function(x, n, i, model, ...) {
  args <- .recycle_cont_vectors(x, n, .check_i_cont(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  A2barxn1(x, n, i, model = model, ...) +
    .pure_endowment_cont(x, n, double_force_i(i), model = model, ...)
}

# ------------------------------------------------------------
# Variances
# ------------------------------------------------------------

#' Variance of continuous whole life insurance PV
#'
#' Computes \eqn{\mathrm{Var}(\bar{Z}_x) =
#' {}^{2}\bar{A}_x - \bar{A}_x^2}.
#'
#' @inheritParams Abarx
#' @return Numeric vector of variances.
#' @export
var_Abarx <- function(x, i, model, ...) {
  A2barx(x, i, model = model, ...) -
    Abarx(x, i, model = model, ...)^2
}

#' Variance of continuous term insurance PV
#'
#' Computes
#' \eqn{\mathrm{Var}(\bar{Z}_{x:\overline{n}|}^{1}) =
#' {}^{2}\bar{A}_{x:\overline{n}|}^{1}
#' - (\bar{A}_{x:\overline{n}|}^{1})^2}.
#'
#' @inheritParams Abarxn1
#' @return Numeric vector of variances.
#' @export
var_Abarxn1 <- function(x, n, i, model, ...) {
  A2barxn1(x, n, i, model = model, ...) -
    Abarxn1(x, n, i, model = model, ...)^2
}

#' Variance of continuous deferred insurance PV
#'
#' @inheritParams nAbarx
#' @return Numeric vector of variances.
#' @export
var_nAbarx <- function(x, n, i, model, ...) {
  A2nAbarx(x, n, i, model = model, ...) -
    nAbarx(x, n, i, model = model, ...)^2
}

#' Variance of continuous endowment insurance PV
#'
#' @inheritParams Abarxn
#' @return Numeric vector of variances.
#' @export
var_Abarxn <- function(x, n, i, model, ...) {
  A2barxn(x, n, i, model = model, ...) -
    Abarxn(x, n, i, model = model, ...)^2
}

test_that("Abarxn decomposes into continuous term insurance plus pure endowment", {
  i <- 0.05

  expect_equal(
    Abarxn(40, 10, i, model = "uniform", omega = 100),
    Abarxn1(40, 10, i, model = "uniform", omega = 100) +
      discount(i, 10) * tpx(10, x = 40, model = "uniform", omega = 100)
  )
})

test_that("nAbarx equals v^n * npx * Abarx at age x+n", {
  i <- 0.05

  expect_equal(
    nAbarx(40, 10, i, model = "uniform", omega = 100),
    discount(i, 10) *
      tpx(10, x = 40, model = "uniform", omega = 100) *
      Abarx(50, i, model = "uniform", omega = 100)
  )
})

test_that("A2barx evaluates Abarx at doubled force", {
  i <- 0.05

  expect_equal(
    A2barx(40, i, model = "uniform", omega = 100),
    Abarx(40, double_force_i(i), model = "uniform", omega = 100)
  )
})

test_that("A2barxn1 evaluates Abarxn1 at doubled force", {
  i <- 0.05

  expect_equal(
    A2barxn1(40, 10, i, model = "uniform", omega = 100),
    Abarxn1(40, 10, double_force_i(i), model = "uniform", omega = 100)
  )
})

test_that("A2nAbarx evaluates nAbarx at doubled force", {
  i <- 0.05

  expect_equal(
    A2nAbarx(40, 10, i, model = "uniform", omega = 100),
    nAbarx(40, 10, double_force_i(i), model = "uniform", omega = 100)
  )
})

test_that("A2barxn decomposes into second moments of term and pure endowment", {
  i <- 0.05
  i2 <- double_force_i(i)

  expect_equal(
    A2barxn(40, 10, i, model = "uniform", omega = 100),
    A2barxn1(40, 10, i, model = "uniform", omega = 100) +
      discount(i2, 10) * tpx(10, x = 40, model = "uniform", omega = 100)
  )
})

test_that("var_Abarx equals A2barx - Abarx^2", {
  i <- 0.05

  expect_equal(
    var_Abarx(40, i, model = "uniform", omega = 100),
    A2barx(40, i, model = "uniform", omega = 100) -
      Abarx(40, i, model = "uniform", omega = 100)^2
  )
})

test_that("var_Abarxn1 equals A2barxn1 - Abarxn1^2", {
  i <- 0.05

  expect_equal(
    var_Abarxn1(40, 10, i, model = "uniform", omega = 100),
    A2barxn1(40, 10, i, model = "uniform", omega = 100) -
      Abarxn1(40, 10, i, model = "uniform", omega = 100)^2
  )
})

test_that("var_nAbarx equals A2nAbarx - nAbarx^2", {
  i <- 0.05

  expect_equal(
    var_nAbarx(40, 10, i, model = "uniform", omega = 100),
    A2nAbarx(40, 10, i, model = "uniform", omega = 100) -
      nAbarx(40, 10, i, model = "uniform", omega = 100)^2
  )
})

test_that("var_Abarxn equals A2barxn - Abarxn^2", {
  i <- 0.05

  expect_equal(
    var_Abarxn(40, 10, i, model = "uniform", omega = 100),
    A2barxn(40, 10, i, model = "uniform", omega = 100) -
      Abarxn(40, 10, i, model = "uniform", omega = 100)^2
  )
})

test_that("continuous insurance functions vectorize correctly", {
  i <- 0.05

  expect_length(Abarxn1(c(40, 45), c(10, 5), i, model = "uniform", omega = 100), 2)
  expect_length(nAbarx(c(40, 45), c(10, 5), i, model = "uniform", omega = 100), 2)
  expect_length(Abarxn(c(40, 45), c(10, 5), i, model = "uniform", omega = 100), 2)
})

test_that("continuous insurance functions reject invalid inputs", {
  i <- 0.05

  expect_error(Abarx(-1, i, model = "uniform", omega = 100))
  expect_error(Abarxn1(40, -1, i, model = "uniform", omega = 100))
  expect_error(nAbarx(40, -1, i, model = "uniform", omega = 100))
  expect_error(Abarx(40, -1, model = "uniform", omega = 100))
})

test_that("Abarx matches closed form under exponential survival", {
  i <- 0.05
  lambda <- 0.25
  delta <- interest_convert(i = i)$delta

  expected <- lambda / (delta + lambda)

  expect_equal(
    Abarx(40, i, model = "exponential", lambda = lambda),
    expected,
    tolerance = 1e-8
  )
})

test_that("Abarxn matches closed form under exponential survival", {
  i <- 0.05
  lambda <- 0.25
  delta <- interest_convert(i = i)$delta
  n <- 10

  expected <- (lambda + delta * exp(-(delta + lambda) * n)) / (delta + lambda)

  expect_equal(
    Abarxn(40, n, i, model = "exponential", lambda = lambda),
    expected,
    tolerance = 1e-8
  )
})

test_that("Abarxn1 equals Abarx minus nAbarx", {
  i <- 0.05

  expect_equal(
    Abarxn1(40, 10, i, model = "uniform", omega = 100),
    Abarx(40, i, model = "uniform", omega = 100) -
      nAbarx(40, 10, i, model = "uniform", omega = 100)
  )
})

test_that("continuous whole life APV under uniform survival is between 0 and 1", {
  i <- 0.05
  val <- Abarx(40, i, model = "uniform", omega = 100)

  expect_true(is.numeric(val))
  expect_true(val > 0)
  expect_true(val < 1)
})

test_that("continuous term insurance APV is not greater than continuous whole life APV", {
  i <- 0.05

  term_val <- Abarxn1(40, 10, i, model = "uniform", omega = 100)
  whole_val <- Abarx(40, i, model = "uniform", omega = 100)

  expect_true(term_val <= whole_val)
})

test_that("continuous endowment APV is at least the pure endowment component", {
  i <- 0.05
  pure_val <- discount(i, 10) * tpx(10, x = 40, model = "uniform", omega = 100)
  endow_val <- Abarxn(40, 10, i, model = "uniform", omega = 100)

  expect_true(endow_val >= pure_val)
})


testthat::test_that("continuous insurance functions vectorize over i", {
  out <- Abarx(x = 40, i = c(0.03, 0.05), model = "uniform", omega = 100)

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
  testthat::expect_true(out[1] > out[2])
})

testthat::test_that("continuous term insurance functions recycle x, n, and i", {
  out <- Abarxn1(
    x = c(40, 50),
    n = 10,
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})

#' Insurance utilities
#'
#' Helper functions for insurance models.
#'
#' These utilities focus on:
#' \itemize{
#'   \item doubled-force calculations for second moments,
#'   \item UDD approximation multipliers,
#'   \item UDD approximations for continuous and m-thly insurance functions.
#' }
#'
#' General interest conversion functions such as \code{interest_convert()}
#' and \code{discount()} are defined elsewhere in the package and are reused here.
#'
#' @name insurance_utils
#' @keywords internal
NULL

# ------------------------------------------------------------
# Internal helpers
# ------------------------------------------------------------

.check_numeric_finite_ins_util <- function(x, name) {
  x <- as.numeric(x)

  if (length(x) == 0L || any(!is.finite(x))) {
    stop(name, " must contain finite numeric values.", call. = FALSE)
  }

  x
}

.check_i_ins_util <- function(i) {
  i <- .check_numeric_finite_ins_util(i, "i")

  if (any(i <= -1)) {
    stop("i must be greater than -1.", call. = FALSE)
  }

  i
}

.check_m_ins_util <- function(m) {
  m <- as.numeric(m)

  if (length(m) != 1L || !is.finite(m) || m <= 0 ||
      abs(m - round(m)) > 1e-10) {
    stop("m must be a positive integer.", call. = FALSE)
  }

  as.integer(round(m))
}

# ------------------------------------------------------------
# Doubled-force helpers for second moments
# ------------------------------------------------------------

#' Effective annual interest at doubled force
#'
#' If \eqn{\delta' = 2\delta}, then
#' \eqn{i' = (1+i)^2 - 1}.
#'
#' @param i Numeric vector of effective annual interest rates.
#'
#' @return Numeric vector of effective annual rates corresponding to doubled force.
#' @export
double_force_i <- function(i) {
  i <- .check_i_ins_util(i)
  (1 + i)^2 - 1
}

#' Doubled force of interest
#'
#' Computes \eqn{\delta' = 2\delta}.
#'
#' @param delta Numeric vector of forces of interest.
#'
#' @return Numeric vector of doubled forces of interest.
#' @export
double_force_delta <- function(delta) {
  delta <- .check_numeric_finite_ins_util(delta, "delta")
  2 * delta
}

# ------------------------------------------------------------
# UDD multipliers
# ------------------------------------------------------------

#' UDD multiplier for continuous insurance approximations
#'
#' Under UDD,
#' \eqn{\bar{A}_x = (i/\delta) A_x}
#' and similarly for term and deferred insurance.
#'
#' @param i Numeric vector of effective annual interest rates.
#'
#' @return Numeric vector equal to \eqn{i/\delta}.
#' @export
udd_continuous_multiplier <- function(i) {
  i <- .check_i_ins_util(i)

  delta <- log(1 + i)
  ifelse(abs(delta) < 1e-12, 1, i / delta)
}

#' UDD multiplier for m-thly insurance approximations
#'
#' Under UDD,
#' \eqn{A_x^{(m)} = (i / i^{(m)}) A_x}
#' and similarly for term and deferred insurance.
#'
#' @param i Numeric vector of effective annual interest rates.
#' @param m Positive integer payment frequency.
#'
#' @return Numeric vector equal to \eqn{i / i^{(m)}}.
#' @export
udd_mthly_multiplier <- function(i, m) {
  i <- .check_i_ins_util(i)
  m <- .check_m_ins_util(m)

  im <- m * ((1 + i)^(1 / m) - 1)
  ifelse(abs(im) < 1e-12, 1, i / im)
}

# ------------------------------------------------------------
# UDD approximations: continuous insurance
# ------------------------------------------------------------

#' UDD approximation of continuous whole life insurance
#'
#' Computes \eqn{\bar{A}_x = (i/\delta)A_x}.
#'
#' @param Ax Discrete whole life insurance APV.
#' @param i Effective annual interest rate.
#'
#' @return Continuous whole life insurance APV under UDD.
#' @export
Abarx_udd <- function(Ax, i) {
  Ax <- .check_numeric_finite_ins_util(Ax, "Ax")
  udd_continuous_multiplier(i) * Ax
}

#' UDD approximation of continuous term insurance
#'
#' Computes \eqn{\bar{A}_{x:\overline{n}|}^{1} = (i/\delta)A_{x:\overline{n}|}^{1}}.
#'
#' @param Axn1 Discrete term insurance APV.
#' @param i Effective annual interest rate.
#'
#' @return Continuous term insurance APV under UDD.
#' @export
Abarxn1_udd <- function(Axn1, i) {
  Axn1 <- .check_numeric_finite_ins_util(Axn1, "Axn1")
  udd_continuous_multiplier(i) * Axn1
}

#' UDD approximation of continuous deferred insurance
#'
#' Computes \eqn{{}_{n\mid}\bar{A}_x = (i/\delta)\,{}_{n\mid}A_x}.
#'
#' @param nAx Discrete deferred insurance APV.
#' @param i Effective annual interest rate.
#'
#' @return Continuous deferred insurance APV under UDD.
#' @export
nAbarx_udd <- function(nAx, i) {
  nAx <- .check_numeric_finite_ins_util(nAx, "nAx")
  udd_continuous_multiplier(i) * nAx
}

#' UDD approximation of continuous endowment insurance
#'
#' Computes
#' \eqn{\bar{A}_{x:\overline{n}|} = (i/\delta)A_{x:\overline{n}|}^{1} + {}_nE_x}.
#'
#' @param Axn1 Discrete term insurance APV.
#' @param nEx Pure endowment APV, \eqn{{}_nE_x}.
#' @param i Effective annual interest rate.
#'
#' @return Continuous endowment insurance APV under UDD.
#' @export
Abarxn_udd <- function(Axn1, nEx, i) {
  Axn1 <- .check_numeric_finite_ins_util(Axn1, "Axn1")
  nEx <- .check_numeric_finite_ins_util(nEx, "nEx")

  Abarxn1_udd(Axn1, i) + nEx
}

# ------------------------------------------------------------
# UDD approximations: m-thly insurance
# ------------------------------------------------------------

#' UDD approximation of m-thly whole life insurance
#'
#' Computes \eqn{A_x^{(m)} = (i/i^{(m)})A_x}.
#'
#' @param Ax Discrete whole life insurance APV.
#' @param i Effective annual interest rate.
#' @param m Positive integer payment frequency.
#'
#' @return m-thly whole life insurance APV under UDD.
#' @export
Ax_m_udd <- function(Ax, i, m) {
  Ax <- .check_numeric_finite_ins_util(Ax, "Ax")
  udd_mthly_multiplier(i, m) * Ax
}

#' UDD approximation of m-thly term insurance
#'
#' Computes \eqn{A_{x:\overline{n}|}^{1(m)} = (i/i^{(m)})A_{x:\overline{n}|}^{1}}.
#'
#' @param Axn1 Discrete term insurance APV.
#' @param i Effective annual interest rate.
#' @param m Positive integer payment frequency.
#'
#' @return m-thly term insurance APV under UDD.
#' @export
Axn1_m_udd <- function(Axn1, i, m) {
  Axn1 <- .check_numeric_finite_ins_util(Axn1, "Axn1")
  udd_mthly_multiplier(i, m) * Axn1
}

#' UDD approximation of m-thly deferred insurance
#'
#' Computes \eqn{{}_{n\mid}A_x^{(m)} = (i/i^{(m)})\,{}_{n\mid}A_x}.
#'
#' @param nAx Discrete deferred insurance APV.
#' @param i Effective annual interest rate.
#' @param m Positive integer payment frequency.
#'
#' @return m-thly deferred insurance APV under UDD.
#' @export
nAx_m_udd <- function(nAx, i, m) {
  nAx <- .check_numeric_finite_ins_util(nAx, "nAx")
  udd_mthly_multiplier(i, m) * nAx
}

#' UDD approximation of m-thly endowment insurance
#'
#' Computes
#' \eqn{A_{x:\overline{n}|}^{(m)} = (i/i^{(m)})A_{x:\overline{n}|}^{1} + {}_nE_x}.
#'
#' @param Axn1 Discrete term insurance APV.
#' @param nEx Pure endowment APV.
#' @param i Effective annual interest rate.
#' @param m Positive integer payment frequency.
#'
#' @return m-thly endowment insurance APV under UDD.
#' @export
Axn_m_udd <- function(Axn1, nEx, i, m) {
  Axn1 <- .check_numeric_finite_ins_util(Axn1, "Axn1")
  nEx <- .check_numeric_finite_ins_util(nEx, "nEx")

  Axn1_m_udd(Axn1, i, m) + nEx
}

test_that("double_force_i computes i' = (1+i)^2 - 1", {
  expect_equal(double_force_i(0.05), (1.05)^2 - 1)
  expect_equal(double_force_i(c(0.02, 0.05)),
               c((1.02)^2 - 1, (1.05)^2 - 1))
})

test_that("double_force_i rejects invalid interest rates", {
  expect_error(double_force_i(-1))
  expect_error(double_force_i(c(0.05, NA)))
})

test_that("double_force_delta computes 2 * delta", {
  expect_equal(double_force_delta(0.08), 0.16)
  expect_equal(double_force_delta(c(0.01, 0.03)), c(0.02, 0.06))
})

test_that("double_force_delta rejects non-finite inputs", {
  expect_error(double_force_delta(Inf))
  expect_error(double_force_delta(NA))
})

test_that("udd_continuous_multiplier computes i / delta", {
  i <- 0.05
  delta <- interest_convert(i = i)$delta
  expect_equal(udd_continuous_multiplier(i), i / delta)
})

test_that("udd_continuous_multiplier handles zero interest correctly", {
  expect_equal(udd_continuous_multiplier(0), 1)
})

test_that("udd_continuous_multiplier rejects invalid interest rates", {
  expect_error(udd_continuous_multiplier(-1))
  expect_error(udd_continuous_multiplier(NA))
})

test_that("udd_mthly_multiplier computes i / i^(m)", {
  i <- 0.06
  m <- 12
  im <- interest_convert(i = i, m = m)$im
  expect_equal(udd_mthly_multiplier(i, m), i / im)
})

test_that("udd_mthly_multiplier handles zero interest correctly", {
  expect_equal(udd_mthly_multiplier(0, 12), 1)
})

test_that("udd_mthly_multiplier rejects invalid inputs", {
  expect_error(udd_mthly_multiplier(0.05, 0))
  expect_error(udd_mthly_multiplier(0.05, 2.5))
  expect_error(udd_mthly_multiplier(-1, 12))
})

test_that("Abarx_udd computes (i/delta) * Ax", {
  Ax <- 0.30
  i <- 0.05
  expected <- udd_continuous_multiplier(i) * Ax
  expect_equal(Abarx_udd(Ax, i), expected)
})

test_that("Abarx_udd vectorizes", {
  Ax <- c(0.20, 0.40)
  i <- c(0.03, 0.05)
  expected <- udd_continuous_multiplier(i) * Ax
  expect_equal(Abarx_udd(Ax, i), expected)
})

test_that("Abarx_udd rejects invalid Ax", {
  expect_error(Abarx_udd(NA, 0.05))
})

test_that("Abarxn1_udd computes (i/delta) * Axn1", {
  Axn1 <- 0.18
  i <- 0.04
  expected <- udd_continuous_multiplier(i) * Axn1
  expect_equal(Abarxn1_udd(Axn1, i), expected)
})

test_that("nAbarx_udd computes (i/delta) * nAx", {
  nAx <- 0.11
  i <- 0.05
  expected <- udd_continuous_multiplier(i) * nAx
  expect_equal(nAbarx_udd(nAx, i), expected)
})

test_that("Abarxn_udd computes continuous endowment approximation", {
  Axn1 <- 0.25
  nEx <- 0.40
  i <- 0.05
  expected <- udd_continuous_multiplier(i) * Axn1 + nEx
  expect_equal(Abarxn_udd(Axn1, nEx, i), expected)
})

test_that("Abarxn_udd rejects invalid inputs", {
  expect_error(Abarxn_udd(NA, 0.4, 0.05))
  expect_error(Abarxn_udd(0.2, NA, 0.05))
})

test_that("Ax_m_udd computes (i / i^(m)) * Ax", {
  Ax <- 0.33
  i <- 0.06
  m <- 4
  expected <- udd_mthly_multiplier(i, m) * Ax
  expect_equal(Ax_m_udd(Ax, i, m), expected)
})

test_that("Axn1_m_udd computes (i / i^(m)) * Axn1", {
  Axn1 <- 0.17
  i <- 0.06
  m <- 12
  expected <- udd_mthly_multiplier(i, m) * Axn1
  expect_equal(Axn1_m_udd(Axn1, i, m), expected)
})

test_that("nAx_m_udd computes (i / i^(m)) * nAx", {
  nAx <- 0.09
  i <- 0.04
  m <- 2
  expected <- udd_mthly_multiplier(i, m) * nAx
  expect_equal(nAx_m_udd(nAx, i, m), expected)
})

test_that("Axn_m_udd computes m-thly endowment approximation", {
  Axn1 <- 0.20
  nEx <- 0.50
  i <- 0.05
  m <- 12
  expected <- udd_mthly_multiplier(i, m) * Axn1 + nEx
  expect_equal(Axn_m_udd(Axn1, nEx, i, m), expected)
})

test_that("Axn_m_udd rejects invalid inputs", {
  expect_error(Axn_m_udd(NA, 0.5, 0.05, 12))
  expect_error(Axn_m_udd(0.2, NA, 0.05, 12))
})

test_that("continuous and m-thly multipliers are close for large m", {
  i <- 0.05
  expect_equal(
    udd_mthly_multiplier(i, 10000),
    udd_continuous_multiplier(i),
    tolerance = 1e-4
  )
})

#' m-thly contingent annuity functions
#'
#' Life annuities payable m-thly.
#'
#' @param x Age.
#' @param n Term or deferral period in years.
#' @param m Number of payments per year.
#' @param i Effective annual interest rate.
#' @param model Survival model name.
#' @param ... Additional parameters passed to the survival model.
#' @param k_max Maximum summation horizon for non-terminating models.
#' @param tol Truncation tolerance for non-terminating models.
#'
#' @return
#' Numeric vector containing the requested m-thly annuity value or actuarial
#' accumulated value.
#'
#' @name annuity_mthly
NULL

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.check_ann_m <- function(m) {
  m <- as.numeric(m)

  if (length(m) != 1L || !is.finite(m) || m <= 0 ||
      abs(m - round(m)) > 1e-10) {
    stop("m must be a positive integer.", call. = FALSE)
  }

  as.integer(round(m))
}

.check_i_ann_m <- function(i) {
  i <- as.numeric(i)

  if (length(i) == 0L || any(!is.finite(i)) || any(i <= -1)) {
    stop("i must contain finite values greater than -1.", call. = FALSE)
  }

  i
}

.recycle_ann_m_vectors <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) {
    stop("All arguments must have positive length.", call. = FALSE)
  }

  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)

  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) if (length(z) == 1L) rep(z, common_len) else z)
}

.check_x_ann_m <- function(x) {
  if (any(x < 0)) stop("x must be nonnegative.", call. = FALSE)
  invisible(TRUE)
}

.check_n_ann_m <- function(n) {
  if (any(n < 0)) stop("n must be nonnegative.", call. = FALSE)
  invisible(TRUE)
}

.max_subperiod_ann_m <- function(x, m, model, ...) {
  dots <- list(...)
  model <- tolower(model)

  if (model == "uniform") {
    omega <- dots$omega

    if (is.null(omega)) {
      stop("For model = 'uniform', omega must be supplied.", call. = FALSE)
    }

    return(max(0L, floor((omega - x) * m)))
  }

  Inf
}

.sum_mthly_annuity_terms <- function(x, i, m, model, ..., due = FALSE,
                                     temporary = NULL, k_max = 200000,
                                     tol = 1e-12) {
  x <- as.numeric(x)
  i <- as.numeric(i)

  if (length(x) != 1L || !is.finite(x) || x < 0) {
    stop("x must be a single nonnegative finite number.", call. = FALSE)
  }

  if (length(i) != 1L || !is.finite(i) || i <= -1) {
    stop("i must be a single finite value greater than -1.", call. = FALSE)
  }

  if (!is.null(temporary)) {
    temporary <- as.numeric(temporary)

    if (length(temporary) != 1L || !is.finite(temporary) || temporary < 0) {
      stop("temporary must be a single nonnegative finite number.", call. = FALSE)
    }
  }

  j <- (1 + i)^(1 / m) - 1
  vsub <- 1 / (1 + j)

  max_sub <- .max_subperiod_ann_m(x, m, model, ...)

  if (due) {
    if (!is.null(temporary)) {
      n_pay <- round(temporary * m)

      if (abs(temporary * m - n_pay) > 1e-10) {
        stop("n * m must be an integer for m-thly annuity functions.", call. = FALSE)
      }

      if (n_pay == 0L) return(0)
      idx <- 0:(n_pay - 1L)

    } else if (is.finite(max_sub)) {
      idx <- 0:max_sub
    } else {
      idx <- 0:k_max
    }
  } else {
    if (!is.null(temporary)) {
      n_pay <- round(temporary * m)

      if (abs(temporary * m - n_pay) > 1e-10) {
        stop("n * m must be an integer for m-thly annuity functions.", call. = FALSE)
      }

      if (n_pay == 0L) return(0)
      idx <- 1:n_pay

    } else if (is.finite(max_sub)) {
      if (max_sub <= 0L) return(0)
      idx <- 1:max_sub
    } else {
      idx <- 1:k_max
    }
  }

  times <- idx / m
  terms <- (1 / m) * (vsub^idx) * tpx(times, x = x, model = model, ...)

  terms[!is.finite(terms)] <- 0

  if (!is.null(temporary) || is.finite(max_sub)) {
    return(sum(terms))
  }

  acc <- 0
  small_run <- 0L

  for (jj in seq_along(terms)) {
    acc <- acc + terms[jj]

    if (terms[jj] < tol) {
      small_run <- small_run + 1L
      if (small_run >= 20L) break
    } else {
      small_run <- 0L
    }
  }

  acc
}

# -------------------------------------------------------------------------
# Whole life m-thly annuities
# -------------------------------------------------------------------------

#' Whole life m-thly annuity-immediate
#'
#' @rdname annuity_mthly
#' @export
ax_m <- function(x, m, i, model, ..., k_max = 200000, tol = 1e-12) {
  m <- .check_ann_m(m)

  args <- .recycle_ann_m_vectors(x, .check_i_ann_m(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_x_ann_m(x)

  sapply(seq_along(x), function(j) {
    .sum_mthly_annuity_terms(
      x = x[j], i = i[j], m = m, model = model, ...,
      due = FALSE, temporary = NULL, k_max = k_max, tol = tol
    )
  })
}

#' Whole life m-thly annuity-due
#'
#' @rdname annuity_mthly
#' @export
adotx_m <- function(x, m, i, model, ..., k_max = 200000, tol = 1e-12) {
  m <- .check_ann_m(m)

  args <- .recycle_ann_m_vectors(x, .check_i_ann_m(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_x_ann_m(x)

  sapply(seq_along(x), function(j) {
    .sum_mthly_annuity_terms(
      x = x[j], i = i[j], m = m, model = model, ...,
      due = TRUE, temporary = NULL, k_max = k_max, tol = tol
    )
  })
}

# -------------------------------------------------------------------------
# Temporary m-thly annuities
# -------------------------------------------------------------------------

#' Temporary m-thly annuity-immediate
#'
#' @rdname annuity_mthly
#' @export
axn_m <- function(x, n, m, i, model, ...) {
  m <- .check_ann_m(m)

  args <- .recycle_ann_m_vectors(x, n, .check_i_ann_m(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_ann_m(x)
  .check_n_ann_m(n)

  sapply(seq_along(x), function(j) {
    .sum_mthly_annuity_terms(
      x = x[j], i = i[j], m = m, model = model, ...,
      due = FALSE, temporary = n[j]
    )
  })
}

#' Temporary m-thly annuity-due
#'
#' @rdname annuity_mthly
#' @export
adotxn_m <- function(x, n, m, i, model, ...) {
  m <- .check_ann_m(m)

  args <- .recycle_ann_m_vectors(x, n, .check_i_ann_m(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_ann_m(x)
  .check_n_ann_m(n)

  sapply(seq_along(x), function(j) {
    .sum_mthly_annuity_terms(
      x = x[j], i = i[j], m = m, model = model, ...,
      due = TRUE, temporary = n[j]
    )
  })
}

# -------------------------------------------------------------------------
# Deferred m-thly annuities
# -------------------------------------------------------------------------

#' Deferred whole life m-thly annuity-immediate
#'
#' @rdname annuity_mthly
#' @export
nax_m <- function(x, n, m, i, model, ..., k_max = 200000, tol = 1e-12) {
  m <- .check_ann_m(m)

  args <- .recycle_ann_m_vectors(x, n, .check_i_ann_m(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_ann_m(x)
  .check_n_ann_m(n)

  sapply(seq_along(x), function(j) {
    nEx(x[j], n[j], i[j], model = model, ...) *
      ax_m(
        x[j] + n[j],
        m = m,
        i = i[j],
        model = model,
        ...,
        k_max = k_max,
        tol = tol
      )
  })
}

#' Deferred whole life m-thly annuity-due
#'
#' @rdname annuity_mthly
#' @export
nadotx_m <- function(x, n, m, i, model, ..., k_max = 200000, tol = 1e-12) {
  m <- .check_ann_m(m)

  args <- .recycle_ann_m_vectors(x, n, .check_i_ann_m(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_ann_m(x)
  .check_n_ann_m(n)

  sapply(seq_along(x), function(j) {
    nEx(x[j], n[j], i[j], model = model, ...) *
      adotx_m(
        x[j] + n[j],
        m = m,
        i = i[j],
        model = model,
        ...,
        k_max = k_max,
        tol = tol
      )
  })
}

# -------------------------------------------------------------------------
# Actuarial accumulated values
# -------------------------------------------------------------------------

#' Temporary m-thly annuity-immediate actuarial accumulated value
#'
#' @rdname annuity_mthly
#' @export
sxn_m <- function(x, n, m, i, model, ...) {
  m <- .check_ann_m(m)

  args <- .recycle_ann_m_vectors(x, n, .check_i_ann_m(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_ann_m(x)
  .check_n_ann_m(n)

  sapply(seq_along(x), function(j) {
    ex <- nEx(x[j], n[j], i[j], model = model, ...)

    if (is.na(ex) || ex <= 0) return(Inf)

    axn_m(x[j], n[j], m = m, i = i[j], model = model, ...) / ex
  })
}

#' Temporary m-thly annuity-due actuarial accumulated value
#'
#' @rdname annuity_mthly
#' @export
sdotxn_m <- function(x, n, m, i, model, ...) {
  m <- .check_ann_m(m)

  args <- .recycle_ann_m_vectors(x, n, .check_i_ann_m(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_ann_m(x)
  .check_n_ann_m(n)

  sapply(seq_along(x), function(j) {
    ex <- nEx(x[j], n[j], i[j], model = model, ...)

    if (is.na(ex) || ex <= 0) return(Inf)

    adotxn_m(x[j], n[j], m = m, i = i[j], model = model, ...) / ex
  })
}

test_that("whole life m-thly due and immediate differ by 1/m", {
  i <- 0.05
  m <- 12

  expect_equal(
    adotx_m(40, m, i, model = "uniform", omega = 100),
    ax_m(40, m, i, model = "uniform", omega = 100) + 1 / m,
    tolerance = 1e-10
  )

  expect_equal(
    adotx_m(40, m, i, model = "exponential", lambda = 0.02),
    ax_m(40, m, i, model = "exponential", lambda = 0.02) + 1 / m,
    tolerance = 1e-10
  )
})

test_that("temporary m-thly due and immediate relation matches textbook formula", {
  i <- 0.05
  m <- 4
  x <- 40
  n <- 10

  expect_equal(
    adotxn_m(x, n, m, i, model = "uniform", omega = 100),
    axn_m(x, n, m, i, model = "uniform", omega = 100) + (1 / m) * (1 - nEx(x, n, i, model = "uniform", omega = 100)),
    tolerance = 1e-10
  )
})

test_that("deferred immediate identity holds", {
  i <- 0.05
  m <- 12
  x <- 40
  n <- 10

  expect_equal(
    nax_m(x, n, m, i, model = "uniform", omega = 100),
    nEx(x, n, i, model = "uniform", omega = 100) *
      ax_m(x + n, m, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    nax_m(x, n, m, i, model = "exponential", lambda = 0.02),
    nEx(x, n, i, model = "exponential", lambda = 0.02) *
      ax_m(x + n, m, i, model = "exponential", lambda = 0.02),
    tolerance = 1e-10
  )
})

test_that("deferred due identity holds", {
  i <- 0.05
  m <- 12
  x <- 40
  n <- 10

  expect_equal(
    nadotx_m(x, n, m, i, model = "uniform", omega = 100),
    nEx(x, n, i, model = "uniform", omega = 100) *
      adotx_m(x + n, m, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    nadotx_m(x, n, m, i, model = "exponential", lambda = 0.02),
    nEx(x, n, i, model = "exponential", lambda = 0.02) *
      adotx_m(x + n, m, i, model = "exponential", lambda = 0.02),
    tolerance = 1e-10
  )
})

test_that("whole life decomposition into temporary plus deferred holds", {
  i <- 0.05
  m <- 4
  x <- 40
  n <- 10

  expect_equal(
    ax_m(x, m, i, model = "uniform", omega = 100),
    axn_m(x, n, m, i, model = "uniform", omega = 100) +
      nax_m(x, n, m, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    adotx_m(x, m, i, model = "uniform", omega = 100),
    adotxn_m(x, n, m, i, model = "uniform", omega = 100) +
      nadotx_m(x, n, m, i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("actuarial accumulated values equal APV divided by nEx", {
  i <- 0.05
  m <- 12
  x <- 40
  n <- 10

  ex <- nEx(x, n, i, model = "uniform", omega = 100)

  expect_equal(
    sxn_m(x, n, m, i, model = "uniform", omega = 100),
    axn_m(x, n, m, i, model = "uniform", omega = 100) / ex,
    tolerance = 1e-10
  )

  expect_equal(
    sdotxn_m(x, n, m, i, model = "uniform", omega = 100),
    adotxn_m(x, n, m, i, model = "uniform", omega = 100) / ex,
    tolerance = 1e-10
  )
})

test_that("m-thly annuities vectorize over x", {
  i <- 0.05
  m <- 12
  x <- c(40, 45, 50)

  out1 <- ax_m(x, m, i, model = "uniform", omega = 100)
  out2 <- adotx_m(x, m, i, model = "uniform", omega = 100)

  expect_length(out1, 3)
  expect_length(out2, 3)
  expect_true(all(is.finite(out1)))
  expect_true(all(is.finite(out2)))
})

test_that("m-thly temporary and deferred annuities vectorize over x and n", {
  i <- 0.05
  m <- 4
  x <- c(40, 45)
  n <- c(10, 5)

  out1 <- axn_m(x, n, m, i, model = "uniform", omega = 100)
  out2 <- adotxn_m(x, n, m, i, model = "uniform", omega = 100)
  out3 <- nax_m(x, n, m, i, model = "uniform", omega = 100)
  out4 <- nadotx_m(x, n, m, i, model = "uniform", omega = 100)

  expect_length(out1, 2)
  expect_length(out2, 2)
  expect_length(out3, 2)
  expect_length(out4, 2)

  expect_true(all(is.finite(out1)))
  expect_true(all(is.finite(out2)))
  expect_true(all(is.finite(out3)))
  expect_true(all(is.finite(out4)))
})

test_that("m-thly annuity values lie between annual and continuous values in standard cases", {
  i <- 0.05
  x <- 40

  a_annual <- ax(x, i, model = "uniform", omega = 100)
  a_cont <- abarx(x, i, model = "uniform", omega = 100)
  a_m <- ax_m(x, 12, i, model = "uniform", omega = 100)

  expect_true(a_annual < a_m)
  expect_true(a_m < a_cont)

  ad_annual <- adotx(x, i, model = "uniform", omega = 100)
  ad_m <- adotx_m(x, 12, i, model = "uniform", omega = 100)

  expect_true(a_m < ad_m)
  expect_true(ad_m < ad_annual)
})

test_that("m-thly temporary annuity is not larger than whole life annuity", {
  i <- 0.05
  m <- 12
  x <- 40
  n <- 10

  expect_true(
    axn_m(x, n, m, i, model = "uniform", omega = 100) <=
      ax_m(x, m, i, model = "uniform", omega = 100)
  )

  expect_true(
    adotxn_m(x, n, m, i, model = "uniform", omega = 100) <=
      adotx_m(x, m, i, model = "uniform", omega = 100)
  )
})

test_that("deferred annuity is zero when nEx is zero under finite support", {
  i <- 0.05
  m <- 12

  expect_equal(
    nax_m(99, 2, m, i, model = "uniform", omega = 100),
    0
  )

  expect_equal(
    nadotx_m(99, 2, m, i, model = "uniform", omega = 100),
    0
  )
})

test_that("temporary annuity with n = 0 is zero", {
  i <- 0.05
  m <- 12

  expect_equal(axn_m(40, 0, m, i, model = "uniform", omega = 100), 0)
  expect_equal(adotxn_m(40, 0, m, i, model = "uniform", omega = 100), 0)
})

test_that("functions reject invalid m", {
  expect_error(
    ax_m(40, m = 0, i = 0.05, model = "uniform", omega = 100),
    "m must be a positive integer"
  )

  expect_error(
    ax_m(40, m = 2.5, i = 0.05, model = "uniform", omega = 100),
    "m must be a positive integer"
  )
})

test_that("functions reject incompatible x and n lengths", {
  expect_error(
    axn_m(c(40, 45), c(10, 5, 3), m = 12, i = 0.05, model = "uniform", omega = 100),
    "compatible lengths"
  )
})

test_that("functions reject non-integer n * m", {
  expect_error(
    axn_m(40, n = 10.1, m = 4, i = 0.05, model = "uniform", omega = 100),
    "n \\* m must be an integer"
  )
})



testthat::test_that("m-thly annuities vectorize over interest rates", {
  out <- ax_m(
    x = 40,
    m = 12,
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(out[1] > out[2])
})

testthat::test_that("m-thly temporary annuities recycle x, n, and i", {
  out <- axn_m(
    x = c(40, 50),
    n = 10,
    m = 12,
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})

# -------------------------------------------------------------------------
# Internal validation and recycling helpers
# -------------------------------------------------------------------------

#' @noRd
.decrement_check_probability <- function(x, name = "x") {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(x))) {
    stop(name, " must contain finite values.", call. = FALSE)
  }

  if (any(x < 0 | x > 1)) {
    stop(name, " must contain values in [0, 1].", call. = FALSE)
  }

  x
}

#' @noRd
.decrement_check_nonnegative <- function(x, name = "x") {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(x)) || any(x < 0)) {
    stop(
      name,
      " must contain nonnegative finite values.",
      call. = FALSE
    )
  }

  x
}

#' @noRd
.decrement_check_nonnegative_scalar <- function(x, name = "x") {
  x <- .decrement_check_nonnegative(x, name)

  if (length(x) != 1L) {
    stop(
      name,
      " must be a single nonnegative finite number.",
      call. = FALSE
    )
  }

  x
}

#' @noRd
.decrement_check_integerish <- function(x, name = "x", positive = FALSE) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  invalid_sign <- if (positive) {
    x <= 0
  } else {
    x < 0
  }

  if (any(!is.finite(x)) ||
      any(invalid_sign) ||
      any(abs(x - round(x)) > 1e-10)) {
    descriptor <- if (positive) "positive" else "nonnegative"

    stop(
      name,
      " must contain ",
      descriptor,
      " integer-like values.",
      call. = FALSE
    )
  }

  as.integer(round(x))
}

#' @noRd
.decrement_recycle <- function(..., .names = NULL) {
  values <- list(...)

  if (length(values) == 0L) {
    return(values)
  }

  lengths <- vapply(values, length, integer(1))

  if (any(lengths == 0L)) {
    stop("Inputs must have positive length.", call. = FALSE)
  }

  common_length <- max(lengths)

  if (is.null(.names)) {
    .names <- paste0("Argument ", seq_along(values))
  }

  if (length(.names) != length(values)) {
    stop(
      ".names must have the same length as the supplied arguments.",
      call. = FALSE
    )
  }

  for (k in seq_along(values)) {
    if (!lengths[k] %in% c(1L, common_length)) {
      stop(
        .names[k],
        " must have length 1 or the common length ",
        common_length,
        ".",
        call. = FALSE
      )
    }

    values[[k]] <- rep(values[[k]], length.out = common_length)
  }

  values
}

#' @noRd
.decrement_scalar_or_matrix <- function(rows, column_names) {
  out <- do.call(rbind, rows)
  colnames(out) <- column_names
  rownames(out) <- NULL

  if (nrow(out) == 1L) {
    return(stats::setNames(as.numeric(out[1L, ]), column_names))
  }

  out
}

#' @noRd
.decrement_validate_table <- function(tbl) {
  if (!is.data.frame(tbl)) {
    stop(
      "tbl must be a data frame produced by md_table().",
      call. = FALSE
    )
  }

  required <- c("x", "qtau", "ptau", "ltau", "dtau")
  missing_columns <- setdiff(required, names(tbl))

  if (length(missing_columns) > 0L) {
    stop(
      "tbl is missing required columns: ",
      paste(missing_columns, collapse = ", "),
      ".",
      call. = FALSE
    )
  }

  if (nrow(tbl) == 0L) {
    stop("tbl must contain at least one row.", call. = FALSE)
  }

  x <- as.numeric(tbl$x)

  if (any(!is.finite(x)) ||
      any(abs(x - round(x)) > 1e-10)) {
    stop(
      "tbl$x must contain finite integer-like values.",
      call. = FALSE
    )
  }

  if (anyDuplicated(x)) {
    stop("tbl$x must not contain duplicate values.", call. = FALSE)
  }

  if (length(x) > 1L && any(diff(x) != 1)) {
    stop(
      "tbl$x must contain consecutive increasing ages or durations.",
      call. = FALSE
    )
  }

  q_columns <- grep("^q[0-9]+$", names(tbl), value = TRUE)
  d_columns <- grep("^d[0-9]+$", names(tbl), value = TRUE)

  if (length(q_columns) == 0L) {
    stop(
      "tbl must contain at least one cause-specific q column.",
      call. = FALSE
    )
  }

  if (length(d_columns) != length(q_columns)) {
    stop(
      "tbl must contain matching cause-specific q and d columns.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}

#' @noRd
.decrement_table_rows <- function(tbl, x, n) {
  start <- match(x, tbl$x)

  if (is.na(start)) {
    stop("x must be present in tbl$x.", call. = FALSE)
  }

  if (n == 0L) {
    return(integer(0))
  }

  end_age <- x + n - 1L
  end <- match(end_age, tbl$x)

  if (is.na(end)) {
    stop(
      "The interval from x through x + n - 1 must be contained in tbl$x.",
      call. = FALSE
    )
  }

  start:end
}

# -------------------------------------------------------------------------
# Discrete multiple-decrement quantities
# -------------------------------------------------------------------------

#' Total one-year decrement probability
#'
#' Computes the total one-year multiple-decrement probability
#' \deqn{q_x^{(\tau)}=\sum_j q_x^{(j)}.}
#'
#' @param qxj Numeric vector of cause-specific decrement probabilities.
#'
#' @return A numeric scalar.
#'
#' @examples
#' qxtau(c(0.011, 0.100))
#'
#' @export
qxtau <- function(qxj) {
  qxj <- .decrement_check_probability(qxj, "qxj")

  total <- sum(qxj)

  if (total > 1 + sqrt(.Machine$double.eps)) {
    stop("sum(qxj) cannot exceed 1.", call. = FALSE)
  }

  min(total, 1)
}

#' Total one-year survival probability
#'
#' Computes
#' \deqn{p_x^{(\tau)}=1-q_x^{(\tau)}.}
#'
#' @inheritParams qxtau
#'
#' @return A numeric scalar.
#'
#' @examples
#' pxtau(c(0.011, 0.100))
#'
#' @export
pxtau <- function(qxj) {
  1 - qxtau(qxj)
}

#' Cause-specific numbers of decrements
#'
#' Computes
#' \deqn{d_x^{(j)}=l_x^{(\tau)}q_x^{(j)}.}
#'
#' @param lxtau Number alive at age or duration \eqn{x} in the
#'   multiple-decrement table.
#' @param qxj Numeric vector of cause-specific decrement probabilities.
#'
#' @return A numeric vector containing the number of decrements from each
#'   cause.
#'
#' @examples
#' dxj(1000, c(0.011, 0.100))
#'
#' @export
dxj <- function(lxtau, qxj) {
  lxtau <- .decrement_check_nonnegative_scalar(lxtau, "lxtau")
  qxj <- .decrement_check_probability(qxj, "qxj")

  if (sum(qxj) > 1 + sqrt(.Machine$double.eps)) {
    stop("sum(qxj) cannot exceed 1.", call. = FALSE)
  }

  lxtau * qxj
}

#' Total number of decrements
#'
#' Computes
#' \deqn{d_x^{(\tau)}=\sum_j d_x^{(j)}.}
#'
#' @inheritParams dxj
#'
#' @return A numeric scalar.
#'
#' @examples
#' dxtau(1000, c(0.011, 0.100))
#'
#' @export
dxtau <- function(lxtau, qxj) {
  sum(dxj(lxtau = lxtau, qxj = qxj))
}

#' Construct a multiple-decrement table
#'
#' Constructs a discrete multiple-decrement table from cause-specific
#' one-year decrement probabilities.
#'
#' @param x Integer vector of consecutive ages or durations.
#' @param qxj Numeric matrix or data frame of cause-specific decrement
#'   probabilities. Rows correspond to values in \code{x}, and columns
#'   correspond to causes.
#' @param radix Initial value of \eqn{l_x^{(\tau)}}.
#'
#' @return An object of classes \code{"md_table"} and \code{"data.frame"}
#'   containing age or duration, cause-specific decrement probabilities,
#'   total decrement and survival probabilities, numbers alive, and
#'   cause-specific and total decrements.
#'
#' @examples
#' ages <- 45:50
#' qmat <- cbind(
#'   withdrawal = c(0.011, 0.012, 0.013, 0.014, 0.015, 0.016),
#'   retirement = rep(0.100, 6)
#' )
#'
#' md_table(ages, qmat, radix = 1000)
#'
#' @export
md_table <- function(x, qxj, radix = 100000) {
  x <- .decrement_check_integerish(x, "x")
  radix <- .decrement_check_nonnegative_scalar(radix, "radix")

  if (anyDuplicated(x)) {
    stop("x must not contain duplicate values.", call. = FALSE)
  }

  if (length(x) > 1L && any(diff(x) != 1L)) {
    stop(
      "x must contain consecutive increasing ages or durations.",
      call. = FALSE
    )
  }

  qxj <- as.matrix(qxj)

  if (length(qxj) == 0L || ncol(qxj) == 0L) {
    stop("qxj must contain at least one cause column.", call. = FALSE)
  }

  suppressWarnings(storage.mode(qxj) <- "double")

  if (nrow(qxj) != length(x)) {
    stop("nrow(qxj) must equal length(x).", call. = FALSE)
  }

  if (any(!is.finite(qxj)) ||
      any(qxj < 0) ||
      any(qxj > 1)) {
    stop("qxj entries must lie in [0, 1].", call. = FALSE)
  }

  qtau <- rowSums(qxj)

  if (any(qtau > 1 + sqrt(.Machine$double.eps))) {
    stop(
      "Each row sum of qxj must not exceed 1.",
      call. = FALSE
    )
  }

  qtau <- pmin(qtau, 1)
  ptau <- 1 - qtau

  row_count <- length(x)
  ltau <- numeric(row_count)
  ltau[1L] <- radix

  if (row_count > 1L) {
    for (k in 2:row_count) {
      ltau[k] <- ltau[k - 1L] * ptau[k - 1L]
    }
  }

  decrement_matrix <- qxj * ltau
  dtau <- ltau * qtau

  cause_count <- ncol(qxj)
  cause_names <- colnames(qxj)

  if (is.null(cause_names) ||
      any(!nzchar(cause_names)) ||
      anyDuplicated(cause_names)) {
    cause_names <- paste0("cause", seq_len(cause_count))
  }

  out <- data.frame(
    x = x,
    stringsAsFactors = FALSE
  )

  for (k in seq_len(cause_count)) {
    out[[paste0("q", k)]] <- qxj[, k]
  }

  out$qtau <- qtau
  out$ptau <- ptau
  out$ltau <- ltau

  for (k in seq_len(cause_count)) {
    out[[paste0("d", k)]] <- decrement_matrix[, k]
  }

  out$dtau <- dtau

  attr(out, "cause_names") <- cause_names
  class(out) <- c("md_table", "data.frame")

  out
}

#' Multiple-decrement survival probability from a table
#'
#' Computes
#' \deqn{{}_np_x^{(\tau)}
#' =\prod_{k=0}^{n-1}p_{x+k}^{(\tau)}.}
#'
#' @param tbl A multiple-decrement table produced by \code{md_table()}.
#' @param x Starting age or duration. May be scalar or vector.
#' @param n Nonnegative integer term. May be scalar or vector.
#'
#' @return A numeric vector of survival probabilities.
#'
#' @examples
#' ages <- 45:50
#' qmat <- cbind(
#'   q1 = c(0.011, 0.012, 0.013, 0.014, 0.015, 0.016),
#'   q2 = rep(0.100, 6)
#' )
#' tbl <- md_table(ages, qmat, radix = 1000)
#'
#' npxtau_md(tbl, x = 46, n = 3)
#'
#' @export
npxtau_md <- function(tbl, x, n) {
  .decrement_validate_table(tbl)

  x <- .decrement_check_integerish(x, "x")
  n <- .decrement_check_integerish(n, "n")

  values <- .decrement_recycle(
    x,
    n,
    .names = c("x", "n")
  )

  x <- values[[1]]
  n <- values[[2]]

  vapply(seq_along(x), function(k) {
    if (n[k] == 0L) {
      return(1)
    }

    rows <- .decrement_table_rows(
      tbl = tbl,
      x = x[k],
      n = n[k]
    )

    prod(tbl$ptau[rows])
  }, numeric(1))
}

#' Cause-specific multiple-decrement probability from a table
#'
#' Computes the probability of decrement from cause \code{j} within
#' \code{n} years:
#' \deqn{
#' {}_nq_x^{(j)}
#' =
#' \sum_{k=0}^{n-1}
#' {}_kp_x^{(\tau)}q_{x+k}^{(j)}.
#' }
#'
#' @inheritParams npxtau_md
#' @param j Positive integer cause index. May be scalar or vector.
#'
#' @return A numeric vector of cause-specific decrement probabilities.
#'
#' @examples
#' ages <- 45:50
#' qmat <- cbind(
#'   q1 = c(0.011, 0.012, 0.013, 0.014, 0.015, 0.016),
#'   q2 = rep(0.100, 6)
#' )
#' tbl <- md_table(ages, qmat, radix = 1000)
#'
#' nqxj_md(tbl, x = 46, n = 2, j = 1)
#'
#' @export
nqxj_md <- function(tbl, x, n, j) {
  .decrement_validate_table(tbl)

  x <- .decrement_check_integerish(x, "x")
  n <- .decrement_check_integerish(n, "n")
  j <- .decrement_check_integerish(j, "j", positive = TRUE)

  values <- .decrement_recycle(
    x,
    n,
    j,
    .names = c("x", "n", "j")
  )

  x <- values[[1]]
  n <- values[[2]]
  j <- values[[3]]

  cause_count <- length(grep("^q[0-9]+$", names(tbl)))

  if (any(j > cause_count)) {
    stop(
      "j must identify an available cause in tbl.",
      call. = FALSE
    )
  }

  vapply(seq_along(x), function(k) {
    if (n[k] == 0L) {
      return(0)
    }

    rows <- .decrement_table_rows(
      tbl = tbl,
      x = x[k],
      n = n[k]
    )

    cause_column <- paste0("q", j[k])
    survival_to_start <- c(
      1,
      cumprod(tbl$ptau[rows])[-length(rows)]
    )

    sum(survival_to_start * tbl[[cause_column]][rows])
  }, numeric(1))
}

#' Total multiple-decrement probability from a table
#'
#' Computes
#' \deqn{{}_nq_x^{(\tau)}=1-{}_np_x^{(\tau)}.}
#'
#' @inheritParams npxtau_md
#'
#' @return A numeric vector of total decrement probabilities.
#'
#' @examples
#' ages <- 45:50
#' qmat <- cbind(
#'   q1 = c(0.011, 0.012, 0.013, 0.014, 0.015, 0.016),
#'   q2 = rep(0.100, 6)
#' )
#' tbl <- md_table(ages, qmat, radix = 1000)
#'
#' nqxtau_md(tbl, x = 46, n = 2)
#'
#' @export
nqxtau_md <- function(tbl, x, n) {
  1 - npxtau_md(tbl = tbl, x = x, n = n)
}

# -------------------------------------------------------------------------
# Constant-force multiple-decrement quantities
# -------------------------------------------------------------------------

#' Single-decrement survival under a constant force
#'
#' Computes
#' \deqn{{}_tp_x^{\prime(j)}=\exp(-\mu_jt).}
#'
#' @param mu Nonnegative constant force of decrement. May be scalar or
#'   vector.
#' @param t Nonnegative time. May be scalar or vector.
#'
#' @return A numeric vector.
#'
#' @examples
#' tpxprimej_cf(0.10, 5)
#' tpxprimej_cf(0.10, c(1, 5, 10))
#'
#' @export
tpxprimej_cf <- function(mu, t) {
  mu <- .decrement_check_nonnegative(mu, "mu")
  t <- .decrement_check_nonnegative(t, "t")

  values <- .decrement_recycle(
    mu,
    t,
    .names = c("mu", "t")
  )

  mu <- values[[1]]
  t <- values[[2]]

  exp(-mu * t)
}

#' Single-decrement failure under a constant force
#'
#' Computes
#' \deqn{{}_tq_x^{\prime(j)}
#' =1-\exp(-\mu_jt).}
#'
#' @inheritParams tpxprimej_cf
#'
#' @return A numeric vector.
#'
#' @examples
#' tqxprimej_cf(0.10, 5)
#'
#' @export
tqxprimej_cf <- function(mu, t) {
  1 - tpxprimej_cf(mu = mu, t = t)
}

#' Total survival under constant cause-specific forces
#'
#' Computes
#' \deqn{
#' {}_tp_x^{(\tau)}
#' =
#' \exp\left(-t\sum_j\mu_j\right).
#' }
#'
#' @param mu Numeric vector of nonnegative cause-specific forces.
#' @param t Nonnegative time. May be scalar or vector.
#'
#' @return A numeric vector.
#'
#' @examples
#' tpx_tau_cf(c(0.10, 0.20), 5)
#' tpx_tau_cf(c(0.10, 0.20), c(1, 5, 10))
#'
#' @export
tpx_tau_cf <- function(mu, t) {
  mu <- .decrement_check_nonnegative(mu, "mu")
  t <- .decrement_check_nonnegative(t, "t")

  exp(-sum(mu) * t)
}

#' Cause-specific decrement probability under constant forces
#'
#' Computes
#' \deqn{
#' {}_tq_x^{(j)}
#' =
#' \frac{\mu_j}{\sum_k\mu_k}
#' \left[
#' 1-\exp\left(-t\sum_k\mu_k\right)
#' \right].
#' }
#'
#' @param mu Numeric vector of nonnegative cause-specific forces.
#' @param j Positive integer cause index.
#' @param t Nonnegative time. May be scalar or vector.
#'
#' @return A numeric vector.
#'
#' @examples
#' tqxj_cf(c(0.10, 0.20), j = 1, t = 5)
#'
#' @export
tqxj_cf <- function(mu, j, t) {
  mu <- .decrement_check_nonnegative(mu, "mu")
  j <- .decrement_check_integerish(j, "j", positive = TRUE)
  t <- .decrement_check_nonnegative(t, "t")

  if (length(j) != 1L) {
    stop("j must be a single cause index.", call. = FALSE)
  }

  if (j > length(mu)) {
    stop("j must identify an available cause in mu.", call. = FALSE)
  }

  total_force <- sum(mu)

  if (total_force == 0) {
    return(rep(0, length(t)))
  }

  (mu[j] / total_force) *
    (1 - exp(-total_force * t))
}

# -------------------------------------------------------------------------
# Multiple-decrement uniform distribution relationships
# -------------------------------------------------------------------------

#' Associated single-decrement probabilities under MUDD
#'
#' Converts dependent multiple-decrement probabilities to associated
#' single-decrement probabilities under the multiple-decrement uniform
#' distribution assumption.
#'
#' @param qxj Numeric vector of dependent cause-specific decrement
#'   probabilities.
#'
#' @return A numeric vector of associated single-decrement probabilities.
#'
#' @examples
#' qxprime_mudd(c(0.20, 0.10))
#'
#' @export
qxprime_mudd <- function(qxj) {
  qxj <- .decrement_check_probability(qxj, "qxj")
  qtau <- sum(qxj)

  if (qtau > 1 + sqrt(.Machine$double.eps)) {
    stop("sum(qxj) cannot exceed 1.", call. = FALSE)
  }

  if (qtau == 0) {
    return(rep(0, length(qxj)))
  }

  if (qtau >= 1) {
    stop(
      "sum(qxj) must be less than 1 for the MUDD conversion.",
      call. = FALSE
    )
  }

  1 - (1 - qtau)^(qxj / qtau)
}

#' Fractional-year associated single-decrement probabilities under MUDD
#'
#' Computes the associated single-decrement probabilities through time
#' \code{t} under the multiple-decrement uniform distribution assumption.
#'
#' @param qxj Numeric vector of dependent cause-specific decrement
#'   probabilities.
#' @param t A single time in \eqn{[0,1]}.
#'
#' @return A numeric vector of associated single-decrement probabilities.
#'
#' @examples
#' tqxprime_mudd(c(0.20, 0.10), t = 0.5)
#'
#' @export
tqxprime_mudd <- function(qxj, t) {
  qxj <- .decrement_check_probability(qxj, "qxj")
  t <- .decrement_check_probability(t, "t")

  if (length(t) != 1L) {
    stop("t must be a single value in [0, 1].", call. = FALSE)
  }

  qtau <- sum(qxj)

  if (qtau > 1 + sqrt(.Machine$double.eps)) {
    stop("sum(qxj) cannot exceed 1.", call. = FALSE)
  }

  if (qtau == 0 || t == 0) {
    return(rep(0, length(qxj)))
  }

  if (qtau >= 1 && t >= 1) {
    stop(
      "t * sum(qxj) must be less than 1 for the MUDD conversion.",
      call. = FALSE
    )
  }

  1 - (1 - t * qtau)^(qxj / qtau)
}

# -------------------------------------------------------------------------
# Constant-force conversion from associated single-decrement probabilities
# -------------------------------------------------------------------------

#' Multiple-decrement probabilities under constant forces
#'
#' Converts associated single-decrement probabilities to dependent
#' cause-specific decrement probabilities under constant forces.
#'
#' @param qxprime Numeric vector of associated single-decrement
#'   probabilities. Each value must be less than one.
#'
#' @return A numeric vector of dependent cause-specific decrement
#'   probabilities.
#'
#' @examples
#' qx_dep_cf(c(0.20, 0.10))
#'
#' @export
qx_dep_cf <- function(qxprime) {
  qxprime <- .decrement_check_probability(qxprime, "qxprime")

  if (any(qxprime >= 1)) {
    stop(
      "qxprime values must be less than 1 under a finite constant-force model.",
      call. = FALSE
    )
  }

  mu <- -log1p(-qxprime)
  total_force <- sum(mu)

  if (total_force == 0) {
    return(rep(0, length(qxprime)))
  }

  total_probability <- -expm1(-total_force)

  (mu / total_force) * total_probability
}

# -------------------------------------------------------------------------
# Single-decrement uniform distribution relationships
# -------------------------------------------------------------------------

#' Multiple-decrement probabilities under SUDD
#'
#' Converts two associated single-decrement probabilities to dependent
#' multiple-decrement probabilities under the single-decrement uniform
#' distribution assumption.
#'
#' @param q1prime Associated single-decrement probability for cause 1.
#'   May be scalar or vector.
#' @param q2prime Associated single-decrement probability for cause 2.
#'   May be scalar or vector.
#'
#' @return For scalar input, a named numeric vector containing \code{q1}
#'   and \code{q2}. For vectorized input, a numeric matrix with columns
#'   \code{q1} and \code{q2}.
#'
#' @examples
#' qx_dep_sudd(0.20, 0.10)
#'
#' @export
qx_dep_sudd <- function(q1prime, q2prime) {
  q1prime <- .decrement_check_probability(q1prime, "q1prime")
  q2prime <- .decrement_check_probability(q2prime, "q2prime")

  values <- .decrement_recycle(
    q1prime,
    q2prime,
    .names = c("q1prime", "q2prime")
  )

  q1prime <- values[[1]]
  q2prime <- values[[2]]

  rows <- lapply(seq_along(q1prime), function(k) {
    c(
      q1 = q1prime[k] * (1 - 0.5 * q2prime[k]),
      q2 = q2prime[k] * (1 - 0.5 * q1prime[k])
    )
  })

  .decrement_scalar_or_matrix(
    rows,
    column_names = c("q1", "q2")
  )
}

#' Associated single-decrement probabilities under SUDD
#'
#' Converts two dependent multiple-decrement probabilities to associated
#' single-decrement probabilities under the single-decrement uniform
#' distribution assumption.
#'
#' @param q1 Dependent decrement probability for cause 1. May be scalar or
#'   vector.
#' @param q2 Dependent decrement probability for cause 2. May be scalar or
#'   vector.
#'
#' @return For scalar input, a named numeric vector containing
#'   \code{q1prime} and \code{q2prime}. For vectorized input, a numeric
#'   matrix with columns \code{q1prime} and \code{q2prime}.
#'
#' @examples
#' qxprime_sudd(0.20, 0.10)
#'
#' @export
qxprime_sudd <- function(q1, q2) {
  q1 <- .decrement_check_probability(q1, "q1")
  q2 <- .decrement_check_probability(q2, "q2")

  values <- .decrement_recycle(
    q1,
    q2,
    .names = c("q1", "q2")
  )

  q1 <- values[[1]]
  q2 <- values[[2]]

  if (any(q1 + q2 > 1 + sqrt(.Machine$double.eps))) {
    stop(
      "q1 and q2 must satisfy q1 + q2 <= 1.",
      call. = FALSE
    )
  }

  rows <- lapply(seq_along(q1), function(k) {
    if (q1[k] == 0 && q2[k] == 0) {
      return(c(q1prime = 0, q2prime = 0))
    }

    coefficient_b <- -(q1[k] - q2[k] + 2)
    coefficient_c <- 2 * q1[k]
    discriminant <- coefficient_b^2 - 4 * coefficient_c

    if (discriminant < -sqrt(.Machine$double.eps)) {
      stop(
        "No real SUDD solution exists for one or more input pairs.",
        call. = FALSE
      )
    }

    discriminant <- max(discriminant, 0)

    roots <- c(
      (-coefficient_b - sqrt(discriminant)) / 2,
      (-coefficient_b + sqrt(discriminant)) / 2
    )

    roots <- roots[
      roots >= -sqrt(.Machine$double.eps) &
        roots <= 1 + sqrt(.Machine$double.eps)
    ]

    if (length(roots) == 0L) {
      stop(
        "No admissible SUDD solution exists in [0, 1].",
        call. = FALSE
      )
    }

    roots <- pmin(pmax(roots, 0), 1)

    candidate_rows <- lapply(roots, function(q1prime) {
      denominator <- 1 - 0.5 * q1prime

      if (denominator <= 0) {
        return(NULL)
      }

      q2prime <- q2[k] / denominator

      if (!is.finite(q2prime) ||
          q2prime < -sqrt(.Machine$double.eps) ||
          q2prime > 1 + sqrt(.Machine$double.eps)) {
        return(NULL)
      }

      q2prime <- min(max(q2prime, 0), 1)

      reconstructed <- qx_dep_sudd(q1prime, q2prime)

      error <- max(abs(
        reconstructed -
          c(q1 = q1[k], q2 = q2[k])
      ))

      list(
        q1prime = q1prime,
        q2prime = q2prime,
        error = error
      )
    })

    candidate_rows <- Filter(Negate(is.null), candidate_rows)

    if (length(candidate_rows) == 0L) {
      stop(
        "No admissible SUDD solution exists in [0, 1].",
        call. = FALSE
      )
    }

    errors <- vapply(
      candidate_rows,
      function(candidate) candidate$error,
      numeric(1)
    )

    best <- candidate_rows[[which.min(errors)]]

    c(
      q1prime = best$q1prime,
      q2prime = best$q2prime
    )
  })

  .decrement_scalar_or_matrix(
    rows,
    column_names = c("q1prime", "q2prime")
  )
}

test_that("qxtau and pxtau work", {
  expect_equal(qxtau(c(0.011, 0.100)), 0.111)
  expect_equal(pxtau(c(0.011, 0.100)), 0.889)
})

test_that("dxj and dxtau work", {
  expect_equal(dxj(1000, c(0.011, 0.100)), c(11, 100))
  expect_equal(dxtau(1000, c(0.011, 0.100)), 111)
})

test_that("md_table reproduces Example 13.1", {
  x <- 45:50
  qmat <- cbind(
    q1 = c(.011, .012, .013, .014, .015, .016),
    q2 = c(.100, .100, .100, .100, .100, .100)
  )

  tbl <- md_table(x = x, qxj = qmat, radix = 1000)

  expect_equal(round(tbl$qtau[1], 3), 0.111)
  expect_equal(round(tbl$ptau[1], 3), 0.889)
  expect_equal(round(tbl$ltau[2], 2), 889.00)
  expect_equal(round(tbl$d1[1], 2), 11.00)
  expect_equal(round(tbl$d2[1], 2), 100.00)
})

test_that("n-year probabilities from md_table match Example 13.2", {
  x <- 45:50
  qmat <- cbind(
    q1 = c(.011, .012, .013, .014, .015, .016),
    q2 = c(.100, .100, .100, .100, .100, .100)
  )

  tbl <- md_table(x = x, qxj = qmat, radix = 1000)

  expect_equal(round(npxtau_md(tbl, x = 46, n = 3), 4), 0.6979)
  expect_equal(nqxj_md(tbl, x = 47, n = 2, j = 2), (78.94 + 70.02) / 789.43, tolerance = 1e-4)
  expect_equal(round(nqxj_md(tbl, x = 46, n = 2, j = 1), 5), 0.02354)
})

test_that("constant-force formulas match Example 13.4", {
  mu <- c(0.10, 0.20)
  t <- 5

  expect_equal(round(tpx_tau_cf(mu, t), 6), round(exp(-0.30 * t), 6))
  expect_equal(round(tqxprimej_cf(0.10, t), 6), round(1 - exp(-0.10 * t), 6))
  expect_equal(round(tqxj_cf(mu, 1, t), 6), round((1/3) * (1 - exp(-0.30 * t)), 6))
  expect_equal(round(tqxj_cf(mu, 1, 100), 6), round(1/3, 6))
})

test_that("MUDD matches Example 13.5", {
  out <- qxprime_mudd(c(.20, .10))
  expect_equal(round(out[2], 5), 0.11210)
})

test_that("SUDD matches Example 13.6", {
  out <- qxprime_sudd(q1 = .20, q2 = .10)
  expect_equal(unname(out["q2prime"]), 0.11184, tolerance = 1e-4)
})

test_that("constant-force and MUDD inverse from dependent q agree", {
  qdep <- c(.20, .10)
  out_mudd <- qxprime_mudd(qdep)
  expect_equal(length(out_mudd), 2)

  qprime <- c(.20, .10)
  qdep_cf <- qx_dep_cf(qprime)
  expect_true(all(qdep_cf >= 0))
  expect_true(sum(qdep_cf) <= 1)
})


testthat::test_that("multiple-decrement table validates age sequences", {
  qmat <- cbind(
    q1 = c(0.01, 0.02, 0.03),
    q2 = c(0.05, 0.05, 0.05)
  )

  testthat::expect_error(
    md_table(c(40, 42, 43), qmat),
    "consecutive"
  )

  testthat::expect_error(
    md_table(c(40, 40, 41), qmat),
    "duplicate"
  )

  testthat::expect_error(
    md_table(c(40, 41.5, 42), qmat),
    "integer-like"
  )
})


testthat::test_that("multiple-decrement table retains expected structure", {
  tbl <- md_table(
    x = 40:42,
    qxj = cbind(
      withdrawal = c(0.01, 0.02, 0.03),
      retirement = c(0.10, 0.10, 0.10)
    ),
    radix = 1000
  )

  testthat::expect_s3_class(tbl, "md_table")
  testthat::expect_s3_class(tbl, "data.frame")

  testthat::expect_true(all(
    c(
      "x", "q1", "q2", "qtau", "ptau",
      "ltau", "d1", "d2", "dtau"
    ) %in% names(tbl)
  ))

  testthat::expect_equal(
    attr(tbl, "cause_names"),
    c("withdrawal", "retirement")
  )
})


testthat::test_that("table probabilities satisfy total-decrement identities", {
  tbl <- md_table(
    x = 40:45,
    qxj = cbind(
      q1 = c(0.01, 0.02, 0.03, 0.04, 0.05, 0.06),
      q2 = rep(0.10, 6)
    ),
    radix = 1000
  )

  total <- nqxtau_md(tbl, x = 40, n = 4)
  cause_1 <- nqxj_md(tbl, x = 40, n = 4, j = 1)
  cause_2 <- nqxj_md(tbl, x = 40, n = 4, j = 2)

  testthat::expect_equal(
    total,
    cause_1 + cause_2,
    tolerance = 1e-12
  )

  testthat::expect_equal(
    total,
    1 - npxtau_md(tbl, x = 40, n = 4),
    tolerance = 1e-12
  )
})


testthat::test_that("table functions handle zero terms", {
  tbl <- md_table(
    x = 40:42,
    qxj = cbind(
      q1 = c(0.01, 0.02, 0.03),
      q2 = c(0.10, 0.10, 0.10)
    )
  )

  testthat::expect_equal(
    npxtau_md(tbl, x = 42, n = 0),
    1
  )

  testthat::expect_equal(
    nqxtau_md(tbl, x = 42, n = 0),
    0
  )

  testthat::expect_equal(
    nqxj_md(tbl, x = 42, n = 0, j = 1),
    0
  )
})


testthat::test_that("table functions vectorize over x, n, and cause", {
  tbl <- md_table(
    x = 40:45,
    qxj = cbind(
      q1 = c(0.01, 0.02, 0.03, 0.04, 0.05, 0.06),
      q2 = rep(0.10, 6)
    )
  )

  survival <- npxtau_md(
    tbl,
    x = c(40, 42),
    n = c(2, 3)
  )

  causes <- nqxj_md(
    tbl,
    x = c(40, 42),
    n = c(2, 3),
    j = c(1, 2)
  )

  testthat::expect_length(survival, 2)
  testthat::expect_length(causes, 2)
  testthat::expect_true(all(survival >= 0 & survival <= 1))
  testthat::expect_true(all(causes >= 0 & causes <= 1))
})


testthat::test_that("table functions reject non-integer terms", {
  tbl <- md_table(
    x = 40:42,
    qxj = cbind(
      q1 = c(0.01, 0.02, 0.03),
      q2 = c(0.10, 0.10, 0.10)
    )
  )

  testthat::expect_error(
    npxtau_md(tbl, x = 40, n = 1.5),
    "integer-like"
  )

  testthat::expect_error(
    nqxj_md(tbl, x = 40, n = 1.5, j = 1),
    "integer-like"
  )
})


testthat::test_that("constant-force functions satisfy probability identities", {
  mu <- c(0.10, 0.20)
  t <- c(1, 5, 10)

  total_failure <- 1 - tpx_tau_cf(mu, t)

  cause_sum <- tqxj_cf(mu, j = 1, t = t) +
    tqxj_cf(mu, j = 2, t = t)

  testthat::expect_equal(
    cause_sum,
    total_failure,
    tolerance = 1e-12
  )
})


testthat::test_that("constant-force single-decrement function vectorizes", {
  out <- tpxprimej_cf(
    mu = c(0.10, 0.20),
    t = 5
  )

  testthat::expect_length(out, 2)

  testthat::expect_equal(
    out,
    exp(-c(0.10, 0.20) * 5)
  )
})


testthat::test_that("MUDD handles the zero-decrement boundary", {
  testthat::expect_equal(
    qxprime_mudd(c(0, 0)),
    c(0, 0)
  )

  testthat::expect_equal(
    tqxprime_mudd(c(0, 0), t = 0.5),
    c(0, 0)
  )
})


testthat::test_that("constant-force conversion rejects probability one", {
  testthat::expect_error(
    qx_dep_cf(c(1, 0.10)),
    "less than 1"
  )
})


testthat::test_that("SUDD forward and inverse conversions agree", {
  independent <- c(q1prime = 0.20, q2prime = 0.10)

  dependent <- qx_dep_sudd(
    q1prime = independent["q1prime"],
    q2prime = independent["q2prime"]
  )

  recovered <- qxprime_sudd(
    q1 = dependent["q1"],
    q2 = dependent["q2"]
  )

  testthat::expect_equal(
    unname(recovered),
    unname(independent),
    tolerance = 1e-10
  )
})


testthat::test_that("SUDD functions vectorize pairwise", {
  dependent <- qx_dep_sudd(
    q1prime = c(0.20, 0.30),
    q2prime = c(0.10, 0.15)
  )

  testthat::expect_true(is.matrix(dependent))
  testthat::expect_equal(dim(dependent), c(2, 2))
  testthat::expect_equal(
    colnames(dependent),
    c("q1", "q2")
  )

  recovered <- qxprime_sudd(
    q1 = dependent[, "q1"],
    q2 = dependent[, "q2"]
  )

  testthat::expect_true(is.matrix(recovered))

  testthat::expect_equal(
    recovered[, "q1prime"],
    c(0.20, 0.30),
    tolerance = 1e-10
  )

  testthat::expect_equal(
    recovered[, "q2prime"],
    c(0.10, 0.15),
    tolerance = 1e-10
  )
})

# life_table_select.R
#
# Utilities for select and ultimate life tables.

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.validate_select_life_table <- function(tbl) {
  required_cols <- c("x_sel", "duration", "attained_age", "lx")

  if (!all(required_cols %in% names(tbl))) {
    stop(
      "select_life_table object must contain columns: x_sel, duration, attained_age, lx.",
      call. = FALSE
    )
  }

  if (any(!is.finite(tbl$x_sel)) ||
      any(!is.finite(tbl$duration)) ||
      any(!is.finite(tbl$attained_age)) ||
      any(!is.finite(tbl$lx))) {
    stop("All select table columns must be finite.", call. = FALSE)
  }

  if (any(tbl$duration < 0)) {
    stop("duration must be nonnegative.", call. = FALSE)
  }

  if (any(tbl$lx < 0)) {
    stop("lx values must be nonnegative.", call. = FALSE)
  }

  invisible(tbl)
}

.recycle_select_vectors <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) {
    stop("All arguments must have positive length.", call. = FALSE)
  }

  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)

  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) {
    if (length(z) == 1L) rep(z, common_len) else z
  })
}

.get_select_row <- function(tbl, x_sel, duration) {
  idx <- which(tbl$x_sel == x_sel & tbl$duration == duration)
  if (length(idx) == 0L) return(NA_integer_)
  idx[1L]
}

.get_select_lx <- function(tbl, x_sel, duration) {
  idx <- .get_select_row(tbl, x_sel, duration)
  if (is.na(idx)) return(NA_real_)
  tbl$lx[idx]
}

# -------------------------------------------------------------------------
# Constructor
# -------------------------------------------------------------------------

#' Construct a select life table
#'
#' Builds a select-life-table object from vectors of selection age,
#' duration since selection, attained age, and survivor values.
#'
#' @param x_sel Numeric vector of ages at selection.
#' @param duration Numeric vector of durations since selection.
#' @param attained_age Numeric vector of attained ages.
#' @param lx Numeric vector of select-table survivor values.
#'
#' @return A data frame with class \code{"select_life_table"}.
#' @export
select_life_table <- function(x_sel, duration, attained_age, lx) {
  x_sel <- as.numeric(x_sel)
  duration <- as.numeric(duration)
  attained_age <- as.numeric(attained_age)
  lx <- as.numeric(lx)

  n <- length(x_sel)

  if (!all(length(duration) == n,
           length(attained_age) == n,
           length(lx) == n)) {
    stop("x_sel, duration, attained_age, and lx must have the same length.",
         call. = FALSE)
  }

  if (n == 0L) {
    stop("x_sel, duration, attained_age, and lx must have positive length.",
         call. = FALSE)
  }

  if (any(!is.finite(x_sel)) ||
      any(!is.finite(duration)) ||
      any(!is.finite(attained_age)) ||
      any(!is.finite(lx))) {
    stop("x_sel, duration, attained_age, and lx must contain finite numeric values.",
         call. = FALSE)
  }

  if (any(duration < 0)) {
    stop("duration must be nonnegative.", call. = FALSE)
  }

  if (any(attained_age != x_sel + duration)) {
    stop("attained_age must equal x_sel + duration row by row.",
         call. = FALSE)
  }

  if (any(lx < 0)) {
    stop("lx values must be nonnegative.", call. = FALSE)
  }

  out <- data.frame(
    x_sel = x_sel,
    duration = duration,
    attained_age = attained_age,
    lx = lx,
    row.names = NULL
  )

  out <- out[order(out$x_sel, out$duration), , drop = FALSE]

  class(out) <- c("select_life_table", class(out))

  .validate_select_life_table(out)

  out
}

# -------------------------------------------------------------------------
# Core accessor
# -------------------------------------------------------------------------

#' Extract select-table survivor value
#'
#' Returns \eqn{l_{[x]+t}} from a select life table.
#'
#' @param tbl A \code{select_life_table} object.
#' @param x_sel Numeric vector of ages at selection.
#' @param t Numeric vector of durations since selection.
#'
#' @return Numeric vector of survivor values.
#' @export
lx_select <- function(tbl, x_sel, t) {
  .validate_select_life_table(tbl)

  args <- .recycle_select_vectors(x_sel, t)
  x_sel <- args[[1L]]
  t <- args[[2L]]

  if (any(t < 0)) {
    stop("t must be nonnegative.", call. = FALSE)
  }

  out <- numeric(length(x_sel))

  for (j in seq_along(x_sel)) {
    out[j] <- .get_select_lx(tbl, x_sel[j], t[j])
  }

  out
}

# -------------------------------------------------------------------------
# Select probabilities
# -------------------------------------------------------------------------

#' Select-life survival probability
#'
#' Computes
#' \eqn{{}_n p_{[x]+t} = l_{[x]+t+n} / l_{[x]+t}}
#' in the discrete select-table setting.
#'
#' @param tbl A \code{select_life_table} object.
#' @param x_sel Numeric vector of ages at selection.
#' @param t Numeric vector of current durations since selection.
#' @param n Numeric vector of nonnegative integer future durations.
#'
#' @return Numeric vector of survival probabilities.
#' @export
npx_select <- function(tbl, x_sel, t, n) {
  .validate_select_life_table(tbl)

  args <- .recycle_select_vectors(x_sel, t, n)
  x_sel <- args[[1L]]
  t <- args[[2L]]
  n <- args[[3L]]

  if (any(t < 0)) {
    stop("t must be nonnegative.", call. = FALSE)
  }

  if (any(n < 0 | n != floor(n))) {
    stop("n must be a nonnegative integer.", call. = FALSE)
  }

  lx_now <- lx_select(tbl, x_sel, t)
  lx_future <- lx_select(tbl, x_sel, t + n)

  out <- lx_future / lx_now
  out[is.na(lx_now) | is.na(lx_future) | lx_now <= 0] <- NA_real_

  out
}

#' Select-life death probability
#'
#' Computes
#' \eqn{{}_n q_{[x]+t} = 1 - {}_n p_{[x]+t}}.
#'
#' @inheritParams npx_select
#'
#' @return Numeric vector of death probabilities.
#' @export
nqx_select <- function(tbl, x_sel, t, n) {
  p <- npx_select(tbl, x_sel, t, n)
  out <- 1 - p
  out[is.na(p)] <- NA_real_
  out
}

#' Deferred select-life death probability
#'
#' Computes
#' \eqn{{}_{n|m} q_{[x]+t} = {}_n p_{[x]+t} \cdot {}_m q_{[x]+t+n}}.
#'
#' @param tbl A \code{select_life_table} object.
#' @param x_sel Numeric vector of ages at selection.
#' @param t Numeric vector of current durations since selection.
#' @param n Numeric vector of nonnegative integer deferred periods.
#' @param m Numeric vector of nonnegative integer death windows.
#'
#' @return Numeric vector of deferred death probabilities.
#' @export
nmxq_select <- function(tbl, x_sel, t, n, m) {
  .validate_select_life_table(tbl)

  args <- .recycle_select_vectors(x_sel, t, n, m)
  x_sel <- args[[1L]]
  t <- args[[2L]]
  n <- args[[3L]]
  m <- args[[4L]]

  if (any(t < 0)) {
    stop("t must be nonnegative.", call. = FALSE)
  }

  if (any(n < 0 | n != floor(n))) {
    stop("n must be a nonnegative integer.", call. = FALSE)
  }

  if (any(m < 0 | m != floor(m))) {
    stop("m must be a nonnegative integer.", call. = FALSE)
  }

  npx_select(tbl, x_sel, t, n) *
    nqx_select(tbl, x_sel, t + n, m)
}

# -------------------------------------------------------------------------
# Print method
# -------------------------------------------------------------------------

#' @export
print.select_life_table <- function(x, ...) {
  cat("<select_life_table>\n")
  print.data.frame(as.data.frame(x), row.names = FALSE, ...)
  invisible(x)
}

# reserve_extended.R
#
# Retrospective reserves, deferred-insurance reserves, annuity reserves,
# continuous-style gain and loss calculations, and numerical helpers for
# Thiele's differential equation.

# -------------------------------------------------------------------------
# Internal validation and recycling helpers
# -------------------------------------------------------------------------

#' @noRd
.reserve_ext_check_finite <- function(x, name) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  if (any(!is.finite(x))) {
    stop(name, " must contain finite numeric values.", call. = FALSE)
  }

  x
}

#' @noRd
.reserve_ext_check_nonnegative <- function(x, name) {
  x <- .reserve_ext_check_finite(x, name)

  if (any(x < 0)) {
    stop(name, " must contain nonnegative finite values.", call. = FALSE)
  }

  x
}

#' @noRd
.reserve_ext_check_positive <- function(x, name) {
  x <- .reserve_ext_check_finite(x, name)

  if (any(x <= 0)) {
    stop(name, " must contain positive finite values.", call. = FALSE)
  }

  x
}

#' @noRd
.reserve_ext_check_integerish <- function(x, name, positive = FALSE) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }

  invalid_numeric <- any(!is.finite(x))
  invalid_sign <- if (positive) any(x <= 0) else any(x < 0)
  invalid_integer <- any(abs(x - round(x)) > 1e-10)

  if (invalid_numeric || invalid_sign || invalid_integer) {
    descriptor <- if (positive) "positive" else "nonnegative"

    stop(
      name,
      " must contain ",
      descriptor,
      " integer-like values.",
      call. = FALSE
    )
  }

  as.integer(round(x))
}

#' @noRd
.reserve_ext_check_probability <- function(x, name) {
  x <- .reserve_ext_check_finite(x, name)

  if (any(x < 0 | x > 1)) {
    stop(name, " must contain values in [0, 1].", call. = FALSE)
  }

  x
}

#' @noRd
.reserve_ext_check_interest <- function(i, name = "i") {
  i <- .reserve_ext_check_finite(i, name)

  if (any(i <= -1)) {
    stop(name, " must contain values greater than -1.", call. = FALSE)
  }

  i
}

#' @noRd
.reserve_ext_check_basis <- function(tbl, model) {
  if (is.null(tbl) && is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }

  if (!is.null(tbl) && !is.null(model)) {
    stop("Supply only one of tbl or model.", call. = FALSE)
  }

  invisible(TRUE)
}

#' @noRd
.reserve_ext_recycle <- function(..., .names = NULL) {
  values <- list(...)
  lengths <- vapply(values, length, integer(1))

  if (length(values) == 0L || any(lengths == 0L)) {
    stop("Inputs must have positive length.", call. = FALSE)
  }

  common_length <- max(lengths)

  if (is.null(.names)) {
    .names <- rep("Input", length(values))
  }

  if (length(.names) != length(values)) {
    stop("Internal error: incorrect number of argument names.", call. = FALSE)
  }

  for (j in seq_along(values)) {
    if (!lengths[j] %in% c(1L, common_length)) {
      stop(
        .names[j],
        " must have length 1 or the common length ",
        common_length,
        ".",
        call. = FALSE
      )
    }
  }

  lapply(values, rep, length.out = common_length)
}

#' @noRd
.reserve_ext_recycle_to <- function(x, size, name) {
  if (!length(x) %in% c(1L, size)) {
    stop(
      name,
      " must have length 1 or ",
      size,
      ".",
      call. = FALSE
    )
  }

  rep(x, length.out = size)
}

#' @noRd
.reserve_ext_check_t_le_n <- function(t, n, allow_equal = TRUE) {
  invalid <- if (allow_equal) t > n else t >= n

  if (any(invalid)) {
    relation <- if (allow_equal) "t <= n" else "t < n"
    stop("t must satisfy ", relation, ".", call. = FALSE)
  }

  invisible(TRUE)
}

#' @noRd
.reserve_ext_check_positive_denominator <- function(x, name) {
  if (any(!is.finite(x)) || any(x <= 0)) {
    stop(name, " must be positive and finite.", call. = FALSE)
  }

  x
}


# Internal compatibility helpers
#
# These wrappers preserve internal helper names used by reserve and
# gain/loss functions in existing package source files. They are not
# exported and do not create documentation pages.

#' @noRd
.check_nonneg_numeric <- function(x, name) {
  .reserve_ext_check_nonnegative(x, name)
}

#' @noRd
.check_nonneg_integerish <- function(x, name) {
  .reserve_ext_check_integerish(
    x = x,
    name = name,
    positive = FALSE
  )
}

#' @noRd
.check_prob_ch10 <- function(x, name) {
  .reserve_ext_check_probability(x, name)
}

#' @noRd
.check_interest_ch10 <- function(i, name = "i") {
  .reserve_ext_check_interest(i, name)
}

#' @noRd
.common_length_ch10 <- function(...) {
  lengths <- vapply(list(...), length, integer(1))

  if (length(lengths) == 0L) {
    return(0L)
  }

  max(lengths)
}

#' @noRd
.recycle_common_ch10 <- function(...) {
  values <- list(...)
  argument_names <- names(values)

  if (is.null(argument_names) ||
      any(!nzchar(argument_names))) {
    argument_names <- rep("Input", length(values))
  }

  .reserve_ext_recycle(
    ...,
    .names = argument_names
  )
}

#' @noRd
.recycle2_ch10 <- function(a,
                           b,
                           a_name = "a",
                           b_name = "b") {
  values <- .reserve_ext_recycle(
    a,
    b,
    .names = c(a_name, b_name)
  )

  list(
    a = values[[1]],
    b = values[[2]]
  )
}

#' @noRd
.recycle3_ch10 <- function(a,
                           b,
                           c,
                           a_name = "a",
                           b_name = "b",
                           c_name = "c") {
  values <- .reserve_ext_recycle(
    a,
    b,
    c,
    .names = c(a_name, b_name, c_name)
  )

  list(
    a = values[[1]],
    b = values[[2]],
    c = values[[3]]
  )
}

#' @noRd
.check_t_le_n <- function(t, n, allow_equal = TRUE) {
  .reserve_ext_check_t_le_n(
    t = t,
    n = n,
    allow_equal = allow_equal
  )
}


# -------------------------------------------------------------------------
# Retrospective reserves
# -------------------------------------------------------------------------

#' Retrospective whole life reserve
#'
#' Computes the retrospective net level premium reserve
#' \deqn{{}_tV_x =
#' P_x\ddot{s}_{x:\overline{t}|}
#' -
#' \frac{A_{x:\overline{t}|}^{1}}{{}_tE_x}.}
#'
#' The mortality basis may be supplied through either a life table or a
#' parametric survival model.
#'
#' @param x Issue age. May be scalar or vector.
#' @param t Nonnegative integer duration. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the actuarial functions.
#' @param tbl Optional life table object. Supply by name.
#'
#' @return Numeric vector of retrospective reserve values.
#' @examples
#' tVx_ret(
#'   40,
#'   t = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#' @export
tVx_ret <- function(x, t, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  t <- .reserve_ext_check_integerish(t, "t")
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    t,
    i,
    .names = c("x", "t", "i")
  )

  x <- values[[1]]
  t <- values[[2]]
  i <- values[[3]]

  vapply(seq_along(x), function(j) {
    premium <- Px(
      x = x[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    accumulated_premiums <- sdotxn(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    accumulated_claims <- Axn1(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    ) / nEx(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    premium * accumulated_premiums - accumulated_claims
  }, numeric(1))
}

#' Retrospective endowment insurance reserve
#'
#' Computes the retrospective reserve for an endowment insurance at a
#' duration satisfying \eqn{0 \le t \le n}. At maturity, the reserve is 1.
#'
#' @inheritParams tVx_ret
#' @param n Nonnegative integer contract term. May be scalar or vector.
#'
#' @return Numeric vector of retrospective reserve values.
#' @examples
#' tVxn_ret(
#'   40,
#'   n = 20,
#'   t = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#' @export
tVxn_ret <- function(x, n, t, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  n <- .reserve_ext_check_integerish(n, "n")
  t <- .reserve_ext_check_integerish(t, "t")
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    n,
    t,
    i,
    .names = c("x", "n", "t", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  t <- values[[3]]
  i <- values[[4]]

  .reserve_ext_check_t_le_n(t, n)

  vapply(seq_along(x), function(j) {
    if (t[j] == n[j]) {
      return(1)
    }

    premium <- Pxn(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    accumulated_premiums <- sdotxn(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    accumulated_claims <- Axn1(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    ) / nEx(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    premium * accumulated_premiums - accumulated_claims
  }, numeric(1))
}

#' Retrospective term insurance reserve
#'
#' Computes the retrospective reserve for a term insurance at a duration
#' satisfying \eqn{0 \le t \le n}. At expiry, the reserve is 0.
#'
#' @inheritParams tVxn_ret
#'
#' @return Numeric vector of retrospective reserve values.
#' @examples
#' tVxn1_ret(
#'   40,
#'   n = 20,
#'   t = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#' @export
tVxn1_ret <- function(x, n, t, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  n <- .reserve_ext_check_integerish(n, "n")
  t <- .reserve_ext_check_integerish(t, "t")
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    n,
    t,
    i,
    .names = c("x", "n", "t", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  t <- values[[3]]
  i <- values[[4]]

  .reserve_ext_check_t_le_n(t, n)

  vapply(seq_along(x), function(j) {
    if (t[j] == n[j]) {
      return(0)
    }

    premium <- Pxn1(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    accumulated_premiums <- sdotxn(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    accumulated_claims <- Axn1(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    ) / nEx(
      x = x[j],
      n = t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    premium * accumulated_premiums - accumulated_claims
  }, numeric(1))
}


# -------------------------------------------------------------------------
# Deferred insurance reserves
# -------------------------------------------------------------------------

#' Deferred insurance reserves
#'
#' Computes prospective reserves for deferred whole life insurance contracts.
#'
#' \code{tVnAx()} computes the reserve at duration \code{t} for an
#' \code{n}-year deferred whole life insurance funded by level annual
#' premiums during the deferral period.
#'
#' \code{htVnAx()} computes the corresponding reserve when premiums are
#' limited to the first \code{h} years, where \code{h <= n}.
#'
#' @param x Issue age.
#' @param n Deferral period in years.
#' @param t Duration in years.
#' @param i Effective annual interest rate.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model name.
#' @param ... Additional arguments passed to the underlying actuarial
#'   functions.
#'
#' @return A numeric vector of prospective reserve values.
#'
#' @examples
#' tVnAx(
#'   x = 40, n = 20, t = 10, i = 0.05,
#'   model = "uniform", omega = 100
#' )
#'
#' htVnAx(
#'   x = 40, n = 20, h = 10, t = 5, i = 0.05,
#'   model = "uniform", omega = 100
#' )
#'
#' @name deferred_insurance_reserves
#' @aliases tVnAx htVnAx
#' @export
tVnAx <- function(x, n, t, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  n <- .reserve_ext_check_integerish(n, "n")
  t <- .reserve_ext_check_integerish(t, "t")
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    n,
    t,
    i,
    .names = c("x", "n", "t", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  t <- values[[3]]
  i <- values[[4]]

  vapply(seq_along(x), function(j) {
    if (t[j] < n[j]) {
      future_benefit <- nAx(
        x = x[j] + t[j],
        n = n[j] - t[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      )

      future_premiums <- PnAx(
        x = x[j],
        n = n[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      ) * adotxn(
        x = x[j] + t[j],
        n = n[j] - t[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      )

      return(future_benefit - future_premiums)
    }

    Ax(
      x = x[j] + t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )
  }, numeric(1))
}


#' @rdname deferred_insurance_reserves
#' @param h Premium-paying period in years.
#' @export
htVnAx <- function(x, n, h, t, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  n <- .reserve_ext_check_integerish(n, "n")
  h <- .reserve_ext_check_integerish(h, "h", positive = TRUE)
  t <- .reserve_ext_check_integerish(t, "t")
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    n,
    h,
    t,
    i,
    .names = c("x", "n", "h", "t", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  h <- values[[3]]
  t <- values[[4]]
  i <- values[[5]]

  if (any(h > n)) {
    stop("h must satisfy h <= n.", call. = FALSE)
  }

  vapply(seq_along(x), function(j) {
    if (t[j] < h[j]) {
      future_benefit <- nAx(
        x = x[j] + t[j],
        n = n[j] - t[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      )

      future_premiums <- tPnAx(
        x = x[j],
        n = n[j],
        t = h[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      ) * adotxn(
        x = x[j] + t[j],
        n = h[j] - t[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      )

      return(future_benefit - future_premiums)
    }

    if (t[j] < n[j]) {
      return(nAx(
        x = x[j] + t[j],
        n = n[j] - t[j],
        i = i[j],
        tbl = tbl,
        model = model,
        ...
      ))
    }

    Ax(
      x = x[j] + t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )
  }, numeric(1))
}

# -------------------------------------------------------------------------
# Annuity reserves
# -------------------------------------------------------------------------

#' Net premium for a deferred annuity-due
#'
#' Computes
#' \deqn{
#' P({}_{n|}\ddot{a}_x)
#' =
#' \frac{{}_{n|}\ddot{a}_x}
#' {\ddot{a}_{x:\overline{n}|}}.
#' }
#'
#' @param x Issue age. May be scalar or vector.
#' @param n Positive integer deferral period. May be scalar or vector.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param model Optional parametric survival model.
#' @param ... Additional parameters passed to the actuarial functions.
#' @param tbl Optional life table object. Supply by name.
#'
#' @return Numeric vector of net annual premiums.
#' @examples
#' PnAdotx(
#'   40,
#'   n = 20,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#' @export
PnAdotx <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  n <- .reserve_ext_check_integerish(n, "n", positive = TRUE)
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    n,
    i,
    .names = c("x", "n", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  i <- values[[3]]

  vapply(seq_along(x), function(j) {
    denominator <- adotxn(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    .reserve_ext_check_positive_denominator(
      denominator,
      "The premium annuity denominator"
    )

    nadotx(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    ) / denominator
  }, numeric(1))
}

#' Reserve for a deferred annuity-due
#'
#' Computes the reserve during the deferral period, where \eqn{0 \le t < n}.
#'
#' @inheritParams PnAdotx
#' @param t Nonnegative integer duration. May be scalar or vector.
#'
#' @return Numeric vector of reserve values.
#' @examples
#' tVnAdotx(
#'   40,
#'   n = 20,
#'   t = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#' @export
tVnAdotx <- function(x, n, t, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  n <- .reserve_ext_check_integerish(n, "n", positive = TRUE)
  t <- .reserve_ext_check_integerish(t, "t")
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    n,
    t,
    i,
    .names = c("x", "n", "t", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  t <- values[[3]]
  i <- values[[4]]

  .reserve_ext_check_t_le_n(t, n, allow_equal = FALSE)

  vapply(seq_along(x), function(j) {
    future_benefits <- nadotx(
      x = x[j] + t[j],
      n = n[j] - t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    future_premiums <- PnAdotx(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    ) * adotxn(
      x = x[j] + t[j],
      n = n[j] - t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    future_benefits - future_premiums
  }, numeric(1))
}

#' Net premium for a deferred annuity-immediate
#'
#' Computes
#' \deqn{
#' P({}_{n|}a_x)
#' =
#' \frac{{}_{n|}a_x}
#' {\ddot{a}_{x:\overline{n}|}}.
#' }
#'
#' @inheritParams PnAdotx
#'
#' @return Numeric vector of net annual premiums.
#' @examples
#' Pnax(
#'   40,
#'   n = 20,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#' @export
Pnax <- function(x, n, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  n <- .reserve_ext_check_integerish(n, "n", positive = TRUE)
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    n,
    i,
    .names = c("x", "n", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  i <- values[[3]]

  vapply(seq_along(x), function(j) {
    denominator <- adotxn(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    .reserve_ext_check_positive_denominator(
      denominator,
      "The premium annuity denominator"
    )

    nax(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    ) / denominator
  }, numeric(1))
}

#' Reserve for a deferred annuity-immediate
#'
#' Computes the reserve during the deferral period, where \eqn{0 \le t < n}.
#'
#' @inheritParams tVnAdotx
#'
#' @return Numeric vector of reserve values.
#' @examples
#' tVnax(
#'   40,
#'   n = 20,
#'   t = 10,
#'   i = 0.05,
#'   model = "uniform",
#'   omega = 100
#' )
#' @export
tVnax <- function(x, n, t, i, model = NULL, ..., tbl = NULL) {
  .reserve_ext_check_basis(tbl, model)

  x <- .reserve_ext_check_nonnegative(x, "x")
  n <- .reserve_ext_check_integerish(n, "n", positive = TRUE)
  t <- .reserve_ext_check_integerish(t, "t")
  i <- .reserve_ext_check_interest(i)

  values <- .reserve_ext_recycle(
    x,
    n,
    t,
    i,
    .names = c("x", "n", "t", "i")
  )

  x <- values[[1]]
  n <- values[[2]]
  t <- values[[3]]
  i <- values[[4]]

  .reserve_ext_check_t_le_n(t, n, allow_equal = FALSE)

  vapply(seq_along(x), function(j) {
    future_benefits <- nax(
      x = x[j] + t[j],
      n = n[j] - t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    future_premiums <- Pnax(
      x = x[j],
      n = n[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    ) * adotxn(
      x = x[j] + t[j],
      n = n[j] - t[j],
      i = i[j],
      tbl = tbl,
      model = model,
      ...
    )

    future_benefits - future_premiums
  }, numeric(1))
}

# -------------------------------------------------------------------------
# Continuous-style gain and loss functions
# -------------------------------------------------------------------------

#' Total gain for a continuous-style one-step recursion
#'
#' Computes the amount accumulated during a step, less the expected reserve
#' required at the end of the step.
#'
#' Reserves, premium rates, benefits, and forces of interest may be negative
#' when such values are meaningful for the application. The step length must
#' be positive, and the survival probability must lie in \eqn{[0,1]}.
#'
#' @param Vt Reserve at time `t`.
#' @param Vt1 Reserve at time `t + h`.
#' @param P Premium rate.
#' @param delta_actual Actual force of interest.
#' @param p_actual Actual survival probability over the step.
#' @param benefit Benefit paid at the start of the step.
#' @param h Positive step length.
#'
#' @return Numeric vector of gain values.
#' @examples
#' GT_cont(
#'   Vt = 10,
#'   Vt1 = 11,
#'   P = 1,
#'   delta_actual = 0.05,
#'   p_actual = 0.99
#' )
#' @export
GT_cont <- function(Vt,
                    Vt1,
                    P,
                    delta_actual,
                    p_actual,
                    benefit = 0,
                    h = 1) {
  Vt <- .reserve_ext_check_finite(Vt, "Vt")
  Vt1 <- .reserve_ext_check_finite(Vt1, "Vt1")
  P <- .reserve_ext_check_finite(P, "P")
  delta_actual <- .reserve_ext_check_finite(
    delta_actual,
    "delta_actual"
  )
  p_actual <- .reserve_ext_check_probability(p_actual, "p_actual")
  benefit <- .reserve_ext_check_finite(benefit, "benefit")
  h <- .reserve_ext_check_positive(h, "h")

  values <- .reserve_ext_recycle(
    Vt,
    Vt1,
    P,
    delta_actual,
    p_actual,
    benefit,
    h,
    .names = c(
      "Vt",
      "Vt1",
      "P",
      "delta_actual",
      "p_actual",
      "benefit",
      "h"
    )
  )

  Vt <- values[[1]]
  Vt1 <- values[[2]]
  P <- values[[3]]
  delta_actual <- values[[4]]
  p_actual <- values[[5]]
  benefit <- values[[6]]
  h <- values[[7]]

  (Vt - benefit + P * h) * exp(delta_actual * h) -
    p_actual * Vt1
}

#' Mortality gain for a continuous-style recursion
#'
#' Evaluates the one-step gain using the assumed force of interest and the
#' actual survival probability.
#'
#' @inheritParams GT_cont
#' @param delta_assumed Assumed force of interest.
#'
#' @return Numeric vector of mortality gain values.
#' @examples
#' GM_cont(
#'   Vt = 10,
#'   Vt1 = 11,
#'   P = 1,
#'   delta_assumed = 0.05,
#'   p_actual = 0.99
#' )
#' @export
GM_cont <- function(Vt,
                    Vt1,
                    P,
                    delta_assumed,
                    p_actual,
                    benefit = 0,
                    h = 1) {
  GT_cont(
    Vt = Vt,
    Vt1 = Vt1,
    P = P,
    delta_actual = delta_assumed,
    p_actual = p_actual,
    benefit = benefit,
    h = h
  )
}

#' Interest gain for a continuous-style recursion
#'
#' Evaluates the one-step gain using the actual force of interest and the
#' assumed survival probability.
#'
#' @inheritParams GT_cont
#' @param p_assumed Assumed survival probability over the step.
#'
#' @return Numeric vector of interest gain values.
#' @examples
#' GI_cont(
#'   Vt = 10,
#'   Vt1 = 11,
#'   P = 1,
#'   delta_actual = 0.05,
#'   p_assumed = 0.99
#' )
#' @export
GI_cont <- function(Vt,
                    Vt1,
                    P,
                    delta_actual,
                    p_assumed,
                    benefit = 0,
                    h = 1) {
  GT_cont(
    Vt = Vt,
    Vt1 = Vt1,
    P = P,
    delta_actual = delta_actual,
    p_actual = p_assumed,
    benefit = benefit,
    h = h
  )
}

# -------------------------------------------------------------------------
# Thiele equation helpers
# -------------------------------------------------------------------------

#' One backward numerical step for Thiele's equation
#'
#' Approximates the reserve at time `t` from a known reserve at time
#' `t + h`.
#'
#' The implemented step is
#' \deqn{
#' V_t =
#' \frac{
#' V_{t+h} - hP + h\mu B
#' }{
#' 1 + h(\delta + \mu)
#' }.
#' }
#'
#' @param V_next Reserve at time `t + h`.
#' @param P Premium rate.
#' @param delta Force of interest.
#' @param mu Nonnegative force of mortality at time `t`.
#' @param benefit Benefit amount.
#' @param h Positive step size.
#'
#' @return Numeric vector of reserve approximations.
#' @examples
#' thiele_backward_step(
#'   V_next = 1000,
#'   P = 26.96,
#'   delta = 0.058,
#'   mu = 0.002,
#'   benefit = 1000,
#'   h = 1
#' )
#' @export
thiele_backward_step <- function(V_next,
                                 P,
                                 delta,
                                 mu,
                                 benefit = 1,
                                 h = 1) {
  V_next <- .reserve_ext_check_finite(V_next, "V_next")
  P <- .reserve_ext_check_finite(P, "P")
  delta <- .reserve_ext_check_finite(delta, "delta")
  mu <- .reserve_ext_check_nonnegative(mu, "mu")
  benefit <- .reserve_ext_check_finite(benefit, "benefit")
  h <- .reserve_ext_check_positive(h, "h")

  values <- .reserve_ext_recycle(
    V_next,
    P,
    delta,
    mu,
    benefit,
    h,
    .names = c(
      "V_next",
      "P",
      "delta",
      "mu",
      "benefit",
      "h"
    )
  )

  V_next <- values[[1]]
  P <- values[[2]]
  delta <- values[[3]]
  mu <- values[[4]]
  benefit <- values[[5]]
  h <- values[[6]]

  denominator <- 1 + h * (delta + mu)

  if (any(abs(denominator) < .Machine$double.eps^0.5)) {
    stop(
      "The backward-step denominator is zero or numerically indistinguishable from zero.",
      call. = FALSE
    )
  }

  (V_next - h * P + h * mu * benefit) / denominator
}

#' Reserve derivative from Thiele's equation
#'
#' Computes
#' \deqn{
#' \frac{dV}{dt}
#' =
#' P + \delta V - \mu(B - V).
#' }
#'
#' @param V Reserve at time `t`.
#' @param P Premium rate.
#' @param delta Force of interest.
#' @param mu Nonnegative force of mortality.
#' @param benefit Benefit amount.
#'
#' @return Numeric vector of reserve derivatives.
#' @examples
#' thiele_dVdt(
#'   V = 900,
#'   P = 25,
#'   delta = 0.05,
#'   mu = 0.002,
#'   benefit = 1000
#' )
#' @export
thiele_dVdt <- function(V, P, delta, mu, benefit = 1) {
  V <- .reserve_ext_check_finite(V, "V")
  P <- .reserve_ext_check_finite(P, "P")
  delta <- .reserve_ext_check_finite(delta, "delta")
  mu <- .reserve_ext_check_nonnegative(mu, "mu")
  benefit <- .reserve_ext_check_finite(benefit, "benefit")

  values <- .reserve_ext_recycle(
    V,
    P,
    delta,
    mu,
    benefit,
    .names = c("V", "P", "delta", "mu", "benefit")
  )

  V <- values[[1]]
  P <- values[[2]]
  delta <- values[[3]]
  mu <- values[[4]]
  benefit <- values[[5]]

  P + delta * V - mu * (benefit - V)
}

#' Backward reserve path from a terminal value
#'
#' Starting from a terminal reserve at the final time, computes reserves
#' backward over a strictly increasing time grid using
#' `thiele_backward_step()`.
#'
#' @param times Finite numeric vector of strictly increasing times.
#' @param V_terminal Finite scalar reserve at the final time.
#' @param P Premium rate, scalar or vector with one value per time step.
#' @param delta Force of interest, scalar or vector with one value per step.
#' @param mu Nonnegative force of mortality, scalar or vector with one value
#'   per step.
#' @param benefit Benefit amount, scalar or vector with one value per step.
#'
#' @return Numeric vector of reserve values corresponding to `times`.
#' @examples
#' times <- seq(19, 20, by = 0.25)
#'
#' thiele_backward_path(
#'   times,
#'   V_terminal = 1000,
#'   P = 26.96,
#'   delta = 0.058,
#'   mu = 0.002,
#'   benefit = 1000
#' )
#' @export
thiele_backward_path <- function(times,
                                 V_terminal,
                                 P,
                                 delta,
                                 mu,
                                 benefit = 1) {
  times <- .reserve_ext_check_finite(times, "times")

  if (length(times) < 2L) {
    stop("times must have length at least 2.", call. = FALSE)
  }

  if (any(diff(times) <= 0)) {
    stop("times must be strictly increasing.", call. = FALSE)
  }

  V_terminal <- .reserve_ext_check_finite(
    V_terminal,
    "V_terminal"
  )

  if (length(V_terminal) != 1L) {
    stop("V_terminal must be a finite scalar.", call. = FALSE)
  }

  P <- .reserve_ext_check_finite(P, "P")
  delta <- .reserve_ext_check_finite(delta, "delta")
  mu <- .reserve_ext_check_nonnegative(mu, "mu")
  benefit <- .reserve_ext_check_finite(benefit, "benefit")

  n_steps <- length(times) - 1L

  P <- .reserve_ext_recycle_to(P, n_steps, "P")
  delta <- .reserve_ext_recycle_to(delta, n_steps, "delta")
  mu <- .reserve_ext_recycle_to(mu, n_steps, "mu")
  benefit <- .reserve_ext_recycle_to(
    benefit,
    n_steps,
    "benefit"
  )

  reserves <- numeric(length(times))
  reserves[length(times)] <- V_terminal

  for (k in seq.int(n_steps, 1L, by = -1L)) {
    h <- times[k + 1L] - times[k]

    reserves[k] <- thiele_backward_step(
      V_next = reserves[k + 1L],
      P = P[k],
      delta = delta[k],
      mu = mu[k],
      benefit = benefit[k],
      h = h
    )
  }

  reserves
}

test_that("whole life retrospective reserve equals prospective reserve", {
  x <- 40
  t <- 10
  i <- 0.05

  expect_equal(
    tVx_ret(x, t = t, i = i, model = "uniform", omega = 100),
    tVx(x, t = t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("whole life retrospective reserve is zero at issue", {
  expect_equal(
    tVx_ret(40, t = 0, i = 0.05, model = "uniform", omega = 100),
    0,
    tolerance = 1e-10
  )
})

test_that("term retrospective reserve equals prospective reserve", {
  x <- 40
  n <- 20
  t <- 10
  i <- 0.05

  expect_equal(
    tVxn1_ret(x, n = n, t = t, i = i, model = "uniform", omega = 100),
    tVxn1(x, n = n, t = t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("term retrospective reserve is zero at expiry", {
  expect_equal(
    tVxn1_ret(40, n = 20, t = 20, i = 0.05, model = "uniform", omega = 100),
    0,
    tolerance = 1e-10
  )
})

test_that("endowment retrospective reserve equals prospective reserve", {
  x <- 40
  n <- 20
  t <- 10
  i <- 0.05

  expect_equal(
    tVxn_ret(x, n = n, t = t, i = i, model = "uniform", omega = 100),
    tVxn(x, n = n, t = t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("endowment retrospective reserve is one at maturity", {
  expect_equal(
    tVxn_ret(40, n = 20, t = 20, i = 0.05, model = "uniform", omega = 100),
    1,
    tolerance = 1e-10
  )
})

test_that("deferred insurance reserve matches prospective formula before deferral ends", {
  x <- 40
  n <- 20
  t <- 10
  i <- 0.05

  direct <- nAx(x = x + t, n = n - t, i = i, model = "uniform", omega = 100) -
    PnAx(x = x, n = n, i = i, model = "uniform", omega = 100) *
    adotxn(x = x + t, n = n - t, i = i, model = "uniform", omega = 100)

  expect_equal(
    tVnAx(x, n = n, t = t, i = i, model = "uniform", omega = 100),
    direct,
    tolerance = 1e-10
  )
})

test_that("deferred insurance reserve becomes whole life reserve after deferral period", {
  x <- 40
  n <- 20
  t <- 25
  i <- 0.05

  expect_equal(
    tVnAx(x, n = n, t = t, i = i, model = "uniform", omega = 100),
    Ax(x = x + t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("h-pay deferred insurance reserve behaves correctly in all regions", {
  x <- 40
  n <- 20
  h <- 10
  i <- 0.05

  t1 <- 5
  val1 <- htVnAx(x, n = n, h = h, t = t1, i = i, model = "uniform", omega = 100)
  rhs1 <- nAx(x = x + t1, n = n - t1, i = i, model = "uniform", omega = 100) -
    tPnAx(x = x, n = n, t = h, i = i, model = "uniform", omega = 100) *
    adotxn(x = x + t1, n = h - t1, i = i, model = "uniform", omega = 100)

  expect_equal(val1, rhs1, tolerance = 1e-10)

  t2 <- 15
  val2 <- htVnAx(x, n = n, h = h, t = t2, i = i, model = "uniform", omega = 100)
  rhs2 <- nAx(x = x + t2, n = n - t2, i = i, model = "uniform", omega = 100)

  expect_equal(val2, rhs2, tolerance = 1e-10)

  t3 <- 25
  val3 <- htVnAx(x, n = n, h = h, t = t3, i = i, model = "uniform", omega = 100)
  rhs3 <- Ax(x = x + t3, i = i, model = "uniform", omega = 100)

  expect_equal(val3, rhs3, tolerance = 1e-10)
})

test_that("deferred annuity-due premium matches definition", {
  x <- 40
  n <- 20
  i <- 0.05

  expect_equal(
    PnAdotx(x, n = n, i = i, model = "uniform", omega = 100),
    nadotx(x, n = n, i = i, model = "uniform", omega = 100) /
      adotxn(x, n = n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("deferred annuity-due reserve matches prospective formula", {
  x <- 40
  n <- 20
  t <- 10
  i <- 0.05

  direct <- nadotx(x = x + t, n = n - t, i = i, model = "uniform", omega = 100) -
    PnAdotx(x = x, n = n, i = i, model = "uniform", omega = 100) *
    adotxn(x = x + t, n = n - t, i = i, model = "uniform", omega = 100)

  expect_equal(
    tVnAdotx(x, n = n, t = t, i = i, model = "uniform", omega = 100),
    direct,
    tolerance = 1e-10
  )
})

test_that("deferred annuity-immediate premium matches definition", {
  x <- 40
  n <- 20
  i <- 0.05

  expect_equal(
    Pnax(x, n = n, i = i, model = "uniform", omega = 100),
    nax(x, n = n, i = i, model = "uniform", omega = 100) /
      adotxn(x, n = n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("deferred annuity-immediate reserve matches prospective formula", {
  x <- 40
  n <- 20
  t <- 10
  i <- 0.05

  direct <- nax(x = x + t, n = n - t, i = i, model = "uniform", omega = 100) -
    Pnax(x = x, n = n, i = i, model = "uniform", omega = 100) *
    adotxn(x = x + t, n = n - t, i = i, model = "uniform", omega = 100)

  expect_equal(
    tVnax(x, n = n, t = t, i = i, model = "uniform", omega = 100),
    direct,
    tolerance = 1e-10
  )
})

test_that("continuous total, mortality, and interest gains coincide when actual equals assumed", {
  Vt <- 10
  Vt1 <- 11
  P <- 1
  delta <- 0.05
  p <- 0.99
  benefit <- 0.5
  h <- 1

  gt <- GT_cont(Vt, Vt1, P, delta_actual = delta, p_actual = p, benefit = benefit, h = h)
  gm <- GM_cont(Vt, Vt1, P, delta_assumed = delta, p_actual = p, benefit = benefit, h = h)
  gi <- GI_cont(Vt, Vt1, P, delta_actual = delta, p_assumed = p, benefit = benefit, h = h)

  expect_equal(gt, gm, tolerance = 1e-12)
  expect_equal(gt, gi, tolerance = 1e-12)
})

test_that("Thiele derivative matches direct formula", {
  expect_equal(
    thiele_dVdt(V = 900, P = 25, delta = 0.05, mu = 0.002, benefit = 1000),
    25 + 0.05 * 900 - 0.002 * (1000 - 900),
    tolerance = 1e-12
  )
})

test_that("backward Thiele step matches explicit formula", {
  V_next <- 1000
  P <- 26.96
  delta <- 0.058
  mu <- 0.002
  benefit <- 1000
  h <- 1

  expected <- (V_next - h * P + h * mu * benefit) / (1 + h * (delta + mu))

  expect_equal(
    thiele_backward_step(
      V_next = V_next,
      P = P,
      delta = delta,
      mu = mu,
      benefit = benefit,
      h = h
    ),
    expected,
    tolerance = 1e-12
  )
})

test_that("backward Thiele path returns terminal value at final time", {
  times <- seq(19, 20, by = 0.25)

  vals <- thiele_backward_path(
    times = times,
    V_terminal = 1000,
    P = 26.96,
    delta = 0.058,
    mu = 0.002,
    benefit = 1000
  )

  expect_length(vals, length(times))
  expect_equal(tail(vals, 1), 1000, tolerance = 1e-12)
})

test_that("backward Thiele path one-step case matches backward step", {
  times <- c(19, 20)

  path_vals <- thiele_backward_path(
    times = times,
    V_terminal = 1000,
    P = 26.96,
    delta = 0.058,
    mu = 0.002,
    benefit = 1000
  )

  one_step <- thiele_backward_step(
    V_next = 1000,
    P = 26.96,
    delta = 0.058,
    mu = 0.002,
    benefit = 1000,
    h = 1
  )

  expect_equal(path_vals[1], one_step, tolerance = 1e-12)
  expect_equal(path_vals[2], 1000, tolerance = 1e-12)
})

test_that("extended reserve functions vectorize cleanly", {
  out1 <- tVx_ret(c(40, 45), t = c(5, 10), i = 0.05, model = "uniform", omega = 100)
  out2 <- tVnAx(c(40, 45), n = c(20, 25), t = c(10, 15), i = 0.05, model = "uniform", omega = 100)
  out3 <- tVnAdotx(c(40, 45), n = c(20, 25), t = c(5, 10), i = 0.05, model = "uniform", omega = 100)
  out4 <- tVnax(c(40, 45), n = c(20, 25), t = c(5, 10), i = 0.05, model = "uniform", omega = 100)

  expect_length(out1, 2)
  expect_length(out2, 2)
  expect_length(out3, 2)
  expect_length(out4, 2)
})

test_that("extended reserve functions reject invalid inputs", {
  expect_error(
    tVnAdotx(40, n = 10, t = 10, i = 0.05, model = "uniform", omega = 100),
    "t must satisfy t < n"
  )

  expect_error(
    tVnax(40, n = 10, t = 12, i = 0.05, model = "uniform", omega = 100),
    "t must satisfy t < n"
  )

  expect_error(
    htVnAx(40, n = 10, h = 12, t = 5, i = 0.05, model = "uniform", omega = 100),
    "h must satisfy h <= n"
  )

  expect_error(
    thiele_backward_path(c(20, 19), V_terminal = 1000, P = 10, delta = 0.05, mu = 0.002),
    "strictly increasing"
  )
})


testthat::test_that("extended reserve functions work with life tables", {
  tbl <- life_table(
    x = 0:5,
    lx = c(100000, 85000, 65000, 40000, 15000, 0)
  )

  whole_life <- tVx_ret(
    x = 0,
    t = 1,
    i = 0.05,
    tbl = tbl
  )

  deferred <- tVnAx(
    x = 0,
    n = 3,
    t = 1,
    i = 0.05,
    tbl = tbl
  )

  testthat::expect_length(whole_life, 1)
  testthat::expect_length(deferred, 1)
  testthat::expect_true(is.finite(whole_life))
  testthat::expect_true(is.finite(deferred))
})


testthat::test_that("retrospective reserve functions vectorize over interest rates", {
  out <- tVxn_ret(
    x = 40,
    n = 20,
    t = 10,
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("retrospective reserves retain maturity values", {
  testthat::expect_equal(
    tVxn_ret(
      x = 40,
      n = 20,
      t = 20,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    1
  )

  testthat::expect_equal(
    tVxn1_ret(
      x = 40,
      n = 20,
      t = 20,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    0
  )
})


testthat::test_that("extended reserve functions reject incompatible lengths", {
  testthat::expect_error(
    tVnAx(
      x = c(40, 45),
      n = c(10, 15, 20),
      t = 5,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "length 1 or the common length"
  )

  testthat::expect_error(
    htVnAx(
      x = c(40, 45),
      n = 20,
      h = c(5, 10, 15),
      t = 2,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "length 1 or the common length"
  )
})


testthat::test_that("extended reserve functions require one mortality basis", {
  testthat::expect_error(
    tVx_ret(
      x = 40,
      t = 10,
      i = 0.05
    ),
    "Supply either tbl or model"
  )

  tbl <- life_table(
    x = 0:3,
    lx = c(100000, 70000, 30000, 0)
  )

  testthat::expect_error(
    tVx_ret(
      x = 0,
      t = 1,
      i = 0.05,
      model = "uniform",
      omega = 100,
      tbl = tbl
    ),
    "Supply only one"
  )
})

testthat::test_that("deferred annuity premiums require positive deferral periods", {
  testthat::expect_error(
    PnAdotx(
      x = 40,
      n = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "positive integer-like"
  )

  testthat::expect_error(
    Pnax(
      x = 40,
      n = 0,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "positive integer-like"
  )
})


testthat::test_that("continuous gain helpers allow finite negative values", {
  out <- GT_cont(
    Vt = -1,
    Vt1 = 0.5,
    P = -0.1,
    delta_actual = -0.01,
    p_actual = 0.99,
    benefit = 0,
    h = 0.25
  )

  expected <- (-1 - 0.1 * 0.25) *
    exp(-0.01 * 0.25) -
    0.99 * 0.5

  testthat::expect_equal(out, expected)
})

testthat::test_that("Thiele derivative allows negative reserves and interest forces", {
  out <- thiele_dVdt(
    V = -10,
    P = -1,
    delta = -0.01,
    mu = 0.02,
    benefit = 100
  )

  expected <- -1 +
    (-0.01) * (-10) -
    0.02 * (100 - (-10))

  testthat::expect_equal(out, expected)
})


testthat::test_that("backward Thiele step matches its defining formula", {
  V_next <- 1000
  P <- 26.96
  delta <- 0.058
  mu <- 0.002
  benefit <- 1000
  h <- 0.25

  expected <- (
    V_next -
      h * P +
      h * mu * benefit
  ) / (
    1 + h * (delta + mu)
  )

  testthat::expect_equal(
    thiele_backward_step(
      V_next = V_next,
      P = P,
      delta = delta,
      mu = mu,
      benefit = benefit,
      h = h
    ),
    expected
  )
})


testthat::test_that("backward Thiele path preserves the terminal reserve", {
  times <- seq(19, 20, by = 0.25)

  out <- thiele_backward_path(
    times = times,
    V_terminal = 1000,
    P = 26.96,
    delta = 0.058,
    mu = 0.002,
    benefit = 1000
  )

  testthat::expect_length(out, length(times))
  testthat::expect_equal(out[length(out)], 1000)
  testthat::expect_true(all(is.finite(out)))
})

testthat::test_that("backward Thiele path checks step-vector lengths", {
  times <- seq(0, 1, by = 0.25)

  testthat::expect_error(
    thiele_backward_path(
      times = times,
      V_terminal = 1000,
      P = c(10, 20),
      delta = 0.05,
      mu = 0.01,
      benefit = 1000
    ),
    "P must have length 1 or 4"
  )

  testthat::expect_error(
    thiele_backward_path(
      times = times,
      V_terminal = c(1000, 1001),
      P = 10,
      delta = 0.05,
      mu = 0.01,
      benefit = 1000
    ),
    "V_terminal must be a finite scalar"
  )
})

#' Survival models: core survival functions and parametric models
#'
#' This file provides actuarial survival-model utilities.
#'
#' @name survival_models
#' @keywords internal
NULL

# -----------------------------
# Internal helpers
# -----------------------------

.check_model <- function(model) {
  if (length(model) != 1L || !is.character(model) || is.na(model)) {
    stop("model must be a single character string.", call. = FALSE)
  }

  model <- tolower(model)
  ok <- model %in% c("uniform", "exponential", "gompertz", "makeham", "weibull")

  if (!ok) {
    stop(
      "Unknown model='", model,
      "'. Use one of: uniform, exponential, gompertz, makeham, weibull.",
      call. = FALSE
    )
  }

  model
}

.check_params <- function(model, ...) {
  p <- list(...)
  model <- .check_model(model)

  need <- switch(
    model,
    "uniform" = "omega",
    "exponential" = "lambda",
    "gompertz" = c("B", "c"),
    "makeham" = c("A", "B", "c"),
    "weibull" = c("shape", "scale")
  )

  miss <- setdiff(need, names(p))
  if (length(miss) > 0L) {
    stop(
      "Missing parameter(s) for model='", model, "': ",
      paste(miss, collapse = ", "),
      call. = FALSE
    )
  }

  if (model == "uniform") {
    if (!is.numeric(p$omega) || length(p$omega) != 1L ||
        !is.finite(p$omega) || p$omega <= 0) {
      stop("For model='uniform', omega must be a single positive number.",
           call. = FALSE)
    }
  }

  if (model == "exponential") {
    if (!is.numeric(p$lambda) || length(p$lambda) != 1L ||
        !is.finite(p$lambda) || p$lambda <= 0) {
      stop("For model='exponential', lambda must be a single positive number.",
           call. = FALSE)
    }
  }

  if (model == "gompertz") {
    if (!is.numeric(p$B) || length(p$B) != 1L ||
        !is.finite(p$B) || p$B <= 0) {
      stop("For model='gompertz', B must be a single positive number.",
           call. = FALSE)
    }
    if (!is.numeric(p$c) || length(p$c) != 1L ||
        !is.finite(p$c) || p$c <= 1) {
      stop("For model='gompertz', c must be a single number greater than 1.",
           call. = FALSE)
    }
  }

  if (model == "makeham") {
    if (!is.numeric(p$A) || length(p$A) != 1L || !is.finite(p$A)) {
      stop("For model='makeham', A must be a single finite number.",
           call. = FALSE)
    }
    if (!is.numeric(p$B) || length(p$B) != 1L ||
        !is.finite(p$B) || p$B <= 0) {
      stop("For model='makeham', B must be a single positive number.",
           call. = FALSE)
    }
    if (!is.numeric(p$c) || length(p$c) != 1L ||
        !is.finite(p$c) || p$c <= 1) {
      stop("For model='makeham', c must be a single number greater than 1.",
           call. = FALSE)
    }
    if (p$A <= -p$B) {
      stop("For model='makeham', A must be greater than -B.",
           call. = FALSE)
    }
  }

  if (model == "weibull") {
    if (!is.numeric(p$shape) || length(p$shape) != 1L ||
        !is.finite(p$shape) || p$shape <= 0) {
      stop("For model='weibull', shape must be a single positive number.",
           call. = FALSE)
    }
    if (!is.numeric(p$scale) || length(p$scale) != 1L ||
        !is.finite(p$scale) || p$scale <= 0) {
      stop("For model='weibull', scale must be a single positive number.",
           call. = FALSE)
    }
  }

  invisible(TRUE)
}

.recycle_tx <- function(t, x) {
  t <- as.numeric(t)
  x <- as.numeric(x)

  if (length(t) == 0L || length(x) == 0L) {
    stop("t and x must have positive length.", call. = FALSE)
  }

  if (length(t) == 1L && length(x) > 1L) t <- rep(t, length(x))
  if (length(x) == 1L && length(t) > 1L) x <- rep(x, length(t))

  if (length(t) != length(x)) {
    stop("t and x must have the same length, or one must have length 1.",
         call. = FALSE)
  }

  list(t = t, x = x)
}

.find_u_max <- function(x, model, ..., tol = 1e-10, u_max_cap = 5000) {
  u <- 10
  tpx_u <- tpx(u, x, model = model, ...)

  if (is.na(tpx_u) || tpx_u < tol) return(u)

  while (u < u_max_cap) {
    u <- u * 2
    tpx_u <- tpx(u, x, model = model, ...)
    if (is.na(tpx_u) || tpx_u < tol) return(u)
  }

  u_max_cap
}

# -----------------------------
# Exported functions
# -----------------------------

#' Survival function for age-at-failure
#'
#' Computes \eqn{S_0(t)=Pr(T_0>t)}.
#'
#' @param t Numeric vector of times.
#' @param model One of \code{"uniform"}, \code{"exponential"},
#'   \code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.
#' @param ... Model parameters.
#'
#' @return Numeric vector of survival probabilities.
#' @export
S0 <- function(t, model, ...) {
  model <- .check_model(model)
  .check_params(model, ...)

  t <- as.numeric(t)

  if (length(t) == 0L || any(!is.finite(t))) {
    stop("t must contain finite numeric values.", call. = FALSE)
  }
  if (any(t < 0)) {
    stop("t must be nonnegative.", call. = FALSE)
  }

  p <- list(...)

  out <- switch(
    model,
    "uniform" = {
      omega <- p$omega
      ifelse(t <= omega, pmax(omega - t, 0) / omega, 0)
    },
    "exponential" = {
      exp(-p$lambda * t)
    },
    "gompertz" = {
      exp((p$B / log(p$c)) * (1 - p$c^t))
    },
    "makeham" = {
      exp((p$B / log(p$c)) * (1 - p$c^t) - p$A * t)
    },
    "weibull" = {
      exp(-(t / p$scale)^p$shape)
    }
  )

  pmin(pmax(out, 0), 1)
}

#' Distribution functions for age-at-failure
#'
#' @name dist0
#' @aliases F0 f0
#' @inheritParams S0
#' @return Numeric vector.
NULL

#' @rdname dist0
#' @export
F0 <- function(t, model, ...) {
  1 - S0(t, model = model, ...)
}

#' @rdname dist0
#' @export
f0 <- function(t, model, ...) {
  model <- .check_model(model)
  .check_params(model, ...)

  t <- as.numeric(t)

  if (length(t) == 0L || any(!is.finite(t))) {
    stop("t must contain finite numeric values.", call. = FALSE)
  }
  if (any(t < 0)) {
    stop("t must be nonnegative.", call. = FALSE)
  }

  p <- list(...)

  out <- switch(
    model,
    "uniform" = {
      omega <- p$omega
      ifelse(t >= 0 & t < omega, 1 / omega, 0)
    },
    "exponential" = {
      p$lambda * exp(-p$lambda * t)
    },
    "gompertz" = {
      mu <- p$B * p$c^t
      mu * S0(t, model = model, ...)
    },
    "makeham" = {
      mu <- p$A + p$B * p$c^t
      mu * S0(t, model = model, ...)
    },
    "weibull" = {
      (p$shape / p$scale) *
        (t / p$scale)^(p$shape - 1) *
        exp(-(t / p$scale)^p$shape)
    }
  )

  pmax(out, 0)
}

#' Hazard or force of mortality for age-at-failure
#'
#' Computes \eqn{\lambda_0(t)}.
#'
#' @inheritParams S0
#' @return Numeric vector.
#' @export
hazard0 <- function(t, model, ...) {
  model <- .check_model(model)
  .check_params(model, ...)

  t <- as.numeric(t)

  if (length(t) == 0L || any(!is.finite(t))) {
    stop("t must contain finite numeric values.", call. = FALSE)
  }
  if (any(t < 0)) {
    stop("t must be nonnegative.", call. = FALSE)
  }

  p <- list(...)

  switch(
    model,
    "uniform" = {
      ifelse(t < p$omega, 1 / (p$omega - t), Inf)
    },
    "exponential" = {
      rep(p$lambda, length(t))
    },
    "gompertz" = {
      p$B * p$c^t
    },
    "makeham" = {
      p$A + p$B * p$c^t
    },
    "weibull" = {
      (p$shape / p$scale) * (t / p$scale)^(p$shape - 1)
    }
  )
}

#' Cumulative hazard for age-at-failure
#'
#' Computes \eqn{\Lambda_0(t)}.
#'
#' @inheritParams S0
#' @return Numeric vector.
#' @export
cumhaz0 <- function(t, model, ...) {
  model <- .check_model(model)
  .check_params(model, ...)

  t <- as.numeric(t)

  if (length(t) == 0L || any(!is.finite(t))) {
    stop("t must contain finite numeric values.", call. = FALSE)
  }
  if (any(t < 0)) {
    stop("t must be nonnegative.", call. = FALSE)
  }

  p <- list(...)

  switch(
    model,
    "uniform" = {
      ifelse(t < p$omega, log(p$omega / (p$omega - t)), Inf)
    },
    "exponential" = {
      p$lambda * t
    },
    "gompertz" = {
      (p$B / log(p$c)) * (p$c^t - 1)
    },
    "makeham" = {
      p$A * t + (p$B / log(p$c)) * (p$c^t - 1)
    },
    "weibull" = {
      (t / p$scale)^p$shape
    }
  )
}

#' Conditional survival probability
#'
#' Computes \eqn{{}_t p_x = S_0(x+t)/S_0(x)}.
#'
#' @param t Numeric vector of durations.
#' @param x Numeric vector of ages.
#' @inheritParams S0
#'
#' @return Numeric vector of survival probabilities.
#' @export
tpx <- function(t, x, model, ...) {
  model <- .check_model(model)
  .check_params(model, ...)

  tx <- .recycle_tx(t, x)
  t <- tx$t
  x <- tx$x

  if (any(!is.finite(t)) || any(!is.finite(x))) {
    stop("t and x must contain finite numeric values.", call. = FALSE)
  }
  if (any(t < 0) || any(x < 0)) {
    stop("t and x must be nonnegative.", call. = FALSE)
  }

  Sx <- S0(x, model = model, ...)
  Sxt <- S0(x + t, model = model, ...)

  out <- ifelse(Sx <= 0, 0, Sxt / Sx)

  pmin(pmax(out, 0), 1)
}

#' Conditional failure probability
#'
#' Computes \eqn{{}_t q_x = 1 - {}_t p_x}.
#'
#' @inheritParams tpx
#' @return Numeric vector of failure probabilities.
#' @export
tqx <- function(t, x, model, ...) {
  1 - tpx(t = t, x = x, model = model, ...)
}

#' Conditional density
#'
#' Computes \eqn{f_x(t)=f_0(x+t)/S_0(x)}.
#'
#' @inheritParams tpx
#' @return Numeric vector of density values.
#' @export
fx <- function(t, x, model, ...) {
  model <- .check_model(model)
  .check_params(model, ...)

  tx <- .recycle_tx(t, x)
  t <- tx$t
  x <- tx$x

  if (any(!is.finite(t)) || any(!is.finite(x))) {
    stop("t and x must contain finite numeric values.", call. = FALSE)
  }
  if (any(t < 0) || any(x < 0)) {
    stop("t and x must be nonnegative.", call. = FALSE)
  }

  Sx <- S0(x, model = model, ...)
  num <- f0(x + t, model = model, ...)

  out <- ifelse(Sx <= 0, 0, num / Sx)

  pmax(out, 0)
}

#' Complete expectation of life
#'
#' Computes \eqn{\overset{\circ}{e}_x=\int_0^\infty {}_t p_x\,dt}.
#'
#' @param x Numeric vector of ages.
#' @inheritParams S0
#' @param tol Tolerance used to choose a finite integration bound.
#'
#' @return Numeric vector of complete expectations.
#' @export
ex_complete <- function(x, model, ..., tol = 1e-10) {
  model <- .check_model(model)
  .check_params(model, ...)

  x <- as.numeric(x)

  if (length(x) == 0L || any(!is.finite(x))) {
    stop("x must contain finite numeric values.", call. = FALSE)
  }
  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  p <- list(...)

  if (model == "uniform") {
    return(ifelse(x >= p$omega, 0, (p$omega - x) / 2))
  }

  if (model == "exponential") {
    return(rep(1 / p$lambda, length(x)))
  }

  sapply(x, function(xx) {
    if (S0(xx, model = model, ...) <= 0) return(0)

    u_max <- .find_u_max(xx, model = model, ..., tol = tol)
    integrand <- function(u) tpx(u, xx, model = model, ...)

    stats::integrate(integrand, lower = 0, upper = u_max,
                     rel.tol = 1e-8)$value
  })
}

#' Curtate expectation of life
#'
#' Computes \eqn{e_x = E[K_x] = \sum_{k=1}^{\infty} {}_k p_x}.
#'
#' @param x Numeric vector of ages.
#' @inheritParams S0
#' @param k_max Maximum integer duration to sum to.
#' @param tol Stop early if the summand is smaller than this tolerance for
#'   several consecutive steps.
#'
#' @return Numeric vector of curtate expectations.
#' @export
ex_curtate <- function(x, model, ..., k_max = 5000, tol = 1e-12) {
  model <- .check_model(model)
  .check_params(model, ...)

  x <- as.numeric(x)

  if (length(x) == 0L || any(!is.finite(x))) {
    stop("x must contain finite numeric values.", call. = FALSE)
  }
  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  sapply(x, function(xx) {
    if (S0(xx, model = model, ...) <= 0) return(0)

    acc <- 0
    small_run <- 0

    for (k in seq_len(k_max)) {
      term <- tpx(k, xx, model = model, ...)
      acc <- acc + term

      if (term < tol) {
        small_run <- small_run + 1L
        if (small_run >= 10L) break
      } else {
        small_run <- 0L
      }
    }

    acc
  })
}

testthat::test_that("Uniform (de Moivre) model basics", {
  omega <- 100

  # endpoints / basic values
  testthat::expect_equal(S0(0, "uniform", omega = omega), 1)
  testthat::expect_equal(F0(0, "uniform", omega = omega), 0)
  testthat::expect_equal(S0(omega, "uniform", omega = omega), 0)
  testthat::expect_equal(F0(omega, "uniform", omega = omega), 1)

  # formula S0(t) = (omega - t)/omega on [0, omega]
  testthat::expect_equal(S0(25, "uniform", omega = omega), (omega - 25) / omega)

  # density constant on (0, omega), 0 outside
  testthat::expect_equal(f0(1, "uniform", omega = omega), 1 / omega)
  testthat::expect_equal(f0(0, "uniform", omega = omega), 1 / omega)
  testthat::expect_equal(f0(omega, "uniform", omega = omega), 0)
  testthat::expect_equal(f0(omega + 1, "uniform", omega = omega), 0)

  # hazard is 1/(omega - t) for t < omega; Inf at/after omega
  testthat::expect_equal(hazard0(10, "uniform", omega = omega), 1 / (omega - 10))
  testthat::expect_true(is.infinite(hazard0(omega, "uniform", omega = omega)))

  # cumulative hazard Lambda(t) = log(omega/(omega - t)) for t < omega
  testthat::expect_equal(
    cumhaz0(10, "uniform", omega = omega),
    log(omega / (omega - 10)),
    tolerance = 1e-12
  )
  testthat::expect_true(is.infinite(cumhaz0(omega, "uniform", omega = omega)))
})

testthat::test_that("Exponential model basics", {
  lam <- 0.02
  t <- 10

  testthat::expect_equal(S0(t, "exponential", lambda = lam), exp(-lam * t), tolerance = 1e-12)
  testthat::expect_equal(F0(t, "exponential", lambda = lam), 1 - exp(-lam * t), tolerance = 1e-12)
  testthat::expect_equal(f0(t, "exponential", lambda = lam), lam * exp(-lam * t), tolerance = 1e-12)

  # constant hazard and cumhaz
  testthat::expect_equal(hazard0(t, "exponential", lambda = lam), lam, tolerance = 1e-12)
  testthat::expect_equal(cumhaz0(t, "exponential", lambda = lam), lam * t, tolerance = 1e-12)

  # identity: f0 = hazard0 * S0
  testthat::expect_equal(
    f0(t, "exponential", lambda = lam),
    hazard0(t, "exponential", lambda = lam) * S0(t, "exponential", lambda = lam),
    tolerance = 1e-12
  )
})

testthat::test_that("Gompertz model: hazard and cumhaz consistency", {
  B <- 0.0005
  c <- 1.08
  t <- 12

  # hazard is B*c^t
  testthat::expect_equal(hazard0(t, "gompertz", B = B, c = c), B * c^t, tolerance = 1e-12)

  # cumhaz is (B/log c)(c^t - 1)
  testthat::expect_equal(
    cumhaz0(t, "gompertz", B = B, c = c),
    (B / log(c)) * (c^t - 1),
    tolerance = 1e-12
  )

  # S0 should match exp(-cumhaz) for Gompertz
  testthat::expect_equal(
    S0(t, "gompertz", B = B, c = c),
    exp(-cumhaz0(t, "gompertz", B = B, c = c)),
    tolerance = 1e-12
  )
})

testthat::test_that("Makeham model: hazard and cumhaz consistency", {
  A <- 0.0002
  B <- 0.0004
  c <- 1.07
  t <- 15

  testthat::expect_equal(hazard0(t, "makeham", A = A, B = B, c = c), A + B * c^t, tolerance = 1e-12)
  testthat::expect_equal(
    cumhaz0(t, "makeham", A = A, B = B, c = c),
    A * t + (B / log(c)) * (c^t - 1),
    tolerance = 1e-12
  )

  testthat::expect_equal(
    S0(t, "makeham", A = A, B = B, c = c),
    exp(-cumhaz0(t, "makeham", A = A, B = B, c = c)),
    tolerance = 1e-12
  )
})

testthat::test_that("Weibull model: hazard/cumhaz consistency", {
  shape <- 2.3
  scale <- 50
  t <- 10

  testthat::expect_equal(
    cumhaz0(t, "weibull", shape = shape, scale = scale),
    (t / scale)^shape,
    tolerance = 1e-12
  )

  testthat::expect_equal(
    hazard0(t, "weibull", shape = shape, scale = scale),
    (shape / scale) * (t / scale)^(shape - 1),
    tolerance = 1e-12
  )

  testthat::expect_equal(
    S0(t, "weibull", shape = shape, scale = scale),
    exp(-cumhaz0(t, "weibull", shape = shape, scale = scale)),
    tolerance = 1e-12
  )
})

testthat::test_that("Conditional relationships: tpx, tqx, fx", {
  omega <- 120
  x <- 30
  t <- 10

  # tpx = S0(x+t)/S0(x)
  testthat::expect_equal(
    tpx(t, x, "uniform", omega = omega),
    S0(x + t, "uniform", omega = omega) / S0(x, "uniform", omega = omega),
    tolerance = 1e-12
  )

  # tqx = 1 - tpx
  testthat::expect_equal(
    tqx(t, x, "uniform", omega = omega),
    1 - tpx(t, x, "uniform", omega = omega),
    tolerance = 1e-12
  )

  # fx(t) = f0(x+t)/S0(x)
  testthat::expect_equal(
    fx(t, x, "uniform", omega = omega),
    f0(x + t, "uniform", omega = omega) / S0(x, "uniform", omega = omega),
    tolerance = 1e-12
  )
})

testthat::test_that("Vectorization over t (and x where relevant)", {
  omega <- 100
  t <- c(0, 10, 25)
  x <- c(20, 30, 40)

  testthat::expect_length(S0(t, "uniform", omega = omega), length(t))
  testthat::expect_length(F0(t, "uniform", omega = omega), length(t))
  testthat::expect_length(f0(t, "uniform", omega = omega), length(t))
  testthat::expect_length(hazard0(t, "uniform", omega = omega), length(t))
  testthat::expect_length(cumhaz0(t, "uniform", omega = omega), length(t))

  testthat::expect_length(tpx(t, x, "uniform", omega = omega), length(t))
  testthat::expect_length(tqx(t, x, "uniform", omega = omega), length(t))
  testthat::expect_length(fx(t, x, "uniform", omega = omega), length(t))
})

testthat::test_that("Monotonicity / sanity checks", {
  lam <- 0.03
  t <- c(0, 5, 10, 20)

  # S0 decreasing; F0 increasing
  S <- S0(t, "exponential", lambda = lam)
  F <- F0(t, "exponential", lambda = lam)
  testthat::expect_true(all(diff(S) <= 0))
  testthat::expect_true(all(diff(F) >= 0))

  # hazard nonnegative
  testthat::expect_true(all(hazard0(t, "exponential", lambda = lam) >= 0))
})

testthat::test_that("ex_complete closed forms: uniform and exponential", {
  # uniform: e_x^o = (omega - x)/2 for x < omega; 0 for x >= omega
  omega <- 100
  testthat::expect_equal(ex_complete(20, "uniform", omega = omega), (omega - 20) / 2, tolerance = 1e-10)
  testthat::expect_equal(ex_complete(100, "uniform", omega = omega), 0)

  # exponential: e_x^o = 1/lambda (memoryless)
  lam <- 0.04
  testthat::expect_equal(ex_complete(0, "exponential", lambda = lam), 1 / lam, tolerance = 1e-10)
  testthat::expect_equal(ex_complete(50, "exponential", lambda = lam), 1 / lam, tolerance = 1e-10)
})

testthat::test_that("ex_curtate basic properties (uniform and exponential)", {
  # uniform: curtate should be finite and <= (omega-x) roughly
  omega <- 120
  x <- 30
  ec <- ex_curtate(x, "uniform", omega = omega, k_max = 5000)
  testthat::expect_true(is.finite(ec))
  testthat::expect_true(ec >= 0)
  testthat::expect_true(ec <= (omega - x) + 1) # loose sanity bound

  # exponential: curtate expectation should be close to sum_{k>=1} exp(-lambda k)
  lam <- 0.05
  approx_geom <- sum(exp(-lam * (1:2000)))  # high truncation
  ec2 <- ex_curtate(10, "exponential", lambda = lam, k_max = 5000)
  testthat::expect_equal(ec2, approx_geom, tolerance = 1e-6)
})

testthat::test_that("Input validation errors", {
  # negative t or x
  testthat::expect_error(S0(-1, "exponential", lambda = 0.1))
  testthat::expect_error(tpx(1, -2, "exponential", lambda = 0.1))
  testthat::expect_error(tpx(-1, 2, "exponential", lambda = 0.1))

  # unknown model
  testthat::expect_error(S0(1, "notamodel", lambda = 0.1))

  # missing parameters
  testthat::expect_error(S0(1, "uniform"))
  testthat::expect_error(S0(1, "exponential"))
  testthat::expect_error(S0(1, "gompertz", B = 0.1))        # missing c
  testthat::expect_error(S0(1, "makeham", A = 0.1, B = 0.1))# missing c

  # invalid parameters
  testthat::expect_error(S0(1, "uniform", omega = -1))
  testthat::expect_error(S0(1, "exponential", lambda = 0))
  testthat::expect_error(S0(1, "gompertz", B = 0, c = 1.1))
  testthat::expect_error(S0(1, "gompertz", B = 0.1, c = 1))
  testthat::expect_error(S0(1, "weibull", shape = 0, scale = 10))
  testthat::expect_error(S0(1, "weibull", shape = 2, scale = 0))
})

# life_table_conversions.R
#
# Conversion utilities linking survival-function notation and life-table notation.
#
# Functions:
#   S0_to_lx()  : convert S0(x) values to l_x values using a radix
#   lx_to_S0()  : convert l_x values to S0(x) values
#   px_to_lx()  : construct l_x values from p_x values and a radix
#   qx_to_lx()  : construct l_x values from q_x values and a radix

# -------------------------------------------------------------------------
# Internal helper
# -------------------------------------------------------------------------

.check_radix <- function(radix) {
  if (!is.numeric(radix) || length(radix) != 1L ||
      !is.finite(radix) || radix <= 0) {
    stop("radix must be a single positive finite number.", call. = FALSE)
  }

  radix
}

# -------------------------------------------------------------------------
# S0 -> lx
# -------------------------------------------------------------------------

#' Convert survival probabilities to life-table values
#'
#' Converts survival function values \eqn{S_0(x)} into life-table survivor
#' values \eqn{l_x = l_0 S_0(x)} using a chosen radix.
#'
#' @param S0 Numeric vector of survival probabilities.
#' @param radix Positive radix \eqn{l_0}.
#'
#' @return Numeric vector of \eqn{l_x} values.
#' @export
S0_to_lx <- function(S0, radix = 100000) {
  radix <- .check_radix(radix)
  S0 <- as.numeric(S0)

  if (length(S0) == 0L) {
    stop("S0 must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(S0))) {
    stop("S0 must contain only finite values.", call. = FALSE)
  }
  if (any(S0 < 0 | S0 > 1)) {
    stop("S0 values must lie in [0, 1].", call. = FALSE)
  }
  if (is.unsorted(-S0)) {
    stop("S0 must be nonincreasing.", call. = FALSE)
  }

  radix * S0
}

# -------------------------------------------------------------------------
# lx -> S0
# -------------------------------------------------------------------------

#' Convert life-table values to survival probabilities
#'
#' Converts life-table survivor values \eqn{l_x} into survival probabilities
#' \eqn{S_0(x) = l_x / l_0}.
#'
#' @param lx Numeric vector of life-table survivor values.
#'
#' @return Numeric vector of \eqn{S_0(x)} values.
#' @export
lx_to_S0 <- function(lx) {
  lx <- as.numeric(lx)

  if (length(lx) == 0L) {
    stop("lx must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(lx))) {
    stop("lx must contain only finite values.", call. = FALSE)
  }
  if (any(lx < 0)) {
    stop("lx values must be nonnegative.", call. = FALSE)
  }
  if (is.unsorted(-lx)) {
    stop("lx must be nonincreasing.", call. = FALSE)
  }
  if (lx[1] <= 0) {
    stop("The radix l0 = lx[1] must be positive.", call. = FALSE)
  }

  lx / lx[1]
}

# -------------------------------------------------------------------------
# px -> lx
# -------------------------------------------------------------------------

#' Construct life-table values from p_x values
#'
#' Builds life-table survivor values recursively from
#' \eqn{l_{x+1} = l_x p_x}, starting from a chosen radix.
#'
#' @param px Numeric vector of one-year survival probabilities \eqn{p_x}.
#' @param radix Positive radix \eqn{l_0}.
#'
#' @return Numeric vector of \eqn{l_x} values of length \code{length(px) + 1}.
#' @export
px_to_lx <- function(px, radix = 100000) {
  radix <- .check_radix(radix)
  px <- as.numeric(px)

  if (length(px) == 0L) {
    stop("px must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(px))) {
    stop("px must contain only finite values.", call. = FALSE)
  }
  if (any(px < 0 | px > 1)) {
    stop("px values must lie in [0, 1].", call. = FALSE)
  }

  lx <- numeric(length(px) + 1L)
  lx[1] <- radix

  for (i in seq_along(px)) {
    lx[i + 1L] <- lx[i] * px[i]
  }

  lx
}

# -------------------------------------------------------------------------
# qx -> lx
# -------------------------------------------------------------------------

#' Construct life-table values from q_x values
#'
#' Builds life-table survivor values recursively from
#' \eqn{l_{x+1} = l_x (1 - q_x)}, starting from a chosen radix.
#'
#' @param qx Numeric vector of one-year death probabilities \eqn{q_x}.
#' @param radix Positive radix \eqn{l_0}.
#'
#' @return Numeric vector of \eqn{l_x} values of length \code{length(qx) + 1}.
#' @export
qx_to_lx <- function(qx, radix = 100000) {
  radix <- .check_radix(radix)
  qx <- as.numeric(qx)

  if (length(qx) == 0L) {
    stop("qx must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(qx))) {
    stop("qx must contain only finite values.", call. = FALSE)
  }
  if (any(qx < 0 | qx > 1)) {
    stop("qx values must lie in [0, 1].", call. = FALSE)
  }

  lx <- numeric(length(qx) + 1L)
  lx[1] <- radix

  for (i in seq_along(qx)) {
    lx[i + 1L] <- lx[i] * (1 - qx[i])
  }

  lx
}

# life_table_core.R
#
# Core life table utilities.
# These functions work in the discrete tabular setting.

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.check_scalar_or_vector_lengths <- function(...) {
  lens <- vapply(list(...), length, integer(1))
  nz <- lens[lens > 0]

  if (length(nz) == 0L) return(TRUE)

  big <- nz[nz != 1L]
  length(unique(big)) <= 1L
}

.recycle_life_table_args <- function(x, n = NULL) {
  x <- as.numeric(x)

  if (length(x) == 0L || any(!is.finite(x))) {
    stop("x must contain finite numeric values.", call. = FALSE)
  }

  if (is.null(n)) {
    return(list(x = x))
  }

  n <- as.numeric(n)

  if (length(n) == 0L || any(!is.finite(n))) {
    stop("n must contain finite numeric values.", call. = FALSE)
  }

  if (!.check_scalar_or_vector_lengths(x, n)) {
    stop("x and n must have compatible lengths.", call. = FALSE)
  }

  len <- max(length(x), length(n))

  list(
    x = rep_len(x, len),
    n = rep_len(n, len)
  )
}

.validate_life_table <- function(tbl) {
  required_cols <- c("x", "lx", "dx", "qx", "px")

  if (!all(required_cols %in% names(tbl))) {
    stop("life_table object must contain columns: x, lx, dx, qx, px.",
         call. = FALSE)
  }

  if (any(!is.finite(tbl$x))) {
    stop("x values must be finite.", call. = FALSE)
  }

  if (any(diff(tbl$x) <= 0)) {
    stop("x values must be strictly increasing.", call. = FALSE)
  }

  if (any(!is.finite(tbl$lx), na.rm = TRUE)) {
    stop("lx values must be finite.", call. = FALSE)
  }

  if (any(tbl$lx < 0, na.rm = TRUE)) {
    stop("lx values must be nonnegative.", call. = FALSE)
  }

  invisible(tbl)
}

.is_closed_life_table <- function(tbl) {
  n <- nrow(tbl)
  isTRUE(tbl$lx[n] == 0)
}

.get_tbl_index <- function(tbl, x) {
  match(x, tbl$x)
}

.get_lx_at <- function(tbl, x) {
  x <- as.numeric(x)

  idx <- .get_tbl_index(tbl, x)
  out <- rep(NA_real_, length(x))

  ok <- !is.na(idx)
  out[ok] <- tbl$lx[idx[ok]]

  beyond <- is.na(idx) & x > max(tbl$x)

  if (any(beyond) && .is_closed_life_table(tbl)) {
    out[beyond] <- 0
  }

  out
}

# -------------------------------------------------------------------------
# Constructor
# -------------------------------------------------------------------------

#' Construct a life table
#'
#' Build a discrete life table from one of \code{lx}, \code{qx}, \code{px},
#' or \code{S0}.
#'
#' @param x Numeric vector of ages.
#' @param lx Numeric vector of \eqn{l_x} values.
#' @param qx Numeric vector of \eqn{q_x} values.
#' @param px Numeric vector of \eqn{p_x} values.
#' @param S0 Numeric vector of \eqn{S_0(x)} values.
#' @param radix Radix used when converting \code{S0} to \code{lx}, or when
#'   building from \code{qx} or \code{px}.
#'
#' @return A data frame with class \code{"life_table"}.
#' @export
life_table <- function(x,
                       lx = NULL,
                       qx = NULL,
                       px = NULL,
                       S0 = NULL,
                       radix = 100000) {
  supplied <- c(
    !is.null(lx),
    !is.null(qx),
    !is.null(px),
    !is.null(S0)
  )

  if (sum(supplied) != 1L) {
    stop("Supply exactly one of lx, qx, px, or S0.", call. = FALSE)
  }

  x <- as.numeric(x)

  if (length(x) == 0L || any(!is.finite(x))) {
    stop("x must contain finite numeric values.", call. = FALSE)
  }

  if (is.null(radix) || length(radix) != 1L ||
      !is.finite(radix) || radix <= 0) {
    stop("radix must be a positive finite scalar.", call. = FALSE)
  }

  if (!all(diff(x) > 0)) {
    stop("x must be strictly increasing.", call. = FALSE)
  }

  if (!is.null(lx)) {
    lx <- as.numeric(lx)

    if (length(lx) != length(x)) {
      stop("When lx is supplied, length(lx) must equal length(x).",
           call. = FALSE)
    }

    if (any(!is.finite(lx))) {
      stop("lx must contain finite numeric values.", call. = FALSE)
    }

    if (any(lx < 0)) {
      stop("lx must be nonnegative.", call. = FALSE)
    }

    if (is.unsorted(-lx)) {
      stop("lx must be nonincreasing with age.", call. = FALSE)
    }

    tbl_x <- x
    tbl_lx <- lx
  }

  if (!is.null(S0)) {
    S0 <- as.numeric(S0)

    if (length(S0) != length(x)) {
      stop("When S0 is supplied, length(S0) must equal length(x).",
           call. = FALSE)
    }

    if (any(!is.finite(S0))) {
      stop("S0 must contain finite numeric values.", call. = FALSE)
    }

    if (any(S0 < 0 | S0 > 1)) {
      stop("S0 values must lie in [0, 1].", call. = FALSE)
    }

    if (is.unsorted(-S0)) {
      stop("S0 must be nonincreasing with age.", call. = FALSE)
    }

    tbl_x <- x
    tbl_lx <- radix * S0
  }

  if (!is.null(qx)) {
    qx <- as.numeric(qx)

    if (length(qx) != length(x)) {
      stop("When qx is supplied, length(qx) must equal length(x).",
           call. = FALSE)
    }

    if (any(!is.finite(qx))) {
      stop("qx must contain finite numeric values.", call. = FALSE)
    }

    if (any(qx < 0 | qx > 1)) {
      stop("qx values must lie in [0, 1].", call. = FALSE)
    }

    lx_vals <- numeric(length(qx) + 1L)
    lx_vals[1] <- radix

    for (i in seq_along(qx)) {
      lx_vals[i + 1L] <- lx_vals[i] * (1 - qx[i])
    }

    tbl_x <- c(x, max(x) + 1)
    tbl_lx <- lx_vals
  }

  if (!is.null(px)) {
    px <- as.numeric(px)

    if (length(px) != length(x)) {
      stop("When px is supplied, length(px) must equal length(x).",
           call. = FALSE)
    }

    if (any(!is.finite(px))) {
      stop("px must contain finite numeric values.", call. = FALSE)
    }

    if (any(px < 0 | px > 1)) {
      stop("px values must lie in [0, 1].", call. = FALSE)
    }

    lx_vals <- numeric(length(px) + 1L)
    lx_vals[1] <- radix

    for (i in seq_along(px)) {
      lx_vals[i + 1L] <- lx_vals[i] * px[i]
    }

    tbl_x <- c(x, max(x) + 1)
    tbl_lx <- lx_vals
  }

  n <- length(tbl_lx)

  dx_vals <- c(tbl_lx[-n] - tbl_lx[-1L], NA_real_)

  qx_vals <- rep(NA_real_, n)
  px_vals <- rep(NA_real_, n)

  denom <- tbl_lx[-n]
  positive <- denom > 0

  qx_vals[-n][positive] <- dx_vals[-n][positive] / denom[positive]
  px_vals[-n][positive] <- tbl_lx[-1L][positive] / denom[positive]

  out <- data.frame(
    x = tbl_x,
    lx = tbl_lx,
    dx = dx_vals,
    qx = qx_vals,
    px = px_vals,
    row.names = NULL
  )

  class(out) <- c("life_table", class(out))
  .validate_life_table(out)
  out
}

# -------------------------------------------------------------------------
# Accessors and derived quantities
# -------------------------------------------------------------------------

#' Extract life-table survivor values
#'
#' @param tbl A \code{life_table} object.
#' @param x Ages.
#'
#' @return Numeric vector of \eqn{l_x} values.
#' @export
lx <- function(tbl, x) {
  .validate_life_table(tbl)
  args <- .recycle_life_table_args(x)
  .get_lx_at(tbl, args$x)
}

#' Compute deaths between ages x and x+1
#'
#' @param tbl A \code{life_table} object.
#' @param x Ages.
#'
#' @return Numeric vector of \eqn{d_x} values.
#' @export
dx <- function(tbl, x) {
  .validate_life_table(tbl)
  args <- .recycle_life_table_args(x)

  lx_x <- .get_lx_at(tbl, args$x)
  lx_x1 <- .get_lx_at(tbl, args$x + 1)

  lx_x - lx_x1
}

#' Compute deaths over an n-year interval from a life table
#'
#' @param tbl A \code{life_table} object.
#' @param x Ages.
#' @param n Nonnegative integer durations.
#'
#' @return Numeric vector of \eqn{{}_n d_x} values.
#' @export
ndx <- function(tbl, x, n) {
  .validate_life_table(tbl)

  args <- .recycle_life_table_args(x, n)
  x <- args$x
  n <- args$n

  if (any(n < 0 | n != floor(n), na.rm = TRUE)) {
    stop("n must be a nonnegative integer.", call. = FALSE)
  }

  lx_x <- .get_lx_at(tbl, x)
  lx_xn <- .get_lx_at(tbl, x + n)

  lx_x - lx_xn
}

#' Compute one-year death probability from a life table
#'
#' @param tbl A \code{life_table} object.
#' @param x Ages.
#'
#' @return Numeric vector of \eqn{q_x} values.
#' @export
qx_tab <- function(tbl, x) {
  .validate_life_table(tbl)
  args <- .recycle_life_table_args(x)

  lx_x <- lx(tbl, args$x)
  dx_x <- dx(tbl, args$x)

  out <- dx_x / lx_x
  out[!is.finite(out)] <- NA_real_

  out
}

#' Compute n-year survival probability from a life table
#'
#' @param tbl A \code{life_table} object.
#' @param x Ages.
#' @param n Nonnegative integer durations.
#'
#' @return Numeric vector of \eqn{{}_n p_x} values.
#' @export
npx <- function(tbl, x, n) {
  .validate_life_table(tbl)

  args <- .recycle_life_table_args(x, n)
  x <- args$x
  n <- args$n

  if (any(n < 0 | n != floor(n), na.rm = TRUE)) {
    stop("n must be a nonnegative integer.", call. = FALSE)
  }

  lx_x <- .get_lx_at(tbl, x)
  lx_xn <- .get_lx_at(tbl, x + n)

  out <- lx_xn / lx_x

  zero_duration <- n == 0 & !is.na(lx_x) & lx_x > 0
  out[zero_duration] <- 1

  out[!is.finite(out)] <- NA_real_

  out
}

#' Compute n-year death probability from a life table
#'
#' @param tbl A \code{life_table} object.
#' @param x Ages.
#' @param n Nonnegative integer durations.
#'
#' @return Numeric vector of \eqn{{}_n q_x} values.
#' @export
nqx <- function(tbl, x, n) {
  .validate_life_table(tbl)

  p <- npx(tbl, x, n)
  out <- 1 - p
  out[is.na(p)] <- NA_real_

  out
}

# -------------------------------------------------------------------------
# Print method
# -------------------------------------------------------------------------

#' @export
print.life_table <- function(x, ...) {
  cat("<life_table>\n")
  print.data.frame(as.data.frame(x), row.names = FALSE, ...)
  invisible(x)
}

#' m-thly insurance models
#'
#' Exact m-thly contingent payment and insurance functions.
#'
#' @name insurance_mthly
#' @keywords internal
NULL

# ------------------------------------------------------------
# Internal helpers
# ------------------------------------------------------------

.check_m_mthly <- function(m) {
  m <- as.numeric(m)

  if (length(m) != 1L || !is.finite(m) || m <= 0 ||
      abs(m - round(m)) > 1e-10) {
    stop("m must be a positive integer.", call. = FALSE)
  }

  as.integer(round(m))
}

.check_i_mthly <- function(i) {
  i <- as.numeric(i)

  if (length(i) == 0L || any(!is.finite(i)) || any(i <= -1)) {
    stop("i must contain finite values greater than -1.", call. = FALSE)
  }

  i
}

.recycle_mthly_vectors <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) {
    stop("All arguments must have positive length.", call. = FALSE)
  }

  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)

  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) {
    if (length(z) == 1L) rep(z, common_len) else z
  })
}

.check_x_mthly <- function(x) {
  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  invisible(TRUE)
}

.check_n_mthly <- function(n) {
  if (any(n < 0)) {
    stop("n must be nonnegative.", call. = FALSE)
  }

  invisible(TRUE)
}

# Kept for backward compatibility with other package files.
.recycle_x_n_m <- function(x, n) {
  args <- .recycle_mthly_vectors(x, n)
  x <- args[[1L]]
  n <- args[[2L]]

  .check_x_mthly(x)
  .check_n_mthly(n)

  list(x = x, n = n)
}

.get_upper_m <- function(x, n = NULL, model, ...) {
  if (tolower(model) == "uniform") {
    dots <- list(...)

    if (is.null(dots$omega)) {
      stop("For model = 'uniform', omega must be supplied.", call. = FALSE)
    }

    upper <- dots$omega - x

    if (!is.null(n)) {
      upper <- pmin(upper, n)
    }

    return(pmax(upper, 0))
  }

  if (is.null(n)) {
    return(Inf)
  }

  n
}

.pure_endowment_m <- function(x, n, i, model, ...) {
  discount(i, n) * tpx(n, x = x, model = model, ...)
}

.max_j_m <- function(x, m, model, ..., tol = 1e-12, j_max = 100000L) {
  upper <- .get_upper_m(x, model = model, ...)

  if (is.finite(upper)) {
    return(max(ceiling(upper * m) - 1L, 0L))
  }

  dt <- 1 / m

  for (j in 0:j_max) {
    s_j <- tpx(j * dt, x = x, model = model, ...)
    if (!is.finite(s_j) || s_j <= tol) {
      return(j)
    }
  }

  j_max
}

# ------------------------------------------------------------
# First moments
# ------------------------------------------------------------

#' m-thly whole life insurance APV
#'
#' Computes
#' \eqn{A_x^{(m)} =
#' \sum_{j=0}^{\infty} v^{(j+1)/m}
#' \Pr(j/m < T_x \le (j+1)/m)}.
#'
#' @param x Age.
#' @param i Effective annual interest rate.
#' @param m Positive integer payment frequency.
#' @param model Parametric survival model name.
#' @param ... Additional model parameters passed to survival-model functions.
#' @param tol Numerical tolerance for truncating the infinite sum.
#' @param j_max Maximum number of m-thly intervals in the sum.
#'
#' @return Numeric vector of APVs.
#' @export
Ax_m <- function(x, i, m, model, ..., tol = 1e-12, j_max = 100000L) {
  m <- .check_m_mthly(m)

  args <- .recycle_mthly_vectors(x, .check_i_mthly(i))
  x <- args[[1L]]
  i <- args[[2L]]

  .check_x_mthly(x)

  dt <- 1 / m

  sapply(seq_along(x), function(k) {
    xx <- x[k]
    ii <- i[k]

    jj_max <- .max_j_m(xx, m, model = model, ..., tol = tol, j_max = j_max)
    j <- 0:jj_max

    p_start <- tpx(j * dt, x = xx, model = model, ...)
    p_end <- tpx((j + 1) * dt, x = xx, model = model, ...)
    probs <- p_start - p_end

    sum(discount(ii, (j + 1) * dt) * probs, na.rm = TRUE)
  })
}

#' m-thly term insurance APV
#'
#' Computes
#' \eqn{A_{x:\overline{n}|}^{1(m)} =
#' \sum_{j=0}^{mn-1} v^{(j+1)/m}
#' \Pr(j/m < T_x \le (j+1)/m)}.
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param m Positive integer payment frequency.
#' @param model Parametric survival model name.
#' @param ... Additional model parameters passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
Axn1_m <- function(x, n, i, m, model, ...) {
  m <- .check_m_mthly(m)

  args <- .recycle_mthly_vectors(x, n, .check_i_mthly(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_mthly(x)
  .check_n_mthly(n)

  dt <- 1 / m

  sapply(seq_along(x), function(k) {
    xx <- x[k]
    nn <- n[k]
    ii <- i[k]

    upper <- .get_upper_m(xx, n = nn, model = model, ...)

    if (upper <= 0) {
      return(0)
    }

    jj_max <- max(ceiling(upper * m) - 1L, 0L)
    j <- 0:jj_max

    p_start <- tpx(j * dt, x = xx, model = model, ...)
    p_end <- tpx((j + 1) * dt, x = xx, model = model, ...)
    probs <- p_start - p_end

    sum(discount(ii, (j + 1) * dt) * probs, na.rm = TRUE)
  })
}

#' m-thly deferred insurance APV
#'
#' Computes
#' \eqn{{}_{n\mid}A_x^{(m)} = v^n {}_np_x A_{x+n}^{(m)}}.
#'
#' @param x Age.
#' @param n Deferral period.
#' @param i Effective annual interest rate.
#' @param m Positive integer payment frequency.
#' @param model Parametric survival model name.
#' @param ... Additional model parameters passed to survival-model functions.
#' @param tol Numerical tolerance for truncating the infinite sum.
#' @param j_max Maximum number of m-thly intervals in the sum.
#'
#' @return Numeric vector of APVs.
#' @export
nAx_m <- function(x, n, i, m, model, ..., tol = 1e-12, j_max = 100000L) {
  m <- .check_m_mthly(m)

  args <- .recycle_mthly_vectors(x, n, .check_i_mthly(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_mthly(x)
  .check_n_mthly(n)

  .pure_endowment_m(x, n, i, model = model, ...) *
    Ax_m(x + n, i, m, model = model, ..., tol = tol, j_max = j_max)
}

#' m-thly endowment insurance APV
#'
#' Computes
#' \eqn{A_{x:\overline{n}|}^{(m)} =
#' A_{x:\overline{n}|}^{1(m)} + v^n {}_np_x}.
#'
#' @param x Age.
#' @param n Term.
#' @param i Effective annual interest rate.
#' @param m Positive integer payment frequency.
#' @param model Parametric survival model name.
#' @param ... Additional model parameters passed to survival-model functions.
#'
#' @return Numeric vector of APVs.
#' @export
Axn_m <- function(x, n, i, m, model, ...) {
  m <- .check_m_mthly(m)

  args <- .recycle_mthly_vectors(x, n, .check_i_mthly(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  .check_x_mthly(x)
  .check_n_mthly(n)

  Axn1_m(x, n, i, m, model = model, ...) +
    .pure_endowment_m(x, n, i, model = model, ...)
}

# ------------------------------------------------------------
# Second moments
# ------------------------------------------------------------

#' Second moment of m-thly whole life insurance PV
#'
#' Computes \eqn{{}^{2}A_x^{(m)}} by evaluating \eqn{A_x^{(m)}}
#' at doubled force.
#'
#' @inheritParams Ax_m
#' @return Numeric vector of second moments.
#' @export
A2x_m <- function(x, i, m, model, ..., tol = 1e-12, j_max = 100000L) {
  Ax_m(x, i = double_force_i(i), m = m, model = model, ...,
       tol = tol, j_max = j_max)
}

#' Second moment of m-thly term insurance PV
#'
#' @inheritParams Axn1_m
#' @return Numeric vector of second moments.
#' @export
A2xn1_m <- function(x, n, i, m, model, ...) {
  Axn1_m(x, n, i = double_force_i(i), m = m, model = model, ...)
}

#' Second moment of m-thly deferred insurance PV
#'
#' @inheritParams nAx_m
#' @return Numeric vector of second moments.
#' @export
A2nAx_m <- function(x, n, i, m, model, ..., tol = 1e-12, j_max = 100000L) {
  nAx_m(x, n, i = double_force_i(i), m = m, model = model, ...,
        tol = tol, j_max = j_max)
}

#' Second moment of m-thly endowment insurance PV
#'
#' @inheritParams Axn_m
#' @return Numeric vector of second moments.
#' @export
A2xn_m <- function(x, n, i, m, model, ...) {
  m <- .check_m_mthly(m)

  args <- .recycle_mthly_vectors(x, n, .check_i_mthly(i))
  x <- args[[1L]]
  n <- args[[2L]]
  i <- args[[3L]]

  A2xn1_m(x, n, i, m, model = model, ...) +
    .pure_endowment_m(x, n, double_force_i(i), model = model, ...)
}

# ------------------------------------------------------------
# Variances
# ------------------------------------------------------------

#' Variance of m-thly whole life insurance PV
#'
#' Computes \eqn{\mathrm{Var}(Z_x^{(m)}) =
#' {}^{2}A_x^{(m)} - (A_x^{(m)})^2}.
#'
#' @inheritParams Ax_m
#' @return Numeric vector of variances.
#' @export
var_Ax_m <- function(x, i, m, model, ..., tol = 1e-12, j_max = 100000L) {
  A2x_m(x, i, m, model = model, ..., tol = tol, j_max = j_max) -
    Ax_m(x, i, m, model = model, ..., tol = tol, j_max = j_max)^2
}

#' Variance of m-thly term insurance PV
#'
#' @inheritParams Axn1_m
#' @return Numeric vector of variances.
#' @export
var_Axn1_m <- function(x, n, i, m, model, ...) {
  A2xn1_m(x, n, i, m, model = model, ...) -
    Axn1_m(x, n, i, m, model = model, ...)^2
}

#' Variance of m-thly deferred insurance PV
#'
#' @inheritParams nAx_m
#' @return Numeric vector of variances.
#' @export
var_nAx_m <- function(x, n, i, m, model, ..., tol = 1e-12, j_max = 100000L) {
  A2nAx_m(x, n, i, m, model = model, ..., tol = tol, j_max = j_max) -
    nAx_m(x, n, i, m, model = model, ..., tol = tol, j_max = j_max)^2
}

#' Variance of m-thly endowment insurance PV
#'
#' @inheritParams Axn_m
#' @return Numeric vector of variances.
#' @export
var_Axn_m <- function(x, n, i, m, model, ...) {
  A2xn_m(x, n, i, m, model = model, ...) -
    Axn_m(x, n, i, m, model = model, ...)^2
}

test_that("Axn_m decomposes into m-thly term insurance plus pure endowment", {
  i <- 0.05
  m <- 4

  expect_equal(
    Axn_m(40, 10, i, m, model = "uniform", omega = 100),
    Axn1_m(40, 10, i, m, model = "uniform", omega = 100) +
      discount(i, 10) * tpx(10, x = 40, model = "uniform", omega = 100)
  )
})

test_that("nAx_m equals v^n * npx * Ax_m at age x+n", {
  i <- 0.05
  m <- 4

  expect_equal(
    nAx_m(40, 10, i, m, model = "uniform", omega = 100),
    discount(i, 10) *
      tpx(10, x = 40, model = "uniform", omega = 100) *
      Ax_m(50, i, m, model = "uniform", omega = 100)
  )
})

test_that("A2x_m evaluates Ax_m at doubled force", {
  i <- 0.05
  m <- 4

  expect_equal(
    A2x_m(40, i, m, model = "uniform", omega = 100),
    Ax_m(40, double_force_i(i), m, model = "uniform", omega = 100)
  )
})

test_that("A2xn1_m evaluates Axn1_m at doubled force", {
  i <- 0.05
  m <- 4

  expect_equal(
    A2xn1_m(40, 10, i, m, model = "uniform", omega = 100),
    Axn1_m(40, 10, double_force_i(i), m, model = "uniform", omega = 100)
  )
})

test_that("A2nAx_m evaluates nAx_m at doubled force", {
  i <- 0.05
  m <- 4

  expect_equal(
    A2nAx_m(40, 10, i, m, model = "uniform", omega = 100),
    nAx_m(40, 10, double_force_i(i), m, model = "uniform", omega = 100)
  )
})

test_that("A2xn_m decomposes into second moments of m-thly term and pure endowment", {
  i <- 0.05
  i2 <- double_force_i(i)
  m <- 4

  expect_equal(
    A2xn_m(40, 10, i, m, model = "uniform", omega = 100),
    A2xn1_m(40, 10, i, m, model = "uniform", omega = 100) +
      discount(i2, 10) * tpx(10, x = 40, model = "uniform", omega = 100)
  )
})

test_that("var_Ax_m equals A2x_m - Ax_m^2", {
  i <- 0.05
  m <- 4

  expect_equal(
    var_Ax_m(40, i, m, model = "uniform", omega = 100),
    A2x_m(40, i, m, model = "uniform", omega = 100) -
      Ax_m(40, i, m, model = "uniform", omega = 100)^2
  )
})

test_that("var_Axn1_m equals A2xn1_m - Axn1_m^2", {
  i <- 0.05
  m <- 4

  expect_equal(
    var_Axn1_m(40, 10, i, m, model = "uniform", omega = 100),
    A2xn1_m(40, 10, i, m, model = "uniform", omega = 100) -
      Axn1_m(40, 10, i, m, model = "uniform", omega = 100)^2
  )
})

test_that("var_nAx_m equals A2nAx_m - nAx_m^2", {
  i <- 0.05
  m <- 4

  expect_equal(
    var_nAx_m(40, 10, i, m, model = "uniform", omega = 100),
    A2nAx_m(40, 10, i, m, model = "uniform", omega = 100) -
      nAx_m(40, 10, i, m, model = "uniform", omega = 100)^2
  )
})

test_that("var_Axn_m equals A2xn_m - Axn_m^2", {
  i <- 0.05
  m <- 4

  expect_equal(
    var_Axn_m(40, 10, i, m, model = "uniform", omega = 100),
    A2xn_m(40, 10, i, m, model = "uniform", omega = 100) -
      Axn_m(40, 10, i, m, model = "uniform", omega = 100)^2
  )
})

test_that("m-thly insurance functions vectorize correctly", {
  i <- 0.05
  m <- 4

  expect_length(Axn1_m(c(40, 45), c(10, 5), i, m, model = "uniform", omega = 100), 2)
  expect_length(nAx_m(c(40, 45), c(10, 5), i, m, model = "uniform", omega = 100), 2)
  expect_length(Axn_m(c(40, 45), c(10, 5), i, m, model = "uniform", omega = 100), 2)
})

test_that("m-thly insurance functions reject invalid inputs", {
  i <- 0.05
  m <- 4

  expect_error(Ax_m(-1, i, m, model = "uniform", omega = 100))
  expect_error(Axn1_m(40, -1, i, m, model = "uniform", omega = 100))
  expect_error(nAx_m(40, -1, i, m, model = "uniform", omega = 100))
  expect_error(Ax_m(40, -1, m, model = "uniform", omega = 100))
  expect_error(Ax_m(40, i, 0, model = "uniform", omega = 100))
  expect_error(Ax_m(40, i, 2.5, model = "uniform", omega = 100))
})

test_that("m-thly whole life APV under uniform survival is between 0 and 1", {
  i <- 0.05
  m <- 4
  val <- Ax_m(40, i, m, model = "uniform", omega = 100)

  expect_true(is.numeric(val))
  expect_true(val > 0)
  expect_true(val < 1)
})

test_that("m-thly term insurance APV is not greater than m-thly whole life APV", {
  i <- 0.05
  m <- 4

  term_val <- Axn1_m(40, 10, i, m, model = "uniform", omega = 100)
  whole_val <- Ax_m(40, i, m, model = "uniform", omega = 100)

  expect_true(term_val <= whole_val)
})

test_that("m-thly endowment APV is at least the pure endowment component", {
  i <- 0.05
  m <- 4
  pure_val <- discount(i, 10) * tpx(10, x = 40, model = "uniform", omega = 100)
  endow_val <- Axn_m(40, 10, i, m, model = "uniform", omega = 100)

  expect_true(endow_val >= pure_val)
})

test_that("Axn1_m equals Ax_m minus nAx_m", {
  i <- 0.05
  m <- 4

  expect_equal(
    Axn1_m(40, 10, i, m, model = "uniform", omega = 100),
    Ax_m(40, i, m, model = "uniform", omega = 100) -
      nAx_m(40, 10, i, m, model = "uniform", omega = 100)
  )
})

test_that("m-thly APV lies between annual and continuous values under uniform survival", {
  i <- 0.05
  x <- 40

  annual_val <- Ax(x, i, model = "uniform", omega = 100)
  mthly_val  <- Ax_m(x, i, 4, model = "uniform", omega = 100)
  cont_val   <- Abarx(x, i, model = "uniform", omega = 100)

  expect_true(mthly_val >= annual_val)
  expect_true(mthly_val <= cont_val)
})

test_that("larger m gives value closer to continuous insurance", {
  i <- 0.05
  x <- 40

  val_m2  <- Ax_m(x, i, 2, model = "uniform", omega = 100)
  val_m12 <- Ax_m(x, i, 12, model = "uniform", omega = 100)
  val_cont <- Abarx(x, i, model = "uniform", omega = 100)

  expect_true(abs(val_m12 - val_cont) <= abs(val_m2 - val_cont))
})

test_that("m-thly whole life APV works for exponential model", {
  i <- 0.05
  m <- 12
  lambda <- 0.25

  val <- Ax_m(40, i, m, model = "exponential", lambda = lambda)

  expect_true(is.numeric(val))
  expect_true(length(val) == 1)
  expect_true(val > 0)
  expect_true(val < 1)
})


testthat::test_that("m-thly insurance functions vectorize over i", {
  out <- Ax_m(x = 40, i = c(0.03, 0.05), m = 12,
              model = "uniform", omega = 100)

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
  testthat::expect_true(out[1] > out[2])
})

testthat::test_that("m-thly term insurance functions recycle x, n, and i", {
  out <- Axn1_m(
    x = c(40, 50),
    n = 10,
    i = c(0.03, 0.05),
    m = 12,
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.recycle2_interest <- function(a, b, a_name = "a", b_name = "b") {
  n <- max(length(a), length(b))

  if (!length(a) %in% c(1L, n)) {
    stop(a_name, " must have length 1 or common length.", call. = FALSE)
  }
  if (!length(b) %in% c(1L, n)) {
    stop(b_name, " must have length 1 or common length.", call. = FALSE)
  }

  if (length(a) == 1L) a <- rep(a, n)
  if (length(b) == 1L) b <- rep(b, n)

  list(a = a, b = b)
}

#' Convert between compound-interest quantities
#'
#' Provides consistent conversions between effective interest rate, effective
#' discount rate, force of interest, and optional nominal interest rate
#' convertible m-thly.
#'
#' Exactly one of `i`, `d`, or `delta` must be provided.
#'
#' @param i Effective interest rate. May be scalar or vector.
#' @param d Effective discount rate. May be scalar or vector.
#' @param delta Force of interest. May be scalar or vector.
#' @param m Optional positive integer compounding frequency for the nominal
#'   rate convertible m-thly.
#'
#' @return A list with elements `i`, `d`, `delta`, and, if `m` is supplied,
#'   `im` and `m`.
#' @export
#'
#' @examples
#' interest_convert(i = 0.05)
#' interest_convert(i = c(0.03, 0.05, 0.07))
#' interest_convert(d = 0.04761905)
#' interest_convert(delta = log(1.05))
interest_convert <- function(i = NULL, d = NULL, delta = NULL, m = NULL) {
  supplied <- sum(!is.null(i), !is.null(d), !is.null(delta))
  if (supplied != 1) {
    stop("Provide exactly one of i, d, or delta.", call. = FALSE)
  }

  if (!is.null(i)) {
    i <- as.numeric(i)
    if (length(i) == 0L || any(!is.finite(i))) {
      stop("i must contain finite numeric values.", call. = FALSE)
    }
    if (any(i <= -1)) {
      stop("i must be greater than -1.", call. = FALSE)
    }

    d <- i / (1 + i)
    delta <- log(1 + i)

  } else if (!is.null(d)) {
    d <- as.numeric(d)
    if (length(d) == 0L || any(!is.finite(d))) {
      stop("d must contain finite numeric values.", call. = FALSE)
    }
    if (any(d >= 1)) {
      stop("d must be less than 1.", call. = FALSE)
    }
    if (any(d < 0)) {
      stop("d must be nonnegative for standard discounting.", call. = FALSE)
    }

    i <- d / (1 - d)
    delta <- log(1 + i)

  } else {
    delta <- as.numeric(delta)
    if (length(delta) == 0L || any(!is.finite(delta))) {
      stop("delta must contain finite numeric values.", call. = FALSE)
    }

    i <- exp(delta) - 1
    d <- i / (1 + i)
  }

  out <- list(i = i, d = d, delta = delta)

  if (!is.null(m)) {
    m <- as.numeric(m)
    if (length(m) != 1L || !is.finite(m) || m <= 0 ||
        abs(m - round(m)) > 1e-10) {
      stop("m must be a positive integer.", call. = FALSE)
    }

    m <- as.integer(round(m))

    out$im <- m * ((1 + i)^(1 / m) - 1)
    out$m <- m
  }

  out
}

#' Discount factor for compound interest
#'
#' Computes the discount factor \eqn{v^t = (1+i)^{-t}}.
#'
#' @param i Effective interest rate. May be scalar or vector.
#' @param t Time. May be scalar or vector.
#'
#' @return Numeric vector of discount factors.
#' @export
#'
#' @examples
#' discount(0.05, 0:5)
#' discount(c(0.03, 0.05), 1)
discount <- function(i, t) {
  i <- as.numeric(i)
  t <- as.numeric(t)

  if (length(i) == 0L || length(t) == 0L) {
    stop("i and t must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(i)) || any(!is.finite(t))) {
    stop("i and t must contain finite numeric values.", call. = FALSE)
  }
  if (any(i <= -1)) {
    stop("i must be greater than -1.", call. = FALSE)
  }

  it <- .recycle2_interest(i, t, "i", "t")
  i <- it$a
  t <- it$b

  (1 + i)^(-t)
}

#' Present value of cash flows at time 0
#'
#' @param cf Cash flow amounts.
#' @param t Times of cash flows.
#' @param i Effective interest rate.
#'
#' @return Present value at time 0.
#' @export
#'
#' @examples
#' pv_cashflows(c(-100, 60, 60), c(0, 1, 2), i = 0.10)
pv_cashflows <- function(cf, t, i) {
  cf <- as.numeric(cf)
  t <- as.numeric(t)
  i <- as.numeric(i)

  if (length(cf) != length(t)) {
    stop("cf and t must have the same length.", call. = FALSE)
  }
  if (length(cf) == 0L) {
    stop("cf and t must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(cf)) || any(!is.finite(t))) {
    stop("cf and t must contain finite numeric values.", call. = FALSE)
  }
  if (length(i) != 1L || !is.finite(i) || i <= -1) {
    stop("i must be a single effective interest rate greater than -1.", call. = FALSE)
  }

  sum(cf * discount(i, t))
}

#' Solve the yield rate by the equation of value
#'
#' Finds the interest rate `i` such that the present value of the cash flows is 0.
#'
#' @param cf Cash flows.
#' @param t Times.
#' @param interval Two-length numeric vector bracketing the root.
#' @param tol Tolerance passed to `uniroot()`.
#'
#' @return Yield rate `i`.
#' @export
#'
#' @examples
#' solve_yield(c(-100, 60, 60), c(0, 1, 2), interval = c(-0.5, 1))
solve_yield <- function(cf, t, interval = c(-0.99, 1), tol = 1e-10) {
  cf <- as.numeric(cf)
  t <- as.numeric(t)
  interval <- as.numeric(interval)

  if (length(cf) != length(t)) {
    stop("cf and t must have the same length.", call. = FALSE)
  }
  if (length(interval) != 2L || any(!is.finite(interval))) {
    stop("interval must be a finite numeric vector of length 2.", call. = FALSE)
  }
  if (interval[1] <= -1 || interval[1] >= interval[2]) {
    stop("interval must satisfy -1 < interval[1] < interval[2].", call. = FALSE)
  }
  if (length(tol) != 1L || !is.finite(tol) || tol <= 0) {
    stop("tol must be a positive numeric scalar.", call. = FALSE)
  }

  f <- function(i) pv_cashflows(cf, t, i)

  a <- interval[1]
  b <- interval[2]
  fa <- f(a)
  fb <- f(b)

  if (!is.finite(fa) || !is.finite(fb)) {
    stop("NPV not finite at interval endpoints. Adjust interval.", call. = FALSE)
  }
  if (fa == 0) return(a)
  if (fb == 0) return(b)
  if (fa * fb > 0) {
    stop("No sign change in NPV over interval; cannot bracket root. Try a wider interval.",
         call. = FALSE)
  }

  stats::uniroot(f, interval = interval, tol = tol)$root
}

#' Present value of a level annuity-certain
#'
#' Computes the present value of an n-period annuity certain with level
#' payments of 1 per period.
#'
#' @param n Number of payments or periods. May be scalar or vector.
#' @param i Effective interest rate per period. May be scalar or vector.
#' @param due If `TRUE`, annuity-due; otherwise annuity-immediate.
#' @param m Payment frequency per period. `m = 1` means annual.
#' @param cont If `TRUE`, use the continuous payment model.
#'
#' @return Numeric vector of present values.
#' @export
#'
#' @examples
#' annuity_certain(n = 10, i = 0.05)
#' annuity_certain(n = c(5, 10), i = 0.05)
#' annuity_certain(n = 10, i = c(0.03, 0.05))
#' annuity_certain(n = 10, i = 0.05, due = TRUE)
#' annuity_certain(n = 10, i = 0.05, cont = TRUE)
annuity_certain <- function(n, i, due = FALSE, m = 1, cont = FALSE) {
  n <- as.numeric(n)
  i <- as.numeric(i)
  m <- as.numeric(m)

  if (length(n) == 0L || any(!is.finite(n)) || any(n < 0)) {
    stop("n must contain nonnegative finite values.", call. = FALSE)
  }
  if (length(i) == 0L || any(!is.finite(i)) || any(i <= -1)) {
    stop("i must contain effective interest rates greater than -1.", call. = FALSE)
  }
  if (length(m) != 1L || !is.finite(m) || m <= 0 ||
      abs(m - round(m)) > 1e-10) {
    stop("m must be a positive integer.", call. = FALSE)
  }
  if (!is.logical(due) || length(due) != 1L || is.na(due)) {
    stop("due must be TRUE or FALSE.", call. = FALSE)
  }
  if (!is.logical(cont) || length(cont) != 1L || is.na(cont)) {
    stop("cont must be TRUE or FALSE.", call. = FALSE)
  }

  ni <- .recycle2_interest(n, i, "n", "i")
  n <- ni$a
  i <- ni$b
  m <- as.integer(round(m))

  out <- numeric(length(n))

  for (j in seq_along(n)) {
    if (cont) {
      delta <- log(1 + i[j])
      out[j] <- if (abs(delta) < 1e-14) {
        n[j]
      } else {
        (1 - exp(-delta * n[j])) / delta
      }
      next
    }

    nm <- n[j] * m
    if (abs(nm - round(nm)) > 1e-10) {
      stop("Each value of n * m must be an integer for m-thly annuity payments.",
           call. = FALSE)
    }

    nm <- as.integer(round(nm))
    if (nm == 0L) {
      out[j] <- 0
      next
    }

    sub_i <- (1 + i[j])^(1 / m) - 1
    v <- 1 / (1 + sub_i)

    out[j] <- if (!due) {
      (1 / m) * sum(v^(1:nm))
    } else {
      (1 / m) * sum(v^(0:(nm - 1L)))
    }
  }

  out
}

test_that("interest conversions are consistent", {
  out <- interest_convert(i = 0.05)
  expect_equal(out$d, 0.05/1.05, tolerance = 1e-12)
  expect_equal(out$delta, log(1.05), tolerance = 1e-12)
})

test_that("annuity due relationship holds (annual)", {
  a  <- annuity_certain(n = 10, i = 0.05, due = FALSE)
  ad <- annuity_certain(n = 10, i = 0.05, due = TRUE)
  expect_equal(ad, (1 + 0.05) * a, tolerance = 1e-10)
})

test_that("pv_cashflows matches manual discounting", {
  cf <- c(-100, 60, 60)
  t <- c(0, 1, 2)
  i <- 0.10
  expect_equal(pv_cashflows(cf, t, i), sum(cf * (1+i)^(-t)), tolerance = 1e-12)
})

test_that("solve_yield finds a root", {
  cf <- c(-100, 60, 60)
  t <- c(0, 1, 2)
  i_hat <- solve_yield(cf, t, interval = c(-0.5, 1))
  expect_lt(abs(pv_cashflows(cf, t, i_hat)), 1e-8)
})


test_that("discount uses explicit recycling", {
  expect_equal(discount(0.05, 0:2), 1.05^(-(0:2)))
  expect_equal(discount(c(0.03, 0.05), 1), c(1.03, 1.05)^(-1))
  expect_error(discount(c(0.03, 0.04), 1:3), "length 1 or common length")
})

test_that("annuity_certain vectorizes over n", {
  out <- annuity_certain(n = c(5, 10), i = 0.05)
  expected <- c(
    sum((1 / 1.05)^(1:5)),
    sum((1 / 1.05)^(1:10))
  )
  expect_equal(out, expected)
})

test_that("annuity_certain vectorizes over i", {
  out <- annuity_certain(n = 10, i = c(0.03, 0.05))
  expected <- c(
    sum((1 / 1.03)^(1:10)),
    sum((1 / 1.05)^(1:10))
  )
  expect_equal(out, expected)
})

test_that("annuity_certain vectorizes over n and i together", {
  out <- annuity_certain(n = c(5, 10), i = c(0.03, 0.05))
  expected <- c(
    sum((1 / 1.03)^(1:5)),
    sum((1 / 1.05)^(1:10))
  )
  expect_equal(out, expected)
})

test_that("annuity_certain handles continuous vectorized values", {
  out <- annuity_certain(n = c(5, 10), i = 0.05, cont = TRUE)
  delta <- log(1.05)
  expect_equal(out, (1 - exp(-delta * c(5, 10))) / delta)
})

#' Premium, loss, and expense functions
#'
#' Functions for net and gross premiums, present-value-of-loss moments,
#' continuous-payment premium rates, and premiums payable more frequently
#' than annually.
#'
#' The functions include:
#' \itemize{
#'   \item net annual premiums under the equivalence principle,
#'   \item limited-payment premiums,
#'   \item continuous-payment premium rates,
#'   \item fully continuous premium rates,
#'   \item true fractional premiums,
#'   \item present-value-of-loss means and variances,
#'   \item a basic gross premium formula for whole life insurance.
#' }
#'
#' The discrete premium functions may be evaluated from either a life table
#' supplied through \code{tbl} or a parametric survival model supplied through
#' \code{model}. Continuous-payment premium functions are evaluated through
#' the corresponding continuous insurance and annuity functions.
#'
#' Scalar inputs retain their existing behavior. Where mathematically
#' meaningful, numeric inputs may also be vectors. Inputs are evaluated
#' elementwise when they have a common length, and scalar inputs are recycled.
#'
#' @param x Age. May be scalar or vector.
#' @param n Term. May be scalar or vector of nonnegative integers.
#' @param t Premium-paying period. May be scalar or vector of nonnegative integers.
#' @param m Number of payments per year. Must be a positive integer scalar.
#' @param i Effective annual interest rate. May be scalar or vector.
#' @param P Premium amount or premium rate. May be scalar or vector.
#' @param tbl Optional life table object.
#' @param model Optional parametric survival model name.
#' @param ... Additional arguments passed to survival-model functions.
#' @param tol Numerical tolerance for functions that truncate infinite sums.
#' @param k_max Maximum summation horizon for functions that truncate infinite sums.
#' @param benefit Benefit amount. May be scalar or vector.
#' @param first_premium_pct First-year premium expense proportion. May be scalar or vector.
#' @param renewal_premium_pct Renewal premium expense proportion. May be scalar or vector.
#' @param first_policy_exp First-year fixed expense. May be scalar or vector.
#' @param renewal_policy_exp Renewal fixed expense after the first year. May be scalar or vector.
#' @param settlement_exp Settlement expense incurred at benefit payment. May be scalar or vector.
#'
#' @return Numeric vector.
#' @name premium_functions
NULL

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.check_premium_interest <- function(i) {
  i <- as.numeric(i)
  if (length(i) == 0L || any(!is.finite(i)) || any(i <= -1)) {
    stop("i must contain finite values greater than -1.", call. = FALSE)
  }
  i
}

#' @noRd
.check_premium_frequency <- function(m) {
  m <- as.numeric(m)
  if (length(m) != 1L || !is.finite(m) || m <= 0 ||
      abs(m - round(m)) > 1e-10) {
    stop("m must be a positive integer.", call. = FALSE)
  }
  as.integer(round(m))
}

#' @noRd
.check_premium_age <- function(x) {
  x <- as.numeric(x)
  if (length(x) == 0L || any(!is.finite(x)) || any(x < 0)) {
    stop("x must contain nonnegative finite values.", call. = FALSE)
  }
  x
}

#' @noRd
.check_premium_integer <- function(z, name) {
  z <- as.numeric(z)
  if (length(z) == 0L || any(!is.finite(z)) || any(z < 0) ||
      any(abs(z - round(z)) > 1e-10)) {
    stop(sprintf("%s must contain nonnegative integer values.", name),
         call. = FALSE)
  }
  as.integer(round(z))
}

#' @noRd
.check_premium_numeric <- function(z, name) {
  z <- as.numeric(z)
  if (length(z) == 0L || any(!is.finite(z))) {
    stop(sprintf("%s must contain finite numeric values.", name),
         call. = FALSE)
  }
  z
}

#' @noRd
.recycle_premium_vectors <- function(...) {
  values <- list(...)
  lens <- vapply(values, length, integer(1))

  if (length(values) == 0L || any(lens == 0L)) {
    stop("Arguments must have positive length.", call. = FALSE)
  }

  target <- max(lens)
  if (!all(lens %in% c(1L, target))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(values, rep_len, length.out = target)
}

#' @noRd
.validate_premium_source <- function(tbl, model) {
  if (is.null(tbl) && is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }
  invisible(TRUE)
}

#' @noRd
.map_premium <- function(args, FUN) {
  n <- length(args[[1L]])
  vapply(seq_len(n), function(j) {
    current <- lapply(args, `[[`, j)
    do.call(FUN, current)
  }, numeric(1))
}

# -------------------------------------------------------------------------
# Annual net premiums
# -------------------------------------------------------------------------

#' Whole life annual net premium
#'
#' @rdname premium_functions
#' @export
Px <- function(x, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, i = i)

  .map_premium(args, function(x, i) {
    Ax(x, i = i, tbl = tbl, model = model, ...) /
      adotx(x, i = i, tbl = tbl, model = model, ...,
            k_max = 5000, tol = 1e-12)
  })
}

#' Term insurance annual net premium
#'
#' @rdname premium_functions
#' @export
Pxn1 <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    Axn1(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn(x, n, i = i, tbl = tbl, model = model, ...)
  })
}

#' Pure endowment annual net premium
#'
#' @rdname premium_functions
#' @export
PnEx <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    nEx(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn(x, n, i = i, tbl = tbl, model = model, ...)
  })
}

#' Endowment insurance annual net premium
#'
#' @rdname premium_functions
#' @export
Pxn <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    Axn(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn(x, n, i = i, tbl = tbl, model = model, ...)
  })
}

#' Limited-payment whole life annual net premium
#'
#' @rdname premium_functions
#' @export
tPx <- function(x, t, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  t <- .check_premium_integer(t, "t")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, t = t, i = i)

  .map_premium(args, function(x, t, i) {
    Ax(x, i = i, tbl = tbl, model = model, ...) /
      adotxn(x, t, i = i, tbl = tbl, model = model, ...)
  })
}

#' Limited-payment term insurance annual net premium
#'
#' @rdname premium_functions
#' @export
tPxn1 <- function(x, n, t, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  t <- .check_premium_integer(t, "t")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, t = t, i = i)

  if (any(args$t > args$n)) {
    stop("For t-pay n-year contracts, t must satisfy t <= n.", call. = FALSE)
  }

  .map_premium(args, function(x, n, t, i) {
    Axn1(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn(x, t, i = i, tbl = tbl, model = model, ...)
  })
}

#' Limited-payment pure endowment annual net premium
#'
#' @rdname premium_functions
#' @export
tPnEx <- function(x, n, t, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  t <- .check_premium_integer(t, "t")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, t = t, i = i)

  if (any(args$t > args$n)) {
    stop("For t-pay n-year contracts, t must satisfy t <= n.", call. = FALSE)
  }

  .map_premium(args, function(x, n, t, i) {
    nEx(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn(x, t, i = i, tbl = tbl, model = model, ...)
  })
}

#' Limited-payment endowment insurance annual net premium
#'
#' @rdname premium_functions
#' @export
tPxn <- function(x, n, t, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  t <- .check_premium_integer(t, "t")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, t = t, i = i)

  if (any(args$t > args$n)) {
    stop("For t-pay n-year contracts, t must satisfy t <= n.", call. = FALSE)
  }

  .map_premium(args, function(x, n, t, i) {
    Axn(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn(x, t, i = i, tbl = tbl, model = model, ...)
  })
}

#' Deferred insurance annual net premium
#'
#' @rdname premium_functions
#' @export
PnAx <- function(x, n, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    nAx(x, n, i = i, tbl = tbl, model = model, ...) /
      adotx(x, i = i, tbl = tbl, model = model, ...,
            k_max = 5000, tol = 1e-12)
  })
}

#' Limited-payment deferred insurance annual net premium
#'
#' @rdname premium_functions
#' @export
tPnAx <- function(x, n, t, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  t <- .check_premium_integer(t, "t")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, t = t, i = i)

  if (any(args$t > args$n)) {
    stop("For deferred insurance with limited premiums, t must satisfy t <= n.",
         call. = FALSE)
  }

  .map_premium(args, function(x, n, t, i) {
    nAx(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn(x, t, i = i, tbl = tbl, model = model, ...)
  })
}

# -------------------------------------------------------------------------
# Continuous-payment premiums for discrete benefits
# -------------------------------------------------------------------------

#' Continuous-payment premium rate for discrete whole life insurance
#'
#' @rdname premium_functions
#' @export
Pbarx <- function(x, i, model, ..., tol = 1e-10) {
  x <- .check_premium_age(x)
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, i = i)

  .map_premium(args, function(x, i) {
    Ax(x, i = i, model = model, ...) /
      abarx(x, i = i, model = model, ..., tol = tol)
  })
}

#' Continuous-payment premium rate for discrete term insurance
#'
#' @rdname premium_functions
#' @export
Pbarxn1 <- function(x, n, i, model, ...) {
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    Axn1(x, n, i = i, model = model, ...) /
      abarxn(x, n, i = i, model = model, ...)
  })
}

#' Continuous-payment premium rate for discrete endowment insurance
#'
#' @rdname premium_functions
#' @export
Pbarxn <- function(x, n, i, model, ...) {
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    Axn(x, n, i = i, model = model, ...) /
      abarxn(x, n, i = i, model = model, ...)
  })
}

# -------------------------------------------------------------------------
# Fully continuous premium rates
# -------------------------------------------------------------------------

#' Fully continuous premium rate for whole life insurance
#'
#' @rdname premium_functions
#' @export
PbarAbarx <- function(x, i, model, ..., tol = 1e-10) {
  x <- .check_premium_age(x)
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, i = i)

  .map_premium(args, function(x, i) {
    Abarx(x, i = i, model = model, ...) /
      abarx(x, i = i, model = model, ..., tol = tol)
  })
}

#' Fully continuous premium rate for term insurance
#'
#' @rdname premium_functions
#' @export
PbarAbarxn1 <- function(x, n, i, model, ...) {
  x <- .check_premium_age(x)
  n <- .check_premium_numeric(n, "n")
  if (any(n < 0)) stop("n must be nonnegative.", call. = FALSE)
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    Abarxn1(x, n, i = i, model = model, ...) /
      abarxn(x, n, i = i, model = model, ...)
  })
}

#' Fully continuous premium rate for endowment insurance
#'
#' @rdname premium_functions
#' @export
PbarAbarxn <- function(x, n, i, model, ...) {
  x <- .check_premium_age(x)
  n <- .check_premium_numeric(n, "n")
  if (any(n < 0)) stop("n must be nonnegative.", call. = FALSE)
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    Abarxn(x, n, i = i, model = model, ...) /
      abarxn(x, n, i = i, model = model, ...)
  })
}

# -------------------------------------------------------------------------
# True fractional premiums
# -------------------------------------------------------------------------

#' True fractional whole life premium
#'
#' @rdname premium_functions
#' @export
Px_m <- function(x, m, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  m <- .check_premium_frequency(m)
  x <- .check_premium_age(x)
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, i = i)

  .map_premium(args, function(x, i) {
    Ax(x, i = i, tbl = tbl, model = model, ...) /
      adotx_m(x, m = m, i = i, tbl = tbl, model = model, ...)
  })
}

#' True fractional term insurance premium
#'
#' @rdname premium_functions
#' @export
Pxn1_m <- function(x, n, m, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  m <- .check_premium_frequency(m)
  x <- .check_premium_age(x)
  n <- .check_premium_numeric(n, "n")
  if (any(n < 0)) stop("n must be nonnegative.", call. = FALSE)
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    Axn1(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn_m(x, n, m = m, i = i, tbl = tbl, model = model, ...)
  })
}

#' True fractional endowment insurance premium
#'
#' @rdname premium_functions
#' @export
Pxn_m <- function(x, n, m, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  m <- .check_premium_frequency(m)
  x <- .check_premium_age(x)
  n <- .check_premium_numeric(n, "n")
  if (any(n < 0)) stop("n must be nonnegative.", call. = FALSE)
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    Axn(x, n, i = i, tbl = tbl, model = model, ...) /
      adotxn_m(x, n, m = m, i = i, tbl = tbl, model = model, ...)
  })
}

#' True fractional deferred insurance premium
#'
#' @rdname premium_functions
#' @export
PnAx_m <- function(x, n, m, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  m <- .check_premium_frequency(m)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, i = i)

  .map_premium(args, function(x, n, i) {
    nAx(x, n, i = i, tbl = tbl, model = model, ...) /
      adotx_m(x, m = m, i = i, tbl = tbl, model = model, ...)
  })
}

# -------------------------------------------------------------------------
# Present-value-of-loss means and variances
# -------------------------------------------------------------------------

#' Expected present value of loss for whole life insurance
#'
#' @rdname premium_functions
#' @export
EL0x <- function(x, P, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  P <- .check_premium_numeric(P, "P")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, P = P, i = i)

  .map_premium(args, function(x, P, i) {
    Ax(x, i = i, tbl = tbl, model = model, ...) -
      P * adotx(x, i = i, tbl = tbl, model = model, ...)
  })
}

#' Variance of present value of loss for whole life insurance
#'
#' @rdname premium_functions
#' @export
varL0x <- function(x, P, i, tbl = NULL, model = NULL, ...,
                   tol = 1e-12, k_max = 5000) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  P <- .check_premium_numeric(P, "P")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, P = P, i = i)

  .map_premium(args, function(x, P, i) {
    d <- i / (1 + i)
    (1 + P / d)^2 *
      var_Ax(x, i = i, tbl = tbl, model = model, ...,
             tol = tol, k_max = k_max)
  })
}

#' Expected present value of loss for term insurance
#'
#' @rdname premium_functions
#' @export
EL0xn1 <- function(x, n, P, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  P <- .check_premium_numeric(P, "P")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, P = P, i = i)

  .map_premium(args, function(x, n, P, i) {
    Axn1(x, n, i = i, tbl = tbl, model = model, ...) -
      P * adotxn(x, n, i = i, tbl = tbl, model = model, ...)
  })
}

#' Variance of present value of loss for term insurance
#'
#' @rdname premium_functions
#' @export
varL0xn1 <- function(x, n, P, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  P <- .check_premium_numeric(P, "P")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, P = P, i = i)

  .map_premium(args, function(x, n, P, i) {
    d <- i / (1 + i)
    (1 + P / d)^2 *
      var_Axn1(x, n, i = i, tbl = tbl, model = model, ...)
  })
}

#' Expected present value of loss for endowment insurance
#'
#' @rdname premium_functions
#' @export
EL0xn <- function(x, n, P, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  P <- .check_premium_numeric(P, "P")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, P = P, i = i)

  .map_premium(args, function(x, n, P, i) {
    Axn(x, n, i = i, tbl = tbl, model = model, ...) -
      P * adotxn(x, n, i = i, tbl = tbl, model = model, ...)
  })
}

#' Variance of present value of loss for endowment insurance
#'
#' @rdname premium_functions
#' @export
varL0xn <- function(x, n, P, i, tbl = NULL, model = NULL, ...) {
  .validate_premium_source(tbl, model)
  x <- .check_premium_age(x)
  n <- .check_premium_integer(n, "n")
  P <- .check_premium_numeric(P, "P")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, n = n, P = P, i = i)

  .map_premium(args, function(x, n, P, i) {
    d <- i / (1 + i)
    (1 + P / d)^2 *
      var_Axn(x, n, i = i, tbl = tbl, model = model, ...)
  })
}

#' Expected present value of loss for fully continuous whole life insurance
#'
#' @rdname premium_functions
#' @export
EL0barAbarx <- function(x, P, i, model, ..., tol = 1e-10) {
  x <- .check_premium_age(x)
  P <- .check_premium_numeric(P, "P")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, P = P, i = i)

  .map_premium(args, function(x, P, i) {
    Abarx(x, i = i, model = model, ...) -
      P * abarx(x, i = i, model = model, ..., tol = tol)
  })
}

#' Variance of present value of loss for fully continuous whole life insurance
#'
#' @rdname premium_functions
#' @export
varL0barAbarx <- function(x, P, i, model, ...) {
  x <- .check_premium_age(x)
  P <- .check_premium_numeric(P, "P")
  i <- .check_premium_interest(i)
  args <- .recycle_premium_vectors(x = x, P = P, i = i)

  .map_premium(args, function(x, P, i) {
    delta <- log(1 + i)
    (1 + P / delta)^2 *
      var_Abarx(x, i = i, model = model, ...)
  })
}

# -------------------------------------------------------------------------
# Gross premium with expenses: whole life, fully discrete
# -------------------------------------------------------------------------

#' Gross premium for whole life insurance with expenses
#'
#' @rdname premium_functions
#' @export
Gx <- function(x,
               i,
               benefit = 1,
               first_premium_pct = 0,
               renewal_premium_pct = 0,
               first_policy_exp = 0,
               renewal_policy_exp = 0,
               settlement_exp = 0,
               tbl = NULL,
               model = NULL,
               ...) {
  .validate_premium_source(tbl, model)

  x <- .check_premium_age(x)
  i <- .check_premium_interest(i)
  benefit <- .check_premium_numeric(benefit, "benefit")
  first_premium_pct <- .check_premium_numeric(
    first_premium_pct, "first_premium_pct"
  )
  renewal_premium_pct <- .check_premium_numeric(
    renewal_premium_pct, "renewal_premium_pct"
  )
  first_policy_exp <- .check_premium_numeric(
    first_policy_exp, "first_policy_exp"
  )
  renewal_policy_exp <- .check_premium_numeric(
    renewal_policy_exp, "renewal_policy_exp"
  )
  settlement_exp <- .check_premium_numeric(
    settlement_exp, "settlement_exp"
  )

  args <- .recycle_premium_vectors(
    x = x,
    i = i,
    benefit = benefit,
    first_premium_pct = first_premium_pct,
    renewal_premium_pct = renewal_premium_pct,
    first_policy_exp = first_policy_exp,
    renewal_policy_exp = renewal_policy_exp,
    settlement_exp = settlement_exp
  )

  .map_premium(
    args,
    function(x, i, benefit, first_premium_pct, renewal_premium_pct,
             first_policy_exp, renewal_policy_exp, settlement_exp) {
      Ax_val <- Ax(x, i = i, tbl = tbl, model = model, ...)
      ax_val <- ax(x, i = i, tbl = tbl, model = model, ...)
      adotx_val <- adotx(x, i = i, tbl = tbl, model = model, ...)

      numerator <- (benefit + settlement_exp) * Ax_val +
        first_policy_exp + renewal_policy_exp * ax_val

      denominator <- adotx_val -
        first_premium_pct - renewal_premium_pct * ax_val

      if (!is.finite(denominator) || denominator <= 0) {
        stop("Gross premium denominator must be positive.", call. = FALSE)
      }

      numerator / denominator
    }
  )
}

test_that("Px matches Chapter 9 whole life premium identity", {
  x <- 40
  i <- 0.05

  expect_equal(
    Px(x, i = i, model = "uniform", omega = 100),
    Ax(x, i = i, model = "uniform", omega = 100) /
      adotx(x, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  d <- i / (1 + i)

  expect_equal(
    Px(x, i = i, model = "uniform", omega = 100),
    1 / adotx(x, i = i, model = "uniform", omega = 100) - d,
    tolerance = 1e-10
  )
})

test_that("Pxn1, PnEx, and Pxn match Chapter 9 temporary formulas", {
  x <- 40
  n <- 20
  i <- 0.05

  p_term <- Pxn1(x, n, i = i, model = "uniform", omega = 100)
  p_pe   <- PnEx(x, n, i = i, model = "uniform", omega = 100)
  p_end  <- Pxn(x, n, i = i, model = "uniform", omega = 100)

  expect_equal(
    p_term,
    Axn1(x, n, i = i, model = "uniform", omega = 100) /
      adotxn(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    p_pe,
    nEx(x, n, i = i, model = "uniform", omega = 100) /
      adotxn(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    p_end,
    Axn(x, n, i = i, model = "uniform", omega = 100) /
      adotxn(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(p_end, p_term + p_pe, tolerance = 1e-10)
})

test_that("limited-payment premiums use temporary premium annuity denominator", {
  x <- 35
  n <- 20
  t <- 10
  i <- 0.05

  expect_equal(
    tPx(x, t, i = i, model = "uniform", omega = 100),
    Ax(x, i = i, model = "uniform", omega = 100) /
      adotxn(x, t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    tPxn1(x, n, t, i = i, model = "uniform", omega = 100),
    Axn1(x, n, i = i, model = "uniform", omega = 100) /
      adotxn(x, t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    tPnEx(x, n, t, i = i, model = "uniform", omega = 100),
    nEx(x, n, i = i, model = "uniform", omega = 100) /
      adotxn(x, t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    tPxn(x, n, t, i = i, model = "uniform", omega = 100),
    Axn(x, n, i = i, model = "uniform", omega = 100) /
      adotxn(x, t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("deferred insurance premium functions match Chapter 9 identities", {
  x <- 40
  n <- 15
  t <- 10
  i <- 0.05

  expect_equal(
    PnAx(x, n, i = i, model = "uniform", omega = 100),
    nAx(x, n, i = i, model = "uniform", omega = 100) /
      adotx(x, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    tPnAx(x, n, t, i = i, model = "uniform", omega = 100),
    nAx(x, n, i = i, model = "uniform", omega = 100) /
      adotxn(x, t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("continuous payment premium functions match formulas", {
  x <- 40
  n <- 20
  i <- 0.05

  expect_equal(
    Pbarx(x, i = i, model = "uniform", omega = 100),
    Ax(x, i = i, model = "uniform", omega = 100) /
      abarx(x, i = i, model = "uniform", omega = 100),
    tolerance = 1e-9
  )

  expect_equal(
    Pbarxn1(x, n, i = i, model = "uniform", omega = 100),
    Axn1(x, n, i = i, model = "uniform", omega = 100) /
      abarxn(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-9
  )

  expect_equal(
    Pbarxn(x, n, i = i, model = "uniform", omega = 100),
    Axn(x, n, i = i, model = "uniform", omega = 100) /
      abarxn(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-9
  )
})

test_that("fully continuous premium functions match formulas and identity", {
  x <- 40
  n <- 20
  i <- 0.05
  delta <- log(1 + i)

  expect_equal(
    PbarAbarx(x, i = i, model = "uniform", omega = 100),
    Abarx(x, i = i, model = "uniform", omega = 100) /
      abarx(x, i = i, model = "uniform", omega = 100),
    tolerance = 1e-9
  )

  expect_equal(
    PbarAbarx(x, i = i, model = "uniform", omega = 100),
    1 / abarx(x, i = i, model = "uniform", omega = 100) - delta,
    tolerance = 1e-9
  )

  expect_equal(
    PbarAbarxn1(x, n, i = i, model = "uniform", omega = 100),
    Abarxn1(x, n, i = i, model = "uniform", omega = 100) /
      abarxn(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-9
  )

  expect_equal(
    PbarAbarxn(x, n, i = i, model = "uniform", omega = 100),
    Abarxn(x, n, i = i, model = "uniform", omega = 100) /
      abarxn(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-9
  )
})

test_that("true fractional premiums match Chapter 9 formulas", {
  x <- 40
  n <- 20
  i <- 0.05
  m <- 12

  expect_equal(
    Px_m(x, m = m, i = i, model = "uniform", omega = 100),
    Ax(x, i = i, model = "uniform", omega = 100) /
      adotx_m(x, m = m, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    Pxn1_m(x, n, m = m, i = i, model = "uniform", omega = 100),
    Axn1(x, n, i = i, model = "uniform", omega = 100) /
      adotxn_m(x, n, m = m, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    Pxn_m(x, n, m = m, i = i, model = "uniform", omega = 100),
    Axn(x, n, i = i, model = "uniform", omega = 100) /
      adotxn_m(x, n, m = m, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    PnAx_m(x, n, m = m, i = i, model = "uniform", omega = 100),
    nAx(x, n, i = i, model = "uniform", omega = 100) /
      adotx_m(x, m = m, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("expected loss is zero under equivalence-principle premium", {
  x <- 40
  n <- 20
  i <- 0.05

  expect_equal(
    EL0x(x, P = Px(x, i = i, model = "uniform", omega = 100),
         i = i, model = "uniform", omega = 100),
    0,
    tolerance = 1e-10
  )

  expect_equal(
    EL0xn1(x, n, P = Pxn1(x, n, i = i, model = "uniform", omega = 100),
           i = i, model = "uniform", omega = 100),
    0,
    tolerance = 1e-10
  )

  expect_equal(
    EL0xn(x, n, P = Pxn(x, n, i = i, model = "uniform", omega = 100),
          i = i, model = "uniform", omega = 100),
    0,
    tolerance = 1e-10
  )

  expect_equal(
    EL0barAbarx(x, P = PbarAbarx(x, i = i, model = "uniform", omega = 100),
                i = i, model = "uniform", omega = 100),
    0,
    tolerance = 1e-9
  )
})

test_that("loss variances match Chapter 9 formulas", {
  x <- 40
  n <- 20
  i <- 0.05
  d <- i / (1 + i)
  delta <- log(1 + i)

  P_wl <- Px(x, i = i, model = "uniform", omega = 100)
  P_term <- Pxn1(x, n, i = i, model = "uniform", omega = 100)
  P_end <- Pxn(x, n, i = i, model = "uniform", omega = 100)
  P_cont <- PbarAbarx(x, i = i, model = "uniform", omega = 100)

  expect_equal(
    varL0x(x, P = P_wl, i = i, model = "uniform", omega = 100),
    (1 + P_wl / d)^2 * var_Ax(x, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    varL0xn1(x, n, P = P_term, i = i, model = "uniform", omega = 100),
    (1 + P_term / d)^2 * var_Axn1(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    varL0xn(x, n, P = P_end, i = i, model = "uniform", omega = 100),
    (1 + P_end / d)^2 * var_Axn(x, n, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )

  expect_equal(
    varL0barAbarx(x, P = P_cont, i = i, model = "uniform", omega = 100),
    (1 + P_cont / delta)^2 * var_Abarx(x, i = i, model = "uniform", omega = 100),
    tolerance = 1e-9
  )
})

test_that("gross premium whole life matches Example 9.10 formula structure", {
  x <- 40
  i <- 0.05

  Ax_val <- Ax(x, i = i, model = "uniform", omega = 100)
  ax_val <- ax(x, i = i, model = "uniform", omega = 100)
  adotx_val <- adotx(x, i = i, model = "uniform", omega = 100)

  G_manual <- ((1000 + 20) * Ax_val + 10 + 2 * ax_val) /
    (adotx_val - 0.75 - 0.10 * ax_val)

  G_pkg <- Gx(
    x = x,
    i = i,
    benefit = 1000,
    first_premium_pct = 0.75,
    renewal_premium_pct = 0.10,
    first_policy_exp = 10,
    renewal_policy_exp = 2,
    settlement_exp = 20,
    model = "uniform",
    omega = 100
  )

  expect_equal(G_pkg, G_manual, tolerance = 1e-10)
})

test_that("premium functions vectorize over x and n", {
  x <- c(30, 40, 50)
  n <- c(10, 15, 20)
  t <- c(5, 10, 10)
  i <- 0.05

  expect_length(Px(x, i = i, model = "uniform", omega = 100), 3)
  expect_length(Pxn1(x, n, i = i, model = "uniform", omega = 100), 3)
  expect_length(PnEx(x, n, i = i, model = "uniform", omega = 100), 3)
  expect_length(Pxn(x, n, i = i, model = "uniform", omega = 100), 3)
  expect_length(tPxn1(x, n, t, i = i, model = "uniform", omega = 100), 3)
  expect_length(PnAx(x, n, i = i, model = "uniform", omega = 100), 3)
})

test_that("functions reject invalid t-pay structures", {
  expect_error(
    tPxn1(40, n = 10, t = 12, i = 0.05, model = "uniform", omega = 100),
    "t must satisfy t <= n"
  )

  expect_error(
    tPnAx(40, n = 10, t = 12, i = 0.05, model = "uniform", omega = 100),
    "t must satisfy t <= n"
  )
})

test_that("premium functions return finite values in standard cases", {
  i <- 0.05

  vals <- c(
    Px(40, i = i, model = "uniform", omega = 100),
    Pxn1(40, 20, i = i, model = "uniform", omega = 100),
    PnEx(40, 20, i = i, model = "uniform", omega = 100),
    Pxn(40, 20, i = i, model = "uniform", omega = 100),
    PbarAbarx(40, i = i, model = "uniform", omega = 100),
    Px_m(40, m = 12, i = i, model = "uniform", omega = 100)
  )

  expect_true(all(is.finite(vals)))
  expect_true(all(vals > 0))
})




test_that("annual premium functions vectorize over interest rates", {
  out <- Px(
    x = 40,
    i = c(0.03, 0.05),
    model = "uniform",
    omega = 100
  )

  expect_length(out, 2)
  expect_true(all(is.finite(out)))
})

test_that("term premium functions vectorize over x, n, and i", {
  out <- Pxn1(
    x = c(40, 45),
    n = c(10, 15),
    i = c(0.04, 0.05),
    model = "uniform",
    omega = 100
  )

  expect_length(out, 2)
  expect_true(all(is.finite(out)))
})

test_that("limited-payment premium functions vectorize over t", {
  out <- tPxn(
    x = 40,
    n = c(10, 20),
    t = c(5, 10),
    i = 0.05,
    model = "uniform",
    omega = 100
  )

  expect_length(out, 2)
  expect_true(all(is.finite(out)))
})

test_that("limited-payment premiums require t not to exceed n", {
  expect_error(
    tPxn1(
      x = 40,
      n = 10,
      t = 15,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "t must satisfy t <= n"
  )
})

test_that("premium functions reject incompatible vector lengths", {
  expect_error(
    Pxn(
      x = c(40, 45),
      n = c(10, 15, 20),
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "compatible lengths"
  )

  expect_error(
    EL0xn(
      x = c(40, 45),
      n = c(10, 15),
      P = c(0.01, 0.02, 0.03),
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "compatible lengths"
  )
})

test_that("loss functions vectorize over premium amounts", {
  out <- EL0x(
    x = 40,
    P = c(0.01, 0.02),
    i = 0.05,
    model = "uniform",
    omega = 100
  )

  expect_length(out, 2)
  expect_true(all(is.finite(out)))
})

test_that("gross premium function vectorizes expense inputs", {
  out <- Gx(
    x = c(40, 45),
    i = 0.05,
    benefit = c(100000, 150000),
    first_premium_pct = 0.5,
    renewal_premium_pct = 0.05,
    first_policy_exp = c(100, 125),
    renewal_policy_exp = 25,
    settlement_exp = 200,
    model = "uniform",
    omega = 100
  )

  expect_length(out, 2)
  expect_true(all(is.finite(out)))
  expect_true(all(out > 0))
})

test_that("premium functions work with a finite life table", {
  tbl <- life_table(
    x = 0:3,
    lx = c(100000, 80000, 50000, 0)
  )

  out <- Px(
    x = 0,
    i = 0.05,
    tbl = tbl
  )

  expect_length(out, 1)
  expect_true(is.finite(out))
  expect_true(out > 0)
})

test_that("scalar premium calls preserve scalar behavior", {
  out <- Px(
    x = 40,
    i = 0.05,
    model = "uniform",
    omega = 100
  )

  expect_length(out, 1)
  expect_true(is.numeric(out))
  expect_true(is.finite(out))
})

# reserve_premium_functions.R
#
# Net level premium reserve, duration-t loss, and gain-analysis functions.
#
# The exported functions in this file support parametric survival models and,
# where the underlying insurance, annuity, and premium functions permit it,
# life table objects supplied through `tbl`.
#
# Existing exported function names and the traditional argument order are
# preserved for backward compatibility.

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.check_nonneg_integerish <- function(x, name) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(x)) || any(x < 0) || any(abs(x - round(x)) > 1e-10)) {
    stop(name, " must contain nonnegative integer-like values.", call. = FALSE)
  }

  as.integer(round(x))
}

#' @noRd
.check_nonneg_numeric <- function(x, name) {
  x <- as.numeric(x)

  if (length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(x)) || any(x < 0)) {
    stop(name, " must contain nonnegative finite values.", call. = FALSE)
  }

  x
}

#' @noRd
.recycle2_ch10 <- function(a, b, a_name = "a", b_name = "b") {
  if (length(a) == 1L && length(b) > 1L) a <- rep(a, length(b))
  if (length(b) == 1L && length(a) > 1L) b <- rep(b, length(a))

  if (length(a) != length(b)) {
    stop(
      a_name, " and ", b_name,
      " must have the same length, or one must have length 1.",
      call. = FALSE
    )
  }

  list(a = a, b = b)
}

#' @noRd
.recycle3_ch10 <- function(a, b, c,
                           a_name = "a", b_name = "b", c_name = "c") {
  n <- max(length(a), length(b), length(c))

  if (!length(a) %in% c(1L, n)) {
    stop(a_name, " must have length 1 or common length.", call. = FALSE)
  }
  if (!length(b) %in% c(1L, n)) {
    stop(b_name, " must have length 1 or common length.", call. = FALSE)
  }
  if (!length(c) %in% c(1L, n)) {
    stop(c_name, " must have length 1 or common length.", call. = FALSE)
  }

  if (length(a) == 1L) a <- rep(a, n)
  if (length(b) == 1L) b <- rep(b, n)
  if (length(c) == 1L) c <- rep(c, n)

  list(a = a, b = b, c = c)
}

#' @noRd
.check_t_le_n <- function(t, n, allow_equal = TRUE) {
  if (allow_equal) {
    if (any(t > n)) stop("t must satisfy t <= n.", call. = FALSE)
  } else {
    if (any(t >= n)) stop("t must satisfy t < n.", call. = FALSE)
  }
}

#' @noRd
.check_m_scalar_ch10 <- function(m) {
  m <- as.numeric(m)

  if (length(m) != 1L || !is.finite(m) || m <= 0 || abs(m - round(m)) > 1e-10) {
    stop("m must be a positive integer.", call. = FALSE)
  }

  as.integer(round(m))
}

#' @noRd
.check_reserve_source <- function(tbl, model) {
  if (!is.null(tbl) && !is.null(model)) {
    stop("Supply either tbl or model, not both.", call. = FALSE)
  }
  if (is.null(tbl) && is.null(model)) {
    stop("Supply either tbl or model.", call. = FALSE)
  }
  invisible(TRUE)
}

#' @noRd
.check_reserve_interest <- function(i) {
  i <- as.numeric(i)
  if (length(i) != 1L || !is.finite(i) || i <= -1) {
    stop("i must be a single finite number greater than -1.", call. = FALSE)
  }
  i
}

# -------------------------------------------------------------------------
# Annual premium reserves
# -------------------------------------------------------------------------

#' Whole life net level premium reserve
#'
#' Computes the prospective reserve
#' for a whole life insurance with annual premiums:
#' reserve at duration t equals future APV of benefits minus future APV
#' of net premiums.
#'
#' @param x Issue age.
#' @param t Duration.
#' @param i Effective annual interest rate.
#' @param model Optional parametric survival model name.
#' @param ... Additional model parameters.
#' @param tbl Optional life table object.
#'
#' @return A numeric vector of values.
#' @examples
#' tVx(40, t = 10, i = 0.05, model = "uniform", omega = 100)
#' @export
tVx <- function(x, t, i, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  t <- .check_nonneg_integerish(t, "t")

  xt <- .recycle2_ch10(x, t, "x", "t")
  x <- xt$a
  t <- xt$b

  sapply(seq_along(x), function(j) {
    Ax(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...) -
      Px(x = x[j], i = i, tbl = tbl, model = model, ...) *
      adotx(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...)
  })
}

#' Term insurance net level premium reserve
#'
#' Computes the prospective reserve for an n-year term insurance.
#'
#' @param x Issue age.
#' @param n Term in years.
#' @param t Duration.
#' @param i Effective annual interest rate.
#' @param model Optional parametric survival model name.
#' @param ... Additional model parameters.
#' @param tbl Optional life table object.
#'
#' @return A numeric vector of values.
#' @examples
#' tVxn1(40, n = 20, t = 10, i = 0.05, model = "uniform", omega = 100)
#' @export
tVxn1 <- function(x, n, t, i, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  n <- .check_nonneg_integerish(n, "n")
  t <- .check_nonneg_integerish(t, "t")

  xnt <- .recycle3_ch10(x, n, t, "x", "n", "t")
  x <- xnt$a
  n <- xnt$b
  t <- xnt$c

  .check_t_le_n(t, n, allow_equal = TRUE)

  sapply(seq_along(x), function(j) {
    if (t[j] == n[j]) return(0)

    Axn1(x = x[j] + t[j], n = n[j] - t[j], i = i, tbl = tbl, model = model, ...) -
      Pxn1(x = x[j], n = n[j], i = i, tbl = tbl, model = model, ...) *
      adotxn(x = x[j] + t[j], n = n[j] - t[j], i = i, tbl = tbl, model = model, ...)
  })
}

#' Pure endowment net level premium reserve
#'
#' Computes the prospective reserve for an n-year pure endowment.
#'
#' @inheritParams tVxn1
#' @return A numeric vector of values.
#' @examples
#' tVnEx(40, n = 20, t = 10, i = 0.05, model = "uniform", omega = 100)
#' @export
tVnEx <- function(x, n, t, i, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  n <- .check_nonneg_integerish(n, "n")
  t <- .check_nonneg_integerish(t, "t")

  xnt <- .recycle3_ch10(x, n, t, "x", "n", "t")
  x <- xnt$a
  n <- xnt$b
  t <- xnt$c

  .check_t_le_n(t, n, allow_equal = TRUE)

  sapply(seq_along(x), function(j) {
    if (t[j] == n[j]) return(1)

    nEx(x = x[j] + t[j], n = n[j] - t[j], i = i, tbl = tbl, model = model, ...) -
      PnEx(x = x[j], n = n[j], i = i, tbl = tbl, model = model, ...) *
      adotxn(x = x[j] + t[j], n = n[j] - t[j], i = i, tbl = tbl, model = model, ...)
  })
}

#' Endowment insurance net level premium reserve
#'
#' Computes the prospective reserve for an n-year endowment insurance.
#'
#' @inheritParams tVxn1
#' @return A numeric vector of values.
#' @examples
#' tVxn(40, n = 20, t = 10, i = 0.05, model = "uniform", omega = 100)
#' @export
tVxn <- function(x, n, t, i, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  n <- .check_nonneg_integerish(n, "n")
  t <- .check_nonneg_integerish(t, "t")

  xnt <- .recycle3_ch10(x, n, t, "x", "n", "t")
  x <- xnt$a
  n <- xnt$b
  t <- xnt$c

  .check_t_le_n(t, n, allow_equal = TRUE)

  sapply(seq_along(x), function(j) {
    if (t[j] == n[j]) return(1)

    Axn(x = x[j] + t[j], n = n[j] - t[j], i = i, tbl = tbl, model = model, ...) -
      Pxn(x = x[j], n = n[j], i = i, tbl = tbl, model = model, ...) *
      adotxn(x = x[j] + t[j], n = n[j] - t[j], i = i, tbl = tbl, model = model, ...)
  })
}

#' h-pay whole life net level premium reserve
#'
#' Computes the prospective reserve for an h-pay whole life policy.
#'
#' @param h Premium-paying period in years.
#' @inheritParams tVx
#'
#' @return A numeric vector of values.
#' @examples
#' htVx(40, h = 10, t = 5, i = 0.05, model = "uniform", omega = 100)
#' @export
htVx <- function(x, h, t, i, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  h <- .check_nonneg_integerish(h, "h")
  t <- .check_nonneg_integerish(t, "t")

  xht <- .recycle3_ch10(x, h, t, "x", "h", "t")
  x <- xht$a
  h <- xht$b
  t <- xht$c

  sapply(seq_along(x), function(j) {
    if (t[j] < h[j]) {
      Ax(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...) -
        tPx(x = x[j], t = h[j], i = i, tbl = tbl, model = model, ...) *
        adotxn(x = x[j] + t[j], n = h[j] - t[j], i = i, tbl = tbl, model = model, ...)
    } else {
      Ax(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...)
    }
  })
}

# -------------------------------------------------------------------------
# Duration-t loss moments
# -------------------------------------------------------------------------

#' Mean present value of loss at duration t for whole life insurance
#'
#' Computes the conditional mean
#' \eqn{E[{}_tL_x \mid K_x \ge t]} for a fully discrete whole life insurance.
#'
#' Under the equivalence-principle premium, this equals the prospective
#' reserve \eqn{{}_tV_x}.
#'
#' @param x Issue age.
#' @param t Duration.
#' @param i Effective annual interest rate.
#' @param P Annual premium.
#' @param model Optional parametric survival model name.
#' @param ... Additional model parameters.
#' @param tbl Optional life table object.
#'
#' @return A numeric vector of values.
#' @examples
#' prem <- Px(40, i = 0.05, model = "uniform", omega = 100)
#' ELtx(40, t = 10, i = 0.05, P = prem, model = "uniform", omega = 100)
#' @export
ELtx <- function(x, t, i, P, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  t <- .check_nonneg_integerish(t, "t")
  P <- .check_nonneg_numeric(P, "P")

  xtp <- .recycle3_ch10(x, t, P, "x", "t", "P")
  x <- xtp$a
  t <- xtp$b
  P <- xtp$c

  sapply(seq_along(x), function(j) {
    Ax(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...) -
      P[j] * adotx(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...)
  })
}

#' Variance of present value of loss at duration t for whole life insurance
#'
#' Computes the conditional variance
#' \eqn{\mathrm{Var}({}_tL_x \mid K_x \ge t)}
#' for a fully discrete whole life insurance.
#'
#' @inheritParams ELtx
#' @return A numeric vector of values.
#' @examples
#' prem <- Px(40, i = 0.05, model = "uniform", omega = 100)
#' varLtx(40, t = 10, i = 0.05, P = prem, model = "uniform", omega = 100)
#' @export
varLtx <- function(x, t, i, P, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  t <- .check_nonneg_integerish(t, "t")
  P <- .check_nonneg_numeric(P, "P")

  xtp <- .recycle3_ch10(x, t, P, "x", "t", "P")
  x <- xtp$a
  t <- xtp$b
  P <- xtp$c

  d <- i / (1 + i)

  sapply(seq_along(x), function(j) {
    (1 + P[j] / d)^2 *
      (A2x(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...) -
         Ax(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...)^2)
  })
}

# -------------------------------------------------------------------------
# Continuous and m-thly premium reserves
# -------------------------------------------------------------------------

#' Whole life reserve with continuous premiums
#'
#' Computes the reserve for a discrete whole life insurance
#' funded by continuous premiums.
#'
#' @inheritParams tVx
#' @return A numeric vector of values.
#' @examples
#' tVbarx(40, t = 10, i = 0.05, model = "uniform", omega = 100)
#' @export
tVbarx <- function(x, t, i, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  t <- .check_nonneg_integerish(t, "t")

  xt <- .recycle2_ch10(x, t, "x", "t")
  x <- xt$a
  t <- xt$b

  sapply(seq_along(x), function(j) {
    Ax(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...) -
      Pbarx(x = x[j], i = i, tbl = tbl, model = model, ...) *
      abarx(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...)
  })
}

#' Fully continuous whole life reserve
#'
#' Computes the reserve for a whole life insurance with
#' continuous premiums and immediate payment of claims.
#'
#' In the fully continuous setting, reserve time \eqn{t} may be any
#' nonnegative real value.
#'
#' @param x Issue age.
#' @param t Duration, allowed to be any nonnegative numeric value.
#' @param i Effective annual interest rate.
#' @param model Optional parametric survival model name.
#' @param ... Additional model parameters.
#' @param tbl Optional life table object.
#'
#' @return A numeric vector of values.
#' @examples
#' tVbarAbarx(40, t = 10, i = 0.05, model = "uniform", omega = 100)
#' tVbarAbarx(40, t = c(19, 19.25, 19.5, 19.75, 20), i = 0.06,
#'            model = "uniform", omega = 100)
#' @export
tVbarAbarx <- function(x, t, i, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  t <- .check_nonneg_numeric(t, "t")

  xt <- .recycle2_ch10(x, t, "x", "t")
  x <- xt$a
  t <- xt$b

  sapply(seq_along(x), function(j) {
    Abarx(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...) -
      PbarAbarx(x = x[j], i = i, tbl = tbl, model = model, ...) *
      abarx(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...)
  })
}

#' Whole life reserve with m-thly premiums
#'
#' Computes the reserve for a whole life insurance funded
#' by true m-thly premiums.
#'
#' @param m Number of premium payments per year.
#' @inheritParams tVx
#'
#' @return A numeric vector of values.
#' @examples
#' tVx_m(40, t = 10, m = 12, i = 0.05, model = "uniform", omega = 100)
#' @export
tVx_m <- function(x, t, m, i, model = NULL, ..., tbl = NULL) {
  .check_reserve_source(tbl, model)
  i <- .check_reserve_interest(i)
  x <- .check_nonneg_numeric(x, "x")
  t <- .check_nonneg_integerish(t, "t")
  m <- .check_m_scalar_ch10(m)

  xt <- .recycle2_ch10(x, t, "x", "t")
  x <- xt$a
  t <- xt$b

  sapply(seq_along(x), function(j) {
    Ax(x = x[j] + t[j], i = i, tbl = tbl, model = model, ...) -
      Px_m(x = x[j], m = m, i = i, tbl = tbl, model = model, ...) *
      adotx_m(x = x[j] + t[j], m = m, i = i, tbl = tbl, model = model, ...)
  })
}

# -------------------------------------------------------------------------
# Gain / loss analysis
# -------------------------------------------------------------------------

#' Total gain for a discrete insurance contract
#'
#' Computes the total gain:
#' amount on hand at year-end minus amount required.
#'
#' @param Vt Reserve at duration t.
#' @param Vt1 Reserve at duration t+1.
#' @param P Net premium for the year.
#' @param i_actual Actual annual effective interest rate.
#' @param q_actual Actual mortality rate for the year.
#' @param B Benefit amount. Defaults to 1.
#'
#' @return A numeric vector of values.
#' @examples
#' GT_disc(Vt = 0.1, Vt1 = 0.11, P = 0.02, i_actual = 0.05, q_actual = 0.01)
#' @export
GT_disc <- function(Vt, Vt1, P, i_actual, q_actual, B = 1) {
  Vt <- .check_nonneg_numeric(Vt, "Vt")
  Vt1 <- .check_nonneg_numeric(Vt1, "Vt1")
  P <- .check_nonneg_numeric(P, "P")
  i_actual <- as.numeric(i_actual)
  q_actual <- as.numeric(q_actual)
  B <- .check_nonneg_numeric(B, "B")

  n <- max(length(Vt), length(Vt1), length(P), length(i_actual), length(q_actual), length(B))

  rep_len_checked <- function(z, nm) {
    if (!length(z) %in% c(1L, n)) stop(nm, " must have length 1 or common length.", call. = FALSE)
    rep(z, length.out = n)
  }

  Vt <- rep_len_checked(Vt, "Vt")
  Vt1 <- rep_len_checked(Vt1, "Vt1")
  P <- rep_len_checked(P, "P")
  i_actual <- rep_len_checked(i_actual, "i_actual")
  q_actual <- rep_len_checked(q_actual, "q_actual")
  B <- rep_len_checked(B, "B")

  if (any(!is.finite(i_actual)) || any(i_actual <= -1)) {
    stop("i_actual must be finite and greater than -1.", call. = FALSE)
  }
  if (any(!is.finite(q_actual)) || any(q_actual < 0) || any(q_actual > 1)) {
    stop("q_actual must lie in [0, 1].", call. = FALSE)
  }

  (Vt + P) * (1 + i_actual) - (q_actual * B + (1 - q_actual) * Vt1)
}

#' Mortality gain for a discrete insurance contract
#'
#' @param i_assumed Assumed annual effective interest rate.
#' @inheritParams GT_disc
#'
#' @return A numeric vector of values.
#' @examples
#' GM_disc(Vt = 0.1, Vt1 = 0.11, P = 0.02, i_assumed = 0.04, q_actual = 0.01)
#' @export
GM_disc <- function(Vt, Vt1, P, i_assumed, q_actual, B = 1) {
  GT_disc(
    Vt = Vt,
    Vt1 = Vt1,
    P = P,
    i_actual = i_assumed,
    q_actual = q_actual,
    B = B
  )
}

#' Interest gain for a discrete insurance contract
#'
#' @param i_actual Actual annual effective interest rate.
#' @param q_assumed Assumed mortality rate for the year.
#' @inheritParams GT_disc
#'
#' @return A numeric vector of values.
#' @examples
#' GI_disc(Vt = 0.1, Vt1 = 0.11, P = 0.02, i_actual = 0.05, q_assumed = 0.01)
#' @export
GI_disc <- function(Vt, Vt1, P, i_actual, q_assumed, B = 1) {
  GT_disc(
    Vt = Vt,
    Vt1 = Vt1,
    P = P,
    i_actual = i_actual,
    q_actual = q_assumed,
    B = B
  )
}

test_that("whole life reserve is zero at issue", {
  expect_equal(
    tVx(40, t = 0, i = 0.05, model = "uniform", omega = 100),
    0,
    tolerance = 1e-10
  )
})

test_that("whole life reserve matches alternative Chapter 10 formulas", {
  x <- 40
  t <- 10
  i <- 0.05

  v1 <- tVx(x, t = t, i = i, model = "uniform", omega = 100)

  v2 <- 1 - adotx(x + t, i = i, model = "uniform", omega = 100) /
    adotx(x, i = i, model = "uniform", omega = 100)

  v3 <- (Px(x + t, i = i, model = "uniform", omega = 100) -
           Px(x, i = i, model = "uniform", omega = 100)) *
    adotx(x + t, i = i, model = "uniform", omega = 100)

  v4 <- (Ax(x + t, i = i, model = "uniform", omega = 100) -
           Ax(x, i = i, model = "uniform", omega = 100)) /
    (1 - Ax(x, i = i, model = "uniform", omega = 100))

  expect_equal(v1, v2, tolerance = 1e-10)
  expect_equal(v1, v3, tolerance = 1e-10)
  expect_equal(v1, v4, tolerance = 1e-10)
})

test_that("term reserve is zero at expiry", {
  expect_equal(
    tVxn1(40, n = 20, t = 20, i = 0.05, model = "uniform", omega = 100),
    0,
    tolerance = 1e-10
  )
})

test_that("pure endowment reserve is one at maturity", {
  expect_equal(
    tVnEx(40, n = 20, t = 20, i = 0.05, model = "uniform", omega = 100),
    1,
    tolerance = 1e-10
  )
})

test_that("endowment reserve is one at maturity", {
  expect_equal(
    tVxn(40, n = 20, t = 20, i = 0.05, model = "uniform", omega = 100),
    1,
    tolerance = 1e-10
  )
})

test_that("endowment reserve equals term plus pure endowment reserve", {
  val_endow <- tVxn(40, n = 20, t = 10, i = 0.05, model = "uniform", omega = 100)
  val_term  <- tVxn1(40, n = 20, t = 10, i = 0.05, model = "uniform", omega = 100)
  val_pe    <- tVnEx(40, n = 20, t = 10, i = 0.05, model = "uniform", omega = 100)

  expect_equal(val_endow, val_term + val_pe, tolerance = 1e-10)
})

test_that("limited pay whole life switches to benefit-only reserve after premium period", {
  x <- 40
  h <- 10
  i <- 0.05

  expect_equal(
    htVx(x, h = h, t = 12, i = i, model = "uniform", omega = 100),
    Ax(x + 12, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("continuous premium reserve matches Chapter 10 identity", {
  x <- 40
  t <- 10
  i <- 0.05

  val <- tVbarAbarx(x, t = t, i = i, model = "uniform", omega = 100)
  rhs <- 1 - abarx(x + t, i = i, model = "uniform", omega = 100) /
    abarx(x, i = i, model = "uniform", omega = 100)

  expect_equal(val, rhs, tolerance = 1e-10)
})

test_that("m-thly reserve is zero at issue", {
  expect_equal(
    tVx_m(40, t = 0, m = 12, i = 0.05, model = "uniform", omega = 100),
    0,
    tolerance = 1e-10
  )
})

test_that("gain decomposition holds for discrete insurance", {
  Vt <- 0.085697
  Vt1 <- 0.096115
  P <- 0.008013
  i_actual <- 0.065
  i_assumed <- 0.06
  q_actual <- 0.00300
  q_assumed <- 0.00356

  gt <- GT_disc(Vt, Vt1, P, i_actual = i_actual, q_actual = q_actual, B = 1)
  gm <- GM_disc(Vt, Vt1, P, i_assumed = i_assumed, q_actual = q_actual, B = 1)
  gi <- GI_disc(Vt, Vt1, P, i_actual = i_actual, q_assumed = q_assumed, B = 1)

  expect_lt(abs((gm + gi) - gt), 5e-7)
})

test_that("reserve functions vectorize over x, n, and t", {
  out1 <- tVx(c(40, 45), t = c(5, 10), i = 0.05, model = "uniform", omega = 100)
  out2 <- tVxn1(c(40, 45), n = 20, t = c(5, 10), i = 0.05, model = "uniform", omega = 100)
  out3 <- tVxn(c(40, 45), n = c(15, 20), t = c(5, 10), i = 0.05, model = "uniform", omega = 100)

  expect_length(out1, 2)
  expect_length(out2, 2)
  expect_length(out3, 2)
})

test_that("reserve functions reject invalid durations", {
  expect_error(
    tVxn1(40, n = 10, t = 11, i = 0.05, model = "uniform", omega = 100),
    "t must satisfy t <= n"
  )

  expect_error(
    tVxn(40, n = 10, t = -1, i = 0.05, model = "uniform", omega = 100),
    "nonnegative integer-like"
  )
})


test_that("ELtx equals tVx when premium is the equivalence-principle premium", {
  x <- 40
  t <- 10
  i <- 0.05

  premium <- Px(x, i = i, model = "uniform", omega = 100)

  expect_equal(
    ELtx(x, t = t, i = i, P = premium, model = "uniform", omega = 100),
    tVx(x, t = t, i = i, model = "uniform", omega = 100),
    tolerance = 1e-10
  )
})

test_that("varLtx matches Chapter 10 variance formula", {
  x <- 40
  t <- 10
  i <- 0.05

  premium <- Px(x, i = i, model = "uniform", omega = 100)
  d <- i / (1 + i)

  rhs <- (1 + premium / d)^2 *
    (A2x(x + t, i = i, model = "uniform", omega = 100) -
       Ax(x + t, i = i, model = "uniform", omega = 100)^2)

  expect_equal(
    varLtx(x, t = t, i = i, P = premium, model = "uniform", omega = 100),
    rhs,
    tolerance = 1e-10
  )
})

test_that("ELtx and varLtx vectorize over x, t, and P", {
  i <- 0.05
  x <- c(40, 45)
  t <- c(5, 10)
  P <- c(
    Px(40, i = i, model = "uniform", omega = 100),
    Px(45, i = i, model = "uniform", omega = 100)
  )

  out_mean <- ELtx(x, t = t, i = i, P = P, model = "uniform", omega = 100)
  out_var  <- varLtx(x, t = t, i = i, P = P, model = "uniform", omega = 100)

  expect_length(out_mean, 2)
  expect_length(out_var, 2)
})

test_that("ELtx and varLtx reject invalid inputs", {
  premium <- Px(40, i = 0.05, model = "uniform", omega = 100)

  expect_error(
    ELtx(40, t = -1, i = 0.05, P = premium, model = "uniform", omega = 100),
    "nonnegative integer-like"
  )

  expect_error(
    varLtx(40, t = 5, i = 0.05, P = -0.01, model = "uniform", omega = 100),
    "nonnegative finite"
  )
})

testthat::test_that("annual reserve functions work with life tables", {
  tbl <- life_table(
    x = 0:4,
    lx = c(100000, 80000, 50000, 20000, 0)
  )

  out <- tVx(
    x = 0,
    t = 1,
    i = 0.05,
    tbl = tbl
  )

  testthat::expect_length(out, 1)
  testthat::expect_true(is.finite(out))
})


testthat::test_that("term and endowment reserves have correct maturity values", {
  testthat::expect_equal(
    tVxn1(
      x = 40,
      n = 20,
      t = 20,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    0
  )

  testthat::expect_equal(
    tVnEx(
      x = 40,
      n = 20,
      t = 20,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    1
  )

  testthat::expect_equal(
    tVxn(
      x = 40,
      n = 20,
      t = 20,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    1
  )
})


testthat::test_that("reserve functions vectorize over ages and durations", {
  out <- tVx(
    x = c(40, 45),
    t = c(5, 10),
    i = 0.05,
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("reserve functions recycle scalar arguments", {
  out <- tVxn(
    x = c(40, 45),
    n = 20,
    t = c(5, 10),
    i = 0.05,
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("reserve functions reject incompatible lengths", {
  testthat::expect_error(
    tVxn1(
      x = c(40, 45),
      n = c(10, 15, 20),
      t = 5,
      i = 0.05,
      model = "uniform",
      omega = 100
    ),
    "length 1.*common length"
  )
})


testthat::test_that("reserve functions require a mortality basis", {
  testthat::expect_error(
    tVx(
      x = 40,
      t = 10,
      i = 0.05
    ),
    "Supply either tbl or model"
  )
})


testthat::test_that("reserve functions reject two mortality bases", {
  tbl <- life_table(
    x = 0:2,
    lx = c(100000, 50000, 0)
  )

  testthat::expect_error(
    tVx(
      x = 0,
      t = 1,
      i = 0.05,
      tbl = tbl,
      model = "uniform",
      omega = 100
    ),
    "not both"
  )
})


testthat::test_that("reserve functions preserve positional model syntax", {
  named_result <- tVx(
    x = 40,
    t = 10,
    i = 0.05,
    model = "uniform",
    omega = 100
  )

  positional_result <- tVx(
    40,
    10,
    0.05,
    "uniform",
    omega = 100
  )

  testthat::expect_equal(
    positional_result,
    named_result,
    tolerance = 1e-12
  )
})


testthat::test_that("duration-t loss functions support vectors", {
  out <- ELtx(
    x = c(40, 45),
    t = c(5, 10),
    i = 0.05,
    P = c(0.02, 0.025),
    model = "uniform",
    omega = 100
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})


testthat::test_that("gain functions vectorize consistently", {
  out <- GT_disc(
    Vt = c(0.10, 0.15),
    Vt1 = c(0.11, 0.16),
    P = 0.02,
    i_actual = c(0.04, 0.05),
    q_actual = 0.01,
    B = 1
  )

  testthat::expect_length(out, 2)
  testthat::expect_true(all(is.finite(out)))
})

# =========================================================
# Variable-interest models
# =========================================================

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

#' @noRd
.variable_interest_check_numeric <- function(x, name, positive_length = TRUE) {
  if (!is.numeric(x)) {
    stop(name, " must be numeric.", call. = FALSE)
  }
  if (positive_length && length(x) == 0L) {
    stop(name, " must have positive length.", call. = FALSE)
  }
  if (any(!is.finite(x))) {
    stop(name, " must contain finite values.", call. = FALSE)
  }
  as.numeric(x)
}

#' @noRd
.variable_interest_check_probability <- function(x, name) {
  x <- .variable_interest_check_numeric(x, name)
  if (any(x < 0 | x > 1)) {
    stop(name, " must contain values in [0, 1].", call. = FALSE)
  }
  x
}

#' @noRd
.variable_interest_check_rate <- function(x, name) {
  x <- .variable_interest_check_numeric(x, name)
  if (any(x <= -1)) {
    stop(name, " must contain values greater than -1.", call. = FALSE)
  }
  x
}

#' @noRd
.variable_interest_check_nonnegative <- function(x, name) {
  x <- .variable_interest_check_numeric(x, name)
  if (any(x < 0)) {
    stop(name, " must contain nonnegative values.", call. = FALSE)
  }
  x
}

#' @noRd
.variable_interest_check_scalar <- function(
    x,
    name,
    lower = -Inf,
    strict_lower = FALSE) {
  x <- .variable_interest_check_numeric(x, name)

  if (length(x) != 1L) {
    stop(name, " must be a numeric scalar.", call. = FALSE)
  }

  invalid <- if (strict_lower) x <= lower else x < lower
  if (invalid) {
    comparator <- if (strict_lower) "greater than" else "at least"
    stop(name, " must be ", comparator, " ", lower, ".", call. = FALSE)
  }

  x
}

#' @noRd
.variable_interest_check_integer <- function(
    x,
    name,
    positive = FALSE) {
  x <- .variable_interest_check_scalar(
    x,
    name,
    lower = 0,
    strict_lower = positive
  )

  if (abs(x - round(x)) > 1e-10) {
    descriptor <- if (positive) "positive" else "nonnegative"
    stop(name, " must be a ", descriptor, " integer.", call. = FALSE)
  }

  as.integer(round(x))
}

#' @noRd
.variable_interest_check_same_length <- function(x, y, x_name, y_name) {
  if (length(x) != length(y)) {
    stop(x_name, " and ", y_name, " must have the same length.", call. = FALSE)
  }
  invisible(TRUE)
}

#' @noRd
.variable_interest_survival_start <- function(qx) {
  c(1, cumprod(1 - qx))[seq_along(qx)]
}

#' @noRd
.variable_interest_survival_end <- function(qx) {
  cumprod(1 - qx)
}

#' @noRd
.variable_interest_benefit_scalar <- function(benefit) {
  benefit <- .variable_interest_check_nonnegative(benefit, "benefit")
  if (length(benefit) != 1L) {
    stop("benefit must be a numeric scalar.", call. = FALSE)
  }
  benefit
}

# -------------------------------------------------------------------------
# Variable-interest discount factors
# -------------------------------------------------------------------------

#' Discount factors under variable annual interest rates
#'
#' Computes cumulative discount factors for a sequence of annual effective
#' interest rates:
#' \deqn{
#' v_t = \prod_{k=1}^{t}(1+i_k)^{-1},
#' \qquad t=1,\ldots,n.
#' }
#'
#' @param i Numeric vector of annual effective interest rates. Each value must
#'   be greater than \code{-1}.
#'
#' @return A numeric vector of cumulative discount factors with the same
#'   length as \code{i}.
#'
#' @examples
#' vt_var(c(0.06, 0.07, 0.08))
#' vt_var(c(-0.01, 0.02, 0.03))
#'
#' @export
vt_var <- function(i) {
  i <- .variable_interest_check_rate(i, "i")
  cumprod(1 / (1 + i))
}


# -------------------------------------------------------------------------
# APVs under a sequence of annual effective interest rates
# -------------------------------------------------------------------------

#' Actuarial present values under variable annual interest rates
#'
#' Computes life-contingent actuarial present values using a specified
#' sequence of annual effective interest rates.
#'
#' The vectors \code{qx} and \code{i} represent one valuation scenario and
#' must have the same positive length.
#'
#' \code{nEx_var()} computes a pure endowment.
#'
#' \code{Axn1_var()} computes term insurance payable at the end of the year
#' of death.
#'
#' \code{Axn_var()} computes endowment insurance.
#'
#' \code{axn_var()} computes a temporary annuity-immediate or
#' annuity-due.
#'
#' @param qx Numeric vector of one-year mortality probabilities.
#' @param i Numeric vector of annual effective interest rates. Each value
#'   must be greater than \code{-1}.
#' @param benefit Nonnegative scalar benefit or annuity payment amount.
#' @param type Character string equal to \code{"immediate"} or
#'   \code{"due"}.
#'
#' @return A numeric scalar.
#'
#' @details
#' Each year's payment is discounted using the cumulative product of the
#' annual effective interest rates supplied in \code{i}.
#'
#' The pure endowment is
#' \deqn{
#' {}_nE
#' =
#' {}_np_x\,v_n,
#' }
#' where \eqn{v_n} is the cumulative discount factor implied by the sequence
#' of annual effective interest rates.
#'
#' Term insurance is obtained by discounting each possible death benefit
#' using the cumulative discount factor applicable to its payment year.
#'
#' Endowment insurance equals the sum of the corresponding term insurance
#' and pure endowment.
#'
#' Temporary annuities discount each payment using the cumulative discount
#' factors derived from the interest-rate sequence.
#'
#' @examples
#' qx <- c(0.03, 0.04, 0.05, 0.06, 0.07)
#' rates <- c(0.06, 0.07, 0.08, 0.09, 0.10)
#'
#' nEx_var(qx, rates, benefit = 1000)
#' Axn1_var(qx, rates)
#' Axn_var(qx, rates)
#' axn_var(qx, rates, type = "due")
#'
#' @name variable_interest_apvs
#' @rdname variable_interest_apvs
#' @export
nEx_var <- function(qx, i, benefit = 1) {
  qx <- .variable_interest_check_probability(qx, "qx")
  i <- .variable_interest_check_rate(i, "i")
  benefit <- .variable_interest_benefit_scalar(benefit)

  .variable_interest_check_same_length(qx, i, "qx", "i")

  benefit * prod(1 - qx) * tail(vt_var(i), 1L)
}

#' @rdname variable_interest_apvs
#' @export
Axn1_var <- function(qx, i, benefit = 1) {
  qx <- .variable_interest_check_probability(qx, "qx")
  i <- .variable_interest_check_rate(i, "i")
  benefit <- .variable_interest_benefit_scalar(benefit)

  .variable_interest_check_same_length(qx, i, "qx", "i")

  discount <- vt_var(i)
  survival_start <- .variable_interest_survival_start(qx)

  benefit * sum(discount * survival_start * qx)
}

#' @rdname variable_interest_apvs
#' @export
Axn_var <- function(qx, i, benefit = 1) {
  qx <- .variable_interest_check_probability(qx, "qx")
  i <- .variable_interest_check_rate(i, "i")
  benefit <- .variable_interest_benefit_scalar(benefit)

  .variable_interest_check_same_length(qx, i, "qx", "i")

  Axn1_var(qx = qx, i = i, benefit = benefit) +
    nEx_var(qx = qx, i = i, benefit = benefit)
}

#' @rdname variable_interest_apvs
#' @export
axn_var <- function(qx, i, type = c("immediate", "due"), benefit = 1) {
  type <- match.arg(type)

  qx <- .variable_interest_check_probability(qx, "qx")
  i <- .variable_interest_check_rate(i, "i")
  benefit <- .variable_interest_benefit_scalar(benefit)

  .variable_interest_check_same_length(qx, i, "qx", "i")

  discount <- vt_var(i)

  if (type == "immediate") {
    survival_end <- .variable_interest_survival_end(qx)
    return(benefit * sum(discount * survival_end))
  }

  survival_start <- .variable_interest_survival_start(qx)
  discount_due <- c(1, head(discount, -1L))

  benefit * sum(discount_due * survival_start)
}

# -------------------------------------------------------------------------
# APVs under spot rates
# -------------------------------------------------------------------------

#' Actuarial present values under spot rates
#'
#' Computes life-contingent actuarial present values using annual effective
#' spot rates by maturity.
#'
#' If \eqn{z_t} denotes the annual effective spot rate for maturity
#' \eqn{t}, the corresponding discount factor is
#' \deqn{(1+z_t)^{-t}.}
#'
#' \code{nEx_spot()} computes a pure endowment.
#'
#' \code{Axn1_spot()} computes term insurance payable at the end of the year
#' of death.
#'
#' \code{Axn_spot()} computes endowment insurance.
#'
#' \code{axn_spot()} computes a temporary annuity-immediate or
#' annuity-due.
#'
#' @param qx Numeric vector of one-year mortality probabilities.
#' @param z Numeric vector of annual effective spot rates for maturities
#'   \code{1, \dots, n}. Each value must be greater than \code{-1}.
#' @param benefit Nonnegative scalar benefit or annuity payment amount.
#' @param type Character string equal to \code{"immediate"} or
#'   \code{"due"}.
#'
#' @return A numeric scalar.
#'
#' @details
#' Each payment is discounted using the spot rate corresponding to its
#' maturity rather than a single level interest rate.
#'
#' The pure endowment is
#' \deqn{
#' {}_nE
#' =
#' {}_np_x(1+z_n)^{-n}.
#' }
#'
#' Term insurance is obtained by discounting each death benefit using the
#' spot rate corresponding to its payment year.
#'
#' Endowment insurance equals the sum of the corresponding term insurance
#' and pure endowment.
#'
#' Temporary annuities discount each payment using the spot rate for its
#' payment time.
#'
#' @examples
#' qx <- c(0.02, 0.03, 0.04, 0.05, 0.06)
#' spot <- c(0.03, 0.04, 0.05, 0.06, 0.07)
#'
#' nEx_spot(qx, spot, benefit = 1000)
#' Axn1_spot(qx, spot)
#' Axn_spot(qx, spot)
#' axn_spot(qx, spot, type = "due")
#'
#' @name spot_interest_apvs
#' @rdname spot_interest_apvs
#' @export
nEx_spot <- function(qx, z, benefit = 1) {
  qx <- .variable_interest_check_probability(qx, "qx")
  z <- .variable_interest_check_rate(z, "z")
  benefit <- .variable_interest_benefit_scalar(benefit)

  .variable_interest_check_same_length(qx, z, "qx", "z")

  n <- length(z)
  benefit * prod(1 - qx) * (1 + z[n])^(-n)
}

#' @rdname spot_interest_apvs
#' @export
Axn1_spot <- function(qx, z, benefit = 1) {
  qx <- .variable_interest_check_probability(qx, "qx")
  z <- .variable_interest_check_rate(z, "z")
  benefit <- .variable_interest_benefit_scalar(benefit)

  .variable_interest_check_same_length(qx, z, "qx", "z")

  times <- seq_along(z)
  discount <- (1 + z)^(-times)
  survival_start <- .variable_interest_survival_start(qx)

  benefit * sum(discount * survival_start * qx)
}

#' @rdname spot_interest_apvs
#' @export
Axn_spot <- function(qx, z, benefit = 1) {
  qx <- .variable_interest_check_probability(qx, "qx")
  z <- .variable_interest_check_rate(z, "z")
  benefit <- .variable_interest_benefit_scalar(benefit)

  .variable_interest_check_same_length(qx, z, "qx", "z")

  Axn1_spot(qx = qx, z = z, benefit = benefit) +
    nEx_spot(qx = qx, z = z, benefit = benefit)
}

#' @rdname spot_interest_apvs
#' @export
axn_spot <- function(qx, z, type = c("immediate", "due"), benefit = 1) {
  type <- match.arg(type)

  qx <- .variable_interest_check_probability(qx, "qx")
  z <- .variable_interest_check_rate(z, "z")
  benefit <- .variable_interest_benefit_scalar(benefit)

  .variable_interest_check_same_length(qx, z, "qx", "z")

  n <- length(z)

  if (type == "immediate") {
    discount <- (1 + z)^(-seq_len(n))
    survival_end <- .variable_interest_survival_end(qx)
    return(benefit * sum(discount * survival_end))
  }

  survival_start <- .variable_interest_survival_start(qx)
  discount_due <- c(
    1,
    (1 + head(z, -1L))^(-seq_len(n - 1L))
  )

  benefit * sum(discount_due * survival_start)
}

# -------------------------------------------------------------------------
# Spot-rate valuation and bootstrapping
# -------------------------------------------------------------------------

#' Present value of deterministic cash flows using spot rates
#'
#' Discounts each cash flow using the spot rate matched to its payment time.
#' Time-zero cash flows are not discounted.
#'
#' @param amounts Numeric vector of cash-flow amounts.
#' @param times Numeric vector of payment times in years.
#' @param spot Numeric vector of spot rates matched elementwise to
#'   \code{times}. The value corresponding to a time-zero cash flow is ignored.
#' @param compounding Character string equal to \code{"annual"} or
#'   \code{"semiannual"}.
#'
#' @return A numeric scalar.
#'
#' @examples
#' pv_spot_cashflows(
#'   amounts = c(200000, 50000, 50000, 100000),
#'   times = c(0, 0.5, 1, 2),
#'   spot = c(0, 0.02440, 0.02601, 0.02936),
#'   compounding = "semiannual"
#' )
#'
#' @export
pv_spot_cashflows <- function(
    amounts,
    times,
    spot,
    compounding = c("annual", "semiannual")) {
  compounding <- match.arg(compounding)

  amounts <- .variable_interest_check_numeric(amounts, "amounts")
  times <- .variable_interest_check_nonnegative(times, "times")
  spot <- .variable_interest_check_numeric(spot, "spot")

  if (!(length(amounts) == length(times) &&
        length(times) == length(spot))) {
    stop(
      "amounts, times, and spot must have the same length.",
      call. = FALSE
    )
  }

  positive_time <- times > 0
  if (compounding == "annual" &&
      any(1 + spot[positive_time] <= 0)) {
    stop(
      "Annual spot rates at positive times must be greater than -1.",
      call. = FALSE
    )
  }
  if (compounding == "semiannual" &&
      any(1 + spot[positive_time] / 2 <= 0)) {
    stop(
      "Semiannual nominal spot rates at positive times must be greater than -2.",
      call. = FALSE
    )
  }

  discount <- rep(1, length(times))

  if (compounding == "annual") {
    discount[positive_time] <-
      (1 + spot[positive_time])^(-times[positive_time])
  } else {
    discount[positive_time] <-
      (1 + spot[positive_time] / 2)^(-2 * times[positive_time])
  }

  sum(amounts * discount)
}

#' Bootstrap semiannual nominal spot rates
#'
#' Bootstraps nominal annual spot rates convertible semiannually from par
#' coupon yields at consecutive half-year maturities.
#'
#' @param maturity Numeric vector of positive maturities in years, in strictly
#'   increasing order. Maturities must be consecutive multiples of
#'   \code{0.5}.
#' @param coupon_yield Numeric vector of nominal annual par coupon yields
#'   convertible semiannually. Values must be greater than \code{-2}.
#' @param par Positive scalar par value.
#'
#' @return A numeric vector of nominal annual spot rates convertible
#'   semiannually.
#'
#' @examples
#' maturity <- c(0.5, 1.0, 1.5, 2.0)
#' coupon_yield <- c(0.0244, 0.0260, 0.0276, 0.0293)
#' z_from_coupon_semi(maturity, coupon_yield)
#'
#' @export
z_from_coupon_semi <- function(maturity, coupon_yield, par = 1000) {
  maturity <- .variable_interest_check_numeric(maturity, "maturity")
  coupon_yield <- .variable_interest_check_numeric(
    coupon_yield,
    "coupon_yield"
  )
  par <- .variable_interest_check_scalar(
    par,
    "par",
    lower = 0,
    strict_lower = TRUE
  )

  .variable_interest_check_same_length(
    maturity,
    coupon_yield,
    "maturity",
    "coupon_yield"
  )

  if (any(maturity <= 0) || is.unsorted(maturity, strictly = TRUE)) {
    stop(
      "maturity must be positive and strictly increasing.",
      call. = FALSE
    )
  }
  if (any(coupon_yield <= -2)) {
    stop(
      "coupon_yield must contain values greater than -2.",
      call. = FALSE
    )
  }

  periods <- round(2 * maturity)
  if (any(abs(2 * maturity - periods) > 1e-10)) {
    stop(
      "maturity must contain multiples of 0.5 years.",
      call. = FALSE
    )
  }
  if (!identical(as.integer(periods), seq_len(length(periods)))) {
    stop(
      "maturity must contain consecutive half-year maturities beginning at 0.5.",
      call. = FALSE
    )
  }

  spot <- numeric(length(maturity))

  for (j in seq_along(maturity)) {
    coupon <- par * coupon_yield[j] / 2
    m <- as.integer(periods[j])

    if (m == 1L) {
      spot[j] <- coupon_yield[j]
      next
    }

    prior_periods <- seq_len(m - 1L)
    pv_prior <- sum(
      coupon / (1 + spot[prior_periods] / 2)^prior_periods
    )
    final_cash_flow <- par + coupon
    remaining_price <- par - pv_prior

    if (remaining_price <= 0) {
      stop(
        "The supplied coupon yields do not produce a positive final discount factor.",
        call. = FALSE
      )
    }

    accumulation <- final_cash_flow / remaining_price
    spot[j] <- 2 * (accumulation^(1 / m) - 1)
  }

  spot
}

#' Bootstrap annual effective spot rates
#'
#' Bootstraps annual effective spot rates from par coupon yields at consecutive
#' integer maturities.
#'
#' @param maturity Numeric vector of positive integer maturities in strictly
#'   increasing order. Maturities must be consecutive and begin at 1.
#' @param coupon_yield Numeric vector of annual effective par coupon yields.
#'   Values must be greater than \code{-1}.
#' @param par Positive scalar par value.
#'
#' @return A numeric vector of annual effective spot rates.
#'
#' @examples
#' maturity <- 1:4
#' coupon_yield <- c(0.02, 0.04, 0.06, 0.08)
#' z_from_coupon_annual(maturity, coupon_yield)
#'
#' @export
z_from_coupon_annual <- function(maturity, coupon_yield, par = 1000) {
  maturity <- .variable_interest_check_numeric(maturity, "maturity")
  coupon_yield <- .variable_interest_check_rate(
    coupon_yield,
    "coupon_yield"
  )
  par <- .variable_interest_check_scalar(
    par,
    "par",
    lower = 0,
    strict_lower = TRUE
  )

  .variable_interest_check_same_length(
    maturity,
    coupon_yield,
    "maturity",
    "coupon_yield"
  )

  if (any(maturity <= 0) || is.unsorted(maturity, strictly = TRUE)) {
    stop(
      "maturity must be positive and strictly increasing.",
      call. = FALSE
    )
  }
  if (any(abs(maturity - round(maturity)) > 1e-10)) {
    stop("maturity must contain integer years.", call. = FALSE)
  }

  maturity <- as.integer(round(maturity))
  if (!identical(maturity, seq_len(length(maturity)))) {
    stop(
      "maturity must contain consecutive integer maturities beginning at 1.",
      call. = FALSE
    )
  }

  spot <- numeric(length(maturity))

  for (j in seq_along(maturity)) {
    coupon <- par * coupon_yield[j]
    m <- maturity[j]

    if (m == 1L) {
      spot[j] <- coupon_yield[j]
      next
    }

    prior_periods <- seq_len(m - 1L)
    pv_prior <- sum(coupon / (1 + spot[prior_periods])^prior_periods)
    final_cash_flow <- par + coupon
    remaining_price <- par - pv_prior

    if (remaining_price <= 0) {
      stop(
        "The supplied coupon yields do not produce a positive final discount factor.",
        call. = FALSE
      )
    }

    accumulation <- final_cash_flow / remaining_price
    spot[j] <- accumulation^(1 / m) - 1
  }

  spot
}

# -------------------------------------------------------------------------
# Forward and spot rate conversions
# -------------------------------------------------------------------------

#' Forward rate implied by spot rates
#'
#' Computes the annual effective forward rate for the interval from time
#' \code{n} to time \code{n + k}:
#' \deqn{
#' (1+z_{n+k})^{n+k}
#' =
#' (1+z_n)^n(1+f_{n,k})^k.
#' }
#'
#' @param z Numeric vector of annual effective spot rates for maturities
#'   \code{1, ..., length(z)}. Each value must be greater than \code{-1}.
#' @param n Nonnegative integer forward-start time.
#' @param k Positive integer forward period.
#'
#' @return A numeric scalar.
#'
#' @examples
#' z <- c(0.03, 0.04, 0.05, 0.06, 0.07)
#' fnk_from_z(z, n = 1, k = 4)
#' fnk_from_z(z, n = 2, k = 2)
#'
#' @export
fnk_from_z <- function(z, n, k) {
  z <- .variable_interest_check_rate(z, "z")
  n <- .variable_interest_check_integer(n, "n")
  k <- .variable_interest_check_integer(k, "k", positive = TRUE)

  if (n + k > length(z)) {
    stop("Spot rates are required through maturity n + k.", call. = FALSE)
  }

  z_n <- if (n == 0L) 0 else z[n]
  z_nk <- z[n + k]

  ((1 + z_nk)^(n + k) / (1 + z_n)^n)^(1 / k) - 1
}

#' Matrix of forward rates implied by spot rates
#'
#' Constructs a matrix of annual effective forward rates. Rows correspond to
#' forward-start times \eqn{n=1,\ldots,m-1}; columns correspond to forward
#' periods \eqn{k=1,\ldots,m-1}. Entries requiring maturities beyond
#' \eqn{m} are returned as \code{NA}.
#'
#' @param z Numeric vector of at least two annual effective spot rates. Each
#'   value must be greater than \code{-1}.
#'
#' @return A numeric matrix.
#'
#' @examples
#' forward_matrix_from_z(c(0.03, 0.04, 0.05, 0.06, 0.07))
#'
#' @export
forward_matrix_from_z <- function(z) {
  z <- .variable_interest_check_rate(z, "z")

  if (length(z) < 2L) {
    stop("z must contain at least two spot rates.", call. = FALSE)
  }

  m <- length(z)
  out <- matrix(
    NA_real_,
    nrow = m - 1L,
    ncol = m - 1L,
    dimnames = list(
      paste0("n=", seq_len(m - 1L)),
      paste0("k=", seq_len(m - 1L))
    )
  )

  for (n in seq_len(m - 1L)) {
    for (k in seq_len(m - n)) {
      out[n, k] <- fnk_from_z(z, n = n, k = k)
    }
  }

  out
}

#' Spot rates from one-year forward rates
#'
#' Converts annual effective one-year forward rates
#' \eqn{f_{0,1},f_{1,1},\ldots,f_{n-1,1}} into annual effective spot rates:
#' \deqn{
#' (1+z_n)^n
#' =
#' \prod_{j=0}^{n-1}(1+f_{j,1}).
#' }
#'
#' @param fn1 Numeric vector of annual effective one-year forward rates. Each
#'   value must be greater than \code{-1}.
#'
#' @return A numeric vector of annual effective spot rates.
#'
#' @examples
#' z_from_fn1(c(0.04, 0.05, 0.06, 0.07, 0.08))
#'
#' @export
z_from_fn1 <- function(fn1) {
  fn1 <- .variable_interest_check_rate(fn1, "fn1")

  accumulation <- cumprod(1 + fn1)
  accumulation^(1 / seq_along(fn1)) - 1
}

test_that("nEx_var reproduces Example 15.1", {
  qx <- c(.03, .04, .05, .06, .07)

  val1 <- nEx_var(qx = qx, i = c(.06, .07, .08, .09, .10), benefit = 1000)
  val2 <- nEx_var(qx = qx, i = c(.06, .06, .06, .06, .06), benefit = 1000)
  val3 <- nEx_var(qx = qx, i = c(.06, .05, .04, .03, .02), benefit = 1000)

  expect_equal(val1, 526.61, tolerance = 1e-2)
  expect_equal(val2, 577.93, tolerance = 1e-2)
  expect_equal(val3, 635.97, tolerance = 1e-2)
})

test_that("Axn1_var reproduces Example 15.2", {
  qx <- c(.03, .04, .05, .06, .07)

  val1 <- Axn1_var(qx = qx, i = c(.06, .07, .08, .09, .10))
  val2 <- Axn1_var(qx = qx, i = c(.06, .06, .06, .06, .06))
  val3 <- Axn1_var(qx = qx, i = c(.06, .05, .04, .03, .02))

  expect_equal(val1, 0.1799, tolerance = 1e-4)
  expect_equal(val2, 0.1875, tolerance = 1e-3)
  expect_equal(val3, 0.1958, tolerance = 1e-3)
})

test_that("Axn_var equals Axn1_var plus nEx_var", {
  qx <- c(.03, .04, .05, .06, .07)
  i  <- c(.06, .07, .08, .09, .10)

  term_val  <- Axn1_var(qx = qx, i = i)
  pe_val    <- nEx_var(qx = qx, i = i)
  endow_val <- Axn_var(qx = qx, i = i)

  expect_equal(endow_val, term_val + pe_val, tolerance = 1e-12)
})

test_that("axn_var reproduces Example 15.3 annuity values", {
  qx <- rep(.02, 5)

  level <- axn_var(qx = qx, i = c(.06, .06, .06, .06, .06), type = "immediate")
  fall  <- axn_var(qx = qx, i = c(.06, .05, .04, .03, .03), type = "immediate")
  rise  <- axn_var(qx = qx, i = c(.06, .07, .08, .09, .10), type = "immediate")

  expect_equal(level, 3.9756, tolerance = 1e-4)
  expect_equal(fall, 4.1102, tolerance = 1e-4)
  expect_equal(rise, 3.8461, tolerance = 1e-4)
})

test_that("z_from_coupon_semi reproduces Table 15.6 approximately", {
  maturity <- c(0.5, 1.0, 1.5, 2.0)
  coupon_yield <- c(0.0244, 0.0260, 0.0276, 0.0293)

  z <- z_from_coupon_semi(maturity, coupon_yield)

  expect_equal(100 * z[1], 2.440, tolerance = 1e-3)
  expect_equal(100 * z[2], 2.601, tolerance = 1e-3)
  expect_equal(100 * z[3], 2.763, tolerance = 1e-3)
  expect_equal(100 * z[4], 2.936, tolerance = 1e-3)
})

test_that("pv_spot_cashflows reproduces Example 15.4", {
  val <- pv_spot_cashflows(
    amounts = c(200000, 50000, 50000, 100000),
    times   = c(0, 0.5, 1, 2),
    spot    = c(0, 0.02440, 0.02601, 0.02936),
    compounding = "semiannual"
  )

  expect_equal(val, 392459.12, tolerance = 1e-2)
})

test_that("spot-rate Chapter 15 functions reproduce Example 15.5 premium pieces", {
  qx <- c(.02, .03, .04, .05, .06)
  z  <- c(.03, .04, .05, .06, .07)

  apvb <- Axn1_spot(qx = qx, z = z)
  apvp_factor <- axn_spot(qx = qx, z = z, type = "due")

  premium <- 1000000 * apvb / apvp_factor

  expect_equal(apvb, 0.1527, tolerance = 5e-4)
  expect_equal(apvp_factor, 4.3054, tolerance = 1e-4)
  expect_equal(premium, 35467.09, tolerance = 1e-2)
})




test_that("spot-rate identities hold", {
  qx <- c(.02, .03, .04, .05, .06)
  z  <- c(.03, .04, .05, .06, .07)

  expect_equal(
    Axn_spot(qx = qx, z = z),
    Axn1_spot(qx = qx, z = z) + nEx_spot(qx = qx, z = z),
    tolerance = 1e-12
  )
})

test_that("fnk_from_z reproduces Example 15.6 values", {
  z <- c(0.03, 0.04, 0.05, 0.06, 0.07)

  expect_equal(100 * fnk_from_z(z, n = 1, k = 4), 8.024, tolerance = 1e-3)
  expect_equal(100 * fnk_from_z(z, n = 2, k = 2), 8.038, tolerance = 1e-3)
})

test_that("forward_matrix_from_z gives expected selected entries", {
  z <- c(0.03, 0.04, 0.05, 0.06, 0.07)
  fmat <- forward_matrix_from_z(z)

  expect_equal(100 * fmat["n=1", "k=1"], 5.01, tolerance = 1e-2)
  expect_equal(100 * fmat["n=1", "k=2"], 6.01, tolerance = 1e-2)
  expect_equal(100 * fmat["n=2", "k=1"], 7.03, tolerance = 1e-2)
  expect_true(is.na(fmat["n=4", "k=2"]))
})

test_that("z_from_fn1 converts one-year forward rates to spot rates", {
  fn1 <- c(0.04, 0.05, 0.06, 0.07, 0.08)
  z <- z_from_fn1(fn1)

  expected <- c(
    0.0400000000,
    ((1.04 * 1.05)^(1 / 2)) - 1,
    ((1.04 * 1.05 * 1.06)^(1 / 3)) - 1,
    ((1.04 * 1.05 * 1.06 * 1.07)^(1 / 4)) - 1,
    ((1.04 * 1.05 * 1.06 * 1.07 * 1.08)^(1 / 5)) - 1
  )

  expect_equal(z, expected, tolerance = 1e-12)
})

test_that("z_from_coupon_annual works on a simple annual table", {
  maturity <- 1:4
  coupon_yield <- c(0.02, 0.04, 0.06, 0.08)

  z <- z_from_coupon_annual(maturity, coupon_yield)

  expect_equal(z[1], 0.02, tolerance = 1e-12)
  expect_true(all(diff(z) > 0))
})

test_that("vt_var returns cumulative variable-rate discount factors", {
  v <- vt_var(c(0.06, 0.07, 0.08))

  expect_equal(v[1], 1 / 1.06, tolerance = 1e-12)
  expect_equal(v[2], 1 / (1.06 * 1.07), tolerance = 1e-12)
  expect_equal(v[3], 1 / (1.06 * 1.07 * 1.08), tolerance = 1e-12)
})




testthat::test_that("variable discount factors allow valid negative rates", {
  rates <- c(-0.01, 0.02, 0.03)

  expected <- cumprod(1 / (1 + rates))

  testthat::expect_equal(
    vt_var(rates),
    expected,
    tolerance = 1e-12
  )
})


testthat::test_that("variable-interest APVs satisfy standard identities", {
  qx <- c(0.02, 0.03, 0.04, 0.05)
  rates <- c(0.03, 0.04, 0.05, 0.06)

  testthat::expect_equal(
    Axn_var(qx, rates),
    Axn1_var(qx, rates) + nEx_var(qx, rates),
    tolerance = 1e-12
  )
})


testthat::test_that("spot-rate APVs satisfy standard identities", {
  qx <- c(0.02, 0.03, 0.04, 0.05)
  spot <- c(0.03, 0.04, 0.05, 0.06)

  testthat::expect_equal(
    Axn_spot(qx, spot),
    Axn1_spot(qx, spot) + nEx_spot(qx, spot),
    tolerance = 1e-12
  )
})


testthat::test_that("annuity-due includes the time-zero payment", {
  qx <- rep(0.02, 5)
  rates <- rep(0.05, 5)

  immediate <- axn_var(qx, rates, type = "immediate")
  due <- axn_var(qx, rates, type = "due")

  testthat::expect_true(due > immediate)
  testthat::expect_equal(
    due,
    sum(
      c(1, head(vt_var(rates), -1)) *
        c(1, cumprod(1 - qx))[seq_along(qx)]
    ),
    tolerance = 1e-12
  )
})


testthat::test_that("constant variable rates agree with direct discounting", {
  qx <- c(0.02, 0.03, 0.04)
  rate <- 0.05
  rates <- rep(rate, length(qx))

  survival_start <- c(1, cumprod(1 - qx))[seq_along(qx)]
  expected <- sum(
    (1 + rate)^(-seq_along(qx)) *
      survival_start *
      qx
  )

  testthat::expect_equal(
    Axn1_var(qx, rates),
    expected,
    tolerance = 1e-12
  )
})


testthat::test_that("spot cash-flow valuation supports valid negative rates", {
  out <- pv_spot_cashflows(
    amounts = c(100, 100),
    times = c(1, 2),
    spot = c(-0.01, 0.02),
    compounding = "annual"
  )

  expected <- 100 / 0.99 + 100 / 1.02^2

  testthat::expect_equal(out, expected, tolerance = 1e-12)
})


testthat::test_that("annual spot bootstrapping reprices par bonds", {
  maturity <- 1:4
  coupon_yield <- c(0.02, 0.04, 0.06, 0.08)
  par <- 1000

  spot <- z_from_coupon_annual(
    maturity = maturity,
    coupon_yield = coupon_yield,
    par = par
  )

  for (j in seq_along(maturity)) {
    coupon <- par * coupon_yield[j]
    cash_flows <- c(
      rep(coupon, maturity[j] - 1L),
      par + coupon
    )

    price <- sum(
      cash_flows /
        (1 + spot[seq_len(maturity[j])])^seq_len(maturity[j])
    )

    testthat::expect_equal(price, par, tolerance = 1e-8)
  }
})


testthat::test_that("semiannual spot bootstrapping reprices par bonds", {
  maturity <- c(0.5, 1.0, 1.5, 2.0)
  coupon_yield <- c(0.0244, 0.0260, 0.0276, 0.0293)
  par <- 1000

  spot <- z_from_coupon_semi(
    maturity = maturity,
    coupon_yield = coupon_yield,
    par = par
  )

  periods <- seq_along(maturity)

  for (j in seq_along(maturity)) {
    coupon <- par * coupon_yield[j] / 2
    cash_flows <- c(
      rep(coupon, periods[j] - 1L),
      par + coupon
    )

    price <- sum(
      cash_flows /
        (1 + spot[seq_len(periods[j])] / 2)^seq_len(periods[j])
    )

    testthat::expect_equal(price, par, tolerance = 1e-8)
  }
})


testthat::test_that("forward and spot conversions are internally consistent", {
  forward <- c(0.04, 0.05, 0.06, 0.07, 0.08)
  spot <- z_from_fn1(forward)

  recovered <- c(
    spot[1],
    vapply(
      1:(length(spot) - 1L),
      function(n) fnk_from_z(spot, n = n, k = 1),
      numeric(1)
    )
  )

  testthat::expect_equal(
    recovered,
    forward,
    tolerance = 1e-12
  )
})


testthat::test_that("forward rate matrix has expected structure", {
  spot <- c(0.03, 0.04, 0.05, 0.06)

  out <- forward_matrix_from_z(spot)

  testthat::expect_true(is.matrix(out))
  testthat::expect_equal(dim(out), c(3, 3))
  testthat::expect_equal(out[1, 1], fnk_from_z(spot, 1, 1))
  testthat::expect_true(is.na(out[3, 2]))
})


testthat::test_that("variable-interest functions reject invalid inputs", {
  testthat::expect_error(
    vt_var(numeric(0)),
    "positive length"
  )

  testthat::expect_error(
    vt_var(c(0.05, -1)),
    "greater than -1"
  )

  testthat::expect_error(
    Axn1_var(
      qx = c(0.02, 0.03),
      i = c(0.04, 0.05, 0.06)
    ),
    "same length"
  )

  testthat::expect_error(
    nEx_var(
      qx = c(0.02, 0.03),
      i = c(0.04, 0.05),
      benefit = c(1, 2)
    ),
    "benefit must be a numeric scalar"
  )

  testthat::expect_error(
    forward_matrix_from_z(0.05),
    "at least two"
  )
})


testthat::test_that("bootstrapping requires consecutive maturities", {
  testthat::expect_error(
    z_from_coupon_annual(
      maturity = c(1, 3),
      coupon_yield = c(0.03, 0.04)
    ),
    "consecutive integer maturities"
  )

  testthat::expect_error(
    z_from_coupon_semi(
      maturity = c(0.5, 1.5),
      coupon_yield = c(0.03, 0.04)
    ),
    "consecutive half-year maturities"
  )
})

# life_table_fractional.R
#
# Fractional-age life table utilities.
# These functions work from a discrete life table and a within-year assumption:
#   - "udd"       : uniform distribution of deaths
#   - "cf"        : constant force
#   - "balducci"  : hyperbolic / Balducci

# -------------------------------------------------------------------------
# Internal helpers
# -------------------------------------------------------------------------

.match_assumption <- function(assumption) {
  assumption <- match.arg(assumption, choices = c("udd", "cf", "balducci"))
  assumption
}

.recycle_fractional_args <- function(...) {
  args <- lapply(list(...), as.numeric)
  lens <- vapply(args, length, integer(1))

  if (any(lens == 0L)) {
    stop("All arguments must have positive length.", call. = FALSE)
  }

  if (any(vapply(args, function(z) any(!is.finite(z)), logical(1)))) {
    stop("All arguments must contain finite numeric values.", call. = FALSE)
  }

  common_len <- max(lens)

  if (any(!(lens %in% c(1L, common_len)))) {
    stop("Arguments must have compatible lengths.", call. = FALSE)
  }

  lapply(args, function(z) {
    if (length(z) == 1L) rep(z, common_len) else z
  })
}

.check_fractional_inputs <- function(tbl, x, t) {
  .validate_life_table(tbl)

  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  if (any(t < 0) || any(t > 1)) {
    stop("t must satisfy 0 <= t <= 1.", call. = FALSE)
  }

  qx_vals <- qx_tab(tbl, x)

  if (any(is.na(qx_vals))) {
    stop("Some x values are not available in the life table.", call. = FALSE)
  }

  invisible(TRUE)
}

# -------------------------------------------------------------------------
# Fractional-age survival probabilities
# -------------------------------------------------------------------------

#' Fractional survival probability from a life table
#'
#' Computes \eqn{{}_t p_x} for \eqn{0 \le t \le 1} from a discrete life table
#' under UDD, constant force, or Balducci.
#'
#' @param tbl A life_table object.
#' @param x Numeric vector of integer ages.
#' @param t Numeric vector of fractional durations with \eqn{0 \le t \le 1}.
#' @param assumption One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.
#'
#' @return Numeric vector of \eqn{{}_t p_x} values.
#' @export
tpx_tab <- function(tbl, x, t, assumption = c("udd", "cf", "balducci")) {
  assumption <- .match_assumption(assumption)

  args <- .recycle_fractional_args(x, t)
  x <- args[[1L]]
  t <- args[[2L]]

  .check_fractional_inputs(tbl, x, t)

  qx <- qx_tab(tbl, x)
  px <- 1 - qx

  out <- switch(
    assumption,
    "udd" = 1 - t * qx,
    "cf" = px^t,
    "balducci" = {
      val <- px / (1 - (1 - t) * qx)
      val[t == 0] <- 1
      val
    }
  )

  pmin(pmax(out, 0), 1)
}

#' Fractional failure probability from a life table
#'
#' Computes \eqn{{}_t q_x = 1 - {}_t p_x} for \eqn{0 \le t \le 1}.
#'
#' @inheritParams tpx_tab
#' @return Numeric vector of \eqn{{}_t q_x} values.
#' @export
tqx_tab <- function(tbl, x, t, assumption = c("udd", "cf", "balducci")) {
  1 - tpx_tab(tbl, x, t, assumption = assumption)
}

# -------------------------------------------------------------------------
# Force of mortality / hazard within the year
# -------------------------------------------------------------------------

#' Fractional force of mortality from a life table
#'
#' Computes \eqn{\mu_{x+t}} under UDD, constant force, or Balducci.
#'
#' @inheritParams tpx_tab
#' @return Numeric vector of \eqn{\mu_{x+t}} values.
#' @export
mux_tab <- function(tbl, x, t, assumption = c("udd", "cf", "balducci")) {
  assumption <- .match_assumption(assumption)

  args <- .recycle_fractional_args(x, t)
  x <- args[[1L]]
  t <- args[[2L]]

  .check_fractional_inputs(tbl, x, t)

  qx <- qx_tab(tbl, x)
  px <- 1 - qx

  switch(
    assumption,
    "udd" = qx / (1 - t * qx),
    "cf" = -log(px),
    "balducci" = qx / (1 - (1 - t) * qx)
  )
}

# -------------------------------------------------------------------------
# Fractional conditional density within the year
# -------------------------------------------------------------------------

#' Fractional conditional density from a life table
#'
#' Computes the conditional density
#' \eqn{f_x(t \mid T_0 > x) = {}_t p_x \mu_{x+t}}.
#'
#' @inheritParams tpx_tab
#' @return Numeric vector of conditional density values.
#' @export
fx_tab <- function(tbl, x, t, assumption = c("udd", "cf", "balducci")) {
  assumption <- .match_assumption(assumption)

  args <- .recycle_fractional_args(x, t)
  x <- args[[1L]]
  t <- args[[2L]]

  .check_fractional_inputs(tbl, x, t)

  out <- tpx_tab(tbl, x, t, assumption = assumption) *
    mux_tab(tbl, x, t, assumption = assumption)

  pmax(out, 0)
}

# -------------------------------------------------------------------------
# Deferred probabilities
# -------------------------------------------------------------------------

#' Deferred death probability from a life table
#'
#' Computes the probability that a life aged x survives n years and then
#' dies within the following m years:
#' \eqn{{}_{n|m} q_x = {}_n p_x \cdot {}_m q_{x+n}}.
#'
#' This function is for integer n and m in the discrete tabular setting.
#'
#' @param tbl A life_table object.
#' @param x Numeric vector of ages.
#' @param n Nonnegative integer deferred period.
#' @param m Nonnegative integer subsequent period.
#'
#' @return Numeric vector of \eqn{{}_{n|m} q_x} values.
#' @export
nmxq <- function(tbl, x, n, m) {
  .validate_life_table(tbl)

  args <- .recycle_fractional_args(x, n, m)
  x <- args[[1L]]
  n <- args[[2L]]
  m <- args[[3L]]

  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  if (any(n < 0 | n != floor(n))) {
    stop("n must be a nonnegative integer.", call. = FALSE)
  }

  if (any(m < 0 | m != floor(m))) {
    stop("m must be a nonnegative integer.", call. = FALSE)
  }

  npx(tbl, x, n) * nqx(tbl, x + n, m)
}

#' Curtate death probability from a life table
#'
#' Computes \eqn{{}_{k|} q_x = {}_k p_x - {}_{k+1} p_x}.
#'
#' @param tbl A life_table object.
#' @param x Numeric vector of ages.
#' @param k Nonnegative integer.
#'
#' @return Numeric vector of \eqn{{}_{k|} q_x} values.
#' @export
nkqx <- function(tbl, x, k) {
  .validate_life_table(tbl)

  args <- .recycle_fractional_args(x, k)
  x <- args[[1L]]
  k <- args[[2L]]

  if (any(x < 0)) {
    stop("x must be nonnegative.", call. = FALSE)
  }

  if (any(k < 0 | k != floor(k))) {
    stop("k must be a nonnegative integer.", call. = FALSE)
  }

  npx(tbl, x, k) - npx(tbl, x, k + 1)
}

#' Internal: safe numerical integration helper
#'
#' @keywords internal
.safe_integrate <- function(f, lower, upper, ..., rel.tol = 1e-10) {
  out <- stats::integrate(f, lower = lower, upper = upper, ..., rel.tol = rel.tol)
  unname(out$value)
}

#' Internal: choose a practical upper bound for integrals to infinity
#'
#' Strategy: find t such that S0(t) is very small, or fall back to a cap.
#'
#' @keywords internal
.find_upper_t <- function(x, model, ..., eps = 1e-10, t_max = 500) {
  # we want S0(x + t) / S0(x) ~ 0, equivalently S0(x+t) small
  # Start with a modest grid and expand until survival is tiny.
  grid <- c(10, 20, 50, 100, 200, 300, 500)
  for (t in grid) {
    s <- S0(x + t, model = model, ...)
    if (is.finite(s) && s < eps) return(t)
  }
  t_max
}

---
title: "Getting Started with mqriskR"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with mqriskR}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

## Overview

The `mqriskR` package provides functions for actuarial risk modeling,
including survival models, insurance and annuity values, premium calculations,
reserves, multiple-decrement models, and mortality improvement projections.

The functions are designed to align with standard actuarial notation
and support teaching, exam preparation, and reproducible actuarial analysis.

This vignette provides a brief introduction to the main functionality
of the package.

For detailed derivations and additional examples, see *Models for Quantifying Risk*.


## Survival Models

We begin by computing survival probabilities under a simple model.

```{r}
library(mqriskR)

# Probability of surviving 10 years from age 40
tpx(10, x = 40, model = "uniform", omega = 100)

```

This gives the probability that a life aged 40 survives 10 years under the uniform distribution of deaths with limiting age 100.


## Insurance Functions

We next compute insurance actuarial present values.

```{r}
# Whole life insurance
Ax(40, i = 0.05, model = "uniform", omega = 100)

# 10-year term insurance
Axn1(40, n = 10, i = 0.05, model = "uniform", omega = 100)

```

The first value is the actuarial present value of a whole life insurance paying 1 at the end of the year of death, while the second value is the present value of a 10-year term insurance that pays 1 if death occurs within the term.

## Life Annuities

We now compute annuity values.

```{r}
# Whole life annuity-immediate
ax(40, i = 0.05, model = "uniform", omega = 100)

# 10-year temporary annuity
axn(40, n = 10, i = 0.05, model = "uniform", omega = 100)
```

The first value represents the present value of a whole life annuity-immediate, while the second gives the present value of a 10-year temporary annuity-immediate.

## Premium Calculation

Using the equivalence principle, we compute a level annual net premium.

```{r}
Ax_term <- Axn1(40, n = 10, i = 0.05, model = "uniform", omega = 100)
adotx_term <- adotxn(40, n = 10, i = 0.05, model = "uniform", omega = 100)

premium <- Ax_term / adotx_term
premium
```

This is the level annual net premium calculated using the equivalence principle, where expected present value of benefits equals expected present value of premiums. This corresponds to the identity

$$
P_{x:\overline{n}|}^{1} = A_{x:\overline{n}|}^{1} / \ddot{a}_{x:\overline{n}|}.
$$


## Reserve Calculation

We now compute a simple prospective reserve at time \(t = 5\) for the same contract.

```{r}
t <- 5

# Prospective reserve: V_t = A_{x+t:n-t} - P * ä_{x+t:n-t}
Ax_future <- Axn1(40 + t, n = 10 - t, i = 0.05, model = "uniform", omega = 100)
adotx_future <- adotxn(40 + t, n = 10 - t, i = 0.05, model = "uniform", omega = 100)

V_t <- Ax_future - premium * adotx_future
V_t
```


This is the prospective reserve at time $t = 5$, representing the expected future loss at that time based on remaining benefits and premiums. This illustrates the standard prospective reserve formula:

$$
{}_{t}V_{x:\overline{n}|}^{1} = A_{x+t:\overline{n-t}|}^{1} - P_{x:\overline{n}|}^{1} \ddot{a}_{x+t:\overline{n-t}|}.
$$


## Multiple-Decrement Models

The package supports multiple-decrement tables.

```{r}
x <- 45:50

qmat <- cbind(
  q1 = c(.011, .012, .013, .014, .015, .016),
  q2 = rep(0.10, 6)
)

tbl <- md_table(x, qmat, radix = 1000)

tbl
```

This table summarizes the multiple-decrement model, including cause-specific decrement probabilities, total decrement probabilities, survival probabilities, and the number of lives remaining at each age.

We can compute survival probabilities:

```{r}
npxtau_md(tbl, x = 46, n = 3)
```

This gives the probability that a life aged 46 remains in force for 3 years when all causes of decrement are considered.

## Mortality Improvement

The package also includes mortality improvement projections.

```{r}
qx_proj(
  qx_base = 0.02,
  AAx = 0.01,
  base_year = 2020,
  proj_year = 2030
)
```

This gives the projected one-year death probability after allowing for mortality improvement between the base year and the projection year.


## Summary

The `mqriskR` package provides a unified framework for actuarial modeling,
covering survival models, insurance and annuity functions, premium calculations,
reserves, multiple-decrement models, and mortality improvement.

Users can combine these functions to build more complex actuarial models
in a transparent and reproducible way.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{reserve_ch10}
\alias{reserve_ch10}
\title{Reserve functions for Chapter 10}
\description{
Net level premium reserve functions aligned with Chapter 10 notation.
}
\details{
These functions implement prospective reserve formulas for whole life,
term insurance, pure endowment, endowment insurance, selected continuous
premium cases, selected m-thly premium cases, basic gain/loss
calculations, and duration-t loss moments.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_varying_benefits.R
\name{insurance_varying_benefits}
\alias{insurance_varying_benefits}
\title{Insurance models with varying benefits}
\description{
Functions for discrete and continuous contingent payment models with
varying benefits.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{reversionary_annuities}
\alias{reversionary_annuities}
\alias{ax_y}
\alias{ay_x}
\title{Reversionary annuity functions}
\usage{
ax_y(x, y, i, tbl = NULL, model = NULL, ..., k_max = 5000L, tol = 1e-12)

ay_x(x, y, i, tbl = NULL, model = NULL, ..., k_max = 5000L, tol = 1e-12)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the survival model or
life-table functions.}

\item{k_max}{Maximum number of terms used for an infinite series.}

\item{tol}{Numerical tolerance used to assess convergence.}
}
\value{
A numeric vector of actuarial present values.
}
\description{
Computes reversionary whole-life annuities payable to one life after the
death of the other life.
}
\details{
\code{ax_y()} computes an annuity payable to the second life after the
death of the first life.

\code{ay_x()} computes an annuity payable to the first life after the
death of the second life.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R, R/reserve_ch11_ext.R
\name{tVFx}
\alias{tVFx}
\title{Full preliminary term reserve for whole life insurance}
\usage{
tVFx(x, t, i, tbl = NULL, model = NULL, ...)

tVFx(x, t, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.

Numeric vector.
}
\description{
Computes the FPT reserve for a whole life insurance.
For whole life insurance,
\eqn{{}_1V^F = 0}
and for \eqn{t \ge 1},
\eqn{{}_tV^F = {}_{t-1}V_{x+1}^{NLP}}.

Computes the FPT reserve for a whole life insurance.
For whole life insurance,
\eqn{{}_1V^F = 0}
and for \eqn{t \ge 1},
\eqn{{}_tV^F = {}_{t-1}V_{x+1}^{NLP}}.
}
\examples{
tVFx(40, t = 5, i = 0.05, model = "uniform", omega = 100)
tVFx(40, t = 5, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{ex_curtate}
\alias{ex_curtate}
\title{Curtate expectation of life}
\usage{
ex_curtate(x, model, ..., k_max = 5000, tol = 1e-12)
}
\arguments{
\item{x}{Numeric vector of ages.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}

\item{k_max}{Maximum integer duration to sum to.}

\item{tol}{Stop early if the summand is smaller than this tolerance for
several consecutive steps.}
}
\value{
Numeric vector of curtate expectations.
}
\description{
Computes \eqn{e_x = E[K_x] = \sum_{k=1}^{\infty} {}_k p_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{nqxj_md}
\alias{nqxj_md}
\title{Cause-specific multiple-decrement probability from a table}
\usage{
nqxj_md(tbl, x, n, j)
}
\arguments{
\item{tbl}{A multiple-decrement table produced by \code{md_table()}.}

\item{x}{Starting age or duration. May be scalar or vector.}

\item{n}{Nonnegative integer term. May be scalar or vector.}

\item{j}{Positive integer cause index. May be scalar or vector.}
}
\value{
A numeric vector of cause-specific decrement probabilities.
}
\description{
Computes the probability of decrement from cause \code{j} within
\code{n} years:
\deqn{
{}_nq_x^{(j)}
=
\sum_{k=0}^{n-1}
{}_kp_x^{(\tau)}q_{x+k}^{(j)}.
}
}
\examples{
ages <- 45:50
qmat <- cbind(
  q1 = c(0.011, 0.012, 0.013, 0.014, 0.015, 0.016),
  q2 = rep(0.100, 6)
)
tbl <- md_table(ages, qmat, radix = 1000)

nqxj_md(tbl, x = 46, n = 2, j = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{reserve_deferred_insurance_ch10}
\alias{reserve_deferred_insurance_ch10}
\alias{tVnAx}
\alias{htVnAx}
\title{Deferred insurance reserve functions}
\usage{
tVnAx(x, n, t, i, model, ...)

htVnAx(x, n, h, t, i, model, ...)
}
\arguments{
\item{x}{Issue age.}

\item{n}{Deferral period.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}

\item{h}{Premium-paying period.}
}
\value{
Numeric vector.

Numeric vector.
}
\description{
Chapter 10 reserve functions for deferred insurance contracts.
}
\details{
\code{tVnAx()} computes the reserve for an \eqn{n}-year deferred insurance
funded over the deferred period.

\code{htVnAx()} computes the reserve when premiums are limited to the first
\eqn{h} years, with \eqn{h \le n}.
}
\examples{
tVnAx(40, n = 20, t = 10, i = 0.05, model = "uniform", omega = 100)
htVnAx(40, n = 20, h = 10, t = 5, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{var_Axn1}
\alias{var_Axn1}
\title{Variance of term insurance PV}
\usage{
var_Axn1(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}
}
\value{
Numeric vector of variances.
}
\description{
Computes \eqn{\mathrm{Var}(Z_{x:\overline{n}|}^{1}) =
{}^{2}A_{x:\overline{n}|}^{1} - (A_{x:\overline{n}|}^{1})^2}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{A2x}
\alias{A2x}
\title{Second moment of whole life insurance PV}
\usage{
A2x(x, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}

\item{tol}{Numerical tolerance for truncating infinite sums.}

\item{k_max}{Maximum number of terms in the sum.}
}
\value{
Numeric vector of second moments.
}
\description{
Computes \eqn{{}^{2}A_x} by evaluating \eqn{A_x} at doubled force.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{rt_ul}
\alias{rt_ul}
\title{Account-value to guaranteed-fund ratio}
\usage{
rt_ul(AV, GMF)
}
\arguments{
\item{AV}{Nonnegative account value.}

\item{GMF}{Positive guaranteed maturity fund.}
}
\value{
A numeric vector.
}
\description{
Computes the ratio of account value to guaranteed maturity fund, capped at
one.
}
\examples{
rt_ul(AV = 4918.20, GMF = 14678.57)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{GT_cont}
\alias{GT_cont}
\title{Total gain for a continuous-style one-step recursion}
\usage{
GT_cont(Vt, Vt1, P, delta_actual, p_actual, benefit = 0, h = 1)
}
\arguments{
\item{Vt}{Reserve at time `t`.}

\item{Vt1}{Reserve at time `t + h`.}

\item{P}{Premium rate.}

\item{delta_actual}{Actual force of interest.}

\item{p_actual}{Actual survival probability over the step.}

\item{benefit}{Benefit paid at the start of the step.}

\item{h}{Positive step length.}
}
\value{
Numeric vector of gain values.
}
\description{
Computes the amount accumulated during a step, less the expected reserve
required at the end of the step.
}
\details{
Reserves, premium rates, benefits, and forces of interest may be negative
when such values are meaningful for the application. The step length must
be positive, and the survival probability must lie in \eqn{[0,1]}.
}
\examples{
GT_cont(
  Vt = 10,
  Vt1 = 11,
  P = 1,
  delta_actual = 0.05,
  p_actual = 0.99
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{A2xn}
\alias{A2xn}
\title{Second moment of endowment insurance PV}
\usage{
A2xn(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}
}
\value{
Numeric vector of second moments.
}
\description{
Computes \eqn{{}^{2}A_{x:\overline{n}|}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{Axn_m}
\alias{Axn_m}
\title{m-thly endowment insurance APV}
\usage{
Axn_m(x, n, i, m, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes
\eqn{A_{x:\overline{n}|}^{(m)} =
A_{x:\overline{n}|}^{1(m)} + v^n {}_np_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_mthly.R
\name{annuity_mthly_deferred_immediate}
\alias{annuity_mthly_deferred_immediate}
\alias{nax_m}
\title{Deferred whole life m-thly annuity-immediate}
\usage{
nax_m(x, n, m, i, model, ..., k_max = 2e+05, tol = 1e-12)
}
\arguments{
\item{x}{Age.}

\item{n}{Deferral period in years.}

\item{m}{Number of payments per year.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model name.}

\item{...}{Additional parameters passed to the survival model.}

\item{k_max}{Maximum summation horizon for non-terminating models.}

\item{tol}{Truncation tolerance for non-terminating models.}
}
\value{
Numeric vector of annuity values.
}
\description{
Computes the deferred whole life m-thly annuity-immediate using
\deqn{{}_nE_x \, a_{x+n}^{(m)}}
}
\examples{
nax_m(40, n = 10, m = 12, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mortality_improvement.R
\name{px_proj}
\alias{px_proj}
\title{Project one-year survival probability under mortality improvement}
\usage{
px_proj(qx_base, AAx, base_year, proj_year)
}
\arguments{
\item{qx_base}{Base-year one-year death probability \eqn{q_x^{[B]}}.}

\item{AAx}{Mortality improvement factor \eqn{AA_x}.}

\item{base_year}{Base year \eqn{B}.}

\item{proj_year}{Projection year \eqn{Y}. May be scalar or vector.}
}
\value{
Numeric vector of projected one-year survival probabilities.
}
\description{
Computes projected one-year survival probability
\deqn{p_x^{[Y]} = 1 - q_x^{[Y]}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_relationships.R
\name{annuity_relationships}
\alias{annuity_relationships}
\alias{annuity_identity_ax}
\alias{annuity_identity_adotx}
\alias{annuity_identity_abarx}
\alias{annuity_identity_axn}
\alias{annuity_identity_adotxn}
\alias{annuity_identity_abarxn}
\alias{annuity_identity_nax}
\alias{annuity_identity_nadotx}
\alias{annuity_identity_nabarx}
\title{Annuity-insurance relationships}
\usage{
annuity_identity_ax(x, i, model = NULL, ..., tbl = NULL)

annuity_identity_adotx(x, i, model = NULL, ..., tbl = NULL)

annuity_identity_abarx(x, i, model, ...)

annuity_identity_axn(x, n, i, model = NULL, ..., tbl = NULL)

annuity_identity_adotxn(x, n, i, model = NULL, ..., tbl = NULL)

annuity_identity_abarxn(x, n, i, model, ...)

annuity_identity_nax(
  x,
  n,
  i,
  model = NULL,
  ...,
  tbl = NULL,
  k_max = 5000,
  tol = 1e-12
)

annuity_identity_nadotx(
  x,
  n,
  i,
  model = NULL,
  ...,
  tbl = NULL,
  k_max = 5000,
  tol = 1e-12
)

annuity_identity_nabarx(x, n, i, model, ..., tol = 1e-10)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object for discrete identities.}

\item{n}{Term or deferral period in years.}

\item{k_max}{Maximum summation horizon for non-terminating models.}

\item{tol}{Truncation tolerance for non-terminating models.}
}
\value{
Numeric vector containing the annuity value computed from the corresponding
annuity-insurance identity.
}
\description{
Identities linking annual and continuous annuity functions to the
corresponding insurance functions.

Computes \eqn{a_x = (v - A_x)/d}.

Computes \eqn{\ddot{a}_x = (1 - A_x)/d}.

Computes \eqn{\bar{a}_x = (1 - \bar{A}_x)/\delta}.

Computes
\eqn{a_{x:\overline{n}|} = (1 - A_{x:\overline{n}|})/d - 1 + {}_nE_x}.

Computes \eqn{\ddot{a}_{x:\overline{n}|} =
(1 - A_{x:\overline{n}|})/d}.

Computes \eqn{\bar{a}_{x:\overline{n}|} =
(1 - \bar{A}_{x:\overline{n}|})/\delta}.

Computes \eqn{{}_{n|}a_x = {}_nE_x a_{x+n}}.

Computes \eqn{{}_{n|}\ddot{a}_x = {}_nE_x \ddot{a}_{x+n}}.

Computes \eqn{{}_{n|}\bar{a}_x = {}_nE_x \bar{a}_{x+n}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{tVbarx}
\alias{tVbarx}
\title{Whole life reserve with continuous premiums}
\usage{
tVbarx(x, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object.}
}
\value{
A numeric vector of values.
}
\description{
Computes the reserve for a discrete whole life insurance
funded by continuous premiums.
}
\examples{
tVbarx(40, t = 10, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_varying_benefits.R
\name{IbarAbarx}
\alias{IbarAbarx}
\title{Fully continuous increasing whole life insurance}
\usage{
IbarAbarx(x, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes
\deqn{(\bar{I}\bar{A})_x =
\int_0^\infty t v^t {}_tp_x \mu_{x+t}\,dt.}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_conversions.R
\name{px_to_lx}
\alias{px_to_lx}
\title{Construct life-table values from p_x values}
\usage{
px_to_lx(px, radix = 1e+05)
}
\arguments{
\item{px}{Numeric vector of one-year survival probabilities \eqn{p_x}.}

\item{radix}{Positive radix \eqn{l_0}.}
}
\value{
Numeric vector of \eqn{l_x} values of length \code{length(px) + 1}.
}
\description{
Builds life-table survivor values recursively from
\eqn{l_{x+1} = l_x p_x}, starting from a chosen radix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{dist0}
\alias{dist0}
\alias{F0}
\alias{f0}
\title{Distribution functions for age-at-failure}
\usage{
F0(t, model, ...)

f0(t, model, ...)
}
\arguments{
\item{t}{Numeric vector of times.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}
}
\value{
Numeric vector.
}
\description{
Distribution functions for age-at-failure
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{fx}
\alias{fx}
\title{Conditional density}
\usage{
fx(t, x, model, ...)
}
\arguments{
\item{t}{Numeric vector of durations.}

\item{x}{Numeric vector of ages.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}
}
\value{
Numeric vector of density values.
}
\description{
Computes \eqn{f_x(t)=f_0(x+t)/S_0(x)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{A2xn1_m}
\alias{A2xn1_m}
\title{Second moment of m-thly term insurance PV}
\usage{
A2xn1_m(x, n, i, m, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of second moments.
}
\description{
Second moment of m-thly term insurance PV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{Pr_vector_disc}
\alias{Pr_vector_disc}
\title{Profit vector for a discrete profit-analysis model}
\usage{
Pr_vector_disc(
  V,
  G,
  i,
  r = 0,
  e = 0,
  q1,
  q2 = 0,
  b1,
  b2 = 0,
  s1 = 0,
  s2 = 0,
  p_tau = NULL,
  pre_contract_expense = 0
)
}
\arguments{
\item{V}{Numeric vector of gross premium reserves with length
\code{n + 1}, including the issue-time and terminal reserves.}

\item{G}{Gross premium by policy year.}

\item{i}{Annual effective interest rate by policy year. Values must be
greater than \code{-1}.}

\item{r}{Percent-of-premium expense rate by policy year. Values must lie in
\code{[0, 1]}.}

\item{e}{Fixed expense by policy year.}

\item{q1}{Probability of the first decrement by policy year.}

\item{q2}{Probability of the second decrement by policy year.}

\item{b1}{Benefit payable on the first decrement.}

\item{b2}{Benefit payable on the second decrement.}

\item{s1}{Settlement expense associated with the first decrement.}

\item{s2}{Settlement expense associated with the second decrement.}

\item{p_tau}{Optional in-force probability by policy year. If omitted, it is
calculated as \code{1 - q1 - q2}.}

\item{pre_contract_expense}{Nonnegative scalar pre-contract expense.}
}
\value{
A named numeric vector of length \code{n + 1}.
}
\description{
Computes expected profit by policy year for a discrete contract with up to
two decrements. The first element is the negative pre-contract expense.
}
\details{
For policy year \eqn{k}, the expected profit is
\deqn{
[V_{k-1} + G_k(1-r_k)-e_k](1+i_k)
-
[(b_k^{(1)}+s_k^{(1)})q_k^{(1)}
+(b_k^{(2)}+s_k^{(2)})q_k^{(2)}
+V_kp_k^{(\tau)}].
}

Scalar yearly inputs are recycled to the number of policy years determined
by \code{length(V) - 1}.
}
\examples{
V <- c(0, 5.66, 6.17, 0)
qx <- c(0.00142, 0.00153, 0.00166)
Pr_vector_disc(
  V = V, G = 95, i = 0.06, r = 0.05, e = 10,
  q1 = qx, b1 = 50000, pre_contract_expense = 15
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{qxprime_sudd}
\alias{qxprime_sudd}
\title{Associated single-decrement probabilities under SUDD}
\usage{
qxprime_sudd(q1, q2)
}
\arguments{
\item{q1}{Dependent decrement probability for cause 1. May be scalar or
vector.}

\item{q2}{Dependent decrement probability for cause 2. May be scalar or
vector.}
}
\value{
For scalar input, a named numeric vector containing
  \code{q1prime} and \code{q2prime}. For vectorized input, a numeric
  matrix with columns \code{q1prime} and \code{q2prime}.
}
\description{
Converts two dependent multiple-decrement probabilities to associated
single-decrement probabilities under the single-decrement uniform
distribution assumption.
}
\examples{
qxprime_sudd(0.20, 0.10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mortality_improvement.R
\name{mortality_improvement}
\alias{mortality_improvement}
\title{Mortality improvement projection functions (Chapter 8)}
\description{
Functions for Chapter 8 Section 8.8 mortality improvement projection.
}
\details{
These functions project one-year death and survival probabilities forward
from a base year and evaluate annuity values under projected mortality.

The standard projection used is
\deqn{q_x^{[Y]} = q_x^{[B]} (1 - AA_x)^{Y-B}}
where:
\itemize{
  \item \eqn{B} is the base year,
  \item \eqn{Y} is the projection year,
  \item \eqn{AA_x} is the mortality improvement factor at age \eqn{x}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{tqxj_cf}
\alias{tqxj_cf}
\title{Cause-specific decrement probability under constant forces}
\usage{
tqxj_cf(mu, j, t)
}
\arguments{
\item{mu}{Numeric vector of nonnegative cause-specific forces.}

\item{j}{Positive integer cause index.}

\item{t}{Nonnegative time. May be scalar or vector.}
}
\value{
A numeric vector.
}
\description{
Computes
\deqn{
{}_tq_x^{(j)}
=
\frac{\mu_j}{\sum_k\mu_k}
\left[
1-\exp\left(-t\sum_k\mu_k\right)
\right].
}
}
\examples{
tqxj_cf(c(0.10, 0.20), j = 1, t = 5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{multilife_ch12_ext}
\alias{multilife_ch12_ext}
\title{Extended multi-life functions for Chapter 12}
\description{
Additional Chapter 12 functions for contingent probabilities,
continuous contingent insurances, and continuous multi-life annuities.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{APV_NR_db}
\alias{APV_NR_db}
\title{Actuarial present value of a normal retirement benefit}
\usage{
APV_NR_db(PABz, v_to_ret, p_surv, adue_ret)
}
\arguments{
\item{PABz}{Nonnegative projected annual benefit at retirement.}

\item{v_to_ret}{Nonnegative discount factor from the valuation date to
retirement.}

\item{p_surv}{Survival or active-service probability to retirement in
\code{[0, 1]}.}

\item{adue_ret}{Positive retirement annuity-due factor.}
}
\value{
A numeric vector.
}
\description{
Computes the actuarial present value of a projected annual retirement
benefit. Scalar arguments are recycled to a common length.
}
\examples{
APV_NR_db(
  PABz = 108008.66,
  v_to_ret = 1 / 1.06^30,
  p_surv = 0.8,
  adue_ret = 12
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{md_table}
\alias{md_table}
\title{Construct a multiple-decrement table}
\usage{
md_table(x, qxj, radix = 1e+05)
}
\arguments{
\item{x}{Integer vector of consecutive ages or durations.}

\item{qxj}{Numeric matrix or data frame of cause-specific decrement
probabilities. Rows correspond to values in \code{x}, and columns
correspond to causes.}

\item{radix}{Initial value of \eqn{l_x^{(\tau)}}.}
}
\value{
An object of classes \code{"md_table"} and \code{"data.frame"}
  containing age or duration, cause-specific decrement probabilities,
  total decrement and survival probabilities, numbers alive, and
  cause-specific and total decrements.
}
\description{
Constructs a discrete multiple-decrement table from cause-specific
one-year decrement probabilities.
}
\examples{
ages <- 45:50
qmat <- cbind(
  withdrawal = c(0.011, 0.012, 0.013, 0.014, 0.015, 0.016),
  retirement = rep(0.100, 6)
)

md_table(ages, qmat, radix = 1000)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{nqxtau_md}
\alias{nqxtau_md}
\title{Total multiple-decrement probability from a table}
\usage{
nqxtau_md(tbl, x, n)
}
\arguments{
\item{tbl}{A multiple-decrement table produced by \code{md_table()}.}

\item{x}{Starting age or duration. May be scalar or vector.}

\item{n}{Nonnegative integer term. May be scalar or vector.}
}
\value{
A numeric vector of total decrement probabilities.
}
\description{
Computes
\deqn{{}_nq_x^{(\tau)}=1-{}_np_x^{(\tau)}.}
}
\examples{
ages <- 45:50
qmat <- cbind(
  q1 = c(0.011, 0.012, 0.013, 0.014, 0.015, 0.016),
  q2 = rep(0.100, 6)
)
tbl <- md_table(ages, qmat, radix = 1000)

nqxtau_md(tbl, x = 46, n = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{z_from_coupon_annual}
\alias{z_from_coupon_annual}
\title{Bootstrap annual effective spot rates}
\usage{
z_from_coupon_annual(maturity, coupon_yield, par = 1000)
}
\arguments{
\item{maturity}{Numeric vector of positive integer maturities in strictly
increasing order. Maturities must be consecutive and begin at 1.}

\item{coupon_yield}{Numeric vector of annual effective par coupon yields.
Values must be greater than \code{-1}.}

\item{par}{Positive scalar par value.}
}
\value{
A numeric vector of annual effective spot rates.
}
\description{
Bootstraps annual effective spot rates from par coupon yields at consecutive
integer maturities.
}
\examples{
maturity <- 1:4
coupon_yield <- c(0.02, 0.04, 0.06, 0.08)
z_from_coupon_annual(maturity, coupon_yield)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{Axj_md}
\alias{Axj_md}
\title{Discrete multiple-decrement insurance present value}
\usage{
Axj_md(qj, ptau, i, benefit = 1)
}
\arguments{
\item{qj}{Cause-specific decrement probabilities by policy year.}

\item{ptau}{In-force probabilities at the beginning of each policy year.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{benefit}{Benefit payable on decrement. May be scalar or vector.}
}
\value{
A numeric vector of actuarial present values.
}
\description{
Computes the actuarial present value of a benefit payable at the end of
the year of decrement from a specified cause.
}
\examples{
qj <- c(0.02, 0.02, 0.02, 0.02, 0.02)
ptau <- c(1, 0.95, 0.89, 0.82, 0.74)
Axj_md(qj, ptau, i = 0.06, benefit = 1000)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{nExybar}
\alias{nExybar}
\title{Last-survivor pure endowment}
\usage{
nExybar(x, y, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{{}_nE_{\overline{xy}} = v^n\,{}_np_{\overline{xy}}}.
}
\examples{
nExybar(40, 50, n = 10, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{thiele_path_01}
\alias{thiele_path_01}
\title{Backward reserve path for a disability model with recovery}
\usage{
thiele_path_01(
  h,
  n,
  delta,
  Pbar,
  B,
  R,
  mu01,
  mu02,
  mu10,
  mu12,
  V0_n = 0,
  V1_n = 0
)
}
\arguments{
\item{h}{Positive step size.}

\item{n}{Nonnegative final time.}

\item{delta}{Force of interest.}

\item{Pbar}{Continuous premium rate.}

\item{B}{Death benefit.}

\item{R}{Continuous disability income rate.}

\item{mu01}{Healthy-to-disabled intensity function.}

\item{mu02}{Healthy-to-deceased intensity function.}

\item{mu10}{Disabled-to-healthy intensity function.}

\item{mu12}{Disabled-to-deceased intensity function.}

\item{V0_n}{Terminal healthy-state reserve.}

\item{V1_n}{Terminal disabled-state reserve.}
}
\value{
A data frame with columns \code{t}, \code{tV0}, and \code{tV1}.
}
\description{
Computes a backward Euler reserve path from terminal healthy-state and
disabled-state reserves.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_expectations.R
\name{ex_curtate_tab}
\alias{ex_curtate_tab}
\title{Curtate expectation of life from a life table}
\usage{
ex_curtate_tab(tbl, x)
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Numeric vector of integer ages.}
}
\value{
Numeric vector of curtate expectations \eqn{e_x}.
}
\description{
Computes the curtate expectation of life
\eqn{e_x = \sum_{k=1}^{\infty} {}_k p_x}
in the discrete tabular setting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{ELtx}
\alias{ELtx}
\title{Mean present value of loss at duration t for whole life insurance}
\usage{
ELtx(x, t, i, P, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{P}{Annual premium.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object.}
}
\value{
A numeric vector of values.
}
\description{
Computes the conditional mean
\eqn{E[{}_tL_x \mid K_x \ge t]} for a fully discrete whole life insurance.
}
\details{
Under the equivalence-principle premium, this equals the prospective
reserve \eqn{{}_tV_x}.
}
\examples{
prem <- Px(40, i = 0.05, model = "uniform", omega = 100)
ELtx(40, t = 10, i = 0.05, P = prem, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{V_zeroized}
\alias{V_zeroized}
\title{Zeroized reserves for a discrete death-benefit contract}
\usage{
V_zeroized(qx, i, G, benefit, r = 0, e = 0, V_terminal = 0, floor_zero = TRUE)
}
\arguments{
\item{qx}{Mortality probability by policy year.}

\item{i}{Annual effective interest rate by policy year. Values must be
greater than \code{-1}.}

\item{G}{Gross premium by policy year.}

\item{benefit}{Death benefit by policy year.}

\item{r}{Percent-of-premium expense rate by policy year. Values must lie in
\code{[0, 1]}.}

\item{e}{Fixed expense by policy year.}

\item{V_terminal}{Nonnegative scalar terminal reserve.}

\item{floor_zero}{Logical scalar. If \code{TRUE}, negative reserves are
replaced by zero.}
}
\value{
A named numeric vector of length \code{length(qx) + 1}.
}
\description{
Computes reserves backward by setting expected profit in each policy year
equal to zero. Negative reserves may optionally be floored at zero.
}
\examples{
V_zeroized(
  qx = c(0.015, 0.017, 0.019, 0.021, 0.024),
  i = 0.06,
  G = 19279,
  benefit = 1000000,
  e = 240
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{ag38_prefunding_ratio}
\alias{ag38_prefunding_ratio}
\title{AG 38 prefunding ratio}
\usage{
ag38_prefunding_ratio(excess_payment, nsp_required)
}
\arguments{
\item{excess_payment}{Nonnegative excess payment or shadow-fund amount.}

\item{nsp_required}{Positive net single premium required to fully fund the
guarantee.}
}
\value{
A numeric vector.
}
\description{
Computes the excess-payment-to-required-net-single-premium ratio, capped at
one.
}
\examples{
ag38_prefunding_ratio(60000, 100000)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_approximations.R
\name{annuity_approximations}
\alias{annuity_approximations}
\title{Annuity approximations}
\description{
Approximation formulas for m-thly and continuous life annuities.
}
\details{
This file implements:
\itemize{
  \item UDD approximations for m-thly annuities,
  \item UDD approximations for continuous annuities,
  \item Woolhouse 2-term approximations,
  \item Woolhouse 3-term approximations.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{Abarxj_md}
\alias{Abarxj_md}
\title{Continuous multiple-decrement insurance present value}
\usage{
Abarxj_md(t, ptau, muj, delta, benefit = 1)
}
\arguments{
\item{t}{Strictly increasing nonnegative time points.}

\item{ptau}{In-force probabilities at the supplied time points.}

\item{muj}{Cause-specific decrement intensities at the supplied time points.}

\item{delta}{Force of interest. May be scalar or vector.}

\item{benefit}{Benefit payable on decrement. May be scalar or vector.}
}
\value{
A numeric vector of actuarial present values.
}
\description{
Approximates the actuarial present value of a benefit payable at the moment
of decrement using the trapezoidal rule.
}
\examples{
t <- seq(0, 20, by = 0.1)
ptau <- exp(-0.012 * t)
muj <- rep(0.002, length(t))
Abarxj_md(t, ptau, muj, delta = 0.05, benefit = 2000)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{qxprime_mudd}
\alias{qxprime_mudd}
\title{Associated single-decrement probabilities under MUDD}
\usage{
qxprime_mudd(qxj)
}
\arguments{
\item{qxj}{Numeric vector of dependent cause-specific decrement
probabilities.}
}
\value{
A numeric vector of associated single-decrement probabilities.
}
\description{
Converts dependent multiple-decrement probabilities to associated
single-decrement probabilities under the multiple-decrement uniform
distribution assumption.
}
\examples{
qxprime_mudd(c(0.20, 0.10))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_expectations.R
\name{ex_complete_tab}
\alias{ex_complete_tab}
\title{Complete expectation of life from a life table}
\usage{
ex_complete_tab(tbl, x, assumption = c("udd", "cf", "balducci"))
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Numeric vector of integer ages.}

\item{assumption}{One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.}
}
\value{
Numeric vector of complete expectations \eqn{\overset{\circ}{e}_x}.
}
\description{
Computes
\eqn{\overset{\circ}{e}_x = \int_0^\infty {}_t p_x \, dt}
using a within-year assumption:
\code{"udd"}, \code{"cf"}, or \code{"balducci"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_fractional.R
\name{fx_tab}
\alias{fx_tab}
\title{Fractional conditional density from a life table}
\usage{
fx_tab(tbl, x, t, assumption = c("udd", "cf", "balducci"))
}
\arguments{
\item{tbl}{A life_table object.}

\item{x}{Numeric vector of integer ages.}

\item{t}{Numeric vector of fractional durations with \eqn{0 \le t \le 1}.}

\item{assumption}{One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.}
}
\value{
Numeric vector of conditional density values.
}
\description{
Computes the conditional density
\eqn{f_x(t \mid T_0 > x) = {}_t p_x \mu_{x+t}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mqriskR-package.R
\docType{package}
\name{mqriskR-package}
\alias{mqriskR}
\alias{mqriskR-package}
\title{mqriskR: Models for Quantifying Risk}
\description{
Actuarial functions for survival models, life contingencies, reserves,
multiple-decrement models, variable interest, universal life, profit
analysis, and pension mathematics.
}
\author{
\strong{Maintainer}: Nii Okine \email{okinean@appstate.edu}

Authors:
\itemize{
  \item Nii Okine \email{okinean@appstate.edu}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{Abarx}
\alias{Abarx}
\title{Continuous whole life insurance APV}
\usage{
Abarx(x, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes
\eqn{\bar{A}_x = \int_0^\infty v^t {}_t p_x \mu_{x+t}\,dt}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{double_force_i}
\alias{double_force_i}
\title{Effective annual interest at doubled force}
\usage{
double_force_i(i)
}
\arguments{
\item{i}{Numeric vector of effective annual interest rates.}
}
\value{
Numeric vector of effective annual rates corresponding to doubled force.
}
\description{
If \eqn{\delta' = 2\delta}, then
\eqn{i' = (1+i)^2 - 1}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{Vprefloor_crvm_ul}
\alias{Vprefloor_crvm_ul}
\title{Pre-floor CRVM reserve}
\usage{
Vprefloor_crvm_ul(r, pvfb_minus_pvfp)
}
\arguments{
\item{r}{Funding ratio. Values must lie in \code{[0, 1]}.}

\item{pvfb_minus_pvfp}{Numeric difference between the present value of
future benefits and future premiums.}
}
\value{
A numeric vector.
}
\description{
Computes the pre-floor reserve by multiplying the funding ratio by the
difference between the present value of future benefits and future
premiums.
}
\examples{
Vprefloor_crvm_ul(r = 0.33506, pvfb_minus_pvfp = 70)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_fractional.R
\name{mux_tab}
\alias{mux_tab}
\title{Fractional force of mortality from a life table}
\usage{
mux_tab(tbl, x, t, assumption = c("udd", "cf", "balducci"))
}
\arguments{
\item{tbl}{A life_table object.}

\item{x}{Numeric vector of integer ages.}

\item{t}{Numeric vector of fractional durations with \eqn{0 \le t \le 1}.}

\item{assumption}{One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.}
}
\value{
Numeric vector of \eqn{\mu_{x+t}} values.
}
\description{
Computes \eqn{\mu_{x+t}} under UDD, constant force, or Balducci.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{pv_spot_cashflows}
\alias{pv_spot_cashflows}
\title{Present value of deterministic cash flows using spot rates}
\usage{
pv_spot_cashflows(
  amounts,
  times,
  spot,
  compounding = c("annual", "semiannual")
)
}
\arguments{
\item{amounts}{Numeric vector of cash-flow amounts.}

\item{times}{Numeric vector of payment times in years.}

\item{spot}{Numeric vector of spot rates matched elementwise to
\code{times}. The value corresponding to a time-zero cash flow is ignored.}

\item{compounding}{Character string equal to \code{"annual"} or
\code{"semiannual"}.}
}
\value{
A numeric scalar.
}
\description{
Discounts each cash flow using the spot rate matched to its payment time.
Time-zero cash flows are not discounted.
}
\examples{
pv_spot_cashflows(
  amounts = c(200000, 50000, 50000, 100000),
  times = c(0, 0.5, 1, 2),
  spot = c(0, 0.02440, 0.02601, 0.02936),
  compounding = "semiannual"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{z_from_coupon_semi}
\alias{z_from_coupon_semi}
\title{Bootstrap semiannual nominal spot rates}
\usage{
z_from_coupon_semi(maturity, coupon_yield, par = 1000)
}
\arguments{
\item{maturity}{Numeric vector of positive maturities in years, in strictly
increasing order. Maturities must be consecutive multiples of
\code{0.5}.}

\item{coupon_yield}{Numeric vector of nominal annual par coupon yields
convertible semiannually. Values must be greater than \code{-2}.}

\item{par}{Positive scalar par value.}
}
\value{
A numeric vector of nominal annual spot rates convertible
  semiannually.
}
\description{
Bootstraps nominal annual spot rates convertible semiannually from par
coupon yields at consecutive half-year maturities.
}
\examples{
maturity <- c(0.5, 1.0, 1.5, 2.0)
coupon_yield <- c(0.0244, 0.0260, 0.0276, 0.0293)
z_from_coupon_semi(maturity, coupon_yield)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{var_Abarxn}
\alias{var_Abarxn}
\title{Variance of continuous endowment insurance PV}
\usage{
var_Abarxn(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of variances.
}
\description{
Variance of continuous endowment insurance PV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_core.R
\name{nqx}
\alias{nqx}
\title{Compute n-year death probability from a life table}
\usage{
nqx(tbl, x, n)
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Ages.}

\item{n}{Nonnegative integer durations.}
}
\value{
Numeric vector of \eqn{{}_n q_x} values.
}
\description{
Compute n-year death probability from a life table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{A2barxn1}
\alias{A2barxn1}
\title{Second moment of continuous term insurance PV}
\usage{
A2barxn1(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of second moments.
}
\description{
Computes \eqn{{}^{2}\bar{A}_{x:\overline{n}|}^{1}} by evaluating
\eqn{\bar{A}_{x:\overline{n}|}^{1}} at doubled force.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interest_theory.R
\name{pv_cashflows}
\alias{pv_cashflows}
\title{Present value of cash flows at time 0}
\usage{
pv_cashflows(cf, t, i)
}
\arguments{
\item{cf}{Cash flow amounts.}

\item{t}{Times of cash flows.}

\item{i}{Effective interest rate.}
}
\value{
Present value at time 0.
}
\description{
Present value of cash flows at time 0
}
\examples{
pv_cashflows(c(-100, 60, 60), c(0, 1, 2), i = 0.10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{nAx}
\alias{nAx}
\title{Deferred insurance APV}
\usage{
nAx(x, n, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000)
}
\arguments{
\item{x}{Age.}

\item{n}{Deferral period.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}

\item{tol}{Numerical tolerance for truncating infinite sums.}

\item{k_max}{Maximum number of terms in the sum.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes \eqn{{}_{n\mid}A_x = {}_nE_x \cdot A_{x+n}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{deferred_insurance_reserves}
\alias{deferred_insurance_reserves}
\alias{tVnAx}
\alias{htVnAx}
\title{Deferred insurance reserves}
\usage{
tVnAx(x, n, t, i, model = NULL, ..., tbl = NULL)

htVnAx(x, n, h, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{n}{Deferral period in years.}

\item{t}{Duration in years.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to the underlying actuarial
functions.}

\item{tbl}{Optional life table object.}

\item{h}{Premium-paying period in years.}
}
\value{
A numeric vector of prospective reserve values.
}
\description{
Computes prospective reserves for deferred whole life insurance contracts.
}
\details{
\code{tVnAx()} computes the reserve at duration \code{t} for an
\code{n}-year deferred whole life insurance funded by level annual
premiums during the deferral period.

\code{htVnAx()} computes the corresponding reserve when premiums are
limited to the first \code{h} years, where \code{h <= n}.
}
\examples{
tVnAx(
  x = 40, n = 20, t = 10, i = 0.05,
  model = "uniform", omega = 100
)

htVnAx(
  x = 40, n = 20, h = 10, t = 5, i = 0.05,
  model = "uniform", omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{NC_PUC_db}
\alias{NC_PUC_db}
\title{Projected Unit Credit normal cost}
\usage{
NC_PUC_db(projected_benefit, total_service, v_to_ret, p_surv, adue_ret)
}
\arguments{
\item{projected_benefit}{Nonnegative projected annual benefit at retirement.}

\item{total_service}{Positive total service at retirement.}

\item{v_to_ret}{Nonnegative discount factor from the valuation date to
retirement.}

\item{p_surv}{Survival or active-service probability to retirement in
\code{[0, 1]}.}

\item{adue_ret}{Positive retirement annuity-due factor.}
}
\value{
A numeric vector.
}
\description{
Computes the actuarial present value of the portion of the projected
benefit attributed to the current year of service.
}
\examples{
NC_PUC_db(
  projected_benefit = 30000,
  total_service = 30,
  v_to_ret = 0.5,
  p_surv = 0.9,
  adue_ret = 12
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{iMA_eiul}
\alias{iMA_eiul}
\title{Monthly-average index growth rate}
\usage{
iMA_eiul(index)
}
\arguments{
\item{index}{Numeric vector of length 13 containing a strictly positive
initial index value followed by twelve nonnegative monthly closing values.}
}
\value{
A numeric scalar.
}
\description{
Computes the monthly-average growth rate from an initial index value and
twelve monthly closing values.
}
\examples{
index <- c(
  1000, 1020, 1100, 1150, 1080, 1040, 960,
  1030, 1000, 1070, 1150, 1200, 1150
)
iMA_eiul(index)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{contribution_rate_target}
\alias{contribution_rate_target}
\title{Target contribution rate for a defined contribution plan}
\usage{
contribution_rate_target(x, z, Sx, RR_target, i, adue_z, g = NULL, s = NULL)
}
\arguments{
\item{x}{Scalar entry age.}

\item{z}{Scalar retirement age.}

\item{Sx}{Positive scalar salary at age \code{x}.}

\item{RR_target}{Scalar target replacement ratio in \code{[0, 1]}.}

\item{i}{Scalar annual effective investment return greater than \code{-1}.}

\item{adue_z}{Positive scalar whole-life annuity-due factor at retirement.}

\item{g}{Optional scalar annual salary growth rate greater than \code{-1}.}

\item{s}{Optional positive salary-scale vector of length \code{z - x}.}
}
\value{
A numeric scalar. The result may exceed one when the target cannot
  be achieved with a contribution rate no greater than 100 percent.
}
\description{
Calculates the contribution rate required to achieve a target replacement
ratio.
}
\examples{
contribution_rate_target(
  x = 30,
  z = 65,
  Sx = 60000,
  RR_target = 0.50,
  i = 0.06,
  adue_z = 11,
  g = 0.04
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{AB_fas}
\alias{AB_fas}
\title{Accrued benefit under a final-average-salary plan}
\usage{
AB_fas(salary_history, p, fas_years = 3)
}
\arguments{
\item{salary_history}{Positive numeric vector of annual salaries.}

\item{p}{Nonnegative scalar accrual percentage.}

\item{fas_years}{Positive integer number of years in the salary average.}
}
\value{
A numeric scalar.
}
\description{
Computes the accrued benefit using salary and service history through the
valuation date.
}
\examples{
AB_fas(
  salary_history = c(150000, 156000),
  p = 1,
  fas_years = 2
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{var_nAbarx}
\alias{var_nAbarx}
\title{Variance of continuous deferred insurance PV}
\usage{
var_nAbarx(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Deferral period.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of variances.
}
\description{
Variance of continuous deferred insurance PV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_varying_benefits.R
\name{DbarAbarxn1}
\alias{DbarAbarxn1}
\title{Fully continuous decreasing n-year term insurance}
\usage{
DbarAbarxn1(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes
\deqn{(\bar{D}\bar{A})_{x:\overline{n}|}^{1}
= \int_0^n (n-t) v^t {}_tp_x \mu_{x+t}\,dt.}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{coi_ul_typeB}
\alias{coi_ul_typeB}
\title{Cost of insurance for Type B universal life}
\usage{
coi_ul_typeB(B, qx, iq)
}
\arguments{
\item{B}{Face amount.}

\item{qx}{Mortality probability for the period.}

\item{iq}{Interest rate used in the cost-of-insurance calculation. Values
must be greater than \code{-1}.}
}
\value{
A numeric vector of cost-of-insurance charges.
}
\description{
Computes the one-period cost of insurance for a Type B universal life
contract:
\deqn{
\mathrm{COI}_t = \frac{B_t q_{x+t-1}}{1+i_t^q}.
}
}
\details{
The arguments may be scalars or vectors. Scalar arguments are recycled to
the common length.
}
\examples{
coi_ul_typeB(B = 100000, qx = 0.00076, iq = 0.03)
coi_ul_typeB(
  B = 100000,
  qx = c(0.00076, 0.00081),
  iq = c(0.03, 0.035)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_conversions.R
\name{qx_to_lx}
\alias{qx_to_lx}
\title{Construct life-table values from q_x values}
\usage{
qx_to_lx(qx, radix = 1e+05)
}
\arguments{
\item{qx}{Numeric vector of one-year death probabilities \eqn{q_x}.}

\item{radix}{Positive radix \eqn{l_0}.}
}
\value{
Numeric vector of \eqn{l_x} values of length \code{length(qx) + 1}.
}
\description{
Builds life-table survivor values recursively from
\eqn{l_{x+1} = l_x (1 - q_x)}, starting from a chosen radix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{nAx_m}
\alias{nAx_m}
\title{m-thly deferred insurance APV}
\usage{
nAx_m(x, n, i, m, model, ..., tol = 1e-12, j_max = 100000L)
}
\arguments{
\item{x}{Age.}

\item{n}{Deferral period.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}

\item{tol}{Numerical tolerance for truncating the infinite sum.}

\item{j_max}{Maximum number of m-thly intervals in the sum.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes
\eqn{{}_{n\mid}A_x^{(m)} = v^n {}_np_x A_{x+n}^{(m)}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_conversions.R
\name{lx_to_S0}
\alias{lx_to_S0}
\title{Convert life-table values to survival probabilities}
\usage{
lx_to_S0(lx)
}
\arguments{
\item{lx}{Numeric vector of life-table survivor values.}
}
\value{
Numeric vector of \eqn{S_0(x)} values.
}
\description{
Converts life-table survivor values \eqn{l_x} into survival probabilities
\eqn{S_0(x) = l_x / l_0}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{Abarxn_udd}
\alias{Abarxn_udd}
\title{UDD approximation of continuous endowment insurance}
\usage{
Abarxn_udd(Axn1, nEx, i)
}
\arguments{
\item{Axn1}{Discrete term insurance APV.}

\item{nEx}{Pure endowment APV, \eqn{{}_nE_x}.}

\item{i}{Effective annual interest rate.}
}
\value{
Continuous endowment insurance APV under UDD.
}
\description{
Computes
\eqn{\bar{A}_{x:\overline{n}|} = (i/\delta)A_{x:\overline{n}|}^{1} + {}_nE_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{htVx}
\alias{htVx}
\title{h-pay whole life net level premium reserve}
\usage{
htVx(x, h, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{h}{Premium-paying period in years.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object.}
}
\value{
A numeric vector of values.
}
\description{
Computes the prospective reserve for an h-pay whole life policy.
}
\examples{
htVx(40, h = 10, t = 5, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{thiele_dVdt}
\alias{thiele_dVdt}
\title{Reserve derivative from Thiele's equation}
\usage{
thiele_dVdt(V, P, delta, mu, benefit = 1)
}
\arguments{
\item{V}{Reserve at time `t`.}

\item{P}{Premium rate.}

\item{delta}{Force of interest.}

\item{mu}{Nonnegative force of mortality.}

\item{benefit}{Benefit amount.}
}
\value{
Numeric vector of reserve derivatives.
}
\description{
Computes
\deqn{
\frac{dV}{dt}
=
P + \delta V - \mu(B - V).
}
}
\examples{
thiele_dVdt(
  V = 900,
  P = 25,
  delta = 0.05,
  mu = 0.002,
  benefit = 1000
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interest_theory.R
\name{solve_yield}
\alias{solve_yield}
\title{Solve the yield rate by the equation of value}
\usage{
solve_yield(cf, t, interval = c(-0.99, 1), tol = 1e-10)
}
\arguments{
\item{cf}{Cash flows.}

\item{t}{Times.}

\item{interval}{Two-length numeric vector bracketing the root.}

\item{tol}{Tolerance passed to `uniroot()`.}
}
\value{
Yield rate `i`.
}
\description{
Finds the interest rate `i` such that the present value of the cash flows is 0.
}
\examples{
solve_yield(c(-100, 60, 60), c(0, 1, 2), interval = c(-0.5, 1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{z_from_fn1}
\alias{z_from_fn1}
\title{Spot rates from one-year forward rates}
\usage{
z_from_fn1(fn1)
}
\arguments{
\item{fn1}{Numeric vector of annual effective one-year forward rates. Each
value must be greater than \code{-1}.}
}
\value{
A numeric vector of annual effective spot rates.
}
\description{
Converts annual effective one-year forward rates
\eqn{f_{0,1},f_{1,1},\ldots,f_{n-1,1}} into annual effective spot rates:
\deqn{
(1+z_n)^n
=
\prod_{j=0}^{n-1}(1+f_{j,1}).
}
}
\examples{
z_from_fn1(c(0.04, 0.05, 0.06, 0.07, 0.08))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{cumhaz0}
\alias{cumhaz0}
\title{Cumulative hazard for age-at-failure}
\usage{
cumhaz0(t, model, ...)
}
\arguments{
\item{t}{Numeric vector of times.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{\Lambda_0(t)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{spot_interest_apvs}
\alias{spot_interest_apvs}
\alias{nEx_spot}
\alias{Axn1_spot}
\alias{Axn_spot}
\alias{axn_spot}
\title{Actuarial present values under spot rates}
\usage{
nEx_spot(qx, z, benefit = 1)

Axn1_spot(qx, z, benefit = 1)

Axn_spot(qx, z, benefit = 1)

axn_spot(qx, z, type = c("immediate", "due"), benefit = 1)
}
\arguments{
\item{qx}{Numeric vector of one-year mortality probabilities.}

\item{z}{Numeric vector of annual effective spot rates for maturities
\code{1, \dots, n}. Each value must be greater than \code{-1}.}

\item{benefit}{Nonnegative scalar benefit or annuity payment amount.}

\item{type}{Character string equal to \code{"immediate"} or
\code{"due"}.}
}
\value{
A numeric scalar.
}
\description{
Computes life-contingent actuarial present values using annual effective
spot rates by maturity.
}
\details{
If \eqn{z_t} denotes the annual effective spot rate for maturity
\eqn{t}, the corresponding discount factor is
\deqn{(1+z_t)^{-t}.}

\code{nEx_spot()} computes a pure endowment.

\code{Axn1_spot()} computes term insurance payable at the end of the year
of death.

\code{Axn_spot()} computes endowment insurance.

\code{axn_spot()} computes a temporary annuity-immediate or
annuity-due.


Each payment is discounted using the spot rate corresponding to its
maturity rather than a single level interest rate.

The pure endowment is
\deqn{
{}_nE
=
{}_np_x(1+z_n)^{-n}.
}

Term insurance is obtained by discounting each death benefit using the
spot rate corresponding to its payment year.

Endowment insurance equals the sum of the corresponding term insurance
and pure endowment.

Temporary annuities discount each payment using the spot rate for its
payment time.
}
\examples{
qx <- c(0.02, 0.03, 0.04, 0.05, 0.06)
spot <- c(0.03, 0.04, 0.05, 0.06, 0.07)

nEx_spot(qx, spot, benefit = 1000)
Axn1_spot(qx, spot)
Axn_spot(qx, spot)
axn_spot(qx, spot, type = "due")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{abarxybar_ch12}
\alias{abarxybar_ch12}
\alias{abarxybar}
\title{Continuous last-survivor whole life annuity}
\usage{
abarxybar(x, y, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{\overline{a}_{\overline{xy}} = \overline{a}_x + \overline{a}_y - \overline{a}_{xy}}.
}
\details{
Shared documentation topic used to avoid filename collisions with
case-distinct function names on case-insensitive file systems.
}
\examples{
abarxybar(40, 50, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{tpxtau_ul}
\alias{tpxtau_ul}
\title{Cumulative persistency to the end of each policy year}
\usage{
tpxtau_ul(qd, qw, year_end_withdrawal = TRUE)
}
\arguments{
\item{qd}{Mortality probabilities.}

\item{qw}{Withdrawal probabilities.}

\item{year_end_withdrawal}{Logical; if \code{TRUE}, use Equation (16.15).}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{{}_tp_x^{(\tau)}} from the one-year persistency rates.
}
\examples{
qd <- c(.001, .002, .003)
qw <- c(.02, .02, .03)
tpxtau_ul(qd, qw)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{Ax_m}
\alias{Ax_m}
\title{m-thly whole life insurance APV}
\usage{
Ax_m(x, i, m, model, ..., tol = 1e-12, j_max = 100000L)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}

\item{tol}{Numerical tolerance for truncating the infinite sum.}

\item{j_max}{Maximum number of m-thly intervals in the sum.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes
\eqn{A_x^{(m)} =
\sum_{j=0}^{\infty} v^{(j+1)/m}
\Pr(j/m < T_x \le (j+1)/m)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{A2xn_m}
\alias{A2xn_m}
\title{Second moment of m-thly endowment insurance PV}
\usage{
A2xn_m(x, n, i, m, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of second moments.
}
\description{
Second moment of m-thly endowment insurance PV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mortality_improvement.R
\name{tpx_improved}
\alias{tpx_improved}
\title{Multi-year survival probability under mortality improvement}
\usage{
tpx_improved(x0, n, qx_base_vec, AAx_vec, base_year, issue_year)
}
\arguments{
\item{x0}{Issue age.}

\item{n}{Number of years.}

\item{qx_base_vec}{Base-year one-year death probabilities for successive ages.}

\item{AAx_vec}{Mortality improvement factors for successive ages.}

\item{base_year}{Base year.}

\item{issue_year}{Issue year.}
}
\value{
Numeric scalar.
}
\description{
Computes survival over \code{n} years starting at age \code{x0} in year
\code{issue_year}, using base-year mortality \code{qx_base_vec} and
improvement factors \code{AAx_vec}.
}
\details{
The vectors should correspond to ages
\code{x0, x0+1, ..., x0+n-1}.

The survival probability is
\deqn{\prod_{j=0}^{n-1} \left(1 - q_{x+j}^{[issue\_year+j]}\right)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{cov_term_endow}
\alias{cov_term_endow}
\title{Covariance of term insurance and pure endowment PVs}
\usage{
cov_term_endow(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}
}
\value{
Numeric vector of covariances.
}
\description{
Computes
\eqn{\mathrm{Cov}(Z_{x:\overline{n}|}^{1},
Z_{x:\overline{n}|}^{1\text{(pure endow)}}) =
-A_{x:\overline{n}|}^{1} \cdot {}_nE_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{tp00_tp01_euler}
\alias{tp00_tp01_euler}
\title{Euler approximation of disability-state probabilities}
\usage{
tp00_tp01_euler(h, n, mu01, mu02, mu10, mu12, p00_0 = 1, p01_0 = 0)
}
\arguments{
\item{h}{Positive step size.}

\item{n}{Nonnegative final time.}

\item{mu01}{Healthy-to-disabled intensity function.}

\item{mu02}{Healthy-to-deceased intensity function.}

\item{mu10}{Disabled-to-healthy intensity function.}

\item{mu12}{Disabled-to-deceased intensity function.}

\item{p00_0}{Initial healthy-state probability.}

\item{p01_0}{Initial disabled-state probability.}
}
\value{
A data frame with columns \code{t}, \code{tp00}, \code{tp01},
  and \code{tp02}.
}
\description{
Approximates probabilities of being healthy, disabled, or deceased in a
three-state model that allows recovery from disability. The final time is
always included, with a shorter final step when needed.
}
\examples{
mu01 <- function(t) 0.10 * t + 0.20
mu02 <- function(t) 0.20
mu10 <- function(t) 0.50
mu12 <- function(t) 0.125 * t + 0.20
tp00_tp01_euler(0.10, 2, mu01, mu02, mu10, mu12)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R
\name{fractional_duration_term_endowment_reserves}
\alias{fractional_duration_term_endowment_reserves}
\alias{tsVxn}
\alias{tsVxn1}
\title{Fractional-duration term and endowment reserves}
\usage{
tsVxn(x, n, t, s, i, tbl = NULL, model = NULL, ...)

tsVxn1(x, n, t, s, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Issue age. May be scalar or vector.}

\item{n}{Positive integer term. May be scalar or vector.}

\item{t}{Nonnegative integer duration satisfying \eqn{t < n}. May be
scalar or vector.}

\item{s}{Fractional duration in \eqn{[0,1]}. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the actuarial functions.}
}
\value{
A numeric vector of reserve values.
}
\description{
Computes fractional-duration reserves for term and endowment insurance
using linear interpolation between the reserve immediately after the
premium at duration \eqn{t} and the reserve at duration \eqn{t+1}.
}
\details{
\code{tsVxn()} computes the fractional-duration reserve for endowment
insurance.

\code{tsVxn1()} computes the fractional-duration reserve for term
insurance.


The fractional reserve is computed using

\deqn{
{}_{t+s}V
=
({}_tV + P)(1-s)
+
{}_{t+1}V s.
}

The premium \eqn{P} is the net annual premium for the corresponding
insurance contract (term or endowment).
}
\examples{
tsVxn(
  40,
  n = 20,
  t = 10,
  s = 0.5,
  i = 0.05,
  model = "uniform",
  omega = 100
)

tsVxn1(
  40,
  n = 20,
  t = 10,
  s = 0.5,
  i = 0.05,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{tpxprimej_cf}
\alias{tpxprimej_cf}
\title{Single-decrement survival under a constant force}
\usage{
tpxprimej_cf(mu, t)
}
\arguments{
\item{mu}{Nonnegative constant force of decrement. May be scalar or
vector.}

\item{t}{Nonnegative time. May be scalar or vector.}
}
\value{
A numeric vector.
}
\description{
Computes
\deqn{{}_tp_x^{\prime(j)}=\exp(-\mu_jt).}
}
\examples{
tpxprimej_cf(0.10, 5)
tpxprimej_cf(0.10, c(1, 5, 10))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{pxtau_ul}
\alias{pxtau_ul}
\alias{tpxtau_ul}
\title{Universal life persistency probabilities}
\usage{
pxtau_ul(qd, qw, year_end_withdrawal = TRUE)

tpxtau_ul(qd, qw, year_end_withdrawal = TRUE)
}
\arguments{
\item{qd}{Mortality probabilities.}

\item{qw}{Withdrawal probabilities.}

\item{year_end_withdrawal}{Logical scalar. If \code{TRUE}, withdrawal is
modeled at year-end and persistency is
\eqn{(1-q^{(d)})(1-q^{(w)})}. Otherwise persistency is
\eqn{1-q^{(d)}-q^{(w)}}.}
}
\value{
A numeric vector.
}
\description{
\code{pxtau_ul()} computes one-year persistency probabilities under
mortality and withdrawal.
}
\details{
\code{tpxtau_ul()} computes cumulative persistency through the end of each
policy year.
}
\examples{
qd <- c(0.001, 0.002, 0.003)
qw <- c(0.02, 0.02, 0.03)

pxtau_ul(qd, qw)
tpxtau_ul(qd, qw)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mortality_improvement.R
\name{qx_proj}
\alias{qx_proj}
\title{Project one-year death probability under mortality improvement}
\usage{
qx_proj(qx_base, AAx, base_year, proj_year)
}
\arguments{
\item{qx_base}{Base-year one-year death probability \eqn{q_x^{[B]}}.}

\item{AAx}{Mortality improvement factor \eqn{AA_x}.}

\item{base_year}{Base year \eqn{B}.}

\item{proj_year}{Projection year \eqn{Y}. May be scalar or vector.}
}
\value{
Numeric vector of projected one-year death probabilities.
}
\description{
Computes projected one-year death probability
\deqn{q_x^{[Y]} = q_x^{[B]} (1-AA_x)^{Y-B}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{thiele_backward_step}
\alias{thiele_backward_step}
\title{One backward numerical step for Thiele's equation}
\usage{
thiele_backward_step(V_next, P, delta, mu, benefit = 1, h = 1)
}
\arguments{
\item{V_next}{Reserve at time `t + h`.}

\item{P}{Premium rate.}

\item{delta}{Force of interest.}

\item{mu}{Nonnegative force of mortality at time `t`.}

\item{benefit}{Benefit amount.}

\item{h}{Positive step size.}
}
\value{
Numeric vector of reserve approximations.
}
\description{
Approximates the reserve at time `t` from a known reserve at time
`t + h`.
}
\details{
The implemented step is
\deqn{
V_t =
\frac{
V_{t+h} - hP + h\mu B
}{
1 + h(\delta + \mu)
}.
}
}
\examples{
thiele_backward_step(
  V_next = 1000,
  P = 26.96,
  delta = 0.058,
  mu = 0.002,
  benefit = 1000,
  h = 1
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{dxj}
\alias{dxj}
\title{Cause-specific numbers of decrements}
\usage{
dxj(lxtau, qxj)
}
\arguments{
\item{lxtau}{Number alive at age or duration \eqn{x} in the
multiple-decrement table.}

\item{qxj}{Numeric vector of cause-specific decrement probabilities.}
}
\value{
A numeric vector containing the number of decrements from each
  cause.
}
\description{
Computes
\deqn{d_x^{(j)}=l_x^{(\tau)}q_x^{(j)}.}
}
\examples{
dxj(1000, c(0.011, 0.100))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{insurance_continuous}
\alias{insurance_continuous}
\title{Continuous insurance models}
\description{
Continuous contingent payment and insurance functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{dxtau}
\alias{dxtau}
\title{Total number of decrements}
\usage{
dxtau(lxtau, qxj)
}
\arguments{
\item{lxtau}{Number alive at age or duration \eqn{x} in the
multiple-decrement table.}

\item{qxj}{Numeric vector of cause-specific decrement probabilities.}
}
\value{
A numeric scalar.
}
\description{
Computes
\deqn{d_x^{(\tau)}=\sum_j d_x^{(j)}.}
}
\examples{
dxtau(1000, c(0.011, 0.100))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_varying_payments.R
\name{annuity_varying_payments}
\alias{annuity_varying_payments}
\alias{Iax}
\alias{Iaxn}
\alias{Daxn}
\alias{Iadotx}
\alias{Iadotxn}
\alias{Dadotxn}
\alias{Iabarx}
\alias{Iabarxn}
\alias{Dabarxn}
\title{Varying-payment annuity functions}
\usage{
Iax(x, i, model = NULL, ..., tbl = NULL, k_max = 5000, tol = 1e-12)

Iaxn(x, n, i, model = NULL, ..., tbl = NULL)

Daxn(x, n, i, model = NULL, ..., tbl = NULL)

Iadotx(x, i, model = NULL, ..., tbl = NULL, k_max = 5000, tol = 1e-12)

Iadotxn(x, n, i, model = NULL, ..., tbl = NULL)

Dadotxn(x, n, i, model = NULL, ..., tbl = NULL)

Iabarx(x, i, model, ..., tol = 1e-10)

Iabarxn(x, n, i, model, ...)

Dabarxn(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object for discrete functions.}

\item{k_max}{Maximum summation horizon for non-terminating models.}

\item{tol}{Truncation tolerance for non-terminating models.}

\item{n}{Term in years.}
}
\value{
Numeric vector containing the requested increasing or decreasing annuity value.
}
\description{
Increasing and decreasing life annuity functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{joint_life_annuities}
\alias{joint_life_annuities}
\alias{adotxyn}
\alias{axyn}
\alias{adotxy}
\alias{axy}
\title{Joint-life annuities}
\usage{
adotxyn(x, y, n, i, tbl = NULL, model = NULL, ...)

axyn(x, y, n, i, tbl = NULL, model = NULL, ...)

adotxy(x, y, i, tbl = NULL, model = NULL, ..., k_max = 5000L, tol = 1e-12)

axy(x, y, i, tbl = NULL, model = NULL, ..., k_max = 5000L, tol = 1e-12)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{n}{Nonnegative integer term. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the survival model.}

\item{k_max}{Maximum summation horizon used for non-terminating parametric
survival models.}

\item{tol}{Positive convergence tolerance for whole-life summations.}
}
\value{
A numeric vector of annuity actuarial present values.
}
\description{
Computes temporary and whole-life joint-life annuities-due and
annuities-immediate for two independent lives.
}
\details{
\code{adotxyn()} computes an \eqn{n}-year temporary joint-life
annuity-due.

\code{axyn()} computes an \eqn{n}-year temporary joint-life
annuity-immediate.

\code{adotxy()} computes a whole-life joint-life annuity-due.

\code{axy()} computes a whole-life joint-life annuity-immediate.


All calculations assume independence between the two future lifetimes.

The temporary annuity-due is
\deqn{
\ddot{a}_{xy:\overline{n}|}
=
\sum_{k=0}^{n-1} v^k\,{}_kp_{xy}.
}

The temporary annuity-immediate is
\deqn{
a_{xy:\overline{n}|}
=
\sum_{k=1}^{n} v^k\,{}_kp_{xy}.
}

For life-table calculations, the whole-life sums terminate at the last
duration supported jointly by the two lives. For parametric models, the
sums are evaluated until convergence or until \code{k_max} is reached.
}
\examples{
adotxyn(
  x = 40,
  y = 50,
  n = 10,
  i = 0.05,
  model = "uniform",
  omega = 100
)

axyn(
  x = 40,
  y = 50,
  n = 10,
  i = 0.05,
  model = "uniform",
  omega = 100
)

adotxy(
  x = 40,
  y = 50,
  i = 0.05,
  model = "uniform",
  omega = 100
)

axy(
  x = 40,
  y = 50,
  i = 0.05,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{insurance_discrete}
\alias{insurance_discrete}
\title{Discrete insurance models}
\description{
Discrete contingent payment and insurance functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{hazard0}
\alias{hazard0}
\title{Hazard or force of mortality for age-at-failure}
\usage{
hazard0(t, model, ...)
}
\arguments{
\item{t}{Numeric vector of times.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{\lambda_0(t)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_core.R
\name{ndx}
\alias{ndx}
\title{Compute deaths over an n-year interval from a life table}
\usage{
ndx(tbl, x, n)
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Ages.}

\item{n}{Nonnegative integer durations.}
}
\value{
Numeric vector of \eqn{{}_n d_x} values.
}
\description{
Compute deaths over an n-year interval from a life table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{udd_continuous_multiplier}
\alias{udd_continuous_multiplier}
\title{UDD multiplier for continuous insurance approximations}
\usage{
udd_continuous_multiplier(i)
}
\arguments{
\item{i}{Numeric vector of effective annual interest rates.}
}
\value{
Numeric vector equal to \eqn{i/\delta}.
}
\description{
Under UDD,
\eqn{\bar{A}_x = (i/\delta) A_x}
and similarly for term and deferred insurance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{continuous_multilife_annuities}
\alias{continuous_multilife_annuities}
\alias{abarxy}
\alias{abarxybar}
\alias{abarx_y}
\alias{abary_x}
\title{Continuous multi-life annuities}
\usage{
abarxy(x, y, i, tbl = NULL, model = NULL, ...)

abarxybar(x, y, i, tbl = NULL, model = NULL, ...)

abarx_y(x, y, i, tbl = NULL, model = NULL, ...)

abary_x(x, y, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object retained for backward compatibility.
Continuous calculations currently require \code{model}.}

\item{model}{Parametric survival model.}

\item{...}{Additional parameters passed to the survival functions.}
}
\value{
A numeric vector of actuarial present values.
}
\description{
Computes continuous joint-life, last-survivor, and reversionary whole-life
annuities for two independent lives.
}
\details{
\code{abarxy()} computes the joint-life annuity.

\code{abarxybar()} computes the last-survivor annuity.

\code{abarx_y()} computes the reversionary annuity payable to the second
life after the death of the first.

\code{abary_x()} computes the reversionary annuity payable to the first
life after the death of the second.

These functions require a parametric survival model.


Under independence, \code{abarxy()} represents the continuous joint-life
annuity, payable while both lives survive.

The last-survivor annuity satisfies
\deqn{\bar{a}_{\overline{xy}} =
\bar{a}_x + \bar{a}_y - \bar{a}_{xy}.}

The reversionary annuities are obtained as the difference between the
corresponding single-life annuity and the joint-life annuity.
}
\examples{
abarxy(
  x = 40,
  y = 50,
  i = 0.05,
  model = "uniform",
  omega = 100
)

abarxybar(
  x = 40,
  y = 50,
  i = 0.05,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R, R/reserve_ch11_ext.R
\name{tsVxn1}
\alias{tsVxn1}
\title{Fractional-duration term reserve}
\usage{
tsVxn1(x, n, t, s, i, tbl = NULL, model = NULL, ...)

tsVxn1(x, n, t, s, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Issue age.}

\item{n}{Term in years.}

\item{t}{Integer duration with t < n.}

\item{s}{Fractional part in [0, 1].}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.

Numeric vector.
}
\description{
Computes the interpolated reserve
\eqn{{}_{t+s}V = ({}_tV + P)(1-s) + {}_{t+1}V \cdot s}
for an n-year term insurance.

Computes the interpolated reserve
\eqn{{}_{t+s}V = ({}_tV + P)(1-s) + {}_{t+1}V \cdot s}
for an n-year term insurance.
}
\examples{
tsVxn1(40, n = 20, t = 10, s = 0.5, i = 0.05, model = "uniform", omega = 100)
tsVxn1(40, n = 20, t = 10, s = 0.5, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R, R/reserve_ch11_ext.R
\name{meanVx}
\alias{meanVx}
\title{Mean reserve for whole life insurance}
\usage{
meanVx(x, t, i, tbl = NULL, model = NULL, ...)

meanVx(x, t, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Integer contract duration.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.

Numeric vector.
}
\description{
Computes the mean reserve
\eqn{{}_{t+1/2}V}.

Computes the mean reserve
\eqn{{}_{t+1/2}V}.
}
\examples{
meanVx(40, t = 10, i = 0.05, model = "uniform", omega = 100)
meanVx(40, t = 10, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{NPV_partial}
\alias{NPV_partial}
\title{Partial net present values}
\usage{
NPV_partial(Pi, r)
}
\arguments{
\item{Pi}{Numeric profit-signature vector.}

\item{r}{Annual effective risk discount rate. May be scalar or vector;
values must be greater than \code{-1}.}
}
\value{
A named numeric vector for scalar \code{r}, or a numeric matrix for
  vectorized \code{r}.
}
\description{
Computes cumulative discounted profits through each duration. A scalar
discount rate returns a named vector; vectorized rates return a matrix.
}
\examples{
Pi <- c(-15.00, 8.42, 8.39, 8.58)
NPV_partial(Pi, r = 0.10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R
\name{gross_premium_expense_reserves}
\alias{gross_premium_expense_reserves}
\alias{tVGx}
\alias{tVEx}
\title{Whole life gross premium and expense reserves}
\usage{
tVGx(
  x,
  t,
  i,
  G,
  benefit = 1,
  renewal_premium_pct = 0,
  renewal_policy_exp = 0,
  settlement_exp = 0,
  tbl = NULL,
  model = NULL,
  ...
)

tVEx(
  x,
  t,
  i,
  G,
  benefit = 1,
  renewal_premium_pct = 0,
  renewal_policy_exp = 0,
  settlement_exp = 0,
  tbl = NULL,
  model = NULL,
  ...
)
}
\arguments{
\item{x}{Issue age. May be scalar or vector.}

\item{t}{Nonnegative integer duration. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{G}{Gross annual premium. May be scalar or vector.}

\item{benefit}{Insurance benefit amount.}

\item{renewal_premium_pct}{Renewal percent-of-premium expense in
\eqn{[0,1]}.}

\item{renewal_policy_exp}{Renewal per-policy expense.}

\item{settlement_exp}{Settlement expense paid at death.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the actuarial functions.}
}
\value{
A numeric vector of reserve values.
}
\description{
Computes prospective gross premium and expense reserves for fully
discrete whole life insurance after issue.
}
\details{
\code{tVGx()} computes the prospective gross premium reserve.

\code{tVEx()} computes the corresponding expense reserve, defined as the
difference between the gross premium reserve and the net benefit reserve.


The gross premium reserve is calculated as

\deqn{
{}_tV_x^G
=
(b+s)A_{x+t}
-
\left[(1-r)G-e\right]\ddot{a}_{x+t},
}

where
\itemize{
  \item \eqn{b} is the insurance benefit,
  \item \eqn{s} is the settlement expense,
  \item \eqn{G} is the gross annual premium,
  \item \eqn{r} is the renewal percent-of-premium expense, and
  \item \eqn{e} is the renewal per-policy expense.
}

The expense reserve is obtained as the gross premium reserve minus the
corresponding net benefit reserve.
}
\examples{
tVGx(
  x = 40,
  t = 10,
  i = 0.05,
  G = 0.03,
  benefit = 1,
  renewal_premium_pct = 0.10,
  renewal_policy_exp = 0.002,
  settlement_exp = 0.02,
  model = "uniform",
  omega = 100
)

tVEx(
  x = 40,
  t = 10,
  i = 0.05,
  G = 0.03,
  benefit = 1,
  renewal_premium_pct = 0.10,
  renewal_policy_exp = 0.002,
  settlement_exp = 0.02,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_core.R
\name{dx}
\alias{dx}
\title{Compute deaths between ages x and x+1}
\usage{
dx(tbl, x)
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Ages.}
}
\value{
Numeric vector of \eqn{d_x} values.
}
\description{
Compute deaths between ages x and x+1
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{NC_EAN_db}
\alias{NC_EAN_db}
\title{Entry Age Normal normal cost}
\usage{
NC_EAN_db(APV_total, adue_active)
}
\arguments{
\item{APV_total}{Nonnegative total actuarial present value of benefits.}

\item{adue_active}{Positive active-service annuity-due factor.}
}
\value{
A numeric vector.
}
\description{
Computes Entry Age Normal normal cost as total benefit APV divided by an
active-service annuity-due factor.
}
\examples{
NC_EAN_db(APV_total = 25000, adue_active = 15)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{var_nAx_m}
\alias{var_nAx_m}
\title{Variance of m-thly deferred insurance PV}
\usage{
var_nAx_m(x, n, i, m, model, ..., tol = 1e-12, j_max = 100000L)
}
\arguments{
\item{x}{Age.}

\item{n}{Deferral period.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}

\item{tol}{Numerical tolerance for truncating the infinite sum.}

\item{j_max}{Maximum number of m-thly intervals in the sum.}
}
\value{
Numeric vector of variances.
}
\description{
Variance of m-thly deferred insurance PV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/premium_ch9.R
\name{premium_functions}
\alias{premium_functions}
\alias{Px}
\alias{Pxn1}
\alias{PnEx}
\alias{Pxn}
\alias{tPx}
\alias{tPxn1}
\alias{tPnEx}
\alias{tPxn}
\alias{PnAx}
\alias{tPnAx}
\alias{Pbarx}
\alias{Pbarxn1}
\alias{Pbarxn}
\alias{PbarAbarx}
\alias{PbarAbarxn1}
\alias{PbarAbarxn}
\alias{Px_m}
\alias{Pxn1_m}
\alias{Pxn_m}
\alias{PnAx_m}
\alias{EL0x}
\alias{varL0x}
\alias{EL0xn1}
\alias{varL0xn1}
\alias{EL0xn}
\alias{varL0xn}
\alias{EL0barAbarx}
\alias{varL0barAbarx}
\alias{Gx}
\title{Premium, loss, and expense functions}
\usage{
Px(x, i, tbl = NULL, model = NULL, ...)

Pxn1(x, n, i, tbl = NULL, model = NULL, ...)

PnEx(x, n, i, tbl = NULL, model = NULL, ...)

Pxn(x, n, i, tbl = NULL, model = NULL, ...)

tPx(x, t, i, tbl = NULL, model = NULL, ...)

tPxn1(x, n, t, i, tbl = NULL, model = NULL, ...)

tPnEx(x, n, t, i, tbl = NULL, model = NULL, ...)

tPxn(x, n, t, i, tbl = NULL, model = NULL, ...)

PnAx(x, n, i, tbl = NULL, model = NULL, ...)

tPnAx(x, n, t, i, tbl = NULL, model = NULL, ...)

Pbarx(x, i, model, ..., tol = 1e-10)

Pbarxn1(x, n, i, model, ...)

Pbarxn(x, n, i, model, ...)

PbarAbarx(x, i, model, ..., tol = 1e-10)

PbarAbarxn1(x, n, i, model, ...)

PbarAbarxn(x, n, i, model, ...)

Px_m(x, m, i, tbl = NULL, model = NULL, ...)

Pxn1_m(x, n, m, i, tbl = NULL, model = NULL, ...)

Pxn_m(x, n, m, i, tbl = NULL, model = NULL, ...)

PnAx_m(x, n, m, i, tbl = NULL, model = NULL, ...)

EL0x(x, P, i, tbl = NULL, model = NULL, ...)

varL0x(x, P, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000)

EL0xn1(x, n, P, i, tbl = NULL, model = NULL, ...)

varL0xn1(x, n, P, i, tbl = NULL, model = NULL, ...)

EL0xn(x, n, P, i, tbl = NULL, model = NULL, ...)

varL0xn(x, n, P, i, tbl = NULL, model = NULL, ...)

EL0barAbarx(x, P, i, model, ..., tol = 1e-10)

varL0barAbarx(x, P, i, model, ...)

Gx(
  x,
  i,
  benefit = 1,
  first_premium_pct = 0,
  renewal_premium_pct = 0,
  first_policy_exp = 0,
  renewal_policy_exp = 0,
  settlement_exp = 0,
  tbl = NULL,
  model = NULL,
  ...
)
}
\arguments{
\item{x}{Age. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}

\item{n}{Term. May be scalar or vector of nonnegative integers.}

\item{t}{Premium-paying period. May be scalar or vector of nonnegative integers.}

\item{tol}{Numerical tolerance for functions that truncate infinite sums.}

\item{m}{Number of payments per year. Must be a positive integer scalar.}

\item{P}{Premium amount or premium rate. May be scalar or vector.}

\item{k_max}{Maximum summation horizon for functions that truncate infinite sums.}

\item{benefit}{Benefit amount. May be scalar or vector.}

\item{first_premium_pct}{First-year premium expense proportion. May be scalar or vector.}

\item{renewal_premium_pct}{Renewal premium expense proportion. May be scalar or vector.}

\item{first_policy_exp}{First-year fixed expense. May be scalar or vector.}

\item{renewal_policy_exp}{Renewal fixed expense after the first year. May be scalar or vector.}

\item{settlement_exp}{Settlement expense incurred at benefit payment. May be scalar or vector.}
}
\value{
Numeric vector.
}
\description{
Functions for net and gross premiums, present-value-of-loss moments,
continuous-payment premium rates, and premiums payable more frequently
than annually.
}
\details{
The functions include:
\itemize{
  \item net annual premiums under the equivalence principle,
  \item limited-payment premiums,
  \item continuous-payment premium rates,
  \item fully continuous premium rates,
  \item true fractional premiums,
  \item present-value-of-loss means and variances,
  \item a basic gross premium formula for whole life insurance.
}

The discrete premium functions may be evaluated from either a life table
supplied through \code{tbl} or a parametric survival model supplied through
\code{model}. Continuous-payment premium functions are evaluated through
the corresponding continuous insurance and annuity functions.

Scalar inputs retain their existing behavior. Where mathematically
meaningful, numeric inputs may also be vectors. Inputs are evaluated
elementwise when they have a common length, and scalar inputs are recycled.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_core.R
\name{lx}
\alias{lx}
\title{Extract life-table survivor values}
\usage{
lx(tbl, x)
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Ages.}
}
\value{
Numeric vector of \eqn{l_x} values.
}
\description{
Extract life-table survivor values
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{replacement_ratio_db}
\alias{replacement_ratio_db}
\title{Replacement ratio for a defined benefit plan}
\usage{
replacement_ratio_db(benefit, salary)
}
\arguments{
\item{benefit}{Nonnegative annual retirement benefit.}

\item{salary}{Positive salary measure used in the denominator.}
}
\value{
A numeric vector.
}
\description{
Computes annual benefit divided by a selected salary measure. Scalar
arguments are recycled to a common length.
}
\examples{
replacement_ratio_db(
  benefit = 108008.66,
  salary = 187119.09
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{Axn}
\alias{Axn}
\title{Endowment insurance APV}
\usage{
Axn(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes \eqn{A_{x:\overline{n}|} = A_{x:\overline{n}|}^{1} + {}_nE_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{AS_path}
\alias{AS_path}
\title{Projected asset-share path for two decrement causes}
\usage{
AS_path(AS0, G, r, e, b1, b2, q1, q2, p_tau, i, b3 = NULL)
}
\arguments{
\item{AS0}{Initial asset share.}

\item{G}{Premium amount by policy year.}

\item{r}{Percent-of-premium expense rate by policy year.}

\item{e}{Fixed expense by policy year.}

\item{b1}{Cause 1 benefit by policy year.}

\item{b2}{Cause 2 benefit by policy year.}

\item{q1}{Cause 1 probability by policy year.}

\item{q2}{Cause 2 probability by policy year.}

\item{p_tau}{In-force probability by policy year.}

\item{i}{Effective annual interest rate by policy year.}

\item{b3}{Survival benefit by policy year.}
}
\value{
A data frame with policy year \code{k} and asset share \code{AS}.
}
\description{
Convenience interface to \code{AS_path_md()} for two causes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{thiele_backward_path}
\alias{thiele_backward_path}
\title{Backward reserve path from a terminal value}
\usage{
thiele_backward_path(times, V_terminal, P, delta, mu, benefit = 1)
}
\arguments{
\item{times}{Finite numeric vector of strictly increasing times.}

\item{V_terminal}{Finite scalar reserve at the final time.}

\item{P}{Premium rate, scalar or vector with one value per time step.}

\item{delta}{Force of interest, scalar or vector with one value per step.}

\item{mu}{Nonnegative force of mortality, scalar or vector with one value
per step.}

\item{benefit}{Benefit amount, scalar or vector with one value per step.}
}
\value{
Numeric vector of reserve values corresponding to `times`.
}
\description{
Starting from a terminal reserve at the final time, computes reserves
backward over a strictly increasing time grid using
`thiele_backward_step()`.
}
\examples{
times <- seq(19, 20, by = 0.25)

thiele_backward_path(
  times,
  V_terminal = 1000,
  P = 26.96,
  delta = 0.058,
  mu = 0.002,
  benefit = 1000
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{ag38_reserve_ul}
\alias{ag38_reserve_ul}
\title{AG 38 reserve calculation}
\usage{
ag38_reserve_ul(
  basic_reserve,
  deficiency_reserve = 0,
  excess_payment,
  nsp_required,
  valuation_nsp,
  surrender_charge = 0
)
}
\arguments{
\item{basic_reserve}{Nonnegative basic reserve.}

\item{deficiency_reserve}{Nonnegative deficiency reserve.}

\item{excess_payment}{Nonnegative excess payment or shadow-fund amount.}

\item{nsp_required}{Positive net single premium required to fully fund the
guarantee.}

\item{valuation_nsp}{Nonnegative valuation net single premium.}

\item{surrender_charge}{Nonnegative surrender charge.}
}
\value{
For scalar inputs, a named list. For vectorized inputs, a data frame
  containing the same calculated quantities.
}
\description{
Computes the prefunding ratio, net additional amount, reduced deficiency
reserve, intermediate reserve, and increased basic reserve.
}
\details{
Scalar inputs preserve the original named-list output. Vectorized inputs
return a data frame with one row per calculation.
}
\examples{
ag38_reserve_ul(
  basic_reserve = 10000,
  deficiency_reserve = 0,
  excess_payment = 60000,
  nsp_required = 100000,
  valuation_nsp = 150000,
  surrender_charge = 5000
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{gain_loss_md}
\alias{gain_loss_md}
\title{Gain or loss in a two-cause multiple-decrement model}
\usage{
gain_loss_md(
  Vt,
  G,
  r,
  e,
  i,
  b1,
  b2,
  s1 = 0,
  s2 = 0,
  q1,
  q2,
  Vt1,
  year_end_cause2 = FALSE,
  q1prime = NULL,
  q2prime = NULL
)
}
\arguments{
\item{Vt}{Gross reserve at the beginning of the year.}

\item{G}{Gross premium.}

\item{r}{Percent-of-premium expense rate.}

\item{e}{Fixed beginning-of-year expense.}

\item{i}{Earned effective annual interest rate.}

\item{b1}{Cause 1 benefit.}

\item{b2}{Cause 2 benefit.}

\item{s1}{Cause 1 settlement expense.}

\item{s2}{Cause 2 settlement expense.}

\item{q1}{Cause 1 decrement probability.}

\item{q2}{Cause 2 decrement probability.}

\item{Vt1}{Gross reserve at the end of the year.}

\item{year_end_cause2}{Whether Cause 2 occurs only at year-end.}

\item{q1prime}{Single-decrement Cause 1 probability for the ordered case.}

\item{q2prime}{Single-decrement Cause 2 probability for the ordered case.}
}
\value{
A numeric vector of gains or losses.
}
\description{
Computes one-year gain or loss under simultaneous within-year decrement
probabilities or an ordered case in which Cause 2 occurs at year-end.
Numeric arguments may be scalar or compatible vectors.
}
\examples{
gain_loss_md(
  Vt = 115, G = 16, r = 0, e = 3, i = 0.06,
  b1 = 1000, b2 = 110, q1 = 0.01, q2 = 0.10, Vt1 = 128.83
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{survival_models}
\alias{survival_models}
\title{Survival models: core survival functions and parametric models}
\description{
This file provides actuarial survival-model utilities.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_select.R
\name{select_life_table}
\alias{select_life_table}
\title{Construct a select life table}
\usage{
select_life_table(x_sel, duration, attained_age, lx)
}
\arguments{
\item{x_sel}{Numeric vector of ages at selection.}

\item{duration}{Numeric vector of durations since selection.}

\item{attained_age}{Numeric vector of attained ages.}

\item{lx}{Numeric vector of select-table survivor values.}
}
\value{
A data frame with class \code{"select_life_table"}.
}
\description{
Builds a select-life-table object from vectors of selection age,
duration since selection, attained age, and survivor values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{udd_mthly_multiplier}
\alias{udd_mthly_multiplier}
\title{UDD multiplier for m-thly insurance approximations}
\usage{
udd_mthly_multiplier(i, m)
}
\arguments{
\item{i}{Numeric vector of effective annual interest rates.}

\item{m}{Positive integer payment frequency.}
}
\value{
Numeric vector equal to \eqn{i / i^{(m)}}.
}
\description{
Under UDD,
\eqn{A_x^{(m)} = (i / i^{(m)}) A_x}
and similarly for term and deferred insurance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{tpxy}
\alias{tpxy}
\title{Joint-life survival probability}
\usage{
tpxy(x, y, t, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{t}{Time.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{{}_tp_{xy} = {}_tp_x\,{}_tp_y} under independence.
}
\examples{
tpxy(40, 50, t = 10, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{discounted_payback_period}
\alias{discounted_payback_period}
\title{Discounted payback period}
\usage{
discounted_payback_period(Pi, r)
}
\arguments{
\item{Pi}{Numeric profit-signature vector.}

\item{r}{Annual effective risk discount rate. May be scalar or vector;
values must be greater than \code{-1}.}
}
\value{
An integer vector with one value for each discount rate. An element
  is \code{NA_integer_} when payback is not reached.
}
\description{
Returns the first duration at which cumulative discounted profit is
nonnegative.
}
\examples{
Pi <- c(-15.00, 8.42, 8.39, 8.58)
discounted_payback_period(Pi, r = 0.10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{tVx_m}
\alias{tVx_m}
\title{Whole life reserve with m-thly premiums}
\usage{
tVx_m(x, t, m, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Duration.}

\item{m}{Number of premium payments per year.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object.}
}
\value{
A numeric vector of values.
}
\description{
Computes the reserve for a whole life insurance funded
by true m-thly premiums.
}
\examples{
tVx_m(40, t = 10, m = 12, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{Axn1_m_udd}
\alias{Axn1_m_udd}
\title{UDD approximation of m-thly term insurance}
\usage{
Axn1_m_udd(Axn1, i, m)
}
\arguments{
\item{Axn1}{Discrete term insurance APV.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}
}
\value{
m-thly term insurance APV under UDD.
}
\description{
Computes \eqn{A_{x:\overline{n}|}^{1(m)} = (i/i^{(m)})A_{x:\overline{n}|}^{1}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{IRR_profit}
\alias{IRR_profit}
\title{Internal rate of return}
\usage{
IRR_profit(Pi, interval = c(0, 1), tol = .Machine$double.eps^0.5)
}
\arguments{
\item{Pi}{Numeric profit-signature vector.}

\item{interval}{Numeric vector of length two giving the root-search
interval. Its lower endpoint must be greater than \code{-1}.}

\item{tol}{Positive scalar tolerance passed to \code{stats::uniroot()}.}
}
\value{
A numeric scalar.
}
\description{
Computes a root of the profit-signature net present value.
}
\examples{
Pi <- c(-15.00, 8.42, 8.39, 8.58)
IRR_profit(Pi)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{multilife_survival_probabilities}
\alias{multilife_survival_probabilities}
\alias{tpxy}
\alias{tqxy}
\alias{tpxybar}
\alias{tqxybar}
\title{Multi-life survival and failure probabilities}
\usage{
tpxy(x, y, t, tbl = NULL, model = NULL, ...)

tqxy(x, y, t, tbl = NULL, model = NULL, ...)

tpxybar(x, y, t, tbl = NULL, model = NULL, ...)

tqxybar(x, y, t, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{t}{Nonnegative duration. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the survival model.}
}
\value{
A numeric vector of probabilities.
}
\description{
Computes joint-life and last-survivor survival and failure probabilities
for two independent lives.
}
\details{
\code{tpxy()} computes the joint-life survival probability
\deqn{
{}_tp_{xy}
=
{}_tp_x\,{}_tp_y.
}

\code{tqxy()} computes the joint-life failure probability
\deqn{
{}_tq_{xy}
=
1-{}_tp_{xy}.
}

\code{tpxybar()} computes the last-survivor survival probability
\deqn{
{}_tp_{\overline{xy}}
=
{}_tp_x + {}_tp_y - {}_tp_x\,{}_tp_y.
}

\code{tqxybar()} computes the last-survivor failure probability
\deqn{
{}_tq_{\overline{xy}}
=
1-{}_tp_{\overline{xy}}.
}


All calculations assume the future lifetimes are independent.

Joint-life probabilities require both lives to satisfy the survival
condition, whereas last-survivor probabilities require at least one life
to survive.
}
\examples{
tpxy(
  40, 50,
  t = 10,
  model = "uniform",
  omega = 100
)

tqxy(
  40, 50,
  t = 10,
  model = "uniform",
  omega = 100
)

tpxybar(
  40, 50,
  t = 10,
  model = "uniform",
  omega = 100
)

tqxybar(
  40, 50,
  t = 10,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{tVnAdotx}
\alias{tVnAdotx}
\title{Reserve for a deferred annuity-due}
\usage{
tVnAdotx(x, n, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age. May be scalar or vector.}

\item{n}{Positive integer deferral period. May be scalar or vector.}

\item{t}{Nonnegative integer duration. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the actuarial functions.}

\item{tbl}{Optional life table object. Supply by name.}
}
\value{
Numeric vector of reserve values.
}
\description{
Computes the reserve during the deferral period, where \eqn{0 \le t < n}.
}
\examples{
tVnAdotx(
  40,
  n = 20,
  t = 10,
  i = 0.05,
  model = "uniform",
  omega = 100
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{Ax}
\alias{Ax}
\title{Whole life insurance APV}
\usage{
Ax(x, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}

\item{tol}{Numerical tolerance for truncating infinite sums.}

\item{k_max}{Maximum number of terms in the sum.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes \eqn{A_x = \sum_{k=0}^\infty v^{k+1} {}_{k\mid}q_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{nAbarx_udd}
\alias{nAbarx_udd}
\title{UDD approximation of continuous deferred insurance}
\usage{
nAbarx_udd(nAx, i)
}
\arguments{
\item{nAx}{Discrete deferred insurance APV.}

\item{i}{Effective annual interest rate.}
}
\value{
Continuous deferred insurance APV under UDD.
}
\description{
Computes \eqn{{}_{n\mid}\bar{A}_x = (i/\delta)\,{}_{n\mid}A_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{qxtau}
\alias{qxtau}
\title{Total one-year decrement probability}
\usage{
qxtau(qxj)
}
\arguments{
\item{qxj}{Numeric vector of cause-specific decrement probabilities.}
}
\value{
A numeric scalar.
}
\description{
Computes the total one-year multiple-decrement probability
\deqn{q_x^{(\tau)}=\sum_j q_x^{(j)}.}
}
\examples{
qxtau(c(0.011, 0.100))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{tqx}
\alias{tqx}
\title{Conditional failure probability}
\usage{
tqx(t, x, model, ...)
}
\arguments{
\item{t}{Numeric vector of durations.}

\item{x}{Numeric vector of ages.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}
}
\value{
Numeric vector of failure probabilities.
}
\description{
Computes \eqn{{}_t q_x = 1 - {}_t p_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{tVxn_ret}
\alias{tVxn_ret}
\title{Retrospective endowment insurance reserve}
\usage{
tVxn_ret(x, n, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age. May be scalar or vector.}

\item{n}{Nonnegative integer contract term. May be scalar or vector.}

\item{t}{Nonnegative integer duration. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the actuarial functions.}

\item{tbl}{Optional life table object. Supply by name.}
}
\value{
Numeric vector of retrospective reserve values.
}
\description{
Computes the retrospective reserve for an endowment insurance at a
duration satisfying \eqn{0 \le t \le n}. At maturity, the reserve is 1.
}
\examples{
tVxn_ret(
  40,
  n = 20,
  t = 10,
  i = 0.05,
  model = "uniform",
  omega = 100
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{Ax_m_udd}
\alias{Ax_m_udd}
\title{UDD approximation of m-thly whole life insurance}
\usage{
Ax_m_udd(Ax, i, m)
}
\arguments{
\item{Ax}{Discrete whole life insurance APV.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}
}
\value{
m-thly whole life insurance APV under UDD.
}
\description{
Computes \eqn{A_x^{(m)} = (i/i^{(m)})A_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interest_theory.R
\name{discount}
\alias{discount}
\title{Discount factor for compound interest}
\usage{
discount(i, t)
}
\arguments{
\item{i}{Effective interest rate. May be scalar or vector.}

\item{t}{Time. May be scalar or vector.}
}
\value{
Numeric vector of discount factors.
}
\description{
Computes the discount factor \eqn{v^t = (1+i)^{-t}}.
}
\examples{
discount(0.05, 0:5)
discount(c(0.03, 0.05), 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{PAB_fas}
\alias{PAB_fas}
\title{Projected annual benefit under a final-average-salary plan}
\usage{
PAB_fas(x, z, CASx, p, fas_years = 3, past_service = 0, g = NULL, s = NULL)
}
\arguments{
\item{x}{Scalar current or entry age.}

\item{z}{Scalar retirement age.}

\item{CASx}{Positive scalar current annual salary.}

\item{p}{Nonnegative scalar accrual percentage, such as \code{2} for
2 percent.}

\item{fas_years}{Positive integer number of years in the final salary
average.}

\item{past_service}{Nonnegative scalar years of service already completed.}

\item{g}{Optional scalar annual salary growth rate greater than \code{-1}.}

\item{s}{Optional positive salary-scale vector of length \code{z - x}.}
}
\value{
A numeric scalar.
}
\description{
Projects the annual benefit using a final-average-salary formula.
}
\examples{
PAB_fas(
  x = 35,
  z = 65,
  CASx = 60000,
  p = 2,
  fas_years = 3,
  g = 0.04
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_select.R
\name{npx_select}
\alias{npx_select}
\title{Select-life survival probability}
\usage{
npx_select(tbl, x_sel, t, n)
}
\arguments{
\item{tbl}{A \code{select_life_table} object.}

\item{x_sel}{Numeric vector of ages at selection.}

\item{t}{Numeric vector of current durations since selection.}

\item{n}{Numeric vector of nonnegative integer future durations.}
}
\value{
Numeric vector of survival probabilities.
}
\description{
Computes
\eqn{{}_n p_{[x]+t} = l_{[x]+t+n} / l_{[x]+t}}
in the discrete select-table setting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11_ext.R
\name{reserve_ch11_ext}
\alias{reserve_ch11_ext}
\title{Extended reserve functions for Chapter 11}
\description{
Additional Chapter 11 reserve functions:
FPT modified reserves, fractional-duration reserves,
gross premium reserves, expense reserves, and
ordered gross gain decomposition.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{Income_dc}
\alias{Income_dc}
\title{Retirement income from a defined contribution accumulation}
\usage{
Income_dc(AVz, adue_z)
}
\arguments{
\item{AVz}{Nonnegative accumulated value at retirement.}

\item{adue_z}{Positive whole-life annuity-due factor at retirement.}
}
\value{
A numeric vector.
}
\description{
Converts an accumulated account value into annual annuity-due income.
Scalar arguments are recycled to a common length.
}
\examples{
Income_dc(AVz = 824211.35, adue_z = 12)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{continuous_multilife_insurance}
\alias{continuous_multilife_insurance}
\alias{Abarxy}
\alias{Abarxybar}
\alias{Abarxy1}
\alias{Abaryx1}
\alias{Abarxy2}
\alias{Abaryx2}
\title{Continuous multi-life insurance}
\usage{
Abarxy(x, y, i, tbl = NULL, model = NULL, ...)

Abarxybar(x, y, i, tbl = NULL, model = NULL, ...)

Abarxy1(x, y, i, tbl = NULL, model = NULL, ...)

Abaryx1(x, y, i, tbl = NULL, model = NULL, ...)

Abarxy2(x, y, i, tbl = NULL, model = NULL, ...)

Abaryx2(x, y, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object retained for backward compatibility.
Continuous calculations currently require \code{model}.}

\item{model}{Parametric survival model.}

\item{...}{Additional parameters passed to the survival and hazard
functions.}
}
\value{
A numeric vector of actuarial present values.
}
\description{
Computes continuous joint-life, last-survivor, and contingent whole-life
insurance values for two independent lives.
}
\details{
\code{Abarxy()} computes joint-life insurance payable at the first death.

\code{Abarxybar()} computes last-survivor insurance payable at the second
death.

\code{Abarxy1()} and \code{Abaryx1()} compute contingent insurance payable
when the specified life dies first.

\code{Abarxy2()} and \code{Abaryx2()} compute contingent insurance payable
when the specified life dies second.

These functions require a parametric survival model.


Under independence, \code{Abarxy()} represents insurance payable at the
first death and \code{Abarxybar()} represents insurance payable at the
second death.

The contingent functions distinguish both the life whose death triggers
payment and whether that death occurs first or second.
}
\examples{
Abarxy(
  x = 40,
  y = 50,
  i = 0.05,
  model = "uniform",
  omega = 100
)

Abarxy1(
  x = 40,
  y = 50,
  i = 0.05,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R
\name{decompGg_disc}
\alias{decompGg_disc}
\title{Ordered decomposition of gross gain}
\usage{
decompGg_disc(
  VtG,
  Vt1G,
  G,
  i_assumed,
  q_assumed,
  r_assumed = 0,
  e_assumed = 0,
  s_assumed = 0,
  i_actual,
  q_actual,
  r_actual = 0,
  e_actual = 0,
  s_actual = 0,
  b = 1,
  order = c("interest", "mortality", "expense")
)
}
\arguments{
\item{VtG}{Gross reserve at duration \code{t}.}

\item{Vt1G}{Gross reserve at duration \code{t + 1}.}

\item{G}{Gross premium.}

\item{i_assumed}{Assumed annual effective interest rate.}

\item{q_assumed}{Assumed mortality probability.}

\item{r_assumed}{Assumed percent-of-premium expense rate.}

\item{e_assumed}{Assumed per-policy expense.}

\item{s_assumed}{Assumed settlement expense.}

\item{i_actual}{Actual annual effective interest rate.}

\item{q_actual}{Actual mortality probability.}

\item{r_actual}{Actual percent-of-premium expense rate.}

\item{e_actual}{Actual per-policy expense.}

\item{s_actual}{Actual settlement expense.}

\item{b}{Benefit amount.}

\item{order}{Character vector containing \code{"interest"},
\code{"mortality"}, and \code{"expense"} exactly once.}
}
\value{
For scalar input, a named numeric vector. For vectorized input,
  a numeric matrix with columns \code{total_gain}, \code{interest},
  \code{mortality}, \code{expense}, and \code{check}.
}
\description{
Decomposes gross gain into interest, mortality, and expense components
in a user-specified order.
}
\details{
For scalar input, the function returns a named numeric vector. For
vectorized input, it returns a numeric matrix with one row per calculation.
}
\examples{
decompGg_disc(
  VtG = 3950.73,
  Vt1G = 4607.07,
  G = 685,
  i_assumed = 0.06,
  q_assumed = 0.00592,
  r_assumed = 0.05,
  e_assumed = 0,
  s_assumed = 300,
  i_actual = 0.065,
  q_actual = 0.005,
  r_actual = 0.06,
  e_actual = 0,
  s_actual = 100,
  b = 50000
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_conversions.R
\name{S0_to_lx}
\alias{S0_to_lx}
\title{Convert survival probabilities to life-table values}
\usage{
S0_to_lx(S0, radix = 1e+05)
}
\arguments{
\item{S0}{Numeric vector of survival probabilities.}

\item{radix}{Positive radix \eqn{l_0}.}
}
\value{
Numeric vector of \eqn{l_x} values.
}
\description{
Converts survival function values \eqn{S_0(x)} into life-table survivor
values \eqn{l_x = l_0 S_0(x)} using a chosen radix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{tVx}
\alias{tVx}
\title{Whole life net level premium reserve}
\usage{
tVx(x, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object.}
}
\value{
A numeric vector of values.
}
\description{
Computes the prospective reserve
for a whole life insurance with annual premiums:
reserve at duration t equals future APV of benefits minus future APV
of net premiums.
}
\examples{
tVx(40, t = 10, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{qx_dep_sudd}
\alias{qx_dep_sudd}
\title{Multiple-decrement probabilities under SUDD}
\usage{
qx_dep_sudd(q1prime, q2prime)
}
\arguments{
\item{q1prime}{Associated single-decrement probability for cause 1.
May be scalar or vector.}

\item{q2prime}{Associated single-decrement probability for cause 2.
May be scalar or vector.}
}
\value{
For scalar input, a named numeric vector containing \code{q1}
  and \code{q2}. For vectorized input, a numeric matrix with columns
  \code{q1} and \code{q2}.
}
\description{
Converts two associated single-decrement probabilities to dependent
multiple-decrement probabilities under the single-decrement uniform
distribution assumption.
}
\examples{
qx_dep_sudd(0.20, 0.10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{AVz_dc}
\alias{AVz_dc}
\title{Accumulated value of defined contribution plan contributions}
\usage{
AVz_dc(x, z, Sx, c, i, g = NULL, s = NULL)
}
\arguments{
\item{x}{Scalar entry age.}

\item{z}{Scalar retirement age.}

\item{Sx}{Positive scalar salary at age \code{x}.}

\item{c}{Scalar contribution rate in \code{[0, 1]}.}

\item{i}{Scalar annual effective investment return greater than \code{-1}.}

\item{g}{Optional scalar annual salary growth rate greater than \code{-1}.}

\item{s}{Optional positive salary-scale vector of length \code{z - x}.}
}
\value{
A numeric scalar.
}
\description{
Computes the accumulated value at retirement from contributions paid at
the beginning of each year and accumulated to retirement.
}
\examples{
AVz_dc(
  x = 30,
  z = 65,
  Sx = 50000,
  c = 0.10,
  i = 0.05,
  g = 0.04
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{varLtx}
\alias{varLtx}
\title{Variance of present value of loss at duration t for whole life insurance}
\usage{
varLtx(x, t, i, P, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{P}{Annual premium.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object.}
}
\value{
A numeric vector of values.
}
\description{
Computes the conditional variance
\eqn{\mathrm{Var}({}_tL_x \mid K_x \ge t)}
for a fully discrete whole life insurance.
}
\examples{
prem <- Px(40, i = 0.05, model = "uniform", omega = 100)
varLtx(40, t = 10, i = 0.05, P = prem, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{cov_term_deferred}
\alias{cov_term_deferred}
\title{Covariance of term and deferred insurance PVs}
\usage{
cov_term_deferred(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}
}
\value{
Numeric vector of covariances.
}
\description{
Computes
\eqn{\mathrm{Cov}(Z_{x:\overline{n}|}^{1}, {}_{n\mid}Z_x) =
-A_{x:\overline{n}|}^{1} \cdot {}_{n\mid}A_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_annual.R
\name{annuity_annual}
\alias{annuity_annual}
\alias{ax}
\alias{adotx}
\alias{abarx}
\alias{axn}
\alias{adotxn}
\alias{abarxn}
\alias{nax}
\alias{nadotx}
\alias{nabarx}
\alias{sxn}
\alias{sdotxn}
\alias{sbarxn}
\title{Annual annuity functions}
\usage{
ax(x, i, model = NULL, ..., tbl = NULL, k_max = 5000, tol = 1e-12)

adotx(x, i, model = NULL, ..., tbl = NULL, k_max = 5000, tol = 1e-12)

abarx(x, i, model, ..., tol = 1e-10)

axn(x, n, i, model = NULL, ..., tbl = NULL)

adotxn(x, n, i, model = NULL, ..., tbl = NULL)

abarxn(x, n, i, model, ...)

nax(x, n, i, model = NULL, ..., tbl = NULL, k_max = 5000, tol = 1e-12)

nadotx(x, n, i, model = NULL, ..., tbl = NULL, k_max = 5000, tol = 1e-12)

nabarx(x, n, i, model, ..., tol = 1e-10)

sxn(x, n, i, model = NULL, ..., tbl = NULL)

sdotxn(x, n, i, model = NULL, ..., tbl = NULL)

sbarxn(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object for annual discrete annuity functions.}

\item{k_max}{Maximum summation horizon for non-terminating models.}

\item{tol}{Truncation tolerance for non-terminating models.}

\item{n}{Term in years.}
}
\value{
Numeric vector containing the requested annuity present value or actuarial
accumulated value.
}
\description{
Annual whole life, temporary, deferred, and actuarial accumulated value
annuity functions in immediate, due, and continuous forms.

Computes \eqn{a_x = \sum_{t=1}^{\infty} v^t {}_t p_x}.

Computes \eqn{\ddot{a}_x = \sum_{t=0}^{\infty} v^t {}_t p_x}.

Computes \eqn{\bar{a}_x = \int_0^{\infty} v^t {}_t p_x dt}.

Computes \eqn{a_{x:\overline{n}|} = \sum_{t=1}^{n} v^t {}_t p_x}.

Computes \eqn{\ddot{a}_{x:\overline{n}|} =
\sum_{t=0}^{n-1} v^t {}_t p_x}.

Computes \eqn{\bar{a}_{x:\overline{n}|} =
\int_0^n v^t {}_t p_x dt}.

Computes \eqn{{}_{n|}a_x = {}_nE_x a_{x+n}}.

Computes \eqn{{}_{n|}\ddot{a}_x = {}_nE_x \ddot{a}_{x+n}}.

Computes \eqn{{}_{n|}\bar{a}_x = {}_nE_x \bar{a}_{x+n}}.

Computes \eqn{s_{x:\overline{n}|} = a_{x:\overline{n}|} / {}_nE_x}.

Computes \eqn{\ddot{s}_{x:\overline{n}|} =
\ddot{a}_{x:\overline{n}|} / {}_nE_x}.

Computes \eqn{\bar{s}_{x:\overline{n}|} =
\bar{a}_{x:\overline{n}|} / {}_nE_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{pxtau_ul}
\alias{pxtau_ul}
\title{One-year persistency rates for universal life}
\usage{
pxtau_ul(qd, qw, year_end_withdrawal = TRUE)
}
\arguments{
\item{qd}{Mortality probabilities.}

\item{qw}{Withdrawal probabilities.}

\item{year_end_withdrawal}{Logical; if \code{TRUE}, use
\eqn{(1-q^{(d)})(1-q^{(w)})}. Otherwise use \eqn{1-q^{(d)}-q^{(w)}}.}
}
\value{
Numeric vector.
}
\description{
Computes the one-year persistency rates \eqn{p_x^{(\tau)}} under either
Equation (16.14) or Equation (16.15).
}
\examples{
qd <- c(.001, .002, .003)
qw <- c(.02, .02, .03)
pxtau_ul(qd, qw)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_approximations.R
\name{annuity_approximations_udd}
\alias{annuity_approximations_udd}
\alias{ax_m_udd}
\alias{axn_m_udd}
\alias{nax_m_udd}
\alias{adotx_m_udd}
\alias{adotxn_m_udd}
\alias{nadotx_m_udd}
\alias{sxn_m_udd}
\alias{sdotxn_m_udd}
\alias{abarx_udd}
\alias{abarxn_udd}
\alias{nabarx_udd}
\title{UDD annuity approximations}
\usage{
adotx_m_udd(x, m, i, model, ...)

adotxn_m_udd(x, n, m, i, model, ...)

nadotx_m_udd(x, n, m, i, model, ...)

ax_m_udd(x, m, i, model, ...)

axn_m_udd(x, n, m, i, model, ...)

nax_m_udd(x, n, m, i, model, ...)

sdotxn_m_udd(x, n, m, i, model, ...)

sxn_m_udd(x, n, m, i, model, ...)

abarx_udd(x, i, model, ...)

abarxn_udd(x, n, i, model, ...)

nabarx_udd(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{m}{Number of payments per year.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}

\item{n}{Term.}
}
\value{
Numeric vector.
}
\description{
UDD-based approximations for annual, fractional-payment, and continuous annuity functions.

Computes
\deqn{\ddot{a}_x^{(m)} \approx \alpha(m)\ddot{a}_x - \beta(m).}

Computes
\deqn{\ddot{a}_{x:\overline{n}|}^{(m)} \approx
\alpha(m)\ddot{a}_{x:\overline{n}|} - \beta(m)(1-{}_nE_x).}

Computes
\deqn{{}_{n\mid}\ddot{a}_x^{(m)} \approx
\alpha(m)\,{}_{n\mid}\ddot{a}_x - \beta(m)\,{}_nE_x.}

Computes
\deqn{a_x^{(m)} \approx \alpha(m)a_x + \gamma(m).}

Computes
\deqn{a_{x:\overline{n}|}^{(m)} \approx
\alpha(m)a_{x:\overline{n}|} + \gamma(m)(1-{}_nE_x).}

Computes
\deqn{{}_{n\mid}a_x^{(m)} \approx
\alpha(m)\,{}_{n\mid}a_x + \gamma(m)\,{}_nE_x.}

Computes
\deqn{\ddot{s}_{x:\overline{n}|}^{(m)} \approx
\alpha(m)\ddot{s}_{x:\overline{n}|} - \beta(m)\left(\frac{1}{{}_nE_x}-1\right).}

Computes
\deqn{s_{x:\overline{n}|}^{(m)} \approx
\alpha(m)s_{x:\overline{n}|} + \gamma(m)\left(\frac{1}{{}_nE_x}-1\right).}

Computes
\deqn{\bar{a}_x \approx \frac{id}{\delta^2}\ddot{a}_x - \frac{i-\delta}{\delta^2}.}

Uses the identity
\deqn{\bar{a}_{x:\overline{n}|} \approx \frac{1-\bar{A}_{x:\overline{n}|}}{\delta}}
together with the package's existing Abarxn_udd() approximation.

Computes
\deqn{{}_{n\mid}\bar{a}_x \approx {}_nE_x \, \bar{a}_{x+n}.}
}
\details{
These functions implement the standard Uniform Distribution of Deaths
approximations linking annual, m-thly, and continuous annuity values.

The exported functions documented on this page are:
\itemize{
  \item \code{ax_m_udd()}
  \item \code{axn_m_udd()}
  \item \code{nax_m_udd()}
  \item \code{adotx_m_udd()}
  \item \code{adotxn_m_udd()}
  \item \code{nadotx_m_udd()}
  \item \code{sxn_m_udd()}
  \item \code{sdotxn_m_udd()}
  \item \code{abarx_udd()}
  \item \code{abarxn_udd()}
  \item \code{nabarx_udd()}
}

The abarxn_udd() calculation uses the existing Abarxn_udd() insurance
approximation. Consequently, additional survival-model arguments are not used
in that calculation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{tqxprimej_cf}
\alias{tqxprimej_cf}
\title{Single-decrement failure under a constant force}
\usage{
tqxprimej_cf(mu, t)
}
\arguments{
\item{mu}{Nonnegative constant force of decrement. May be scalar or
vector.}

\item{t}{Nonnegative time. May be scalar or vector.}
}
\value{
A numeric vector.
}
\description{
Computes
\deqn{{}_tq_x^{\prime(j)}
=1-\exp(-\mu_jt).}
}
\examples{
tqxprimej_cf(0.10, 5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{i_credit_eiul}
\alias{i_credit_eiul}
\title{Credited rates from index growth rates}
\usage{
i_credit_eiul(
  i_raw,
  part = 1,
  floor = 0,
  cap = Inf,
  margin = 0,
  margin_after_participation = TRUE
)
}
\arguments{
\item{i_raw}{Numeric vector of raw index growth rates.}

\item{part}{Nonnegative scalar participation rate.}

\item{floor}{Scalar minimum credited rate.}

\item{cap}{Scalar maximum credited rate. The default is \code{Inf}.}

\item{margin}{Nonnegative scalar index margin.}

\item{margin_after_participation}{Logical scalar. If \code{TRUE}, the margin
is subtracted after applying participation; otherwise it is subtracted
before applying participation.}
}
\value{
A numeric vector of credited rates.
}
\description{
Applies a participation rate, floor, cap, and optional margin to raw index
growth rates.
}
\examples{
raw <- iP_eiul(c(1000, 1050, 1200, 1100))
i_credit_eiul(raw, part = 1.10, floor = 0.01, cap = 0.10)
i_credit_eiul(raw)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{last_survivor_insurance}
\alias{last_survivor_insurance}
\alias{Axybarn1}
\alias{Axybarn}
\alias{Axybar}
\title{Last-survivor insurance functions}
\usage{
Axybarn1(x, y, n, i, tbl = NULL, model = NULL, ...)

Axybarn(x, y, n, i, tbl = NULL, model = NULL, ...)

Axybar(x, y, i, tbl = NULL, model = NULL, ..., k_max = 5000L, tol = 1e-12)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{n}{Term in years. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the survival model or
life-table functions.}

\item{k_max}{Maximum number of terms used for an infinite series.}

\item{tol}{Numerical tolerance used to assess convergence.}
}
\value{
A numeric vector of actuarial present values.
}
\description{
Computes temporary and whole-life insurance values for two lives under
the last-survivor status.
}
\details{
\code{Axybarn1()} computes temporary insurance payable at the second
death within the term.

\code{Axybarn()} computes last-survivor endowment insurance.

\code{Axybar()} computes last-survivor whole-life insurance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{multilife_pure_endowments}
\alias{multilife_pure_endowments}
\alias{nExy}
\alias{nExybar}
\title{Multi-life pure endowments}
\usage{
nExy(x, y, n, i, tbl = NULL, model = NULL, ...)

nExybar(x, y, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{n}{Nonnegative integer term. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the survival model.}
}
\value{
A numeric vector of actuarial present values.
}
\description{
Computes pure endowment actuarial present values for joint-life and
last-survivor statuses for two independent lives.
}
\details{
\code{nExy()} computes an \eqn{n}-year joint-life pure endowment, payable
at time \eqn{n} if both lives survive.

\code{nExybar()} computes an \eqn{n}-year last-survivor pure endowment,
payable at time \eqn{n} if at least one life survives.


The joint-life pure endowment is
\deqn{
{}_nE_{xy}
=
v^n\,{}_np_{xy}.
}

The last-survivor pure endowment is
\deqn{
{}_nE_{\overline{xy}}
=
v^n\,{}_np_{\overline{xy}}.
}

Under independence,
\deqn{
{}_np_{xy}
=
{}_np_x\,{}_np_y
}
and
\deqn{
{}_np_{\overline{xy}}
=
{}_np_x + {}_np_y - {}_np_x\,{}_np_y.
}
}
\examples{
nExy(
  x = 40,
  y = 50,
  n = 10,
  i = 0.05,
  model = "uniform",
  omega = 100
)

nExybar(
  x = 40,
  y = 50,
  n = 10,
  i = 0.05,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{insurance_mthly}
\alias{insurance_mthly}
\title{m-thly insurance models}
\description{
Exact m-thly contingent payment and insurance functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_core.R
\name{qx_tab}
\alias{qx_tab}
\title{Compute one-year death probability from a life table}
\usage{
qx_tab(tbl, x)
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Ages.}
}
\value{
Numeric vector of \eqn{q_x} values.
}
\description{
Compute one-year death probability from a life table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{AAL_PUC_db}
\alias{AAL_PUC_db}
\title{Projected Unit Credit accrued liability for a DB plan}
\usage{
AAL_PUC_db(
  projected_benefit,
  past_service,
  total_service,
  v_to_ret,
  p_surv,
  adue_ret
)
}
\arguments{
\item{projected_benefit}{Projected benefit at retirement.}

\item{past_service}{Past service completed.}

\item{total_service}{Total service at retirement.}

\item{v_to_ret}{Discount factor to retirement.}

\item{p_surv}{Active-service survival probability to retirement.}

\item{adue_ret}{Retirement annuity factor.}
}
\value{
PUC accrued liability.
}
\description{
Computes the PUC accrued liability as the APV of the portion of the
projected benefit attributed to past service.
}
\examples{
AAL_PUC_db(projected_benefit = 30000, past_service = 10, total_service = 30,
v_to_ret = 0.5, p_surv = 0.9, adue_ret = 12)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R
\name{full_preliminary_term}
\alias{full_preliminary_term}
\alias{alphaF}
\alias{betaF}
\alias{tVFx}
\title{Full preliminary term modified premiums and reserves}
\usage{
alphaF(x, i, tbl = NULL, model = NULL, ...)

betaF(x, i, tbl = NULL, model = NULL, ...)

tVFx(x, t, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Issue age. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the actuarial functions.}

\item{t}{Nonnegative integer duration. May be scalar or vector.}
}
\value{
A numeric vector of modified premiums or reserves.
}
\description{
Computes modified premiums and reserves under the full preliminary term
method for whole-life insurance.
}
\details{
\code{alphaF()} computes the first-year modified premium
\eqn{\alpha^F = vq_x}.

\code{betaF()} computes the renewal modified premium
\eqn{\beta^F = P_{x+1}}.

\code{tVFx()} computes the full preliminary term reserve. The reserve is
zero at durations 0 and 1. For \eqn{t > 1}, it equals the net level premium
reserve at duration \eqn{t - 1} for a policy issued at age \eqn{x + 1}.


Under the full preliminary term method, the first policy year is treated
as one-year term insurance. The first-year modified premium is therefore
the actuarial present value of one-year term insurance at age \eqn{x}.

Renewal premiums are based on a whole-life policy issued one year later,
at age \eqn{x + 1}. Accordingly, reserves after the first policy year are
obtained from the corresponding net level premium reserve for that
deferred issue age.
}
\examples{
alphaF(
  x = 40,
  i = 0.05,
  model = "uniform",
  omega = 100
)

betaF(
  x = 40,
  i = 0.05,
  model = "uniform",
  omega = 100
)

tVFx(
  x = 40,
  t = 5,
  i = 0.05,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{variable_interest_apvs}
\alias{variable_interest_apvs}
\alias{nEx_var}
\alias{Axn1_var}
\alias{Axn_var}
\alias{axn_var}
\title{Actuarial present values under variable annual interest rates}
\usage{
nEx_var(qx, i, benefit = 1)

Axn1_var(qx, i, benefit = 1)

Axn_var(qx, i, benefit = 1)

axn_var(qx, i, type = c("immediate", "due"), benefit = 1)
}
\arguments{
\item{qx}{Numeric vector of one-year mortality probabilities.}

\item{i}{Numeric vector of annual effective interest rates. Each value
must be greater than \code{-1}.}

\item{benefit}{Nonnegative scalar benefit or annuity payment amount.}

\item{type}{Character string equal to \code{"immediate"} or
\code{"due"}.}
}
\value{
A numeric scalar.
}
\description{
Computes life-contingent actuarial present values using a specified
sequence of annual effective interest rates.
}
\details{
The vectors \code{qx} and \code{i} represent one valuation scenario and
must have the same positive length.

\code{nEx_var()} computes a pure endowment.

\code{Axn1_var()} computes term insurance payable at the end of the year
of death.

\code{Axn_var()} computes endowment insurance.

\code{axn_var()} computes a temporary annuity-immediate or
annuity-due.


Each year's payment is discounted using the cumulative product of the
annual effective interest rates supplied in \code{i}.

The pure endowment is
\deqn{
{}_nE
=
{}_np_x\,v_n,
}
where \eqn{v_n} is the cumulative discount factor implied by the sequence
of annual effective interest rates.

Term insurance is obtained by discounting each possible death benefit
using the cumulative discount factor applicable to its payment year.

Endowment insurance equals the sum of the corresponding term insurance
and pure endowment.

Temporary annuities discount each payment using the cumulative discount
factors derived from the interest-rate sequence.
}
\examples{
qx <- c(0.03, 0.04, 0.05, 0.06, 0.07)
rates <- c(0.06, 0.07, 0.08, 0.09, 0.10)

nEx_var(qx, rates, benefit = 1000)
Axn1_var(qx, rates)
Axn_var(qx, rates)
axn_var(qx, rates, type = "due")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{nAx_m_udd}
\alias{nAx_m_udd}
\title{UDD approximation of m-thly deferred insurance}
\usage{
nAx_m_udd(nAx, i, m)
}
\arguments{
\item{nAx}{Discrete deferred insurance APV.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}
}
\value{
m-thly deferred insurance APV under UDD.
}
\description{
Computes \eqn{{}_{n\mid}A_x^{(m)} = (i/i^{(m)})\,{}_{n\mid}A_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{GM_cont}
\alias{GM_cont}
\title{Mortality gain for a continuous-style recursion}
\usage{
GM_cont(Vt, Vt1, P, delta_assumed, p_actual, benefit = 0, h = 1)
}
\arguments{
\item{Vt}{Reserve at time `t`.}

\item{Vt1}{Reserve at time `t + h`.}

\item{P}{Premium rate.}

\item{delta_assumed}{Assumed force of interest.}

\item{p_actual}{Actual survival probability over the step.}

\item{benefit}{Benefit paid at the start of the step.}

\item{h}{Positive step length.}
}
\value{
Numeric vector of mortality gain values.
}
\description{
Evaluates the one-step gain using the assumed force of interest and the
actual survival probability.
}
\examples{
GM_cont(
  Vt = 10,
  Vt1 = 11,
  P = 1,
  delta_assumed = 0.05,
  p_actual = 0.99
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_core.R
\name{npx}
\alias{npx}
\title{Compute n-year survival probability from a life table}
\usage{
npx(tbl, x, n)
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Ages.}

\item{n}{Nonnegative integer durations.}
}
\value{
Numeric vector of \eqn{{}_n p_x} values.
}
\description{
Compute n-year survival probability from a life table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_core.R
\name{life_table}
\alias{life_table}
\title{Construct a life table}
\usage{
life_table(x, lx = NULL, qx = NULL, px = NULL, S0 = NULL, radix = 1e+05)
}
\arguments{
\item{x}{Numeric vector of ages.}

\item{lx}{Numeric vector of \eqn{l_x} values.}

\item{qx}{Numeric vector of \eqn{q_x} values.}

\item{px}{Numeric vector of \eqn{p_x} values.}

\item{S0}{Numeric vector of \eqn{S_0(x)} values.}

\item{radix}{Radix used when converting \code{S0} to \code{lx}, or when
building from \code{qx} or \code{px}.}
}
\value{
A data frame with class \code{"life_table"}.
}
\description{
Build a discrete life table from one of \code{lx}, \code{qx}, \code{px},
or \code{S0}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{tVbarAbarx}
\alias{tVbarAbarx}
\title{Fully continuous whole life reserve}
\usage{
tVbarAbarx(x, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Duration, allowed to be any nonnegative numeric value.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object.}
}
\value{
A numeric vector of values.
}
\description{
Computes the reserve for a whole life insurance with
continuous premiums and immediate payment of claims.
}
\details{
In the fully continuous setting, reserve time \eqn{t} may be any
nonnegative real value.
}
\examples{
tVbarAbarx(40, t = 10, i = 0.05, model = "uniform", omega = 100)
tVbarAbarx(40, t = c(19, 19.25, 19.5, 19.75, 20), i = 0.06,
           model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{fnk_from_z}
\alias{fnk_from_z}
\title{Forward rate implied by spot rates}
\usage{
fnk_from_z(z, n, k)
}
\arguments{
\item{z}{Numeric vector of annual effective spot rates for maturities
\code{1, ..., length(z)}. Each value must be greater than \code{-1}.}

\item{n}{Nonnegative integer forward-start time.}

\item{k}{Positive integer forward period.}
}
\value{
A numeric scalar.
}
\description{
Computes the annual effective forward rate for the interval from time
\code{n} to time \code{n + k}:
\deqn{
(1+z_{n+k})^{n+k}
=
(1+z_n)^n(1+f_{n,k})^k.
}
}
\examples{
z <- c(0.03, 0.04, 0.05, 0.06, 0.07)
fnk_from_z(z, n = 1, k = 4)
fnk_from_z(z, n = 2, k = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{AB_cae}
\alias{AB_cae}
\title{Accrued benefit under a career-average-earnings plan}
\usage{
AB_cae(salary_history, p)
}
\arguments{
\item{salary_history}{Positive numeric vector of annual salaries.}

\item{p}{Nonnegative scalar accrual percentage.}
}
\value{
A numeric scalar.
}
\description{
Computes the accrued benefit using salary history through the valuation
date.
}
\examples{
AB_cae(
  salary_history = c(100000, 104000, 108160),
  p = 1
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{chapter15_spot_interest_apv}
\alias{chapter15_spot_interest_apv}
\alias{nEx_spot}
\alias{Axn1_spot}
\alias{Axn_spot}
\alias{axn_spot}
\title{Spot-rate actuarial present value functions}
\usage{
nEx_spot(qx, z, benefit = 1)

Axn1_spot(qx, z, benefit = 1)

Axn_spot(qx, z, benefit = 1)

axn_spot(qx, z, type = c("immediate", "due"), benefit = 1)
}
\arguments{
\item{qx}{Numeric vector of one-year mortality rates.}

\item{z}{Numeric vector of annual effective spot rates for maturities
\eqn{1,\dots,n}.}

\item{benefit}{Amount of each annuity payment.}

\item{type}{Either \code{"immediate"} or \code{"due"}.}
}
\value{
A numeric scalar.

A numeric scalar.

A numeric scalar.

A numeric scalar.
}
\description{
Chapter 15 functions for actuarial present values when discounting uses
spot rates by maturity. If \eqn{z_t} denotes the annual effective spot rate
for maturity \eqn{t}, then the discount factor is \eqn{(1+z_t)^{-t}}.

Computes
\deqn{
{}_nE_x = (1+z_n)^{-n}\cdot {}_np_x.
}

Computes
\deqn{
A_{x:\overline{n}|}^1 = \sum_{t=1}^{n}(1+z_t)^{-t}\cdot {}_{t-1}p_x \cdot q_{x+t-1}.
}

Computes
\deqn{
A_{x:\overline{n}|} = A_{x:\overline{n}|}^1 + {}_nE_x
}
using spot-rate discount factors.

For an immediate annuity,
\deqn{
a_{x:\overline{n}|} = \sum_{t=1}^{n}(1+z_t)^{-t}\cdot {}_tp_x.
}
}
\details{
For an annuity-due,
\deqn{
\ddot{a}_{x:\overline{n}|} = \sum_{t=0}^{n-1}(1+z_t)^{-t}\cdot {}_tp_x,
}
where the time-0 discount factor is 1.
}
\examples{
qx <- c(.02, .03, .04, .05, .06)
z  <- c(.03, .04, .05, .06, .07)
nEx_spot(qx, z, benefit = 1000000)

qx <- c(.02, .03, .04, .05, .06)
z  <- c(.03, .04, .05, .06, .07)
Axn1_spot(qx, z)

qx <- c(.02, .03, .04, .05, .06)
z  <- c(.03, .04, .05, .06, .07)
Axn_spot(qx, z)

qx <- c(.02, .03, .04, .05, .06)
z  <- c(.03, .04, .05, .06, .07)
axn_spot(qx, z, type = "due")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{qx_dep_cf}
\alias{qx_dep_cf}
\title{Multiple-decrement probabilities under constant forces}
\usage{
qx_dep_cf(qxprime)
}
\arguments{
\item{qxprime}{Numeric vector of associated single-decrement
probabilities. Each value must be less than one.}
}
\value{
A numeric vector of dependent cause-specific decrement
  probabilities.
}
\description{
Converts associated single-decrement probabilities to dependent
cause-specific decrement probabilities under constant forces.
}
\examples{
qx_dep_cf(c(0.20, 0.10))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{double_force_delta}
\alias{double_force_delta}
\title{Doubled force of interest}
\usage{
double_force_delta(delta)
}
\arguments{
\item{delta}{Numeric vector of forces of interest.}
}
\value{
Numeric vector of doubled forces of interest.
}
\description{
Computes \eqn{\delta' = 2\delta}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mortality_improvement.R
\name{naxn_improved}
\alias{naxn_improved}
\title{Deferred temporary annuity-immediate under mortality improvement}
\usage{
naxn_improved(x0, u, n, i, qx_base_vec, AAx_vec, base_year, issue_year)
}
\arguments{
\item{x0}{Issue age.}

\item{u}{Deferral period in years.}

\item{n}{Temporary payment period in years.}

\item{i}{Effective annual interest rate.}

\item{qx_base_vec}{Base-year one-year death probabilities for ages
\code{x0, x0+1, ..., x0+u+n-1}.}

\item{AAx_vec}{Mortality improvement factors for ages
\code{x0, x0+1, ..., x0+u+n-1}.}

\item{base_year}{Base year.}

\item{issue_year}{Issue year.}
}
\value{
Numeric scalar.
}
\description{
Computes
\deqn{{}_{u\mid}a_{x:\overline{n}|} = \sum_{t=u+1}^{u+n} v^t \, {}_tp_x^{(\mathrm{improved})}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_select.R
\name{nmxq_select}
\alias{nmxq_select}
\title{Deferred select-life death probability}
\usage{
nmxq_select(tbl, x_sel, t, n, m)
}
\arguments{
\item{tbl}{A \code{select_life_table} object.}

\item{x_sel}{Numeric vector of ages at selection.}

\item{t}{Numeric vector of current durations since selection.}

\item{n}{Numeric vector of nonnegative integer deferred periods.}

\item{m}{Numeric vector of nonnegative integer death windows.}
}
\value{
Numeric vector of deferred death probabilities.
}
\description{
Computes
\eqn{{}_{n|m} q_{[x]+t} = {}_n p_{[x]+t} \cdot {}_m q_{[x]+t+n}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{A2barxn}
\alias{A2barxn}
\title{Second moment of continuous endowment insurance PV}
\usage{
A2barxn(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of second moments.
}
\description{
Computes \eqn{{}^{2}\bar{A}_{x:\overline{n}|}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_varying_benefits.R
\name{IAbarx}
\alias{IAbarx}
\title{Piecewise-continuous increasing whole life insurance}
\usage{
IAbarx(x, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes
\deqn{(I\bar{A})_x =
\int_0^\infty \lfloor t+1 \rfloor v^t {}_tp_x \mu_{x+t}\,dt.}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{ex_complete}
\alias{ex_complete}
\title{Complete expectation of life}
\usage{
ex_complete(x, model, ..., tol = 1e-10)
}
\arguments{
\item{x}{Numeric vector of ages.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}

\item{tol}{Tolerance used to choose a finite integration bound.}
}
\value{
Numeric vector of complete expectations.
}
\description{
Computes \eqn{\overset{\circ}{e}_x=\int_0^\infty {}_t p_x\,dt}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{GI_cont}
\alias{GI_cont}
\title{Interest gain for a continuous-style recursion}
\usage{
GI_cont(Vt, Vt1, P, delta_actual, p_assumed, benefit = 0, h = 1)
}
\arguments{
\item{Vt}{Reserve at time `t`.}

\item{Vt1}{Reserve at time `t + h`.}

\item{P}{Premium rate.}

\item{delta_actual}{Actual force of interest.}

\item{p_assumed}{Assumed survival probability over the step.}

\item{benefit}{Benefit paid at the start of the step.}

\item{h}{Positive step length.}
}
\value{
Numeric vector of interest gain values.
}
\description{
Evaluates the one-step gain using the actual force of interest and the
assumed survival probability.
}
\examples{
GI_cont(
  Vt = 10,
  Vt1 = 11,
  P = 1,
  delta_actual = 0.05,
  p_assumed = 0.99
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numerical_utils.R
\name{.find_upper_t}
\alias{.find_upper_t}
\title{Internal: choose a practical upper bound for integrals to infinity}
\usage{
.find_upper_t(x, model, ..., eps = 1e-10, t_max = 500)
}
\description{
Strategy: find t such that S0(t) is very small, or fall back to a cap.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{salary_scale}
\alias{salary_scale}
\title{Salary scale under constant annual growth}
\usage{
salary_scale(k, g, base_age = min(k), s_base = 1)
}
\arguments{
\item{k}{Numeric vector of ages or durations.}

\item{g}{Annual salary growth rate greater than \code{-1}.}

\item{base_age}{Scalar age or duration at which the scale is normalized.}

\item{s_base}{Positive scalar salary-scale value at \code{base_age}.}
}
\value{
A numeric vector with the same length as \code{k}.
}
\description{
Constructs salary-scale values under a constant annual growth rate.
}
\examples{
salary_scale(k = 30:34, g = 0.04, base_age = 30)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{var_Axn1_m}
\alias{var_Axn1_m}
\title{Variance of m-thly term insurance PV}
\usage{
var_Axn1_m(x, n, i, m, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of variances.
}
\description{
Variance of m-thly term insurance PV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{NPV_profit}
\alias{NPV_profit}
\title{Net present value of a profit signature}
\usage{
NPV_profit(Pi, r)
}
\arguments{
\item{Pi}{Numeric profit-signature vector.}

\item{r}{Annual effective risk discount rate. May be scalar or vector;
values must be greater than \code{-1}.}
}
\value{
A numeric vector with one value for each rate in \code{r}.
}
\description{
Discounts a profit signature at one or more annual effective risk discount
rates.
}
\examples{
Pi <- c(-15.00, 8.42, 8.39, 8.58)
NPV_profit(Pi, r = 0.10)
NPV_profit(Pi, r = c(0.08, 0.10, 0.12))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{tpx}
\alias{tpx}
\title{Conditional survival probability}
\usage{
tpx(t, x, model, ...)
}
\arguments{
\item{t}{Numeric vector of durations.}

\item{x}{Numeric vector of ages.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}
}
\value{
Numeric vector of survival probabilities.
}
\description{
Computes \eqn{{}_t p_x = S_0(x+t)/S_0(x)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{tVxn1}
\alias{tVxn1}
\title{Term insurance net level premium reserve}
\usage{
tVxn1(x, n, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age.}

\item{n}{Term in years.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional model parameters.}

\item{tbl}{Optional life table object.}
}
\value{
A numeric vector of values.
}
\description{
Computes the prospective reserve for an n-year term insurance.
}
\examples{
tVxn1(40, n = 20, t = 10, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mortality_improvement.R
\name{ax_improved}
\alias{ax_improved}
\title{Whole life annuity-immediate under mortality improvement}
\usage{
ax_improved(x0, i, qx_base_vec, AAx_vec, base_year, issue_year)
}
\arguments{
\item{x0}{Issue age.}

\item{i}{Effective annual interest rate.}

\item{qx_base_vec}{Base-year one-year death probabilities for successive ages.}

\item{AAx_vec}{Mortality improvement factors for successive ages.}

\item{base_year}{Base year.}

\item{issue_year}{Issue year.}
}
\value{
Numeric scalar.
}
\description{
Computes a truncated whole life annuity-immediate under projected mortality.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{nEx}
\alias{nEx}
\title{Pure endowment APV}
\usage{
nEx(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes \eqn{{}_nE_x = v^n {}_n p_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{AV_path_ul_typeA}
\alias{AV_path_ul_typeA}
\title{Type A universal life account-value path}
\usage{
AV_path_ul_typeA(G, r, e, qx, ic, B, iq = ic, AV0 = 0)
}
\arguments{
\item{G}{Premium amount by period.}

\item{r}{Percent-of-premium expense rate by period. Values must lie in
\code{[0, 1]}.}

\item{e}{Fixed expense by period.}

\item{qx}{Mortality probability by period.}

\item{ic}{Credited annual effective interest rate by period. Values must be
greater than \code{-1}.}

\item{B}{Face amount by period.}

\item{iq}{Interest rate used in the cost-of-insurance calculation. Defaults
to \code{ic}; values must be greater than \code{-1}.}

\item{AV0}{Nonnegative scalar initial account value.}
}
\value{
A data frame containing the policy duration, premium, and account
  value.
}
\description{
Computes a year-by-year Type A universal life account-value path. The death
benefit is fixed, so the net amount at risk depends on the ending account
value and the roll-forward is solved explicitly each period.
}
\examples{
qx <- c(0.00076, 0.00081, 0.00085, 0.00090, 0.00095)
r <- c(0.75, rep(0.10, 4))
e <- c(100, rep(20, 4))

AV_path_ul_typeA(
  G = 5000,
  r = r,
  e = e,
  qx = qx,
  ic = 0.03,
  B = 100000
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{Axn_m_udd}
\alias{Axn_m_udd}
\title{UDD approximation of m-thly endowment insurance}
\usage{
Axn_m_udd(Axn1, nEx, i, m)
}
\arguments{
\item{Axn1}{Discrete term insurance APV.}

\item{nEx}{Pure endowment APV.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}
}
\value{
m-thly endowment insurance APV under UDD.
}
\description{
Computes
\eqn{A_{x:\overline{n}|}^{(m)} = (i/i^{(m)})A_{x:\overline{n}|}^{1} + {}_nE_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{axybar_ch12}
\alias{axybar_ch12}
\alias{axybar}
\title{Last-survivor whole life annuity-immediate}
\usage{
axybar(x, y, i, tbl = NULL, model = NULL, ..., k_max = 5000, tol = 1e-12)
}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}

\item{k_max}{Maximum number of terms.}

\item{tol}{Convergence tolerance.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{a_{\overline{xy}}}.
}
\details{
Shared documentation topic used to avoid filename collisions with
case-distinct function names on case-insensitive file systems.
}
\examples{
axybar(40, 50, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/premium_ch9.R
\name{premium_ch9}
\alias{premium_ch9}
\alias{Px}
\alias{Pxn1}
\alias{PnEx}
\alias{Pxn}
\alias{tPx}
\alias{tPxn1}
\alias{tPnEx}
\alias{tPxn}
\alias{PnAx}
\alias{tPnAx}
\alias{Pbarx}
\alias{Pbarxn1}
\alias{Pbarxn}
\alias{PbarAbarx}
\alias{PbarAbarxn1}
\alias{PbarAbarxn}
\alias{Px_m}
\alias{Pxn1_m}
\alias{Pxn_m}
\alias{PnAx_m}
\alias{EL0x}
\alias{varL0x}
\alias{EL0xn1}
\alias{varL0xn1}
\alias{EL0xn}
\alias{varL0xn}
\alias{EL0barAbarx}
\alias{varL0barAbarx}
\alias{Gx}
\title{Chapter 9 premium, loss, and expense functions}
\usage{
Px(x, i, tbl = NULL, model = NULL, ...)

Pxn1(x, n, i, tbl = NULL, model = NULL, ...)

PnEx(x, n, i, tbl = NULL, model = NULL, ...)

Pxn(x, n, i, tbl = NULL, model = NULL, ...)

tPx(x, t, i, tbl = NULL, model = NULL, ...)

tPxn1(x, n, t, i, tbl = NULL, model = NULL, ...)

tPnEx(x, n, t, i, tbl = NULL, model = NULL, ...)

tPxn(x, n, t, i, tbl = NULL, model = NULL, ...)

PnAx(x, n, i, tbl = NULL, model = NULL, ...)

tPnAx(x, n, t, i, tbl = NULL, model = NULL, ...)

Pbarx(x, i, model, ..., tol = 1e-10)

Pbarxn1(x, n, i, model, ...)

Pbarxn(x, n, i, model, ...)

PbarAbarx(x, i, model, ..., tol = 1e-10)

PbarAbarxn1(x, n, i, model, ...)

PbarAbarxn(x, n, i, model, ...)

Px_m(x, m, i, tbl = NULL, model = NULL, ...)

Pxn1_m(x, n, m, i, tbl = NULL, model = NULL, ...)

Pxn_m(x, n, m, i, tbl = NULL, model = NULL, ...)

PnAx_m(x, n, m, i, tbl = NULL, model = NULL, ...)

EL0x(x, P, i, tbl = NULL, model = NULL, ...)

varL0x(x, P, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000)

EL0xn1(x, n, P, i, tbl = NULL, model = NULL, ...)

varL0xn1(x, n, P, i, tbl = NULL, model = NULL, ...)

EL0xn(x, n, P, i, tbl = NULL, model = NULL, ...)

varL0xn(x, n, P, i, tbl = NULL, model = NULL, ...)

EL0barAbarx(x, P, i, model, ..., tol = 1e-10)

varL0barAbarx(x, P, i, model, ...)

Gx(
  x,
  i,
  benefit = 1,
  first_premium_pct = 0,
  renewal_premium_pct = 0,
  first_policy_exp = 0,
  renewal_policy_exp = 0,
  settlement_exp = 0,
  tbl = NULL,
  model = NULL,
  ...
)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}

\item{n}{Term.}

\item{t}{Premium-paying period.}

\item{tol}{Numerical tolerance for functions that truncate infinite sums.}

\item{m}{Number of payments per year.}

\item{P}{Premium amount or premium rate.}

\item{k_max}{Maximum summation horizon for functions that truncate infinite sums.}

\item{benefit}{Benefit amount.}

\item{first_premium_pct}{First-year premium expense proportion.}

\item{renewal_premium_pct}{Renewal premium expense proportion.}

\item{first_policy_exp}{First-year fixed expense.}

\item{renewal_policy_exp}{Renewal fixed expense each year after the first.}

\item{settlement_exp}{Settlement expense incurred at benefit payment.}
}
\value{
Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.

Numeric vector.
}
\description{
Functions in this file implement Chapter 9 funding-plan formulas,
including:
\itemize{
  \item net annual premiums under the equivalence principle,
  \item limited-payment premiums,
  \item continuous-payment premium rates,
  \item fully continuous premium rates,
  \item true fractional premiums,
  \item present-value-of-loss means and variances,
  \item a basic gross premium formula for whole life insurance.
}
}
\details{
Naming follows Chapter 9 notation as closely as possible:
\itemize{
  \item \code{Px()} = whole life annual premium
  \item \code{Pxn1()} = term insurance annual premium
  \item \code{PnEx()} = pure endowment annual premium
  \item \code{Pxn()} = endowment insurance annual premium
  \item \code{tPx()} = limited-payment whole life annual premium
  \item \code{tPxn1()} = limited-payment term insurance annual premium
  \item \code{tPnEx()} = limited-payment pure endowment annual premium
  \item \code{tPxn()} = limited-payment endowment insurance annual premium
  \item \code{PnAx()} = deferred insurance annual premium
  \item \code{tPnAx()} = limited-payment deferred insurance annual premium
  \item \code{Pbarx()} = continuous-payment premium for discrete whole life insurance
  \item \code{Pbarxn1()} = continuous-payment premium for discrete term insurance
  \item \code{Pbarxn()} = continuous-payment premium for discrete endowment insurance
  \item \code{PbarAbarx()} = fully continuous premium for continuous whole life insurance
  \item \code{PbarAbarxn1()} = fully continuous premium for continuous term insurance
  \item \code{PbarAbarxn()} = fully continuous premium for continuous endowment insurance
  \item \code{Px_m()} = true fractional whole life annual premium
  \item \code{Pxn1_m()} = true fractional term insurance annual premium
  \item \code{Pxn_m()} = true fractional endowment insurance annual premium
  \item \code{PnAx_m()} = true fractional deferred insurance annual premium
}

The discrete premium functions can be evaluated from either a life table
via \code{tbl = ...} or from a parametric model via \code{model = ...}.

The continuous-premium-rate functions use the continuous annuity functions
already in the package, so they are written for the parametric survival
model framework.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_select.R
\name{nqx_select}
\alias{nqx_select}
\title{Select-life death probability}
\usage{
nqx_select(tbl, x_sel, t, n)
}
\arguments{
\item{tbl}{A \code{select_life_table} object.}

\item{x_sel}{Numeric vector of ages at selection.}

\item{t}{Numeric vector of current durations since selection.}

\item{n}{Numeric vector of nonnegative integer future durations.}
}
\value{
Numeric vector of death probabilities.
}
\description{
Computes
\eqn{{}_n q_{[x]+t} = 1 - {}_n p_{[x]+t}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_mthly.R
\name{annuity_mthly_whole_immediate}
\alias{annuity_mthly_whole_immediate}
\alias{ax_m}
\title{Whole life m-thly annuity-immediate}
\usage{
ax_m(x, m, i, model, ..., k_max = 2e+05, tol = 1e-12)
}
\arguments{
\item{x}{Age.}

\item{m}{Number of payments per year.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model name.}

\item{...}{Additional parameters passed to the survival model.}

\item{k_max}{Maximum summation horizon for non-terminating models.}

\item{tol}{Truncation tolerance for non-terminating models.}
}
\value{
Numeric vector of annuity values.
}
\description{
Computes the exact whole life m-thly annuity-immediate,
with annual payment rate 1 split into m equal payments of size 1/m.
}
\details{
\deqn{a_x^{(m)} = \frac{1}{m}\sum_{t=1}^{\infty} v^{t/m}\,{}_{t/m}p_x}
}
\examples{
ax_m(40, m = 12, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_varying_benefits.R
\name{DAxn1}
\alias{DAxn1}
\title{Decreasing n-year term insurance}
\usage{
DAxn1(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes
\deqn{(DA)_{x:\overline{n}|}^{1}
= \sum_{k=0}^{n-1} (n-k) v^{k+1} \Pr(K_x = k).}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_select.R
\name{lx_select}
\alias{lx_select}
\title{Extract select-table survivor value}
\usage{
lx_select(tbl, x_sel, t)
}
\arguments{
\item{tbl}{A \code{select_life_table} object.}

\item{x_sel}{Numeric vector of ages at selection.}

\item{t}{Numeric vector of durations since selection.}
}
\value{
Numeric vector of survivor values.
}
\description{
Returns \eqn{l_{[x]+t}} from a select life table.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{GMF_rollforward_ul}
\alias{GMF_rollforward_ul}
\title{Guaranteed maturity fund roll-forward}
\usage{
GMF_rollforward_ul(GMF_prev, GMP, r, policy_charge, i)
}
\arguments{
\item{GMF_prev}{Prior guaranteed maturity fund.}

\item{GMP}{Guaranteed maturity premium.}

\item{r}{Percent-of-premium expense rate in \code{[0, 1]}.}

\item{policy_charge}{Guaranteed policy charge.}

\item{i}{Guaranteed annual effective interest rate greater than \code{-1}.}
}
\value{
A numeric vector.
}
\description{
Computes a one-period guaranteed maturity fund roll-forward. Arguments may
be scalars or vectors and follow common-length recycling.
}
\examples{
GMF_rollforward_ul(140.40, 14.49, 0.04, 11.80, 0.03)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_fractional.R
\name{nkqx}
\alias{nkqx}
\title{Curtate death probability from a life table}
\usage{
nkqx(tbl, x, k)
}
\arguments{
\item{tbl}{A life_table object.}

\item{x}{Numeric vector of ages.}

\item{k}{Nonnegative integer.}
}
\value{
Numeric vector of \eqn{{}_{k|} q_x} values.
}
\description{
Computes \eqn{{}_{k|} q_x = {}_k p_x - {}_{k+1} p_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{nAbarx}
\alias{nAbarx}
\title{Continuous deferred insurance APV}
\usage{
nAbarx(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Deferral period.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes
\eqn{{}_{n\mid}\bar{A}_x = v^n {}_n p_x \bar{A}_{x+n}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{NC_TUC_db}
\alias{NC_TUC_db}
\title{Traditional Unit Credit normal cost for a DB plan}
\usage{
NC_TUC_db(accrual_benefit, v_to_ret, p_surv, adue_ret)
}
\arguments{
\item{accrual_benefit}{Benefit accrued in the current year.}

\item{v_to_ret}{Discount factor to retirement.}

\item{p_surv}{Active-service survival probability to retirement.}

\item{adue_ret}{Retirement annuity factor.}
}
\value{
TUC normal cost.
}
\description{
Computes the TUC normal cost as the APV of the current year's accrual.
}
\examples{
NC_TUC_db(accrual_benefit = 1560, v_to_ret = 0.5, p_surv = 0.9, adue_ret = 12)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{replacement_ratio_dc}
\alias{replacement_ratio_dc}
\title{Replacement ratio for a defined contribution plan}
\usage{
replacement_ratio_dc(x, z, Sx, c, i, adue_z, g = NULL, s = NULL)
}
\arguments{
\item{x}{Scalar entry age.}

\item{z}{Scalar retirement age.}

\item{Sx}{Positive scalar salary at age \code{x}.}

\item{c}{Scalar contribution rate in \code{[0, 1]}.}

\item{i}{Scalar annual effective investment return greater than \code{-1}.}

\item{adue_z}{Positive scalar whole-life annuity-due factor at retirement.}

\item{g}{Optional scalar annual salary growth rate greater than \code{-1}.}

\item{s}{Optional positive salary-scale vector of length \code{z - x}.}
}
\value{
A numeric scalar.
}
\description{
Computes annual retirement income divided by salary in the final
pre-retirement year.
}
\examples{
replacement_ratio_dc(
  x = 30,
  z = 65,
  Sx = 50000,
  c = 0.10,
  i = 0.05,
  adue_z = 12,
  g = 0.04
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{A2x_m}
\alias{A2x_m}
\title{Second moment of m-thly whole life insurance PV}
\usage{
A2x_m(x, i, m, model, ..., tol = 1e-12, j_max = 100000L)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}

\item{tol}{Numerical tolerance for truncating the infinite sum.}

\item{j_max}{Maximum number of m-thly intervals in the sum.}
}
\value{
Numeric vector of second moments.
}
\description{
Computes \eqn{{}^{2}A_x^{(m)}} by evaluating \eqn{A_x^{(m)}}
at doubled force.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{Abarxn1}
\alias{Abarxn1}
\title{Continuous term insurance APV}
\usage{
Abarxn1(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes
\eqn{\bar{A}_{x:\overline{n}|}^{1} = \int_0^n v^t \, {}_t p_x \mu_{x+t}\,dt}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{Pbar_trapz_ms}
\alias{Pbar_trapz_ms}
\title{Continuous premium approximation in a disability model}
\usage{
Pbar_trapz_ms(t, tp00, tp01, delta, mu02, mu12, B02 = 1, B12 = 1, R = 0)
}
\arguments{
\item{t}{Strictly increasing nonnegative time points.}

\item{tp00}{Healthy-state probabilities.}

\item{tp01}{Disabled-state probabilities.}

\item{delta}{Force of interest.}

\item{mu02}{Healthy-to-deceased intensity function.}

\item{mu12}{Disabled-to-deceased intensity function.}

\item{B02}{Benefit on death while healthy.}

\item{B12}{Benefit on death while disabled.}

\item{R}{Continuous disability income rate.}
}
\value{
A numeric scalar.
}
\description{
Approximates a continuous premium rate by trapezoidal integration.
}
\examples{
mu01 <- function(t) 0.10 * t + 0.20
mu02 <- function(t) 0.20
mu10 <- function(t) 0.50
mu12 <- function(t) 0.125 * t + 0.20

probs <- tp00_tp01_euler(
  h = 0.10,
  n = 2,
  mu01 = mu01,
  mu02 = mu02,
  mu10 = mu10,
  mu12 = mu12
)

Pbar_trapz_ms(
  t = probs$t,
  tp00 = probs$tp00,
  tp01 = probs$tp01,
  delta = 0.04,
  mu02 = mu02,
  mu12 = mu12,
  B02 = 1000,
  B12 = 1000,
  R = 1000
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{Abarxn}
\alias{Abarxn}
\title{Continuous endowment insurance APV}
\usage{
Abarxn(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes
\eqn{\bar{A}_{x:\overline{n}|} =
\bar{A}_{x:\overline{n}|}^{1} + v^n {}_n p_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_varying_benefits.R
\name{DAbarxn1}
\alias{DAbarxn1}
\title{Piecewise-continuous decreasing n-year term insurance}
\usage{
DAbarxn1(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes
\deqn{(D\bar{A})_{x:\overline{n}|}^{1}
= \int_0^n \lfloor n+1-t \rfloor v^t {}_tp_x \mu_{x+t}\,dt.}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{AAL_TUC_db}
\alias{AAL_TUC_db}
\title{Traditional Unit Credit accrued liability}
\usage{
AAL_TUC_db(accrued_benefit, v_to_ret, p_surv, adue_ret)
}
\arguments{
\item{accrued_benefit}{Nonnegative accrued benefit.}

\item{v_to_ret}{Nonnegative discount factor from the valuation date to
retirement.}

\item{p_surv}{Survival or active-service probability to retirement in
\code{[0, 1]}.}

\item{adue_ret}{Positive retirement annuity-due factor.}
}
\value{
A numeric vector.
}
\description{
Computes the actuarial present value of the benefit accrued through the
valuation date.
}
\examples{
AAL_TUC_db(
  accrued_benefit = 12000,
  v_to_ret = 0.5,
  p_surv = 0.9,
  adue_ret = 12
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{thiele_dVdt_01}
\alias{thiele_dVdt_01}
\title{Reserve derivatives for a disability model with recovery}
\usage{
thiele_dVdt_01(t, V0, V1, delta, Pbar, B, R, mu01, mu02, mu10, mu12)
}
\arguments{
\item{t}{Time.}

\item{V0}{Healthy-state reserve.}

\item{V1}{Disabled-state reserve.}

\item{delta}{Force of interest.}

\item{Pbar}{Continuous premium rate.}

\item{B}{Death benefit.}

\item{R}{Continuous disability income rate.}

\item{mu01}{Healthy-to-disabled intensity function.}

\item{mu02}{Healthy-to-deceased intensity function.}

\item{mu10}{Disabled-to-healthy intensity function.}

\item{mu12}{Disabled-to-deceased intensity function.}
}
\value{
A named vector for scalar input or a two-column matrix for
  vectorized input.
}
\description{
Computes the coupled Thiele reserve derivatives. Numeric arguments may be
scalar or compatible vectors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{AV_path_ul_typeB}
\alias{AV_path_ul_typeB}
\title{Type B universal life account-value path}
\usage{
AV_path_ul_typeB(G, r, e, qx, ic, B, iq = ic, AV0 = 0)
}
\arguments{
\item{G}{Premium amount by period.}

\item{r}{Percent-of-premium expense rate by period. Values must lie in
\code{[0, 1]}.}

\item{e}{Fixed expense by period.}

\item{qx}{Mortality probability by period.}

\item{ic}{Credited annual effective interest rate by period. Values must be
greater than \code{-1}.}

\item{B}{Face amount by period.}

\item{iq}{Interest rate used in the cost-of-insurance calculation. Defaults
to \code{ic}; values must be greater than \code{-1}.}

\item{AV0}{Nonnegative scalar initial account value.}
}
\value{
A data frame containing the policy duration, premium,
  net contribution, cost-of-insurance charge, and account value.
}
\description{
Computes a year-by-year Type B universal life account-value path. Premiums
and expenses are applied at the beginning of each period, followed by the
cost-of-insurance charge and credited interest.
}
\examples{
qx <- c(0.00076, 0.00081, 0.00085, 0.00090, 0.00095)
r <- c(0.75, rep(0.10, 4))
e <- c(100, rep(20, 4))

AV_path_ul_typeB(
  G = 5000,
  r = r,
  e = e,
  qx = qx,
  ic = 0.03,
  B = 100000
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{APV_gross_premiums}
\alias{APV_gross_premiums}
\title{Actuarial present value of gross premiums}
\usage{
APV_gross_premiums(G, r, p_tau)
}
\arguments{
\item{G}{Nonnegative gross premium vector.}

\item{r}{Annual effective risk discount rate. May be scalar or vector;
values must be greater than \code{-1}.}

\item{p_tau}{One-year in-force probabilities. For \code{n > 1}, this may
have length \code{n - 1} or \code{n}; the final value is ignored when
length \code{n}. For one premium, use \code{numeric(0)} or one value.}
}
\value{
A numeric vector with one value for each rate in \code{r}.
}
\description{
Computes the actuarial present value of premiums weighted by contract
persistency at the start of each policy year.
}
\examples{
APV_gross_premiums(
  G = rep(95, 3),
  r = 0.10,
  p_tau = c(0.99858, 0.99847, 0.99834)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{tVx_ret}
\alias{tVx_ret}
\title{Retrospective whole life reserve}
\usage{
tVx_ret(x, t, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age. May be scalar or vector.}

\item{t}{Nonnegative integer duration. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the actuarial functions.}

\item{tbl}{Optional life table object. Supply by name.}
}
\value{
Numeric vector of retrospective reserve values.
}
\description{
Computes the retrospective net level premium reserve
\deqn{{}_tV_x =
P_x\ddot{s}_{x:\overline{t}|}
-
\frac{A_{x:\overline{t}|}^{1}}{{}_tE_x}.}
}
\details{
The mortality basis may be supplied through either a life table or a
parametric survival model.
}
\examples{
tVx_ret(
  40,
  t = 10,
  i = 0.05,
  model = "uniform",
  omega = 100
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{var_Abarxn1}
\alias{var_Abarxn1}
\title{Variance of continuous term insurance PV}
\usage{
var_Abarxn1(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of variances.
}
\description{
Computes
\eqn{\mathrm{Var}(\bar{Z}_{x:\overline{n}|}^{1}) =
{}^{2}\bar{A}_{x:\overline{n}|}^{1}
- (\bar{A}_{x:\overline{n}|}^{1})^2}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{vt_var}
\alias{vt_var}
\title{Discount factors under variable annual interest rates}
\usage{
vt_var(i)
}
\arguments{
\item{i}{Numeric vector of annual effective interest rates. Each value must
be greater than \code{-1}.}
}
\value{
A numeric vector of cumulative discount factors with the same
  length as \code{i}.
}
\description{
Computes cumulative discount factors for a sequence of annual effective
interest rates:
\deqn{
v_t = \prod_{k=1}^{t}(1+i_k)^{-1},
\qquad t=1,\ldots,n.
}
}
\examples{
vt_var(c(0.06, 0.07, 0.08))
vt_var(c(-0.01, 0.02, 0.03))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_expectations.R
\name{ex_temp_complete_tab}
\alias{ex_temp_complete_tab}
\title{Temporary complete expectation of life from a life table}
\usage{
ex_temp_complete_tab(tbl, x, n, assumption = c("udd", "cf", "balducci"))
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Numeric vector of integer ages.}

\item{n}{Numeric vector of nonnegative durations.}

\item{assumption}{One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.}
}
\value{
Numeric vector of temporary complete expectations.
}
\description{
Computes
\eqn{\overset{\circ}{e}_{x:\overline{n}|} =
\int_0^n {}_t p_x \, dt}
using a within-year assumption:
\code{"udd"}, \code{"cf"}, or \code{"balducci"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{multilife_contingent_probabilities}
\alias{multilife_contingent_probabilities}
\alias{tqxy1}
\alias{tqyx1}
\alias{tqxy2}
\alias{tqyx2}
\title{Contingent multi-life probabilities}
\usage{
tqxy1(x, y, n, tbl = NULL, model = NULL, ...)

tqyx1(x, y, n, tbl = NULL, model = NULL, ...)

tqxy2(x, y, n, tbl = NULL, model = NULL, ...)

tqyx2(x, y, n, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{n}{Term in years. May be scalar or vector.}

\item{tbl}{Optional life table object retained for backward compatibility.
Continuous calculations currently require \code{model}.}

\item{model}{Parametric survival model.}

\item{...}{Additional parameters passed to the survival and hazard
functions.}
}
\value{
A numeric vector of contingent probabilities.
}
\description{
Computes probabilities associated with the order of death of two
independent lives over a specified term.
}
\details{
\code{tqxy1()} computes the probability that the first life dies before
the second life within \eqn{n} years.

\code{tqyx1()} computes the probability that the second life dies before
the first life within \eqn{n} years.

\code{tqxy2()} computes the probability that the first life dies after
the second life but within \eqn{n} years.

\code{tqyx2()} computes the probability that the second life dies after
the first life but within \eqn{n} years.

These functions require a parametric survival model because an annual
life table does not uniquely determine the within-year order of death
without an additional interpolation assumption.


All calculations assume the two future lifetimes are independent.

The probabilities are obtained by integrating the joint survival
function together with the appropriate force of mortality.

The four functions partition the probability that one of the two lives
dies within the specified term according to the order of death.
}
\examples{
tqxy1(
  40, 50,
  n = 10,
  model = "uniform",
  omega = 100
)

tqyx1(
  40, 50,
  n = 10,
  model = "uniform",
  omega = 100
)

tqxy2(
  40, 50,
  n = 10,
  model = "uniform",
  omega = 100
)

tqyx2(
  40, 50,
  n = 10,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{tqxprime_mudd}
\alias{tqxprime_mudd}
\title{Fractional-year associated single-decrement probabilities under MUDD}
\usage{
tqxprime_mudd(qxj, t)
}
\arguments{
\item{qxj}{Numeric vector of dependent cause-specific decrement
probabilities.}

\item{t}{A single time in \eqn{[0,1]}.}
}
\value{
A numeric vector of associated single-decrement probabilities.
}
\description{
Computes the associated single-decrement probabilities through time
\code{t} under the multiple-decrement uniform distribution assumption.
}
\examples{
tqxprime_mudd(c(0.20, 0.10), t = 0.5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R, R/reserve_ch11_ext.R
\name{tVGx}
\alias{tVGx}
\title{Whole life gross premium reserve}
\usage{
tVGx(
  x,
  t,
  i,
  G,
  benefit = 1,
  renewal_premium_pct = 0,
  renewal_policy_exp = 0,
  settlement_exp = 0,
  tbl = NULL,
  model = NULL,
  ...
)

tVGx(
  x,
  t,
  i,
  G,
  benefit = 1,
  renewal_premium_pct = 0,
  renewal_policy_exp = 0,
  settlement_exp = 0,
  tbl = NULL,
  model = NULL,
  ...
)
}
\arguments{
\item{x}{Issue age.}

\item{t}{Duration.}

\item{i}{Effective annual interest rate.}

\item{G}{Gross annual premium.}

\item{benefit}{Insurance amount. Default 1.}

\item{renewal_premium_pct}{Renewal percent-of-premium expense.}

\item{renewal_policy_exp}{Renewal per-policy expense.}

\item{settlement_exp}{Settlement expense at death.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.

Numeric vector.
}
\description{
Computes the Chapter 11 prospective gross premium reserve for a
fully discrete whole life insurance with annual premiums and
renewal expenses.

Computes the Chapter 11 prospective gross premium reserve for a
fully discrete whole life insurance with annual premiums and
renewal expenses.
}
\details{
This function is intended for durations after issue, where future
expenses are modeled through renewal premium expenses, renewal
per-policy expenses, and settlement expense.
}
\examples{
tVGx(
  x = 40, t = 10, i = 0.05, G = 0.03,
  benefit = 1, renewal_premium_pct = 0.10,
  renewal_policy_exp = 0.002, settlement_exp = 0.02,
  model = "uniform", omega = 100
)
tVGx(
  x = 40, t = 10, i = 0.05, G = 0.03,
  benefit = 1, renewal_premium_pct = 0.10,
  renewal_policy_exp = 0.002, settlement_exp = 0.02,
  model = "uniform", omega = 100
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mortality_improvement.R
\name{mortality_improvement_projection}
\alias{mortality_improvement_projection}
\alias{qx_proj}
\alias{px_proj}
\alias{tpx_improved}
\alias{axn_improved}
\alias{naxn_improved}
\alias{ax_improved}
\title{Mortality improvement projection functions}
\usage{
qx_proj(qx_base, AAx, base_year, proj_year)

px_proj(qx_base, AAx, base_year, proj_year)

tpx_improved(x0, n, qx_base_vec, AAx_vec, base_year, issue_year)

axn_improved(x0, n, i, qx_base_vec, AAx_vec, base_year, issue_year)

naxn_improved(x0, u, n, i, qx_base_vec, AAx_vec, base_year, issue_year)

ax_improved(x0, i, qx_base_vec, AAx_vec, base_year, issue_year)
}
\arguments{
\item{qx_base}{Base-year one-year death probability.}

\item{AAx}{Mortality improvement factor.}

\item{base_year}{Base year.}

\item{proj_year}{Projection year. May be scalar or vector.}

\item{x0}{Issue age.}

\item{n}{Number of years.}

\item{qx_base_vec}{Base-year death probabilities for successive ages.}

\item{AAx_vec}{Mortality improvement factors for successive ages.}

\item{issue_year}{Issue year.}

\item{i}{Effective annual interest rate.}

\item{u}{Deferral period in years.}
}
\value{
Numeric vector of projected one-year death probabilities.
}
\description{
Functions for projecting one-year death and survival probabilities under
mortality improvement and evaluating annuity values under projected
mortality.
}
\details{
The standard projection used is
\deqn{q_x^{[Y]} = q_x^{[B]} (1 - AA_x)^{Y-B},}
where \eqn{B} is the base year, \eqn{Y} is the projection year, and
\eqn{AA_x} is the mortality improvement factor at age \eqn{x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_approximations.R
\name{annuity_approximations_woolhouse2}
\alias{annuity_approximations_woolhouse2}
\alias{ax_m_woolhouse2}
\alias{adotx_m_woolhouse2}
\alias{axn_m_woolhouse2}
\alias{adotxn_m_woolhouse2}
\alias{nax_m_woolhouse2}
\alias{nadotx_m_woolhouse2}
\alias{sxn_m_woolhouse2}
\alias{sdotxn_m_woolhouse2}
\alias{abarx_woolhouse2}
\title{Woolhouse 2-term annuity approximations}
\usage{
ax_m_woolhouse2(x, m, i, model, ...)

adotx_m_woolhouse2(x, m, i, model, ...)

nax_m_woolhouse2(x, n, m, i, model, ...)

nadotx_m_woolhouse2(x, n, m, i, model, ...)

axn_m_woolhouse2(x, n, m, i, model, ...)

adotxn_m_woolhouse2(x, n, m, i, model, ...)

sxn_m_woolhouse2(x, n, m, i, model, ...)

sdotxn_m_woolhouse2(x, n, m, i, model, ...)

abarx_woolhouse2(x, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{m}{Number of payments per year.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}

\item{n}{Term.}
}
\value{
Numeric vector.
}
\description{
Woolhouse 2-term approximations for fractional-payment annuity functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{GI_disc}
\alias{GI_disc}
\title{Interest gain for a discrete insurance contract}
\usage{
GI_disc(Vt, Vt1, P, i_actual, q_assumed, B = 1)
}
\arguments{
\item{Vt}{Reserve at duration t.}

\item{Vt1}{Reserve at duration t+1.}

\item{P}{Net premium for the year.}

\item{i_actual}{Actual annual effective interest rate.}

\item{q_assumed}{Assumed mortality rate for the year.}

\item{B}{Benefit amount. Defaults to 1.}
}
\value{
A numeric vector of values.
}
\description{
Interest gain for a discrete insurance contract
}
\examples{
GI_disc(Vt = 0.1, Vt1 = 0.11, P = 0.02, i_actual = 0.05, q_assumed = 0.01)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interest_theory.R
\name{annuity_certain}
\alias{annuity_certain}
\title{Present value of a level annuity-certain}
\usage{
annuity_certain(n, i, due = FALSE, m = 1, cont = FALSE)
}
\arguments{
\item{n}{Number of payments or periods. May be scalar or vector.}

\item{i}{Effective interest rate per period. May be scalar or vector.}

\item{due}{If `TRUE`, annuity-due; otherwise annuity-immediate.}

\item{m}{Payment frequency per period. `m = 1` means annual.}

\item{cont}{If `TRUE`, use the continuous payment model.}
}
\value{
Numeric vector of present values.
}
\description{
Computes the present value of an n-period annuity certain with level
payments of 1 per period.
}
\examples{
annuity_certain(n = 10, i = 0.05)
annuity_certain(n = c(5, 10), i = 0.05)
annuity_certain(n = 10, i = c(0.03, 0.05))
annuity_certain(n = 10, i = 0.05, due = TRUE)
annuity_certain(n = 10, i = 0.05, cont = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_mthly.R
\name{annuity_mthly}
\alias{annuity_mthly}
\alias{ax_m}
\alias{adotx_m}
\alias{axn_m}
\alias{adotxn_m}
\alias{nax_m}
\alias{nadotx_m}
\alias{sxn_m}
\alias{sdotxn_m}
\title{m-thly contingent annuity functions}
\usage{
ax_m(x, m, i, model, ..., k_max = 2e+05, tol = 1e-12)

adotx_m(x, m, i, model, ..., k_max = 2e+05, tol = 1e-12)

axn_m(x, n, m, i, model, ...)

adotxn_m(x, n, m, i, model, ...)

nax_m(x, n, m, i, model, ..., k_max = 2e+05, tol = 1e-12)

nadotx_m(x, n, m, i, model, ..., k_max = 2e+05, tol = 1e-12)

sxn_m(x, n, m, i, model, ...)

sdotxn_m(x, n, m, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{m}{Number of payments per year.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model name.}

\item{...}{Additional parameters passed to the survival model.}

\item{k_max}{Maximum summation horizon for non-terminating models.}

\item{tol}{Truncation tolerance for non-terminating models.}

\item{n}{Term or deferral period in years.}
}
\value{
Numeric vector containing the requested m-thly annuity value or actuarial
accumulated value.
}
\description{
Life annuities payable m-thly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{abarx_y}
\alias{abarx_y}
\title{Continuous reversionary annuity to (y) after death of (x)}
\usage{
abarx_y(x, y, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{\overline{a}_{x|y} = \overline{a}_y - \overline{a}_{xy}}.
}
\examples{
abarx_y(40, 50, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{tpx_tau_cf}
\alias{tpx_tau_cf}
\title{Total survival under constant cause-specific forces}
\usage{
tpx_tau_cf(mu, t)
}
\arguments{
\item{mu}{Numeric vector of nonnegative cause-specific forces.}

\item{t}{Nonnegative time. May be scalar or vector.}
}
\value{
A numeric vector.
}
\description{
Computes
\deqn{
{}_tp_x^{(\tau)}
=
\exp\left(-t\sum_j\mu_j\right).
}
}
\examples{
tpx_tau_cf(c(0.10, 0.20), 5)
tpx_tau_cf(c(0.10, 0.20), c(1, 5, 10))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R
\name{fractional_duration_reserves}
\alias{fractional_duration_reserves}
\alias{tsVx}
\alias{meanVx}
\title{Fractional-duration whole life reserves}
\usage{
tsVx(x, t, s, i, tbl = NULL, model = NULL, ...)

meanVx(x, t, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Issue age. May be scalar or vector.}

\item{t}{Nonnegative integer duration. May be scalar or vector.}

\item{s}{Fractional duration in \eqn{[0,1]}. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the actuarial functions.}
}
\value{
A numeric vector of reserve values.
}
\description{
Computes fractional-duration whole life reserves using linear
interpolation between the reserve immediately after the premium at
duration \eqn{t} and the reserve at duration \eqn{t+1}.
}
\details{
\code{tsVx()} computes the reserve at fractional duration
\eqn{t+s}, where \eqn{0 \le s \le 1}.

\code{meanVx()} computes the reserve at the midpoint of the policy
year (\eqn{s=0.5}).


The fractional reserve is computed using

\deqn{
{}_{t+s}V_x
=
({}_tV_x + P_x)(1-s)
+
{}_{t+1}V_x s.
}

The function \code{meanVx()} is a convenience wrapper corresponding to
\eqn{s=0.5}.
}
\examples{
tsVx(
  40,
  t = 10,
  s = 0.5,
  i = 0.05,
  model = "uniform",
  omega = 100
)

meanVx(
  40,
  t = 10,
  i = 0.05,
  model = "uniform",
  omega = 100
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{reserve_ch10_ext}
\alias{reserve_ch10_ext}
\title{Extended reserve functions for Chapter 10}
\description{
Additional Chapter 10 reserve functions:
retrospective reserves, deferred insurance reserves, annuity reserves,
continuous-time gain/loss helpers, and Euler/Thiele approximation helpers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{var_Abarx}
\alias{var_Abarx}
\title{Variance of continuous whole life insurance PV}
\usage{
var_Abarx(x, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of variances.
}
\description{
Computes \eqn{\mathrm{Var}(\bar{Z}_x) =
{}^{2}\bar{A}_x - \bar{A}_x^2}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_varying_benefits.R
\name{IAx}
\alias{IAx}
\title{Increasing whole life insurance}
\usage{
IAx(x, i, tbl = NULL, model = NULL, ..., tol = 1e-12, k_max = 5000)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional model parameters.}

\item{tol}{Numerical tolerance for truncation.}

\item{k_max}{Maximum number of terms.}
}
\value{
Numeric vector.
}
\description{
Computes
\deqn{(IA)_x = \sum_{k=0}^{\infty} (k+1) v^{k+1} \Pr(K_x = k).}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_approximations.R
\name{annuity_approximations_woolhouse3}
\alias{annuity_approximations_woolhouse3}
\alias{ax_m_woolhouse3}
\alias{adotx_m_woolhouse3}
\alias{axn_m_woolhouse3}
\alias{adotxn_m_woolhouse3}
\alias{nax_m_woolhouse3}
\alias{nadotx_m_woolhouse3}
\alias{abarx_woolhouse3}
\title{Woolhouse 3-term annuity approximations}
\usage{
ax_m_woolhouse3(x, m, i, model, ...)

adotx_m_woolhouse3(x, m, i, model, ...)

nax_m_woolhouse3(x, n, m, i, model, ...)

nadotx_m_woolhouse3(x, n, m, i, model, ...)

axn_m_woolhouse3(x, n, m, i, model, ...)

adotxn_m_woolhouse3(x, n, m, i, model, ...)

abarx_woolhouse3(x, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{m}{Number of payments per year.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}

\item{n}{Term.}
}
\value{
Numeric vector.
}
\description{
Woolhouse 3-term approximations for fractional-payment annuity functions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_varying_benefits.R
\name{IAxn1}
\alias{IAxn1}
\title{Increasing n-year term insurance}
\usage{
IAxn1(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes
\deqn{(IA)_{x:\overline{n}|}^{1}
= \sum_{k=0}^{n-1} (k+1) v^{k+1} \Pr(K_x = k).}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interest_theory.R
\name{interest_convert}
\alias{interest_convert}
\title{Convert between compound-interest quantities}
\usage{
interest_convert(i = NULL, d = NULL, delta = NULL, m = NULL)
}
\arguments{
\item{i}{Effective interest rate. May be scalar or vector.}

\item{d}{Effective discount rate. May be scalar or vector.}

\item{delta}{Force of interest. May be scalar or vector.}

\item{m}{Optional positive integer compounding frequency for the nominal
rate convertible m-thly.}
}
\value{
A list with elements `i`, `d`, `delta`, and, if `m` is supplied,
  `im` and `m`.
}
\description{
Provides consistent conversions between effective interest rate, effective
discount rate, force of interest, and optional nominal interest rate
convertible m-thly.
}
\details{
Exactly one of `i`, `d`, or `delta` must be provided.
}
\examples{
interest_convert(i = 0.05)
interest_convert(i = c(0.03, 0.05, 0.07))
interest_convert(d = 0.04761905)
interest_convert(delta = log(1.05))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{var_Axn_m}
\alias{var_Axn_m}
\title{Variance of m-thly endowment insurance PV}
\usage{
var_Axn_m(x, n, i, m, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of variances.
}
\description{
Variance of m-thly endowment insurance PV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{ch12_multilife_cont_params}
\alias{ch12_multilife_cont_params}
\title{Shared parameters for Chapter 12 continuous multi-life functions}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{i}{Effective annual interest rate.}

\item{n}{Term in years.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\description{
Shared parameters for Chapter 12 continuous multi-life functions
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R, R/reserve_ch11_ext.R
\name{betaF}
\alias{betaF}
\title{Full preliminary term renewal modified premium}
\usage{
betaF(x, i, tbl = NULL, model = NULL, ...)

betaF(x, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Issue age.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.

Numeric vector.
}
\description{
Computes the FPT renewal premium
\eqn{\beta^F = P_{x+1}}
for whole life insurance.

Computes the FPT renewal premium
\eqn{\beta^F = P_{x+1}}
for whole life insurance.
}
\examples{
betaF(40, i = 0.05, model = "uniform", omega = 100)
betaF(40, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R
\name{reserve_ch11}
\alias{reserve_ch11}
\title{Reserve functions for Chapter 11}
\description{
Chapter 11 functions for modified reserves, fractional-duration reserves,
gross premium reserves, expense reserves, and gross gain analysis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_mthly.R
\name{Axn1_m}
\alias{Axn1_m}
\title{m-thly term insurance APV}
\usage{
Axn1_m(x, n, i, m, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{m}{Positive integer payment frequency.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of APVs.
}
\description{
Computes
\eqn{A_{x:\overline{n}|}^{1(m)} =
\sum_{j=0}^{mn-1} v^{(j+1)/m}
\Pr(j/m < T_x \le (j+1)/m)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{A2barx}
\alias{A2barx}
\title{Second moment of continuous whole life insurance PV}
\usage{
A2barx(x, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of second moments.
}
\description{
Computes \eqn{{}^{2}\bar{A}_x} by evaluating \eqn{\bar{A}_x}
at doubled force.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{PnAdotx}
\alias{PnAdotx}
\title{Deferred annuity-due premium}
\usage{
PnAdotx(x, n, i, model, ...)
}
\arguments{
\item{x}{Issue age.}

\item{n}{Deferral period.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes
\eqn{P({}_{n|}\ddot{a}_x) = {}_{n|}\ddot{a}_x / \ddot{a}_{x:\overline{n}|}}.
}
\examples{
PnAdotx(40, n = 20, i = 0.05, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{insurance_utils}
\alias{insurance_utils}
\title{Insurance utilities}
\description{
Helper functions for insurance models.
}
\details{
These utilities focus on:
\itemize{
  \item doubled-force calculations for second moments,
  \item UDD approximation multipliers,
  \item UDD approximations for continuous and m-thly insurance functions.
}

General interest conversion functions such as \code{interest_convert()}
and \code{discount()} are defined elsewhere in the package and are reused here.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{markov_nstep_prob}
\alias{markov_nstep_prob}
\title{Multi-step transition probability}
\usage{
markov_nstep_prob(P, n, i, j)
}
\arguments{
\item{P}{Square transition-probability matrix.}

\item{n}{Nonnegative integer number of steps.}

\item{i}{Starting-state index.}

\item{j}{Ending-state index.}
}
\value{
A numeric scalar.
}
\description{
Computes an entry of the matrix power \eqn{P^n}.
}
\examples{
P <- matrix(c(0.9, 0.1, 0, 1), nrow = 2, byrow = TRUE)
markov_nstep_prob(P, n = 3, i = 1, j = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter14_multiple_decrement.R
\name{AS_path_md}
\alias{AS_path_md}
\title{General projected asset-share path}
\usage{
AS_path_md(AS0, G, r, e, b_mat, q_mat, p_tau, i, b_surv = NULL)
}
\arguments{
\item{AS0}{Initial asset share.}

\item{G}{Premium amount by policy year.}

\item{r}{Percent-of-premium expense rate by policy year.}

\item{e}{Fixed expense by policy year.}

\item{b_mat}{Matrix of decrement benefits.}

\item{q_mat}{Matrix of decrement probabilities.}

\item{p_tau}{In-force probability by policy year.}

\item{i}{Effective annual interest rate by policy year.}

\item{b_surv}{Survival benefit by policy year.}
}
\value{
A data frame with policy year \code{k} and asset share \code{AS}.
}
\description{
Computes projected asset shares for a multiple-decrement contract. Rows of
\code{b_mat} and \code{q_mat} represent policy years and columns represent
decrement causes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{var_Axn}
\alias{var_Axn}
\title{Variance of endowment insurance PV}
\usage{
var_Axn(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}
}
\value{
Numeric vector of variances.
}
\description{
Variance of endowment insurance PV
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/survival_models.R
\name{S0}
\alias{S0}
\title{Survival function for age-at-failure}
\usage{
S0(t, model, ...)
}
\arguments{
\item{t}{Numeric vector of times.}

\item{model}{One of \code{"uniform"}, \code{"exponential"},
\code{"gompertz"}, \code{"makeham"}, or \code{"weibull"}.}

\item{...}{Model parameters.}
}
\value{
Numeric vector of survival probabilities.
}
\description{
Computes \eqn{S_0(t)=Pr(T_0>t)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decrement_ch13_ext.R
\name{pxtau}
\alias{pxtau}
\title{Total one-year survival probability}
\usage{
pxtau(qxj)
}
\arguments{
\item{qxj}{Numeric vector of cause-specific decrement probabilities.}
}
\value{
A numeric scalar.
}
\description{
Computes
\deqn{p_x^{(\tau)}=1-q_x^{(\tau)}.}
}
\examples{
pxtau(c(0.011, 0.100))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{Pi_signature}
\alias{Pi_signature}
\title{Profit signature}
\usage{
Pi_signature(Pr, p_tau)
}
\arguments{
\item{Pr}{Profit vector of length \code{n + 1}.}

\item{p_tau}{One-year in-force probabilities. For \code{n > 1}, this may
have length \code{n - 1} or \code{n}; the final value is ignored when
length \code{n}. For a one-year contract, use \code{numeric(0)} or a
single probability.}
}
\value{
A named numeric vector with the same length as \code{Pr}.
}
\description{
Converts policy-year expected profits into a profit signature by weighting
each future expected profit by the probability that the contract is in
force at the start of that policy year.
}
\examples{
Pr <- c(-15.00, 8.42, 8.40, 8.61)
Pi_signature(Pr, p_tau = c(0.99858, 0.99847, 0.99834))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_expectations.R
\name{ex_temp_curtate_tab}
\alias{ex_temp_curtate_tab}
\title{Temporary curtate expectation of life from a life table}
\usage{
ex_temp_curtate_tab(tbl, x, n)
}
\arguments{
\item{tbl}{A \code{life_table} object.}

\item{x}{Numeric vector of integer ages.}

\item{n}{Numeric vector of nonnegative integers.}
}
\value{
Numeric vector of temporary curtate expectations.
}
\description{
Computes
\eqn{e_{x:\overline{n}|} = \sum_{k=1}^{n} {}_k p_x}
for integer \eqn{n} in the discrete tabular setting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{chapter15_variable_interest_apv}
\alias{chapter15_variable_interest_apv}
\alias{nEx_var}
\alias{Axn1_var}
\alias{Axn_var}
\alias{axn_var}
\title{Variable-interest actuarial present value functions}
\usage{
nEx_var(qx, i, benefit = 1)

Axn1_var(qx, i, benefit = 1)

Axn_var(qx, i, benefit = 1)

axn_var(qx, i, type = c("immediate", "due"), benefit = 1)
}
\arguments{
\item{qx}{Numeric vector of one-year mortality rates
\eqn{q_x, q_{x+1}, \dots, q_{x+n-1}}.}

\item{i}{Numeric vector of annual effective interest rates
\eqn{i_1, i_2, \dots, i_n}.}

\item{benefit}{Amount of each annuity payment.}

\item{type}{Either \code{"immediate"} or \code{"due"}.}
}
\value{
A numeric scalar.

A numeric scalar.

A numeric scalar.

A numeric scalar.
}
\description{
Chapter 15 functions for actuarial present values under variable annual
effective interest rates interpreted as a yearly scenario
\eqn{i_1, i_2, \dots, i_n}.

Computes the APV of an \eqn{n}-year pure endowment under a variable annual
interest scenario:
\deqn{
{}_nE_x = v_n \cdot {}_np_x.
}

Computes the APV of an \eqn{n}-year term insurance with benefit paid at the
end of the year of death under a variable annual interest scenario:
\deqn{
A_{x:\overline{n}|}^1 = \sum_{t=1}^{n} v_t \cdot {}_{t-1}p_x \cdot q_{x+t-1}.
}

Computes the APV of an \eqn{n}-year endowment insurance under a variable
annual interest scenario:
\deqn{
A_{x:\overline{n}|} = A_{x:\overline{n}|}^1 + {}_nE_x.
}

Computes the APV of an \eqn{n}-year temporary life annuity under a variable
annual interest scenario.
}
\details{
If a benefit amount is supplied, the function returns that benefit times the
APV factor.

If a benefit amount is supplied, the function returns that benefit times the
APV factor.

If a benefit amount is supplied, the function returns that benefit times the
APV factor.

For an immediate annuity,
\deqn{
a_{x:\overline{n}|} = \sum_{t=1}^{n} v_t \cdot {}_tp_x.
}

For an annuity-due,
\deqn{
\ddot{a}_{x:\overline{n}|} = \sum_{t=0}^{n-1} v_t \cdot {}_tp_x,
}
with \eqn{v_0 = 1}.
}
\examples{
qx <- c(.03, .04, .05, .06, .07)
nEx_var(qx = qx, i = c(.06, .07, .08, .09, .10), benefit = 1000)

qx <- c(.03, .04, .05, .06, .07)
Axn1_var(qx = qx, i = c(.06, .07, .08, .09, .10))

qx <- c(.03, .04, .05, .06, .07)
Axn_var(qx = qx, i = c(.06, .07, .08, .09, .10))

qx <- rep(.02, 5)
axn_var(qx = qx, i = c(.06, .05, .04, .03, .03), type = "immediate")
axn_var(qx = qx, i = c(.03, .04, .05, .06, .07), type = "due")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10_ext.R
\name{Pnax}
\alias{Pnax}
\title{Net premium for a deferred annuity-immediate}
\usage{
Pnax(x, n, i, model = NULL, ..., tbl = NULL)
}
\arguments{
\item{x}{Issue age. May be scalar or vector.}

\item{n}{Positive integer deferral period. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the actuarial functions.}

\item{tbl}{Optional life table object. Supply by name.}
}
\value{
Numeric vector of net annual premiums.
}
\description{
Computes
\deqn{
P({}_{n|}a_x)
=
\frac{{}_{n|}a_x}
{\ddot{a}_{x:\overline{n}|}}.
}
}
\examples{
Pnax(
  40,
  n = 20,
  i = 0.05,
  model = "uniform",
  omega = 100
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter17_profit_analysis.R
\name{profit_margin}
\alias{profit_margin}
\title{Profit margin}
\usage{
profit_margin(NPV, APV_GP)
}
\arguments{
\item{NPV}{Net present value of profits.}

\item{APV_GP}{Positive actuarial present value of gross premiums.}
}
\value{
A numeric vector.
}
\description{
Computes net present value divided by the actuarial present value of gross
premiums. Scalar arguments are recycled to the common length.
}
\examples{
profit_margin(6.03, 259.52)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_fractional.R
\name{tpx_tab}
\alias{tpx_tab}
\title{Fractional survival probability from a life table}
\usage{
tpx_tab(tbl, x, t, assumption = c("udd", "cf", "balducci"))
}
\arguments{
\item{tbl}{A life_table object.}

\item{x}{Numeric vector of integer ages.}

\item{t}{Numeric vector of fractional durations with \eqn{0 \le t \le 1}.}

\item{assumption}{One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.}
}
\value{
Numeric vector of \eqn{{}_t p_x} values.
}
\description{
Computes \eqn{{}_t p_x} for \eqn{0 \le t \le 1} from a discrete life table
under UDD, constant force, or Balducci.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{multilife_ch12}
\alias{multilife_ch12}
\title{Multi-life functions for Chapter 12}
\description{
Core Chapter 12 functions for two-life joint-life and last-survivor models,
including survival probabilities, annuities, insurances, and reversionary annuities.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter16_universal_life.R
\name{iP_eiul}
\alias{iP_eiul}
\title{Point-to-point index growth rates}
\usage{
iP_eiul(index)
}
\arguments{
\item{index}{Numeric vector of strictly positive index values.}
}
\value{
A numeric vector with length one less than \code{index}.
}
\description{
Computes consecutive point-to-point growth rates from index values.
}
\examples{
iP_eiul(c(1000, 1050, 1200, 1100))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_discrete.R
\name{A2nEx}
\alias{A2nEx}
\title{Second moment of pure endowment PV}
\usage{
A2nEx(x, n, i, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term.}

\item{i}{Effective annual interest rate.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model name.}

\item{...}{Additional arguments passed to survival-model functions.}
}
\value{
Numeric vector of second moments.
}
\description{
Computes \eqn{{}^{2}{}_nE_x = (v')^n {}_n p_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_fractional.R
\name{nmxq}
\alias{nmxq}
\title{Deferred death probability from a life table}
\usage{
nmxq(tbl, x, n, m)
}
\arguments{
\item{tbl}{A life_table object.}

\item{x}{Numeric vector of ages.}

\item{n}{Nonnegative integer deferred period.}

\item{m}{Nonnegative integer subsequent period.}
}
\value{
Numeric vector of \eqn{{}_{n|m} q_x} values.
}
\description{
Computes the probability that a life aged x survives n years and then
dies within the following m years:
\eqn{{}_{n|m} q_x = {}_n p_x \cdot {}_m q_{x+n}}.
}
\details{
This function is for integer n and m in the discrete tabular setting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch11.R
\name{GTg_disc}
\alias{GTg_disc}
\title{Total gross gain for a discrete insurance contract}
\usage{
GTg_disc(
  VtG,
  Vt1G,
  G,
  i_actual,
  q_actual,
  r_actual = 0,
  e_actual = 0,
  s_actual = 0,
  b = 1
)
}
\arguments{
\item{VtG}{Gross reserve at duration \code{t}.}

\item{Vt1G}{Gross reserve at duration \code{t + 1}.}

\item{G}{Gross premium.}

\item{i_actual}{Actual annual effective interest rate.}

\item{q_actual}{Actual mortality probability.}

\item{r_actual}{Actual percent-of-premium expense rate.}

\item{e_actual}{Actual per-policy expense.}

\item{s_actual}{Actual settlement expense.}

\item{b}{Benefit amount.}
}
\value{
A numeric vector of total gains.
}
\description{
Computes total gain during one policy year under gross premiums,
gross reserves, actual mortality, actual interest, and actual expenses.
}
\examples{
GTg_disc(
  VtG = 0.10,
  Vt1G = 0.12,
  G = 0.02,
  i_actual = 0.05,
  q_actual = 0.01,
  r_actual = 0.03,
  e_actual = 0,
  s_actual = 0.01,
  b = 1
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{tqyx1}
\alias{tqyx1}
\title{Probability that (y) fails before (x) within n years}
\usage{
tqyx1(x, y, n, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{n}{Term in years.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{{}_n q_{xy}^{\hspace{1mm}1} = \int_0^n {}_tp_{xy}\mu_{y+t}\,dt}
under independence.
}
\examples{
tqyx1(40, 50, n = 10, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{last_survivor_annuities}
\alias{last_survivor_annuities}
\alias{adotxybarn}
\alias{axybarn}
\alias{adotxybar}
\alias{axybar}
\title{Last-survivor annuity functions}
\usage{
adotxybarn(x, y, n, i, tbl = NULL, model = NULL, ...)

axybarn(x, y, n, i, tbl = NULL, model = NULL, ...)

adotxybar(x, y, i, tbl = NULL, model = NULL, ..., k_max = 5000L, tol = 1e-12)

axybar(x, y, i, tbl = NULL, model = NULL, ..., k_max = 5000L, tol = 1e-12)
}
\arguments{
\item{x}{Age of the first life. May be scalar or vector.}

\item{y}{Age of the second life. May be scalar or vector.}

\item{n}{Term in years. May be scalar or vector.}

\item{i}{Effective annual interest rate. May be scalar or vector.}

\item{tbl}{Optional life table object.}

\item{model}{Optional parametric survival model.}

\item{...}{Additional parameters passed to the survival model or
life-table functions.}

\item{k_max}{Maximum number of terms used for an infinite series.}

\item{tol}{Numerical tolerance used to assess convergence.}
}
\value{
A numeric vector of actuarial present values.
}
\description{
Computes temporary and whole-life annuities for two lives under the
last-survivor status.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter15_variable_interest.R
\name{forward_matrix_from_z}
\alias{forward_matrix_from_z}
\title{Matrix of forward rates implied by spot rates}
\usage{
forward_matrix_from_z(z)
}
\arguments{
\item{z}{Numeric vector of at least two annual effective spot rates. Each
value must be greater than \code{-1}.}
}
\value{
A numeric matrix.
}
\description{
Constructs a matrix of annual effective forward rates. Rows correspond to
forward-start times \eqn{n=1,\ldots,m-1}; columns correspond to forward
periods \eqn{k=1,\ldots,m-1}. Entries requiring maturities beyond
\eqn{m} are returned as \code{NA}.
}
\examples{
forward_matrix_from_z(c(0.03, 0.04, 0.05, 0.06, 0.07))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12.R
\name{tqxy}
\alias{tqxy}
\title{Joint-life failure probability}
\usage{
tqxy(x, y, t, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{t}{Time.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{{}_tq_{xy} = 1 - {}_tp_{xy}}.
}
\examples{
tqxy(40, 50, t = 10, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chapter18_pensions.R
\name{PAB_cae}
\alias{PAB_cae}
\title{Projected annual benefit under a career-average-earnings plan}
\usage{
PAB_cae(x, z, CASx, p, past_salary_total = 0, g = NULL, s = NULL)
}
\arguments{
\item{x}{Scalar current or entry age.}

\item{z}{Scalar retirement age.}

\item{CASx}{Positive scalar current annual salary.}

\item{p}{Nonnegative scalar accrual percentage.}

\item{past_salary_total}{Nonnegative total of actual prior salaries.}

\item{g}{Optional scalar annual salary growth rate greater than \code{-1}.}

\item{s}{Optional positive salary-scale vector of length \code{z - x}.}
}
\value{
A numeric scalar.
}
\description{
Projects the annual benefit using a career-average-earnings formula.
}
\examples{
PAB_cae(
  x = 30,
  z = 65,
  CASx = 100000,
  p = 1,
  g = 0.04
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserve_ch10.R
\name{GT_disc}
\alias{GT_disc}
\title{Total gain for a discrete insurance contract}
\usage{
GT_disc(Vt, Vt1, P, i_actual, q_actual, B = 1)
}
\arguments{
\item{Vt}{Reserve at duration t.}

\item{Vt1}{Reserve at duration t+1.}

\item{P}{Net premium for the year.}

\item{i_actual}{Actual annual effective interest rate.}

\item{q_actual}{Actual mortality rate for the year.}

\item{B}{Benefit amount. Defaults to 1.}
}
\value{
Numeric vector.
}
\description{
Computes the Chapter 10 total gain:
amount on hand at year-end minus amount required.
}
\examples{
GT_disc(Vt = 0.1, Vt1 = 0.11, P = 0.02, i_actual = 0.05, q_actual = 0.01)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{Abarx_udd}
\alias{Abarx_udd}
\title{UDD approximation of continuous whole life insurance}
\usage{
Abarx_udd(Ax, i)
}
\arguments{
\item{Ax}{Discrete whole life insurance APV.}

\item{i}{Effective annual interest rate.}
}
\value{
Continuous whole life insurance APV under UDD.
}
\description{
Computes \eqn{\bar{A}_x = (i/\delta)A_x}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_continuous.R
\name{A2nAbarx}
\alias{A2nAbarx}
\title{Second moment of continuous deferred insurance PV}
\usage{
A2nAbarx(x, n, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Deferral period.}

\item{i}{Effective annual interest rate.}

\item{model}{Parametric survival model name.}

\item{...}{Additional model parameters passed to survival-model functions.}
}
\value{
Numeric vector of second moments.
}
\description{
Computes \eqn{{}^{2}{}_{n\mid}\bar{A}_x} by evaluating
\eqn{{}_{n\mid}\bar{A}_x} at doubled force.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/life_table_fractional.R
\name{tqx_tab}
\alias{tqx_tab}
\title{Fractional failure probability from a life table}
\usage{
tqx_tab(tbl, x, t, assumption = c("udd", "cf", "balducci"))
}
\arguments{
\item{tbl}{A life_table object.}

\item{x}{Numeric vector of integer ages.}

\item{t}{Numeric vector of fractional durations with \eqn{0 \le t \le 1}.}

\item{assumption}{One of \code{"udd"}, \code{"cf"}, \code{"balducci"}.}
}
\value{
Numeric vector of \eqn{{}_t q_x} values.
}
\description{
Computes \eqn{{}_t q_x = 1 - {}_t p_x} for \eqn{0 \le t \le 1}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/numerical_utils.R
\name{.safe_integrate}
\alias{.safe_integrate}
\title{Internal: safe numerical integration helper}
\usage{
.safe_integrate(f, lower, upper, ..., rel.tol = 1e-10)
}
\description{
Internal: safe numerical integration helper
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mortality_improvement.R
\name{axn_improved}
\alias{axn_improved}
\title{Temporary annuity-immediate under mortality improvement}
\usage{
axn_improved(x0, n, i, qx_base_vec, AAx_vec, base_year, issue_year)
}
\arguments{
\item{x0}{Issue age.}

\item{n}{Term in years.}

\item{i}{Effective annual interest rate.}

\item{qx_base_vec}{Base-year one-year death probabilities for ages
\code{x0, x0+1, ..., x0+n-1}.}

\item{AAx_vec}{Mortality improvement factors for ages
\code{x0, x0+1, ..., x0+n-1}.}

\item{base_year}{Base year.}

\item{issue_year}{Issue year.}
}
\value{
Numeric scalar.
}
\description{
Computes
\deqn{a_{x:\overline{n}|} = \sum_{t=1}^n v^t \, {}_tp_x^{(\mathrm{improved})}}
using projected mortality rates.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insurance_utils.R
\name{Abarxn1_udd}
\alias{Abarxn1_udd}
\title{UDD approximation of continuous term insurance}
\usage{
Abarxn1_udd(Axn1, i)
}
\arguments{
\item{Axn1}{Discrete term insurance APV.}

\item{i}{Effective annual interest rate.}
}
\value{
Continuous term insurance APV under UDD.
}
\description{
Computes \eqn{\bar{A}_{x:\overline{n}|}^{1} = (i/\delta)A_{x:\overline{n}|}^{1}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multilife_ch12_ext.R
\name{tqxy2}
\alias{tqxy2}
\title{Probability that (x) fails after (y) within n years}
\usage{
tqxy2(x, y, n, tbl = NULL, model = NULL, ...)
}
\arguments{
\item{x}{Age of first life.}

\item{y}{Age of second life.}

\item{n}{Term in years.}

\item{tbl}{Life table.}

\item{model}{Survival model.}

\item{...}{Additional model parameters.}
}
\value{
Numeric vector.
}
\description{
Computes \eqn{{}_n q_{xy}^{2} = {}_n q_x - {}_n q_{xy}^{1}}.
}
\examples{
tqxy2(40, 50, n = 10, model = "uniform", omega = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annuity_mthly.R
\name{annuity_mthly_temp_due}
\alias{annuity_mthly_temp_due}
\alias{adotxn_m}
\title{Temporary m-thly annuity-due}
\usage{
adotxn_m(x, n, m, i, model, ...)
}
\arguments{
\item{x}{Age.}

\item{n}{Term in years.}

\item{m}{Number of payments per year.}

\item{i}{Effective annual interest rate.}

\item{model}{Survival model name.}

\item{...}{Additional parameters passed to the survival model.}
}
\value{
Numeric vector of annuity values.
}
\description{
Computes the exact temporary m-thly annuity-due.
}
\details{
\deqn{\ddot{a}_{x:\overline{n}|}^{(m)} = \frac{1}{m}\sum_{t=0}^{mn-1} v^{t/m}\,{}_{t/m}p_x}
}
\examples{
adotxn_m(40, n = 10, m = 12, i = 0.05, model = "uniform", omega = 100)
}
